:- module(bliss_interface, [init_bliss_interface/0,
                            init_directed_graph/0,
                            init_undirected_graph/0,
                            add_node/2,
                            add_node/3,
                            add_edge/2,
                            find_automorphisms/1,
                            get_canonical_form/1,
                            graph_to_dot_file/1]).

:- use_module(library(random), [random/3]).

%% find_automorphisms(-PermGroup).
% Find graph automorphisms yielding a permutation group,
% each permutation in the group is given by a generator which is represented in cycle notation.

%% get_canonical_form(-CanonicalForm).
% Fet the canonical labeling of the graph in cycle notation.

foreign_resource('bliss_interface', [init_directed_graph_wrapper,
                                     init_undirected_graph_wrapper,
                                     add_node_wrapper,
                                     add_edge_wrapper,
                                     find_automorphisms,
                                     get_canonical_form]).

foreign(init_directed_graph_wrapper, c, init_directed_graph_wrapper).
foreign(init_undirected_graph_wrapper, c, init_undirected_graph_wrapper).
foreign(add_node_wrapper, c, add_node_wrapper(+integer, [-integer])).
foreign(add_edge_wrapper, c, add_edge_wrapper(+integer, +integer)).
foreign(find_automorphisms, c, find_automorphisms([-term])).
foreign(get_canonical_form, c, get_canonical_form([-term])).

:- dynamic is_initialised/0, amount_of_nodes/1, dot_edge_atom/1, is_directed_graph/0, node_color/2, node_label/2, color_code/2.
:- volatile is_initialised/0, amount_of_nodes/1, dot_edge_atom/1, is_directed_graph/0, node_color/2, node_label/2, color_code/2.

%% init_bliss_interface.
init_bliss_interface :-
    \+ is_initialised,
    reset_state,
    !,
    Error = '~n! Could not load ~w library!~n! Check that it is available in the lib folder of ProB.~n~n',
    catch(load_foreign_resource(library(bliss_interface)),_,
          (format(user_error, Error, [library(bliss_interface)]), fail)),
    asserta(is_initialised).
init_bliss_interface.

%% init_directed_graph.
% Clear graph and start creating a directed graph.
init_directed_graph :-
    reset_state,
    asserta(is_directed_graph),
    init_directed_graph_wrapper.

%% init_undirected_graph.
% Clear graph and start creating an undirected graph.
init_undirected_graph :-
    reset_state,
    init_undirected_graph_wrapper.

%% add_node(+Color, -Id).
% Create a new node with the given colour and return the id.
add_node(Color, Id) :-
    inc_nodes,
    add_node_wrapper(Color, Id),
    asserta(node_color(Id,Color)).

%% add_node(+Label, +Color, -Id).
% Label is used for dot graph visualization.
add_node(Label, Color, Id) :-
    add_node(Color, Id),
    asserta(node_label(Id,Label)).

%% add_edge(+Id1, +Id2).
% Create an edge between two ids.
add_edge(Id1, Id2) :-
    amount_of_nodes(Nodes),
    Id1 >= 0,
    Id2 >= 0,
    Id1 < Nodes,
    Id2 < Nodes,
    add_dot_edge_atom(Id1, Id2),
    add_edge_wrapper(Id1, Id2).

%% graph_to_dot_file(+FilePath).
graph_to_dot_file(FilePath) :-
    open(FilePath, write, S),
    dot_file_header(S),
    dot_node_attributes(S),
    dot_file_edges(S),
    write(S, '}'),
    close(S).

inc_nodes :-
    retract(amount_of_nodes(Old)),
    New is Old + 1,
    asserta(amount_of_nodes(New)).

add_dot_edge_atom(Id1, Id2) :-
    number_to_atom(Id1, AId1),
    number_to_atom(Id2, AId2),
    atom_concat(AId1, ' -> ', T1),
    atom_concat(T1, AId2, T2),
    atom_concat(T2, ';', DotEdgeAtom),
    (   dot_edge_atom(DotEdgeAtom)
    ->  true
    ;   asserta(dot_edge_atom(DotEdgeAtom))
    ).

reset_state :-
    retractall(is_directed_graph),
    retractall(amount_of_nodes(_)),
    retractall(node_color(_,_)),
    retractall(node_label(_,_)),
    retractall(dot_edge_atom(_)),
    asserta(amount_of_nodes(0)).

dot_node_attributes(S) :-
    findall((Id,Color), node_color(Id, Color), NodeColors),
    retractall(color_code(_,_)),
    dot_node_attributes(NodeColors, S).

dot_node_attributes([], _).
dot_node_attributes([(Id,Color)|T], S) :-
    get_color_code_for_id(Id, Color, ColorCode),
    number_to_atom(Id, AId),
    atom_concat(AId, ' [style="filled" color="', T1),
    atom_concat(T1, ColorCode, T2),
    (   node_label(Id, Label)
    ->  atom_concat(T2, '" label="', T3),
        atom_concat(T3, Label, T4),
        atom_concat(T4, '"]', DotNodeColor)
    ;   atom_concat(T2, '"]', DotNodeColor)
    ),
    write(S, DotNodeColor),
    nl(S),
    dot_node_attributes(T, S).

get_color_code_for_id(Id, Color, ColorCode) :-
    node_color(Id, Color),
    (   color_code(Color, TColorCode)
    ->  ColorCode = TColorCode
    ;   random(100000, 999999, RGB),
        number_to_atom(RGB, ARGB),
        atom_concat('#', ARGB, ColorCode),
        asserta(color_code(Color, ColorCode))
    ).

number_to_atom(Number, Atom) :-
    number_codes(Number, C),
    atom_codes(Atom, C).

dot_file_edges(S) :-
    findall(DotEdgeAtom, dot_edge_atom(DotEdgeAtom), DotEdgeAtoms),
    dot_file_edges(DotEdgeAtoms, S).

dot_file_edges([], _).
dot_file_edges([DotEdgeAtom|T], S) :-
    write(S, DotEdgeAtom),
    nl(S),
    dot_file_edges(T, S).

dot_file_header(S) :-
    is_directed_graph,
    !,
    write(S, 'digraph bliss_graph {'),
    nl(S).
dot_file_header(S) :-
    write(S, 'graph bliss_graph {'),
    nl(S).
