#include <stdint.h>

#include "ltlc.h"
#include "loop_fairness.h"
#include "fair_counterexample.h"

/*
 * This module contains code that takes a counter-example in the form of an SCC (and a path to it)
 * and generates a sequence of atoms that form a loop and can serve as a user readable 
 * counter-example.
 */

typedef enum {item_with_all_matched_untils_in_loop,
              item_with_more_goals_in_loop,
              no_similiar_item_in_loop} in_loop_result;

// transition_id should be 0 if no particular transition id is specified
modelitem create_model_item(ltlatom *atom, unsigned int transition_id) {
    modelitem item;
    item.atom = atom;
    // the transition id is sometimes needed to know by which node transition the
    // atom is reached. especially when a fair loop should be found one has to
    // make sure that the relevant fairness transition are executed along the loop
    item.transition_id = transition_id;
    return item;
}

/* returns the bitfield that represents the goals encountered
 * in the atom of the corresponding item */
bitfield* until_goals_of_loopitem(loopitem *item) {
  return (bitfield*) (((uintptr_t) item) + sizeof(loopitem));
}

/* returns the bitfield that represents the used untils (untils which are evaluated to true)
 * in the atom of the corresponding item */
bitfield* used_untils_of_loopitem(ltl_formula *f, loopitem *item) {
  return (bitfield*) (((uintptr_t) item) + sizeof(loopitem) + get_bitfield_memsize(f->number_of_untils));
}




static ltlatom** rewind_path(ltlarray *bf_items, unsigned int solution,int *resultlength) {
    int i,size = 0;
    sccitem *item;
    unsigned int index = solution;
    ltlatom ** result;
    /* determine the size of the path */
    do {
        size++;
        item = (sccitem*) get_array_element(bf_items, index);
        index = item->previous;
    } while(index != 0);

    *resultlength = size;
    result = (ltlatom**) ltl_malloc(sizeof(ltlatom*) * (*resultlength));

    item = (sccitem*) get_array_element(bf_items, solution);
    index = item->previous;
    /* copy the path into the SCC to the beginning of the counter-example */
    for (i=0; i<size; i++) {
        item = (sccitem*) get_array_element(bf_items, index);
        result[size-i-1] = item->atom;
        index = item->previous;
    }
    return result;
}


#ifdef LTLLOOP

void print_bitfield(bitfield *field, int size) {
    int i;
    printf("[");
    for (i=0; i<size; i++) {
        if (i>0) {
            printf(",");
        }
        if (getbit(field,i))
            printf("1");
        else
            printf("0");
    }
    printf("]");
}

void print_loopitem(ltl_formula *f, loopitem *item, char const *s) {
    ltlatom *atom = item->atom;
    int nodeid = (atom->node)->id.id;
    int atomid = atom->tarjan_number;
    unsigned int transition_id = item->transition_id;
    bitfield *goalbuffer, *usedbuffer;
    printf("(%s) loopitem (%p) { previous: %d, nodeid: %d, atomid: % d, transition_id: %d, number_of_encounters: %d, ",s,item,item->previous,nodeid,atomid,transition_id, item->number_of_encounters);
    
    goalbuffer = until_goals_of_loopitem(item);
    printf("goalbuffer: ");
    print_bitfield(goalbuffer, f->number_of_untils);
    printf(", usedbuffer: ");
    usedbuffer = used_untils_of_loopitem(f,item);
    print_bitfield(usedbuffer, f->number_of_untils);
    printf("}\n");
}

static void debug_print_true_untils(ltl_formula *f, bitfield *untils, const char *text) {
  unsigned int i;
  printf(" (%s:", text);
  for (i = 0; i<f->number_of_untils; i++) {
    if (getbit(untils, i)) {
      printf("%u",i);
    }
  }
  printf(" )");
}


static void debug_print_solution_atoms(ltl_formula *f, ltlarray *bf_nodes, int start_item, int first, int include_first) {
  unsigned int itemnr = start_item, atomid;
  int first_item_encountered = 0;
  loopitem *item;
  while (!first_item_encountered) {
    first_item_encountered = (itemnr == first);
    if (include_first || !first_item_encountered) {
      item = (loopitem *) get_array_element(bf_nodes, itemnr);
      atomid = find_atom_id(f,item->atom);
      printf(" %u (node id: %d)",atomid, (item->atom)->node->id.id);
      debug_print_true_untils(f, used_untils_of_loopitem(f, item), "untils");
      debug_print_true_untils(f, until_goals_of_loopitem(item), "goals");
    }
    itemnr = item->previous;
  }
}

void print_out_buffer(ltl_formula *f, ltlarray *bf_nodes,unsigned int bf_size) {
    unsigned int index = bf_size -1;
    loopitem* item;
    printf("Buffer array:\n");
    while (index > 0) {
        item = (loopitem*) get_array_element(bf_nodes,index);
        print_loopitem(f, item, "buffer array");
        index--;
    }
    printf("*****************************************\n");
}

static void debug_print_loop(ltl_formula *f, ltlarray *bf_nodes, int start_item, int circle_start) {
  ltldeb_inc();
  printf("Found a loop where all untils are matched.\n");
  printf("  Atoms in loop (reverse order): ");
  debug_print_solution_atoms(f,bf_nodes,start_item,circle_start,1);
  printf("\n  Atoms in path to loop (but still in SCC, reverse order): ");
  debug_print_solution_atoms(f,bf_nodes,circle_start,0,1);
  printf("\n");
  ltldeb_dec();
}

#endif




static int scc_item_in_array(ltlatom*atom, ltlarray *bf_nodes, unsigned int size) {
    unsigned int i;
    sccitem *item;
    for (i=0; i<size; i++) {
        item = (sccitem*) get_array_element(bf_nodes,i);
        if (item->atom == atom) {
            return 1;
        }
    }
    return 0;
}

static sccitem *save_sccitem(ltlarray *bf_nodes, unsigned int pos,
                             unsigned int previous_item_pos, ltlatom *atom) {
    
    assure_arraysize(bf_nodes, pos+1);
    sccitem *nextitem = (sccitem*) get_array_element(bf_nodes, pos);
    if (nextitem != NULL) {
        nextitem->atom = atom;
        nextitem->previous = previous_item_pos;
    } else {
        fprintf(stderr, "save_sccitem(): nextitem is NULL.\n");
    }
    return nextitem;
}

/*
 get path through the main SCC from one node to another
 */
ltlatom** get_path_to_scc(ltl_formula *f, ltlatom** main_scc, int main_scc_size, ltlatom *entry_atom, ltlatom **scc, int scc_size, ltlatom* entry_atom_scc, int *length) {
    ltlatom *atom, *nextatom;
    unsigned int bf_index, bf_size, nextitem_index;
    unsigned int transition_id;
    sccitem *item, *nextitem;
    int i;
    ltlarray bf_nodes;
    ltlatom **path;
    
    const size_t sccitem_size = sizeof(sccitem);
    init_array(&bf_nodes, sccitem_size, NULL, "bf_nodes");
    
    /* put first atom into the buffer of the breadth-first search */
    bf_size = 1;
    bf_index = 0;

    item = save_sccitem(&bf_nodes, bf_index, 0, entry_atom);
    
    if (entry_atom != entry_atom_scc) {
        while (bf_index < bf_size) {
            item = (sccitem*) get_array_element(&bf_nodes, bf_index);
            atom = item->atom;
            for (i=0; i<atom->number_of_transitions; i++) {
                transition_id = atom->first_transition_id + i;
                nextatom = get_atom_trans(f, transition_id);
                
                if (is_in_scc(nextatom, main_scc_size, main_scc) && !scc_item_in_array(nextatom,&bf_nodes,bf_size)) {
                    nextitem_index = bf_size;
                    nextitem = save_sccitem(&bf_nodes, nextitem_index, bf_index, nextatom);
                    //ltl_printnode_with_info("node get_path (node): ",nextatom->node);
                    if (is_in_scc(nextitem->atom,scc_size,scc) && entry_atom_scc == nextitem->atom ) {
                        path = rewind_path(&bf_nodes, nextitem_index, length);
                        delete_array(&bf_nodes);
                        return path;
                    } else {
                        bf_size++;
                    }
                }
            }
            bf_index++;
        }
    }
    *length = 0;
    path = 0;
    delete_array(&bf_nodes);
    return path;
}



int check_fairness(found_scc_type *found) {
    return (found->fairness_assumptions.weak_size>0
            || found->fairness_assumptions.strong_size>0);
}

void get_path_through_scc(ltl_formula *f, found_scc_type *found, modelitem **result, int *resultlength, int *is_deadlock, int *loop_length) {
  IFLTLDEBUG( ltl_print_prolog_graph(f,found,0,0); );
  if (found==0) {
    *result = 0;
    *resultlength = 0;
    *is_deadlock = 0;
  } else {
    *is_deadlock = (found->number_of_scc_atoms == 1
                    && found->scc[0]->node->number_of_transitions == 0);
    if (*is_deadlock) {
      /* SCC consisting of one deadlock state */
      *result = (modelitem*) ltl_malloc(sizeof(modelitem) * (found->number_of_path_elements + 1));
      int i;
      for (i=0; i<found->number_of_path_elements; i++) (*result)[i] = create_model_item(found->path[i],0);
      (*result)[found->number_of_path_elements] = create_model_item(found->scc[0],0);
      *resultlength = found->number_of_path_elements + 1;
    } else {
        if (check_fairness(found)) {
            get_fair_loop(f,found,result,resultlength,loop_length);
        } else {
            // get_loop2 is extremely inefficient, need to be improved
            // (consider removing or replacing with a more efficient implementation)
            //get_loop2(f,found,result,resultlength,loop_length);
            get_fair_loop(f,found,result,resultlength,loop_length);
        }
    }
  }
  IFLTLDEBUG(
    if (*result) {
      debug_print_trace(f,*result,*resultlength,*is_deadlock);
    });
}

