% (c) 2009-2020 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(testltlc,[testltlc/0]).

user:file_search_path(library,'.').

:- use_module(ltlc).
:- use_module(library(lists)).

testltlc :-
    findall(N,testcase(N,_,_,_,_),Names),
    testall(Names, Verdict),
    (Verdict=passed -> print('all test passed\n');
        print('**** TEST FAILED ****\n')).

/* the callback */
:- public testcallback/5.
testcallback(Node, Aps, Out, APProps, TPProps) :-
    node(Node, Nodestate),
    ( Nodestate=fail -> fail
    ; Nodestate=exception(E) -> raise_exception(E)
    ; otherwise -> evalprops(APProps, Nodestate, Aps),
        findall(trans(Next,Op,Tp),
                (trans(Node,Op,Next,P),evalprops(TPProps,P,Tp)),
                Out)).

:- public update/5, fairness/2.
update(States,Atoms,Trans,VAtoms,VTrans) :-
    print(States/Atoms/Trans/VAtoms/VTrans),nl.
fairness(Arg,SCC) :-
    print('--- fairness called.\n'),
    print(Arg),nl,print(SCC),nl.

evalprops([],_,[]).
evalprops([P|PRest],State,[R|RRest]) :-
    (member(P,State) -> R=1; R=0),!,
    evalprops(PRest,State,RRest).

test(Name, Verdict) :-
    ltlc_init,
    format('~nRunning Test ~w:~n',[Name]),
    testcase(Name,Formula,Inits,Exp,SccCheck),
    on_exception(Exception,
                 c_ltl_modelcheck(Formula, Inits, Res,
                                  testltlc:testcallback, % Callback to explore state space
                                  none,                  % Callback for periodic updates
                                  none,                  % Callback transition ids
                                  none,                  % Callback Enabled Actions
                                  %SccCheck,              % Callback for valid SCCs  % TO DO: update to new format
                                  %Name,                  % Data for SCC check callback
                                  [],[]),                % Weak and Strong Fairness args
  % c_ltl_modelcheck(Formula, Startnodes, Result, Callback, Update, GetTransIdsPred, GetEnabledActionsPred, WeakFairnessArgs, StrongFairnessArgs)
                 Res=exception(Exception)),
    ( Exp=check(Call) ->
        ( call(Call,Res) -> Verdict=passed; Verdict=failed )
    ; Res=Exp   -> Verdict=passed
    ; otherwise -> Verdict=failed),
    format('Test verdict for ~w: ~w~n',[Name,Verdict]),
    (Verdict=passed -> true;
        print('\n  Result was: '),print(Res),
        print('\n  expected: '),print(Exp)),
    nl.

testall([],passed).
testall([T|Rest],Verdict) :-
    test(T,V),
    testall(Rest,RestV),
    combineverdict(V,RestV,Verdict).

combineverdict(passed,passed,passed) :- !.
combineverdict(_,_,failed).


%*******************************************************************************
% Test cases
:- discontiguous node/2, trans/4, testcase/4, testcase/5.

% by default, a testcase has no callback defined for checking the SCC
testcase(Name,Formula,Inits,Exp,none) :-
    testcase(Name,Formula,Inits,Exp).

% graph
node(root,[1,0]).
node(1,   [1,0]).
node(2,   [1,0]).
node(3,   [0,1]).
trans(root, tr1, 1,    [0,0]).
trans(1,    t13, 3,    [0,0]).
trans(1,    t12, 2,    [0,0]).
trans(2,    t2r, root, [0,0]).
testcase(no_yesterday_in_start,not(yesterday(true)),[root],model(_,_)).

node(10, []).
node(11, []).
node(12, []).
node(13, []).
node(14, [start]).
node(15, []).
node(16, [end]).
trans(10,t01,11,[]).
trans(11,t12,12,[]).
trans(12,t20,10,[]).
trans(12,t23,13,[]).
trans(13,t34,14,[]).
trans(14,t45,15,[]).
trans(14,t46,16,[]).
trans(15,t53,13,[]).
trans(16,t63,13,[]).
testcase(simple_until,
         finally(and(ap(start),finally(ap(end)))),
         [10],
         model(_,_)).
testcase(trivial_false_globally,
         not(globally(ap(never))),
         [10],
         model([atom(10,_,none)],no_loop)).

node(20, []).
node(21, fail).
node(22, [target]).
trans(20,a,21,[]).
trans(20,b,22,[]).
testcase(partial_counter,
         finally(ap(target)),
         [20],
         model(_,_)).
testcase(partial_incomplete,not(F),Inits,incomplete) :-
    testcase(partial_counter,F,Inits,_).

node(40, []).
node(41, exception(testexception)).
trans(40,a,41,[]).
testcase(exception,F,[40],exception(testexception)) :-
    testcase(partial_counter,F,_,_).

% 0 -> 1 -> 2 (final)
%          -> 3 (deadlock)
%          -> 4 <-> 5

node(50, []).
node(51, []).
node(52, [final]).
node(53, []).
node(54, []).
node(55, []).
trans(50,t01,51,[]).
trans(51,t12,52,[]).
trans(51,t13,53,[]).
trans(51,t14,54,[]).
trans(52,t22,52,[]).
trans(54,t45,55,[]).
trans(55,t54,54,[]).
% test if loop is found correctly
testcase(finally_loop,not(finally(or(ap(final),not(next(true))))),[50],
         model([atom(50,_,t01),atom(51,_,t14),atom(54,_,t45),atom(55,_,t54)],loop(2))).

% 0 -> 1 -> 2 (final)
%          -> 3 (deadlock)
%
node(60, []).
node(61, []).
node(62, [final]).
node(63, []).
trans(60,t01,61,[]).
trans(61,t12,62,[]).
trans(61,t13,63,[]).
% test if deadlock state is found correctly
testcase(finally,not(finally(ap(final))),[60],
         model([atom(60,_,t01),atom(61,_,t13),atom(63,_,none)],deadlock)).


node(70, [a_enabled]).
node(71, [a_enabled]).
node(72, []).
trans(70,t71,71,[is_a]).
trans(70,t72,72,[]).
trans(71,t73,71,[is_a]).
trans(71,t74,72,[]).
trans(72,t75,71,[is_c]).
testcase(ticket_3187a,
         not(implies(implies(globally(finally(ap(a_enabled))),
                             globally(finally(action(is_a)))),
                     finally(action(is_c)))),
         [70],
         model(_,_)).
testcase(ticket_3187b,Formula,[71], model(_,_)) :-
    testcase(ticket_3187a,Formula,_,_).

node(80, [a_enabled]).
trans(80,t81,80,[is_a]).
testcase(ticket_3187_simplified,
         not(implies(implies(globally(finally(ap(a_enabled))),
                             globally(finally(action(is_a)))),
                     finally(action(is_c)))),
         [80],
         model(_,_)).
testcase(ticket_3187_even_more_simplified,
         globally(finally(action(is_a))),
         [80],
         model(_,_)).

node(90,[]).
node(91,[]).
node(92,[]).
node(93,[]).
node(94,[]).
node(95,[]).
trans(90,t90,91,[]).
trans(91,t91,93,[start]).
trans(92,t92,93,[]).
trans(93,t93,94,[]).
trans(94,t94,95,[]).
trans(95,t95,94,[]).
trans(95,t96,92,[]).
testcase(loopsize,
         not(globally(implies(action(start),next(finally(ap(goal)))))),
         [90],
         model([atom(90,_,t90),
                atom(91,_,t91),
                atom(93,_,t93),
                atom(94,_,t94),
                atom(95,_,t95)],
               loop(3))).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% checking the SCC check callback
node(100,[]).
node(101,[]).
node(102,[]).
node(103,[]).
node(104,[]).
node(105,[goal]).
trans(100,t100, 101,[]).
trans(101,t101, 102,[]).
trans(102,t102a,101,[]).
trans(102,t102b,103,[]).
trans(103,t103, 104,[]).
trans(104,t104a,103,[]).
trans(104,t104b,105,[]).

:- public test_scc/2.
test_scc(check_scc_callback_firstloop,SCC) :-
    % check if this is the first loop
    length(SCC,2),
    memberchk(anode(_A101,101,_),SCC),
    memberchk(anode(_A102,102,_),SCC).
test_scc(check_scc_callback_secondloop,SCC) :-
    % check if this is the second loop
    length(SCC,2),
    memberchk(anode(_A103,103,_),SCC),
    memberchk(anode(_A104,104,_),SCC).
test_scc(check_scc_callback_exception,_SCC) :-
    throw(scc_exception).

testcase(check_scc_callback_firstloop,
         not(finally(ap(goal))),
         [100],
         model([atom(100,_,t100),
                atom(101,_,t101),
                atom(102,_,t102a)],
               loop(1)),
         testltlc:test_scc).
testcase(check_scc_callback_secondloop,
         not(finally(ap(goal))),
         [100],
         model([atom(100,_,t100),
                atom(101,_,t101),
                atom(102,_,t102b),
                atom(103,_,t103),
                atom(104,_,t104a)],
               loop(3)),
         testltlc:test_scc).
testcase(check_scc_callback_exception,
         not(finally(ap(goal))),
         [100],
         exception(scc_exception),
         testltlc:test_scc).

% user:runtime_entry(start) :- testltlc:test(finally_loop,_).
user:runtime_entry(start) :- testltlc:testltlc.

:- public show/1.
show(model(Trace,loop(Entry))) :-
    show2(Trace,0,Entry).
show2([],_,_).
show2([Atom|Arest],Index,Entry) :-
    show_loop(Index,Entry),
    show_atom(Atom),
    I2 is Index+1,
    show2(Arest,I2,Entry).
show_loop(I,E) :- I<E,!,write('      ').
show_loop(I,E) :- I=E,!,write(' /--> ').
show_loop(I,E) :- I>E,!,write(' |    ').

show_atom(atom(Node,_,NextOp)) :-
    write(Node),write('  '),
    ( trans(Node,NextOp,_NextNode,TP) -> write(TP)
    ; otherwise -> write('no op for '),write(NextOp)),
    nl.
