#include "z3++.h"
#include <sicstus/sicstus.h>
#include <unordered_map>
#include <vector>
#include <algorithm>
#include <string>

using namespace z3;
using namespace std;

static context* c;
static solver* smtt;
static model* stored_model;

vector<expr> prolog_list_to_expr_vector(SP_term_ref list);
vector<z3::sort> prolog_type_list_to_sort_vector(SP_term_ref list);
string prolog_term_to_string(SP_term_ref term);

z3::sort mk_type(SP_term_ref type);
SP_integer insert_id(expr x);
SP_integer insert_id(func_decl x);

extern "C" {
    static unordered_map<string, z3::sort> *type_table;

    static unordered_map<SP_integer, expr> *id_table;
    static unordered_map<string, unordered_map<string, Z3_func_decl>> *couple_proj_decls_table;
    static unordered_map<string, string> *inverse_couple_type_table;
    static unordered_map<SP_integer, const func_decl> *func_table;
    static SP_integer id_counter;

    SP_term_ref term_ref_from_string(string str) {
        SP_term_ref return_value = SP_new_term_ref();
        int ret = SP_cons_functor(return_value, SP_atom_from_string(str.c_str()), 0);
        assert(ret);
        return return_value;
    }

    void init(void) {
        // Z3_open_log("z3_prob.log");
        // comment in line above to generate a log file which can be replayed with same version of Z3
        c = new context();
        smtt = new solver(*c);
        stored_model = new model(*c);

        id_counter = 2;

        id_table = new unordered_map<SP_integer, expr>();
        id_table->insert(make_pair<SP_integer,expr>(0, c->bool_val(false)));
        id_table->insert(make_pair<SP_integer,expr>(1, c->bool_val(true)));

        couple_proj_decls_table = new unordered_map<string, unordered_map<string, Z3_func_decl>>();
        inverse_couple_type_table = new unordered_map<string, string>();

        // types
        type_table = new unordered_map<string, z3::sort>();
        type_table->insert(pair<string,z3::sort>("real", c->real_sort()));
        type_table->insert(pair<string,z3::sort>("integer", c->int_sort()));
        type_table->insert(pair<string,z3::sort>("boolean", c->bool_sort()));

        func_table = new unordered_map<SP_integer, const func_decl>();
    }

    void pretty_print_smt1(void) {
            cout << "; Z3 Solver Store (SMT format):" << endl;
            cout << *smtt << endl;
            // some constraints seem to be missing in this output !
            //cout << "; assertions:" << endl;
            //cout << smtt->assertions() << endl;
    }

    void pretty_print_smt2(void) {
            cout << "; Z3 Solver Store (SMT2 format):" << endl;
            cout << smtt->to_smt2() << endl;
            // does not show constraints added to id_table nor sort declarations
            cout << "; Context\n" << c << endl;
    }

    void raise_Z3_exception(z3::exception& e, std::string name) {
            cout << "z3_exception in " << name << " : " << e.msg() << endl;
            SP_term_ref eterm = SP_new_term_ref();
            SP_term_ref arg = SP_new_term_ref();

            int ret = SP_put_atom(arg,SP_atom_from_string(e.msg()));
            assert(ret);
            ret = SP_cons_functor(eterm, SP_atom_from_string("z3_exception"),1, arg);
            assert(ret);

            SP_raise_exception(eterm);// From SICStus manual: The exception will be stored as pending.
            // This function is normally used in the context of a call from Prolog to C,
            // and will cause the exception to be propagated to Prolog on return from the call.
            return;
    }
    void pretty_print_smt2_for_id(SP_integer spe) {
        try {
            expr x = id_table->at(spe);
            smtt->push();
            smtt->add(x);
            pretty_print_smt2();
            smtt->pop();
        } catch(z3::exception& e) {
            raise_Z3_exception(e, "pretty_print_smt2_for_id");
            return; //assert(false);
        }
    }


    SP_integer mk_var(SP_term_ref type, char *varname) {
        z3::sort t = mk_type(type);
        expr x = c->constant(varname, t);
        return insert_id(x);
    }

    SP_integer mk_bounded_var(SP_term_ref type, char *varname) {
        return mk_var(type,varname);
    }

    SP_integer mk_empty_set(SP_term_ref type) {
        // z3 api: mk_empty_set gets the domains / subtypes sort!
        SP_term_ref arg = SP_new_term_ref();
        int ret = SP_get_arg(1,type,arg);
        assert(ret);
        z3::sort t = mk_type(arg);
        expr emptyset = to_expr(*c,Z3_mk_empty_set(*c,t));
        return insert_id(emptyset);
    }

    SP_integer mk_full_set(SP_term_ref type) {
        try {
            // z3 api: mk_full_set gets the domains / subtypes sort!
            SP_term_ref arg = SP_new_term_ref();
            int ret = SP_get_arg(1,type,arg);
            assert(ret);
            z3::sort t = mk_type(arg);
            expr fullset = to_expr(*c,Z3_mk_full_set(*c,t));
            return insert_id(fullset);
        } catch(z3::exception& e) {
            raise_Z3_exception(e, "mk_full_set");
            return 0; //assert(false);
        }
    }

    SP_integer mk_int_const(int64_t val) {
        try {
            expr x = c->int_val(val);
            return insert_id(x);
        }
        catch(z3::exception& e) {
            raise_Z3_exception(e, "mk_int_const");
            return 0; //assert(false);
        }
    }

    SP_integer mk_real_const(char *val_in) {
        try {
            expr x = c->real_val(val_in);
            return insert_id(x);
        }
        catch(z3::exception& e) {
            raise_Z3_exception(e, "mk_real_const");
            return 0;
        }
    }

    SP_integer mk_bool_const(char *val) {
        if (0 == strcmp(val, "boolean_true")) {
            return 1;
        } else if (0 == strcmp(val, "boolean_false")) {
            return 0;
        } else {
            // explode
            assert(0);
        }
    }

    SP_integer mk_record_const(SP_term_ref type, SP_term_ref fields) {
        z3::sort record_type = mk_type(type);
        Z3_func_decl constructor = Z3_get_tuple_sort_mk_decl(*c,record_type);

        vector<expr> fields_exprs = prolog_list_to_expr_vector(fields);

        z3::array<Z3_ast> arr(fields_exprs.size());
        for(unsigned int i = 0; i < fields_exprs.size(); i++) {
            arr[i] = fields_exprs[i];
        }

        expr x = to_expr(*c,Z3_mk_app(*c,constructor,arr.size(),arr.ptr()));
        return insert_id(x);
    }

    SP_integer mk_record_field(SP_integer record, SP_integer field_nr, SP_atom field_name) {
        expr record_expr = id_table->at(record);

        func_decl record_select = to_func_decl(*c,Z3_get_tuple_sort_field_decl(*c,record_expr.get_sort(),field_nr));
        expr x = record_select(record_expr);
        return insert_id(x);
    }

    SP_integer mk_couple(SP_term_ref type, SP_integer arg1, SP_integer arg2) {
        try {
            z3::sort sort = mk_type(type);
            // cout << "calling mk_couple with sort '" << sort << "' and arg1 = " << arg1 << " and arg2= " << arg2 << endl;

            expr a = id_table->at(arg1);
            expr b = id_table->at(arg2);
            //cout << "expr a = " << a << endl; cout << "expr b = " << b << endl;

            Z3_func_decl z3d = Z3_get_tuple_sort_mk_decl(*c,sort);
            // cout << "z3 func decl = " << z3d << endl;

            // std::cout << "Z3:" << endl << Z3_solver_to_string(smtt->ctx(), *smtt) << endl;
            // std::cout << *smtt << endl << endl;
            //pretty_print_smt1();
            //pretty_print_smt2();
            func_decl constructor = to_func_decl(*c,z3d);
            return insert_id(constructor(a,b));
        } catch(z3::exception& e) {
            raise_Z3_exception(e, "mk_couple");
            return 0; //assert(false);
        }
    }

    SP_integer mk_op_image(SP_integer arg1, SP_integer arg2, SP_term_ref couple_type_term, SP_term_ref prj1_type_term, SP_term_ref prj2_type_term) {
        // (lambda ((y B)) (exists ((x A)) (and (select set2 x) (select (set1 (|couple(A,B)| x y))))))
        expr fun = id_table->at(arg1);
        expr set = id_table->at(arg2);
        z3::sort fun_sort = fun.get_sort();
        z3::sort set_sort = set.get_sort();
        z3::sort prj1_couple_sort = mk_type(prj1_type_term);
        z3::sort prj2_couple_sort = mk_type(prj2_type_term);
        z3::sort couple_sort = mk_type(couple_type_term);
        expr prj1_couple_cst = c->constant("prj1_couple_cst", prj1_couple_sort);
        expr prj2_couple_cst = c->constant("prj2_couple_cst", prj2_couple_sort);
        Z3_func_decl z3d_couple = Z3_get_tuple_sort_mk_decl(*c, couple_sort);
        func_decl couple_constructor = to_func_decl(*c, z3d_couple);

        expr_vector exists_vars(*c);
        exists_vars.push_back(prj1_couple_cst);
        expr exists_body =
            to_expr(*c, Z3_mk_select(*c, set, prj1_couple_cst)) &&
            to_expr(*c, Z3_mk_select(*c, fun, couple_constructor(prj1_couple_cst, prj2_couple_cst)));
        expr body = to_expr(*c, exists(exists_vars, exists_body));
        Z3_app lambda_vars[] = {(Z3_app) prj2_couple_cst};
        expr lambda = to_expr(*c, Z3_mk_lambda_const(*c, 1, lambda_vars, body));
        return insert_id(lambda);
    }

    SP_integer mk_op(char *opname, SP_integer arg1, SP_integer arg2) {
        try {
            if(id_table->count(arg1) == 0) {
                func_decl a = func_table->at(arg1);
                expr b = id_table->at(arg2);
                if(strcmp(opname, "function") == 0) {
                    return insert_id(a(b));
                }
            } else {
                expr a = id_table->at(arg1);
                expr b = id_table->at(arg2);

                if(strcmp(opname, "modulo") == 0) {
                    return insert_id(to_expr(*c,Z3_mk_rem(*c,a,b)));
                }
                if(strcmp(opname, "add") == 0 || strcmp(opname, "add_real") == 0) {
                    return insert_id(a + b);
                }
                if(strcmp(opname, "minus") == 0 || strcmp(opname, "minus_real") == 0) {
                    return insert_id(a - b);
                }
                if(strcmp(opname, "power_of") == 0 || strcmp(opname, "power_of_real") == 0) {
                    return insert_id(pw(a,b));
                }
                if(strcmp(opname, "div") == 0 || strcmp(opname, "div_real") == 0) {
                    // smtt->add(b > 0);
                    return insert_id(ite(a > 0,a / b,ite(b > 0,(a / b)+1,(a / b)-1)));
                }
                if(strcmp(opname, "multiplication") == 0 || strcmp(opname, "multiplication_real") == 0) {
                    return insert_id(a * b);
                }
                if(strcmp(opname, "equal") == 0) {
                    return insert_id(a == b);
                }
                if(strcmp(opname, "implication") == 0) {
                    return insert_id(implies(a,b));
                }
                if(strcmp(opname, "equivalence") == 0) {
                    return insert_id(a == b);
                }
                if(strcmp(opname, "greater") == 0 || strcmp(opname, "greater_real") == 0) {
                    return insert_id(a > b);
                }
                if(strcmp(opname, "greater_equal") == 0 || strcmp(opname, "greater_equal_real") == 0) {
                    return insert_id(a >= b);
                }
                if(strcmp(opname, "less") == 0 || strcmp(opname, "less_real") == 0) {
                    return insert_id(a < b);
                }
                if(strcmp(opname, "less_equal") == 0 || strcmp(opname, "less_equal_real") == 0) {
                    return insert_id(a <= b);
                }
                if(strcmp(opname, "member") == 0) {
                    return insert_id(to_expr(*c,Z3_mk_set_member(*c,a,b)));
                }
                if(strcmp(opname, "subset") == 0) {
                    return insert_id(to_expr(*c,Z3_mk_set_subset(*c,a,b)));
                }
                if(strcmp(opname, "union") == 0) {
                    Z3_ast uargs[2] = {a,b};
                    return insert_id(to_expr(*c,Z3_mk_set_union(*c,2,uargs)));
                }
                if(strcmp(opname, "intersection") == 0) {
                    Z3_ast uargs[2] = {a,b};
                    return insert_id(to_expr(*c,Z3_mk_set_intersect(*c,2,uargs)));
                }
                if(strcmp(opname, "set_subtraction") == 0) {
                    return insert_id(to_expr(*c,Z3_mk_set_difference(*c,a,b)));
                }
            }
        } catch(z3::exception& e) {
            raise_Z3_exception(e, "mk_op");
            return 0;
        }

        cout << "Operator missing in mk_op: " << opname << endl;
        assert(false); // operator missing
    }

    SP_integer mk_quantifier(char *opname, SP_term_ref list, SP_integer ex) {
        vector<expr> args = prolog_list_to_expr_vector(list);
        expr inner = id_table->at(ex);
        try {
            expr_vector vars(*c);
            for(unsigned int i = 0; i< args.size(); i++) {
                vars.push_back(args[i]);
            }

            if(strcmp(opname, "forall") == 0) {
                expr r = forall(vars, inner);
                return insert_id(r);
            }
            if(strcmp(opname, "exists") == 0) {
                expr r = exists(vars, inner);
                return insert_id(r);
            }
        } catch(z3::exception& e) {
            raise_Z3_exception(e, "mk_quantifier");
            return 0;
        }
        cout << "Operator missing in mk_quantifier: " << opname << endl;
        assert(false);
    }

    string b_couple_type_to_z3(string b_couple_type) {
        return "|" + b_couple_type + "|";
    }

    /**
    * Given a list of arguments of a B comprehension set and corresponding couple types.
    * For instance, arguments ((i A) (j B) (k C)) and types [couple(A,B),couple(couple(A,B),C))].
    * Reduce the list to a single couple, e.g., (|couple(couple(A,B),C)| (|couple(A,B)| i j) k).
    */
    expr reduce_comp_set_args_to_couple(vector<expr> &arg_ids, vector<z3::sort> &couple_type_sorts) {
        func_decl first_couple_constructor = to_func_decl(*c, Z3_get_tuple_sort_mk_decl(*c, couple_type_sorts.at(0)));
        expr x = first_couple_constructor(arg_ids.at(0), arg_ids.at(1));
        assert(arg_ids.size() - 1 == couple_type_sorts.size());
        for (unsigned int i=2; i<arg_ids.size(); i++) {
            Z3_func_decl z3d_couple = Z3_get_tuple_sort_mk_decl(*c, couple_type_sorts.at(i-1));
            func_decl couple_constructor = to_func_decl(*c, z3d_couple);
            x = couple_constructor(x, arg_ids.at(i));
        }
        return x;
    }

    SP_integer mk_op_comprehension_set_multi(SP_term_ref arg_id_list, SP_term_ref couple_types, SP_integer body_id) {
        // (lambda ((c T)) (exists (args) body && (= c (couple args))))
        // we have to create B couples for comprehension sets with multiple arguments
        vector<expr> arg_ids = prolog_list_to_expr_vector(arg_id_list);
        vector<z3::sort> couple_type_sorts = prolog_type_list_to_sort_vector(couple_types);
        expr return_couple = reduce_comp_set_args_to_couple(arg_ids, couple_type_sorts);
        expr return_couple_cst = c->constant("couple_cst", return_couple.get_sort());
        expr body = id_table->at(body_id);
        expr_vector exists_vars(*c);
        for (unsigned int i=0; i<arg_ids.size(); i++) {
            exists_vars.push_back(arg_ids.at(i));
        }
        expr exists_expr = to_expr(*c, exists(exists_vars, body && return_couple_cst == return_couple));
        Z3_app lambda_vars[1] = {(Z3_app) return_couple_cst};
        expr lambda = to_expr(*c, Z3_mk_lambda_const(*c, 1, lambda_vars, exists_expr));
        return insert_id(lambda);
    }

    SP_integer mk_op_comprehension_set_singleton(SP_integer arg_id, SP_integer body_id) {
        // (lambda ((arg T)) body)
        expr arg = id_table->at(arg_id);
        expr body = id_table->at(body_id);
        Z3_app lambda_vars[1] = {(Z3_app) arg};
        expr lambda = to_expr(*c, Z3_mk_lambda_const(*c, 1, lambda_vars, body));
        return insert_id(lambda);
    }

    SP_integer mk_op_domain(SP_integer arg_id, SP_term_ref couple_type_term, SP_term_ref prj1_type_term, SP_term_ref prj2_type_term) {
        // (lambda ((x A)) (exists ((y B)) (select set1 (|couple(A,B)| x y))))
        expr arg = id_table->at(arg_id);

        z3::sort prj1_couple_sort = mk_type(prj1_type_term);
        expr prj1_couple_cst = c->constant("prj1_couple_cst", prj1_couple_sort);
        z3::sort prj2_couple_sort = mk_type(prj2_type_term);
        expr prj2_couple_cst = c->constant("prj2_couple_cst", prj2_couple_sort);
        z3::sort couple_sort = mk_type(couple_type_term);
        Z3_func_decl z3d_couple = Z3_get_tuple_sort_mk_decl(*c, couple_sort);
        func_decl couple_constructor = to_func_decl(*c, z3d_couple);

        expr_vector exists_vars(*c);
        exists_vars.push_back(prj2_couple_cst);
        expr body = to_expr(*c, exists(exists_vars, to_expr(*c, Z3_mk_select(*c, arg, couple_constructor(prj1_couple_cst, prj2_couple_cst)))));
        Z3_app lambda_vars[] = {(Z3_app) prj1_couple_cst};
        expr lambda = to_expr(*c, Z3_mk_lambda_const(*c, 1, lambda_vars, body));
        return insert_id(lambda);
    }

    SP_integer mk_op_range(SP_integer arg_id, SP_term_ref couple_type_term, SP_term_ref prj1_type_term, SP_term_ref prj2_type_term) {
        // (lambda ((y B)) (exists ((x A)) (select set1 (|couple(A,B)| x y))))
        expr arg = id_table->at(arg_id);

        z3::sort prj1_couple_sort = mk_type(prj1_type_term);
        expr prj1_couple_cst = c->constant("prj1_couple_cst", prj1_couple_sort);
        z3::sort prj2_couple_sort = mk_type(prj2_type_term);
        expr prj2_couple_cst = c->constant("prj2_couple_cst", prj2_couple_sort);
        z3::sort couple_sort = mk_type(couple_type_term);
        Z3_func_decl z3d_couple = Z3_get_tuple_sort_mk_decl(*c, couple_sort);
        func_decl couple_constructor = to_func_decl(*c, z3d_couple);

        expr_vector exists_vars(*c);
        exists_vars.push_back(prj1_couple_cst);
        expr body = to_expr(*c, exists(exists_vars, to_expr(*c, Z3_mk_select(*c, arg, couple_constructor(prj1_couple_cst, prj2_couple_cst)))));
        Z3_app lambda_vars[] = {(Z3_app) prj2_couple_cst};
        expr lambda = to_expr(*c, Z3_mk_lambda_const(*c, 1, lambda_vars, body));
        return insert_id(lambda);
    }

    SP_integer mk_op_reverse(SP_integer arg_id, SP_term_ref couple_type_term) {
        // (lambda ((xx |couple(B,A)|)) (select ff (|couple(A,B)| (get_y xx) (get_x xx))))
        string inv_couple_type = b_couple_type_to_z3(prolog_term_to_string(couple_type_term));
        string couple_type = inverse_couple_type_table->at(inv_couple_type);
        expr arg = id_table->at(arg_id);

        z3::sort couple_sort = type_table->at(couple_type);
        z3::sort inv_couple_sort = type_table->at(inv_couple_type);
        Z3_func_decl z3d = Z3_get_tuple_sort_mk_decl(*c, couple_sort);
        expr inv_couple_cst = c->constant("inv_couple_cst", inv_couple_sort);
        func_decl couple_constructor = to_func_decl(*c, z3d);

        unordered_map<string, Z3_func_decl> inv_couple_proj_decls = couple_proj_decls_table->at(string(inv_couple_type));
        func_decl get_x = to_func_decl(*c, inv_couple_proj_decls["x"]);
        func_decl get_y = to_func_decl(*c, inv_couple_proj_decls["y"]);

        expr body = to_expr(*c, Z3_mk_select(*c, arg, couple_constructor(get_y(inv_couple_cst), get_x(inv_couple_cst))));
        Z3_app vars[] = {(Z3_app) inv_couple_cst};
        expr lambda = to_expr(*c, Z3_mk_lambda_const(*c, 1, vars, body));
        return insert_id(lambda);
    }

    SP_integer mk_op_arglist(char *opname, SP_term_ref list) {
        vector<expr> args = prolog_list_to_expr_vector(list);
        try {
            if(strcmp(opname, "negation") == 0) {
                return insert_id(!args.at(0));
            }
            if(strcmp(opname, "unary_minus") == 0 || strcmp(opname, "unary_minus_real") == 0) {
                return insert_id(-args.at(0));
            }
            if(strcmp(opname, "if_then_else") == 0) {
                return insert_id(ite(args.at(0),args.at(1),args.at(2)));
            }
            if(strcmp(opname, "conjunct") == 0) {
                expr x = expr(args.at(0));
                for (unsigned int i = 1; i < args.size(); i++) {
                    x = x && args.at(i);
                }
                return insert_id(x);
            }
            if(strcmp(opname, "disjunct") == 0) {
                expr x = expr(args.at(0));
                for (unsigned int i = 1; i < args.size(); i++) {
                    x = x || args.at(i);
                }
                return insert_id(x);
            }
            if(strcmp(opname, "distinct") == 0) {
                z3::array<Z3_ast> arr(args.size());
                for(unsigned int i = 0; i< args.size(); i++) {
                    arr[i] = args[i];
                }

                return insert_id(to_expr(*c,Z3_mk_distinct(*c, args.size(), arr.ptr())));
            }
        }
        catch(z3::exception& e) {
            raise_Z3_exception(e, "mk_op_arglist");
            return 0;
        }
        cout << "Operator missing in mk_op_arglist: " << opname << endl;
        assert(false); // operator missing
    }

    SP_term_ref add_interpreter_constraint(SP_integer spe, SP_atom name, SP_integer timeout) {
        SP_term_ref return_value = SP_new_term_ref();
        try{
            // receives a new constraint with a given name,
            // adds it to the smt solver,
            // checks for a contradiction
            // returns unsat(ListOfNamesInUnsatCore), sat or unknown

            expr x = id_table->at(spe);
            // cout << "expr x is " << x << endl;
            smtt->add(x,SP_string_from_atom(name));

            params p(*c);
            p.set("timeout", (unsigned int)timeout);
            p.set("unsat-core", true);
             //p.set("mbqi", true); // default is true anyway ?
             //p.set("pull_nested_quantifiers",true);
            smtt->set(p);

            check_result r = smtt->check();
            if(r == check_result::sat) {
                return term_ref_from_string("sat");
            } else if (r == check_result::unknown) {
                if(smtt->reason_unknown().compare("canceled") == 0) {
                    return term_ref_from_string("time_out");
                }
                return term_ref_from_string("unknown");
            } else if(r == check_result::unsat) {
                SP_term_ref list = SP_new_term_ref();
                expr_vector core = smtt->unsat_core();
                for (unsigned i = 0; i < core.size(); i++) {
                    SP_term_ref head = SP_new_term_ref();
                    std::ostringstream out;
                    out << core[i];
                    int ret = SP_put_atom(head, SP_atom_from_string(out.str().c_str()));
                    assert(ret);
                    ret = SP_cons_list(list,head,list);
                    assert(ret);
                }
                int ret = SP_cons_functor(return_value, SP_atom_from_string("unsat"), 1, list);
                assert(ret);
                return return_value;
            }
        }
        catch(z3::exception& e) {
            cout << "z3_exception in add_interpreter_constraint: " << e.msg() << endl;

            if(strcmp(e.msg(),"canceled") == 0) { // occurs on timeout
              return term_ref_from_string("time_out");
            } else {
                SP_term_ref eterm = SP_new_term_ref();
                SP_term_ref arg = SP_new_term_ref();

                int ret = SP_put_atom(arg,SP_atom_from_string(e.msg()));
                assert(ret);
                ret = SP_cons_functor(eterm, SP_atom_from_string("z3_exception"),1, arg);
                assert(ret);

                SP_raise_exception(eterm);
                return term_ref_from_string("no_atom_here"); // make compiler happy
            }
        }
        assert(false);
    }

    SP_atom smt_solve_query(SP_atom use_inc_solver_atom, SP_integer spe, SP_integer timeout) {
        bool use_inc_solver = strcmp(SP_string_from_atom(use_inc_solver_atom), "true") == 0;
        try {
            expr x = id_table->at(spe);
            // cout << "expr x is " << x << endl;

            params p(*c);
            z3::set_param("pp.decimal", true);
            z3::set_param("pp.decimal-precision", 20);

            p.set("timeout", (unsigned int)timeout);
            p.set("mbqi", true); // added by leuschel <--------- Model Based Quantifier Instantiation, default true
            p.set("pull_nested_quantifiers",true); // get other options with z3 -p
            //cout << "setting mbqi and pull_nested_quantifiers" << endl;
            smtt->set(p);

            if (use_inc_solver) smtt->push();
            smtt->add(x);
            // cout << "calling smtt->check()" << x << endl;
            // pretty_print_smt2();
            check_result r = smtt->check();
            // cout << "smtt->check() done" << x << endl;
            if(r == check_result::sat) {
                // cout << "Z3 found model: " << *stored_model << endl;
                stored_model = new model(*c, smtt->get_model()); // pop removes a model
                if (use_inc_solver) smtt->pop();
                return SP_atom_from_string("sat");
            } else if(r == check_result::unsat) {
                if (use_inc_solver) smtt->pop();
                return SP_atom_from_string("unsat");
            } else if(r == check_result::unknown && smtt->reason_unknown().compare("cancelled") == 0) {
                if (use_inc_solver) smtt->pop();
                return SP_atom_from_string("time_out");
            }
            return SP_atom_from_string("unknown");
        } catch(z3::exception& e) {
            cout << "z3_exception in query: " << e.msg() << endl;

            if(strcmp(e.msg(),"canceled") == 0) { // occurs on timeout
              return SP_atom_from_string("time_out");
            } else {
                SP_term_ref eterm = SP_new_term_ref();
                SP_term_ref arg = SP_new_term_ref();

                int ret = SP_put_atom(arg,SP_atom_from_string(e.msg()));
                assert(ret);
                ret = SP_cons_functor(eterm, SP_atom_from_string("z3_exception"),1, arg);
                assert(ret);
                SP_raise_exception(eterm);
                return SP_atom_from_string("no_atom_here"); // exception will be raised upon return
            }
        }
        assert(false);
    }

    void print_functs(model m) {
        for(unsigned int i=0;i<m.num_funcs();i++) {
            func_decl decl = m.get_func_decl(i);
            func_interp interp = m.get_func_interp(decl);

            cout << "found interp for: " << decl.name() << endl;
            for(unsigned int j=0;j<interp.num_entries();j++) {
                func_entry entry = interp.entry(j);
                cout << "entry: " << entry.arg(0) << " -> " << entry.value() << endl;
            }
            cout << "default: " << interp.else_value() << endl;
        }
    }

    SP_atom get_model_string(SP_integer spexpr) {
        try {
            expr x = id_table->at(spexpr);
            // cout << "for expr:" << x << endl;
            expr model = stored_model->eval(x,true);
            // true means model_completion, i.e., Z3 will bind values of don't cares with arbitrary values
            // true required for for :z3 x:INTEGER & (x>y & y>z) => x>=z; see SSIMT-5

            //print_functs(*stored_model);

            std::ostringstream out;
            out << model;
            return SP_atom_from_string(out.str().c_str());
        } catch(z3::exception& e) {
            raise_Z3_exception(e, "get_model_string");
            return SP_atom_from_string("no_atom_here");
        }
    }

    SP_atom get_header_version(void) {
          //  return SP_atom_from_string("4.8.9");
           #ifdef Z3VERS
              return SP_atom_from_string(Z3VERS);
           #else
              return SP_atom_from_string("unknown");
           #endif
    }

    SP_atom get_full_version(void) {
        unsigned major, minor, build_number, revision_number;
        Z3_get_version(&major, &minor, &build_number, &revision_number);
        string Vers = std::to_string(major) + "." + std::to_string(minor) + "." + std::to_string(build_number)
                      + "."+std::to_string(revision_number);
        // cout << "Z3 Version " << Vers;
        return SP_atom_from_string(Vers.c_str());
    }

    SP_atom get_full_model_string(void) {
        try {
            std::ostringstream out;
            out << *stored_model << endl;
            return SP_atom_from_string(out.str().c_str());
        } catch(z3::exception& e) {
            raise_Z3_exception(e, "get_full_model_string");
            return SP_atom_from_string("no_atom_here"); // exception will be raised upon return
        }
    }

    SP_integer mk_set(SP_term_ref list) {
        try {
            vector<expr> exprs = prolog_list_to_expr_vector(list);

            z3::sort dom_type = exprs[0].get_sort();
            expr theset = to_expr(*c,Z3_mk_empty_set(*c,dom_type));

            for(unsigned int i=0; i < exprs.size(); i++) {
                theset = to_expr(*c,Z3_mk_set_add(*c,theset,exprs[i]));
            }
            return insert_id(theset);
        }
        catch(z3::exception& e) {
            raise_Z3_exception(e, "mk_set");
            return 0; //assert(false);
        }
    }

    void mk_sort(SP_atom name) {
        const char* setname = SP_string_from_atom(name);
        // cout << "mk_sort with name " << setname << endl;
        if(type_table->count(setname) == 0) {
            z3::sort type = to_sort(*c,Z3_mk_uninterpreted_sort(*c,Z3_mk_string_symbol(*c, setname)));
            type_table->insert(pair<string,z3::sort>(string(setname),type));
        }
    }

    void mk_sort_with_cardinality(SP_atom name, SP_integer cardinality) {
        const char* setname = SP_string_from_atom(name);
        // cout << "new sort with cardinality " << cardinality << " named " << setname << endl;
        if(type_table->count(setname) == 0) {
            // cout << "new sort added to table" << endl;
            z3::sort type = to_sort(*c,Z3_mk_finite_domain_sort(*c,Z3_mk_string_symbol(*c, setname),cardinality));

            type_table->insert(pair<string,z3::sort>(string(setname),type));
        }
    }

    void push_frame(void) {
        try {
            smtt->push();
        } catch(z3::exception& e) {
            raise_Z3_exception(e, "push_frame");
            return; //assert(false);
        }
    }

    void pop_frame(void) {
        try {
            smtt->pop();
        } catch(z3::exception& e) {
            raise_Z3_exception(e, "pop_frame");
            return; //assert(false);
        }
    }

    void reset(void) {
        // TODO: does not reset the context!
        //smtt->reset();
        //id_table->clear();
        //func_table->clear();
        //id_counter = 2;
        init();
    }

    void setup_enumerated_set(SP_term_ref list) {
        try {
            // the set identifiers are distinct. the cardinality is already defined
            vector<expr> ids = prolog_list_to_expr_vector(list);
            z3::array<Z3_ast> arr(ids.size());
            for(unsigned int i = 0; i < ids.size(); i++) {
                arr[i] = ids[i];
            }

            // can only be distinct if more than one given
            if(ids.size() > 1) {
                expr distinct = to_expr(*c,Z3_mk_distinct(*c, ids.size(), arr.ptr()));
                smtt->add(distinct);
            }

            // each is enforced to be equal to a corresponding uninterpreted constant
            // if this is a finite domain sort (i.e. was finite all along, not shrunken by prob)
            z3::sort the_sort = ids.begin()->get_sort();
            if(the_sort.is_finite_domain()) {
                int num = 0;
                for (vector<expr>::iterator it = ids.begin(); it != ids.end(); ++it) {
                    expr uc = c->num_val(num,the_sort);
                    expr equal = *it == uc;
                    smtt->add(equal);
                    num++;
                }
            }
        } catch(z3::exception& e) {
            raise_Z3_exception(e, "setup_enumerated_set");
        }
    }

    SP_integer conjoin_negated_state(SP_term_ref list, SP_integer old_expr) {
        try {
            vector<expr> exprs = prolog_list_to_expr_vector(list);
            vector<expr> negated_exprs;

            for (vector<expr>::iterator it = exprs.begin(); it != exprs.end(); ++it) {
                expr not_equal = *it != stored_model->eval(*it);
                negated_exprs.push_back(not_equal);
            }

            if(negated_exprs.size() == 1) {
                expr new_expr = negated_exprs.front() && id_table->at(old_expr);
                return insert_id(new_expr);
            } else {
                expr negstate = expr(negated_exprs.at(0));
                for (unsigned int i = 1; i < negated_exprs.size(); i++) {
                    negstate = negstate || negated_exprs.at(i);
                }
                expr new_expr = negstate && id_table->at(old_expr);
                return insert_id(new_expr);
            }
        } catch(z3::exception& e) {
            raise_Z3_exception(e, "conjoin_negated_state");
            return 0; //assert(false);
        }
    }
}

string get_couple_proj_type_name(string couple_type) {
    // couple(A,B) -> couple_A_B for projection functions, e.g., get_x_couple_A_B
    string couple_proj_type = couple_type;
    std::replace(couple_proj_type.begin(), couple_proj_type.end(), '(', '_');
    std::replace(couple_proj_type.begin(), couple_proj_type.end(), ',', '_');
    couple_proj_type.erase(std::remove(couple_proj_type.begin(), couple_proj_type.end(), '|'), couple_proj_type.end());
    couple_proj_type.erase(std::remove(couple_proj_type.begin(), couple_proj_type.end(), ')'), couple_proj_type.end());
    return couple_proj_type;
}

z3::sort create_couple(SP_term_ref type, SP_term_ref ret, string couple_type) {
    SP_term_ref arg1 = SP_new_term_ref();
    ret = SP_get_arg(1,type,arg1);
    assert(ret);
    SP_term_ref arg2 = SP_new_term_ref();
    ret = SP_get_arg(2,type,arg2);
    assert(ret);

    Z3_func_decl mk_tuple_decl;
    Z3_func_decl proj_decls[2];
    Z3_sort proj_sorts[2];
    Z3_symbol couple_proj_names[2];
    string sarg1 = prolog_term_to_string(arg1);
    string sarg2 = prolog_term_to_string(arg2);
    z3::sort type_arg1 = mk_type(arg1);
    z3::sort type_arg2 = mk_type(arg2);
    proj_sorts[0] = type_arg1;
    proj_sorts[1] = type_arg2;
    // new couple for specific type
    Z3_symbol mk_couple_name = Z3_mk_string_symbol(*c, couple_type.c_str());
    string couple_type_proj_name = get_couple_proj_type_name(couple_type);
    couple_proj_names[0] = Z3_mk_string_symbol(*c, string("get_x_").append(couple_type_proj_name).c_str());
    couple_proj_names[1] = Z3_mk_string_symbol(*c, string("get_y_").append(couple_type_proj_name).c_str());
    Z3_sort pair_sort = Z3_mk_tuple_sort(*c, mk_couple_name, 2, couple_proj_names, proj_sorts, &mk_tuple_decl, proj_decls);
    // store names of type specific projection functions (e.g., get_xcouple_integer_integer
    unordered_map<string, Z3_func_decl> couple_proj_decls_map;
    couple_proj_decls_map["x"] = proj_decls[0];
    couple_proj_decls_map["y"] = proj_decls[1];
    couple_proj_decls_table->insert(pair<string, unordered_map<string, Z3_func_decl>>(string(couple_type), couple_proj_decls_map));
    type_table->insert(make_pair<string,z3::sort>(string(couple_type), to_sort(*c, pair_sort)));
    // new couple for inverse couple type
    std::ostringstream inverse_couple_type;
    inverse_couple_type << "|couple(" << sarg2 << "," << sarg1 << ")|";
    if (inverse_couple_type.str() != string(couple_type)) {
        Z3_func_decl mk_inv_tuple_decl;
        Z3_func_decl inv_proj_decls[2];
        Z3_symbol inv_couple_proj_names[2];
        Z3_sort inv_proj_sorts[2];
        inv_proj_sorts[0] = type_arg2;
        inv_proj_sorts[1] = type_arg1;
        Z3_symbol mk_inv_couple_name = Z3_mk_string_symbol(*c, inverse_couple_type.str().c_str());
        string inv_couple_type_proj_name = get_couple_proj_type_name(inverse_couple_type.str());
        inv_couple_proj_names[0] = Z3_mk_string_symbol(*c, string("get_x_").append(inv_couple_type_proj_name).c_str());
        inv_couple_proj_names[1] = Z3_mk_string_symbol(*c, string("get_y_").append(inv_couple_type_proj_name).c_str());
        Z3_sort inv_pair_sort = Z3_mk_tuple_sort(*c, mk_inv_couple_name, 2, inv_couple_proj_names, inv_proj_sorts, &mk_inv_tuple_decl, inv_proj_decls);
        unordered_map<string, Z3_func_decl> inv_couple_proj_decls_map;
        inv_couple_proj_decls_map["x"] = inv_proj_decls[0];
        inv_couple_proj_decls_map["y"] = inv_proj_decls[1];
        couple_proj_decls_table->insert(pair<string, unordered_map<string, Z3_func_decl>>(inverse_couple_type.str(), inv_couple_proj_decls_map));
        type_table->insert(make_pair<string,z3::sort>(inverse_couple_type.str(), to_sort(*c, inv_pair_sort)));
        inverse_couple_type_table->insert(make_pair<string,string>(inverse_couple_type.str(), string(couple_type)));
    }
    // store inverse couple types for fast access when translating operators such as reverse
    inverse_couple_type_table->insert(make_pair<string,string>(string(couple_type), inverse_couple_type.str()));
    return to_sort(*c, pair_sort);
}

z3::sort create_record(SP_term_ref type, SP_term_ref ret) {
    // first argument is a list of field(Name,Type)
    SP_term_ref list_of_fields = SP_new_term_ref();
    ret = SP_get_arg(1,type,list_of_fields);
    assert(ret);

    // iterate over the list and place fields in vector
    vector<pair<Z3_symbol,z3::sort>> fields;
    SP_term_ref head = SP_new_term_ref();
    SP_term_ref tail = SP_new_term_ref();
    while (SP_get_list(list_of_fields, head, tail)) {
        SP_term_ref name = SP_new_term_ref();
        ret = SP_get_arg(1,head,name);
        assert(ret);
        ret = SP_get_arg(2,head,type);
        assert(ret);

        z3::sort field_type = mk_type(type);

        SP_atom as_atom;
        ret = SP_get_atom(name,&as_atom);
        assert(ret);
        const char *namestr = SP_string_from_atom(as_atom);
        Z3_symbol field_name = Z3_mk_string_symbol(*c,namestr);

        fields.push_back(make_pair(field_name,field_type));
        list_of_fields = tail;
    }

    Z3_symbol const_name = Z3_mk_string_symbol(*c, "record");
    Z3_func_decl record_constructor;

    z3::array<Z3_func_decl> projections(fields.size());

    z3::array<Z3_sort> proj_sorts(fields.size());
    z3::array<Z3_symbol> proj_names(fields.size());
    for(unsigned int i = 0; i < fields.size(); i++) {
        proj_names[i] = fields[i].first;
        proj_sorts[i] = fields[i].second;
    }

    Z3_sort record_sort = Z3_mk_tuple_sort(*c, const_name, fields.size(), proj_names.ptr(), proj_sorts.ptr(), &record_constructor, projections.ptr());

    return to_sort(*c,record_sort);
}

z3::sort mk_type(SP_term_ref type) {
    int ret;
    // simple atom types 'integer' and 'boolean'
    if(SP_is_atom(type)) {
        SP_atom as_atom;
        ret = SP_get_atom(type,&as_atom);
        assert(ret);
        const char *typestr = SP_string_from_atom(as_atom);
        z3::sort to_return = type_table->at(string(typestr));
        if(type_table->count(typestr) == 0) {
            SP_printf("Type missing in table: %s\n",typestr);
        } else {
            return to_return;
        }
    }
    // types that have been constructed before
    string type_as_string = prolog_term_to_string(type);
    string couple_type = b_couple_type_to_z3(type_as_string);
    if(type_table->count(type_as_string) > 0) {
        return type_table->at(string(type_as_string));
    } else if(type_table->count(couple_type) > 0) {
        return type_table->at(string(couple_type));
    }
    // otherwise deconstruct type
    SP_atom name;
    int arity;
    ret = SP_get_functor(type,&name,&arity);
    assert(ret);

    // sets
    const char *atom = SP_string_from_atom(name);
    if(strcmp(atom, "set") == 0) {
        SP_term_ref arg = SP_new_term_ref();
        ret = SP_get_arg(1,type,arg);
        assert(ret);
        z3::sort subtype = mk_type(arg);
        type_table->insert(make_pair<string,z3::sort>(string(type_as_string),to_sort(*c,Z3_mk_set_sort(*c,subtype))));
        // cout << "created type for " << type_as_string << endl;
        return to_sort(*c,Z3_mk_set_sort(*c,subtype));
    }
    if(strcmp(atom, "global") == 0) {
        SP_term_ref arg = SP_new_term_ref();
        ret = SP_get_arg(1,type,arg);
        assert(ret);
        SP_atom as_atom;
        ret = SP_get_atom(arg,&as_atom);
        assert(ret);
        const char *typestr = SP_string_from_atom(as_atom);
        return type_table->at(string(typestr));
    }
    if(strcmp(atom, "couple") == 0) {
        return create_couple(type, ret, couple_type);
    }
    if(strcmp(atom, "record") == 0) {
        return create_record(type, ret);
    }

    // TODO: further types?
    printf("Unsupported type: %s\n",atom);
    assert(false);
}

SP_integer insert_id(expr x) {
    SP_integer id = id_counter;
    id_table->insert(pair<SP_integer, expr>( id,  x));
    id_counter++;

    //cout << id << ": " << x << endl;
    return id;
}

SP_integer insert_id(func_decl x) {
    SP_integer id = id_counter;
    func_table->insert(pair<SP_integer, func_decl>( id,  x));
    id_counter++;

    //cout << id << ": " << x << endl;
    return id;
}

vector<expr> prolog_list_to_expr_vector(SP_term_ref list) {
    assert(SP_is_list(list));

    vector<expr> args;
    SP_term_ref head = SP_new_term_ref();
    SP_term_ref tail = SP_new_term_ref();
    while (SP_get_list(list, head, tail)) {
        assert(SP_is_integer(head));
        SP_integer id;
        SP_get_integer(head, &id);

        expr x = id_table->at(id);
        args.push_back(x);
        list = tail;
    }
    return args;
}

vector<z3::sort> prolog_type_list_to_sort_vector(SP_term_ref list) {
    assert(SP_is_list(list));
    vector<z3::sort> args;
    SP_term_ref head = SP_new_term_ref();
    SP_term_ref tail = SP_new_term_ref();
    while (SP_get_list(list, head, tail)) {
        args.push_back(mk_type(head));
        list = tail;
    }
    return args;
}

string prolog_term_to_string(SP_term_ref term) {
    SP_atom functor;
    int arity;
    SP_term_ref arg = SP_new_term_ref();

    int ret = SP_get_functor(term, &functor, &arity);
    assert(ret);
    if(arity == 0) {
        return string(SP_string_from_atom(functor));
    } else {
        string s = SP_string_from_atom(functor);
        s.append("(");
        for(int i=1;i<arity;i++) {
            ret = SP_get_arg(i,term,arg);
            assert(ret);
            s.append(prolog_term_to_string(arg));
            s.append(",");
        }
        ret = SP_get_arg(arity,term,arg);
        assert(ret);
        s.append(prolog_term_to_string(arg));
        s.append(")");
        return s;
    }
}


