#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif
//#include <sys/mman.h>
//#include <sys/stat.h>
#include <sys/ipc.h>
#include <sys/shm.h>
//#include <fcntl.h>
#include <semaphore.h>

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <assert.h>
#include <math.h>
#include <string.h>
#include "shm_common.h"
#include "hash.h"

#include "assert_helper.h"



#ifdef BLOCKS
unsigned int blocks_read = 0;
#endif


enum segment_types {
    DIGESTS,
    NODES,
    META
};

static char *suffix_str(char const * const pre, const int id) {
    char *res;
    int rc = asprintf(&res, "%s%d", pre, id);
    assert_esf(rc != -1, "asprintf returned %d. This should not happen. Pre was %s and id was %d.", rc, pre, id);
    return res;
}

void init_semaphore_names(struct semap_env * const env, const int id) {
    if (env->names_initialised) {
        return;
    }
    env->semap_meta_s      = suffix_str("/hashes-semap_infos", id);
    env->semap_hashes_s    = suffix_str("/hashes-semap_digests", id);
    env->semap_hashwrite_s = suffix_str("/hashes-semap_digestswrite", id);
    int i;
    for (i = 0; i < BRANCH_FACTOR; i++) {
        char *tmp = suffix_str("/hashes-semap_nodes", i);
        env->semap_nodes_s[i] = suffix_str(tmp, id); // this is somewhat ugly but whatever
        free(tmp);

        tmp = suffix_str("/hashes_semap_meta", i);
        env->semap_multiwrite_s[i] = suffix_str(tmp, id);
        free(tmp);
    }
    env->names_initialised = 1;
}

void free_semaphore_names(struct semap_env * const env) {
    if (!env->names_initialised) {
        return;
    }
    free(env->semap_meta_s);
    free(env->semap_hashes_s);
    free(env->semap_hashwrite_s);
    env->semap_meta_s = NULL;
    env->semap_hashes_s = NULL;
    env->semap_hashwrite_s = NULL;

    int i;
    for (i = 0; i < BRANCH_FACTOR; i++) {
        free(env->semap_nodes_s[i]);
        free(env->semap_multiwrite_s[i]);
        env->semap_nodes_s[i] = NULL;
        env->semap_multiwrite_s[i] = NULL;
    }
    env->names_initialised = 0;
}


#include <stdio.h>
static struct shm_block get_shm(struct shm_env * const env, enum segment_types t) {
    key_t key;
    size_t size; 
    switch(t) {
        case META:
            key = env->meta_infos;
            size = sizeof(struct hash_meta_infos);
            break;
        case DIGESTS:
            key = env->next_digests_key++;
            size = env->elements_per_digests_block * HASH_INFO;
            //assert(next_digest_key <= last_meta_information);
            break;
        case NODES:
            key = env->next_nodes_key++;
            size = env->elements_per_nodes_block * sizeof(tNode);
            //assert(next_nodes_key <= NODES_KEY_START + 1000);
            break;
        default:
            assert_es(0, "Requested an invalid shared memory segment type. Aborting.");
    }
    
    int shmid = shmget(key, size, IPC_CREAT | 0666);
    assert_esf(shmid >= 0, "Getting a shared memory segment failed. shmid was %d. Key was %d.\n"
       "Most likely, your operating system does not allow enough shared memory for parB to run properly. Try running:\n"
       "    sudo sysctl -w kern.sysv.shmseg=4096\n"
       "    sudo sysctl -w kern.sysv.shmmax=`perl -e \"print 1073741824*4\"`\n"
       "    sudo sysctl -w kern.sysv.shmall=`perl -e \"print 262144*4\"`\n"
       "If the run did not fail immediately, try increasing the values for kern.sysv.shmmax and kern.sysv.shmall and re-run parB.", shmid, key);

    void *ptr = shmat(shmid, NULL, 0);
    assert_esf(ptr != (void*) -1, "Attaching shared memory segment with id %d failed.", shmid);


    struct shm_block res;
    res.ptr = ptr;
    res.shmid = shmid;

    return res;
}

#include <errno.h>

void shm_cleanup(struct shm_env * const env, int server) {
    int rc = shmdt(env->shm_meta_infos.ptr);
    assert_es(rc == 0, "Detaching a shared memory segment failed. This is not too bad since we were done anyway, but should be looked into.");
    if (server) {
        rc = shmctl(env->shm_meta_infos.shmid, IPC_RMID, NULL);
        assert_es(rc != -1, "Marking a shared memory segment to be destroyed failed. Check your run configuration for multiple proxies with overlapping shared memory regions.");
    }
    
    int i;
    for (i = env->next_digests_key - 1 - env->digests_start; i >= 0; i--) {
        rc = shmdt(env->shm_digests[i].ptr);
        assert_es(rc == 0, "Detaching a shared memory segment failed. This is not too bad since we were done anyway, but should be looked into.");
        if (server) {
            rc = shmctl(env->shm_digests[i].shmid, IPC_RMID, NULL);
            assert_es(rc != -1, "Marking a shared memory segment to be destroyed failed. Check your run configuration for multiple proxies with overlapping shared memory regions.");
        }
    }
      
    for (i = env->next_nodes_key - 1 - env->nodes_start; i >= 0; i--) {
        rc = shmdt(env->shm_nodes[i].ptr);
        assert_es(rc == 0, "Detaching a shared memory segment failed. This is not too bad since we were done anyway, but should be looked into.");
        if (server) {
            rc = shmctl(env->shm_nodes[i].shmid, IPC_RMID, NULL);
            assert_es(rc != -1, "Marking a shared memory segment to be destroyed failed. Check your run configuration for multiple proxies with overlapping shared memory regions.");
        }
    }

    free(env->shm_digests);
    free(env->shm_nodes);

    *env = (struct shm_env) {0};
}


void get_pos_tuple(size_t elements_per_block, int64_t index, size_t *block, size_t *offset) {
    int64_t input = (index >= 0) ? index : -index;
    *block  = input / elements_per_block;
    *offset = input % elements_per_block;
}

struct shm_env *init_shm_common(struct semap_env sem_env,
                                size_t elements_per_nodes_block,
                                size_t elements_per_digests_block,
                                key_t meta_info,
                                key_t digests_start,
                                key_t nodes_start) {
    struct shm_env *res = (struct shm_env*) malloc(sizeof(struct shm_env));
    res->semaphores = sem_env;

    res->elements_per_nodes_block = elements_per_nodes_block;
    res->elements_per_digests_block = elements_per_digests_block;

    res->meta_infos = meta_info;

    res->digests_start = digests_start;
    res->next_digests_key = digests_start;

    res->nodes_start = nodes_start;
    res->next_nodes_key = nodes_start;

    res->shm_meta_infos = get_shm(res, META); // HACK: res is not yet fully initialised here
                                              //       this is really scary and should not be done at all

    // these two arrays will be expanded later
    res->shm_digests_len = 1;
    res->shm_digests = calloc(sizeof(struct shm_block), res->shm_digests_len);

    res->shm_nodes_len = 1;
    res->shm_nodes = calloc(sizeof(struct shm_block), res->shm_nodes_len);

    // initialise blocks
    res->shm_nodes[0] = get_shm(res, NODES);
    res->shm_digests[0] = get_shm(res, DIGESTS);

    return res;
}

inline void lock_digests_read(struct shm_env const * const env) {
    sem_wait(env->semaphores.semap_hashes);
}

inline void release_digests_read(struct shm_env const * const env) {
    sem_post(env->semaphores.semap_hashes);
}


inline static void lock_infos(struct shm_env const * const env) {
    sem_wait(env->semaphores.semap_meta);
}

inline static void release_infos(struct shm_env const * const env) {
    sem_post(env->semaphores.semap_meta);
}

struct hash_meta_infos get_meta_infos(struct shm_env const * const env) {
    struct hash_meta_infos res;
    lock_infos(env);
    res = *((struct hash_meta_infos*) (env->shm_meta_infos.ptr));
    release_infos(env);
    return res;
}

char *get_shm_digests(struct shm_env * const env, size_t amount) {
    // ensure all required shared memory blocks are attached
    (void) lookup_digest(env, amount);

    char *buf = calloc(HASH_INFO, amount);
    size_t pos = 0;


    int i;
    for (i = env->digests_start; i < env->next_digests_key; i++) {
        size_t blocklen = env->elements_per_digests_block;
        if (amount < blocklen) {
            blocklen = amount; // copy only as many digests as needed in last block
        }
        memcpy(buf + pos * HASH_INFO, (env->shm_digests[i - env->digests_start]).ptr, blocklen * HASH_INFO);
        amount = amount - blocklen;
        if (amount <= 0)
            break;
    }
    return buf;
}


tNode *lookup_node(struct shm_env * const env, int64_t index) {
    size_t block, offset;
    get_pos_tuple(env->elements_per_nodes_block, index, &block, &offset);
    
    int i;
    for (i = env->shm_nodes_len; i <= block; i++) { // load necessary blocks
        env->shm_nodes = realloc(env->shm_nodes, (env->shm_nodes_len + 1) * sizeof(struct shm_block));
        env->shm_nodes[env->shm_nodes_len] = get_shm(env, NODES);
        env->shm_nodes_len++;
    }

    return ((tNode*) env->shm_nodes[block].ptr) + offset; // pointer arithmetic does its magic
}

char *lookup_digest(struct shm_env * const env, int64_t index) {
    size_t block, offset;
    get_pos_tuple(env->elements_per_digests_block, index -1, &block, &offset);
    
    int i;
    for (i = env->shm_digests_len; i <= block; i++) { // load necessary blocks
        env->shm_digests = realloc(env->shm_digests, (env->shm_digests_len + 1) * sizeof(struct shm_block));
        env->shm_digests[env->shm_digests_len] = get_shm(env, DIGESTS);
        env->shm_digests_len++;
    }

    return ((char*) env->shm_digests[block].ptr) + (HASH_INFO * offset);
}

inline void lock_nodes_read(struct shm_env const * const env, const int pos) {
#ifdef BLOCKS
    int rc = sem_trywait(env->semaphores.semap_nodes[pos]);
    if (rc == 0)
        return;
    blocks_read++;
#endif
    sem_wait(env->semaphores.semap_nodes[pos]);
}


inline void release_nodes_read(struct shm_env const * const env, const int pos) {
    sem_post(env->semaphores.semap_nodes[pos]);
}

void inc_checked_states(struct shm_env const * const env) {
    lock_infos(env);
    ((struct hash_meta_infos*) (env->shm_meta_infos.ptr))->checked_states++;
    release_infos(env);
}


void update_max_depth(struct shm_env const * const env, int depth) {
    lock_infos(env);
    int max = ((struct hash_meta_infos*) env->shm_meta_infos.ptr)->max_depth;
    if (max < depth)
        ((struct hash_meta_infos*) env->shm_meta_infos.ptr)->max_depth = depth;
    release_infos(env);
}




struct shm_node alloc_node(struct shm_env * const env) {
    lock_infos(env);
    size_t res = ((struct hash_meta_infos*) env->shm_meta_infos.ptr)->pos_nodes++ + PRE_INIT_NODES;
    // allocating nodes right after the pre-allocated nodes should render a seperate
    // initialisation redundant, since the shared memory will be initialised with
    // zeroes anyway
    release_infos(env);

    struct shm_node r;
    r.nodeptr = lookup_node(env, res);
    r.index = - (int64_t) res;

    return r;
}

struct shm_digest alloc_digest(struct shm_env * const env) {
    lock_infos(env);
    size_t res = ((struct hash_meta_infos*) env->shm_meta_infos.ptr)->pos_hashes + 1;
    ((struct hash_meta_infos*) env->shm_meta_infos.ptr)->pos_hashes++;
    release_infos(env);

    struct shm_digest r;
    r.ptr = lookup_digest(env, res);
    r.index = res;

    return r;
}


inline void lock_nodes_write(struct shm_env const * const env, const int pos) {
    sem_wait(env->semaphores.semap_multiwrite[pos]);
    int i;
    for (i = 0; i < SEMAP_NODES_VAL; i++) {
        sem_wait(env->semaphores.semap_nodes[pos]);
    }
    sem_post(env->semaphores.semap_multiwrite[pos]);
}

inline void lock_digests_write(struct shm_env const * const env) {
    sem_wait(env->semaphores.semap_hashwrite);
    int i;
    for (i = 0; i < SEMAP_DIGESTS_VAL; i++) {
        sem_wait(env->semaphores.semap_hashes);
    }
    sem_post(env->semaphores.semap_hashwrite);
}

inline void release_nodes_write(struct shm_env const * const env, const int pos) {
    int i;
    for (i = 0; i < SEMAP_NODES_VAL; i++) {
        sem_post(env->semaphores.semap_nodes[pos]);
    }
}

inline void release_digests_write(struct shm_env const * const env) {
    int i;
    for (i = 0; i < SEMAP_DIGESTS_VAL; i++) {
        sem_post(env->semaphores.semap_hashes);
    }
}

