% (c) 2009-2020 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(avl_ugraphs,[avl_set_to_ugraph/2,
                       ugraph_to_avl_set/2,
                       avl_transitive_closure/2,
                       avl_scc_sets/2]).

%

:- use_module(module_information).
:- module_info(group,external_functions).
:- module_info(description,'A module to interface ProB\'s avl_set relations with the SICStus ugraph library (unweighted graphs).').


:- use_module(library(avl)).
:- use_module(library(lists)).
:- use_module(library(ugraphs),[vertices_edges_to_ugraph/3,
        transitive_closure/2, edges/2,
        reduce/2, vertices/2]).

pair_to_minus((A,B),A-B).
%pair_to_minus((A,B),CA-CB) :- convert_val(A,CA), convert_val(B,CB).
minus_to_pair(A-B,(A,B)-true).
%minus_to_pair(A-B,(CA,CB)-true) :- reconvert_val(A,CA), reconvert_val(B,CB).

%convert_val(int(X),R) :- !, R=X.
%convert_val(R,R).
%reconvert_val(X,R) :- number(X),!, R=int(X).
%reconvert_val(R,R).

% closure1(%x.(x:1..2000|x*x)) is much slower with this than with the version in custom-explicit sets (1 secs vs 10ms)
% closure1(%x.(x:1..1000|x*x) \/ %x.(x:1..2000|x+x) \/ %x.(x:1..1000|x/2)): 9.3 seconds vs 2 seconds
% others like closure1(nxt \/ nxt~) for TrackUp.mch are faster (110ms vs 380ms)

avl_set_to_ugraph(A,UGraph) :-
    avl_domain(A,AList),
    maplist(pair_to_minus,AList,InEdges),
    vertices_edges_to_ugraph([],InEdges,UGraph).

ugraph_to_avl_set(UGraph,Res) :-
    edges(UGraph,Edges),
    maplist(minus_to_pair,Edges,Res).


avl_transitive_closure(A,Res) :-  % print(start1),debug:nl_time,
    avl_set_to_ugraph(A,UGraph),
    transitive_closure(UGraph,TUG), %print(tc(TUG)),debug:nl_time,
    edges(TUG,Edges), %print(edg(Edges)),debug:nl_time,
    maplist(minus_to_pair,Edges,ListRes),
    sort(ListRes,SListRes),  %print(sl(SListRes)),debug:nl_time,
    ord_list_to_avl(SListRes,Res). % could be empty
 %print(stop1),debug:nl_time,fail.

%Note: result could be empty, use: construct_avl_set on Res

:- use_module(custom_explicit_sets,[convert_to_avl/2]).

% compute strongly connected components, returning a set of sets of vertices
avl_scc_sets(A,Res) :-  %print(start_scc),debug:nl_time,
    avl_set_to_ugraph(A,UGraph),
    reduce(UGraph,SCCGraph),
    vertices(SCCGraph,VList), %print(vl(VList)),nl,
    %print(stop_scc),debug:nl_time,
    %maplist(convert_to_avl,V,SCCList),
    convert_to_avl(VList,avl_set(Res)). % could be empty

/* TO DO:

reduce(+Graph, -Reduced)
is true if Reduced is the reduced graph for Graph. The vertices of the reduced
graph are the strongly connected components of Graph. There is an edge in
Reduced from u to v iff there is an edge in Graph from one of the vertices in u
to one of the vertices in v. A strongly connected component is a maximal set
of vertices where each vertex has a path to every other vertex. Algorithm from
"Algorithms" by Sedgewick, page 482, Tarjan’s algorithm.

| ?- vertices_edges_to_ugraph([a,b,c],[a-b,b-c,c-b,b-d],G), reduce(G,G2), edges(G2,Edges).
G = [a-[b],b-[c,d],c-[b],d-[]],
G2 = [[a]-[[b,c]],[b,c]-[[d]],[d]-[]],
Edges = [[a]-[b,c],[b,c]-[d]] ?

reachable(+Vertex, +Graph, -Reachable)
is given a Graph and a Vertex of that Graph, and returns the set of vertices
that are Reachable from that Vertex. Takes O(N^2) time.

random_ugraph(+P, +N, -Graph)
where P is a probability, unifies Graph with a random graph of N vertices where
each possible edge is included with probability P.

*/
