% (c) 2009-2020 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(b_machine_hierarchy,[analyse_hierarchy/2
                              ,analyse_eventb_hierarchy/2
                              ,main_machine_name/1    % find out name of main machine
                              ,machine_name/1         % can be used to find out machine names
                              ,machine_type/2         % machine_type(Name, R) - get the type
                                                      % (abstract_machine, abstract_model, refinement, implementation) for the machine named Name
                              ,machine_references/2   % get a list of references and the type of the reference.
                                                      % Example: if A refines B, then machine_references('A',X) returns X= [ref(refines'B','')].
                                                      % The last argument in ref/3 is the prefix
                              ,machine_identifiers/7  % gets Params,Sets,Abstract Variables,Concrete Variables,Abstract Constants and Concrete
                                                      %  Constants for a machine. Abstract constants/Variables are introduced in the machine
                                                      % using the ABSTRACT_VARAIABLES/ABSTARCT_CONSTANTS keyword. The concrete versions
                                                      % analogously via CONCRETE_VARAIBLES,CONCRETE_CONSTANTS
                                                      % Example output for
                                                      % MACHINE xx(T,u)
                                                      % CONSTRAINTS u:T
                                                      % SETS A; B={foo, bar}
                                                      % CONCRETE_CONSTANTS cc
                                                      % PROPERTIES cc:B
                                                      % ABSTRACT_VARIABLES xx
                                                      % CONCRETE_VARIABLES yy
                                                      % INVARIANT
                                                      %  xx:INT &
                                                      %  yy : T
                                                      % INITIALISATION xx,yy:= ({1|->2};{2|->4})(1), u
                                                      % END
                                                      %
                                                      % machine_identifiers(A,B,C,D,E,F,G).
                                                      % A = xx,
                                                      % B = [identifier(pos(4,1,1,12,1,12),'T'),identifier(pos(5,1,1,14,1,14),u)],
                                                      % C = [deferred_set(pos(11,1,3,6,3,6),'A'),enumerated_set(pos(12,1,3,9,3,20),'B',[identifier(pos(13,1,3,12,3,14),foo),identifier(pos(14,1,3,17,3,19),bar)])],
                                                      % D = [identifier(pos(22,1,6,20,6,21),xx)],
                                                      % E = [identifier(pos(24,1,7,20,7,21),yy)],
                                                      % F = [],
                                                      % G = [identifier(pos(16,1,4,20,4,21),cc)]
                              , get_machine_identifier_names/7 % a version returning atomic identifier names
                              ,abstract_constant/2, concrete_constant/2
                              ,possibly_abstract_constant/1
                              ,machine_operations/2   % machine_operations(M, Ops) gets the names of the Operations defined in Machine M
                              ,machine_operation_calls/2
                              ,machine_hash/2         % stores a hash value for the machine
                              ,properties_hash/2      % computes a hash over the constants and properties
                              ,operation_hash/3       % computes a hash for the operation in a machine
                              ,write_dot_hierarchy_to_file/1
                              ,write_dot_op_hierarchy_to_file/1
                              ,write_dot_event_hierarchy_to_file/1
                              ]).

:- use_module(library(lists)).
:- use_module(library(ordsets)).
:- use_module(bmachine,[b_get_definition/5, b_filenumber/4]).
:- use_module(bmachine_construction).
:- use_module(self_check).
:- use_module(specfile,[get_specification_description/2]).
:- use_module(extension('probhash/probhash'),[raw_sha_hash/2]).
:- use_module(input_syntax_tree).

:- use_module(value_persistance,[cache_is_activated/0]).

:- use_module(module_information,[module_info/2]).
:- module_info(group,analysis).
:- module_info(description,'This module provides functionality to visualize the dependencies of a B machine (include and sees relations, etc.).').

:- volatile
    main_machine_name/1,
    machine_type/2,
    machine_references/2,
    machine_identifiers/7,
    machine_operations/2,
    machine_operation_calls/2,
    refines_event/4,
    machine_has_assertions/1,
    raw_machine/2,
    machine_hash_cached/2,
    properties_hash_cached/2, operation_hash_cached/3,
    event_refinement_change/6.
:- dynamic
    main_machine_name/1,
    machine_type/2,
    machine_references/2,
    machine_identifiers/7,
    machine_operations/2,
    machine_operation_calls/2,
    refines_event/4,
    machine_has_assertions/1,
    raw_machine/2,
    machine_hash_cached/2,
    properties_hash_cached/2, operation_hash_cached/3,
    event_refinement_change/6.
:- volatile abstract_constant/2, concrete_constant/2.
:- dynamic abstract_constant/2.
:- dynamic concrete_constant/2.

:- use_module(specfile,[animation_minor_mode/1]).
possibly_abstract_constant(ID) :-
    (abstract_constant(ID,_)  ; animation_minor_mode(eventb),concrete_constant(ID,_) ).

reset_hierarchy :-
    retract_all(main_machine_name/1),
    retract_all(machine_type/2),
    retract_all(machine_references/2),
    retract_all(machine_identifiers/7),
    retract_all(abstract_constant/2),
    retract_all(concrete_constant/2),
    retract_all(machine_operations/2),
    retract_all(machine_operation_calls/2),
    retract_all(refines_event/4),
    retract_all(machine_has_assertions/1),
    retract_all(machine_hash_cached/2),
    retract_all(raw_machine/2),
    retract_all(properties_hash_cached/2),
    retract_all(operation_hash_cached/3),
    retract_all(event_refinement_change/6).

machine_name(Name) :- machine_type(Name,_).

:- use_module(eventhandling,[register_event_listener/3]).
:- register_event_listener(clear_specification,reset_hierarchy,
                           'Reset B Machine Hierarchy Facts.').

retract_all(Functor/Arity) :-
    functor(Pattern,Functor,Arity),
    retractall(Pattern).

analyse_hierarchy(Main,Machines) :- (var(Main) ; var(Machines)),!,
    add_internal_error('Illegal call:',analyse_hierarchy(Main,Machines)).
analyse_hierarchy(Main,Machines) :-
    reset_hierarchy,
    assert(main_machine_name(Main)),
    analyse_machine(Main,Machines,main).

:- use_module(error_manager).
:- use_module(tools_strings,[ajoin/2]).
:- public analyse_machine/3.
analyse_machine(Name,_Machines,_) :-
    % machine already analysed
    machine_type(Name,_),!.
analyse_machine(Name,Machines,_) :-
    debug:debug_println(19,analysing_machine(Name)),
    get_machine(Name,Machines,Type,Header,Refines,Body),
    !,
    ( cache_is_activated -> % we need the machines stored for later analysis
        assert_all_machines(Machines)
    ; otherwise -> true),
    assert(machine_type(Name,Type)),
    ( get_raw_section(assertions,Body,_) ->
        assert(machine_has_assertions(Name))
    ; true),
    store_identifiers(Name,Header,Body),
    store_operations(Name,Body),
    store_references(Name,Refines,Body,Machines).
analyse_machine(Name,Machines,RefType) :-
    b_filenumber(Name,_Ext,Nr,File),
    get_ref_type_name(RefType,Clause),
    !,
    ( member(M,Machines),get_constructed_machine_name_and_filenumber(M,OtherName,Nr)
      -> ajoin(['Cannot use B machine "',Name,'" within ', Clause,
                      ' clause. Rename machine "', OtherName,'" to "', Name, '" in file: '],Msg)
      ;  ajoin(['Cannot find B machine "',Name,'" within ', Clause,
                      ' clause. Check that machine name matches filename in: '],Msg)
    ),
    add_error_fail(invalid_machine_reference,Msg,File).
analyse_machine(Name,_Machines,_) :-
    add_error_fail(invalid_machine_reference,
      'Could not find machine in parsed machine list (check that your machine names match your filenames): ',Name).

% store the un-typed input syntax tree for later analysis
assert_all_machines(Machines) :- maplist(assert_raw_machine,Machines).
assert_raw_machine(Machine) :-
    get_raw_machine_name(Machine,Name),
    assert( raw_machine(Name,Machine) ).


store_identifiers(Name,Header,Body) :-
    get_parameters(Header,Params),
    get_sets(Body,Sets),
    get_identifiers([abstract_variables,variables],Body,AVars),
    get_identifiers([concrete_variables],Body,CVars),
    get_identifiers([abstract_constants],Body,AConsts),
    get_identifiers([concrete_constants,constants],Body,CConsts), % fixed abstract -> concrete
    assert(machine_identifiers(Name,Params,Sets,AVars,CVars,AConsts,CConsts)),
    maplist(assert_raw_id_with_position(abstract_constant),AConsts),
    maplist(assert_raw_id_with_position(concrete_constant),CConsts).


raw_id_is_identifier2(description(_Pos,_Desc,Raw),ID) :- !, raw_id_is_identifier2(Raw,ID).
raw_id_is_identifier2(deferred_set(_,ID),ID) :- !.
raw_id_is_identifier2(enumerated_set(_,ID,_Elements),ID) :- !.
raw_id_is_identifier2(Raw,ID) :- raw_id_is_identifier(Raw,_,ID).

get_machine_identifier_names(Name,Params,Sets,AVars,CVars,AConsts,CConsts) :-
    machine_identifiers(Name,RawParams,RawSets,RawAVars,RawCVars,RawAConsts,RawCConsts),
    maplist(raw_id_is_identifier2,RawParams,Params),
    maplist(raw_id_is_identifier2,RawSets,Sets),
    maplist(raw_id_is_identifier2,RawAVars,AVars),
    maplist(raw_id_is_identifier2,RawCVars,CVars),
    maplist(raw_id_is_identifier2,RawAConsts,AConsts),
    maplist(raw_id_is_identifier2,RawCConsts,CConsts).


machine_hash(Name,Hash) :-
    machine_hash_cached(Name,Hash1),!,Hash=Hash1.
machine_hash(Name,Hash) :-
    compute_machine_hash(Name,Hash1),
    assert( machine_hash_cached(Name,Hash1) ),
    Hash=Hash1.
compute_machine_hash(Name,Digest) :-
    if(raw_machine(Name,Machine),
       raw_sha_hash(Machine,Digest),
       add_error_and_fail(compute_machine_hash,'Machine does not exist or has not been processed:',Name)
      ).

operation_hash(MachName,OpName,Hash) :-
    operation_hash_cached(MachName,OpName,Hash1),!,Hash=Hash1.
operation_hash(MachName,OpName,Hash) :-
    compute_operation_hash(MachName,OpName,Hash1),
    assert( operation_hash_cached(MachName,OpName,Hash1) ),
    Hash=Hash1.
compute_operation_hash(MachName,OpName,Digest) :-
    raw_machine(MachName,Machine),
    get_machine(MachName,[Machine],_Type,_Header,_Refines,Body),
    get_opt_section(operations,Body,Operations),
    Op = operation(_,identifier(_,OpName),_,_,_),
    memberchk(Op,Operations),
    extract_used_np_definitions(Op,Body,UsedDefinitions),
    remove_raw_position_info(Op,RawOperation),!,

    raw_sha_hash(op(RawOperation,UsedDefinitions),Digest).


:- use_module(debug,[debug_println/2]).
assert_raw_id_with_position(PredFunctor,Rid) :-
   (raw_id_is_identifier(Rid,Pos,ID)
    -> true
    ; Rid = definition(Pos,ID,_)
     -> add_error(illegal_definition_use,'Definition cannot be used as identifier here: ',ID,Pos)
     ; add_error_fail(assert_raw_id_with_position,'Not identifier: ',Rid)
   ),
   Fact =.. [PredFunctor,ID,Pos],
   assert(Fact), debug_println(9,Fact).

raw_id_is_identifier(identifier(Pos,ID),Pos,ID).
raw_id_is_identifier(unit(_,_,identifier(Pos,ID)),Pos,ID).
raw_id_is_identifier(new_unit(_,_,identifier(Pos,ID)),Pos,ID).
raw_id_is_identifier(inferred_unit(_,_,identifier(Pos,ID)),Pos,ID).
raw_id_is_identifier(inferredunit(_,_,identifier(Pos,ID)),Pos,ID). % the (new?) parser seems to generate the wrong pragma in the .prob file; TO DO: investigate
raw_id_is_identifier(description(_,_,identifier(Pos,ID)),Pos,ID).

get_raw_identifier(Raw,Res) :- raw_id_is_identifier(Raw,_Pos,Id),!, Res=Id.
get_raw_identifier(definition(_Pos,DID,[]),ID) :-  % see also expand_definition_to_variable_list
   atom(DID),!,
   ajoin([DID,'(DEFINITION)'],ID).
get_raw_identifier(Raw,Res) :- add_internal_error('Cannot get identifier:',Raw),Res='???'.

raw_identifier_member(ID,List) :- member(Raw,List), raw_id_is_identifier(Raw,_Pos,ID).

store_operations(MachName,Body) :-
    get_opt_section(operations,Body,Operations),
    findall(I,(member(Op,Operations),get_raw_operation_id_and_body(Op,I,_)),Ids),
    assert(machine_operations(MachName,Ids)),
    findall(calls(I1,I2),(member(Op,Operations),get_raw_operation_call(Op,I1,I2)),Calls),
    sort(Calls,SCalls),
    assert(machine_operation_calls(MachName,SCalls)).

get_raw_operation_id_and_body(operation(_Pos,Id,_,_,Body),Id,Body).
get_raw_operation_id_and_body(refined_operation(_Pos,Id,_Results,_Args,_RefinesID,Body),Id,Body).

% get operations called in body
get_raw_operation_call(Op,Id,CallsId) :-
    get_raw_operation_id_and_body(Op,identifier(_,Id),Body),
    get_raw_operation_call(Body,identifier(_,CallsId)).

:- use_module(debug,[debug_format/3]).
get_raw_operation_call(block(_,Body),ID) :- !, get_raw_operation_call(Body,ID).
get_raw_operation_call(precondition(_,_,Body),ID) :- !, get_raw_operation_call(Body,ID).
get_raw_operation_call(assertion(_,_,Body),ID) :- !, get_raw_operation_call(Body,ID).
get_raw_operation_call(var(_,_,Body),ID) :- !, get_raw_operation_call(Body,ID).
get_raw_operation_call(select_when(_,_,Body),ID) :- !, get_raw_operation_call(Body,ID).
get_raw_operation_call(if_elsif(_,_,Body),ID) :- !, get_raw_operation_call(Body,ID).
get_raw_operation_call(let(_,_,_,Body),ID) :- !, get_raw_operation_call(Body,ID).
get_raw_operation_call(any(_,_,_,Body),ID) :- !, get_raw_operation_call(Body,ID).
get_raw_operation_call(case(_,_,_,_,_,Body),ID) :- !, get_raw_operation_call(Body,ID).
get_raw_operation_call(while(_,_Cond,Body,_,_),ID) :- !, get_raw_operation_call(Body,ID).
get_raw_operation_call(parallel(_,List),ID) :- !,member(A,List), get_raw_operation_call(A,ID).
get_raw_operation_call(sequence(_,List),ID) :- !,member(A,List), get_raw_operation_call(A,ID).
get_raw_operation_call(if(_,_Test,Then,List,Else),ID) :- !,member(A,[Then,Else|List]), get_raw_operation_call(A,ID).
get_raw_operation_call(select(_,_Cond,Body,List),ID) :- !,member(A,[Body|List]), get_raw_operation_call(A,ID).
get_raw_operation_call(select(_,_Cond,Body,List,Else),ID) :- !,member(A,[Body,Else|List]), get_raw_operation_call(A,ID).
get_raw_operation_call(choice(_,List),ID) :- !,member(A,List), get_raw_operation_call(A,ID).
get_raw_operation_call(choice_or(_,Body),ID) :- !, get_raw_operation_call(Body,ID).
get_raw_operation_call(operation_call(_,ID,_,_),ID) :- !.
get_raw_operation_call(skip(_),_) :- !, fail.
get_raw_operation_call(assign(_,_,_),_) :- !, fail.
get_raw_operation_call(becomes_element_of(_,_,_),_) :- !, fail.
get_raw_operation_call(becomes_such(_,_,_),_) :- !, fail.
get_raw_operation_call(definition(_,Name,_),_) :- !, % TO DO: improve
   debug_format(19,'Ignoring operation calls in DEFINITION ~w for operation call diagram~n',[Name]),
   % b_get_definition(Name,_DefType,_Args,DefBody,_Deps), not yet precompiled !
   fail.
get_raw_operation_call(Subst,_) :- functor(Subst,F,N),print(uncovered_subst(F,N,Subst)),nl,fail.
% we also do not find operation calls in expressions

store_references(Name,Refines,Body,Machines) :-
    debug:debug_println(9,store_references(Name)),
    get_refinements(Refines,Refs1),
    get_references(Body,Refs2),
    append(Refs1,Refs2,Refs),
    assert(machine_references(Name,Refs)),
    follow_refs(Refs,Machines).

get_references(Body,Refs) :-
    findrefs(Body,includes,Includes),
    findrefs(Body,extends,Extends),
    findrefs(Body,imports,Imports),
    findusessees(Body,uses,Uses),
    findusessees(Body,sees,Sees),
    append([Includes,Imports,Extends,Uses,Sees],Refs).

get_refinements([],[]).
get_refinements([Name|NRest],[ref(refines,Name,'')|RRest]) :-
    get_refinements(NRest,RRest).

findrefs(Body,Type,Refs) :-
    get_opt_section(Type,Body,RawRefs),
    findrefs2(RawRefs,Type,Refs).
findrefs2([],_Type,[]).
findrefs2([machine_reference(_Pos,R,_Params)|MRest],Type,[ref(Type,Name,Prefix)|RRest]) :-
    bmachine_construction:split_prefix(R,Prefix,Name),
    findrefs2(MRest,Type,RRest).

findusessees(Body,Type,Refs) :-
    get_opt_section(Type,Body,RawRefs),
    findusessees2(RawRefs,Type,Refs).
findusessees2([],_Type,[]).
findusessees2([identifier(_Pos,Name)|MRest],Type,[ref(Type,Name,'')|RRest]) :-
    findusessees2(MRest,Type,RRest).

follow_refs([],_Machines).
follow_refs([ref(RefType,Name,_Prefix)|Rest],Machines) :-
    analyse_machine(Name,Machines,RefType),
    follow_refs(Rest,Machines).

get_parameters(machine_header(_Pos,_Name,Params),Params).

get_sets(Body,Sets) :-
    get_opt_section(sets,Body,Sets).

get_identifiers([],_Body,[]).
get_identifiers([Sec|Rest],Body,Ids) :-
    get_opt_section(Sec,Body,Ids1),
    append(Ids1,IRest,Ids),
    get_identifiers(Rest,Body,IRest).

get_opt_sections([],_Body,[]).
get_opt_sections([S|Srest],Body,Contents) :-
    get_opt_section(S,Body,L),append(L,Rest,Contents),
    get_opt_sections(Srest,Body,Rest).

get_opt_section(Sec,Body,Result) :-
    ( get_raw_section(Sec,Body,Content) -> Result=Content; Result=[]).
get_raw_section(Sec,Body,Content) :- % look for Sec(_Pos,Content) in Body list
    functor(Pattern,Sec,2),arg(2,Pattern,Content),
    memberchk(Pattern,Body).

get_machine(Name,Machines,Type,Header,Refines,Body) :-
    get_machine1(Name,Machines,_Machine,Type,Header,Refines,Body).
%get_raw_machine(Name,Machines,Machine) :-
%    get_machine1(Name,Machines,Machine,_Type,_Header,_Refines,_Body).
get_machine1(Name,Machines,Machine,Type,Header,Refines,Body) :-
    Header = machine_header(_Pos,Name,_Params),
    member(Machine,Machines),
    get_machine2(Machine,Type,Header,Refines,Body),!.
get_machine2(abstract_machine(_Pos,MS,Header,Body),TypeOfAbstractMachine,Header,[],Body) :-
    get_abstract_machine_type(MS,TypeOfAbstractMachine).
get_machine2(refinement_machine(_Pos,Header,Refines,Body),refinement,Header,[Refines],Body).
get_machine2(implementation_machine(_Pos,Header,Refines,Body),implementation,Header,[Refines],Body).
get_machine2(generated(_Pos,Machine),A,B,C,D) :- % @generated Pragma used at top of file
    get_machine2(Machine,A,B,C,D).
get_machine2(unit_alias(_Pos,_Name,_Alias,Machine),A,B,C,D) :-
    get_machine2(Machine,A,B,C,D).

get_raw_machine_name(Machine,Name) :-
    get_machine2(Machine,_,machine_header(_,Name,_),_,_).

get_abstract_machine_type(machine(_Pos2),R) :- !,R=abstract_machine.
get_abstract_machine_type(system(_Pos2),R) :- !,R=abstract_machine.
get_abstract_machine_type(model(_Pos2),R) :- !,R=abstract_model.
get_abstract_machine_type(X,R) :- atomic(X),!,
    add_error(get_abstract_machine_type,'Your parser seems out-dated. Assuming abstract_machine: ',X),
    R=abstract_machine.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% write dot operation call graph
write_dot_op_hierarchy_to_file(File) :-
    (get_preference(dot_event_hierarchy_horizontal,true) -> PageOpts=[horizontal] ; PageOpts=[]),
    gen_dot_graph(File,PageOpts,b_machine_hierarchy,dot_operation_node,dot_op_calls_op,none,dot_subgraph).


:- use_module(bmachine,[b_top_level_operation/1, b_top_level_feasible_operation/1]).
:- public dot_operation_node/6.
dot_operation_node(OpName,M,Desc,Shape,Style,Color) :-
    machine_operations(M,Operations),
    (machine_promotes_operations(M,Promotes) -> true ; Promotes=[]),
    raw_identifier_member(OpName,Operations),
    (b_top_level_feasible_operation(OpName) -> Color=lightgray
     ; b_top_level_operation(OpName) -> Color='OldLace' % commented out operation
     ; raw_identifier_member(OpName,Promotes) -> Color='gray98' % promoted but not to top_level
     ; Color=white),
    Desc=OpName, Shape=box, Style=filled.

:- public dot_op_calls_op/5.
dot_op_calls_op(Op1,Label,Op2,Color,Style) :- Style=solid, Color=steelblue,
    Label = '',  % TO DO: different colors for local operation calls, detect op calls in expressions?
    machine_operation_calls(_,Operations),
    member(calls(Op1,Op2),Operations).



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% print hierarchy to dot file

maxlinelength(30).

:- use_module(tools_io).

% write hierarchy of machines, with inclusion/refinement links
write_dot_hierarchy_to_file(Filename) :-
    findall(M,machine_type(M,_),Machines),
    safe_open_file(Filename,write,S,[]),
    ( print_header(S),
      print_machines(S,1,Machines,Ids),
      print_refs_for_dot(S,Ids,Ids),
      print_footer(S)
      -> true
      ; add_internal_error('Command failed:',write_dot_hierarchy_to_file(Filename))
    ),
    close(S).

print_header(S) :-
    write(S,'digraph module_hierarchy {\n'),
    write(S,'  graph [page="8.5, 11",ratio=fill,size="7.5,10"];\n').
print_footer(S) :-
    write(S,'}\n').

print_machines(_S,_Nr,[],[]).
print_machines(S,Nr,[M|Machines],[id(M,Nr)|Ids]) :-
    print_machine(S,Nr,M),
    Nr2 is Nr + 1,
    print_machines(S,Nr2,Machines,Ids).
print_machine(S,Nr,M) :-
    write(S,'  '),write(S,Nr),
    (main_machine_name(M) ->
          write(S,' [shape=record, style=bold, color=darkgreen, label=\"|{')
       ;  write(S,' [shape=record, color=steelblue, label=\"|{')),
    print_machine_header(S,M),
    machine_identifiers(M,_Params,Sets,AVars,CVars,AConsts,CConsts),
    %print_hash(S,M),
    print_sets(S,Sets),
    print_identifiers(S,'VARIABLES',AVars),
    print_identifiers(S,'CONCRETE VARIABLES',CVars),
    print_identifiers(S,'ABSTRACT CONSTANTS',AConsts),
    print_identifiers(S,'CONSTANTS',CConsts),
    ( machine_has_assertions(M) ->
        get_specification_description(assertions,AssStr),
        print_title(S,AssStr)
      ; true),
    (machine_promotes_operations(M,Promotes)
     -> print_identifiers(S,'PROMOTES',Promotes) ; true),
    machine_operations(M,Operations),
    delete(Operations,identifier(_,'INITIALISATION'),Operations2),
    exclude(is_refinining_event(M),Operations2,RefiningOperations),
    include(is_refinining_event(M),Operations2,NewOperations),
    ((RefiningOperations=[] ; NewOperations=[])
     -> get_specification_description(operations,OpStr),
        print_identifiers(S,OpStr,Operations2)
     ;  print_identifiers(S,'EVENTS (refining)',RefiningOperations),
        print_identifiers(S,'EVENTS (new)',NewOperations)
    ),

    write(S,'}|\"];\n').

is_refinining_event(M,identifier(_,Event)) :- (refines_event(M,Event,_,_) -> true).

%print_hash(S,M) :- machine_hash(M,Digest),!,print_title(S,'Digest'),
%    maplist(format(S,'~16r'),Digest),write(S,'\\n').
%print_hash(_S,_M).

%get_machine_colour(M,steelblue) :- main_machine_name(M),!.
%get_machine_colour(_M,steelblue).

print_machine_header(S,M) :-
    machine_type(M,Type),
    get_machine_type(Type,P),
    write(S,P),write(S,' '),
    write(S,M),write(S,'\\n').
get_machine_type(abstract_machine,'MACHINE').
get_machine_type(abstract_model,'MODEL').
get_machine_type(refinement,'REFINEMENT').
get_machine_type(implementation,'IMPLEMENTATION').
get_machine_type(context,'CONTEXT').

print_sets(_,[]) :- !.
print_sets(S,Sets) :-
    write(S,'|SETS\\n'),
    print_sets2(Sets,S).
print_sets2([],_).
print_sets2([Set|Rest],S) :-
    print_set(Set,S),
    write(S,'\\n'),
    print_sets2(Rest,S).
print_set(description(_Pos,_Desc,Raw),S) :- print_set(Raw,S).
print_set(deferred_set(_Pos,Name),S) :-
    write_id(S,Name).
print_set(enumerated_set(_Pos,Name,List),S) :-
    write_id(S,Name),write(S,' = \\{'),
    maxlinelength(Max),
    preferences:get_preference(dot_hierarchy_max_ids,MaxIdsToPrint),
    print_set_elements(List,S,Max,MaxIdsToPrint),
    write(S,'\\}').
print_set_elements([],_,_,_).
print_set_elements([identifier(_Pos,Name)],S,_LenSoFar,_) :-
    !,write_id(S,Name).
print_set_elements([identifier(_Pos,Name),B|Rest],S,LenSoFar,MaxIdsToPrint) :-
    dec_atom_length(LenSoFar,Name,NewLen),
    (NewLen<0 -> maxlinelength(NL0),NL is NL0-1,write(S,'\\n ') ; NL is NewLen-1),
    write_id(S,Name),write(S,','),
    M1 is MaxIdsToPrint-1,
    (M1 < 1, Rest \= []
      -> write(S,'...,'),
         last([B|Rest],Last), print_set_elements([Last],S,NL,M1)
      ;  print_set_elements([B|Rest],S,NL,M1)
    ).

:- use_module(tools,[string_escape/2]).
% write identifier and escape it for dot
write_id(S,Name) :- string_escape(Name,EName), write(S,EName).

dec_atom_length(Prev,Atom,New) :- atom_length(Atom,Len),
    New is Prev-Len.

print_title(S,Title) :-
    write(S,'|'),
    write(S,Title),write(S,'\\n').
print_identifiers(_S,_,[]) :- !.
print_identifiers(S,Title,List) :-
    print_title(S,Title),
    preferences:get_preference(dot_hierarchy_max_ids,MaxIdsToPrint), % how many do we print overall; -1 means print all of them
    print_identifiers2(List,0,_,S,MaxIdsToPrint).
print_identifiers2([],Count,Count,_,_MaxIdsToPrint).
print_identifiers2([RawID|Rest],Count,NCount,S,MaxIdsToPrint) :-
    get_raw_identifier(RawID,Name),
    (MaxIdsToPrint = 0,Rest\=[]
     -> length(Rest,Len), RN is Len+1,
        format(S,' (~w more)',[RN])
     ;  MaxIdsToPrint1 is MaxIdsToPrint-1,
        print_identifier(Name,Count,ICount,S),
        (Rest = [] -> true; write(S,',')),
        print_identifiers2(Rest,ICount,NCount,S,MaxIdsToPrint1)
    ).
print_identifier(Name,Count,NewCount,S) :-
    atom_codes(Name,Codes),
    length(Codes,Length),
    NewCount1 is Count+Length,
    maxlinelength(Max),
    ( Count == 0       -> NewCount1=NewCount
    ; NewCount1 =< Max -> NewCount1=NewCount
    ; otherwise       ->  NewCount=0,write(S,'\\n')),
    write_id(S,Name).

print_refs_for_dot(_S,[],_).
print_refs_for_dot(S,[id(M,Nr)|Rest],Ids) :-
    machine_references(M,Refs),
    filter_redundant_refs(Refs,Refs,UsefulRefs),
    print_refs2(UsefulRefs,M,Nr,Ids,S),
    print_refs_for_dot(S,Rest,Ids).
print_refs2([],_,_,_,_).
print_refs2([ref(Type,Dest,_Prefix)|Rest],M,Nr,Ids,S) :-
    member(id(Dest,DestNr),Ids),!,
    get_ref_type(Type,Ref,Dir), !,
    (Dir=reverse ->
        format(S,'  ~w -> ~w ~w;~n',[DestNr,Nr,Ref])
      ; format(S,'  ~w -> ~w ~w;~n',[Nr,DestNr,Ref])
    ),
    print_refs2(Rest,M,Nr,Ids,S).
get_ref_type(includes,'[label=\"INCLUDES\",color=navyblue]',normal).
get_ref_type(imports,'[label=\"IMPORTS\",color=navyblue]',normal).
get_ref_type(extends,'[label=\"EXTENDS\",color=navyblue]',normal).
get_ref_type(uses,'[label=\"USES\",color=navyblue,style=dashed]',normal).
get_ref_type(sees,'[label=\"SEES\",color=navyblue,style=dashed]',normal).
%get_ref_type(sees,'[label=\"SEES\",color=navyblue,style=dashed,dir=back]',reverse).
%get_ref_type(refines,'[label=\"REFINES\",color=navyblue,style=bold,dir=back]',reverse).
get_ref_type(refines,'[label=\"REFINEMENT\",color=navyblue,style=bold]',reverse).  % reverse so that abstract machines are shown on top
get_ref_type(UNKNOWN,'[label=\"UNKNOWN\",color=navyblue]',normal) :- add_internal_error('Unknown : ',get_ref_type(UNKNOWN,_,_)).

get_ref_type_name(includes,'INCLUDES').
get_ref_type_name(imports,'IMPORTS').
get_ref_type_name(extends,'EXTENDS').
get_ref_type_name(uses,'USES').
get_ref_type_name(sees,'SEES').
get_ref_type_name(refines,'REFINEMENT').
get_ref_type_name(main,'MACHINE').
get_ref_type_name(UNKNOWN,UNKNOWN).


% remove redundant references (has to be done after analyse_eventb_machine has asserted all facts)
filter_redundant_refs([],_,[]).
filter_redundant_refs([ref(sees,Dest,_)|T],All,R) :-
    machine_type(Dest,context),
    member(ref(sees,Other,_),All),
    machine_references(Other,Refs),
    member(ref(extends,Dest,_),Refs), % Dest already seen by other seen context;
    % Note: Rodin export contains transitive sees relation;
    %   if the user had included Dest in the sees clause we would have a warning "Redundant seen context"
    !,
    %format(user_output,'Redundant sees of ~w (~w)~n',[Dest,Other]),
    filter_redundant_refs(T,All,R).
filter_redundant_refs([H|T],All,[H|R]) :-
    filter_redundant_refs(T,All,R).

/************************************************************************/
/* The same for Event-B                                                 */
/************************************************************************/

analyse_eventb_hierarchy(Machines,Contextes) :-
    reset_hierarchy,
    get_eventb_name(Machines,Contextes,MainName),
    assert(main_machine_name(MainName)),
    maplist(analyse_eventb_machine,Machines),
    analyse_eventb_refinement_types(Machines),
    maplist(analyse_eventb_context,Contextes),!.
analyse_eventb_hierarchy(Machines,Contextes) :-
    add_internal_error('Analyzing Event-B Hierarchy Failed: ',analyse_eventb_hierarchy(Machines,Contextes)).

get_eventb_name([MainMachine|_AbstractMachines],_Contextes,Name) :-
    event_b_model(MainMachine,Name,_),!.
get_eventb_name(_Machines,[MainContext|_AbstractContextes],Name) :-
    event_b_context(MainContext,Name,_).

event_b_model(event_b_model(_,Name,Sections),Name,Sections).
event_b_context(event_b_context(_,Name,Sections),Name,Sections).

analyse_eventb_machine(Machine) :-
    event_b_model(Machine,Name,Sections),
    % print(analyzing(Name)),nl,
    ( memberchk(refines(_,RName),Sections) -> Type=refinement, RRefs=[ref(refines,RName,'')]
    ; otherwise                            -> Type=abstract_model, RRefs=[], RName='$none'),
    get_identifiers([variables],Sections,Vars),
    get_sees_context_refs(Sections,SRefs),
    append(RRefs,SRefs,Refs),
    get_events(Name,Sections,Events,RName),
    raw_sha_hash(Machine,Digest),
    assert(machine_type(Name,Type)),
    assert(machine_identifiers(Name,[],[],Vars,[],[],[])),
    assert(machine_references(Name,Refs)),
    assert(machine_operations(Name,Events)),
    assert(machine_operation_calls(Name,[])), % Event-B events cannot call other events
    assert_if_has_theorems(Name,Sections),
    assert(machine_hash_cached(Name,Digest)).


% analyze which kinds of refinments we have between events
analyse_eventb_refinement_types([]).
analyse_eventb_refinement_types([RefMachine,AbsMachine|_]) :-
    event_b_model(RefMachine,RefName,RefSections),
    memberchk(refines(_,AbsName),RefSections),
    event_b_model(AbsMachine,AbsName,AbsSections),
    get_opt_section(events,RefSections,RawEvents),
    get_opt_section(events,AbsSections,AbsRawEvents),
    member(RawEvent,RawEvents),
    bmachine_eventb:raw_event(RawEvent,_,RefEvName,_St1,Ref,_Prm1,RefGrd,_Thm1,RefAct,_Wit1),
    Ref = [AbsEvName],
    member(AbsRawEvent,AbsRawEvents),
    bmachine_eventb:raw_event(AbsRawEvent,_,AbsEvName,_St2,_,_Prm2,AbsGrd,_Thm2,AbsAct,_Wit2),
    check_raw_prefix(AbsGrd,RefGrd,SameGuard),
    check_raw_prefix(AbsAct,RefAct,SameAct),
    %% format('~nEvent refinement change ~w (~w) -> ~w (~w)  guard: ~w, action: ~w~n',[RefEvName,RefName,AbsEvName,AbsName,SameGuard,SameAct]),
    %print(rawgrd(RefGrd,AbsGrd)), nl, print(rawact(RefAct,AbsAct)),nl,
    assert(event_refinement_change(RefName,RefEvName,AbsName,AbsEvName,SameGuard,SameAct)),
    fail.
analyse_eventb_refinement_types([_|T]) :- analyse_eventb_refinement_types(T).

check_raw_prefix([],[],Res) :- !, Res=unchanged.
check_raw_prefix([],[_|_],Res) :- !, Res=extends. % the refinement has some more actions/guards
check_raw_prefix([Abs|AT],[Ref|RT],Result) :- same_raw_expression(Abs,Ref),!,
   check_raw_prefix(AT,RT,Result).
check_raw_prefix(_,_,refines).

% TO DO: check if we have a more complete version of this predicate; to do: handle @desc description/3 terms
same_raw_expression(identifier(_,A),RHS) :- !, RHS=identifier(_,B), A=B.
same_raw_expression(equal(_,A,B),RHS) :- !, RHS=equal(_,A2,B2), same_raw_expression(A,A2), same_raw_expression(B,B2).
same_raw_expression(assign(_,A1,A2),RHS) :- !, RHS=assign(_,B1,B2),
    maplist(same_raw_expression,A1,B1), maplist(same_raw_expression,A2,B2).
same_raw_expression(A,B) :- atomic(A),!, A=B.
same_raw_expression(A,B) :- A =.. [F,_|AA],  % print(match(F,AA)),nl,
    B=.. [F,_|BB], maplist(same_raw_expression,AA,BB).



get_sees_context_refs(Sections,SRefs) :-
    get_opt_section(sees,Sections,Seen),
    findall(ref(sees,I,''),member(I,Seen),SRefs).

:- use_module(bmachine_eventb,[raw_event/10]).
get_events(Name,Sections,Events,AbsMachineName) :-
    get_opt_section(events,Sections,RawEvents),
    compute_event_refines(Name,RawEvents,AbsMachineName),
    findall( identifier(Pos,EvName),
             ( member(RawEvent,RawEvents),
                 raw_event(RawEvent,Pos,EvName,_St,_Rf,_Prm,_Grd,_Thm,_Act,_Wit)  ),
             Events).

% TO DO: detect when an event extends another one without changing it
compute_event_refines(MachineName,RawEvents,AbsMachineName) :-
    member(RawEvent,RawEvents),
    bmachine_eventb:raw_event(RawEvent,_Pos,EvName,_St,Refines,_Prm,_Grd,_Thm,_Act,_Wit),
    member(RefEvent,Refines),
     % print(refines(MachineName,EvName,AbsMachineName,RefEvent)),nl,
    assert(refines_event(MachineName,EvName,AbsMachineName,RefEvent)),
    fail.
compute_event_refines(_,_,_).


% try and get the name of the machine we refine
%machine_refines(Machine,AbsMachine) :- machine_references(Machine,Refs), member(ref(refines,AbsMachine,_),Refs).

analyse_eventb_context(Context) :-
    event_b_context(Context,Name,Sections),
    get_identifiers([constants],Sections,ConcreteConstants),
    get_identifiers([abstract_constants],Sections,AbstractConstants),
    append([ConcreteConstants,AbstractConstants],AllConstants),
    get_sets(Sections,Sets),
    get_extends_refs(Sections,Refs),
    raw_sha_hash(Context,Digest),
    assert(machine_type(Name,context)),
    assert(machine_identifiers(Name,[],Sets,[],[],[],AllConstants)),
    assert(machine_references(Name,Refs)),
    assert(machine_operations(Name,[])),
    assert(machine_operation_calls(Name,[])),
    maplist(assert_raw_id_with_position(concrete_constant),ConcreteConstants),
    maplist(assert_raw_id_with_position(abstract_constant),AbstractConstants),
    assert_if_has_theorems(Name,Sections),
    assert(machine_hash_cached(Name,Digest)).

get_extends_refs(Sections,Refs) :-
    get_opt_section(extends,Sections,Extended),
    findall(ref(extends,E,''),member(E,Extended),Refs).

assert_if_has_theorems(Name,Sections) :-
    get_opt_section(theorems,Sections,[_|_]),
    assert(machine_has_assertions(Name)).
assert_if_has_theorems(_Name,_Sections).


% compute a hash based on constants and properties
properties_hash(MachineName,Hash) :-
    properties_hash_cached(MachineName,Hash1),!,
    Hash=Hash1.
properties_hash(MachineName,Hash) :-
    compute_properties_hash(MachineName,Hash1),
    assert(properties_hash_cached(MachineName,Hash1)),
    Hash=Hash1.
compute_properties_hash(Name,Hash) :-
    raw_machine(Name,Machine),
    get_machine(Name,[Machine],_Type,_Header,_Refines,Body),
    extract_sorted_np_sets(Body,Sets),
    extract_sorted_np_constants(Body,Constants),
    extract_np_properties(Body,Properties),
    extract_used_np_definitions_from_properties(Body,Definitions),
    ToHash = [Sets,Constants,Properties,Definitions],
    raw_sha_hash(ToHash,Hash).
    % save_properties_hash(Name,ToHash,Hash).

extract_sorted_np_sets(Body,Sets) :-
    get_sets(Body,PosSets),
    remove_raw_position_info(PosSets,UnsortedSets),
    sort(UnsortedSets,Sets).
extract_sorted_np_constants(Body,Constants) :-
    get_opt_sections([constants,concrete_constants,abstract_constants],Body,PosConstants),
    remove_raw_position_info(PosConstants,UnsortedConstants),
    sort(UnsortedConstants,Constants).
extract_np_properties(Body,Properties) :-
    get_opt_section(properties,Body,PosProperties),
    remove_raw_position_info(PosProperties,Properties).
extract_used_np_definitions_from_properties(Body,Definitions) :-
    get_opt_section(properties,Body,Properties),
    extract_used_np_definitions(Properties,Body,Definitions).
extract_used_np_definitions(RawSyntax,Body,Definitions) :-
    extract_raw_identifiers(RawSyntax,UsedIds),
    all_definition_ids(Body,AllDefs),
    ord_intersection(UsedIds,AllDefs,UsedDefNames),
    transitive_used_definitions(UsedDefNames,AllUsedDefs),
    findall( definition(none,Name,Args,DefBody),
             ( member(Name,AllUsedDefs),b_get_definition(Name,_DefType,Args,DefBody,_Deps)),
             PosDefinitions),
    maplist(remove_raw_position_info,PosDefinitions,Definitions).
all_definition_ids(Body,Ids) :-
    get_opt_section(definitions,Body,Definitions),
    convlist(get_definition_name,Definitions,Ids1),
    sort(Ids1,Ids).
transitive_used_definitions(Defs,TransDefs) :-
    findall( D, reachable_definition(Defs,D), TD1),
    sort(TD1,TransDefs).
reachable_definition(Defs,D) :-
    member(N,Defs),
    ( D=N
    ; b_get_definition(N,_DefType,_Args,_Body,Deps),
        reachable_definition(Deps,D)).


/* just for debugging:
save_properties_hash(MachineName,ToHash,Hash) :-
    main_machine_name(Main),
    open('/home/plagge/hashes.pl',append,S),
    writeq(S,hash(Main,MachineName,ToHash,Hash)),
    write(S,'.\n'),
    close(S).
*/

% ---------------------------
:- dynamic event_info/3.
analyze_extends_relation :-
   retractall(event_info(_,_,_)),
   bmachine:b_get_machine_operation(_Name,_Results,_RealParameters,TBody,_OType,_OpPos),
   treat_event_body(TBody),fail.
analyze_extends_relation.

treat_event_body(TBody) :-
    rlevent_info(TBody,EventName,Machine,Status,AbstractEvents),
    % format('Event ~w:~w ~w~n',[Machine,EventName,Status]),
    assert(event_info(Machine,EventName,Status)),
    maplist(treat_event_body,AbstractEvents).

:- use_module(bsyntaxtree,[get_texpr_expr/2]).
rlevent_info(TBody,EventName,Machine,FStatus,AbstractEvents) :-
    get_texpr_expr(TBody,Event),
    Event = rlevent(EventName,Machine,TStatus,_Params,_Guard,_Theorems,_Actions,_VWit,_PWit,_Unmod,AbstractEvents),
    bsyntaxtree:get_texpr_expr(TStatus,Status), %ordinary, convergent, anticipated
    functor(Status,FStatus,_).


% write the event refinement hierarchy to a dot file
% (currently) only makes sense for Event-B models

:- use_module(dot_graph_generator,[gen_dot_graph/7]).
:- use_module(preferences,[get_preference/2]).
:- use_module(tools_strings,[ajoin/2]).

:- public dot_event_node/6.
dot_event_node(M:Ev,M,Desc,Shape,Style,Color) :-
    machine_operations(M,Evs),
    raw_identifier_member(Ev,Evs),
    Ev \= 'INITIALISATION',
    event_info(M,Ev,Status),
    (event_refinement_change(M,Ev,_,_,SameGuard,SameAction)
      -> true ; SameGuard=unknown, SameAction=unknown),
    (Status = convergent -> ajoin([Ev,' (<)'],Desc)
     ; Status = anticipated -> ajoin([Ev, '(<=)'],Desc)
     ; otherwise -> Desc=Ev),
    dot_get_color_style(M,Ev,Status,SameGuard,SameAction,Shape,Color,Style).

dot_get_color_style(M,Ev,_Status,_,_,box,Color,Style) :- \+ refines_event(M,Ev,_,_),!,
     get_preference(dot_event_hierarchy_new_event_colour,Color), Style=none.
dot_get_color_style(M,Ev,_,SameGuard,SameAction,box,Color,Style) :-
     refines_event(M,Ev,_,Ev2), dif(Ev2,Ev),!, % changes name
     ((SameGuard,SameAction)=(unchanged,unchanged)
      -> get_preference(dot_event_hierarchy_rename_unchanged_event_colour,Color)
      ;  get_preference(dot_event_hierarchy_rename_event_colour,Color)), Style=filled.
dot_get_color_style(_M,_Ev,_,unchanged,unchanged,box,Color,Style) :- % keeps name and adds no guard or action
     !,
     get_preference(dot_event_hierarchy_unchanged_event_colour,Color), Style=filled.
dot_get_color_style(_M,_Ev,_,_,unchanged,box,Color,Style) :- % keeps name and adds no guard or action
     !,
     get_preference(dot_event_hierarchy_grd_strengthening_event_colour,Color), Style=filled.
dot_get_color_style(_M,_Ev,_,unchanged,_,box,Color,Style) :- % keeps name and adds action but keeps guard action
     !,
     get_preference(dot_event_hierarchy_grd_keeping_event_colour,Color), Style=filled.
dot_get_color_style(_M,_Ev,_,_,_,box,Color,Style) :- % keeps name but adds or modifies
 % TO DO: distinguish extends from refines
     get_preference(dot_event_hierarchy_refines_colour,Color), Style=filled.

:- public dot_refines_event/5.
dot_refines_event(M2:Ev2,Label,M1:Ev1,Color,Style) :-
  Label = '',  % TO DO: detect refine, extends, identical
  refines_event(M1,Ev1,M2,Ev2),
  Ev1 \= 'INITIALISATION',
  (event_refinement_change(M1,Ev1,_,_,SameGuard,SameAct)
   -> arrow_style(SameGuard,SameAct,Style,ColPref), get_preference(ColPref,Color)
   ; Style=solid, Color=red % should not happen
   ).

arrow_style(unchanged,unchanged,Style,ColPref) :- !,
     Style=arrowhead(none,solid), ColPref=dot_event_hierarchy_extends_colour.
arrow_style(refines,_,Style,ColPref) :- !, Style=solid, ColPref=dot_event_hierarchy_edge_colour.
arrow_style(_,refines,Style,ColPref) :- !, Style=solid, ColPref=dot_event_hierarchy_edge_colour.
%arrow_style(_,_,Style) :- Style = arrowhead(vee,arrowtail(box,solid)). % we have extends
arrow_style(_,_,Style,dot_event_hierarchy_extends_colour) :- Style = arrowhead(vee,solid). % we have extends

:- public dot_subgraph/3.
dot_subgraph(M,filled,Colour) :- get_preference(dot_event_hierarchy_machine_colour,Colour),
  machine_operations(M,[_|_]).

:- public dot_same_rank/1.
dot_same_rank(SameRankVals) :- machine_operations(M,Evs),
    findall(M:Ev,raw_identifier_member(Ev,Evs),SameRankVals).

write_dot_event_hierarchy_to_file(File) :-
    analyze_extends_relation,
    (get_preference(dot_event_hierarchy_horizontal,true) -> PageOpts=[horizontal] ; PageOpts=[]),
    gen_dot_graph(File,PageOpts,b_machine_hierarchy,dot_event_node,dot_refines_event,none,dot_subgraph).
    %gen_dot_graph(File,b_machine_hierarchy,dot_event_node,dot_refines_event,dot_same_rank,dot_subgraph).
