% (c) 2014-2020 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(before_after_predicates, [tcltk_before_after_predicate_for_operation/2,
                                    before_after_predicate_for_operation/2,
                                    before_after_predicate_with_equalities/3,
                                    before_after_predicate_list_conjunct_with_equalities/3,
                                    before_after_predicate_list_disjunct_with_equalities/3,
                                    list_renamings_to_primed_vars/2]).

:- use_module(library(lists)).

:- use_module(bsyntaxtree, [get_texpr_expr/2,
                            get_texpr_id/2,
                            is_truth/1,
                            safe_create_texpr/3,
                            get_texpr_type/2,
                            conjunct_predicates/2,
                            conjunction_to_list/2,
                            create_implication/3,
                            create_couple/3,
                            disjunct_predicates/2,
                            find_typed_identifier_uses/2,
                            remove_all_infos/2,
                            rename_bt/3,
                            create_exists/3,
                            create_negation/2]).
:- use_module(error_manager, [add_error_fail/3]).
:- use_module(bmachine_eventb, [prime_variable/2,
                                prime_variables/2,
                                is_primed_id/1]).
:- use_module(bmachine,[b_get_machine_operation/4,
                        b_get_machine_variables/1,
                        b_get_machine_constants/1]).
%:- use_module(specfile,[animation_minor_mode/1]).
:- use_module(translate,[translate_bexpression/2]).

:- use_module(module_information,[module_info/2]).
:- module_info(group,misc).
:- module_info(description,'This module computes before-after-predicates of B and Event-B actions.').

tcltk_before_after_predicate_for_operation(OpName,list(Res)) :-
    before_after_predicate_for_operation(OpName,BAPredicate),
    %translate:print_bexpr(BAPredicate),nl,
    conjunction_to_list(BAPredicate,LP),
    maplist(translate_bexpression,LP,Res).

before_after_predicate_for_operation(OpName,BAPredicate) :-
    b_get_machine_operation(OpName,_Results,_Parameters,OpBody),
    %print(op(OpName)),nl,
    b_get_machine_variables(Variables),
    b_get_machine_constants(Constants),
    append(Variables,Constants,VarsAndConsts),
    before_after_predicate_with_equalities(OpBody,VarsAndConsts,BAPredicate).

:- mode before_after_predicate_with_equalities(+Action,+AllIdentifiers,-Predicate).
% these add missing y' = y equalities for unchanged variables
before_after_predicate_with_equalities(Action,AllIdentifiers,Predicate) :-
    before_after_predicate_list_conjunct_with_equalities([Action],AllIdentifiers,Predicate).
before_after_predicate_list_conjunct_with_equalities(Actions,AllIdentifiers,BAPredicateWithEq) :-
    maplist(before_after_predicate_no_equalities_tl(top_level),Actions,AllPredicates),
    conjunct_predicates(AllPredicates,BAPredicate),
    add_missing_equalities(AllIdentifiers,BAPredicate,BAPredicateWithEq).
before_after_predicate_list_disjunct_with_equalities(Actions,AllIdentifiers,BAPredicate) :-
    maplist(before_after_predicate_no_equalities_tl(top_level),Actions,AllPredicates),
    maplist(add_missing_equalities(AllIdentifiers),AllPredicates,BAPredicates),
    disjunct_predicates(BAPredicates,BAPredicate).

% does not add missing y' = y equalities for unchanged variables
% top-level version: has a special rule for PRE
before_after_predicate_no_equalities_tl(TopLevel,Action,Predicate) :-
    get_texpr_expr(Action,Expr),
    before_after_predicate_no_equalities_top_level_aux(Expr,Predicate,TopLevel).

:- use_module(preferences,[get_preference/2]).
before_after_predicate_no_equalities_top_level_aux(precondition(Pre,Body),Predicate,top_level) :-
    get_preference(treat_outermost_pre_as_select,true),
    !,
    before_after_predicate_no_equalities(Body,BABody,_),
    conjunct_predicates([Pre,BABody],Predicate). % May sometimes have to be an implication (depending on how BA-predicate is used) !
before_after_predicate_no_equalities_top_level_aux(Expr,Pred,_) :-
   before_after_predicate_no_equalities_aux(Expr,Pred,_).

% does not add missing y' = y equalities for unchanged variables
before_after_predicate_no_equalities(Action,Predicate,ChangedIDs) :-
    get_texpr_expr(Action,Expr),
    before_after_predicate_no_equalities_aux(Expr,Predicate,ChangedIDs).

get_changed_id(b(function(FUN,_),_,_),ID) :- !, get_changed_id(FUN,ID).
get_changed_id(ID,ID).

before_after_predicate_no_equalities_aux(assign_single_id(Id,Pred),Predicate,ChangedIDs) :- !,
    before_after_predicate_no_equalities_aux(assign([Id],[Pred]),Predicate,ChangedIDs).
before_after_predicate_no_equalities_aux(assign(Ids,Preds),Predicate,ChangedIds) :- !,
    maplist(get_changed_id,Ids,ChangedIds),
    maplist(ba_predicate_equal,Ids,Preds,Predicates),
    conjunct_predicates(Predicates,Predicate).
before_after_predicate_no_equalities_aux(becomes_element_of(Ids,Set),Predicate,Ids) :- !,
    maplist(ba_predicate_member(Set),Ids,Predicates),
    conjunct_predicates(Predicates,Predicate). % should we generate Set /= {} => Predicate
before_after_predicate_no_equalities_aux(becomes_such(Ids,Pred),Predicate,Ids) :- !,
    ba_predicate_becomes_such(Ids,Pred,Predicate).
before_after_predicate_no_equalities_aux(choice(Choices),Predicate,ChangedIDs) :- !,
    maplist(before_after_predicate_no_equalities,Choices,BAChoices,Changes),
    append(Changes,ChangedIDs),
    maplist(add_missing_equalities(ChangedIDs),BAChoices,BAChoicesWithEqualities),
    disjunct_predicates(BAChoicesWithEqualities,Predicate).
before_after_predicate_no_equalities_aux(precondition(Pre,Body),Predicate,ChangedIDs) :- !,
    before_after_predicate_no_equalities(Body,BABody,ChangedIDs),
    create_implication(Pre,BABody,Predicate).
before_after_predicate_no_equalities_aux(assertion(Pre,Body),Predicate,ChangedIDs) :-
    !,
    before_after_predicate_no_equalities(Body,BABody,ChangedIDs),
    create_implication(Pre,BABody,Predicate1),
    create_negation(Pre,NotPre),
    add_missing_equalities(ChangedIDs,b(truth,pred,[]),NotPrePredicate),
    create_implication(NotPre,NotPrePredicate,Predicate2),
    conjunct_predicates([Predicate1,Predicate2],Predicate).
before_after_predicate_no_equalities_aux(select_when(Pre,Body),Predicate,ChangedIDs) :- !,
    before_after_predicate_no_equalities(Body,BABody,ChangedIDs),
    conjunct_predicates([Pre,BABody],Predicate).
before_after_predicate_no_equalities_aux(select(Whens),Predicate,ChangedIDs) :- !,
    maplist(before_after_predicate_no_equalities,Whens,WhensPredicates,Changes),
    append(Changes,ChangedIDs),
    maplist(add_missing_equalities(ChangedIDs),WhensPredicates,WhensPredicatesWithEqualities),
    disjunct_predicates(WhensPredicatesWithEqualities,Predicate).
before_after_predicate_no_equalities_aux(select(Whens,Else),Predicate,ChangedIDs) :- !,
    maplist(before_after_predicate_no_equalities,Whens,WhensPredicates,Changes),
    maplist(add_missing_equalities(ChangedIDs),WhensPredicates,WhensPredicatesWithEqualities),
    maplist(get_select_conditions,Whens,SelectConditions),
    disjunct_predicates(SelectConditions,ElseTemp),
    create_negation(ElseTemp,ElseLHS),
    before_after_predicate_no_equalities(Else,ElseRHS,ElseChangedIDs),
    append([ElseChangedIDs|Changes],ChangedIDs),
    conjunct_predicates([ElseLHS,ElseRHS],ElsePred),
    add_missing_equalities(ChangedIDs,ElsePred,ElsePredWithEqualities),
    disjunct_predicates([ElsePredWithEqualities|WhensPredicatesWithEqualities],Predicate).
% TODO: case, handle while
before_after_predicate_no_equalities_aux(sequence(ListOfSteps),Predicate,ChangedIDs) :- !,
    ba_predicate_sequence(ListOfSteps,Predicate,ChangedIDs).
before_after_predicate_no_equalities_aux(if(IfList),Predicate,ChangedIDs) :- !,
    ba_predicate_if(IfList,Predicate,ChangedIDs).
before_after_predicate_no_equalities_aux(let(P,D,B),Predicate,ChangedIDs) :- !,
    before_after_predicate_no_equalities_aux(any(P,D,B),Predicate,ChangedIDs).
before_after_predicate_no_equalities_aux(any(Parameters,PreCond,Body),Predicate,ChangedIDs) :- !,
    before_after_predicate_no_equalities(Body,BABody,ChangedIDs),
    conjunct_predicates([PreCond,BABody],Inner),
    create_exists(Parameters,Inner,Predicate).
before_after_predicate_no_equalities_aux(parallel(P),Predicate,ChangedIDs) :- !,
    maplist(before_after_predicate_no_equalities,P,Preds,Changes),
    append(Changes,ChangedIDs),
    conjunct_predicates(Preds,Predicate).
before_after_predicate_no_equalities_aux(skip,Predicate,[]) :- !,
    Predicate = b(truth,pred,[]).
before_after_predicate_no_equalities_aux(var(Parameters,Body),Predicate,ChangedIDs) :- !,
    before_after_predicate_no_equalities(Body,BABody,ChangedIDs),
    create_exists(Parameters,BABody,Predicate).
before_after_predicate_no_equalities_aux(rlevent(_Name,_Section,_Stat,_Parameters,Guards,_Thm,Actions,_VWit,_PWit,_Unmod,_AbsEvents),Predicate,ChangedIDs) :- !,
    maplist(before_after_predicate_no_equalities,Actions,BAActions,Changes),
    append(Changes,ChangedIDs),
    conjunct_predicates([Guards|BAActions],Predicate). % should probably be an implication? see precondition.
before_after_predicate_no_equalities_aux(operation_call(OpId,[],[]),Predicate,ChangedIDs) :- !,
    get_texpr_id(OpId,op(OpName)),
    b_get_machine_operation(OpName,[],[],OpBody),
    before_after_predicate_no_equalities(OpBody,Predicate,ChangedIDs).
before_after_predicate_no_equalities_aux(Expr,Predicate,ChangedIDs) :-
    add_error_fail(before_after_predicate,'before-after-predicate implementation missing: ',[Expr,Predicate,ChangedIDs]).

get_select_conditions(b(select_when(Condition,_),_,_),Condition).

ba_predicate_equal(LHS,RHS,Res) :-
    % LHS := RHS   trace,
    get_override(LHS,RHS,TPred), !,
    %nl,translate:print_bexpr(TPred),nl,
    Res=TPred.

:- use_module(probsrc(error_manager),[add_internal_error/2]).
% get override Term for f(X)(Y)... := RHS
get_override(b(function(FUN,ARG),Type,_Info),ExprSoFar,Res) :- !, % FUN(ARG) := RHS
    %print(fun(FUN)),nl, translate:print_bexpr(ExprSoFar),nl,
    create_couple(ARG,ExprSoFar,Update), % ARG|-> ExprSoFar
    safe_create_texpr(set_extension([Update]),Type,UpdateSET), % {ARG|->ExprSoFar}
    construct_override(FUN,UpdateSET,Override), % FUN <+ {ARG |-> {...}}
    get_override(FUN,Override,Res).  % Pass Override as new RHS in case FUN itself is a function application
get_override(FUNID,OverrideExpr,TPred) :-
    get_texpr_id(FUNID,_ID), %print(base(ID)),nl,
    prime_variable(FUNID,FUNPrimeId),
    safe_create_texpr(equal(FUNPrimeId,OverrideExpr),pred,TPred). % FUN' = FUN <+ {ARG |-> RHS}
get_override(FUNID,OverrideExpr,TPred) :-
   add_internal_error('Cannot compute override: ',get_override(FUNID,OverrideExpr,TPred)),fail.

construct_override(FunID,Update,Override) :-
    get_texpr_type(Update,Type),
    Override = b(overwrite(FunID,Update),Type,[]).

ba_predicate_member(Set,Id,TPred) :-
    % prime identifiers on the lhs
    prime_variable(Id,PrimedId),
    safe_create_texpr(member(PrimedId,Set),pred,TPred).

ba_predicate_becomes_such(Ids,Pred,Predicate) :-
    list_renamings_to_primed_vars(Ids,RenamingList),
    rename_bt(Pred,RenamingList,Predicate).

list_renamings_to_primed_vars([],[]).
list_renamings_to_primed_vars([Var|Vars],[rename(VarId,PrimedId)|T]) :-
    get_texpr_id(Var,VarId),
    atom_concat(VarId,'\'',PrimedId),
    list_renamings_to_primed_vars(Vars,T).

add_missing_equalities(AllVariables,PredWithoutEqualities,PredWithEqualities) :-
    find_typed_identifier_uses(PredWithoutEqualities,UsedIdentifiers),
    exclude(primed_var_is_used(UsedIdentifiers),AllVariables,VariablesNotOccuringPrimed),
    maplist(ba_predicate_equal,VariablesNotOccuringPrimed,VariablesNotOccuringPrimed,EqualityPredicates),
    conjunct_predicates([PredWithoutEqualities|EqualityPredicates],PredWithEqualities).

primed_var_is_used(UsedIdentifiers,Identifier) :-
    prime_variable(Identifier,PrimedIdentifier),
    remove_all_infos(PrimedIdentifier,PrimedIdentifierWithoutInfos),
    member(PrimedIdentifierWithoutInfos,UsedIdentifiers).

add_else_if_missing([],[DefaultElse]) :-
    safe_create_texpr(if_elsif(b(truth,pred,[]),b(skip,subst,[])),subst,DefaultElse).
add_else_if_missing([If|Ifs],[If|Ifs]) :-
    get_texpr_expr(If,if_elsif(Test,_)),
    is_truth(Test), !.
add_else_if_missing([If|Ifs],[If|MoreIfs]) :-
    add_else_if_missing(Ifs,MoreIfs).

ba_predicate_if(IF,BA,AllChangedIDs) :-
    add_else_if_missing(IF,IFWithElse),
    maplist(ba_predicate_if_aux,IFWithElse,BAPreds,ChangedIDs),
    append(ChangedIDs,AllChangedIDs),
    ba_predicate_if(IFWithElse,BAPreds,AllChangedIDs,BA).

ba_predicate_if([IF|IFs],[BAPred|BAPreds],IDs,BA) :-
    get_texpr_expr(IF,if_elsif(Test,_)),
    % if-part
    add_missing_equalities(IDs,BAPred,InnerBAPredIf),
    (is_truth(Test)
    -> InnerBAPredIf = BAPredIf
    ;  safe_create_texpr(implication(Test,InnerBAPredIf),pred,BAPredIf)),
    % if there are further else/elseif parts, create pos
    (ba_predicate_if(IFs,BAPreds,IDs,ElsePred)
    -> create_negation(Test,NegTest),
       safe_create_texpr(implication(NegTest,ElsePred),pred,BAPredElse),
       conjunct_predicates([BAPredIf,BAPredElse],BA)
    ;  BA = BAPredIf).

ba_predicate_if_aux(IF,BAPred,ChangedIDs) :-
    get_texpr_expr(IF,if_elsif(_,Body)),
    before_after_predicate_no_equalities(Body,BAPred,ChangedIDs).

ba_predicate_sequence([Last],Predicate,ChangedIDs) :- !,
    before_after_predicate_no_equalities(Last,Predicate,ChangedIDs).
ba_predicate_sequence([First|SomeMore],Predicate,ChangedIDs) :-
    before_after_predicate_no_equalities(First,FirstPred,ChangesFirst),
    find_typed_identifier_uses(FirstPred,Identifiers),
    include(is_primed,Identifiers,PrimedIdentifiers),
    list_renamings_to_primed_vars(PrimedIdentifiers,RenamingToTwicePrimedIDs),
    rename_bt(FirstPred,RenamingToTwicePrimedIDs,FirstPredAfterReplacement),

    ba_predicate_sequence(SomeMore,SecondPred,FurtherChanges),
    list_renamings_remove_prime(PrimedIdentifiers,RenamingToNotPrimed),
    rename_bt(SecondPred,RenamingToNotPrimed,SecondPredAfterReplacement),

    conjunct_predicates([FirstPredAfterReplacement,SecondPredAfterReplacement],FullPred),

    prime_variables(PrimedIdentifiers,TwicePrimedIdentifiers),
    create_exists(TwicePrimedIdentifiers,FullPred,Predicate),
    append(ChangesFirst,FurtherChanges,ChangedIDs).

is_primed(TExpr) :-
    get_texpr_id(TExpr,ID),
    is_primed_id(ID).

list_renamings_remove_prime([],[]).
list_renamings_remove_prime([Var|Vars],[rename(VarId,NonPrimedID)|T]) :-
        get_texpr_id(Var,VarId),
        atom_concat(NonPrimedID,'\'',VarId),
        list_renamings_remove_prime(Vars,T).
