% (c) 2009-2020 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(bmachine_structure,[ create_machine/2           % create a new, empty machine
                             , get_machine_name/2 % get the name of the machine
                             , valid_section/1 % checks if the given section exists
                             , get_section/3 % get the content of a section
                             , write_section/4 % overwrite the content of a section
                             , select_section/5 % modify the content of a section
                             , get_section_of_machines/3 % get one section for multiple machines
                             , get_section_texprs/3 % get all typed expressions of a section
                             , select_section_texprs/5 % modify the typed expressions of a section
                             , append_to_section/4 % append a list to a section
                             , conjunct_to_section/4 % write the conjunction of the new and old value
                             ]).

:- use_module(tools,[is_list_simple/1]).
:- use_module(bsyntaxtree).

:- use_module(module_information,[module_info/2]).
:- module_info(group,typechecker).
:- module_info(description,'Here is the data structure defined that holds a complete B machine.').

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% access to machines

% creates an empty machine
create_machine(Name,M) :-
    create_texpr(truth,pred,[initial],Truth),
    M = machine(Name,
                [deferred_sets/[], enumerated_sets/[], enumerated_elements/[],
                 parameters/[], internal_parameters/[],
                 abstract_constants/[], concrete_constants/[],
                 abstract_variables/[], concrete_variables/[],
                 promoted/[], unpromoted/[],

                 constraints/Truth,
                 properties/Truth,
                 invariant/Truth,
                 linking_invariant/Truth,
                 assertions/[],

                 initialisation/[],
                 operation_bodies/[],
                 definitions/[],
                 used/[],
                 freetypes/[],
                 operators/[],
                 values/[],

                 meta/[]
                ]).

get_machine_name(machine(Name,_),Name).

% checks if the given secion exists
valid_section(Sec) :-
    create_machine(_,machine(_,List)), member(Sec/_,List),!.

% get the content of a section
get_section(Sec,machine(_,List),Content) :-
    memberchk(Sec/C,List),C=Content.

% overwrite the content of a section
write_section(Sec,Content,Old,New) :- select_section(Sec,_,Content,Old,New).

:- use_module(error_manager,[add_internal_error/2]).

% modify the content of a section
select_section(Sec,OldContent,NewContent,machine(Name,OldM),machine(Name,NewM)) :-
    select_section2(OldM,NewM,Sec,O,N),!,
    O=OldContent,N=NewContent.
select_section(Sec,Old,_,M,NewM) :-
    add_internal_error('Section does not exist:',select_section(Sec,Old,_,M,NewM)),
    Old=[],NewM=M.
select_section2([Sec/C1|R1],[Sec/C2|R2],Select,O,N) :-
    ( Sec==Select -> O=C1, N=C2, R1=R2
    ; otherwise -> C1=C2, select_section2(R1,R2,Select,O,N)).

% append the content to a section (section content must be a list)
append_to_section(Sec,Content,Old,New) :-
    select_section(Sec,Before,App,Old,New),
    append(Before,Content,App).

% append the content to a section (section content must be a list)
conjunct_to_section(Sec,Content,Old,New) :-
    select_section(Sec,Before,App,Old,New),
    conjunct_predicates([Before,Content],App).

% get the content of a section
% get_section_of_machines(List of Machine,Name of Section,List of Content)
get_section_of_machines([],_,[]).
get_section_of_machines([Machine|MRest],Section,[Content|CRest]) :-
    get_section(Section,Machine,Content),
    get_section_of_machines(MRest,Section,CRest).
%get_sections([],_,[]).
%get_sections([Sec|Rest],M,[Content|CRest]) :-
%    get_section(Sec,M,Content),get_sections(Rest,M,CRest).

% get all typed expressions of a section
get_section_texprs(Sec,Machine,TExprs) :-
    get_section(Sec,Machine,Content),
    extract_sec_texprs(Sec,Content,TExprs).
extract_sec_texprs(Sec,C,C) :- has_texpr_l(Sec).
extract_sec_texprs(Sec,C,[C]) :- has_texpr(Sec).
extract_sec_texprs(initialisation,C,I) :-
    ( is_texpr(C)       -> I=[C]
    ; is_list_simple(C) -> findall(Subst,member(init(_,Subst),C),I)).

% modify the typed expressions of a section
select_section_texprs(Sec,OldTExprs,NewTExprs,OldMachine,NewMachine) :-
    select_section(Sec,OldContent,NewContent,OldMachine,NewMachine),
    change_sec_texprs(Sec,OldContent,OldTExprs,NewTExprs,NewContent).
change_sec_texprs(Sec,O,O,N,N) :- has_texpr_l(Sec),!.
change_sec_texprs(Sec,O,[O],[N],N) :- has_texpr(Sec),!.
change_sec_texprs(initialisation,O,OI,NI,N) :-
    ( is_list_simple(O)
     -> change_init(O,OI,NI,N)
      ; otherwise -> OI=[O], NI=[N]).

change_init([],[],[],[]).
change_init([init(Name,O)|IOrest],[O|Orest],[N|Nrest],[init(Name,N)|INrest]) :-
    change_init(IOrest,Orest,Nrest,INrest).


has_texpr_l(Sec) :-
    memberchk(Sec,[deferred_sets, enumerated_sets, enumerated_elements,
                   parameters, internal_parameters, abstract_constants, concrete_constants,
                   abstract_variables, concrete_variables,
                   promoted, unpromoted, assertions,
                   operation_bodies,values]).
has_texpr(Sec) :- memberchk(Sec,[constraints, properties, invariant]).
