% (c) 2009-2020 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(bool_pred,[free_var/1, bool_negate_check/2,
                    negate/2, bool_equality/3]).

/*
 a module that supports a boolean type: pred_false, pred_true
 with an efficient negation procedure
*/

:- use_module(tools).
:- use_module(self_check).
:- use_module(error_manager).

:- use_module(module_information,[module_info/2]).
:- module_info(group,kernel).
:- module_info(description,'This module provides reified equality and negation for pred_true/pred_false.').

:- use_module(library(atts)).


:- attribute bool_pred_info/2.
% store for each bool value the negated variable & list of reifications it is involved in
% bool_pred_info(NegationOfVariable,  List of eqcheck(OtherVariable,ReificationVariable)

% check that we have a real variable without any bool_pred constraints attached
free_var(X) :- var(X), \+ get_atts(X,+bool_pred_info(_,_)).

% check if currently X is negation of Y; call only for variable X
bool_negate_check(X,Y) :- get_atts(X,+bool_pred_info(OtherX,_)),Y==OtherX.

negate(X,Y) :-  %enter_log(negate(X,Y)),
 (nonvar(X) -> negate3(X,Y)
  ; nonvar(Y) -> negate3(Y,X)
  ; get_atts(X,+bool_pred_info(OtherX,_)) -> Y=OtherX, X \== Y, update_bool_pred_negation(Y,X)
  ; get_atts(Y,+bool_pred_info(OtherY,_)) -> put_atts(X,+bool_pred_info(Y,[])), X=OtherY, X \== Y
  ; X \== Y, put_atts(Y,+bool_pred_info(X,[])), put_atts(X,+bool_pred_info(Y,[]))
 ).% ,exit_log(negate(X,Y)).

update_bool_pred_negation(X,_NegX) :- nonvar(X),!.
update_bool_pred_negation(X,NegX) :-
   (get_atts(X,+bool_pred_info(OtherX,_)) -> OtherX=NegX
      ; put_atts(X,+bool_pred_info(NegX,[]))).

%negate3(X,Y) :- print(negate3(X,Y)),nl,fail.
negate3(pred_true,pred_false). % for kernel_equality,... results
negate3(pred_false,pred_true).
%negate3(bool_true,bool_false). % for boolean values
%negate3(bool_false,bool_true).

% verify_attributes is called whenever a variable Var that might have attributes is about to be bound to Value (it might have none).
% verify(VariableBeforeUnification, ValueWithWhichVariableUnifies,  GoalWhichSICStusShouldCallAfterUnif)
verify_attributes(ThisVar,Value,Goal) :- get_atts(ThisVar,+bool_pred_info(NegThis,ThisReifList)),!,
    % enter_log(ThisVar=Value), %% print(verify(ThisVar,Value,NegThis,ThisReifList)),nl, %%
   (var(Value) -> (get_atts(Value,+bool_pred_info(NegValue,ValReifList))
                   -> Goal=[NegThis=NegValue,ThisVar\==NegThis,EqGoal1,EqGoal2],
                      check_equalities(ThisReifList,Value,NegValue,EqGoal1,NewReifList,RR),
                      check_equalities(ValReifList,ThisVar,NegThis,EqGoal2,RR,[]),
                      (nonvar(Value) -> print(verify_attribute_instantiated_value(Value)),nl /* check equalities could have forced value to be known now; should not happen */
                       ; put_atts(Value,+bool_pred_info(NegValue,NewReifList)))
                   ;  put_atts(Value,+bool_pred_info(NegThis,ThisReifList)), Goal=[]
                   )
     ; (ThisReifList=[]
         -> Goal=[negate3(Value,NegThis)] %(var(NegThis) -> Goal=[negate3(Value,NegThis)] ; Goal=[])
         ;  Goal=[negate3(Value,NegThis),check_inst_equalities(ThisReifList,Value)])
   ). %,  exit_log(Goal),trace. %, print(done_verify(ThisVar,Value,Goal)),nl.
verify_attributes(_, _, [] ).


:- assert_must_succeed((bool_pred:enter_log(a),bool_pred:enter_log(b),
                        bool_pred:exit_log(b),bool_pred:exit_log(a))).
% a simple logging facility to replay calls later
:- dynamic logging_level/1.
logging_level(0).

enter_log(X) :- retract(logging_level(L)),
    indent_ws(L),print(X), print(','),nl,
    L1 is L+1, assert(logging_level(L1)).
exit_log(G) :- retract(logging_level(L)),
  (L>0 -> L1 is L-1, assert(logging_level(L1)), indent_ws(L1), print('exit '), print(G),nl
    ; add_internal_error('*** negative - logging level !',L),
      assert(logging_level(L))).
indent_ws(X) :- X<1,!.
indent_ws(X) :- print('  '), X1 is X-1, indent_ws(X1).


% check if any reification can now be determined
% Idea: add other boolean operators here as well: imply, and, or ??
check_equalities([],_,_,true) --> [].
check_equalities([eqcheck(V2,Res)|T],V1,NegV1,Goal) -->
   (  {V2==V1}    -> {Goal=(Res=pred_true,G2)}
    ; {V2==NegV1} -> {Goal=(Res=pred_false,G2)}
    ; {V1==Res} -> {Goal=(V2=pred_true,G2)}  % Warning: V2 could expect bool_true !
    ; {NegV1==Res} -> {Goal=(V2=pred_false,G2)}  % Warning: V2 could expect bool_false !
    ; %{print(no_new_info),nl},
      [eqcheck(V2,Res)], {Goal=G2}
   ),check_equalities(T,V1,NegV1,G2).

% will be called with second argument instantiated to pred_false/pred_true

:- public check_inst_equalities/2. % used above
check_inst_equalities([],_).
check_inst_equalities([eqcheck(V2,Res)|T],V1) :-
   (nonvar(V2) -> (V1=V2 -> Res=pred_true ; Res=pred_false)
     ; V1=pred_true -> Res=V2
     ; negate(V2,Res)
   ), check_inst_equalities(T,V1).

:- assert_must_succeed((bool_pred:negate(X,Y), bool_pred:negate(X,Z), Y==Z, Z=pred_true, X==pred_false)).
% bool_pred:negate(X,Y), bool_pred:negate(X,Z), Y==Z, Z=pred_true, X==pred_false.
/*
 CLP(FD) cannot do this:
 | ?- X in 0..1, Y in 0..1, Z in 0..1, X #\= Y, X#\=Z.
X in 0..1,
Y in 0..1,
Z in 0..1 ?
yes
*/

:- assert_must_succeed(( bool_pred:negate(X,Y), bool_pred:negate(Y,Z), X==Z, Z=pred_true, X==pred_true, Y==pred_false)).
:- assert_must_succeed((bool_pred:negate(X,Y), bool_pred:negate(X2,Y2), X=X2, Y==Y2, Y=pred_false, X2==pred_true)).
:- assert_must_succeed((bool_pred:negate(X,Y), bool_pred:negate(X,Z), Y==Z, Z=pred_true, X==pred_false)).
:- assert_must_succeed((bool_pred:negate(X,Y), bool_pred:negate(Y,Z), \+ bool_pred:negate(Z,X))).
:- assert_must_succeed((bool_pred:negate(X,Y), bool_pred:negate(Y,Z), bool_pred:negate(Z,X2), \+ X2=X)).
:- assert_must_succeed((bool_pred:negate(X,Y), bool_pred:negate(Y,Z), X2=X, \+ bool_pred:negate(Z,X2))).
:- assert_must_fail(bool_pred:negate(X,X)).
:- assert_must_succeed(( bool_pred:negate(X,Y), bool_pred:negate(V,Z), bool_pred:bool_equality(X,V,RXV), Z=Y, X==V, RXV==pred_true)).
:- assert_must_succeed(( bool_pred:negate(X,Y), bool_pred:negate(V,Z), bool_pred:bool_equality(X,V,RXV), Z=X, Y==V, RXV==pred_false)).
:- assert_must_succeed(( bool_pred:negate(X,Y), bool_pred:negate(V,Z), bool_pred:bool_equality(X,V,RXV), RXV=pred_false, Y==V, X==Z)).
:- assert_must_succeed(( bool_pred:negate(X,Y), bool_pred:negate(V,Z), bool_pred:bool_equality(X,V,RXV), RXV=pred_true, Y==Z, X==V)).
:- assert_must_succeed((Y=pred_true, bool_pred:bool_equality(X,Y,R) , R==X)).
:- assert_must_fail((Y=pred_false, bool_pred:bool_equality(X,Y,R) , R=X)).


%:- attribute bool_equality/2.
% bool_equality is actually solving: (X<=>Y) <=> R
% Note this is commutative and associative in its three arguments
bool_equality(X,Y,Res) :- %print(bool_equality(X,Y,Res)),nl, %% enter_log(bool_equality(X,Y,Res)), %%
    (  nonvar(X) -> eq1(X,Y,Res)
     ; nonvar(Y) -> eq1(Y,X,Res)
     ; nonvar(Res) -> eq1(Res,X,Y)
     ; X==Y -> Res=pred_true
     ; X==Res -> Y=pred_true  % Warning: Y could expect bool_true !
     ; Y==Res -> X=pred_true  % Warning: X could expect bool_true !
     ; get_atts(X,+bool_pred_info(NX,XList))
        -> (NX==Y -> Res=pred_false % Y is the negation of X
             ; NX==Res -> Y=pred_false % Warning: Y could expect bool_false !
             ; put_atts(X,+bool_pred_info(NX,[eqcheck(Y,Res)|XList])),
               add_equality(Y,X,Res)
            ),
            blocking_force_eq3(Res,X,Y)
     ; add_new_equality(X,Y,Res),add_equality(Y,X,Res),
       blocking_force_eq3(Res,X,Y)
     ). % ,exit_log(eq(X,Y,Res)).

% Missing rules (requires merging bool_true/pred_true, ...)
%                X==Res -> Y=pred_true  &&  Y==Res -> X=pred_true
%                X==~Res -> Y=pred_false && Y==~Res -> X=pred_false [Missing above]

add_equality(X,Y,Res) :-
  (get_atts(X,+bool_pred_info(NX,List))
    -> put_atts(X,+bool_pred_info(NX,[eqcheck(Y,Res)|List]))
    ;  add_new_equality(X,Y,Res)
   ).
add_new_equality(X,Y,Res) :-
  put_atts(NX,+bool_pred_info(X,[])),% NX is a fresh variable, no need to go through negate/2
  put_atts(X,+bool_pred_info(NX,[eqcheck(Y,Res)])).

% X -> bool_equality(Y,Res) : check if X==Y  -> Res=pred_true or if Y==NegatedX -> Res=pred_false
% we could also check if Res==X -> Y=pred_true


eq1(pred_true,X,X).
eq1(pred_false,X,Y) :- negate(X,Y).

% equality if one of the bools (second arg) is known: wait for other bool (arg1) or result
%:- block eq2(-,?,-).
%eq2(X,Y,Res) :- nonvar(Res),!,eq1(Res,X,Y).
%eq2(X,Y,Res) :- (X=Y -> Res=pred_true ; Res=pred_false).

%:- block blocking_force_eq(-,?,?). % when((?=(Res,X) ; nonvar(Res)),  --> we could detect force_eq(A,A,pred_true/false)
%blocking_force_eq(Res,X,Y) :- force_eq(Res,X,Y).

blocking_force_eq3(Res,X,Y) :-
   when((?=(Res,X) ; ?=(Res,Y) ; ?=(X,Y) ; nonvar(Res) ; nonvar(X) ; nonvar(Y)), bool_equality(X,Y,Res)).

%blocking_force_eq2(Res,X,Y) :- when((?=(Res,X) ; ?=(Res,Y) ; ?=(X,Y) ; nonvar(Res)), force_eq2(Res,X,Y)).
%
%force_eq2(Res,X,Y) :- X==Res,!, %print(forcing_eq_eq1(Res,X,Y)),nl,
%                      Y=pred_true.
%force_eq2(Res,X,Y) :- Y==Res,!, %print(forcing_eq_eq2(Res,X,Y)),nl,
%                      X=pred_true.
%force_eq2(Res,X,Y) :- X==Y,!, %aprint(forcing_eq_eq3(Res,X,Y)),nl,
%                      Res=pred_true.
%force_eq2(pred_true,X,X).
%force_eq2(pred_false,X,Y) :- negate(X,Y).

