% (c) 2009-2020 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(chr_integer_inequality,[chr_leq/2, chr_lt/2, chr_eq/2, chr_eq/3, chr_neq/2, chr_in_interval/4,
                                  enable_complex_rules/0, attach_chr_integer_inequality_to_wf_store/1]).

:- use_module(probsrc(module_information),[module_info/2]).
:- module_info(group,kernel).
:- module_info(description,'Provide interface to CHR integer reasoning.').

:- use_module(library(clpfd)).

:- use_module(library(chr)).
:- chr_option(check_guard_bindings,off).
:- chr_option(optimize,full).
:- chr_option(debug,off).
%:- chr_option(generate_debug_info,true). % does not work

% restore operators overwritten by chr library
% otherwise, unary operators might not be parseable
:- op(200, fy, [+,-]).
% :- op(500, yfx, \).
:- op(0,yfx,'#').


:- use_module(probsrc(self_check), [assert_must_succeed/1, assert_must_fail/1]).
:- use_module(probsrc(kernel_waitflags), [get_enumeration_starting_wait_flag/3]).
:- use_module(probsrc(preferences), [get_preference/2]).
:- use_module(probsrc(debug), [debug_println/2]).

:- assert_must_succeed(( chr_leq(X,Y), chr_leq(Y,Z), Z=X, Y==X, Y==Z, X=5)).
:- assert_must_succeed(( chr_leq(X,Y), chr_leq(Y,Z), chr_leq(Z,X), Y==X, Y==Z, X=5)).
:- assert_must_fail(( chr_leq(X,Y), chr_lt(Y,X))).
:- assert_must_succeed(( chr_leq(X,Y), X=5, Y=7)).
:- assert_must_succeed(( chr_leq(X,Y), chr_leq(Y,X), X==Y, X=5)).
:- assert_must_fail(( chr_lt(X,Y), chr_eq(X,Y) )).
:- assert_must_fail(( chr_leq(X,Y), chr_neq(X,Y), chr_leq(Y,X) )).
:- assert_must_succeed(( chr_leq(X*X,4), chr_leq(4,X*X), 0 #< X, X==2)).
% does not work on sicstus 4.3 as #= is handled by a propagator rather than unification
% :- assert_must_succeed(( chr_eq(X,Y), chr_leq(4,Y*Y), chr_leq(X*X,4), %X #= Y, 0 #< X, X==2)).
% :- assert_must_succeed(( chr_eq(X,Y), chr_leq(4,2*Y), chr_leq(X*X,4),
%                         X #= Y, 0 #< X, X==2)).
:- assert_must_succeed((X*X #=< Y*Y, chr_leq(X*X,Y*Y),
                        4 #=< X*X, chr_leq(4,X*X),
                        Y=2, 0 #< X, X==2)).
% This is hard to infer by an easy rule
%:- assert_must_succeed((post_constraint2( X*X #=< Y*Y, _ ), chr_leq(X*X,Y*Y),
%                        post_constraint2( Y*Y #=< X*X, _), chr_leq(X*X,Y*Y),
%                        X==Y)).
:- assert_must_fail(( chr_lt(X*X,Y*Y), chr_lt(Y*Y,X*X) )).



% assumes we use CLP(FD) to complement it

chr_leq(X,Y) :-
    simplify_expr(X,SX), simplify_expr(Y,SY), %print(chr_leq(X,Y,SX,SY)),nl,
    leq(SX,SY).
chr_lt(X,Y) :-
    simplify_expr(X,SX), simplify_expr(Y,SY), %print(chr_lt(X,Y,SX,SY)),nl,
    lt(SX,SY).

:- use_module(probsrc(clpfd_interface),[computable_arith_expression/1]).
simplify_expr(E,R) :-
  % TO DO: avoid redoing the same check in clpfd_interface after doing CHR propagation
  %        pass simplification result back to caller ??!
  (computable_arith_expression(E) -> R is E ; R=E).

chr_eq(X,Y) :- is_idl(X,Y,A,B,C),!,idl(A,B,C). %TODO: should we set up chr_eq/3 in addition to idl/3?
chr_eq(X,Y) :- is_idl(Y,X,A,B,C),!,idl(A,B,C). % Alternative: introduce on demand out of chr_eq/3?
chr_eq(X,Y) :-
     chr_eq(X,Y,pred_true).


is_idl(X,Y, A,B,C) :-  % A = B+C
    simple(X),
    %var(X), % should we also allow numbers ?  100 = x+y ?
    nonvar(Y),
    (Y = -(A,B) -> simple(A),simple(B),C=X % , print(idl(X,Y)),nl.
     ; Y = +(B,C) -> simple(B),simple(C), A=X). %,print(is_idl(X,Y)),print_idl(A,B,C).


chr_eq(X,Y,R) :- simplify_expr(X,SX), simplify_expr(Y,SY),  % print(chr_eq(SX,SY,R)),nl,
   eq(SX,SY,R),
   eq(SY,SX,R). % commutativity, add constraint the other way around
chr_neq(X,Y) :- chr_eq(X,Y,pred_false), chr_eq(Y,X,pred_false).

:- public setup_eq/2.
setup_eq(X,Y) :-  % print(setup_eq(X,Y)),nl,
  (var(X), ground(Y)    -> X is Y ;
   var(Y), ground(X)    -> Y is X ;
   var(X), var(Y)       -> X = Y ;
   fd_var(X), fd_var(Y) -> chr_eq(X,Y), call(X #= Y) ;
   ground(X), ground(Y) -> true ; % X =:= Y check is done by clpfd
   otherwise            -> chr_eq(X,Y),
                           call(X #= Y)). % the call is necessary; otherwise we get an error for complex arguments such as: "expected an integer, but found _109*_109"

:- public setup_lt/2.
setup_lt(X,Y) :- number(X),number(Y),!, X<Y.
setup_lt(X,Y) :- %print(lt(X,Y)),nl,
  lt(X,Y),
  call_dif_when_necesssary(X,Y), % ensure that frozen_dif detects these variables to be different
  call(X #< Y). % Do we need those ?? Isn't CLPFD doing this automatically ? TO DO : check
  %post_constraint2(X #< Y, _Posted).


% set up a dif co-routine only if necessary; assumption: E1,E2 are either number or variable
call_dif_when_necesssary(E1,E2) :- (number(E1) ; number(E2)),!. % nothing to be gained by posting dif
call_dif_when_necesssary(E1,E2) :-
   dif(E1,E2). % ensure that frozen_dif detects these variables to be different

:- public setup_leq/2.
setup_leq(X,Y) :- number(X),number(Y),!, X=<Y.
setup_leq(X,Y) :- %print(leq(X,Y)),nl,
  leq(X,Y),
  call(X #=< Y).
  %post_constraint2(X #=< Y, _Posted).

% this constraints can be added to the store to enable complex
% propagation rules if nessasary
% (i.e. before enumerating infinite domains)
:- chr_constraint enable_idl_rules/0, disable_idl_rules/0.
enable_complex_rules :- % print('ENABLE IDL RULES'),nl,nl,
                        enable_idl_rules.
%enable_complex_rules :- print('BACKTRACK ENABLE IDL'),nl,fail.

disable_complex_rules :- disable_idl_rules.

%enable1 @ enable_idl_rules ==> format('~nEnabling CHR-IDL~n~n',[]).
disable1 @ disable_idl_rules \ enable_idl_rules <=> true. % ,format('~nDisabling CHR-IDL~n~n',[]).
rm_disable @ disable_idl_rules <=> true, debug_println(4,disabled_chr_complex_rules).

attach_chr_integer_inequality_to_wf_store(WF) :-
    get_preference(use_chr_solver,true), !,
    disable_complex_rules,
    get_enumeration_starting_wait_flag(chr_integer_inequality,WF,LWF),
    enable_complex_rules_block(WF,LWF).
attach_chr_integer_inequality_to_wf_store(_).

:- block enable_complex_rules_block(?,-).
enable_complex_rules_block(_WF,_) :- % kernel_waitflags:portray_waitflags(_WF),
    enable_complex_rules.

:- chr_constraint leq/2, lt/2.
:- chr_constraint eq/3.

:- public '@'/2.
reflexivity  @ leq(X,X) <=> true.
antisymmetry @ leq(X,Y), leq(Y,X) <=> setup_eq(X,Y).
idempotence  @ leq(X,Y) \ leq(X,Y) <=> true.
transitivity @ leq(X,Y), leq(Y,Z) ==> setup_leq(X,Z).

% checking is done by CLP(FD); we could use true instead of test.
checking        @ leq(X,Y) <=> ground(X),ground(Y) | X =< Y.
checking        @ lt(X,Y) <=> ground(X),ground(Y) |  X < Y.
checking        @ eq(X,Y,pred_false) <=> ground(X),ground(Y) |  X \= Y.

% Only handled by CLP(FD) if value is known!
antireflexivity @ lt(X,X) <=> %print(inconsistent(X,X)),nl,
                              fail.
idempotence     @ lt(X,Y) \ lt(X,Y) <=> true.
transitivity    @ lt(X,Y), leq(Y,Z) ==> trans_useful(X,Y,Z) | setup_lt(X,Z).
transitivity    @ leq(X,Y), lt(Y,Z) ==> trans_useful(X,Y,Z) | setup_lt(X,Z).
transitivity    @ lt(X,Y), lt(Y,Z)  ==> trans_useful(X,Y,Z) | setup_lt(X,Z).
plus1rule1      @ leq(A+B,Y)  ==> pos_add1(A,B,X) | %print(plus1_leq(X,Y)),nl,
                                  setup_lt(X,Y). % no need to setup CLPFD; is actually <=> ; Discuss with Sebastian; but we need to set up dif
plus1rule2      @ lt(X,A+B) ==> pos_add1(A,B,Y) | %print(plus1_2_lt(X,Y)),nl,
                                leq(X,Y). % ditto
plus1weak1      @ lt(A+B,Y) ==> pos_add(A,B,X) | %print(plus1_weak_lt(X,Y)),nl,
                                setup_lt(X,Y).
% TO DO: add rules for minus; maybe rewrite using idl constraints

% check if transitivity propagation is useful
trans_useful(A,B,C) :- number(B),!, (number(A) -> fail ; \+ number(C)).
trans_useful(_,_,_).

% --------------------

:- chr_constraint in_interval/4.
% assert X..Y <: V..W  either Y<X or V<=X & Y <= W
chr_in_interval(X,Y,V,W) :-
    simplify_expr(X,SX), simplify_expr(Y,SY), simplify_expr(V,SV), simplify_expr(W,SW),
    %print(post(in(SX,SY,SV,SW))),nl,
    in_interval(SX,SY,SV,SW).

pos_in_interval1 @ leq(X,Y) \ in_interval(X,Y,V,W) <=> setup_leq(V,X), setup_leq(Y,W). % | print(in(X,Y,V,W)),nl.
pos_in_interval2 @ lt(X,Y) \ in_interval(X,Y,V,W) <=> setup_leq(V,X), setup_leq(Y,W). % | print(inlt(X,Y,V,W)),nl.
eq_in_interval @ in_interval(X,X,V,W) <=> setup_leq(V,X), setup_leq(X,W). % | print(inlt(X,X,V,W)),nl.
neg_in_interval @ lt(Y,X) \ in_interval(X,Y,_V,_W) <=>  true. % | print(in_empty(X,Y,V,W)),nl.

% --------------------

/*
leads to loop for test 1403
lt_idl_rule1  @  lt(X,Y), idl(X,Y,Z) ==> not_negative(Z) | setup_lt(0,Z).
lt_idl_rule2  @  lt(X,Y), idl(X,Z,Y) ==> not_negative(Z) | setup_lt(0,Z).
lt_idl_rule3  @  lt(Y,X), idl(X,Y,Z) ==> not_positive(Z) | setup_lt(Z,0).
lt_idl_rule4  @  lt(Y,X), idl(X,Z,Y) ==> not_positive(Z) | setup_lt(Z,0).

leq_idl_rule1  @  leq(X,Y), idl(X,Y,Z) ==> not_definitely_geq(Z,0) | setup_leq(0,Z).
leq_idl_rule2  @  leq(X,Y), idl(X,Z,Y) ==> not_definitely_geq(Z,0) | setup_leq(0,Z).
leq_idl_rule3  @  leq(Y,X), idl(X,Y,Z) ==> not_definitely_leq(Z,0) | setup_leq(Z,0).
leq_idl_rule4  @  leq(Y,X), idl(X,Z,Y) ==> not_definitely_leq(Z,0) | setup_leq(Z,0).


not_positive(V) :- \+ positive(V).
positive(Var) :- clpfd_interface:clpfd_domain(Var,Low,_), number(Low), Low>0.
not_definitely_geq(V,N) :- \+ definitely_geq(V,N).
definitely_geq(Var,Bound) :- clpfd_interface:clpfd_domain(Var,Low,_), number(Low), Low>=Bound.
not_negative(V) :- \+ negative(V).
negative(Var) :- clpfd_interface:clpfd_domain(Var,_,Up), number(Up), Up<0.
not_definitely_leq(V,N) :- \+ definitely_geq(V,N).
definitely_leq(Var,Bound) :- clpfd_interface:clpfd_domain(Var,_,Up), number(Up), Up=<Bound.
*/


% idempotence of equality seems to be costly. remove this rule?
% idempotence     @ eq(X,Y,R) \ eq(X,Y,R) <=> true.
became_equal    @ eq(X,X,R) <=>  R=pred_true.
% same @ eq(X,Y,R1) \ eq(X,Y,R2) <=> print(eq_same(X,Y,R1,R2)),nl,R1=R2. % not really required as R1/2 = pred_false or pred_true, if pred_true will force X=Y

strengthen @ eq(X,Y,pred_false) \ leq(X,Y) <=> setup_lt(X,Y).
% strengthen @ eq(Y,X,pred_false) \ leq(X,Y) <=> setup_lt(X,Y). % not required anymore, we add eq both ways

% this reifies to disequality if lt constraint was generated above
eq_lt_contradiction @ eq(X,Y,R), lt(X,Y) ==> R = pred_false.
%eq_lt_contradiction @ eq(X,Y,R), lt(Y,X) ==> R = pred_false. % not required anymore, we add eq both ways

:- public pos_add/3, pos_add1/3. % Spider does not seem to detect usage above
pos_add(A,B,X) :- positive_constant_addition(A,B,X,_).
pos_add1(A,B,X) :- (B==1 -> X=A ; A==1, X=B).
positive_constant_addition(X,Y,Var,Cst) :-
 (var(X) -> pos_aux(X,Y,Var,Cst)
  ; var(Y) -> pos_aux(Y,X,Var,Cst)).
:- use_module(probsrc(clpfd_interface),[clpfd_check_geq_nr/2]).
pos_aux(X,Y,Var,Cst) :-
  clpfd_check_geq_nr(Y,1), % TO DO: in test 2056 Y is -1+0; ensure that such expressions are evaluated
  Var=X, Cst=Y.

% if an expression is ground, we can calculate it (might not be needed?)
% calculate @ leq(X,Y) ==> compound(X), ground(X) | V is X, leq(V,Y).
% calculate @ leq(X,Y) ==> compound(Y), ground(Y) | V is Y, leq(X,V).

:- chr_constraint idl/3.
:- chr_constraint idl_sym/3.

% now that the idl constraints are only evaluated later on
% some of them might be inferred already.
% remove those!
% if two arguments are ground, clp(fd) inferred the third!
remove_idl_done @ idl(A,B,C) <=> (ground(A) -> (ground(B); ground(C)) ; ground(B),ground(C)) | true. %, print(removed_idl(A,B,C)),nl.

% proof rules for IDL like expressions
%introduce_idl @ eq(V,X-Y,pred_true) <=> idl(V,X,Y).

idempotence1     @ enable_idl_rules, idl(X1,Y,R) \ idl(X2,Y,R) <=> %print(idem1(X1,X2,Y+R)),nl,
                                                                   X1=X2.
idempotence2     @ enable_idl_rules, idl(X1,Y,R) \ idl(X2,R,Y) <=> %print(idem2(X1,X2,Y+R)),nl,
                                                                   X1=X2.

neq_idl1 @    enable_idl_rules, idl(X,V,Y), eq(X,Y,pred_false) ==> clpfd_interface:clpfd_neq_expr(V,0). %eq(V,0,pred_false). %, print(idl1(V,X,Y,EQXY)),nl.
neq_idl2 @    enable_idl_rules, idl(X,Y,V), eq(X,Y,pred_false) ==> clpfd_interface:clpfd_neq_expr(V,0).

%idl_eq_zero1 @  enable_idl_rules \ idl(X,V,X) <=>  print(eq_zero1(V,X)),nl, clpfd_eq_expr(V,0).
%idl_eq_zero2 @  enable_idl_rules \ idl(X,X,V) <=>  print(eq_zero2(V,X)),nl,                                       clpfd_eq_expr(V,0).
idl_eq_zero1 @  enable_idl_rules, idl(X,V,X) ==>  %print(eq_zero1(V,X)),nl,
                                       clpfd_eq_expr(V,0).
idl_eq_zero2 @  enable_idl_rules, idl(X,X,V) ==>  %print(eq_zero2(V,X)),nl,
                                       clpfd_eq_expr(V,0).
% the following rules infer lt from idl, e.g., lt(x,y) in  x:1..3 & y=2+x & v={x,y} & card(v)=r
idl_lt1 @  enable_idl_rules, idl(X,N,Y) ==>  number(N), N>0  %,  print(idl_lt(X,N,Y)),nl
                                          | lt(Y,X).
idl_lt2 @  enable_idl_rules, idl(X,N,Y) ==>  number(N), N<0  %,  print(idl_lt(X,N,Y)),nl
                                          | lt(X,Y).
% add rule

%idl_eq_zero @ eq(V,X-Y,pred_true) \ eq(X,Y,pred_true) <=> true | print(idl_eq_zero(V,X,Y)),nl, clpfd_eq_expr(V,0).

:- use_module(probsrc(clpfd_interface),[clpfd_eq_expr/2, clpfd_eq_expr_optimized/2]).
%finite_idl1  @ idl(X0,V2,X1) \ enable_idl_rules <=> finite(V2) | idl_finite(X0,V2,X1), print(finite1(X0,V2,X1)),nl.
%finite_idl2  @ idl(X0,X1,V2) \ enable_idl_rules <=> finite(V2) | idl_finite(X0,V2,X1), print(finite2(X0,V2,X1)),nl.

% generate an idl target goal for deriving new constraints about infinite variables
gen_idl_sym_target1     @ enable_idl_rules, idl(A,B,C)  ==> %print('Check: '),print_idl(A,B,C),
          infinite(B)
         | %print(target1(B,(A,B,C))),nl,
           idl_sym(A,B,C).
gen_idl_sym_target2     @ enable_idl_rules, idl(A,C,B)  ==>
         infinite(B) | %print(target2(B,(A,B,C))),nl,
                       idl_sym(A,B,C).

remove_idl_sym @ idl_sym(_,V,_) <=> finite(V) | true.
% try and infer new constraints for infinite variables
transitivity_idl1  @  enable_idl_rules, idl_sym(X0,V2,X1), idl(X0,V0,X2), idl(X1,V1,X2) ==>
         infinite(V2) | %print(idl3(V2,V0-V1)),nl,
                        setup_eq(V2,V0-V1).
transitivity_idl2  @  enable_idl_rules, idl_sym(X0,V2,X1), idl(X0,V0,X2), idl(X1,X2,V1) ==>
         infinite(V2) | %print(idl4(V2,V0-V1)),nl,
                        setup_eq(V2,V0-V1).
transitivity_idl3  @  enable_idl_rules, idl_sym(X0,V2,X1), idl(X0,X2,V0), idl(X1,X2,V1)  ==>
         infinite(V2) | %print(idl5(V2,V0-V1)),nl,
                        setup_eq(V2,V0-V1).
transitivity_idl4  @ enable_idl_rules, idl_sym(X0,V2,X1), idl(X0,X2,V0), idl(X1,V1,X2)  ==>
         infinite(V2) | %print(idl6(V2,V0-V1)),nl,
                        setup_eq(V2,V0-V1).

% next rule relevant for solving x+y=100 & y<x   or  x=100-y & y<x
transitivity_idl5  @ enable_idl_rules, idl_sym(A,X,Y), lt(X,Y)  ==>
        (number(A);infinite(X);infinite(Y)) | %print(idl7(A,X,Y)),nl,
                        setup_lt(A,Y+Y),setup_lt(X+X,A).

/*

transitivity_idl3  @  enable_idl_rules, idl(X0,X2,V0), idl(X1,X2,V1) \ idl_sym(X0,V2,X1) <=>
         chck_infinite(V2) | print(idl5(V2,V0-V1)),nl,
                        setup_eq(V2,V0-V1).


old version with idl_pair
transitivity_idl1  @ enable_idl_rules, idl(X0,V2,X1), idl_finite(X0,V0,X2), idl_finite(X1,V1,X2) ==>
         infinite(V2) | print(idl3(V2,V1,V0)),nl, setup_eq(V2,V0-V1).
transitivity_idl2  @ enable_idl_rules, idl(X0,V2,X1), idl_finite(X0,V0,X2), idl_finite(X1,X2,V1) ==>
         infinite(V2) | print(idl4(V2,V1,V0)),nl, setup_eq(V2,V0-V1).
:- chr_constraint idl_pair/4.
% idempotence     @ idl_pair(V,W,Y,R) \ idl_pair(V,W,Y,R) <=> true. % not beneficial ?
transitivity_pair1 @   enable_idl_rules, idl(V,X,YY), idl(W,YY,Z)
                  ==> print(pair1(V,W,X,Z)),nl,
                      idl_pair(V,W,X,Z). % V+W = X-Z
transitivity_pair2 @   enable_idl_rules, idl(YY,X,V), idl(W,YY,Z)
                  ==> print(pair2(V,W,X,Z)),nl,
                      idl_pair(V,W,X,Z). % V+W = X-Z
transitivity_idl  @   enable_idl_rules, idl_pair(V,W,X,Z), idl(U,X,Z)
                  ==> idl_worth_it(U,V,W) |  print(idl5(U,V+W)),nl,
                                %% idl(V,U,W).
                                %clpfd_eq_expr(U,V+W).
                                setup_eq(U,V+W).
                                %kernel_objects:int_plus(int(V),int(W),int(U)).
% heuristic: only set up new constraint when we have an infinite domain
%idl_worth_it(U,V,W) :-print_fd(U),print_fd(V),print_fd(W),nl,fail.
idl_worth_it(U,_V,_W) :- clpfd_interface:clpfd_size(U,sup),!.
idl_worth_it(_U,V,_W) :- clpfd_interface:clpfd_size(V,sup),!.
idl_worth_it(_U,_V,W) :- clpfd_interface:clpfd_size(W,sup),!.
*/

:- public finite/1.
infinite(Var) :- clpfd_interface:clpfd_size(Var,sup).
finite(Var) :- \+ infinite(Var).

:- public print_idl/3.
print_idl(A,B,C) :- print(' IDL: '), print_fd(A), print(' = '), print_fd(B), print(' + '), print_fd(C),nl.
print_fd(N) :- number(N),!,print(N).
print_fd(V) :- clpfd_interface:clpfd_domain(V,VA,VB), %clpfd_interface:clpfd_size(V,Sz),
  format('~w:(~w..~w)',[V,VA,VB]). %print(fd(V,size(Sz),VA:VB)).

% much slower:
%%transitivity_pair @ idl(V,X,Y), idl(W,Y,Z) ==> clpfd_eq_expr(V+W,VW), idl(VW,X,Z).
