% (c) 2009-2020 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(delay,[delay_setof_wf/7,
                 delay_setof_check_wf/9,
                 delay_setof_list/4,
                 %delay_call/2, delay_call/3,
                 %delay_not/3, compute_wait_variables/3
                 not_with_enum_warning/4
                 ]).

% meta_predicate annotations should appear before loading any code:
:- meta_predicate my_findall(-,0,-,-).
:- meta_predicate my_findall_catch(-,0,-,-,-).
:- meta_predicate my_findall_catch(-,0,-,-,-,-).
:- meta_predicate my_findall_check(-,0,-,-,0,0,-).

:- meta_predicate delay_setof_wf(-,0,-,-,-,-,-).
%:- meta_predicate block_my_findall_catch_wf(-,-,0,-,-,-).

:- meta_predicate delay_setof_check_wf(-,0,-,-,-,0,0,-,-).
:- meta_predicate block_findall_check(-,-,0,-,0,0,-,-,-,-).

:- meta_predicate delay_setof_list(-,0,-,-).


%:- meta_predicate delay_call(0,-,-).
%:- meta_predicate delay_call(0,-).

%:- meta_predicate delay_not(0,-,-).
:- meta_predicate not_with_enum_warning(0,-,?,?).
:- meta_predicate not_with_enum_warning2(0,-,-).

% --------------


:- use_module(tools).

:- use_module(module_information,[module_info/2]).
:- module_info(group,kernel).
:- module_info(description,'Utilities to delay calls until sufficiently instantiated.').

%:- use_module(self_check).

:- use_module(debug). /* so that calls can call unqualified debug_prints */

:- use_module(error_manager).
:- use_module(kernel_tools).


:- use_module(tools_printing,[print_term_summary/1]).
:- load_files(library(system), [when(compile_time), imports([environ/2])]).
:- use_module(tools_meta,[call_residue/2]).

:- meta_predicate call_residue_check(0).
:- if(debug:global_debug_flag).
call_residue_check(Call) :- call(Call).
my_findall(X,P,L,_) :-
   statistics(runtime,[Start,_]),
   findall(X,P,L),
   statistics(runtime,[End,_]),
   Tot is End-Start,
   (Tot>50 -> nl, %nl,print(Call),nl,
                 print('*** FINDALL: '),print_term_summary(P),
                 print('*** exceeded limit: '), print(Tot), print(' ms'),nl,
                 length(L,Len),
                 print('*** SOLUTIONS: '), print(Len), nl
		 ; true).
:- elif(environ(prob_safe_mode,true)). % not enabled by default; performance impact seems small; except for test 469
% for test 469 we have a call like this: delay:custom_explicit_sets:test_closure_and_convert/5: [_zzzz_unary]: [integer]: PRED(disjunct/2,z_ : #3906:{1,3,...,3906,3907} or z_ : 3908 .. 19532): VARIABLE: _348117: no_wf_available
% with 19531 solutions
call_residue_check(Call) :- % print_term_summary(Call),nl,
    call_residue(Call,Residue),
    (Residue = [] -> true
      ; add_internal_error('Call had residue: ',residue(Residue)/call(Call)),
        print_goal(Residue),nl
    ).
my_findall(X,P,L,ExpectedFinalResult) :-
   (ExpectedFinalResult==[]
    -> \+(call_residue_check(P)),
       L=[]  % we know the end result to be empty; not necessary to try and find all solutions; <--- we should probably treat this somewhere else to ensure that we do not throw all_solutions enumeration warnings
    ; findall(X,call_residue_check(P),L)
   ).
:- else.
call_residue_check(Call) :- call(Call).
my_findall(X,P,L,ExpectedFinalResult) :-
 (ExpectedFinalResult==[]
    -> \+(P),L=[]  % we know the end result to be empty; not necessary to try and find all solutions; <--- we should probably treat this somewhere else to ensure that we do not throw all_solutions enumeration warnings
    ; findall(X,P,L)).
    %; findall(X,(P,print(sol(X)),nl,nl,trace),L)).
:- endif.




my_findall_catch(X,P,L,ExpectedFinalResult,Span) :-
   my_findall_catch(X,P,L,ExpectedFinalResult,EnumWarningWhichOcc,Span),
   (enum_warning_occ(EnumWarningWhichOcc) -> throw(EnumWarningWhichOcc) ; true).

enum_warning_occ(EnumWarningWhichOcc) :-
    nonvar(EnumWarningWhichOcc),EnumWarningWhichOcc=enumeration_warning(_,_,_,_,_).

my_findall_catch(X,P,L,ExpectedFinalResult,EnumWarningWhichOcc,Span) :-
    enter_new_error_scope(Level,my_findall_catch),
    throw_enumeration_warnings_in_scope(Level,critical,Span), % critical enumeration warnings are thrown straight away
    call_cleanup(
      my_findall(X,P,L,ExpectedFinalResult),
      (event_occurred_in_error_scope(enumeration_warning(A,B,C,D,E))
               -> exit_error_scope(Level,_,my_findall_catch_error),
                  % do we need to throw the enumeration warning ?
                  EnumWarningWhichOcc=enumeration_warning(A,B,C,D,E),
                  debug_println(9,my_findall_catch(enumeration_warning(A,B,C,D,E))), %trace,
                  true %throw(enumeration_warning(A,B,C,D,E))
               ;  exit_error_scope(Level,_,my_findall_catch_no_error),
                  EnumWarningWhichOcc=false)).

:- use_module(runtime_profiler,[observe_runtime/3]).
my_findall_check(X,P,L,ExpectedFinalResult,TimeOutCode,VirtualTimeOutCode,Span) :-
  statistics(runtime,[Start,_]),
  get_total_number_of_errors(Nr1), % now also does count stored wd errors, see WD_SetCompr_Error
  %format('Start closure expansion at ~w ms~n',[Start]),print_message_span(Span),nl,
  call_cleanup((my_findall_catch(X,P,L,ExpectedFinalResult,EnumWarningWhichOcc,Span),Ok=true),
      (Ok==true,EnumWarningWhichOcc==false
       -> get_total_number_of_errors(Nr2), %print(f(Nr1,Nr2)),nl,
          (Nr2>Nr1 -> add_error(well_definedness_error,'Error(s) occurred during expansion of set comprehension','',Span),
                      fail % TO DO: see comment below in delay_setof_wf
                   ;  observe_runtime(Start,'==PERF-MESSAGE==> Long closure expansion: ',Span) %, print(done),nl
                      %,statistics(runtime,[End,_]), Tot is End-Start,
                      %(Tot>10 -> length(L,Len),add_message(delay,'Long closure expansion:',ms_sols(Tot,Len),Span) ; true)
          )
        ; /* failure or time-out; --> time-out as findall should never fail */
          statistics(runtime,[End,_]), Tot is End-Start,
          (enum_warning_occ(EnumWarningWhichOcc)
            -> (VirtualTimeOutCode = _:true -> true
                 ; print('### VIRTUAL TIME-OUT after: '), print(Tot), print(' ms'),nl,
                   call(VirtualTimeOutCode)
                ), %print(throw(EnumWarningWhichOcc)),nl,
               throw(EnumWarningWhichOcc) % ??? this is a set-comprehension expression; failure would mean not-well-defined
           ; Tot>1000 -> % TO DO: could also be CTRL-C ! we could use catch_interrupt_assertion_call(my_findall_catch) or on_exception(user_interrupt_signal,my_finall_catch()
              call(TimeOutCode),  % Warning: this call is also called if an outer-time-out is triggered !
              print('Runtime for set comprehension expansion until TIME-OUT: '),print(Tot),nl
              % WHY DON'T WE FAIL HERE ---> because exception is passed on to outer calls after call_cleanup
           ; true) % check if the findall is probably responsible for the time-out
     )).

/* ----------------------------------------- */

/* Below is a findall that delays until all
   non-output variables have become ground.
 Ideally, one would want a delay_findall that
  figures out by itself when it has all the answers
  (by looking at call_residue) and that may
  even progressively instantiate the output list
  such as in:
     delay_findall(X,delay_member(X,[a|T]),R), delay_member(Z,R)
 But this will require more involved machinery.
*/


:- use_module(kernel_objects,[equal_object_wf/4]).
:- if(current_prolog_flag(dialect,sicstus)).
:- if((current_prolog_flag(version_data,sicstus(4,X,_,_,_)),X<3)).
:- use_module(library(terms),[term_variables/2]). % is built-in in SICSTUS 4.3
:- endif.
:- endif.


:- use_module(kernel_tools,[ground_value_check/2]).

% the following is called by expand_normal_closure_direct
delay_setof_wf(V,G,FullSetResult,WVars,Done,WF,Span) :- %print(delay_set_of(V)),nl,
      ground_value_check(WVars,GW),
      block_my_findall_catch_wf2(GW,V,G,FullSetResult,Done,WF,Span).
% comment in to perform idling to allow outer co-routines to run; useful when WD condition attached to Body G
%:- use_module(kernel_waitflags,[get_idle_wait_flag/3]).
%:- block block_my_findall_catch_wf(-,?,?,?,?,?).
%block_my_findall_catch_wf(_,V,G,FullSetResult,Done,WF) :-
%      get_idle_wait_flag(delay_setof_wf,WF,LWF),
%      print(idling(WF,LWF)),nl,trace,
%      block_my_findall_catch_wf2(LWF,V,G,FullSetResult,Done,WF,Span).
:- block block_my_findall_catch_wf2(-,?,?,?,?,?,?).
block_my_findall_catch_wf2(_,V,G,FullSetResult,Done,WF,Span) :-
              get_total_number_of_errors(Nr1),
	            my_findall_catch(V,G,RRes,FullSetResult,Span),
	            transform_result_into_set(RRes,SetRes),
	            get_total_number_of_errors(Nr2), %print(f(Nr1,Nr2)),nl,
              (Nr2>Nr1
                -> add_error(well_definedness_error,'Error(s) occurred during expansion of set comprehension','',Span),
                   fail
                   % TO DO?: we could add_abort error here instead of adding a real error
                   % for this, however, the inner abort_errors should only be recorded, not yet raised for the user
                ; true),
	            equal_object_wf(SetRes,FullSetResult,delay_setof_wf,WF),
	            Done=true.


delay_setof_check_wf(V,G,FullSetResult,WVars,Done,TimeOutCode,VirtualTimeOutCode,WF,Span) :- %print(delay_set_of(V)),nl,
	   %% print(wait_vars(WVars,G)),nl, %%
     ground_value_check(WVars,GW),
	   block_findall_check(GW,V,G,FullSetResult,TimeOutCode,VirtualTimeOutCode,Done,FullSetResult,WF,Span).

:- block block_findall_check(-,?,?,?,?,?,?,?,?,?).
block_findall_check(_,V,G,FullSetResult,TimeOutCode,VirtualTimeOutCode,Done,FullSetResult,WF,Span) :-
          my_findall_check(V,G,RRes,FullSetResult,TimeOutCode,VirtualTimeOutCode,Span),
	        transform_result_into_set(RRes,SetRes),
            % will generate AVL tree
            % print(delay_setof_result(SetRes,FullSetResult)),nl,  translate:print_bvalue(SetRes),nl,
	        equal_object_wf(SetRes,FullSetResult,delay_setof_check_wf,WF),
	        Done=true.

:- use_module(tools_printing,[print_term_summary/1]).
:- use_module(custom_explicit_sets,[convert_to_avl/2]).
transform_result_into_set(RRes,SetRes) :-   % print(transform_result_into_set(RRes,SetRes)),
   (convert_to_avl(RRes,SetRes) % we are sure that RRes is ground !!??
     -> true
     ;  print_term_summary(convert_to_avl_failed(RRes,SetRes)),
        convert_list_of_expressions_into_set(RRes,SetRes)).

:- use_module(kernel_waitflags,[init_wait_flags/2,ground_wait_flags/1]).
:- use_module(b_interpreter,[convert_list_of_expressions_into_set_wf/4]).
convert_list_of_expressions_into_set(List,Set) :-
    init_wait_flags(WF,[convert_list_of_expressions_into_set]),
    convert_list_of_expressions_into_set_wf(List,Set,set(any),WF),
    ground_wait_flags(WF).

 /* same as above but Result is a list of elements; */
 /*   the list should not be interpreted as a set, but each element is a value for a parameter */
 /* There is a very small chance that the same value is represented twice in the list (e.g., once as closure once as explicit list) */
% last argument is simple waitflag

:- block delay_setof_list(?,?,?,-).
delay_setof_list(V,G,FullSetResult,_) :-
          %print_quoted(findall(V,G,RRes)),nl,
		  my_findall(V,G,RRes,FullSetResult),
	      %%  print_message(delay_seotof_list_rres(RRes)), %%
		  ll_norm(RRes,NormRRes),
		  sort(NormRRes,SNormRRes), % removes duplicates
		  FullSetResult=SNormRRes.


% normalise a list of list of values; the inner lists are not sets but values for parameters
ll_norm([],[]).
ll_norm([El1|T1],[El2|T2]) :- l_check_norm_required(El1,El2,NormRequired),
   (var(NormRequired) -> T1=T2  % no normalisation required: simply copy rest of list
% TO DO: we could also statically determine based on parameter types whether normalising is required
     ; ll_norm2(T1,T2)).

:- use_module(store,[l_normalise_values/2]).
ll_norm2([],[]).
ll_norm2([El1|T1],[El2|T2]) :- l_normalise_values(El1,El2), ll_norm2(T1,T2).

% normalise a value and return norm_required in last argument if it is needed
l_check_norm_required([],[],_).
l_check_norm_required([H|T],[NH|NT],NormRequired) :-
   check_norm_required(H,NH,NormRequired),
   l_check_norm_required(T,NT,NormRequired).

check_norm_required(pred_true,R,_) :- !, R=pred_true.
check_norm_required(pred_false,R,_) :- !, R=pred_false.
check_norm_required(int(S),R,_) :- !, R=int(S).
check_norm_required(fd(X,T),R,_) :- !, R=fd(X,T).
check_norm_required(string(S),R,_) :- !, R=string(S).
check_norm_required((A,B),(NA,NB),NormRequired) :- !,
     check_norm_required(A,NA,NormRequired), check_norm_required(B,NB,NormRequired).
% to do: add record and freeval
check_norm_required(V,NV,norm_required) :- store:normalise_value(V,NV).





%:- use_module(tools,[remove_variables/3]).
%compute_wait_variables(WaitTerm,OutputVars,WaitVars) :-
%    term_variables(WaitTerm,Vars),
%	  term_variables(OutputVars,RealOutputVars),
%	  remove_variables(Vars,RealOutputVars,WaitVars).


%delay_call(Call,WaitTerm,OutputVars) :-
%	  compute_wait_variables(WaitTerm,OutputVars,WaitVars),
	  %print_message(delay(Call,WaitVars)),
%	  when(ground(WaitVars),Call).

%delay_call(Call,OutputVars) :-
%	  delay_call(Call,Call,OutputVars).

% LocalWF: created by create_inner_wait_flags; enumeration finished waitflag should not be grounded here
%delay_not(Call,OutputVars, LocalWF) :-  % print_message(informational,delay_not(Call,OutputVars,LocalWF)),
%   delay_call(  not_with_enum_warning(Call, LocalWF) , OutputVars).

not_with_enum_warning(Call, LocalWF, PP, Span) :-
            %%  print_term_summary(neg_call(Call, LocalWF)), %%
            enter_new_clean_error_scope(Level),
            call_cleanup((not_with_enum_warning2(Call,Level,LocalWF) -> Ok=true ; Ok=false),
                         ((var(Ok) -> add_message(delay,'TIME OUT or Exception inside: ',PP,Span) ; true),
                          exit_error_scope(Level,_ErrOccured,not_with_enum_warn))),
            Ok==true.

:- use_module(tools_printing,[print_goal/1]).
:- use_module(kernel_waitflags,[pending_abort_error/1, ground_wait_flags/1]).
:- use_module(preferences,[get_preference/2]).
not_with_enum_warning2(Call,Level,LocalWF) :-
    %print(not_with_enum_warning2(Level,LocalWF)),nl,
    (call_residue(Call,Residue) ->
        %% print(clearing_warnings(Level)),nl, %%
        %print(done_not_with_enum(Level,LocalWF)),nl,
       (pending_abort_error(LocalWF)
         -> debug_println(19,pending_abort_error(LocalWF)),
            PENDING_ABORT = true
            % we do not fail in this case, so that abort errors can be raised
            % relevant for test 1966 or predicate f={1|->2} & not(#s.(s:1..2 & f(s)=3)) with -p TRY_FIND_ABORT TRUE
         ; Residue \= [] ->
             % nl,term_variables(Call,Vars), print(vars(Vars)),nl, print(call(Call)),nl,
             %write('Residue: '),write(Residue),nl,
             (get_preference(use_chr_solver,true)
              -> add_warning(not_with_enum_warning,'Call (in negation) had residue (possibly due to use of CHR making call_residue imprecise): ',residue(Residue)/call(Call))
              ;  add_internal_error('Call (in negation) had residue: ',residue(Residue)/call(Call)),
                 print_goal(Residue),nl
              ),
            fail
         ; clear_events_in_error_scope(Level,enumeration_warning(_,_,_,_,_)), fail
           % call has succeeded without pending abort errors
        )
      ; otherwise ->  true
     ),
      % Note: everything else is ground; so Call should not trigger any other co-routines/enumeration ?!
      Event = enumeration_warning(_,_,_,_,_),
      (event_occurred_in_error_scope(Event) % critical ? only at this level ? or also below ?
       -> PENDING_ABORT==true, %print(grd(LocalWF)),nl, kernel_waitflags:portray_waitflags(LocalWF),nl,
          ground_wait_flags(LocalWF), % raise any potential abort errors
          debug_println(19,grounded_wait_flag_for_pending_abort_error), % THIS print is important due to issue SPRM-20473 in SICStus Prolog which will otherwise *not* activate the pending co-routines attached to LocalWF !
          fail
          %, print_message(throwing(Event)), throw(Event) % or simply re-post enum warning ?
          % we could also simply fail and raise an enumeration warning; meaning we havent explored all possibilities
          % Note: at the moment the error is copied into the outer scope by copy_error_scope_events
        ; true)
             %%  ,print_term_summary(done_neg_call(Call)),nl %%
      .



