:- module(dpllt_preprocessing, [preprocess_predicate/4]).

:- use_module(library(lists)).
:- use_module(library(sets)).
:- use_module(probsrc('dpllt_solver/difference_logic/ast_to_difference_logic')).
:- use_module(probsrc(bsyntaxtree), [find_identifier_uses/3,
                                     remove_all_infos/2,
                                     conjunct_predicates/2]).
:- use_module(probsrc('well_def/well_def_analyser'), [prove_sequent/1]).

:- use_module(probsrc(module_information), [module_info/2]).
:- module_info(group, dpllt).
:- module_info(description,'This module provides several preprocessing steps for the DPLL(T) based solver such as lifting negations from B operators.').

:- dynamic rewrite_to_difference_logic/0.

%% process_candidate_impls(+RewriteToIdl, +Pred, -LiftedPred, -FilteredCandidateImplsConj).
preprocess_predicate(RewriteToIdl, Pred, LiftedPred, FilteredCandidateImplsConj) :-
    (   (RewriteToIdl, \+ rewrite_to_difference_logic)
    ->  asserta(rewrite_to_difference_logic) % TO DO: don't assert
    ;   (   \+ RewriteToIdl,
            retractall(rewrite_to_difference_logic)
        ;   true
        )
    ),
    empty_candidate_impls_acc(CandidateAcc),
    lift_negations_find_impls(Pred, CandidateAcc, LiftedPred, CandidateImpls),
    process_candidate_impls(CandidateImpls, FilteredCandidateImplsConj).

%% process_candidate_impls(+CandidateImpls, -FilteredCandidateImplsConj).
% Given a candidate like x>y, get partner candidates like y<x, y=<x and x=y.
% Filter those partner candidates that exist in the given formula (prevent introducing redundant SAT variables)
% and create the implication, e.g., x>y => (not(y<x) & not(y=<x) & not(x=y)).
process_candidate_impls(CandidateImpls, FilteredCandidateImplsConj) :-
    CandidateImpls = candidate_impls(integer(Integer),integer_ground(IntegerGrnd),set(Set),set_ground(SetGrnd)),
    process_var_candidate_impls_list(Integer, IntImplsConj),
    process_ground_candidate_impls_list(IntegerGrnd, IntGrndImplsConj),
    process_var_candidate_impls_list(Set, SetImplsConj),
    process_ground_candidate_impls_list(SetGrnd, SetGrndImplsConj),
    conjunct_predicates([IntImplsConj,IntGrndImplsConj,SetImplsConj,SetGrndImplsConj], FilteredCandidateImplsConj).

process_ground_candidate_impls_list(GrndCandidateImpls, ImplsConj) :-
    process_ground_candidate_impls_list(GrndCandidateImpls, [], Impls),
    conjunct_predicates(Impls, ImplsConj).

process_ground_candidate_impls_list([], Acc, Acc).
process_ground_candidate_impls_list([GrndCandidateImpl|T], Acc, ImplsConj) :-
    get_ground_partner_candidates(GrndCandidateImpl, T, Partners, _RestT),
    Partners \== [],
    !,
    build_implications_from_ground_candidates(GrndCandidateImpl, Partners, PartialImplsConj),
    process_ground_candidate_impls_list(T, [PartialImplsConj|Acc], ImplsConj).
process_ground_candidate_impls_list([_|T], Acc, ImplsConj) :-
    process_ground_candidate_impls_list(T, Acc, ImplsConj).

%% build_implications_from_ground_candidates(+GrndCandidateImpl, +Partners, -PartialImplsConj).
% Given GrndCandidateImpl and some matching partner candidates, collect proven implications:
% p in Partners, add one of 'GrndCandidateImpl => p', 'p => GrndCandidateImpl' or nothing
build_implications_from_ground_candidates(GrndCandidateImpl, Partners, PartialImplsConj) :-
    build_implications_from_ground_candidates(GrndCandidateImpl, Partners, [], PartialImplsConj1),
    build_conj_implications_from_ground_candidates(GrndCandidateImpl, Partners, Partners, [], PartialImplsConj2),
    conjunct_predicates([PartialImplsConj1,PartialImplsConj2], PartialImplsConj).

build_conj_implications_from_ground_candidates(_, [], _, Acc, PartialImplsConj) :-
    conjunct_predicates(Acc, PartialImplsConj).
build_conj_implications_from_ground_candidates(GrndCandidateImpl, [Partner|T], Partners, Acc, PartialImplsConj) :-
    select(Partner, Partners, Rest),
    NegConj = b(disjunct(b(negation(GrndCandidateImpl),pred,[]),b(negation(Partner),pred,[])),pred,[]),
    remove_zero_var(NegConj, NegConjNoZero),
    build_conj_implications_from_ground_conj(NegConj, NegConjNoZero, Rest, Acc, NewAcc),
    build_conj_implications_from_ground_candidates(GrndCandidateImpl, T, Partners, NewAcc, PartialImplsConj).

build_conj_implications_from_ground_conj(_, _, [], Acc, Acc).
build_conj_implications_from_ground_conj(NegConj, NegConjNoZero, [Partner|T], Acc, Impls) :-
    remove_zero_var(Partner, PartnerNoZero),
    Impl1 = b(disjunct(NegConj,Partner),pred,[]),
    Impl1NoZero = b(disjunct(NegConjNoZero,PartnerNoZero),pred,[]),
    Impl2 = b(disjunct(NegConj,b(negation(Partner),pred,[])),pred,[]),
    Impl2NoZero = b(disjunct(NegConjNoZero,b(negation(PartnerNoZero),pred,[])),pred,[]),
    (   %nl,write('Try Impl1: '), nl, translate:print_bexpr(Impl1NoZero), nl,
        prove_sequent(Impl1NoZero)%, write('True'),nl
    ->  NewAcc = [Impl1|Acc]
    ;   %nl,write('Try Impl2: '), nl, translate:print_bexpr(Impl2NoZero), nl,
        prove_sequent(Impl2NoZero),%, write('True'),nl,
        NewAcc = [Impl2|Acc]
    ;   NewAcc = Acc
    ),
    build_conj_implications_from_ground_conj(NegConj, NegConjNoZero, T, NewAcc, Impls).

/*
prove_with_prob(Constraint) :-
    find_typed_identifier_uses(Constraint, [], TypedIds),
    translate:generate_typing_predicates(TypedIds, TypingPreds),
    conjunct_predicates(TypingPreds, TypingPred),
    Forall = b(forall(TypedIds,TypingPred,Constraint),pred,[]),
    solver_interface:solve_predicate(Forall, _, Res),
    Res == solution([]).
*/

build_implications_from_ground_candidates(_, [], Acc, PartialImplsConj) :-
    conjunct_predicates(Acc, PartialImplsConj).
build_implications_from_ground_candidates(GrndCandidateImpl, [Partner|T], Acc, PartialImplsConj) :-
    NegGrndCandidateImpl = b(negation(GrndCandidateImpl),pred,[]),
    NegPartner = b(negation(Partner),pred,[]),
    % possibly remove zero var from idl solver
    Impl1 = b(disjunct(NegGrndCandidateImpl,Partner),pred,[]),
    remove_zero_var(Impl1, Impl1NoZero),
    Impl2 = b(disjunct(NegPartner,GrndCandidateImpl),pred,[]),
    remove_zero_var(Impl2, Impl2NoZero),
    Impl3 = b(disjunct(NegGrndCandidateImpl,b(negation(Partner),pred,[])),pred,[]),
    remove_zero_var(Impl3, Impl3NoZero),
    Impl4 = b(disjunct(NegPartner,b(negation(GrndCandidateImpl),pred,[])),pred,[]),
    remove_zero_var(Impl4, Impl4NoZero),
    (   %nl,write('Try Impl1: '), nl, translate:print_bexpr(Impl1NoZero  ), nl,
        prove_sequent(Impl1NoZero)%, write('True'),nl
    ->  NewAcc = [Impl1|Acc]
    ;   %nl,write('Try Impl2: '), nl, translate:print_bexpr(Impl2NoZero), nl,
        prove_sequent(Impl2NoZero)%, write('True'),nl
    ->  NewAcc = [Impl2|Acc]
    ;   %nl,write('Try Impl3: '), nl, translate:print_bexpr(Impl3NoZero), nl,
        prove_sequent(Impl3NoZero)%, write('True'),nl
    ->  NewAcc = [Impl3|Acc]
    ;   %nl,write('Try Impl4: '), nl, translate:print_bexpr(Impl4NoZero), nl,
        prove_sequent(Impl4NoZero)%, write('True'),nl
    ->  NewAcc = [Impl4|Acc]
    ;   NewAcc = Acc
    ),
    build_implications_from_ground_candidates(GrndCandidateImpl, T, NewAcc, PartialImplsConj).

uses_same_ids(A, UsedIds) :-
    find_identifier_uses(A, [], UsedA),
    sets:subset(UsedA, UsedIds),
    sets:subset(UsedIds, UsedA), !.

%% get_ground_partner_candidates(+Id, +Rest, -Partners, -RestRest).
% ASTs in Rest have the same type.
% idea: given, e.g., 1<x: search for all comparisons between x and a ground value like 0<x, x<10 etc.
% afterwards, evaluate the constraints and try to build implications like 1<x => 0<x
get_ground_partner_candidates(GrndCandidateImpl, Rest, Partners, RestRest) :-
    find_identifier_uses(GrndCandidateImpl, [], UsedIds),
    get_ground_partner_candidates(UsedIds, Rest, [], Partners, [], RestRest),!.

get_ground_partner_candidates(_, [], PartnersAcc, PartnersAcc, RestAcc, RestAcc).
get_ground_partner_candidates(UsedIds, Rest, PartnersAcc, Partners, RestAcc, RestRest) :-
    select(Partner, Rest, TRest),
    (   uses_same_ids(Partner, UsedIds)
    ->  NPartnersAcc = [Partner|PartnersAcc],
        NRestAcc = RestAcc
    ;   NPartnersAcc = PartnersAcc,
        NRestAcc = [Partner|RestAcc]
    ),
    !,
    get_ground_partner_candidates(UsedIds, TRest, NPartnersAcc, Partners, NRestAcc, RestRest).

process_var_candidate_impls_list(CandidateImpls, ImplsConj) :-
    process_var_candidate_impls_list(CandidateImpls, CandidateImpls, [], Impls),
    conjunct_predicates(Impls, ImplsConj).

process_var_candidate_impls_list(_, [], Acc, Acc).
process_var_candidate_impls_list(CandidateImpls, [CandidateImpl|T], Acc, Impls) :-
    get_partner_candidates(CandidateImpl, PartnerCandidates),
    filter_and_negate_partner_candidates_for_existence(PartnerCandidates, CandidateImpls, [], Partners),
    Partners \== [],
    !,
    conjunct_predicates(Partners, PartnersConj),
    Impl = b(disjunct(b(negation(CandidateImpl),pred,[]),PartnersConj),pred,[]),
    process_var_candidate_impls_list(CandidateImpls, T, [Impl|Acc], Impls).
process_var_candidate_impls_list(CandidateImpls, [_|T], Acc, Impls) :-
    process_var_candidate_impls_list(CandidateImpls, T, Acc, Impls).

filter_and_negate_partner_candidates_for_existence([], _, Acc, Acc).
filter_and_negate_partner_candidates_for_existence([PartnerCandidate|T], CandidateImpls, Acc, Partners) :-
    remove_all_infos(PartnerCandidate, CleanPartnerCandidate),
    member(CleanPartnerCandidate, CandidateImpls),
    !,
    filter_and_negate_partner_candidates_for_existence(T,  CandidateImpls, [b(negation(PartnerCandidate),pred,[])|Acc], Partners).
filter_and_negate_partner_candidates_for_existence([_|T], Rest, Acc, Partners) :-
    filter_and_negate_partner_candidates_for_existence(T, Rest, Acc, Partners).

% More knowledge for the SAT solver, e.g., x<y => (not(y<x) & not(x=y) & not(y=<x))
% Note: negations are lifted, predicates are normalized (e.g., x+-1 is normalized to x-1, only < and =< etc.)
% Note: the constraints in the second argument's list are negated afterwards, info fields are variables for unification
% TO DO: we could consider transitivity but this is most likely too much overhead.
get_partner_candidates(b(less(Arg1,Arg2),pred,_), [b(less(Arg2,Arg1),pred,[]),b(less_equal(Arg2,Arg1),pred,[]),b(equal(Arg1,Arg2),pred,[])|AdditionalCandidates]) :-
    get_partner_candidates_drop_ground_addition_for_less(Arg1, Arg2, AdditionalCandidates).
get_partner_candidates(b(less_equal(Arg1,Arg2),pred,_), [b(less(Arg2,Arg1),pred,[])]).
get_partner_candidates(b(subset(Arg1,Arg2),pred,_), [b(subset(Arg2,Arg1),pred,[]),b(subset_strict(Arg1,Arg2),pred,[]),b(equal(Arg1,Arg2),pred,[])]).
get_partner_candidates(b(subset_strict(Arg1,Arg2),pred,_), [b(subset(Arg2,Arg1),pred,[])]). % TO DO: maybe add card(Arg2) < card(Arg1) etc.

% TO DO: find more implying constraints, what about z:NATURAL & y < x & x + z < y ?
get_partner_candidates_drop_ground_addition_for_less(Arg1, Arg2, Candidates) :-
    Arg1 = b(add(AddArg1, AddArg2),integer,_),
    AddArg2 = b(integer(I),integer,_),
    I > 0,
    !,
    Candidates = [b(less(Arg2,AddArg1),pred,_)].
get_partner_candidates_drop_ground_addition_for_less(_, _, []).

% Assuming that the AST has been normalized.
%candidate_impl_binary_operator(greater, integer).
%candidate_impl_binary_operator(greater_equal, integer).
candidate_impl_binary_operator(equal, integer).
candidate_impl_binary_operator(member, integer).
candidate_impl_binary_operator(equal, set).
candidate_impl_binary_operator(less, integer).
candidate_impl_binary_operator(less_equal, integer).
candidate_impl_binary_operator(member, set).
candidate_impl_binary_operator(subset, set).
candidate_impl_binary_operator(subset_strict, set).

ground_b_ast(b(integer(Int),_,_)) :-
    ground(Int).
ground_b_ast(b(value(int(Int)),_,_)) :-
    ground(Int).
ground_b_ast(b(interval(Min,Max),_,_)) :-
    Min = b(integer(I1),integer,_),
    Max = b(integer(I2),integer,_),
    integer(I1),
    integer(I2).
ground_b_ast(b(value(Set),_,_)) :-
    ground(Set).

%% extend_candidate_impls_acc(+Type, +Functor, +Arg1, +Arg2, +Ast, +CandidateAcc, -NewCandidateAcc) :-
% Collect ASTs with functor in candidate_impl_binary_operator/2 and both arguments being
% an identifier or ground value.
extend_candidate_impls_acc(Type, Functor, Arg1, Arg2, Ast, CandidateAcc, NewCandidateAcc) :-
    Type == pred,
    candidate_impl_binary_operator(Functor, FType),
    (   (ground_b_ast(Arg1),
         is_id_or_binary_containing_id(Arg2))
    ;
        (ground_b_ast(Arg2),
         is_id_or_binary_containing_id(Arg1))
    ),
    remove_all_infos(Ast, AstClean),
    ast_not_in_candidate_acc(ground, FType, AstClean, CandidateAcc),
    !,
    extend_candidate_impls_acc_safe_ground(FType, CandidateAcc, Ast, NewCandidateAcc).
extend_candidate_impls_acc(Type, Functor, Arg1, Arg2, Ast, CandidateAcc, NewCandidateAcc) :-
    Type == pred,
    candidate_impl_binary_operator(Functor, FType),
    is_id_or_binary_containing_id(Arg1),
    is_id_or_binary_containing_id(Arg2),
    remove_all_infos(Ast, AstClean),
    ast_not_in_candidate_acc(var, FType, AstClean, CandidateAcc),
    !,
    extend_candidate_impls_acc_safe(FType, CandidateAcc, Ast, NewCandidateAcc).
extend_candidate_impls_acc(_, _, _, _, _, CandidateAcc, CandidateAcc).

is_id_or_binary_containing_id(b(identifier(_),_,_)) :-
    !.
is_id_or_binary_containing_id(b(unary_minus(b(identifier(_),_,_)),_,_)) :-
    !.
is_id_or_binary_containing_id(b(Node,_,_)) :-
    Node =.. [_,Arg1,Arg2],
    (   is_id_or_binary_containing_id(Arg1),
        !
    ;   is_id_or_binary_containing_id(Arg2)
    ).

%% ast_not_in_candidate_acc(+VarOrGround, +Type, +AstClean, +CandidateAcc).
ast_not_in_candidate_acc(var, integer, AstClean, candidate_impls(integer(Integer),integer_ground(_),set(_),set_ground(_))) :-
    \+ member(AstClean, Integer).
ast_not_in_candidate_acc(ground, integer, AstClean, candidate_impls(integer(_),integer_ground(IntegerGrnd),set(_),set_ground(_))) :-
    \+ member(AstClean, IntegerGrnd).
ast_not_in_candidate_acc(var, set, AstClean, candidate_impls(integer(_),integer_ground(_),set(Set),set_ground(_))) :-
    \+ member(AstClean, Set).
ast_not_in_candidate_acc(ground, set, AstClean, candidate_impls(integer(_),integer_ground(_),set(_),set_ground(SetGrnd))) :-
    \+ member(AstClean, SetGrnd).

% TO DO: use get_sorted_equality/3 for equalities
extend_candidate_impls_acc_safe_ground(integer, candidate_impls(integer(Integer),integer_ground(IntegerGrnd),set(Set),set_ground(SetGrnd)), Ast, candidate_impls(integer(Integer),integer_ground([Ast|IntegerGrnd]),set(Set),set_ground(SetGrnd))).
extend_candidate_impls_acc_safe_ground(set, candidate_impls(integer(Integer),integer_ground(IntegerGrnd),set(Set),set_ground(SetGrnd)), Ast, candidate_impls(integer(Integer),integer_ground(IntegerGrnd),set(Set),set_ground([Ast|SetGrnd]))).

extend_candidate_impls_acc_safe(integer, candidate_impls(integer(Integer),integer_ground(IntegerGrnd),set(Set),set_ground(SetGrnd)), Ast, candidate_impls(integer([Ast|Integer]),integer_ground(IntegerGrnd),set(Set),set_ground(SetGrnd))).
extend_candidate_impls_acc_safe(set, candidate_impls(integer(Integer),integer_ground(IntegerGrnd),set(Set),set_ground(SetGrnd)), Ast, candidate_impls(integer(Integer),integer_ground(IntegerGrnd),set([Ast|Set]),set_ground(SetGrnd))).

empty_candidate_impls_acc(candidate_impls(integer([]),integer_ground([]),set([]),set_ground([]))).

%% lift_negations_find_impls(+BTerm, +CandidateAcc, -LPred, -CandidateImpls).
% Unfold negated operators like not_equal to detect equal ASTs when introducing
% boolean variables for the SAT formula. For instance, it improves performance
% to transform 'x={} & x/={}' to 'x={} & not(x={})' first, since we can then translate to
% 'A & not(A)' in SAT rather than 'A & B'.
% Collect specific operators that might imply each other like 'x>0' and 'x>1' result to 'x>1 => x>0' (see candidate_impl_binary_operator/2).
% Note: only binary operators are checked recursively since we transform SMT to SAT on the level of conjunct, disjunct and implication
% Note: CandidateImpls contains true asts of lifted predicates, e.g., 'x:{1,2}' is stored for 'x/:{1,2}'
lift_negations_find_impls(Pred, CandidateAcc, LPred, CandidateImpls) :-
    Pred = b(equivalence(Lhs,Rhs),pred,_),
    !,
    Impl1 = b(implication(Lhs,Rhs),pred,[]),
    Impl2 = b(implication(Rhs,Lhs),pred,[]),
    Rewritten = b(conjunct(Impl1,Impl2),pred,[]),
    lift_negations_find_impls(Rewritten, CandidateAcc, LPred, CandidateImpls).
lift_negations_find_impls(Pred, CandidateAcc, LPred, CandidateImpls) :-
    Pred = b(implication(Lhs,Rhs),pred,_),
    !,
    NLhs = b(negation(Lhs),pred,[]),
    Rewritten = b(disjunct(NLhs,Rhs),pred,[]),
    lift_negations_find_impls(Rewritten, CandidateAcc, LPred, CandidateImpls).
lift_negations_find_impls(Pred, CandidateAcc, LPred, CandidateImpls) :-
    Pred = b(negation(b(negation(Pos),pred,_)),pred,_),
    !,
    lift_negations_find_impls(Pos, CandidateAcc, LPred, CandidateImpls).
lift_negations_find_impls(Pred, CandidateAcc, LPred, CandidateImpls) :-
    Pred = b(negation(b(truth,pred,_)),pred,Info),
    !,
    LPred = b(falsity,pred,Info),
    CandidateImpls = CandidateAcc.
lift_negations_find_impls(Pred, CandidateAcc, LPred, CandidateImpls) :-
    Pred = b(negation(b(falsity,pred,_)),pred,Info),
    !,
    LPred = b(truth,pred,Info),
    CandidateImpls = CandidateAcc.
lift_negations_find_impls(Inequality, CandidateAcc, LPred, CandidateImpls) :-
    rewrite_to_difference_logic,
    (   Inequality = b(not_equal(_,_),pred,_)
        %IEQ = Inequality
    ;   Inequality = b(negation(b(equal(_,_),pred,_)),pred,_)
        %IEQ = b(not_equal(Arg1,Arg2),pred,EQInfo)
    ),
    rewrite_inequality_to_idl_disj_no_zero(Inequality, DLConstraint),
    !,
    DLConstraint = b(disjunct(Lhs,Rhs),pred,_),
    Lhs = b(negation(LhsPos),pred,_),
    Rhs = b(negation(RhsPos),pred,_),
    LhsPos = b(less_equal(LArg1,LArg2),pred,_),
    RhsPos = b(less_equal(RArg1,RArg2),pred,_),
    LPred = DLConstraint,
    %extend_candidate_impls_acc(pred, not_equal, Arg1, Arg2, IEQ, CandidateAcc, CandidateAcc1),
    extend_candidate_impls_acc(pred, less_equal, LArg1, LArg2, LhsPos, CandidateAcc, CandidateAcc1),
    extend_candidate_impls_acc(pred, less_equal, RArg1, RArg2, RhsPos, CandidateAcc1, CandidateImpls).
lift_negations_find_impls(Equality, CandidateAcc, LPred, CandidateImpls) :-
    rewrite_to_difference_logic,
    (   Equality = b(equal(Arg1,_),pred,_)
        %EQ = Equality
    ;   Equality = b(negation(b(not_equal(Arg1,_),pred,_)),pred,_)
        %EQ = b(equal(Arg1,Arg2),pred,EQInfo)
    ),
    Arg1 = b(_,integer,_),
    rewrite_to_idl_no_zero(Equality, ConjList),
    !,
    ConjList = [TLhs,TRhs],
    % don't create _zero var from idl solver in SAT formula
    remove_idl_origin_from_info(TLhs, Lhs),
    remove_idl_origin_from_info(TRhs, Rhs),
    conjunct_predicates([Lhs,Rhs], DLConstraint),
    Lhs = b(less_equal(LArg1,LArg2),pred,_),
    Rhs = b(less_equal(RArg1,RArg2),pred,_),
    LPred = DLConstraint,
    extend_candidate_impls_acc(pred, less_equal, LArg1, LArg2, Lhs, CandidateAcc, CandidateAcc1),
    extend_candidate_impls_acc(pred, less_equal, RArg1, RArg2, Rhs, CandidateAcc1, CandidateImpls).
lift_negations_find_impls(Pred, CandidateAcc, LPred, CandidateImpls) :-
    Pred = b(Node,Type,Info),
    functor(Node, Functor, 2),
    negated_b_operator(Functor, NFunctor),
    !,
    arg(1, Node, Arg1),
    arg(2, Node, Arg2),
    TrueNode =.. [NFunctor,Arg1,Arg2],
    TrueAst = b(TrueNode,Type,Info),
    extend_candidate_impls_acc(Type, NFunctor, Arg1, Arg2, TrueAst, CandidateAcc, CandidateImpls),
    LPred = b(negation(TrueAst),pred,[]).
% simplify negated conjunction or disjunction possibly introduced by rewriting equivalence and implication
lift_negations_find_impls(b(Node,_,_), CandidateAcc, LPred, CandidateImpls) :-
    Node = negation(Arg),
    Arg = b(conjunct(Conj1,Conj2),pred,ConjInfo),
    !,
    NConj1 = b(negation(Conj1),pred,[]),
    NConj2 = b(negation(Conj2),pred,[]),
    Rewritten = b(disjunct(NConj1,NConj2),pred,ConjInfo),
    lift_negations_find_impls(Rewritten, CandidateAcc, LPred, CandidateImpls).
lift_negations_find_impls(b(Node,_,_), CandidateAcc, LPred, CandidateImpls) :-
    Node = negation(Arg),
    Arg = b(disjunct(Conj1,Conj2),pred,ConjInfo),
    !,
    NConj1 = b(negation(Conj1),pred,[]),
    NConj2 = b(negation(Conj2),pred,[]),
    Rewritten = b(conjunct(NConj1,NConj2),pred,ConjInfo),
    lift_negations_find_impls(Rewritten, CandidateAcc, LPred, CandidateImpls).
lift_negations_find_impls(b(Node,Type,Info), CandidateAcc, LPred, CandidateImpls) :-
    Node = negation(Arg),
    lift_negations_find_impls(Arg, CandidateAcc, LArg, CandidateImpls),
    (   LArg = b(negation(InnerArg),_,_)
    ->  LPred = InnerArg % simplify double negation
    ;   LPred = b(negation(LArg),Type,Info)
    ).
lift_negations_find_impls(b(Node,Type,Info), CandidateAcc, LPred, CandidateImpls) :-
    functor(Node, Functor, 2),
    Functor \== exists, % quantifications will be abstracted by a single SAT variable
    !,
    arg(1, Node, Arg1),
    arg(2, Node, Arg2),
    extend_candidate_impls_acc(Type, Functor, Arg1, Arg2, b(Node,Type,Info), CandidateAcc, CandidateAcc1),
    lift_negations_top_level_functor(Functor, Node, CandidateAcc1, NNode, CandidateImpls),
    LPred = b(NNode,Type,Info).
lift_negations_find_impls(BTerm, CandidateAcc, BTerm, CandidateAcc).

lift_negations_top_level_functor(Functor, Node, CandidateAcc, NNode, CandidateImpls) :-
    arg(1, Node, Lhs),
    arg(2, Node, Rhs),
    lift_negations_find_impls(Lhs, CandidateAcc, LLhs, CandidateAcc1),
    lift_negations_find_impls(Rhs, CandidateAcc1, LRhs, CandidateImpls),
    NNode =.. [Functor,LLhs,LRhs].

get_sorted_equality(Lhs, Rhs, Equality) :-
    Lhs @> Rhs,
    !,
    Equality = b(equal(Rhs,Lhs),pred,[]).
get_sorted_equality(Lhs, Rhs, b(equal(Lhs,Rhs),pred,[])).

negated_b_operator(not_equal, equal).
negated_b_operator(not_member, member).
negated_b_operator(not_subset, subset).
negated_b_operator(not_subset_strict, subset_strict).

%%%%%%%%%%%%%%%%%%%%% Unit Tests %%%%%%%%%%%%%%%%%%%%%

:- use_module(library(plunit)).

:- begin_tests(extend_candidate_impls_acc).

test(extend_candidate_impls_acc_less, [true(NewAcc == candidate_impls(integer([]),integer_ground([b(less(b(identifier(x),integer,[]),b(integer(7),integer,[])),pred,[])]),set([]),set_ground([])))]) :-
    empty_candidate_impls_acc(EmptyAcc),
    Arg1 = b(identifier(x),integer,[]),
    Arg2 = b(integer(7),integer,[]),
    Ast = b(less(Arg1,Arg2),pred,[]),
    extend_candidate_impls_acc(pred, less, Arg1, Arg2, Ast, EmptyAcc, NewAcc).

test(extend_candidate_impls_acc_less_non_ground, [true(NewAcc == candidate_impls(integer([]),integer_ground([]),set([]),set_ground([])))]) :-
    empty_candidate_impls_acc(EmptyAcc),
    Arg1 = b(identifier(x),set(set(integer)),[]),
    Arg2 = b(set_extension([_]),set(set(integer)),[]),
    Ast = b(less(Arg1,Arg2),pred,[]),
    extend_candidate_impls_acc(pred, less, Arg1, Arg2, Ast, EmptyAcc, NewAcc).

test(extend_candidate_impls_acc_less_no_identifier, [true(NewAcc == candidate_impls(integer([]),integer_ground([]),set([]),set_ground([])))]) :-
    empty_candidate_impls_acc(EmptyAcc),
    Arg1 = b(comprehension_set([b(identifier(x),integer,[])],b(member(b(identifier(x),integer,[]), b(set_extension([b(integer(234),integer,[]),b(integer(34),integer,[])]),set(integer),[])),pred,[])),set(integer),[some,info]),
    Arg2 = b(set_extension([b(set_extension([b(integer(27),integer,[])]),set(integer),[])]),set(set(integer)),[]),
    Ast = b(less(Arg1,Arg2),pred,[]),
    extend_candidate_impls_acc(pred, less, Arg1, Arg2, Ast, EmptyAcc, NewAcc).

test(extend_candidate_impls_acc_less_eq, [true(NewAcc == candidate_impls(integer([]),integer_ground([b(less_equal(b(identifier(x),integer,[]),b(integer(7),integer,[])),pred,[])]),set([]),set_ground([])))]) :-
    empty_candidate_impls_acc(EmptyAcc),
    Arg1 = b(identifier(x),integer,[]),
    Arg2 = b(integer(7),integer,[]),
    Ast = b(less_equal(Arg1,Arg2),pred,[]),
    extend_candidate_impls_acc(pred, less_equal, Arg1, Arg2, Ast, EmptyAcc, NewAcc).

test(extend_candidate_impls_acc_less_eq_non_ground, [true(NewAcc == candidate_impls(integer([]),integer_ground([]),set([]),set_ground([])))]) :-
    empty_candidate_impls_acc(EmptyAcc),
    Arg1 = b(identifier(x),set(set(integer)),[]),
    Arg2 = b(set_extension([_]),set(set(integer)),[]),
    Ast = b(less_equal(Arg1,Arg2),pred,[]),
    extend_candidate_impls_acc(pred, less_equal, Arg1, Arg2, Ast, EmptyAcc, NewAcc).

test(extend_candidate_impls_acc_less_eq_no_identifier, [true(NewAcc == candidate_impls(integer([]),integer_ground([]),set([]),set_ground([])))]) :-
    empty_candidate_impls_acc(EmptyAcc),
    Arg1 = b(comprehension_set([b(identifier(x),integer,[])],b(member(b(identifier(x),integer,[]), b(set_extension([b(integer(234),integer,[]),b(integer(34),integer,[])]),set(integer),[])),pred,[])),set(integer),[some,info]),
    Arg2 = b(set_extension([b(set_extension([b(integer(27),integer,[])]),set(integer),[])]),set(set(integer)),[]),
    Ast = b(less_equal(Arg1,Arg2),pred,[]),
    extend_candidate_impls_acc(pred, less_equal, Arg1, Arg2, Ast, EmptyAcc, NewAcc).

test(extend_candidate_impls_acc_subset, [true(NewAcc == candidate_impls(integer([]),integer_ground([]),set([]),set_ground([b(subset(b(identifier(x),set(set(integer)),[]),b(set_extension([b(set_extension([b(integer(27),integer,[])]),set(integer),[])]),set(set(integer)),[])),pred,[])])))]) :-
    empty_candidate_impls_acc(EmptyAcc),
    Arg1 = b(identifier(x),set(set(integer)),[]),
    Arg2 = b(set_extension([b(set_extension([b(integer(27),integer,[])]),set(integer),[])]),set(set(integer)),[]),
    Ast = b(subset(Arg1,Arg2),pred,[]),
    extend_candidate_impls_acc(pred, subset, Arg1, Arg2, Ast, EmptyAcc, NewAcc).

test(extend_candidate_impls_acc_subset_non_ground, [true(NewAcc == candidate_impls(integer([]),integer_ground([]),set([]),set_ground([])))]) :-
    empty_candidate_impls_acc(EmptyAcc),
    Arg1 = b(identifier(x),set(set(integer)),[]),
    Arg2 = b(set_extension([_]),set(set(integer)),[]),
    Ast = b(subset(Arg1,Arg2),pred,[]),
    extend_candidate_impls_acc(pred, subset, Arg1, Arg2, Ast, EmptyAcc, NewAcc).

test(extend_candidate_impls_acc_subset_no_identifier, [true(NewAcc == candidate_impls(integer([]),integer_ground([]),set([]),set_ground([])))]) :-
    empty_candidate_impls_acc(EmptyAcc),
    Arg1 = b(comprehension_set([b(identifier(x),integer,[])],b(member(b(identifier(x),integer,[]), b(set_extension([b(integer(234),integer,[]),b(integer(34),integer,[])]),set(integer),[])),pred,[])),set(integer),[some,info]),
    Arg2 = b(set_extension([b(set_extension([b(integer(27),integer,[])]),set(integer),[])]),set(set(integer)),[]),
    Ast = b(subset(Arg1,Arg2),pred,[]),
    extend_candidate_impls_acc(pred, subset, Arg1, Arg2, Ast, EmptyAcc, NewAcc).

test(extend_candidate_impls_acc_subset_strict, [true(NewAcc == candidate_impls(integer([]),integer_ground([]),set([]),set_ground([b(subset_strict(b(identifier(x),set(set(integer)),[]),b(set_extension([b(set_extension([b(integer(27),integer,[])]),set(integer),[])]),set(set(integer)),[])),pred,[])])))]) :-
    empty_candidate_impls_acc(EmptyAcc),
    Arg1 = b(identifier(x),set(set(integer)),[]),
    Arg2 = b(set_extension([b(set_extension([b(integer(27),integer,[])]),set(integer),[])]),set(set(integer)),[]),
    Ast = b(subset_strict(Arg1,Arg2),pred,[]),
    extend_candidate_impls_acc(pred, subset_strict, Arg1, Arg2, Ast, EmptyAcc, NewAcc).

test(extend_candidate_impls_acc_subset_strict_non_ground, [true(NewAcc == candidate_impls(integer([]),integer_ground([]),set([]),set_ground([])))]) :-
    empty_candidate_impls_acc(EmptyAcc),
    Arg1 = b(identifier(x),set(set(integer)),[]),
    Arg2 = b(set_extension([_]),set(set(integer)),[]),
    Ast = b(subset_strict(Arg1,Arg2),pred,[]),
    extend_candidate_impls_acc(pred, subset_strict, Arg1, Arg2, Ast, EmptyAcc, NewAcc).

test(extend_candidate_impls_acc_subset_strict_no_identifier, [true(NewAcc == candidate_impls(integer([]),integer_ground([]),set([]),set_ground([])))]) :-
    empty_candidate_impls_acc(EmptyAcc),
    Arg1 = b(comprehension_set([b(identifier(x),integer,[])],b(member(b(identifier(x),integer,[]), b(set_extension([b(integer(234),integer,[]),b(integer(34),integer,[])]),set(integer),[])),pred,[])),set(integer),[some,info]),
    Arg2 = b(set_extension([b(set_extension([b(integer(27),integer,[])]),set(integer),[])]),set(set(integer)),[]),
    Ast = b(subset_strict(Arg1,Arg2),pred,[]),
    extend_candidate_impls_acc(pred, subset_strict, Arg1, Arg2, Ast, EmptyAcc, NewAcc).

test(extend_candidate_impls_acc_equal, [true(NewAcc == candidate_impls(integer([]),integer_ground([]),set([]),set_ground([b(equal(b(identifier(x),set(set(integer)),[]),b(set_extension([b(set_extension([b(integer(27),integer,[])]),set(integer),[])]),set(set(integer)),[])),pred,[])])))]) :-
    empty_candidate_impls_acc(EmptyAcc),
    Arg1 = b(identifier(x),set(set(integer)),[]),
    Arg2 = b(set_extension([b(set_extension([b(integer(27),integer,[])]),set(integer),[])]),set(set(integer)),[]),
    Ast = b(equal(Arg1,Arg2),pred,[]),
    extend_candidate_impls_acc(pred, equal, Arg1, Arg2, Ast, EmptyAcc, NewAcc).

test(extend_candidate_impls_acc_member, [true(NewAcc == candidate_impls(integer([]),integer_ground([]),set([]),set_ground([])))]) :-
    empty_candidate_impls_acc(EmptyAcc),
    Arg1 = b(identifier(x),set(integer),[]),
    Arg2 = b(set_extension([b(set_extension([b(integer(27),integer,[])]),set(integer),[])]),set(set(integer)),[]),
    Ast = b(member(Arg1,Arg2),pred,[]),
    extend_candidate_impls_acc(pred, member, Arg1, Arg2, Ast, EmptyAcc, NewAcc).

test(extend_candidate_impls_acc_member_non_ground_type, [true(NewAcc == candidate_impls(integer([]),integer_ground([]),set([]),set_ground([])))]) :-
    empty_candidate_impls_acc(EmptyAcc),
    Arg1 = b(identifier(x),set(integer),[]),
    Arg2 = b(set_extension([b(set_extension([b(integer(27),integer,[])]),set(integer),[])]),set(set(integer)),[]),
    Ast = b(member(Arg1,Arg2),pred,[]),
    extend_candidate_impls_acc(_Type, member, Arg1, Arg2, Ast, EmptyAcc, NewAcc).

test(extend_candidate_impls_acc_add, [true(NewAcc == candidate_impls(integer([]),integer_ground([]),set([]),set_ground([])))]) :-
    empty_candidate_impls_acc(EmptyAcc),
    Arg1 = b(integer(1),integer,[]),
    Arg2 = b(integer(1),integer,[]),
    Ast = b(add(Arg1,Arg2),integer,[]),
    extend_candidate_impls_acc(integer, add, Arg1, Arg2, Ast, EmptyAcc, NewAcc).

test(extend_candidate_impls_acc_overwrite, [true(NewAcc == candidate_impls(integer([]),integer_ground([]),set([]),set_ground([])))]) :-
    empty_candidate_impls_acc(EmptyAcc),
    Arg1 = b(set_extension([b(couple(b(integer(1),integer,[]),b(integer(2),integer,[])),couple(integer,integer),[info(o),o])]),set(couple(integer,integer)),[]),
    Arg2 = b(set_extension([b(couple(b(integer(1),integer,[]),b(integer(2),integer,[])),couple(integer,integer),[ein(e),wei(t,e(re)),info])]),set(couple(integer,integer)),[]),
    Ast = b(overwrite(Arg1,Arg2),set(couple(integer,integer)),[]),
    extend_candidate_impls_acc(integer, add, Arg1, Arg2, Ast, EmptyAcc, NewAcc).

test(extend_candidate_impls_acc_overwrite_non_ground_type, [true(NewAcc == candidate_impls(integer([]),integer_ground([]),set([]),set_ground([])))]) :-
    empty_candidate_impls_acc(EmptyAcc),
    Arg1 = b(set_extension([b(couple(b(integer(1),integer,[]),b(integer(2),integer,[])),couple(integer,integer),[info(o),o])]),set(couple(integer,integer)),[]),
    Arg2 = b(set_extension([b(couple(b(integer(1),integer,[]),b(integer(2),integer,[])),couple(integer,integer),[ein(e),wei(t,e(re)),info])]),set(couple(integer,integer)),[]),
    Ast = b(overwrite(Arg1,Arg2),set(couple(integer,integer)),[]),
    extend_candidate_impls_acc(_Type, add, Arg1, Arg2, Ast, EmptyAcc, NewAcc).

:- end_tests(extend_candidate_impls_acc).

