:- module(dpllt_sat_solver, [b_to_cnf/3,
                             create_solver_state/1,
                             sat/2,
                             sat/3,
                             is_backjumping/0,
                             is_last_propagated_sat_var/2,
                             announce_bt_from_smt_solution/0,
                             store_idl_unsat_core/1,
                             log_theory_propagation_sat_stack/3,
                             set_variable_references_in_clauses/3,
                             remove_idl_unsat_core/0]).

:- use_module(library(lists)).
:- use_module(library(avl)).
:- use_module(library(heaps)).
:- use_module(library(ordsets)).

:- use_module(probsrc('dpllt_solver/dpllt_pred_to_sat')).
:- use_module(probsrc('dpllt_solver/dpllt_settings')).
:- use_module(probsrc('dpllt_solver/symmetry_check/sat_symmetry_breaking')).
:- use_module(probsrc(bsyntaxtree),[get_texpr_expr/2,
                                    create_texpr/4,
                                    conjunction_to_list/2]).
:- use_module(probsrc(error_manager), [add_error_fail/3]).
:- use_module(probsrc(translate), [translate_bexpression/2]).
:- use_module(probsrc(unsat_cores), [unsat_core_with_fixed_conjuncts/3]).
:- use_module(probsrc(module_information), [module_info/2]).

:- module_info(group, dpllt).
:- module_info(description,'This module provides the SAT solver for DPLL(T) implementing CDCL with backjumping.').

:- dynamic backjump_clause/3, backjump_level/1, conflict_stats/3, saved_phase/2, bt_from_smt_solution/0, is_restart/0, exclude_sat_solution_clause/1, total_lbd_sum/1, recent_lbds_dl/2, recent_stack_sizes_dl/2, amount_clause_reductions/1, amount_learned_clauses/1, idl_unsat_core/1.
:- volatile backjump_clause/3, backjump_level/1, conflict_stats/3, saved_phase/2, bt_from_smt_solution/0, is_restart/0, exclude_sat_solution_clause/1, total_lbd_sum/1, recent_lbds_dl/2, recent_stack_sizes_dl/2, amount_clause_reductions/1, amount_learned_clauses/1, idl_unsat_core/1.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% A sat solver utilising delay declaration to implement watched literals.
% Authors: Jacob Howe and Andy King
%
% Extended for DPLL(T) with first UIP conflict-driven clause learning,
% theory propagation, branching heuristics, phase saving for rapid restarts,
% learned clause reduction, random restarts, and static SMT symmetry breaking.
% We implement a variant of VSIDS as follows:
%   - one score for each variable initialized with the number of occurences in all clauses
%   - increase score by one for each variable occuring in a learned clause
%   - divide all scores by 2 after each dpllt_settings:vsids_decay_frequency/1 conflict (favor recently learned clauses)
% Random restarts:
%   restarts using either the Luby sequence or Glucose restarts
% Phase saving:
%   - if backjump level is 0 because of learning a unit clause or a random restart,
%   cache phases/polarities of variable assignments (see saved_phase/2)
%   - when the decision heuristic decides to branch on a variable, pick the cached phase first if one exists
% Reducing learned clauses:
%   - remove half of the learned clauses every "frequency_constant + constant_factor * x" conflicts, where x is the amount of clause reductions so far
%   - keep clauses with small LBD (dpllt_settings:discard_clause_greater_lbd/1) forever
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% TO DO:
%        [] ANRFA, VMTF, (ACIDS) branching heuristics
%        [] polarity for heuristic scores (no branching when selecting max heuristic value?)
%              -> first tests inidicated that this is slower
%        (- FirstNewCutscheme (cf. https://www.aaai.org/Papers/JAIR/Vol22/JAIR-2210.pdf))
%        [] improve implying constraint detection (e.g., :dpllt x > y & gt : {x+1} & y > gt)
%        [] clause minimization (cf. Definition 9, "Towards Understanding and Harnessing the Potential of Clause Learning")
%        [] use profiler
%        [] refactor current level

%% sat(+SolverEnv, +Clauses).
% SolverEnv is a list [sat_vars:_, sat_bindings:_, t2b_env:_, pl_vars:_].
% sat_vars are boolean b/3 identifier ASTs for corresponding SAT variables in Clauses.
% sat_bindings is a list of terms bind/2 as used by the ProB constraint solver.
% t2b_env is the environment mapping SAT variable names and theory formulae used to
% translate B predicates to SAT.
% pl_vars is a list of triples PrologVar-SatVarName-SmtFormula.
sat(SolverEnv, Clauses) :-
    init_dynamic_solver_state,
    create_solver_state(StateMutable),
    %init_heuristic(SolverEnv),
    init_decision_heuristic(StateMutable, Clauses),
    sat_stateful_pre(SolverEnv, StateMutable, Clauses).

%% sat(+SolverEnv, +StateMutable, +Clauses).
sat(SolverEnv, StateMutable, Clauses) :-
    init_dynamic_solver_state,
    init_decision_heuristic(StateMutable, Clauses),
    sat_stateful_pre(SolverEnv, StateMutable, Clauses).

% Wrapper for sat_stateful_restart_cp/3 to enable backjumping to level 0 several times.
sat_stateful_pre(SolverEnv, StateMutable, Clauses) :-
    sat_stateful_restart_cp(SolverEnv, StateMutable, Clauses).

sat_stateful_restart_cp(SolverEnv, StateMutable, Clauses) :-
    (   sat_stateful(SolverEnv, StateMutable, Clauses)
    ;   % we either learned a unit-clause from conflict analysis or randomly restart
        % in both cases, we learn all asserted backjump clauses
        retract(backjump_level(0)),
        (   retract(is_restart)
        ->  format("backjump to level 0 after learning a unit-clause or random restart~n", [])
        ;   true
        ),
        learn_asserted_clauses(0, SolverEnv, Clauses, NewClauses),
        sat_stateful_pre(SolverEnv, StateMutable, NewClauses)
    ).

%% sat_stateful(+Clauses, +StateMutable, +Clauses).
% DPLL with first UIP conflict-driven clause learning, two-watched literals, VSIDS branching heuristic,
% random restarts (Luby sequence with unit length 32), and phase saving for rapid restarts.
% See create_solver_state/1 for variable StateMutable.
sat_stateful(SolverEnv, StateMutable, Clauses) :-
    problem_setup(SolverEnv, StateMutable, Clauses),
    get_from_solver_env(SolverEnv, pl_vars, PlVars), % TO DO: enable constant access by assuming an order of SolverEnv
    %static_sat_symmetry_breaking(Clauses, SolverEnv, CurLevelMut, NewSolverEnv),
    elim_vars(SolverEnv, StateMutable, PlVars).

%% elim_vars(+SolverEnv, +StateMutable, +PlVars).
elim_vars(SolverEnv, StateMutable, PlVars) :-
    get_full_state(StateMutable, cur_level, CurLevel),
    (   % solution found
        get_full_state(StateMutable, heuristic_heap, Heap),
        empty_heap(Heap)
    ;   update_full_state(StateMutable, decision_level, CurLevel),
        %possibly_decay_heuristic_values, % only for vsids; remove for now
        (   elim_vars_from_asserted_heuristic_rapid_restart(SolverEnv, StateMutable, CurLevel, PlVars)
        ;   backjump_level(C),
            C == CurLevel,
            retract(backjump_level(C)),
            get_clauses_to_learn(C, SolverEnv, ClausesToLearn),
            update_heuristic_scores_for_clauses(StateMutable, ClausesToLearn),
            problem_setup(SolverEnv, StateMutable, ClausesToLearn),
            format("~nbackjump: branch prop level ~w~n", [CurLevel]),
            elim_vars(SolverEnv, StateMutable, PlVars)
        )
    ).

%% elim_vars_from_asserted_heuristic_rapid_restart(+SolverEnv, +StateMutable, +CurLevel, +PlVars).
% Variable State Independent Decaying Sum (VSIDS) branching heuristic.
% Initially, each variable is assigned the number of its occurences in all clauses as its score.
% Afterwards, for each conflict clause, the scores of the variables occuring in the clause are increased by bump_heuristic_score_for_type.
% Periodically divides all scores by a constant for vsids (see possibly_decay_heuristic_values/0).
% Possibly choose the polarity to assign first using saved_phase/2 for rapid restarts.
% Note: Assert state before propagation because Var is bound to co-routines.
% OriginalClause not necessary for decision variables in the stack because
% they are not expanded during conflict analysis.
elim_vars_from_asserted_heuristic_rapid_restart(SolverEnv, StateMutable, CurLevel, PlVars) :-
    \+ backjump_level(_),
    max_heuristic_values(StateMutable, MaxScoreAsMin, SatVarNames),
    get_decision_variable_from_list(SatVarNames, SatVarName),
    delete_state_instance(StateMutable, heuristic_heap, (SatVarName, MaxScoreAsMin)),
    delete_state_instance(StateMutable, heuristic_avl, (SatVarName, MaxScoreAsMin)),
    get_pl_var_for_sat_id(SatVarName, PlVars, Var),
    NextLevel is CurLevel + 1,
    (   saved_phase(SatVarName, _)
    ->  elim_var_from_saved_phase(SatVarName, Var, SolverEnv, StateMutable, CurLevel, PlVars)
    ;   % uninformed phase/polarity selection
        (   bt_log_sat_stack(StateMutable, SatVarName, branch, CurLevel, pred_true, _OriginalClause),
            format("Decide variable ~w with polarity ~w on level ~w~n", [SatVarName,pred_true,CurLevel]),
            Var = pred_true,
            update_full_state(StateMutable, cur_level, NextLevel),
            elim_vars(SolverEnv, StateMutable, PlVars)
        ;   \+ backjump_level(_),
            %learn_idl_core(StateMutable, CurLevel, SolverEnv),
            remove_idl_unsat_core,
            bt_log_sat_stack(StateMutable, SatVarName, branch, CurLevel, pred_false, _OriginalClause),
            format("Decide variable ~w with polarity ~w on level ~w~n", [SatVarName,pred_false,CurLevel]),
            Var = pred_false,
            update_full_state(StateMutable, cur_level, NextLevel),
            elim_vars(SolverEnv, StateMutable, PlVars)
        )
    ).

%% elim_var_from_saved_phase(+SatVarName, +Var, +SolverEnv, +StateMutable, +CurLevel, +PlVars).
% Restore phase/polarity for +SatVarName from cache for rapid restart.
elim_var_from_saved_phase(SatVarName, Var, SolverEnv, StateMutable, CurLevel, PlVars) :-
    bt_retract(saved_phase(SatVarName, SavedPol)),
    get_pl_var_for_sat_id(SatVarName, PlVars, Var),
    NextLevel is CurLevel + 1,
    (   bt_log_sat_stack(StateMutable, SatVarName, branch, CurLevel, SavedPol, _OriginalClause),
        format("Phase Restore: Decide variable ~w with polarity ~w on level ~w~n", [SatVarName,SavedPol,CurLevel]),
        Var = SavedPol,
        update_full_state(StateMutable, cur_level, NextLevel),
        elim_vars(SolverEnv, StateMutable, PlVars)
    ;   \+ backjump_level(_),
        remove_idl_unsat_core,
        %learn_idl_core(StateMutable, CurLevel, SolverEnv),
        negate_polarity(SavedPol, NegSavedPol),
        bt_log_sat_stack(StateMutable, SatVarName, branch, CurLevel, NegSavedPol, _OriginalClause),
        format("Phase Restore Failed: Decide variable ~w with polarity ~w on level ~w~n", [SatVarName,NegSavedPol,CurLevel]),
        Var = NegSavedPol,
        update_full_state(StateMutable, cur_level, NextLevel),
        elim_vars(SolverEnv, StateMutable, PlVars)
    ).

problem_setup(_, _, []).
problem_setup(SolverEnv, StateMutable, [Clause|Clauses]) :-
    clause_setup(SolverEnv, StateMutable, Clause),
    problem_setup(SolverEnv, StateMutable, Clauses).

clause_setup(SolverEnv, StateMutable, Clause) :-
    Clause = [Pol-Var-SatVarName|Triples],
    set_watch(SolverEnv, StateMutable, Clause, Triples, Var, Pol, SatVarName).

set_watch(_, StateMutable, OriginalClause, [], Var, Pol, SatVarName) :-
    Var == Pol,
    delete_state_instance(StateMutable, heuristic_avl, (SatVarName, Score)), % avl tree is used to get the key in the heap
    delete_state_instance(StateMutable, heuristic_heap, (SatVarName, Score)),
    bt_retract_if_exists(saved_phase(SatVarName, _)),
    worsen_scores_for_vars_in_clause(StateMutable, OriginalClause).
set_watch(SolverEnv, StateMutable, OriginalClause, [], Var, Pol, _) :-
    ground(Var),
    Var \== Pol,
    format('sat conflict~n', []),
    get_complete_theory_lemma(SolverEnv, TheoryLemmaClause),
    possibly_bt_from_smt_solution(StateMutable, TheoryLemmaClause),
    get_full_state(StateMutable, current_stack_size, StackSize),
    compute_and_assert_backjump_clause(SolverEnv, StateMutable, StackSize, OriginalClause), % TO DO: TheoryLemmaClause?
    format('computed backjump clause~n', []),
    fail.
set_watch(SolverEnv, StateMutable, OriginalClause, [], Var, Pol, SatVarName) :-
    var(Var),
    get_full_state(StateMutable, cur_level, CurLevel),
    format("unit propagation of ~w with polarity ~w on level ~w~n", [SatVarName,Pol,CurLevel]),
    bt_log_sat_stack(StateMutable, SatVarName, unit, CurLevel, Pol, OriginalClause),
    (   Var = Pol,
        delete_state_instance(StateMutable, heuristic_avl, (SatVarName, Score)),
        delete_state_instance(StateMutable, heuristic_heap, (SatVarName, Score)),
        bt_retract_if_exists(saved_phase(SatVarName, _)),
        worsen_scores_for_vars_in_clause(StateMutable, OriginalClause),
        (   true
        ;   % remove idl unsat core from higher decision level on backtracking
            remove_idl_unsat_core,
            fail
        )
    ;   % theory conflict: if Var = Pol triggered a SAT conflict instead,
        % a backjump clause has been computed and backjump_level/1 is asserted
        \+ backjump_level(_), % skip when backjumping
        get_last_decision_level(StateMutable, LastDLevel), % skip if no decision so far
        LastDLevel \== 1, % just backtrack if last decision level is 1
        get_complete_theory_lemma(SolverEnv, TheoryLemmaClause),
        possibly_bt_from_smt_solution(StateMutable, TheoryLemmaClause),
        format("theory conflict on level ~w~n", [CurLevel]),
        get_conflict_clause_for_theory_conflict(SatVarName, Pol, SolverEnv, ConflictClause),
        nl,write('Conflict Clause: '), write(ConflictClause),nl,
        get_full_state(StateMutable, current_stack_size, StackSize),
        compute_and_assert_backjump_clause_get_level(SolverEnv, StateMutable, StackSize, ConflictClause, _BjLevel),
        %get_full_state(StateMutable, sat_stack, SatStack),
        %compute_lbd_score(SatStack, TheoryLemmaClause, LbdScore)
        %asserta(backjump_clause(BjLevel, TheoryLemmaClause, LbdScore)), % always learn theory lemma from unsat SMT formula
        fail
    ).
set_watch(SolverEnv, StateMutable, OriginalClause, [Pol2-Var2-Name2|Triples], Var1, Pol1, Name1):-
    watch(SolverEnv, StateMutable, OriginalClause, Var1, Pol1, Name1, Var2, Pol2, Name2, Triples).

:- block watch(?, ?, ?, -, ?, ?, -, ?, ?, ?).
watch(SolverEnv, StateMutable, OriginalClause, Var1, Pol1, Name1, Var2, Pol2, Name2, Triples) :-
    (   nonvar(Var1)
    ->  update_watch(SolverEnv, StateMutable, OriginalClause, Var1, Pol1, Name1, Var2, Pol2, Name2, Triples)
    ;   update_watch(SolverEnv, StateMutable, OriginalClause, Var2, Pol2, Name2, Var1, Pol1, Name1, Triples)
    ).

update_watch(SolverEnv, StateMutable, OriginalClause, Var1, Pol1, Name1, Var2, Pol2, Name2, Triples) :-
    (   Var1 == Pol1
    ->  delete_state_instance(StateMutable, heuristic_avl, (Name1, Score)),
        delete_state_instance(StateMutable, heuristic_heap, (Name1, Score)),
        bt_retract_if_exists(saved_phase(Name1, _)),
        worsen_scores_for_vars_in_clause(StateMutable, OriginalClause)
    ;   set_watch(SolverEnv, StateMutable, OriginalClause, Triples, Var2, Pol2, Name2)
    ).

retract_once(Assertion) :-
    retract(Assertion),
    !.

%% is_backjumping.
is_backjumping :-
    backjump_level(_).

%% is_last_propagated_sat_var(+StateMutable, +SatVarName).
is_last_propagated_sat_var(StateMutable, SatVarName) :-
    get_full_state(StateMutable, sat_stack, Stack),
    % TO DO: improve
    avl_member(SatVarName, Stack, _).

log_theory_propagation_sat_stack(StateMutable, SatVarName, SatPrologVar) :-
    get_full_state(StateMutable, cur_level, CurLevel),
    !,
    format("Theory propagation of ~w with polarity ~w~n", [SatVarName,SatPrologVar]),
    % explanation (OriginalClause) is computed lazily for conflict analysis if necessary, see explain_theory_propagation/5
    add_state_instance(StateMutable, sat_stack, (SatVarName,theory,CurLevel,SatPrologVar,_OriginalClause)).

% Calculating the unsat core in the idl solver is no overhead so it might be worthwhile to learn
% unsat cores on the decision level as well.
learn_idl_core(StateMutable, CurLevel, SolverEnv) :-
    retract(idl_unsat_core(Core)),
    !,
    get_conflict_clause_from_unsat_core(Core, SolverEnv, ConflictClause),
    get_full_state(StateMutable, sat_stack, SatStack),
    compute_lbd_score(SatStack, ConflictClause, LbdScore),
    asserta(backjump_clause(CurLevel, ConflictClause, LbdScore)).
learn_idl_core(_, _, _).

% Log variable assignments.
% Effect is undone on backtracking. If backjump level is zero on backtracking, i.e.,
% a random restart occured or a unit-clause has been learned, save the phase to restore
% assignments afterwards for a rapid restart.
bt_log_sat_stack(StateMutable, SatVarName, PropType, CurLevel, Pol, OriginalClause) :-
    add_state_instance(StateMutable, sat_stack, (SatVarName, PropType, CurLevel, Pol, OriginalClause)),
    (   increase_stack_size(StateMutable)
    ;   is_restart,
        asserta_once(saved_phase(SatVarName, Pol)),
        fail
    ).

increase_stack_size(StateMutable) :-
    get_full_state(StateMutable, current_stack_size, Old),
    New is Old + 1, !,
    update_full_state(StateMutable, current_stack_size, New).

%% possibly_bt_from_smt_solution(+StateMutable, +TheoryLemmaClause).
% Backtracking from an SMT solution which should not be handled as a theory conflict.
possibly_bt_from_smt_solution(StateMutable, TheoryLemmaClause) :-
    retract(bt_from_smt_solution),
    !,
    % backjump to the last decision level without computing a conflict clause
    % but excluding the last SAT solution when backtracking after finding an SMT solution
    get_last_decision_level(StateMutable, Level),
    asserta(exclude_sat_solution_clause(TheoryLemmaClause)),
    asserta(backjump_level(Level)),
    fail.
possibly_bt_from_smt_solution(_, _).

negate_smt_formula(b(negation(Smt),pred,[]), NegSmt) :-
    !,
    NegSmt = Smt.
negate_smt_formula(Smt, b(negation(Smt),pred,[])).

%% explain_theory_propagation(+SolverEnv, +SatStack, +TheoryPropSatVar, +TheoryPropLevel, +TheoryPropPol, -ExplainingClause).
explain_theory_propagation(SolverEnv, SatStack, TheoryPropSatVar, TheoryPropLevel, TheoryPropPol, ExplainingClause) :-
    avl_to_list(SatStack, SatStackList),
    findall((SatVarName-Pol), (member(SatVarName-(_,Level,Pol,_), SatStackList), Level<TheoryPropLevel), SatVarNameTuples),
    get_from_solver_env(SolverEnv, sat_vars, SatVars),
    get_smt_formula_for_sat_pol(SatVars, TheoryPropSatVar, TheoryPropPol, SmtFormula),
    negate_smt_formula(SmtFormula, NegSmt),
    get_explanation_candidate_formula(SatVars, SatVarNameTuples, CandidateFormula),
    (   SatVarNameTuples = [_]
    ->  UnsatCore = b(conjunct(CandidateFormula,NegSmt),pred,[])
    ;   temporary_set_unsat_core_timeout,
        unsat_core_with_fixed_conjuncts(CandidateFormula, NegSmt, TUnsatCore),
        UnsatCore = b(conjunct(TUnsatCore,NegSmt),pred,[]),
        preferences:reset_temporary_preference(time_out)
    ),
    nl, write('Explanation Core: '), nl,translate:print_bexpr(UnsatCore),nl,
    get_from_solver_env(SolverEnv, t2b_env, T2BEnv),
    predicate_to_sat(only_reuse, SatVars, T2BEnv, UnsatCore, _NewEnv, _AnnotatedPred, UnsatCoreSat, _NewSatVars),
    sat_conj_negate_to_clause(UnsatCoreSat, TExplainingClause),
    get_from_solver_env(SolverEnv, pl_vars, PlVars),
    set_variable_references_in_clause(PlVars, TExplainingClause, ExplainingClause).% TO DO: learn theory lemma?

get_explanation_candidate_formula(SatVars, [SatVarName-Pol|T], CandidateFormula) :-
    get_smt_formula_for_sat_pol(SatVars, SatVarName, Pol, SmtFormula),
    get_explanation_candidate_formula_acc(SatVars, T, SmtFormula, CandidateFormula).

get_explanation_candidate_formula_acc(_, [], Acc, Acc).
get_explanation_candidate_formula_acc(SatVars, [SatVarName-Pol|T], Acc, CandidateFormula) :-
    get_smt_formula_for_sat_pol(SatVars, SatVarName, Pol, SmtFormula),
    NAcc = b(conjunct(SmtFormula,Acc),pred,[]),
    get_explanation_candidate_formula_acc(SatVars, T, NAcc, CandidateFormula).

get_smt_formula_for_sat_pol(SatVars, SatVarName, Pol, SmtFormula) :-
    memberchk(b(identifier(SatVarName),_,Info), SatVars),
    member(smt_formula(TSmtFormula), Info),
    (   Pol == pred_true
    ->  SmtFormula = TSmtFormula
    ;   TSmtFormula = b(negation(SmtFormula),pred,[])
    ->  true
    ;   SmtFormula = b(negation(TSmtFormula),pred,[])
    ).

%% max_heuristic_values(+StateMutable, -MaxScoreAsMin, -SatVarNames).
% Heuristic values are multiplied by -1 since we use a min heap.
max_heuristic_values(StateMutable, MaxScoreAsMin, SatVarNames) :-
    get_full_state(StateMutable, heuristic_heap, Heap),
    get_from_heap(Heap, MaxScoreAsMin, SatVarName, NewHeap),
    get_heuristic_values_for_score(NewHeap, MaxScoreAsMin, RSatVarNames), !,
    SatVarNames = [SatVarName|RSatVarNames].

get_heuristic_values_for_score(Heap, Score, [SatVarName|NT]) :-
    get_from_heap(Heap, MinScore, SatVarName, NewHeap),
    MinScore == Score,
    !,
    get_heuristic_values_for_score(NewHeap, Score, NT).
get_heuristic_values_for_score(_, _, []).

% Skip on backjumping.
get_decision_variable_from_list(SatVarNames, SatVarName) :-
    member(SatVarName, SatVarNames),
    \+ backjump_level(_).

bt_retract_if_exists(Assertion) :-
    (   Assertion
    ->  bt_retract(Assertion)
    ;   true
    ).

% Not undone on backtracking.
/*
If this is used we need to store the mutable since operations are undone on backtracking.
possibly_decay_heuristic_values(StateMutable) :-
    decision_heuristic(Heuristic),
    Heuristic == vsids,
    conflict_stats(total(AC), _, _),
    AC \== 0,
    vsids_decay_frequency(I),
    C is AC mod I,
    C == 0,
    format("~nDecay heuristic scores~n", []),
    !,
    vsids_decay_value(Val),
    decay_heuristic_values(StateMutable, Val).
possibly_decay_heuristic_values(_).

% Not undone on backtracking.
decay_heuristic_values(StateMutable, DecayValue) :-
    retract(heuristic_tuple(SatVarName, Score)),
    !,
    NewScore is Score / DecayValue,
    decay_heuristic_values(DecayValue),
    asserta(heuristic_tuple(SatVarName, NewScore)).
decay_heuristic_values(_, _).*/

% Each variable in each polarity has a counter, initialized to 0.
/*init_heuristic(SolverEnv) :-
    get_from_solver_env(SolverEnv, sat_varnames, SatVarNames),
    init_heuristic_tuples_from_vars(SatVarNames).

init_heuristic_tuples_from_vars([]).
init_heuristic_tuples_from_vars([SatVarName|T]) :-
    asserta(heuristic_tuple(SatVarName, 0)),
    %asserta(heuristic_tuple(SatVarName, pred_true, 0)),
    %asserta(heuristic_tuple(SatVarName, pred_false, 0)),
    init_heuristic_tuples_from_vars(T).*/

% Each variable is initially assigned the number of its occurences in all clauses as its score.
% Variable scores are not split for polarities.
init_decision_heuristic(StateMutable, Clauses) :-
    count_occurences_in_clauses(Clauses, VarScoreTriples),
    add_heuristic_tuples(StateMutable, VarScoreTriples).

add_heuristic_tuples(_, []).
add_heuristic_tuples(StateMutable, [(SatVarName,Score)|T]) :-
    add_state_instance(StateMutable, heuristic_avl, (SatVarName,Score)),
    add_state_instance(StateMutable, heuristic_heap, (SatVarName,Score)),
    add_heuristic_tuples(StateMutable, T).

:- dynamic occurences/2.
:- volatile occurences/2.

count_occurences_in_clauses(Clauses, VarScoreTriples) :-
    retractall(occurences(_, _)),
    assert_count_occurences(Clauses),
    findall((SatVarName,Score), retract(occurences(SatVarName, Score)), VarScoreTriples),!.

assert_count_occurences([]).
assert_count_occurences([Clause|T]) :-
    assert_count_occurences_from_clause(Clause),
    assert_count_occurences(T).

% TO DO: use avl tree instead of assert
assert_count_occurences_from_clause([]).
assert_count_occurences_from_clause([_-_-SatVarName|T]) :-
    (   retract(occurences(SatVarName, Score))
    ->  Score1 is Score + 1,
        asserta(occurences(SatVarName, Score1))
    ;   asserta(occurences(SatVarName, 1))
    ),
    assert_count_occurences_from_clause(T).

% Update heuristic scores from learned clauses for VSIDS branching heuristic.
update_heuristic_scores_for_clauses(StateMutable, ClausesToLearn) :-
    get_non_ground_varnames_occurring_in_clauses(ClausesToLearn, SatVarNames),
    decision_heuristic(Type),
    update_heuristic_scores(StateMutable, Type, SatVarNames).

get_non_ground_varnames_occurring_in_clauses(ClausesToLearn, SatVarNames) :-
    get_non_ground_varnames_occurring_in_clauses(ClausesToLearn, [], SatVarNames).

get_non_ground_varnames_occurring_in_clauses([], Acc, Acc).
get_non_ground_varnames_occurring_in_clauses([Clause|T], Acc, SatVarNames) :-
    get_non_ground_varnames_occurring_in_clause(Clause, Acc, NewAcc),
    get_non_ground_varnames_occurring_in_clauses(T, NewAcc, SatVarNames).

get_non_ground_varnames_occurring_in_clause([], Acc, Acc).
get_non_ground_varnames_occurring_in_clause([_-Var-SatVarName|T], Acc, SatVarNames) :-
    var(Var),
    !,
    ord_add_element(Acc, SatVarName, NewAcc),
    get_non_ground_varnames_occurring_in_clause(T, NewAcc, SatVarNames).
get_non_ground_varnames_occurring_in_clause([_|T], Acc, SatVarNames) :-
    get_non_ground_varnames_occurring_in_clause(T, Acc, SatVarNames).

%% update_heuristic_scores(+StateMutable, +Type, +SatVarNames).
% Type is either vsids or evsids.
% SatVarNames are the variable names occurring in learned clauses.
% Effect is undone on backtracking.
update_heuristic_scores(StateMutable, Type, SatVarNames) :-
    conflict_stats(total(TotalConflicts), _, _),
    update_heuristic_scores(StateMutable, TotalConflicts, Type, SatVarNames).

update_heuristic_scores(_, _, _, []).
update_heuristic_scores(StateMutable, TotalConflicts, Type, [SatVarName|T]) :-
    get_full_state(StateMutable, heuristic_avl, Avl),
    avl_member(SatVarName, Avl, OldScore1),
    delete_state_instance(StateMutable, heuristic_heap, (SatVarName,OldScore1)),
    OldScore is OldScore1 * -1,
    bump_heuristic_score_for_type(Type, TotalConflicts, OldScore, NewScore),
    add_state_instance(StateMutable, heuristic_heap, (SatVarName,NewScore)),
    NewScore1 is NewScore * -1,
    avl_change(SatVarName, Avl, _, NAvl, NewScore1),
    update_full_state(StateMutable, heuristic_avl, NAvl),
    update_heuristic_scores(StateMutable, TotalConflicts, Type, T).

% v_s + 1
bump_heuristic_score_for_type(vsids, _, OldScore, NewScore) :-
    NewScore is OldScore + 1.
% v_s + f^(-k), where f is evsids_f_value/1 and it is the k-th conflict
bump_heuristic_score_for_type(evsids, TotalConflicts, OldScore, NewScore) :-
    evsids_f_value(FVal),
    NewScore is OldScore + FVal**(-TotalConflicts).

get_pl_var_for_sat_id(SatVarName, PlVars, Var) :-
    member((Var-SatVarName-_), PlVars). % TO DO: improve?

% We have positive scores, maximum value is best. Remove if zero.
% Effect is undone on backtracking.
worsen_scores_for_vars_in_clause(_, []).
worsen_scores_for_vars_in_clause(StateMutable, [_-_-SatVarName|T]) :-
    get_full_state(StateMutable, heuristic_avl, Avl),
    (   avl_member(SatVarName, Avl, Score)
    ->  (   Score \== -1,
            NScore is Score + 1,
            NScore1 is NScore * -1,
            avl_change(SatVarName, Avl, _, NAvl, NScore),
            update_full_state(StateMutable, heuristic_avl, NAvl),
            delete_state_instance(StateMutable, heuristic_heap, (SatVarName, Score)),
            add_state_instance(StateMutable, heuristic_heap, (SatVarName, NScore1))
        ;   Score == -1,
            delete_state_instance(StateMutable, heuristic_avl, (SatVarName, Score)),
            delete_state_instance(StateMutable, heuristic_heap, (SatVarName, Score))
        ),
        worsen_scores_for_vars_in_clause(StateMutable, T)
    ;   worsen_scores_for_vars_in_clause(StateMutable, T)
    ).

% Literal Block Distance (LBD) as defined in "Predicting Learnt Clauses Quality in Modern SAT Solvers",
% i.e., the amount of different decision levels in a clause.
compute_lbd_score(SatStack, BjClause, LbdScore) :-
    compute_lbd_score(SatStack, BjClause, [], LbdScore).

compute_lbd_score(_, [], Acc, LbdScore) :-
    length(Acc, LbdScore).
compute_lbd_score(SatStack, [_-_-SatVarName|T], Acc, LbdScore) :-
    avl_member(SatVarName, SatStack, (_,Level,_,_)),
    ord_add_element(Acc, Level, NewAcc),
    compute_lbd_score(SatStack, T, NewAcc, LbdScore).

% Remove half of the clauses every "frequency_constant + constant_factor * x" conflicts,
% where x is the amount of clause reductions so far. We keep clauses with a small LBD forever.
possibly_discard_learned_clauses :-
    conflict_stats(_, _, since_reduction(SRed)),
    dpllt_settings:discard_learned_clauses_frequency_constant(Frequency),
    SRed >= Frequency, % first check to improve performance
    amount_clause_reductions(Reductions),
    dpllt_settings:discard_learned_clauses_constant_factor(Factor),
    Threshold is Frequency + Factor * Reductions,
    SRed == Threshold,
    !,
    format("~nDiscard half of learned clauses..", []),
    dpllt_settings:discard_clause_greater_lbd(LbdLimit),
    amount_learned_clauses(ALC),
    Half is ALC // 2,
    discard_half_of_learned_clauses_greater_lbd(0, Half, LbdLimit, Removed),
    retract(amount_learned_clauses(Old)),
    New is Old - Removed,
    asserta(amount_learned_clauses(New)),
    retract(conflict_stats(total(Total), since_restart(SRed), _)),
    asserta(conflict_stats(total(Total), since_restart(SRed), since_reduction(0))), !.
possibly_discard_learned_clauses.

% Remove half of the clauses learned so far.
% We keep clauses with an LBD smaller than dpllt_settings:discard_clause_greater_lbd/1
% forever. Thus it is not guaranteed that exactly half of the clauses are deleted if
% a lot of high quality clauses have been learned.
discard_half_of_learned_clauses_greater_lbd(C, Half, LbdLimit, Removed) :-
    C =< Half,
    backjump_clause(BjLevel, BjClause, LbdScore),
    !,
    retract(backjump_clause(BjLevel, BjClause, LbdScore)),
    (   LbdScore > LbdLimit
    ->  C1 is C + 1,
        discard_half_of_learned_clauses_greater_lbd(C1, Half, LbdLimit, Removed)
    ;   discard_half_of_learned_clauses_greater_lbd(C, Half, LbdLimit, Removed),
        asserta(backjump_clause(BjLevel, BjClause, LbdScore))
    ).
discard_half_of_learned_clauses_greater_lbd(C, _, _, C).

%% count_found_conflict(-NewTotal, -NewSinceRestart, -NewSinceReduction).
count_found_conflict(Total1, SRes1, SRed1) :-
    retract(conflict_stats(total(Total), since_restart(SRes), since_reduction(SRed))),
    Total1 is Total + 1,
    SRes1 is SRes + 1,
    SRed1 is SRed + 1,
    asserta(conflict_stats(total(Total1), since_restart(SRes1), since_reduction(SRed1))).

%% compute_and_assert_backjump_clause(+SolverEnv, +StateMutable, +StackSize, -ConflictClause).
compute_and_assert_backjump_clause(SolverEnv, StateMutable, StackSize, ConflictClause) :-
    compute_and_assert_backjump_clause_get_level(SolverEnv, StateMutable, StackSize, ConflictClause, _).

%% compute_and_assert_backjump_clause(+SolverEnv, +StateMutable, +StackSize, -ConflictClause, -BjLevel).
compute_and_assert_backjump_clause_get_level(SolverEnv, StateMutable, StackSize, ConflictClause, BjLevel) :-
    possibly_discard_learned_clauses,
    get_backjump_clause_conflict_graph_analysis(SolverEnv, StateMutable, ConflictClause, BjClause, BjLevel),
    get_full_state(StateMutable, sat_stack, SatStack),
    compute_lbd_score(SatStack, BjClause, LbdScore),
    asserta(backjump_clause(BjLevel, BjClause, LbdScore)),
    asserta(backjump_level(BjLevel)),
    count_found_conflict(NewTotal, NewSinceRestart, _),
    store_recent_stack_size(StackSize),
    count_learned_clause,
    nl, writeq('Backjump clause: '), nl, writeq(BjClause), nl,
    nl, writeq('Backjump level: '), writeq(BjLevel), nl,
    (   \+ clear_lbd_values_or_fail(StateMutable),
        BjLevel \== 0
    ->  store_recent_lbd_score(LbdScore),
        update_total_lbd_sum(LbdScore),
        (random_restart(NewTotal, NewSinceRestart), !; true)
    ;   true
    ).

count_learned_clause :-
    retract(amount_learned_clauses(Old)),
    New is Old + 1,
    asserta(amount_learned_clauses(New)).

%% store_recent_lbd_score(+LbdScore).
% The stored list of recent LBDs is a difference list and we insert at the end of list.
% The first LBD is thus the oldest one and removed when exceeding the maximum size
% defined by recent_lbd_threshold/1.
store_recent_lbd_score(LbdScore) :-
    retract(recent_lbds_dl(Lbds-Diff, Amount)),
    glucose_restart_recent_lbds_threshold(Thresh),
    Diff = [LbdScore|NDiff],
    (   Amount < Thresh
    ->  Amount1 is Amount + 1,
        asserta(recent_lbds_dl(Lbds-NDiff, Amount1))
    ;   Lbds = [_|RLbds],
        asserta(recent_lbds_dl(RLbds-NDiff, Amount))
    ).

store_recent_stack_size(StackSize) :-
    retract(recent_stack_sizes_dl(StackSizes-Diff, Amount)),
    glucose_restart_trail_threshold(Thresh),
    Diff = [StackSize|NDiff],
    (   Amount < Thresh
    ->  Amount1 is Amount + 1,
        asserta(recent_stack_sizes_dl(StackSizes-NDiff, Amount1))
    ;   StackSizes = [_|RStackSizes],
        asserta(recent_stack_sizes_dl(RStackSizes-NDiff, Amount))
    ).

%% update_total_lbd_sum(+LbdScore).
update_total_lbd_sum(LbdScore) :-
    retract(total_lbd_sum(Old)),
    New is Old + LbdScore,
    asserta(total_lbd_sum(New)).

%% random_restart(+AmountOfConflicts, +AmountOfConflictsSinceRestart).
% "Randomly" restart after an amount of conflicts since the last restart.
random_restart(_AC, ACR) :-
    restart_policy(luby),
    random_luby_restart(ACR), !.
random_restart(AC, _ACR) :-
    restart_policy(glucose),
    random_glucose_restart(AC), !.

do_random_restart :-
    retract(backjump_level(_)),
    asserta(backjump_level(0)),
    retract(conflict_stats(total(Total), _, since_reduction(SRed))),
    asserta(conflict_stats(total(Total), since_restart(0), since_reduction(SRed))),
    asserta_once(is_restart),
    format("~nRandom restart~n", []).

%% random_glucose_restart(+AmountOfConflicts).
% Restart search if recent_lbds_dl/2 contains exactly glucose_restart_recent_lbds_threshold/1 amount
% of most recent LBDs and RecentAvg * Proportion > TotalAvg is satisfied, i.e., the recently learned clauses'
% LBDs got too large.
random_glucose_restart(AC) :-
    total_lbd_sum(TotalLbdSum), % TO DO: reset the sum at any time?
    TotalAvg is TotalLbdSum / AC,
    recent_lbds_dl(RLbds-[], LbdAmount),
    glucose_restart_recent_lbds_threshold(LbdThresh),
    LbdAmount == LbdThresh, % queue of recent LBDs is full
    sumlist(RLbds, RSum),
    RecentAvg is RSum / LbdAmount,
    glucose_restart_margin_ratio(Proportion),
    format("~nTotalAvg: ~w~nRecentAvg: ~w~n", [TotalAvg,RecentAvg]),
    (   RecentAvg * Proportion > TotalAvg
    ->  do_random_restart,
        clear_recent_lbds_state
    ;   true
    ).

% Favor SAT instances too by possibly clearing the recent LBD queue.
% Assumption: The queue of recent LBDs is full.
clear_lbd_values_or_fail(StateMutable) :-
    glucose_restart_trail_threshold(TrailThresh),
    glucose_restart_stack_avg_factor(StackSizeFactor),
    recent_stack_sizes_dl(StackSizes-[], AmountStackSizes),
    AmountStackSizes == TrailThresh, % queue of recent stack sizes on conflict is full
    get_full_state(StateMutable, current_stack_size, StackSize),
    sumlist(StackSizes, SSum),
    StackSizesAvg is SSum / AmountStackSizes,
    StackSize > StackSizeFactor * StackSizesAvg,
    clear_recent_lbds_state.

clear_recent_lbds_state :-
    retract(recent_lbds_dl(_, _)),
    asserta(recent_lbds_dl(Diff-Diff, 0)).

%% random_luby_restart(+AmountOfConflictsSinceRestart).
% The amount of conflicts when to restart is defined by the Luby sequence,
% i.e., 1,1,2,1,1,2,4,1,1,2,1,1,2,4,8,.. etc., (cf. "Optimal speedup of Las Vegas algorithms")
% using a unit length of 32 (see luby_counter/1).
random_luby_restart(ACR) :-
    restart_policy(luby),
    get_next_luby_value(Luby),
    !,
    (   ACR == Luby
    ->  retract_once(luby_counter(_)),
        do_random_restart
    ;   true
    ).
random_luby_restart(_) :-
    add_error_fail(random_restart, 'No more values in predefined Luby sequence', []).

get_next_luby_value(LubyValue) :-
    luby_counter(Luby),
    luby_restart_unit_length(UnitLength),
    LubyValue is UnitLength * Luby, % TO DO: faster to precompute once a unit length is fixed
    !.

%% learn_asserted_clauses(+CurrentBjLevel, +SolverEnv, +Clauses, -NewClauses).
learn_asserted_clauses(CurrentBjLevel, SolverEnv, Clauses, NewClauses) :-
    get_clauses_to_learn(CurrentBjLevel, SolverEnv, ClausesToLearn),
    append(ClausesToLearn, Clauses, NewClauses).

% We exclude SAT solutions when backtracking from finding an SMT solution to find a different assignment.
get_sat_solution_exclusion_clauses(ExclClausesNoRef) :-
    findall(ExclClause, exclude_sat_solution_clause(ExclClause), ExclClausesNoRef), !.

%% get_clauses_to_learn(+CurrentBjLevel, +SolverEnv, -ClausesToLearn).
% Note that the Prolog variable references are set correctly in here.
% We thus do not have to take care for correct references when analyzing a conflict
% and creating a backjump clause beforehand.
% We learn all backjump clauses which have been learned on a decision level higher or equal to CurrentBjLevel.
% Note that the clauses learned on a lower level are already included in the set of clauses.
% Yet, when backjumping, we do not want to forget clauses learned on a decision level between the current backjump
% level and the last conflict level.
get_clauses_to_learn(CurrentBjLevel, SolverEnv, ClausesToLearn) :-
    get_asserted_backjump_clauses_greater_eq_level(CurrentBjLevel, TempBjClauses),
    length(TempBjClauses, Len),
    format("Learn ~w clauses~n", [Len]),
    get_sat_solution_exclusion_clauses(ExclClauses),
    append(ExclClauses, TempBjClauses, BjClausesNoRef),
    get_from_solver_env(SolverEnv, pl_vars, PlVars),
    set_variable_references_in_clauses(PlVars, BjClausesNoRef, ClausesToLearn).

% Get and remove, but undo on backtracking.
get_asserted_backjump_clauses_greater_eq_level(CurrentBjLevel, BjClauses) :-
    get_asserted_backjump_clauses_greater_eq_level(CurrentBjLevel, [], BjClauses).

get_asserted_backjump_clauses_greater_eq_level(CurrentBjLevel, Acc, BjClauses) :-
    (   backjump_clause(BjLevel, BjClause, _)
    ->  bt_retract(backjump_clause(BjLevel, BjClause, _)),
        (   BjLevel >= CurrentBjLevel
        ->  get_asserted_backjump_clauses_greater_eq_level(CurrentBjLevel, [BjClause|Acc], BjClauses)
        ;   get_asserted_backjump_clauses_greater_eq_level(CurrentBjLevel, Acc, BjClauses),
            bt_asserta(backjump_clause(BjLevel, BjClause, _))
        )
    ;   BjClauses = Acc
    ).

%% set_variable_references_in_clauses(+PlVars, +Clauses, -RefClauses).
set_variable_references_in_clauses(_, [], []).
set_variable_references_in_clauses(PlVars, [Clause|T], [RefClause|NT]) :-
    set_variable_references_in_clause(PlVars, Clause, RefClause),
    set_variable_references_in_clauses(PlVars, T, NT).

% We need to set the correct variable references in a new learned clause
% to trigger co-routines etc.
set_variable_references_in_clause(_, [], []).
set_variable_references_in_clause(PlVars, [Pol-_-Name|T], [Pol-Var-Name|NT]) :-
    member(Var-Name-_, PlVars),
    !,
    set_variable_references_in_clause(PlVars, T, NT).
set_variable_references_in_clause(PlVars, [_|T], NT) :-
    % symmetry breaking might introduce new SAT variables which are not connected to any SMT variable
    set_variable_references_in_clause(PlVars, T, NT).

get_from_solver_env(SolverEnv, Key, OutValue) :-
    member(Key:Value, SolverEnv),
    !,
    OutValue = Value.
get_from_solver_env(SolverEnv, Key, _) :-
    add_error_fail(get_from_solver_env,
        'Wrong key or corrupt solver environment in DPLL(T) SAT solver',
        [env:SolverEnv,key:Key]).

bt_asserta(Assertion) :-
    asserta(Assertion); (retract(Assertion), fail).

bt_retract(Assertion) :-
    (   retract(Assertion)
    ->  (   true
        ;   asserta(Assertion),
            fail
        )
    ;   fail
    ).

%% get_conflict_clause_for_theory_conflict(+LastVarName, +LastVarValue, +SolverEnv, -ConflictClause).
% Translate current sat solution, which is a theory conflict, to the theory and compute
% the (minimum) unsat core. Return the negated unsat core as a conflict clause.
% Note that an SMT formula passed to the theory solver is always a conjunction and
% thus the negated unsat core is a disjunction (clause).
% Assumption: unsat core computation only removes constraints and does not deduce new ones.
get_conflict_clause_for_theory_conflict(LastVarName, LastVarValue, SolverEnv, ConflictClause) :-
    get_from_solver_env(SolverEnv, sat_bindings, SatBindings),
    get_from_solver_env(SolverEnv, sat_vars, SatVars),
    (   retract(idl_unsat_core(UnsatCore))
    ->  true
    ;   temporary_set_unsat_core_timeout,
        select(bind(LastVarName,_),SatBindings,RestBindings),
        sat_to_predicate_from_solution(RestBindings, SatVars, SmtFormula),
        sat_to_predicate_from_solution([bind(LastVarName,LastVarValue)], SatVars, FixedConj),
        unsat_core_with_fixed_conjuncts(SmtFormula, FixedConj, UnsatCore),
        preferences:reset_temporary_preference(time_out)
    ),
    get_conflict_clause_from_unsat_core(UnsatCore, SolverEnv, ConflictClause),
    !.

get_conflict_clause_from_unsat_core(UnsatCore, SolverEnv, ConflictClause) :-
    get_from_solver_env(SolverEnv, sat_vars, SatVars),
    get_from_solver_env(SolverEnv, t2b_env, T2BEnv),
    get_from_solver_env(SolverEnv, pl_vars, PlVars),
    predicate_to_sat(only_reuse, SatVars, T2BEnv, UnsatCore, _NewEnv, _AnnotatedPred, UnsatCoreSat, _NewSatVars),
    nl, write('Core: '),
    translate:print_bexpr(UnsatCore),nl,
    translate:print_bexpr(UnsatCoreSat),nl,
    sat_conj_negate_to_clause(UnsatCoreSat, TConflictClause),
    set_variable_references_in_clause(PlVars, TConflictClause, ConflictClause), !.

% Return the complete unsatisfiable SMT formula as a conflict clause.
get_complete_theory_lemma(SolverEnv, TheoryLemmaClause) :-
    get_from_solver_env(SolverEnv, pl_vars, PlVars),
    get_from_solver_env(SolverEnv, sat_bindings, SatBindings),
    get_clause_from_negated_ground_bindings(SatBindings, TTheoryLemmaClause),
    set_variable_references_in_clause(PlVars, TTheoryLemmaClause, TheoryLemmaClause).

get_clause_from_negated_ground_bindings(SatBindings, TheoryLemmaClause) :-
    get_clause_from_negated_ground_bindings(SatBindings, [], TheoryLemmaClause).

get_clause_from_negated_ground_bindings([], Acc, Acc).
get_clause_from_negated_ground_bindings([bind(SatVarName,Pol)|T], Acc, TheoryLemmaClause) :-
    ground(Pol),
    !,
    negate_polarity(Pol, NegPol),
    get_clause_from_negated_ground_bindings(T, [NegPol-_-SatVarName|Acc], TheoryLemmaClause).
get_clause_from_negated_ground_bindings([_|T], Acc, TheoryLemmaClause) :-
    get_clause_from_negated_ground_bindings(T, Acc, TheoryLemmaClause).

temporary_set_unsat_core_timeout :-
    preferences:temporary_set_preference(time_out, 25).
temporary_set_unsat_core_timeout :-
    preferences:reset_temporary_preference(time_out),
    fail.

% In the conflict graph each node corresponds to a variable assignment.
% The predecessors of node x are the antecedent assignments A^w(x) corresponding
% to the unit clause w (in its original form) that led to the propagation of x (logical implications).
% The directed edges from the nodes in A^w(x) to node x are all labeled with w.
% Nodes that have no predecessor correspond to decision/branching assignments
% (cf. GRASP: A Search Algorithm for Propositional Satisfiability, by Marques-Silva and Sakallah).
% As an improvement, we apply first UIP clause learning (cf. Efficient Conflict Driven Learning in a Boolean Satisfiability Solver, Zhang et al.).
:- dynamic node/3, edge/3.
:- volatile node/3, edge/3.

%% get_backjump_clause_conflict_graph_analysis(+SolverEnv, +StateMutable, +ConflictClause, -BjClause, -BjLevel).
% Compute the conflict graph for a conflicting unit-propagation of a variable
% and extract a backjump clause.
% Note that the variable references in the returned clause are not yet set (cf. get_clauses_to_learn/3).
get_backjump_clause_conflict_graph_analysis(SolverEnv, StateMutable, ConflictClause, BjClause, BjLevel) :-
    get_last_decision_level(StateMutable, _), % fail if no decision has been made at all because computing a backjump clause is redundant
    (assert_conflict_graph(SolverEnv, StateMutable, ConflictClause, FirstUIP, FirstUIPLevel); trace, nl, assert_conflict_graph(SolverEnv, StateMutable, ConflictClause, FirstUIP, FirstUIPLevel)),
    get_backjump_clause_from_asserted_conflict_graph(FirstUIP, FirstUIPLevel, BjClause),
    get_full_state(StateMutable, sat_stack, SatStack),
    (get_backjump_level_from_learned_clause(SatStack, BjClause, BjLevel); trace, get_backjump_level_from_learned_clause(SatStack, BjClause, BjLevel)),
    retractall(node(_, _, _)),
    retractall(edge(_, _, _)), !.

% Given a backjump clause, the backjump level is the second highest decision level of
% a literal in the learned clause. If the learned clause is unary, the backjump level is 0.
get_backjump_level_from_learned_clause(_, [_], BjLevel) :-
    !,
    BjLevel = 0,
    asserta_once(is_restart).
get_backjump_level_from_learned_clause(SatStack, BjClause, SecondHighest) :-
    get_two_highest_decision_levels_for_literals_in_clause(SatStack, BjClause, 0, -1, Highest, TSecondHighest),
    format("~nHighest ~w, Second ~w~n", [Highest, TSecondHighest]),
    (   TSecondHighest == -1
    ->  % it is ensured that clause is not unary: all levels are the same
        SecondHighest = Highest
    ;   SecondHighest = TSecondHighest
    ).

% Fail if no decision has been made so far, i.e., the decision level is set to -1.
get_last_decision_level(StateMutable, Level) :-
    get_full_state(StateMutable, decision_level, TLevel),
    TLevel \== -1,
    !,
    Level = TLevel.

% Given a backjump clause, the backjump level is the second highest decision level of a literal in
% the clause.
get_two_highest_decision_levels_for_literals_in_clause(_, [], HAcc, SAcc, HAcc, SAcc).
get_two_highest_decision_levels_for_literals_in_clause(SatStack, [_-_-Name|T], HAcc, SAcc, Highest, Second) :-
    avl_member(Name, SatStack, (_,Level,_,_)),
    update_two_highest_levels_accs(Level, HAcc, SAcc, NewHAcc, NewSAcc),
    !,
    get_two_highest_decision_levels_for_literals_in_clause(SatStack, T, NewHAcc, NewSAcc, Highest, Second).

update_two_highest_levels_accs(Level, HAcc, _, NewHAcc, NewSAcc) :-
    Level > HAcc,
    !,
    NewHAcc = Level,
    NewSAcc = HAcc.
update_two_highest_levels_accs(Level, HAcc, SAcc, NewHAcc, NewSAcc) :-
    Level == HAcc,
    !,
    NewHAcc = HAcc,
    NewSAcc = SAcc.
update_two_highest_levels_accs(Level, HAcc, SAcc, NewHAcc, NewSAcc) :-
    Level > SAcc,
    !,
    NewHAcc = HAcc,
    NewSAcc = Level.
update_two_highest_levels_accs(_, HAcc, SAcc, NewHAcc, NewSAcc) :-
    NewHAcc = HAcc,
    NewSAcc = SAcc.

%% assert_conflict_graph(+SolverEnv, +StateMutable, +ToExpand, -FirstUIP, -FirstUIPLevel).
% Compute the conflict graph for a conflict clause.
% A directed graph is represented using the dynamic facts node/3 and edge/2.
assert_conflict_graph(SolverEnv, StateMutable, ToExpand, FirstUIP, FirstUIPLevel) :-
    retractall(node(_, _, _)),
    retractall(edge(_, _, _)),
    assert_conflict_graph(SolverEnv, StateMutable, 0, ToExpand, FirstUIP, FirstUIPLevel).

%% assert_conflict_graph(+SolverEnv, +StateMutable, +CurLevel, +ToExpand, -FirstUIP, -FirstUIPLevel).
% FirstUIP corresponds to the UIP which is closest to the conflict clause in the conflict graph.
assert_conflict_graph(_, _, CurLevel, [], _, FirstUIPLevel) :-
    % last level had only branching decisions which have been discarded for expansion in update_predecessor_to_expand_acc/3
    !,
    FirstUIPLevel is CurLevel - 1.
assert_conflict_graph(SolverEnv, StateMutable, CurLevel, ToExpand, FirstUIP, FirstUIPLevel) :-
    map_backward_expand_implications_uip(SolverEnv, StateMutable, CurLevel, ToExpand, PreToExpand, FirstUIP),
    (   compound(FirstUIP)
    ->  FirstUIPLevel = CurLevel
    ;   NextLevel is CurLevel + 1,
        assert_conflict_graph(SolverEnv, StateMutable, NextLevel, PreToExpand, FirstUIP, FirstUIPLevel)
    ).

%% get_from_sat_stack_or_explain_theory_propagation(+SolverEnv, +StateMutable, +SatVarName, +PropType, +Level, +Pol, -OriginalClause).
get_from_sat_stack_or_explain_theory_propagation(SolverEnv, StateMutable, SatVarName, PropType, Level, Pol, OriginalClause) :-
    get_full_state(StateMutable, sat_stack, SatStack),
    avl_member(SatVarName, SatStack, (PropType,Level,Pol,TOriginalClause)),
    (   (PropType == theory, var(TOriginalClause))
    ->  explain_theory_propagation(SolverEnv, SatStack, SatVarName, Level, Pol, OriginalClause),
        avl_change(SatVarName, SatStack, _, NStack, (PropType,Level,Pol,OriginalClause)),
        update_full_state(StateMutable, sat_stack, NStack)
    ;   OriginalClause = TOriginalClause
    ).

%% map_backward_expand_implications_uip(+SolverEnv, +StateMutable, +CurLevel, +ToExpand, -PreToExpand, -FirstUIP).
% A unique implication point (FirstUIP) is any node in the implication graph
% other than the conflict that is on all paths from the current decision/branching literal.
map_backward_expand_implications_uip(SolverEnv, StateMutable, CurLevel, [_-_-Name], ToExpand, FirstUIP) :-
    (get_from_sat_stack_or_explain_theory_propagation(SolverEnv, StateMutable, Name, PropType, Level, Pol, OriginalClause); trace, nl, get_from_sat_stack_or_explain_theory_propagation(SolverEnv, StateMutable, Name, PropType, Level, Pol, OriginalClause)),
    !,
    % terminate; first UIP clause learning (cf. Efficient Conflict Driven Learning in a Boolean Satisfiability Solver, Zhang et al.)
    FirstUIP = (Name,PropType,Level,Pol,OriginalClause),
    % expand first UIP but discard further expansion using PreToExpand
    backward_expand_implications_for_propagation(StateMutable, CurLevel, FirstUIP, [], _PreToExpand),
    ToExpand = [].
map_backward_expand_implications_uip(SolverEnv, StateMutable, CurLevel, ToExpand, PreToExpand, _) :-
    map_backward_expand_implications(SolverEnv, StateMutable, CurLevel, ToExpand, [], PreToExpand).

map_backward_expand_implications(_, _, _, [], Acc, Acc).
map_backward_expand_implications(SolverEnv, StateMutable, CurLevel, [_-_-Name|T], Acc, PreToExpand) :-
    (get_from_sat_stack_or_explain_theory_propagation(SolverEnv, StateMutable, Name, PropType, Level, Pol, OriginalClause); trace, nl, get_from_sat_stack_or_explain_theory_propagation(SolverEnv, StateMutable, Name, PropType, Level, Pol, OriginalClause)),
    Propagation = (Name,PropType,Level,Pol,OriginalClause),
    (   PropType == unit
    ->  backward_expand_implications_for_propagation(StateMutable, CurLevel, Propagation, Acc, NewAcc)
    ;   % the antecedent assignment of an electively assigned variable (decision/branch) is empty
        assert_or_update_node(CurLevel, Name, Pol),
        NewAcc = Acc
    ),
    map_backward_expand_implications(SolverEnv, StateMutable, CurLevel, T, NewAcc, PreToExpand).

backward_expand_implications_for_propagation(StateMutable, CurLevel, Propagation, Acc, PreToExpand) :-
    Propagation = (Name,_,_,Pol,OriginalClause),
    assert_or_update_node(CurLevel, Name, Pol),
    NextLevel is CurLevel + 1,
    assert_or_update_neg_nodes(StateMutable, (Name,Pol), CurLevel, NextLevel, OriginalClause, OriginalClause, Acc, PreToExpand).

assert_or_update_neg_nodes(_, _, _, _, [], _, Acc, Acc).
assert_or_update_neg_nodes(StateMutable, (ImplName,ImplPol), CurLevel, NextLevel, [Pol-_-Name|T], OriginalClauseRef, Acc, PreToExpand) :-
    Name \== ImplName,
    !,
    negate_polarity(Pol, NegPol),
    assert_or_update_node(NextLevel, Name, NegPol),
    asserta(edge(node(Name,NextLevel,NegPol),OriginalClauseRef,node(ImplName,CurLevel,ImplPol))),
    update_predecessor_to_expand_acc(StateMutable, NegPol-NegPol-Name, Acc, NewAcc),
    assert_or_update_neg_nodes(StateMutable, (ImplName,ImplPol), CurLevel, NextLevel, T, OriginalClauseRef, NewAcc, PreToExpand).
assert_or_update_neg_nodes(StateMutable, (ImplName,ImplPol), CurLevel, NextLevel, [_|T], OriginalClauseRef, Acc, PreToExpand) :-
    assert_or_update_neg_nodes(StateMutable, (ImplName,ImplPol), CurLevel, NextLevel, T, OriginalClauseRef, Acc, PreToExpand).

assert_or_update_node(CurLevel, Name, Pol) :-
    (   retract(node(Name,_,Pol)),
        !
    ;   true
    ),
    asserta(node(Name,CurLevel,Pol)).

negate_polarity(pred_true, pred_false).
negate_polarity(pred_false, pred_true).

update_predecessor_to_expand_acc(StateMutable, NegPol-Var-Name, Acc, NewAcc) :-
    get_full_state(StateMutable, sat_stack, SatStack),
    (   memberchk(NegPol-Var-Name, Acc)
    ;   avl_member(Name, SatStack, (PropType,_,_,_)),
        % nodes that have no predecessor correspond to decision/branching assignments
        PropType == branch
    ),
    !,
    NewAcc = Acc.
update_predecessor_to_expand_acc(_, NegPol-Var-Name, Acc, [NegPol-Var-Name|Acc]).

%% get_backjump_clause_from_asserted_conflict_graph(+FirstUIP, +UIPLevel, -BjClause).
% Extract a backjump clause from an asserted conflict graph.
% If the first UIP has been derived by unit-propagation, the BjClause contains the negated
% literal and all negated branching literals on the same or a lower level in the backward conflict graph (shortest conflict clause).
% Otherwise, the conflict clause consists of all sources of the conflict graph.
% Every cut that separates sources from the sink defines a valid conflict clause.
% Sources in the conflict graph correspond to decision/branching assignments.
get_backjump_clause_from_asserted_conflict_graph(FirstUIP, UIPLevel, BjClause) :-
    FirstUIP = (UIPName,PropType,_,UIPPol,_),
    PropType == unit,
    !,
    get_sources_from_asserted_conflict_graph_lower_or_eq_level(UIPLevel, Sources),
    negate_nodes_to_clause(Sources, NegSourcesClause),
    negate_polarity(UIPPol, NegUIPPol),
    BjClause = [NegUIPPol-_-UIPName|NegSourcesClause].
get_backjump_clause_from_asserted_conflict_graph(_, _, BjClause) :-
    get_sources_from_asserted_conflict_graph(Sources),
    negate_nodes_to_clause(Sources, BjClause).

%% negate_nodes_to_clause(+Nodes, -BjClause).
% Negate literals and create a clause without original variable references.
negate_nodes_to_clause([], []).
negate_nodes_to_clause([node(Name,_,Pol)|T], [NegPol-_-Name|Rest]) :-
    negate_polarity(Pol, NegPol),
    negate_nodes_to_clause(T, Rest).

get_sources_from_asserted_conflict_graph(Sources) :-
    findall(Source,
        (node(Name,Level,Pol),
         Source = node(Name,Level,Pol),
         \+ edge(_,_,Source)
        ), Sources).

get_sources_from_asserted_conflict_graph_lower_or_eq_level(UIPLevel, Sources) :-
    findall(Source,
        (node(Name,Level,Pol),
         Level =< UIPLevel,
         Source = node(Name,Level,Pol),
         \+ edge(_,_,Source)
        ), Sources).

asserta_once(Assertion) :-
    call(Assertion),
    !.
asserta_once(Assertion) :-
    asserta(Assertion).

b_pred_or_bool_true(X) :- X == truth; X == boolean_true; X == value(pred_true).
b_pred_or_bool_false(X) :- X == falsity; X == boolean_false; X == value(pred_false).

sat_conj_negate_to_clause(Pred, Clause) :-
    conjunction_to_list(Pred, List),
    maplist(sat_equality_negate_to_literal, List, Clause).

sat_equality_negate_to_literal(b(equal(ID,Pol),pred,_), NegPol-_-Name) :-
    ID = b(identifier(Name),_,_),
    Pol = b(Value,boolean,_),
    (   b_pred_or_bool_true(Value)
    ->  NegPol = pred_false
    ;   NegPol = pred_true
    ).

%%%%% Solver State
create_solver_state(StateMutable) :-
    empty_heap(EmptyHeap),
    empty_avl(EmptyAvl),
    State = [sat_stack:EmptyAvl,cur_level:1,':'(decision_level,-1),current_stack_size:0,heuristic_heap:EmptyHeap,heuristic_avl:EmptyAvl],
    create_mutable(solver_state(State), StateMutable).

%% get_full_state(+StateMutable, +Key, -Datum).
get_full_state(StateMutable, Key, Datum) :-
    get_mutable(solver_state(SolverState), StateMutable),
    member(Key:Datum, SolverState).

%% add_state_instance(+StateMutable, +Key, +Datum).
% Add an instance to iterable data.
add_state_instance(StateMutable, Key, Datum) :-
    get_mutable(solver_state(SolverState), StateMutable),
    select(Key:Data, SolverState, Rest),
    (   Key == heuristic_heap
    ->  Datum = (SatVarName,Score),
        Score1 is Score * -1, % its a min heap
        add_to_heap(Data, Score1, SatVarName, NewData)
    ;   Key == heuristic_avl
    ->  Datum = (SatVarName,Score),
        Score1 is Score * -1, % its a min heap
        avl_store(SatVarName, Data, Score1, NewData)
    ;   Key == sat_stack,
        Datum = (SatVarName,A,B,C,D),
        avl_store(SatVarName, Data, (A,B,C,D), NewData)
    ), !,
    update_mutable(solver_state([Key:NewData|Rest]), StateMutable).

%% delete_state_instance(+StateMutable, +Key, ?Datum).
delete_state_instance(StateMutable, Key, Datum) :-
    get_mutable(solver_state(SolverState), StateMutable),
    select(Key:Data, SolverState, Rest),
    (   Key == heuristic_heap
    ->  Datum = (_,Score),
        (   delete_from_heap(Data, Score, _, NewData),
            !
        ;   NewData = Data)
    ;   Key == heuristic_avl
    ->  Datum = (SatVarName,Score),
        (   avl_delete(SatVarName, Data, Score, NewData),
            !
        ;   NewData = Data
        )
    ;   Key == sat_stack,
        Datum = (SatVarName,A,B,C,D),
        (   avl_delete(SatVarName, Data, (A,B,C,D), NewData),
            !
        ;   NewData = Data
        )
    ), !,
    update_mutable(solver_state([Key:NewData|Rest]), StateMutable).

%% update_full_state(+StateMutable, +Key, +Datum).
update_full_state(StateMutable, Key, Datum) :-
    get_mutable(solver_state(SolverState), StateMutable),
    select(Key:_, SolverState, Rest), !,
    update_mutable(solver_state([Key:Datum|Rest]), StateMutable).

init_dynamic_solver_state :-
    clean_dynamic_solver_state,
    asserta(amount_clause_reductions(0)),
    asserta(amount_learned_clauses(0)),
    asserta(conflict_stats(total(0), since_restart(0), since_reduction(0))),
    asserta(recent_lbds_dl(Diff1-Diff1, 0)),
    asserta(recent_stack_sizes_dl(Diff2-Diff2, 0)),
    asserta(total_lbd_sum(0)),
    init_luby_sequence.

clean_dynamic_solver_state :-
    retractall(is_restart),
    retractall(bt_from_smt_solution),
    retractall(recent_lbds_dl(_, _)),
    retractall(recent_stack_sizes_dl(_, _)),
    retractall(saved_phase(_, _)),
    retractall(amount_clause_reductions(_)),
    retractall(amount_learned_clauses(_)),
    retractall(luby_counter(_)),
    retractall(exclude_sat_solution_clause(_)),
    retractall(total_lbd_sum(_)),
    retractall(conflict_stats(_, _, _)),
    retractall(backjump_clause(_, _, _)),
    retractall(idl_unsat_core(_)),
    retractall(backjump_level(_)).

remove_idl_unsat_core :-
    retractall(idl_unsat_core(_)).

store_idl_unsat_core(Core) :-
    retractall(idl_unsat_core(_)),
    asserta(idl_unsat_core(Core)).

announce_bt_from_smt_solution :-
    asserta(bt_from_smt_solution).
%%%%%

%% b_to_cnf(+In, +State, -Out).
% Accepts predicates with conjunction and disjunction.
b_to_cnf(In, State, Out) :-
    get_texpr_expr(In, Expr),
    b_to_cnf_aux(Expr, State, Out),
    !.
b_to_cnf(In, _, _) :-
    translate_bexpression(In, PPIn),
    !,
    add_error_fail(b_to_cnf, 'CNF conversion failed', PPIn).

b_to_cnf_aux(identifier(Name), State, [[pred_true-Res-Name]]) :-
    member(bind(Name,Res), State).
b_to_cnf_aux(X, _, [[pred_true-pred_true-nonvar]]) :-
    b_pred_or_bool_true(X).
b_to_cnf_aux(X, _, [[pred_false-pred_false-nonvar]]) :-
    b_pred_or_bool_false(X).

b_to_cnf_aux(equal(A,B), State, CNF) :-
    ground(B),
    get_texpr_expr(B, Expr),
    b_pred_or_bool_true(Expr),
    b_to_cnf(A, State, CNF).
b_to_cnf_aux(equal(A,B), State, CNF) :-
    ground(B),
    get_texpr_expr(B, Expr),
    b_pred_or_bool_false(Expr),
    create_texpr(value(pred_true), boolean, [], True),
    create_texpr(equal(A,True), pred, [], EqualTrue),
    b_to_cnf_aux(negation(EqualTrue), State, CNF).

b_to_cnf_aux(negation(A), State, [[Res]]) :-
    b_to_cnf(A, State, [[Pol-Var-Name]]),
    (   Pol == pred_true
    ->  Res = pred_false-Var-Name
    ;   Pol == pred_false
    ->  Res = pred_true-Var-Name
    ).

b_to_cnf_aux(disjunct(D1,D2), State, Res) :-
    get_texpr_expr(D1, conjunct(A,B)),
    !,
    create_texpr(disjunct(A,D2), pred, [], DJ1),
    create_texpr(disjunct(B,D2), pred, [], DJ2),
    b_to_cnf_aux(conjunct(DJ1,DJ2), State, Res).
b_to_cnf_aux(disjunct(D1,D2), State, Res) :-
    get_texpr_expr(D2, conjunct(A,B)),
    !,
    create_texpr(disjunct(D1,A), pred, [], DJ1),
    create_texpr(disjunct(D1,B), pred, [], DJ2),
    b_to_cnf_aux(conjunct(DJ1,DJ2), State, Res).

b_to_cnf_aux(disjunct(D1,D2),State,[Res]) :-
    b_to_cnf(D1, State, ResD1),
    b_to_cnf(D2, State, ResD2),
    append(ResD1, ResD2, TempRes),
    append(TempRes, Res).

b_to_cnf_aux(conjunct(A,B), State, Res) :-
    b_to_cnf(A, State, RA),
    b_to_cnf(B, State, RB),
    append(RA, RB, Res).

:- dynamic luby_counter/1.
:- volatile luby_counter/1.

% Used for random restarts.
% First restart is after 32nd conflict, second after additional 32 conflicts,
% third after additional 64 conflicts etc.
init_luby_sequence :-
    retractall(luby_counter(_)),
    % unit length of 32
    % 1,1,2,1,1,2,4,1,1,2,1,1,2,4,8,..
    assertz(luby_counter(1)),
    assertz(luby_counter(1)),
    assertz(luby_counter(2)),
    assertz(luby_counter(1)),
    assertz(luby_counter(1)),
    assertz(luby_counter(2)),
    assertz(luby_counter(4)),
    assertz(luby_counter(1)),
    assertz(luby_counter(1)),
    assertz(luby_counter(2)),
    assertz(luby_counter(1)),
    assertz(luby_counter(1)),
    assertz(luby_counter(2)),
    assertz(luby_counter(4)),
    assertz(luby_counter(8)),
    assertz(luby_counter(1)),
    assertz(luby_counter(1)),
    assertz(luby_counter(2)),
    assertz(luby_counter(1)),
    assertz(luby_counter(1)),
    assertz(luby_counter(2)),
    assertz(luby_counter(4)),
    assertz(luby_counter(8)),
    assertz(luby_counter(16)),
    assertz(luby_counter(1)),
    assertz(luby_counter(1)),
    assertz(luby_counter(2)),
    assertz(luby_counter(1)),
    assertz(luby_counter(1)),
    assertz(luby_counter(2)),
    assertz(luby_counter(4)),
    assertz(luby_counter(8)),
    assertz(luby_counter(16)),
    assertz(luby_counter(32)),
    assertz(luby_counter(1)),
    assertz(luby_counter(1)),
    assertz(luby_counter(2)),
    assertz(luby_counter(1)),
    assertz(luby_counter(1)),
    assertz(luby_counter(2)),
    assertz(luby_counter(4)),
    assertz(luby_counter(8)),
    assertz(luby_counter(16)),
    assertz(luby_counter(32)),
    assertz(luby_counter(64)),
    assertz(luby_counter(1)),
    assertz(luby_counter(1)),
    assertz(luby_counter(2)),
    assertz(luby_counter(1)),
    assertz(luby_counter(1)),
    assertz(luby_counter(2)),
    assertz(luby_counter(4)),
    assertz(luby_counter(8)),
    assertz(luby_counter(16)),
    assertz(luby_counter(32)),
    assertz(luby_counter(64)),
    assertz(luby_counter(128)).

/*
%% static_sat_symmetry_breaking(+Clauses, +SolverEnv, +CurLevelMut, -NewSolverEnv).
% Static SAT symmetry breaking after unit propagation of problem_setup/3 but only once, e.g., not after a restart or bj to level 0
% Can only be used for pure SAT solving.
static_sat_symmetry_breaking(_, SolverEnv, _, NewSolverEnv) :-
    static_sat_symmetry_breaking(false),
    !,
    NewSolverEnv = SolverEnv.
static_sat_symmetry_breaking(Clauses, SolverEnv, CurLevelMut, NewSolverEnv) :-
    \+ symmetry_breaking_pred(_, _),
    !,
    get_symmetry_breaking_predicates(Clauses, SBPsConjNoRef, NewSatVars),
    asserta(symmetry_breaking_pred(SBPsConjNoRef,NewSatVars)),
    static_sat_symmetry_breaking(Clauses, SolverEnv, CurLevelMut, NewSolverEnv).
static_sat_symmetry_breaking(_, SolverEnv, CurLevelMut, NewSolverEnv) :-
    get_asserted_symmetry_breaking_pred(SolverEnv, SBPsConj, NewSolverEnv),
    problem_setup(NewSolverEnv, CurLevelMut, SBPsConj).

get_asserted_symmetry_breaking_pred(SolverEnv, SBPsConj, NewSolverEnv) :-
    symmetry_breaking_pred(SBPsConjNoRef, NewSatVars),
    add_symmetry_breaking_eqs_to_solver_env(NewSatVars, SolverEnv, NewSolverEnv),
    get_from_solver_env(NewSolverEnv, pl_vars, PlVars),
    set_variable_references_in_clauses(PlVars, SBPsConjNoRef, SBPsConj),
    findall(Var, (member(Var, NewSatVars), bt_asserta(heuristic_tuple(Var,1))), _).

%% add_symmetry_breaking_eqs_to_solver_env(+NewSatVars, +SolverEnv, -NewSolverEnv).
% Symmetry breaking might introduce new SAT variables for equalities inbetween cycles of permutations.
% We update the solver environment and create the corresponding SMT formulas for equalities here.
% This is necessary, e.g., if a SAT variable corresponding to an equality introduced
% from symmetry breaking is involved in a conflict.
add_symmetry_breaking_eqs_to_solver_env(NewSatVars, SolverEnv, NewSolverEnv) :-
    SolverEnv = [sat_vars:SatVars,sat_varnames:SatVarNames,sat_bindings:SatBindings,t2b_env:Env,pl_vars:PrologSatVarTriple],
    add_symmetry_breaking_eqs_to_solver_env(NewSatVars, SatVars, SatVarNames, SatBindings, Env, PrologSatVarTriple, NSatVars, NSatVarNames, NSatBindings, NEnv, NPrologSatVarTriple),
    NewSolverEnv = [sat_vars:NSatVars,sat_varnames:NSatVarNames,sat_bindings:NSatBindings,t2b_env:NEnv,pl_vars:NPrologSatVarTriple].

add_symmetry_breaking_eqs_to_solver_env([], SatVars, SatVarNames, SatBindings, Env, PrologSatVarTriple, SatVars, SatVarNames, SatBindings, Env, PrologSatVarTriple).
add_symmetry_breaking_eqs_to_solver_env([NewSatVarName-EqClauses|T], SatVars, SatVarNames, SatBindings, Env, PrologSatVarTriple, NSatVars, NSatVarNames, NSatBindings, NEnv, NPrologSatVarTriple) :-
    maplist(clause_to_smt_disjunction(SatVars), EqClauses, EqSmtClauses),
    conjunct_predicates(EqSmtClauses, EqSmt1),      % all true
    maplist(negate_smt_formula, EqSmtClauses, NegEqSmtClauses),
    conjunct_predicates(NegEqSmtClauses, EqSmt2),   % or all false
    EqSmt = b(disjunct(EqSmt1,EqSmt2),pred,[]),
    SatVars1 = [b(identifier(NewSatVarName),boolean,[smt_formula(EqSmt),eq_symmetry_break])|SatVars],
    SatVarNames1 = [NewSatVarName|SatVarNames],
    SatBindings1 = [bind(NewSatVarName,Var)|SatBindings],
    PrologSatVarTriple1 = [Var-NewSatVarName-EqSmt|PrologSatVarTriple],
    Env1 = [(EqSmt,b(identifier(NewSatVarName),boolean,[]))|Env],
    add_symmetry_breaking_eqs_to_solver_env(T, SatVars1, SatVarNames1, SatBindings1, Env1, PrologSatVarTriple1, NSatVars, NSatVarNames, NSatBindings, NEnv, NPrologSatVarTriple).

%clause_to_smt_disjunction(_, [], b(truth,pred,[])).
clause_to_smt_disjunction(SatVars, [Pol-_-SatVarName|T], SmtDisj) :-
    get_smt_formula_for_sat_pol(SatVars, SatVarName, Pol, SmtFormula),
    clause_to_smt_disjunction(SatVars, T, SmtFormula, SmtDisj).

clause_to_smt_disjunction(_, [], Acc, Acc).
clause_to_smt_disjunction(SatVars, [Pol-_-SatVarName|T], Acc, SmtDisj) :-
    get_smt_formula_for_sat_pol(SatVars, SatVarName, Pol, SmtFormula),
    NAcc = b(disjunct(SmtFormula,Acc),pred,[]),
    clause_to_smt_disjunction(SatVars, T, NAcc, SmtDisj).
*/
