:- module(smt_symmetry_breaking, [init_graph/0,
                                  get_symmetry_breaking_predicates/2,
                                  add_symmetry_breaking_predicates/2]).

:- use_module(library(codesio), [write_to_codes/2]).
:- use_module(library(samsort), [samsort/3]).
:- use_module(library(terms), [term_hash/2]).
:- use_module(library(lists), [select/3, maplist/3]).
:- use_module(extension('bliss/bliss_interface')).
:- use_module(probsrc('dpllt_solver/symmetry_check/unique_quantified_identifiers')).
:- use_module(probsrc(error_manager), [add_error_and_fail/2, add_error_and_fail/3]).
:- use_module(probsrc(bsyntaxtree), [remove_all_infos/2,
                                     remove_all_infos_and_ground/2,
                                     find_typed_identifier_uses/3]).

% Foundation: "SyMT: Finding Symmetries in SMT formulas" by Areces et al.
% Improvement: "Advances in Symmetry Breaking for SAT Modulo Theories" by Dinliwal et al. % TO DO

:- dynamic next_color/1, color/2, seen_pred/2, seen_upred/2, seen_expr/2, seen_uexpr/2, node_id_to_ast/2, ast_to_node_id/2.
:- volatile next_color/1, color/2, seen_pred/2, seen_upred/2, seen_expr/2, seen_uexpr/2, node_id_to_ast/2, ast_to_node_id/2.

init_graph :-
    bliss_interface:init_bliss_interface,
    bliss_interface:init_directed_graph,
    retractall(node_id_to_ast(_,_)),
    retractall(ast_to_node_id(_,_)),
    retractall(seen_pred(_,_)),
    retractall(seen_upred(_,_)),
    retractall(seen_expr(_,_)),
    retractall(seen_uexpr(_,_)),
    retractall(color(_,_)),
    asserta(color(arg, 0)), % all argument nodes have the same unique color
    retractall(next_color(_)),
    asserta(next_color(1)).

get_texpr_expr_functor_and_type(b(Expr,Type,_), Expr, Type, Functor) :-
    functor(Expr, Functor, _).

% TO DO: variable ordering using information from SAT formula

%% get_symmetry_breaking_predicates(+SmtFormula, -SBPs).
% Assuming that the AST has been normalized by B AST cleanup and
% quantified identifers have been renamed to unique names (see module unique_quantified_identifiers).
get_symmetry_breaking_predicates(SmtFormula, SBPs) :-
    get_symmetry_breaking_predicates(SmtFormula, SBPs, _).

%% get_symmetry_breaking_predicates(+SmtFormula, +SBPs, -Generators).
get_symmetry_breaking_predicates(SmtFormula, SBPs, Generators) :-
    find_typed_identifier_uses(SmtFormula, [], UsedIds),
    remove_all_infos_and_ground(SmtFormula, CSmtFormula),
    order_variables(UsedIds, _TypeOrdering, VariableOrdering),
    (   VariableOrdering == []
    ->  SBPs = b(truth,pred,[])
    ;   init_graph,
        build_colored_graph(CSmtFormula),
        bliss_interface:find_automorphisms(Generators),
        nl, write('Generators: '), write(Generators),nl,
        get_sbps_from_generators(Generators, VariableOrdering, TSBPs),
        !,
        SBPs = TSBPs
    ).

%% add_symmetry_breaking_predicates(+SmtFormula, -NSmtFormula).
% Adds symmetry breaking predicates to quantifiers as well.
add_symmetry_breaking_predicates(SmtFormula, NSmtFormula) :-
    set_unique_quantified_identifiers(SmtFormula, UniqueQIdsFormula),
    get_symmetry_breaking_predicates(UniqueQIdsFormula, SBPs, Generators),
    add_sbps_to_quantifiers(UniqueQIdsFormula, [], Generators, TSmtFormula),
    NSmtFormula = b(conjunct(SBPs, TSmtFormula),pred,[]).

%% add_sbps_to_quantifiers(+Ast, +ScopeIds, +Generators, -SmtFormula).
add_sbps_to_quantifiers(b(Expr,Type,Info), ScopeIds, Generators, NSmtFormula) :-
    !,
    add_sbps_to_quantifiers_expr(Expr, ScopeIds, Generators, NExpr),
    NSmtFormula = b(NExpr,Type,Info).
add_sbps_to_quantifiers(Formula, _, _, Formula).

add_sbps_to_quantifiers_expr(forall(Ids,Lhs,Rhs), ScopeIds, Generators, NExpr) :-
    !,
    append(ScopeIds, Ids, NScopeIds),
    order_variables(NScopeIds, _TypeOrdering, VariableOrdering),
    get_sbps_from_generators(Generators, VariableOrdering, SBPs),
    % empty scope for nested quantifiers since we break symmetry here
    add_sbps_to_quantifiers(Rhs, [], Generators, NRhs),
    (   SBPs = b(truth,pred,_)
    ->  NExpr = forall(Ids,Lhs,NRhs)
    ;   NExpr = forall(Ids,Lhs,b(conjunct(SBPs,NRhs),pred,[]))
    ).
add_sbps_to_quantifiers_expr(exists(Ids,Body), ScopeIds, Generators, NExpr) :-
    !,
    append(ScopeIds, Ids, NScopeIds),
    order_variables(NScopeIds, _TypeOrdering, VariableOrdering),
    get_sbps_from_generators(Generators, VariableOrdering, SBPs),
    add_sbps_to_quantifiers(Body, [], Generators, NBody),
    (   SBPs = b(truth,pred,_)
    ->  NExpr = exists(Ids,NBody)
    ;   NExpr = exists(Ids,b(conjunct(SBPs,NBody),pred,[]))
    ).
add_sbps_to_quantifiers_expr(Expr, _, _, NExpr) :-
    (   is_interpreted_symbol(Expr)
    ;   is_uninterpreted_symbol(Expr)
    ),
    !,
    NExpr = Expr.
add_sbps_to_quantifiers_expr(Expr, Scopes, Generators, NExpr) :-
    functor(Expr, Functor, 2),
    !,
    arg(1, Expr, Arg1),
    arg(2, Expr, Arg2),
    add_sbps_to_quantifiers(Arg1, Scopes, Generators, NArg1),
    add_sbps_to_quantifiers(Arg2, Scopes, Generators, NArg2),
    functor(NExpr, Functor, 2),
    arg(1, NExpr, NArg1),
    arg(2, NExpr, NArg2).
add_sbps_to_quantifiers_expr(Expr, Scopes, Generators, NExpr) :-
    functor(Expr, Functor, 1),
    !,
    arg(1, Expr, Arg),
    add_sbps_to_quantifiers(Arg, Scopes, Generators, NArg),
    functor(NExpr, Functor, 1),
    arg(1, NExpr, NArg).
add_sbps_to_quantifiers_expr(Expr, _, _, Expr).

%% get_sbps_from_generators(+Generators, +VariableOrdering, -SBPs).
get_sbps_from_generators(Generators, VariableOrdering, SBPs) :-
    add_sbps_from_generators(Generators, VariableOrdering, b(truth,pred,[]), SBPs).

%% add_sbps_from_generators(+Generators, +VariableOrdering, +Acc, -SBPs).
add_sbps_from_generators([], _, Acc, Acc).
add_sbps_from_generators([Generator|T], VariableOrdering, Acc, SBPs) :-
    add_sbps_from_generator(VariableOrdering, Generator, Acc , NAcc),
    add_sbps_from_generators(T, VariableOrdering, NAcc, SBPs).

%% add_sbps_from_generator(+Generators, +VariableOrdering, +Acc, -SBPs).
add_sbps_from_generator([], _, Acc , Acc).
add_sbps_from_generator([FirstVarAst|T], Generator, Acc , NAcc) :-
    remove_all_infos(FirstVarAst, CFirstVarAst),
    ast_to_node_id(CFirstVarAst, FirstVarId),
    image_of_generator(Generator, FirstVarId, ImageId),
    node_id_to_ast(ImageId, ImageAst),
    (   ImageAst = CFirstVarAst
    ->  % tautology
        add_sbps_from_generator(T, Generator, Acc, NAcc)
    ;   Eq = b(equal(FirstVarAst,ImageAst),pred,[]),
        less_eq_for_type(FirstVarAst, ImageAst, LEq),
        extend_conj_acc(Acc, LEq, NAcc1),
        add_sbps_from_generator_eq_acc(T, Generator, Eq, NAcc1, NAcc)
    ).

%% image_of_generator(+Generator, +Id, -ImageId).
image_of_generator([], Id, Id). % identity
image_of_generator([Cycle|_], Id, ImageId) :-
    image_of_cycle(false, Cycle, Id, TImageId),
    !,
    ImageId = TImageId.
image_of_generator([_|T], Id, ImageId) :-
    image_of_generator(T, Id, ImageId).

image_of_cycle(MapLastToFirst, [FirstCycleId|T], Id, ImageId) :-
    image_of_cycle(MapLastToFirst, [FirstCycleId|T], FirstCycleId, Id, ImageId).

%% image_of_cycle(+MapLastToFirst, +FirstCycleId, +Id, -ImageId).
% Tautologies are generated in symmetry breaking if last element maps to first.
image_of_cycle(true, [Id], ImageId, Id, ImageId).
image_of_cycle(false, [Id], _, Id, Id).
image_of_cycle(_, [Id,TImageId|_], _, Id, ImageId) :-
    ImageId = TImageId.
image_of_cycle(MapLastToFirst, [_|T], FirstCycleId, Id, ImageId) :-
    image_of_cycle(MapLastToFirst, T, FirstCycleId, Id, ImageId).

%% add_sbps_from_generator(+VariableOrdering, +Generator, +EqAcc, +Acc, -NAcc).
add_sbps_from_generator_eq_acc([], _, _, Acc, Acc).
add_sbps_from_generator_eq_acc([VarAst|T], Generator, EqAcc, Acc, NAcc) :-
    remove_all_infos(VarAst, CVarAst),
    ast_to_node_id(CVarAst, VarId),
    image_of_generator(Generator, VarId, ImageId),
    node_id_to_ast(ImageId, ImageAst),
    (   ImageAst = CVarAst
    ->  % tautology
        add_sbps_from_generator_eq_acc(T, Generator, EqAcc, Acc, NAcc)
    ;   Eq = b(equal(VarAst,ImageAst),pred,[]),
        less_eq_for_type(VarAst, ImageAst, LEq),
        Impl = b(implication(EqAcc,LEq),pred,[]),
        extend_conj_acc(Acc, Impl, NAcc1),
        NEqAcc = b(conjunct(EqAcc,Eq),pred,[]),
        add_sbps_from_generator_eq_acc(T, Generator, NEqAcc, NAcc1, NAcc)
    ).

%% less_eq_for_type(+VarAst, +ImageAst, -LEq).
less_eq_for_type(VarAst, ImageAst, LEq) :-
    VarAst = b(_,Type,_),
    ImageAst = b(_,Type,_),
    LEq = b(external_pred_call('LEQ_SYM_BREAK', [VarAst,ImageAst]),pred,[]).
less_eq_for_type(VarAst, ImageAst, _) :-
    VarAst = b(_,TypeA,_),
    ImageAst = b(_,TypeB,_),
    TypeA \== TypeB,
    error_manager:add_error_and_fail(less_eq_for_type, 'Differently typed permutation mapping. Encoding for SMT symmetry breaking is defective.', [VarAst,ImageAst]).

%% get_equality_from_cycle(+Cycle, -Eq).
% Pairwise equality.
get_equality_from_cycle([A,B|T], Eq) :-
    EqAcc = b(equal(A,B),pred,[]),
    get_equality_from_cycle([B|T], EqAcc, Eq).

get_equality_from_cycle([_], EqAcc, EqAcc).
get_equality_from_cycle([A,B|T], EqAcc, Eq) :-
    NEqAcc = b(conjunct(b(equal(A,B),pred,[]),EqAcc),pred,[]),
    get_equality_from_cycle([B|T], NEqAcc, Eq).

%% extend_conj_acc(+Acc, +TConj, -Acc).
extend_conj_acc(b(truth,pred,[]), TConj, Conj) :-
    !,
    Conj = TConj.
extend_conj_acc(Acc, Conj, b(conjunct(Acc,Conj),pred,[])).

%% build_colored_graph(+SmtFormula).
build_colored_graph(SmtFormula) :-
    get_texpr_expr_functor_and_type(SmtFormula, Expr, Type, ExprFunctor),
    add_term_and_symbol_nodes_to_graph(Expr, Type, ExprFunctor, RootNodeId),
    build_colored_graph(Expr, Type, RootNodeId).

%% build_colored_graph(+Term, +TermRootNodeId).
build_colored_graph(truth, pred,  _).
build_colored_graph(falsity, pred, _).
build_colored_graph(Expr, _, RootNodeId) :-
    (   Expr = set_extension(List)
    ;   Expr = sequence_extension(List)
    ),
    !,
    build_colored_graph_from_set(List, RootNodeId).
build_colored_graph(Expr, Type, RootNodeId) :-
    (   Expr = let_predicate(Ids, EqVals, Body)
    ;   Expr = let_expression(Ids, EqVals, Body)
    ),
    !,
    % non-commutative
    zip_to_equalities_conj(Ids, EqVals, EqsConj),
    get_texpr_expr_functor_and_type(EqsConj, EqsExpr, EqsType, EqsFunctor),
    get_texpr_expr_functor_and_type(Body, BodyExpr, BodyType, BodyFunctor),
    add_term_and_symbol_nodes_to_graph(EqsExpr, pred, EqsFunctor, EqsRootNodeId),
    add_term_and_symbol_nodes_to_graph(BodyExpr, pred, BodyFunctor, BodyRootNodeId),
    add_node_to_colored_graph(arg, EqsExpr, EqsType, EqsNodeId),
    add_node_to_colored_graph(arg, BodyExpr, BodyType, BodyNodeId),
    bliss_interface:add_edge(EqsNodeId, EqsRootNodeId),
    bliss_interface:add_edge(BodyNodeId, BodyRootNodeId),
    bliss_interface:add_edge(EqsNodeId, BodyNodeId),
    bliss_interface:add_edge(RootNodeId, EqsNodeId),
    build_colored_graph(EqsExpr, EqsType, EqsRootNodeId),
    build_colored_graph(BodyExpr, Type, BodyRootNodeId).
build_colored_graph(if_then_else(Cond,If,Else), _, RootNodeId) :-
    % non-commutative
    !,
    get_texpr_expr_functor_and_type(Cond, CondExpr, CondType, CondFunctor),
    get_texpr_expr_functor_and_type(If, IfExpr, IfType, IfFunctor),
    get_texpr_expr_functor_and_type(Else, ElseExpr, ElseType, ElseFunctor),
    add_term_and_symbol_nodes_to_graph(CondExpr, CondType, CondFunctor, CondRootNodeId),
    add_term_and_symbol_nodes_to_graph(IfExpr, IfType, IfFunctor, IfRootNodeId),
    add_term_and_symbol_nodes_to_graph(ElseExpr, ElseType, ElseFunctor, ElseRootNodeId),
    add_node_to_colored_graph(arg, CondExpr, CondType, CondNodeId),
    add_node_to_colored_graph(arg, IfExpr, IfType, IfNodeId),
    add_node_to_colored_graph(arg, ElseExpr, ElseType, ElseNodeId),
    bliss_interface:add_edge(CondNodeId, CondRootNodeId),
    bliss_interface:add_edge(IfNodeId, IfRootNodeId),
    bliss_interface:add_edge(ElseNodeId, ElseRootNodeId),
    bliss_interface:add_edge(CondNodeId, IfNodeId),
    bliss_interface:add_edge(IfNodeId, ElseNodeId),
    bliss_interface:add_edge(RootNodeId, CondNodeId),
    build_colored_graph(CondExpr, CondType, CondRootNodeId),
    build_colored_graph(IfExpr, IfType, IfRootNodeId),
    build_colored_graph(ElseExpr, ElseType, ElseRootNodeId).
build_colored_graph(forall(Ids,Lhs,Rhs), pred, RootNodeId) :-
    !,
    (   Lhs = b(truth,pred,_) % typing information might have been removed
    ->  Body = Rhs
    ;   Body = b(implication(Lhs,Rhs),pred,[])
    ),
    create_nodes_for_ids(Ids),
    get_texpr_expr_functor_and_type(Body, BodyExpr, BodyType, BodyFunctor),
    add_term_and_symbol_nodes_to_graph(BodyExpr, BodyType, BodyFunctor, ArgRootNodeId),
    add_node_to_colored_graph(arg, BodyExpr, pred, ArgNodeId),
    bliss_interface:add_edge(ArgNodeId, ArgRootNodeId),
    bliss_interface:add_edge(RootNodeId, ArgNodeId),
    build_colored_graph(BodyExpr, BodyType, ArgRootNodeId).
build_colored_graph(Expr, _, RootNodeId) :-
    (   Expr = comprehension_set(Ids,Body)
    ;   Expr = exists(Ids,Body)
    ),
    !,
    create_nodes_for_ids(Ids),
    get_texpr_expr_functor_and_type(Body, BodyExpr, BodyType, BodyFunctor),
    add_term_and_symbol_nodes_to_graph(BodyExpr, BodyType, BodyFunctor, ArgRootNodeId),
    add_node_to_colored_graph(arg, BodyExpr, BodyType, ArgNodeId),
    bliss_interface:add_edge(ArgNodeId, ArgRootNodeId),
    bliss_interface:add_edge(RootNodeId, ArgNodeId),
    build_colored_graph(BodyExpr, BodyType, ArgRootNodeId).
build_colored_graph(lambda(_Ids,LExpr,Pred), _, RootNodeId) :-
    % non-commutative
    !,
    get_texpr_expr_functor_and_type(LExpr, LExprExpr, LExprType, LExprFunctor),
    get_texpr_expr_functor_and_type(Pred, PredExpr, PredType, PredFunctor),
    add_term_and_symbol_nodes_to_graph(LExpr, pred, LExprFunctor, LExprRootNodeId),
    add_term_and_symbol_nodes_to_graph(Pred, pred, PredFunctor, PredRootNodeId),
    add_node_to_colored_graph(arg, LExprExpr, LExprType, LExprNodeId),
    add_node_to_colored_graph(arg, PredExpr, PredType, PredNodeId),
    bliss_interface:add_edge(LExprNodeId, LExprRootNodeId),
    bliss_interface:add_edge(PredNodeId, PredRootNodeId),
    bliss_interface:add_edge(LExprNodeId, PredNodeId),
    bliss_interface:add_edge(RootNodeId, LExprNodeId),
    build_colored_graph(LExprExpr, LExprType, LExprRootNodeId),
    build_colored_graph(PredExpr, PredType, PredRootNodeId).
build_colored_graph(Binary, _Type, RootNodeId) :-
    functor(Binary, Functor, Arity),
    Arity == 2,
    \+ is_binary_interpreted_symbol(Binary),
    !,
    arg(1, Binary, Arg1),
    arg(2, Binary, Arg2),
    get_texpr_expr_functor_and_type(Arg1, Arg1Expr, Arg1Type, Arg1Functor),
    get_texpr_expr_functor_and_type(Arg2, Arg2Expr, Arg2Type, Arg2Functor),
    % root nodes for arguments
    add_term_and_symbol_nodes_to_graph(Arg1Expr, Arg1Type, Arg1Functor, Arg1RootNodeId),
    add_term_and_symbol_nodes_to_graph(Arg2Expr, Arg2Type, Arg2Functor, Arg2RootNodeId),
    (   is_commutative(Functor)
    ->  % add an edge from the root node to the root node of each argument
        bliss_interface:add_edge(RootNodeId, Arg1RootNodeId),
        bliss_interface:add_edge(RootNodeId, Arg2RootNodeId)
    ;   % special argument node for each argument
        add_node_to_colored_graph(arg, Arg1Expr, Arg1Type, Arg1NodeId),
        add_node_to_colored_graph(arg, Arg2Expr, Arg2Type, Arg2NodeId),
        % one edge from the argument node to the argument's root node
        bliss_interface:add_edge(Arg1NodeId, Arg1RootNodeId),
        bliss_interface:add_edge(Arg2NodeId, Arg2RootNodeId),
        % edge from first to second argument to represent the ordering
        bliss_interface:add_edge(Arg1NodeId, Arg2NodeId),
        % add an edge from the root node to the first argument's argument node
        bliss_interface:add_edge(RootNodeId, Arg1NodeId)
    ),
    build_colored_graph(Arg1Expr, Arg1Type, Arg1RootNodeId),
    build_colored_graph(Arg2Expr, Arg2Type, Arg2RootNodeId).
build_colored_graph(Term, _, _) :-
    (   is_interpreted_symbol(Term)
    ;   is_uninterpreted_symbol(Term)
    ),
    !.
build_colored_graph(Unary, _Type, RootNodeId) :-
    functor(Unary, _Functor, Arity),
    Arity == 1,
    !,
    arg(1, Unary, Arg),
    get_texpr_expr_functor_and_type(Arg, ArgExpr, ArgType, ArgFunctor),
    % root node for argument
    add_term_and_symbol_nodes_to_graph(ArgExpr, ArgType, ArgFunctor, ArgRootNodeId),
    % special argument node for each argument
    add_node_to_colored_graph(arg, ArgExpr, ArgType, ArgNodeId),
    % one edge from the argument node to the argument's root node
    bliss_interface:add_edge(ArgNodeId, ArgRootNodeId),
    % add an edge from the root node to the first argument's argument node
    bliss_interface:add_edge(RootNodeId, ArgNodeId),
    build_colored_graph(ArgExpr, ArgType, ArgRootNodeId).
build_colored_graph(Expr, _Type, _ExprRootNodeId) :-trace,
    add_error_and_fail(build_colored_graph, "Missing implementation for: ", [Expr]).

%% build_colored_graph_from_set(+List, +RootNodeId).
build_colored_graph_from_set([], _).
build_colored_graph_from_set([Elm|T], RootNodeId) :-
    % commutative
    get_texpr_expr_functor_and_type(Elm, ElmExpr, ElmType, ElmFunctor),
    add_term_and_symbol_nodes_to_graph(ElmExpr, ElmType, ElmFunctor, ElmRootNodeId),
    bliss_interface:add_edge(RootNodeId, ElmRootNodeId),
    build_colored_graph(ElmExpr, ElmType, ElmRootNodeId),
    build_colored_graph_from_set(T, RootNodeId).

create_nodes_for_ids([]).
create_nodes_for_ids([b(identifier(Id),Type,_)|T]) :-
    add_term_and_symbol_nodes_to_graph(identifier(Id), Type, identifier, _),
    create_nodes_for_ids(T).

have_seen_expr(Term, NodeId) :-
    term_hash(Term, TermHash),
    seen_expr(TermHash, NodeId).

have_seen_pred(Term, NodeId) :-
    term_hash(Term, TermHash),
    seen_pred(TermHash, NodeId).

log_seen_expr(Term, RootNodeId) :-
    term_hash(Term, TermHash),
    asserta(seen_expr(TermHash, RootNodeId)).

log_seen_pred(Term, RootNodeId) :-
    term_hash(Term, TermHash),
    asserta(seen_pred(TermHash, RootNodeId)).

%% add_term_and_symbol_nodes_to_graph(+Term, +Type, +Functor, -RootNodeId).
% Add a root node for compound term f(t1,...,tn). Add an edge from the root node to the (unique) symbol node for f.
% Add a node for (un)interpreted symbols.
add_term_and_symbol_nodes_to_graph(Term, _, _, RootNodeId) :-
    is_interpreted_symbol(Term),
    (   have_seen_expr(Term, TNodeId)
    ->  RootNodeId = TNodeId
    ;   get_next_color(Color),
        term_to_label(Term, ATerm),
        bliss_interface:add_node(ATerm, Color, RootNodeId),
        log_seen_expr(Term, RootNodeId)
        %asserta(ast_to_node_id(b(Term,Type,[]), RootNodeId))
        %asserta(node_id_to_ast(RootNodeId,b(Term,Type,[])))
    ), !.
add_term_and_symbol_nodes_to_graph(Term, Type, _, RootNodeId) :-
    is_uninterpreted_symbol(Term),
    (   have_seen_expr(Term, TNodeId)
    ->  RootNodeId = TNodeId
    ;   get_color_for_type(Type, Term, Color),
        term_to_label(Term, ATerm),
        bliss_interface:add_node(ATerm, Color, RootNodeId),
        log_seen_expr(Term, RootNodeId),
        asserta(ast_to_node_id(b(Term,Type,[]),RootNodeId)),
        asserta(node_id_to_ast(RootNodeId,b(Term,Type,[])))
    ), !.
add_term_and_symbol_nodes_to_graph(Term, Type, Functor, RootNodeId) :-
    categorize_type(Type, Category, SymbolCategory),
    add_node_to_colored_graph(SymbolCategory, Functor, Type, SymbolNodeId),
    add_node_to_colored_graph(Category, Term, Type, RootNodeId),
    % add an edge from the root node to the symbol node
    bliss_interface:add_edge(RootNodeId, SymbolNodeId).

%% add_node_to_colored_graph(+Category, +Symbol, +Type, -NodeId).
% Symbol can be a compound term like conjunct(_,_) or uninterpreted symbol like conjunct.
% Argument nodes are assigned a specific, unique color.
% Uninterpreted symbol nodes and root nodes are assigned a color based on their type.
% Each interpreted symbol such as the integer 1 is assigned a unique color.
add_node_to_colored_graph(arg, _, _, ArgNodeId) :-
    color(arg, Color),
    bliss_interface:add_node('arg', Color, ArgNodeId), !.
    %asserta(node_id_to_ast(ArgNodeId,b(Symbol,pred,[]))).
add_node_to_colored_graph(Category, Symbol, Type, NodeId) :-
    (   Category == pred, IsPred = true, IsExpr = false
    ;   Category == expr, IsExpr = true, IsPred = false
    ),
    (   (   (IsPred, have_seen_pred(Symbol, TNodeId))
        ;   (IsExpr, have_seen_expr(Symbol, TNodeId))
        )
    ->  NodeId = TNodeId
    ;   get_color_for_type(Type, Symbol, Color),
        term_to_label(Symbol, ASymbol),
        bliss_interface:add_node(ASymbol, Color, NodeId),
        (   IsExpr
        ->  log_seen_expr(Symbol, NodeId)
        ;   log_seen_pred(Symbol, NodeId)
        )
        %asserta(node_id_to_ast(NodeId,b(Symbol,Type,[])))
    ), !.
add_node_to_colored_graph(upred, USymbol, pred, NodeId) :-
    seen_upred(USymbol, TNodeId),
    !,
    NodeId = TNodeId.
add_node_to_colored_graph(upred, USymbol, pred, NodeId) :-
    get_next_color(Color),
    bliss_interface:add_node(USymbol, Color, NodeId),
    asserta(seen_upred(USymbol,NodeId)), !.
add_node_to_colored_graph(uexpr, USymbol, _, NodeId) :-
    seen_uexpr(USymbol, TNodeId),
    !,
    NodeId = TNodeId.
add_node_to_colored_graph(uexpr, USymbol, _, NodeId) :-
    get_next_color(Color),
    bliss_interface:add_node(USymbol, Color, NodeId),
    asserta(seen_uexpr(USymbol,NodeId)), !.

expr_functor(b(Expr,_,_), Functor) :-
    !,
    functor(Expr, Functor, _).
expr_functor(Expr, Functor) :-
    functor(Expr, Functor, _).

term_to_label(Term, Atom) :-
    Term =.. [Functor|Args],
    maplist(expr_functor, Args, ArgFunctors),
    NTerm =.. [Functor|ArgFunctors],
    write_to_codes(NTerm, Codes),
    atom_codes(Atom, Codes).

%% get_color_for_type(+Type, +Symbol, -Color).
get_color_for_type(Type, _, Color) :-
    color(Type, TColor),
    !,
    Color = TColor.
get_color_for_type(Type, _, Color) :-
    get_next_color(Color),
    asserta(color(Type,Color)).

%% get_next_color(-Color).
get_next_color(Color) :-
    retract(next_color(Color)),
    Color1 is Color + 1,
    asserta(next_color(Color1)).

%% categorize_type(+Type, -Category, -SymbolCategory).
% Only used for a faster lookup of asserted facts.
categorize_type(Type, Category, SymbolCategory) :-
    Type == pred,
    !,
    Category = pred,
    SymbolCategory = upred.
categorize_type(_, expr, uexpr).

%% is_commutative(+Term).
is_commutative(conjunct).
is_commutative(disjunct).
is_commutative(equivalence).
is_commutative(equal).
is_commutative(not_equal).
is_commutative(add).
is_commutative(multiplication).
is_commutative(union).
is_commutative(intersection).

is_uninterpreted_symbol(identifier(_)).

is_interpreted_symbol(Symbol) :-
    is_binary_interpreted_symbol(Symbol).
is_interpreted_symbol(boolean_true).
is_interpreted_symbol(boolean_false).
is_interpreted_symbol(max_int).
is_interpreted_symbol(min_int).
is_interpreted_symbol(empty_set).
is_interpreted_symbol(bool_set).
is_interpreted_symbol(real_set).
is_interpreted_symbol(string_set).
is_interpreted_symbol(real(_)).
is_interpreted_symbol(integer(_)).
is_interpreted_symbol(string(_)).
is_interpreted_symbol(value(_)).
is_interpreted_symbol(integer_set(_)).
is_interpreted_symbol(event_b_identity).

is_binary_interpreted_symbol(interval(_,_)).

%% zip_to_equalities_conj(+Ids, +EqVals, -Eqs).
zip_to_equalities_conj([Id|T], [EqVal|VT], Eqs) :-
    zip_to_equalities_conj(T, VT, b(equal(Id,EqVal),pred,[]), Eqs).

zip_to_equalities_conj([], [], Acc, Acc).
zip_to_equalities_conj([Id|T], [EqVal|VT], Acc, Eqs) :-
    NAcc = b(conjunct(b(equal(Id,EqVal),pred,[]),Acc),pred,[]),
    zip_to_equalities_conj(T, VT, NAcc, Eqs).

%% order_variables(+TypedVars, -TypeOrdering, -VariableOrdering).
% Assume that the variable set used in the SMT formula is an ordered set.
% Assume that there is some pre-defined ordering over types, and all the
% variables of a certain type ti appear before all the variables of another
% type tj in the variable ordering if type ti appears before type tj in the type ordering.
order_variables(TypedVars, TypeOrdering, VariableOrdering) :-
    order_variables_by_type(TypedVars, [], VarsOrderedByType),
    order_variables_within_typed_group(VarsOrderedByType, [], [], TypeOrdering, VariableOrdering).

%% order_variables_by_type(+TypedVars, +Acc, -VarsOrderedByType).
% Deterministic ordering of variables and corresponding types. The variables within the group of a type
% are sorted by their name using @<.
order_variables_by_type([], Acc, Acc).
order_variables_by_type([TypedVar|T], Acc, VarsOrderedByType) :-
    TypedVar = b(_,Type,_),
    extend_type_acc(Acc, Type, TypedVar, NAcc),
    order_variables_by_type(T, NAcc, VarsOrderedByType).

%% order_variables_within_typed_group(+TypeVarsTuples, +VarAcc, +TypeAcc, -TypeOrdering, -VariableOrdering).
order_variables_within_typed_group([], VAcc, TAcc, TAcc, VAcc).
order_variables_within_typed_group([(Type,TypeVars)|T], VAcc, TAcc, TypeOrdering, VariableOrdering) :-
    samsort(cmp_typed_identifier, TypeVars, STypeVars),
    append(STypeVars, VAcc, NVAcc),
    order_variables_within_typed_group(T, NVAcc, [Type|TAcc], TypeOrdering, VariableOrdering).

cmp_typed_identifier(b(identifier(Id1),_,_), b(identifier(Id2),_,_)) :-
    Id1 @< Id2.

%% extend_type_acc(+Acc, +Type, +TypedVar, -NAcc).
% Acc is a list of tuples (Type,TypeVars).
extend_type_acc(Acc, Type, TypedVar, NAcc) :-
    select((Type,TypedVars), Acc, RAcc),
    !,
    NAcc = [(Type,[TypedVar|TypedVars])|RAcc].
extend_type_acc(Acc, Type, TypedVar, NAcc) :-
    NAcc = [(Type,[TypedVar])|Acc].

:- use_module(library(plunit)).

:- begin_tests(order_variables).

test(order_variables_empty, [true(VariableOrdering == ExpectedVars), true(TypeOrdering == ExpectedTypes)]) :-
    Variables = [],
    order_variables(Variables, TypeOrdering, VariableOrdering),
    ExpectedVars = [],
    ExpectedTypes = [].

test(order_variables_single_type, [true(VariableOrdering == ExpectedVars), true(TypeOrdering == ExpectedTypes)]) :-
    Variables = [b(identifier(c),integer,[]),b(identifier(a),integer,[]),b(identifier(b),integer,[])],
    order_variables(Variables, TypeOrdering, VariableOrdering),
    ExpectedVars = [b(identifier(a),integer,[]),b(identifier(b),integer,[]),b(identifier(c),integer,[])],
    ExpectedTypes = [integer].

test(order_variables_two_types, [true(VariableOrdering == ExpectedVars), true(TypeOrdering == ExpectedTypes)]) :-
    Variables = [b(identifier(c),integer,[]),b(identifier(d),set(integer),[]),b(identifier(a),integer,[]),b(identifier(e),set(integer),[]),b(identifier(b),integer,[]),b(identifier(f),set(integer),[])],
    order_variables(Variables, TypeOrdering, VariableOrdering),
    ExpectedVars = [b(identifier(a),integer,[]),b(identifier(b),integer,[]),b(identifier(c),integer,[]),b(identifier(d),set(integer),[]),b(identifier(e),set(integer),[]),b(identifier(f),set(integer),[])],
    ExpectedTypes = [integer,set(integer)].

test(order_variables_two_types_set_first, [true(VariableOrdering == ExpectedVars), true(TypeOrdering == ExpectedTypes)]) :-
    Variables = [b(identifier(d),set(integer),[]),b(identifier(a),integer,[])],
    order_variables(Variables, TypeOrdering, VariableOrdering),
    ExpectedVars = [b(identifier(d),set(integer),[]),b(identifier(a),integer,[])],
    ExpectedTypes = [set(integer),integer].

test(order_variables_two_types_integer_first, [true(VariableOrdering == ExpectedVars), true(TypeOrdering == ExpectedTypes)]) :-
    Variables = [b(identifier(a),integer,[]),b(identifier(d),set(integer),[])],
    order_variables(Variables, TypeOrdering, VariableOrdering),
    ExpectedVars = [b(identifier(a),integer,[]),b(identifier(d),set(integer),[])],
    ExpectedTypes = [integer,set(integer)].

test(order_variables_three_types, [true(VariableOrdering == ExpectedVars), true(TypeOrdering == ExpectedTypes)]) :-
    Variables = [b(identifier(c),integer,[]),b(identifier(d),set(integer),[]),b(identifier(a),integer,[]),b(identifier(e),set(integer),[]),b(identifier(b),integer,[]),b(identifier(f),set(set(string)),[])],
    order_variables(Variables, TypeOrdering, VariableOrdering),
    ExpectedVars = [b(identifier(d),set(integer),[]),b(identifier(e),set(integer),[]),b(identifier(a),integer,[]),b(identifier(b),integer,[]),b(identifier(c),integer,[]),b(identifier(f),set(set(string)),[])],
    ExpectedTypes = [set(integer),integer,set(set(string))].

:- end_tests(order_variables).
