% (c) 2012-2020 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(external_function_declarations,[external_function_library/2,
               get_external_function_definition/3,
               available_external_library/1]).

:- use_module(module_information,[module_info/2]).
:- module_info(group,external_functions).
:- module_info(description,'This module provides utilities to query the available external functions.').

:- use_module(external_functions, [external_fun_type/3]).
:- use_module(error_manager).

% construct a raw DEFINITION for a given external function name
% works without access to the .def files in stdlib
get_external_function_definition(FunName,LibraryFilename,DEFINITION) :-
    external_function_library(FunName,Arity,PredOrExpr,LibraryFilename),
    external_fun_type(FunName,TypeParameters,ListTypeExpr),
    Pos = src_position_with_filename(0,0,-1,LibraryFilename),
    % print(def_ext_def(FunName,Arity,PredOrExpr,LibraryFilename,TypeParameters)),nl,
    inst_type_arguments(TypeParameters,Pos,1),
    get_external_fun_def_aux(PredOrExpr,FunName,Arity,TypeParameters,ListTypeExpr,Pos,DEFINITION).

get_external_fun_def_aux(expression,FunName,Arity,TypeParameters,ListTypeExpr,Pos,DEFINITION) :-
    DummyDef = string(Pos,'???'), % dummy definition in .def file; not used ??
    get_targs(1,Arity,Pos,FArgs),
    convert_to_function_type(ListTypeExpr,Pos,TypeExpr),
    !,
    %format('Adding ~w/~w of type: ',[FunName,Arity]),translate:print_raw_bexpr(TypeExpr),nl,
    DEFINITION = definition(expression,FArgs,
                            external_function_call(Pos,FunName,FArgs,DummyDef,
                                                   rewrite_protected(TypeParameters),
                                                   rewrite_protected(TypeExpr))).
get_external_fun_def_aux(predicate,FunName,Arity,TypeParameters,ListTypeExpr,Pos,DEFINITION) :-
    DummyDef = truth(Pos), % dummy definition in .def file; not used ??
    PredArity is Arity-1,
    get_targs(1,PredArity,Pos,FArgs),
    append(Types,[boolean],ListTypeExpr),
    convert_to_type_expr(Types,Pos,TypeExpr),
    !,
    DEFINITION = definition(predicate,FArgs,
                            external_pred_call(Pos,FunName,FArgs,DummyDef,
                                                   rewrite_protected(TypeParameters),
                                                   rewrite_protected(TypeExpr))).
get_external_fun_def_aux(T,F,Arity,TP,ListTypeExpr,Pos,_) :-
   add_internal_error('Could not add DEFINITION:',get_external_fun_def_aux(T,F,Arity,TP,ListTypeExpr,Pos,_)).

% instantiate type arguments
inst_type_arguments([],_,_).
inst_type_arguments([identifier(Pos,TArgID)|T],Pos,Nr) :- N1 is Nr+1,!,
  number_codes(Nr,NC), append("TYPE_ARG_",NC,C),
  atom_codes(TArgID,C),
  inst_type_arguments(T,Pos,N1).

% create formal argument list:
get_targs(N,Nr,_,Res) :- N>Nr,!, Res=[].
get_targs(Nr,Tot,POS,[identifier(POS,ArgID)|TA]) :- N1 is Nr+1,!,
  number_codes(Nr,NC), append("ARGUMENT_",NC,C),
  atom_codes(ArgID,C),
  get_targs(N1,Tot,POS,TA).


convert_to_type_expr([A],Pos,TA) :- !,
    convert_to_rawtype(A,Pos,TA).
convert_to_type_expr([A,B|T],Pos,Res) :-
    convert_to_type_expr([couple(A,B)|T],Pos,Res).

% convert a list of argument types to a total_function type expression:
convert_to_function_type([A],Pos,TA) :- !, % we have a constant function without arguments
    convert_to_rawtype(A,Pos,TA).
convert_to_function_type([A,B|T],Pos,Res) :- T \= [], !,
    convert_to_function_type([couple(A,B)|T],Pos,Res).
convert_to_function_type([A,B],Pos,total_function(Pos,TA,TB)) :-
    convert_to_rawtype(A,Pos,TA), convert_to_rawtype(B,Pos,TB).

:- use_module(library(lists),[maplist/3]).
% convert a Prolog type term to a raw AST type term
convert_to_rawtype(identifier(P,I),_,identifier(P,I)).
convert_to_rawtype(real,Pos,real_set(Pos)).
convert_to_rawtype(string,Pos,string_set(Pos)).
convert_to_rawtype(integer,Pos,integer_set(Pos)).
convert_to_rawtype(boolean,Pos,bool_set(Pos)).
convert_to_rawtype(set(Type),Pos,pow_subset(Pos,RT)) :- convert_to_rawtype(Type,Pos,RT).
convert_to_rawtype(seq(Type),Pos,seq(Pos,RT)) :- convert_to_rawtype(Type,Pos,RT).
convert_to_rawtype(couple(A,B),Pos,mult_or_cart(Pos,TA,TB)) :-
    convert_to_rawtype(A,Pos,TA), convert_to_rawtype(B,Pos,TB).
convert_to_rawtype(record(Fields),Pos,struct(Pos,TF)) :-
    maplist(conv_field(Pos),Fields,TF).

conv_field(Pos,field(ID,A),rec_entry(Pos,identifier(Pos,ID),TA)) :- convert_to_rawtype(A,Pos,TA).


available_external_library('LibraryStrings.def').
available_external_library('LibraryHash.def').
available_external_library('LibraryIO.def').
available_external_library('LibraryMath.def').
available_external_library('LibraryMeta.def').
available_external_library('LibraryReals.def').
available_external_library('LibraryRegex.def').
available_external_library('CHOOSE.def').
available_external_library('SCCS.def').
available_external_library('SORT.def').

% associate external functions with libraries (as found in stdlib)
external_function_library(Id,File) :- external_function_library(Id,_,_,File).

% we provide all predicates as BOOL expression, parser does not deal well with external predicates without access to DEFINITIONS ?!

external_function_library('CODES_TO_STRING',1,expression,'LibraryStrings.def').
external_function_library('DEC_STRING_TO_INT',2,expression,'LibraryStrings.def').
external_function_library('FORMAT_TO_STRING',2,expression,'LibraryStrings.def').
external_function_library('INT_TO_DEC_STRING',2,expression,'LibraryStrings.def').
external_function_library('INT_TO_HEX_STRING',1,expression,'LibraryStrings.def').
external_function_library('INT_TO_STRING',1,expression,'LibraryStrings.def').
external_function_library('STRING_APPEND',2,expression,'LibraryStrings.def').
external_function_library('STRING_CHARS',1,expression,'LibraryStrings.def').
external_function_library('STRING_CODES',1,expression,'LibraryStrings.def').
external_function_library('STRING_CONC',1,expression,'LibraryStrings.def').
external_function_library('STRING_CONTAINS_STRING',2,expression,'LibraryStrings.def'). % predicate
external_function_library('STRING_EQUAL_CASE_INSENSITIVE',2,expression,'LibraryStrings.def'). % predicate
external_function_library('STRING_IS_ALPHANUMERIC',1,expression,'LibraryStrings.def'). % predicate
external_function_library('STRING_IS_DECIMAL',1,expression,'LibraryStrings.def'). % predicate
external_function_library('STRING_IS_INT',1,expression,'LibraryStrings.def'). % predicate
external_function_library('STRING_IS_NUMBER',1,expression,'LibraryStrings.def'). % predicate
external_function_library('STRING_JOIN',2,expression,'LibraryStrings.def').
external_function_library('STRING_LENGTH',1,expression,'LibraryStrings.def').
external_function_library('STRING_REV',1,expression,'LibraryStrings.def').
external_function_library('STRING_SPLIT',2,expression,'LibraryStrings.def').
external_function_library('STRING_TO_INT',1,expression,'LibraryStrings.def').
external_function_library('STRING_TO_LOWER',1,expression,'LibraryStrings.def').
external_function_library('STRING_TO_UPPER',1,expression,'LibraryStrings.def').
external_function_library('STRINGIFY',1,expression,'LibraryStrings.def').
external_function_library('TO_INT',1,expression,'LibraryStrings.def').
external_function_library('TO_STRING',1,expression,'LibraryStrings.def').
external_function_library('STRING_PADLEFT',3,expression,'LibraryStrings.def').

external_function_library('HASH',1,expression,'LibraryHash.def').
external_function_library('SHA_HASH',1,expression,'LibraryHash.def').
external_function_library('SHA_HASH_HEX',1,expression,'LibraryHash.def').

external_function_library('printf',2,expression,'LibraryIO.def'). % predicate, provide as BOOL expression
external_function_library('vprintf',2,expression,'LibraryIO.def').
external_function_library('observe',1,expression,'LibraryIO.def'). % predicate, provide as BOOL expression

external_function_library('CHOOSE',1,expression,'CHOOSE.def').
external_function_library('MU',1,expression,'CHOOSE.def').

external_function_library('SCCS',1,expression,'SCCS.def').
external_function_library('CLOSURE1',1,expression,'SCCS.def').

external_function_library('SORT',1,expression,'SORT.def').
external_function_library('SQUASH',1,expression,'SORT.def').
external_function_library('REPLACE',3,expression,'SORT.def'). %provide external_fun def

external_function_library('GCD',2,expression,'LibraryMath.def').
external_function_library('MSB',1,expression,'LibraryMath.def').
external_function_library('ABS',1,expression,'LibraryMath.def').
external_function_library('FDIV',1,expression,'LibraryMath.def').
external_function_library('CDIV',1,expression,'LibraryMath.def').
external_function_library('SQRT',1,expression,'LibraryMath.def').
external_function_library('FACTORIAL',1,expression,'LibraryMath.def').

external_function_library('RADD',2,expression,'LibraryReals.def'). % could be useful in Event-B context
external_function_library('RSUB',2,expression,'LibraryReals.def'). % ditto
external_function_library('RMUL',2,expression,'LibraryReals.def'). % ditto
external_function_library('RDIV',2,expression,'LibraryReals.def'). % ditto
external_function_library('RZERO',0,expression,'LibraryReals.def'). % ditto
external_function_library('RONE',0,expression,'LibraryReals.def'). % ditto
external_function_library('RSIN',1,expression,'LibraryReals.def').
external_function_library('RCOS',1,expression,'LibraryReals.def').
external_function_library('RTAN',1,expression,'LibraryReals.def').
external_function_library('RCOT',1,expression,'LibraryReals.def').
external_function_library('RSINH',1,expression,'LibraryReals.def').
external_function_library('RCOSH',1,expression,'LibraryReals.def').
external_function_library('RTANH',1,expression,'LibraryReals.def').
external_function_library('RCOTH',1,expression,'LibraryReals.def').
external_function_library('RASIN',1,expression,'LibraryReals.def').
external_function_library('RACOS',1,expression,'LibraryReals.def').
external_function_library('RATAN',1,expression,'LibraryReals.def').
external_function_library('RACOT',1,expression,'LibraryReals.def').
external_function_library('RASINH',1,expression,'LibraryReals.def').
external_function_library('RACOSH',1,expression,'LibraryReals.def').
external_function_library('RATANH',1,expression,'LibraryReals.def').
external_function_library('RACOTH',1,expression,'LibraryReals.def').
external_function_library('REXP',1,expression,'LibraryReals.def').
external_function_library('RLOGe',1,expression,'LibraryReals.def').
external_function_library('RSQRT',1,expression,'LibraryReals.def').
external_function_library('RABS',1,expression,'LibraryReals.def').
external_function_library('ROUND',1,expression,'LibraryReals.def').
external_function_library('RSIGN',1,expression,'LibraryReals.def').
external_function_library('RFRACTION',1,expression,'LibraryReals.def').
external_function_library('RMAX',2,expression,'LibraryReals.def').
external_function_library('RMIN',2,expression,'LibraryReals.def').
external_function_library('RPOW',2,expression,'LibraryReals.def').
external_function_library('RLOG',2,expression,'LibraryReals.def').
external_function_library('RPI',0,expression,'LibraryReals.def').
external_function_library('REULER',0,expression,'LibraryReals.def').
external_function_library('RNORMAL',2,expression,'LibraryReals.def').
external_function_library('RAND',2,expression,'LibraryReals.def').


external_function_library('REGEX_MATCH',2,expression,'LibraryRegex.def'). % predicate, provide as BOOL expression
external_function_library('IS_REGEX',1,expression,'LibraryRegex.def'). % ditto
external_function_library('REGEX_REPLACE',3,expression,'LibraryRegex.def').
external_function_library('REGEX_SEARCH_STR',2,expression,'LibraryRegex.def').
external_function_library('REGEX_SEARCH',3,expression,'LibraryRegex.def').
external_function_library('REGEX_SEARCH_ALL',2,expression,'LibraryRegex.def').

external_function_library('ENABLED',1,expression,'LibraryMeta.def'). % predicate, provide as BOOL expression
external_function_library('PROB_INFO_STR',1,expression,'LibraryMeta.def'). % e.g., prob-version
external_function_library('PROB_STATISTICS',1,expression,'LibraryMeta.def').
external_function_library('PROJECT_INFO',1,expression,'LibraryMeta.def').
external_function_library('PROJECT_STATISTICS',1,expression,'LibraryMeta.def').
external_function_library('MACHINE_INFO',2,expression,'LibraryMeta.def').
external_function_library('GET_PREF',1,expression,'LibraryMeta.def').
external_function_library('HISTORY',0,expression,'LibraryMeta.def').



% TO DO: support externel substitutions

/*
  TO_STRING :
  definition(expression,[identifier(pos(215,1,102,13,102,16),sss)],
  external_function_call(pos(214,1,102,3,102,24),TO_STRING,
  [identifier(pos(215,1,102,13,102,16),sss)],
  string(pos(216,1,102,21,102,24),0),
  rewrite_protected([identifier(pos(218,1,103,31,103,45),TO_STRING_TYPE)]),
  rewrite_protected(total_function(pos(219,1,103,51,103,76),identifier(pos(220,1,103,51,103,65),TO_STRING_TYPE),string_set(pos(221,1,103,70,103,76))))))

def_ext_def(TO_STRING,1,LibraryStrings.def,[identifier(src_position_with_filename(0,0,-1,LibraryStrings.def),ARGUMENT_1)],[_3682515])
storing_ext_def(TO_STRING,LibraryStrings.def,

definition(expression,[identifier(src_position_with_filename(0,0,-1,LibraryStrings.def),ARGUMENT_1)],
external_function_call(src_position_with_filename(0,0,-1,LibraryStrings.def),TO_STRING,
[identifier(src_position_with_filename(0,0,-1,LibraryStrings.def),ARGUMENT_1)],
string(src_position_with_filename(0,0,-1,LibraryStrings.def),[63,63,63]),
rewrite_protected([identifier(src_position_with_filename(0,0,-1,LibraryStrings.def),TYPE_ARG_1)]),
rewrite_protected([identifier(src_position_with_filename(0,0,-1,LibraryStrings.def),TYPE_ARG_1),string]))))
  */
