% (c) 2009-2020 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)


:- module(junit_tests,[set_junit_dir/1,print_junit/2,create_junit_result/6,junit_mode/1,
   create_and_print_junit_result/5]).

:- use_module(library(xml)).
:- use_module(library(lists)).
:- use_module(library(codesio)).
:- use_module(library(file_systems)).
:- use_module(library(process)).
:- use_module(library(sets)).
:- use_module(library(system)).

:- use_module(module_information,[module_info/2]).
:- module_info(group,infrastructure).
:- module_info(description,'This module provides functionality to output test case results in a format compatible to junit.').


:- dynamic junit_mode/1.

set_junit_dir(Directory) :- assert(junit_mode(Directory)).

iso8601_datetime(ISODatime) :-  datime(datime(Year,Month,Day,Hour,Min,Sec)),
  format_to_codes('~`0t~d~2|-~`0t~d~3+-~`0t~d~3+T~`0t~d~3+:~`0t~d~3+:~`0t~d~3+',
                  [Year,Month,Day,Hour,Min,Sec], ISODatime).



%   <testcase name="Call" classname="Selfcheck.Module" time="2"><error type="error"></testcase>

create_junit_result(Name,Section, Module, Time, Verdict, result(Name,Section,Module,TR,Verdict)) :-
    junit_mode(_),!,
    convert_to_junit_time(Time,TR).

create_junit_result(_Call, _Section, _Module, _Time, _Verdict, none).

convert_to_junit_time(T,R) :- number(T),!, R is T / 1000.

create_and_print_junit_result(Name,Section, Module, Time, Verdict) :-
    create_junit_result(Name,Section, Module, Time, Verdict,Result),
    print_junit([Result],Section).


print_junit(Results,Module) :-
          junit_mode(Dir),!,print_junit2(Dir,Results,Module).
print_junit(_Results,_Module).

:- use_module(tools_strings,[ajoin/2]).
:- use_module(tools_files,[put_codes/2]).
:- use_module(error_manager).
print_junit2(Dir,Results,Module) :-
          prepare_xml(Results,Codes),
          open_file(0,Dir,Module,Stream),
          put_codes(Codes,Stream),
          close(Stream),!.
print_junit2(Dir,_Results,Module) :- add_error(junit_tests,'print_junit failed',Dir:Module).

:- use_module(tools_meta,[safe_on_exception/3]).
open_file(C,Dir,Module,Stream) :- process_id(PID),
    number_codes(C,CC), atom_codes(Num,CC),
    ajoin([Dir,'/',Num,'_',PID,'_',Module,'.xml'], File),
    open_file2(C,File,Dir,Module,Stream).
open_file2(C,F,D,M,Stream) :- file_exists(F),!, C1 is C+1, open_file(C1,D,M,Stream).
open_file2(_C,File,_D,_M,Stream) :-
   safe_on_exception(E,open(File, write,Stream),
            (print('### Exception while opening Junit File: '), print(File),nl,
             add_error(junit_tests,'Junit File Opening Exception: ',File:E), fail)).

prepare_xml(Results,Codes) :-
   maplist(prepare_testcase_xml, Results,X),
   % count tests
   length(Results,TestsNr),
   number_codes(TestsNr, Tests),
    % count tests marked as error
   count_results(error(_), Results, ErrorsNr),
   number_codes(ErrorsNr, Errors),
   % count skipped tests
   count_results(skip, Results, SkippedNr),
   number_codes(SkippedNr, Skipped),
   % sum all test runtimes
   maplist(extract_times, Results,Times),
   sumlist(Times, TimeNr),
   format_to_codes('~6F', [TimeNr], Time),
   append([element(properties, [], [])|X],
          [element('system-out',[], []), element('system-err', [], [])], Children),
   maplist(extract_testcase_section, Results, SectionsL),
   list_to_set(SectionsL, Sections),
   %format('Sections ~w~n',[Sections]),
   [Section|_] = Sections,
   atom_codes(Section,SC),
   iso8601_datetime(DT),
   xml_parse(Codes,
     xml([version="1.0",encoding="UTF-8"],
       element(testsuite,
               [name=SC,hostname="Test Runner",tests=Tests,
                skipped=Skipped,failures="0",
                errors=Errors,time=Time,timestamp=DT],
               Children))).
prepare_xml(Results,_Codes) :- add_error_fail(junit_tests,'prepare_xml failed',Results).

extract_testcase_section(Result,_) :- var(Result),!,
          add_error_fail(junit_tests,'extract_testcase_section called on non-ground result',Result).
extract_testcase_section(result(_,Section,_,_,_),Section).

prepare_testcase_xml(Result,_) :- var(Result),!,
          add_error_fail(junit_tests,'prepare_testcase_xml called on non-ground result',Result).
prepare_testcase_xml(result(Call,Section,Module,Time,Verdict),element(testcase,R,Error)) :-
          R = [name=Name, classname=Classname, time=T],
           write_to_codes(Call,Name),
           atom_codes(Module,MC),
           atom_codes(Section,SC),
           ( Verdict=pass     -> Error=[]
           ; Verdict=error(E) -> (createError(E,Err), Error=[element(error,['='(type,"Error")],Err)])
           ; Verdict=skip     -> Error=[element(skipped,[],[])]),
           append([SC,".",MC],Classname),
           format_to_codes('~6F', [Time], T).

createError([],[]).
createError([E|T],[pcdata(Error)|R]) :- write_to_codes(E,Codes), append(["\n",Codes,"\n"],Error), createError(T,R).

extract_times(Result,_) :- var(Result),!,
          add_error_fail(junit_tests,'extract_times called on non-ground result',Result).
extract_times(result(_,_,_,Time,_),Time).

count_results(_,Results,_) :- var(Results),!,
          add_error_fail(junit_tests,'count_results called on non-ground list of results',Results).
count_results(Type,Results,Sum) :- count_results2(Type, Results, Sum).

count_results2(_, [], 0).
count_results2(Type,[result(_,_,_,_,Type)|Tail],Result) :- !, count_results2(Type,Tail,Temp), Result is 1 + Temp.
count_results2(Type,[result(_,_,_,_,_)|Tail],Result) :- !, count_results2(Type,Tail,Result).
