% (c) 2020-2020 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(kernel_reals, [construct_real/2,
                         is_real/1, is_real/2, is_real_wf/2,
                         is_float/1, is_float_wf/2,
                         is_not_real/1, is_not_float/1,
                         is_ground_real/1,
                         convert_int_to_real/2,
                         real_floor/2, real_ceiling/2,
                         real_addition_wf/4, real_subtraction_wf/4,
                         real_multiplication_wf/4, real_division_wf/4,
                         real_unary_minus_wf/3,
                         real_round_wf/3,
                         real_unop_wf/4, real_binop_wf/5,
                         real_power_of_wf/5,
                         real_less_than_wf/3, real_less_than_equal_wf/3,
                         real_comp_wf/5
  ]).

% Reals/Floats in ProB are represented by terms of the form term(floating(Number))
% in future a proper wrapper such as real(Number) might be created

:- use_module(error_manager).
:- use_module(self_check).
:- use_module(library(lists)).

:- use_module(module_information,[module_info/2]).
:- module_info(group,kernel).
:- module_info(description,'This module provides (external) functions to manipulate B reals and floats.').


% used to construct a Value from real(Atom) AST node
construct_real(Atom,term(floating(Float))) :-
    atom_codes(Atom,C),
    number_codes(Nr,C),
    Float is float(Nr). % make sure we store a float; not required for AST literals

is_real_wf(X,_WF) :- is_real(X,_).
is_real(X) :- is_real(X,_).

is_real(term(floating(Nr)),Nr).
% TO DO: other formats

is_float_wf(X,_WF) :- is_float(X).

is_float(term(floating(_))).

is_not_real(_) :- fail.
is_not_float(_) :- fail.

is_ground_real(term(floating(Nr))) :- number(Nr).

% convert an integer to a real, corresponds to the real(.) function in Atelier-B; convert_real as AST
convert_int_to_real(int(X),term(floating(R))) :-
    convert_int_to_real_aux(X,R).

:- block convert_int_to_real_aux(-,-).
convert_int_to_real_aux(Int,Real) :- number(Int),!,
   Real is float(Int).
convert_int_to_real_aux(Int,Real) :- Int1 is floor(Real),
   Real is float(Int1),
   Int1=Int.


% convert_int_floor as AST, Atelier-B floor(.)
real_floor(term(floating(R)),int(X)) :- real_floor_aux(R,X).
:- block real_floor_aux(-,?).
real_floor_aux(Real,Int) :- Int is floor(Real).

% convert_int_ceiling as AST, Atelier-B ceiling(.)
real_ceiling(term(floating(R)),int(X)) :- real_ceiling_aux(R,X).
:- block real_ceiling_aux(-,?).
real_ceiling_aux(Real,Int) :- Int is ceiling(Real).


real_addition_wf(term(floating(X)),term(floating(Y)),term(floating(R)),WF) :-
    real_add_wf_aux(X,Y,R,WF).
:- block real_add_wf_aux(-,?,?,?), real_add_wf_aux(?,-,?,?).
real_add_wf_aux(X,Y,R,WF) :- safe_is(R,X+Y,WF).

real_subtraction_wf(term(floating(X)),term(floating(Y)),term(floating(R)),WF) :-
    real_sub_wf_aux(X,Y,R,WF).
:- block real_sub_wf_aux(-,?,?,?), real_sub_wf_aux(?,-,?,?).
real_sub_wf_aux(X,Y,R,WF) :- safe_is(R,X-Y,WF).

real_multiplication_wf(term(floating(X)),term(floating(Y)),term(floating(R)),WF) :-
    real_mul_wf_aux(X,Y,R,WF).
:- block real_mul_wf_aux(-,?,?,?), real_mul_wf_aux(?,-,?,?).
real_mul_wf_aux(X,Y,R,WF) :- safe_is(R,X*Y,WF).

real_division_wf(term(floating(X)),term(floating(Y)),term(floating(R)),WF) :-
    real_div_wf_aux(X,Y,R,WF).
:- block real_div_wf_aux(-,?,?,?), real_div_wf_aux(?,-,?,?).
real_div_wf_aux(X,Y,R,WF) :- safe_is_div(R,X/Y,WF).


real_unary_minus_wf(term(floating(X)),term(floating(R)),WF) :-
    real_um_wf_aux(X,R,WF).
:- block real_um_wf_aux(-,?,?).
real_um_wf_aux(X,R,_) :- R is -X.

real_round_wf(term(floating(X)),int(R),_WF) :-
    when(nonvar(X), R is round(X)).

:- use_module(kernel_waitflags,[add_wd_error/3]).
% a version of is/2 which catches overflows
safe_is(R,Expr,WF) :-
     on_exception(error(evaluation_error(float_overflow),_),
                  R is Expr,
                  add_wd_error('Float Overflow while computing:',Expr,WF)).


safe_is_div(R,Expr,WF) :-
     %print(safe_is_div(R,Expr,WF)),nl,trace,
     on_exception(error(evaluation_error(zero_divisor),_),
                  safe_is(R,Expr,WF),
                  add_wd_error('Division by zero while computing:',Expr,WF)).

% call a Prolog unary artihmetic operator
real_unop_wf(OP,term(floating(X)),term(floating(R)),WF) :-
    real_unop_wf_aux(OP,X,R,WF).
:- block real_unop_wf_aux(-,?,?,?), real_unop_wf_aux(?,-,?,?).
real_unop_wf_aux(OP,X,R,WF) :-
     Expr =.. [OP,X],
     safe_is(R,Expr,WF).

real_power_of_wf(X,Y,Res,_Span,WF) :-
    real_binop_wf('exp',X,Y,Res,WF).

% call a Prolog binary artihmetic operator
real_binop_wf(OP,term(floating(X)),term(floating(Y)),term(floating(R)),WF) :-
    real_binnop_wf_aux(OP,X,Y,R,WF).
:- block real_binnop_wf_aux(-,?,?,?,?), real_binnop_wf_aux(?,-,?,?,?), real_binnop_wf_aux(?,?,-,?,?).
real_binnop_wf_aux(OP,X,Y,R,WF) :-
     Expr =.. [OP,X,Y],
     safe_is(R,Expr,WF).

real_less_than_wf(X,Y,WF) :- real_comp_wf('<',X,Y,pred_true,WF).
real_less_than_equal_wf(X,Y,WF) :- real_comp_wf('=<',X,Y,pred_true,WF).

% call a Prolog bianry artihmetic operator
real_comp_wf(OP,term(floating(X)),term(floating(Y)),R,WF) :-
    real_comp_wf_aux(OP,X,Y,R,WF).
:- block real_comp_wf_aux(-,?,?,?,?), real_comp_wf_aux(?,-,?,?,?), real_comp_wf_aux(?,?,-,?,?).
real_comp_wf_aux(OP,X,Y,R,_) :-
     (call(OP,X,Y) -> R=pred_true ; R=pred_false).



% not used yet: useful for kernel_strings ?
%:- block real_to_string(-,?).
%real_to_string(term(floating(I)),S) :- real_to_string2(I,S).
%
%:- block real_to_string2(-,?).
%real_to_string2(Num,Res) :-
%   number_codes(Num,C),
%   atom_codes(S,C), Res=string(S).


