% (c) 2009-2020 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)


:- module(kernel_waitflags,
   [init_wait_flags/1, init_wait_flags/2,
    waitflag0_is_set/1,
    get_wait_flag_infos/2, get_wait_flag_info/2, is_wait_flag_info/2,
    add_wait_flag_info/3,
    get_wait_flag/3, get_wait_flag/4, get_bounded_wait_flag/4,
    block_get_wait_flag/4,
    get_wait_flag0/2,
    get_wait_flag1/2,  get_wait_flag1/3, % deterministic wait flag
    get_last_wait_flag/3,
    get_idle_wait_flag/3,
    get_binary_choice_wait_flag/3,
    get_pow2_binary_choice_priority/2,
    get_binary_choice_wait_flag_exp_backoff/3, get_binary_choice_wait_flag_exp_backoff/4,
    %refresh_wait_flag/4,
    add_fd_variable_for_labeling/3, add_fd_variable_for_labeling/2, add_fd_variables_for_labeling/2,
    get_new_subsidiary_wait_flag/4,
    update_waitflag/4,
    ground_constraintprop_wait_flags/1,
    check_is_wait_flag/1,
    no_pending_waitflags/1,
    create_inner_wait_flags/3, copy_wf01e_wait_flags/2,
    copy_wfe_from_inner_if_necessary/2,
    %copy_wf_start/2,
    copy_wf_start/3,
    copy_wf_finish/2,
    copy_waitflag_store/2,

    clone_wait_flags_from1/2, clone_wait_flags_from1_finish/2,
    ground_wait_flag0/1,
    %create_wdguarded_wait_flags/4,
    get_large_finite_wait_flag/3,
    get_enumeration_starting_wait_flag/3,
    %get_enumeration_almost_finished_wait_flag/3,
    get_enumeration_finished_wait_flag/2,
    get_integer_enumeration_wait_flag/3,
    ground_det_wait_flag/1, ground_wait_flag_to/2,
    ground_wait_flags/1, ground_inner_wait_flags/1,
    deref_wf/2,

    add_abort_error_span/5,
    add_wd_error_span/4,
    add_wd_error/3, add_wd_error_set_result/6,
    pending_abort_error/1, pending_abort_error/4,
    get_wait_flags_context_msg/2,
    %add_wd_error_span_now/3,
    assert_must_abort_wf/2,

    get_minimum_waitflag_prio/3,
    portray_waitflags/1, quick_portray_waitflags/1, quick_portray_waitflags_store/1,
    find_waitflag_info/4,

    integer_priority/1, large_finite_priority/1, integer_pow2_priority/1,

    my_get_kodkod_predicates/3, my_add_predicate_to_kodkod/3,
    my_get_satsolver_predicates/3, my_add_predicate_to_satsolver/3,

    set_silent/1]).
/* File: kernel_waitflags.pl */
/* Created: 17/3/06 by Michael Leuschel */

:- meta_predicate assert_must_abort_wf(0,*).
%:- meta_predicate assert_must_abort2_wf(0,-).

/*
Phases of the ProB Kernel:

 -> all Waitflags unbound

* Phase 0:  propagation of completely known ground values in a deterministic fashion

 -> WF0 is bound (first to s(VAR) and then to s(0))
                  when WF0 is not ground one knows that the WF0 phase is still ongoing
                  (relevant for getting 1.0 waitflag)

* Phase 1:  deterministic propagation (not necessarily of completely known full values)

 -> WF1 is bound (1.0)

* Phase 2:  boolean propagation (e.g., P or Q will now try P or Q)

 -> WF2 is bound

* Phase 3:  non-deterministic propagation (e.g, x: {1,2,3})

 -> Enumeration Starting WF is bound

* Enumeration Phase:  (later to be split into enumeration of finite and infinite types (e.g., integer)

  -> Enumeration Finished Waitflag bound

* Error Generation Phase: errors for, e.g., division by zero, f(x) with x outside domain of f,... are generated
* This phase should not create choice points and should not detect failure
  (e.g., in the
  case of negated existential quantifiers the Enumeration Finished Waitflag will be set outside of delay_not)


We now distinguish
 - Enumeration Wait Flags for Decision Variables with estimated enumeration domain size as priority
 - Binary Choice Point Wait Flags (stemming e.g. from disjunctions)

Note: all floats like 1.0, 1.5, 2.0 ... are dealt with before all integers (even 1) !

There are three ways to create a waitflag store:
 - init_wait_flags : to create a fresh empty store
 - create_inner_wait_flags: the inner WFE should not be instantiated and will be instantiated from the outside
     useful for a nested quantifier where well-definedness errors should only be raised when the outer quantifier
     is set and where you want to enumerate the inner quantifier in isolation (e.g., for exists to be able to use the Prolog cut)
 - copy_wf_start to create a temporary copy with WF0 unset, which will be unified with copy_wf_finish
    useful for adding a new predicate to the constraint solver and ensure that WF0 is not set until
    predicate has been fully interpreted, to avoid enumerations/inefficient computations (e.g., there
    could be an obvious inconsistency at the end of the predicate)
*/

:- use_module(error_manager).
:- use_module(self_check).
:- use_module(typechecker).
:- use_module(library(clpfd),[fd_size/2, fd_degree/2]).
:- use_module(tools).
:- use_module(coverage_tools_annotations,['$NOT_COVERED'/1]).
:- use_module(preferences,[preference/2]).

:- use_module(module_information,[module_info/2]).
:- module_info(group,kernel).
:- module_info(description,'This module manages the various choice points and enumerators of ProB, by maintaining the kernel Waitflags store.').

:- type wait_flag +--> (var ; wfx(simple,mutable,simple,list(any)) ; no_wf_available).
% a mutable either contains:
%  wfm_store(AVLWaitflagHeap,NewFDVARS,KodSatPredicateList)
%  wfm_ref(Mutable) a reference pointing to another Mutable which should be used to store waitflags

:- use_module(library(avl)).

% -------------- ATTRIBUTES ------------------

:- use_module(library(atts)).
:- attribute %wf_store/1,wf_fdvars/1, wf_kodkod/2, wf_satsolver/2, % now stored in mutable
             wf_abort_pending/0.

% verify_attributes is called whenever a variable Var that might have attributes is about to be bound to Value (it might have none).
% verify(VariableBeforeUnification, ValueWithWhichVariableUnifies,  GoalWhichSICStusShouldCallAfterUnif)
verify_attributes(Var, Value, [] ) :- get_atts(Var,+wf_abort_pending),!,
   % we are unifying the EWF enumeration finished waitflag, not the store
   (nonvar(Value) -> true % EWF bound, enumeration finished
    ; get_atts(Value,+wf_abort_pending) -> true % other EWF Value already has same attribute
    ; put_atts(Value,+wf_abort_pending)
   ).
verify_attributes(Var, Value, [] ) :-
  (nonvar(Value) -> print('Trying to unify WFE flags'),nl
        ;  print(verify_attributes(Var,Value)),nl).
        % The SICStus documentation is not entirely clear: it seems we only have to copy attributes from Var to Value, not the other way around

% -------------- MUTABLES ------------------

% Should we use several mutables?
put_mutable_attr(Store,wf_store(New)) :- !,
    get_mutable_store_for_update(Store,_S,FD,KodSat,StoreToUpdate),
    update_mutable(wfm_store(New,FD,KodSat),StoreToUpdate).
put_mutable_attr(Store,wf_fdvars(NewFDVARS)) :- !,
    get_mutable_store_for_update(Store,S,_FD,KodSat,StoreToUpdate),
    update_mutable(wfm_store(S,NewFDVARS,KodSat),StoreToUpdate).
put_mutable_attr(Store,wf_preds(ID,Type,PredList)) :- !,
    get_mutable_store_for_update(Store,S,FD,KodSat,StoreToUpdate),
    update_preds(KodSat,ID,Type,PredList,KodSat2),
    update_mutable(wfm_store(S,FD,KodSat2),StoreToUpdate).
put_mutable_attr(Store,New) :-
    add_internal_error('Uncovered put_mutable_attr',put_mutable_attr(Store,New)).

% TO DO: improve my_get_wf_store_fdvars_atts / put_mutable_attr pairs of calls to make them more efficient

% update Kodkod or Satsolver predicates attached in store:
update_preds([],ID,Type,PredList,[wf_preds(ID,Type,PredList)]).
update_preds([wf_preds(ID,Type,_)|T],ID,Type,PredList,Res) :- !,
    Res = [wf_preds(ID,Type,PredList)|T].
update_preds([H|T],ID,Type,PredList,[H|RT]) :-
    update_preds(T,ID,Type,PredList,RT).

% a utility predicate that always returns a mutable with wfm attribute
get_mutable_store_for_update(Store, S,FD,KodSat, MutableToUpdate) :- mutable(Store),!,
    get_mutable(MutableInfo,Store),
    (get_mutable_infos__for_update_aux(MutableInfo,S,FD,KodSat,Store,MutableToUpdate) -> true
      ; add_internal_error('get_mutable_infos__for_update_aux failed for:',Store),fail).
get_mutable_store_for_update(Store, S,[],[], Res) :- var(Store),!, Res=Store,
    empty_avl(S),
    create_mutable(wfm_store(S,[],[]),Store).
get_mutable_store_for_update(Store, _,_,_, _) :-
    add_internal_error('get_mutable_store_for_update failed for:',Store),fail.

get_mutable_infos__for_update_aux(wfm_store(S,FD,KodSat),S,FD,KodSat,Store,Store).
get_mutable_infos__for_update_aux(wfm_ref(OuterStore),S,FD,KodSat,_,MutableToUpdate) :-
    get_mutable_store_for_update(OuterStore,S,FD,KodSat,MutableToUpdate).

% follow links of wfm_ref to get access to proper mutable store to update
deref_store(Store,MutableToUpdate) :- mutable(Store), !,
    (Store=wfm_ref(Store2) -> deref_store(Store2,MutableToUpdate)
      ; MutableToUpdate=Store).
deref_store(Store,MutableToUpdate) :- var(Store),!, MutableToUpdate=Store,
    empty_avl(S),
    create_mutable(wfm_store(S,[],[]),Store).
deref_store(Store, _) :-
    add_internal_error('deref_store failed for:',Store),fail.

% get the associated infos for a mutable store; follow chain of wfm_ref's:
get_mutable_infos(S,FD,KodSat,Store) :- mutable(Store),
    get_mutable(MutableInfo,Store),
    get_mutable_infos_aux(MutableInfo,S,FD,KodSat).
get_mutable_infos_aux(wfm_store(S,FD,KodSat),S,FD,KodSat).
get_mutable_infos_aux(wfm_ref(OuterStore),S,FD,KodSat) :- get_mutable_infos(S,FD,KodSat,OuterStore).

my_get_wf_store_att(Store,Heap) :- get_mutable_infos(S,_,_,Store),!,
   Heap=S.
my_get_wf_store_att(_,E) :- empty_avl(E).

my_get_wf_preds_att(Store,ID,Type,PredList) :- get_mutable_infos(_,_,KodSat,Store),!,
   (member(wf_preds(ID,Type,R),KodSat) -> PredList=R).

my_get_fdvars_att(Store,FDList) :- get_mutable_infos(_,FD,_,Store),!,
   FDList=FD.
my_get_fdvars_att(_,[]).

my_get_wf_store_fdvars_atts(Store,Heap,FDList) :- get_mutable_infos(S,FD,_,Store),!,
   Heap=S, FDList=FD.
my_get_wf_store_fdvars_atts(_,E,[]) :- empty_avl(E).

% Mutable store should now point to another mutable store (OuterStore)
% ensures that calls which add waitflags to Store now get-redirected to OuterStore
set_mutable_ref(Store,OuterStore) :- mutable(Store),
    !,
    update_mutable(wfm_ref(OuterStore),Store).
set_mutable_ref(Store,OuterStore) :- % not instantiated yet, we can simply unify
    Store=OuterStore.


% -------------------------

% the Call must succeed and generate one abort_error
assert_must_abort_wf(M:Call,WF) :- assert_must_succeed_any(M:(kernel_waitflags:assert_must_abort2_wf(M:Call,WF))).

:- use_module(tools_printing,[format_with_colour/4]).
assert_must_abort2_wf(_,_) :-
   get_error(well_definedness_error,M),
   add_internal_error('Abort error prior to call in assert_must_abort: ',M),fail.
assert_must_abort2_wf(Call,WF) :-
  (real_error_occurred -> RE=true ; RE=false),
  init_wait_flags(WF,[assert_must_abort2_wf]),
  call(Call),
  (ground_wait_flags(WF)
  -> add_internal_error('Call did not fail after grounding WF: ',assert_must_abort2_wf(Call,WF)),fail
  ;  get_error(well_definedness_error,_), % we have 1 abort_error
     format_with_colour(user_output,[blue],'Got expected well-definedness error~n',[]),
     (RE=false -> reset_real_error_occurred ; true)
  ).


:- use_module(library(lists),[is_list/1]).
:- assert_must_succeed(( kernel_waitflags:init_wait_flags(WF), kernel_waitflags:check_is_wait_flag(WF) )).

check_is_wait_flag(WFX) :- var(WFX),!,
  add_internal_error('Variable as waitflag: ',check_is_wait_flag(WFX)).
check_is_wait_flag(no_wf_available) :- !.
check_is_wait_flag(WFX) :-
  WFX = wfx(WF0,Store,WFE,Infos)
    -> (get_mutable_infos(_,_,_,Store)
         -> true
          ; print('WF does not yet contain heap: '), print(WFX),nl),
            %add_internal_error('WF does not contain heap: ',check_is_wait_flag(WFX)))
       (var(WF0) -> true ; WF0 = s(WF00), \+compound(WF00) -> true
        ; compound(WF0) -> add_internal_error('WF0 is compound: ',WF0) ; true
       ),
       (compound(WFE) -> add_internal_error('WFE is compound: ',WFE) ; true),
       (nonvar(Infos),is_list(Infos) -> true ; add_internal_error('WF Infos not list: ',Infos))
    ;  add_internal_error('Illegal WF Format: ',check_is_wait_flag(WFX)).


% initialise a new waitflag store
init_wait_flags(WF,Info) :-
   (nonvar(Info) -> true ; add_internal_error('Illegal WF Format: ',init_wait_flags(WF,Info))),
   init_debug_wait_flag_info(Info,Infos),
   WF=wfx(_,_Store,_,Infos).
init_wait_flags(wfx(_,_Store,_,[no_expansion_context_available])).
%init_wait_flags(wfx(_,Store,_,[])) :- init_wait_flags_store(Store).
% we no longer do this because of performance, (see, e.g., LotsOfInvariants.mch) :
init_wait_flags_store(Store) :- empty_avl(Heap),
   create_mutable(wfm_store(Heap,[],[]),Store).
%init_wait_flags_store(Store) :-
%	empty_avl(Heap),
%	put_atts(Store,wf_store(Heap)),
%	put_atts(Store,wf_fdvars([])).


:- assert_must_succeed(( kernel_waitflags:init_wait_flags(WF),
     kernel_waitflags:ground_wait_flag0(WF), kernel_waitflags:waitflag0_is_set(WF) )).
:- assert_must_fail(( kernel_waitflags:init_wait_flags(WF), kernel_waitflags:waitflag0_is_set(WF) )).
waitflag0_is_set(wfx(WF0,_Store,_EF,_INFOS)) :- ground(WF0).
waitflag0_is_set(no_wf_available).

get_new_subsidiary_wait_flag(OldWF,Info,WFX,NewWF) :-
   (ground(OldWF) /* then we have created a choice point with the old WF: double new WF prio */
      -> NewPrio is OldWF*2, get_wait_flag(NewPrio,Info,WFX,NewWF)
      ;  NewWF = OldWF
   ).

% not used at the moment:
%refresh_wait_flag(OldWF,Info,WFX,NewWF) :-
%   (ground(OldWF) /* then we have created a choice point with the old WF: double new WF prio */
%      -> get_wait_flag(OldWF,Info,WFX,NewWF)
%      ;  NewWF = OldWF
%   ).

:- volatile silent/0.
:- dynamic silent/0.
set_silent(true) :- !,(silent -> true ; assert(silent)).
set_silent(false) :- retractall(silent).
%waitflag_not_init :-  silent -> true ; print('waitflag-store not initialised'),nl, '$NOT_COVERED'('This should not happen').

:- use_module(eventhandling,[register_event_listener/3]).
:- register_event_listener(start_unit_tests,set_silent(true),'Allow wf not to be set up without printing a warning').
:- register_event_listener(stop_unit_tests, set_silent(false), 'Printing wf warnings again').

:- use_module(library(lists),[maplist/2]).
add_fd_variables_for_labeling(Vars,WF) :- WF==no_wf_available,!,
   add_internal_error('Cannot add FDVars to waitflag store: ',Vars).
add_fd_variables_for_labeling(Vars,WF) :-
   WF=wfx(_,Store,_EnumFinished,_INFOS),
   !,
   my_get_fdvars_att(Store,FDVARS),
   l_add_fd_var_to_FDVARS(Vars,FDVARS,NewFDVARS),
   put_mutable_attr(Store,wf_fdvars(NewFDVARS)).
add_fd_variables_for_labeling(Vars,WF) :-
   add_internal_error('Illegal waitflag store: ',add_fd_variables_for_labeling(Vars,WF)).

l_add_fd_var_to_FDVARS([],Acc,Acc).
l_add_fd_var_to_FDVARS([Var|VT],FDVARS,NewFDVARS) :-
    (nonvar(Var) -> F2=FDVARS;  add_fd_var_to_FDVARS(Var,FDVARS,F2)),
    l_add_fd_var_to_FDVARS(VT,F2,NewFDVARS).

% adds a CLP(FD) variable with given domain Size, will be enumerated by labeling (even if clpfd_solver preference is false !)
% Size could be computed by fd_size(FDVariable,Size)
add_fd_variable_for_labeling(FDVariable,WF) :-
    add_fd_variable_for_labeling_aux(FDVariable,WF).
add_fd_variable_for_labeling(FDVariable,Size,WF) :- Size = not_used_anymore,
    add_fd_variable_for_labeling_aux(FDVariable,WF).


add_fd_variable_for_labeling_aux(FDVariable,_WF) :- nonvar(FDVariable),!.
add_fd_variable_for_labeling_aux(FDVariable,WF) :-
  % otherwise: store the FDVariable in the separate list of FD Variables (all mixed together)
  WF=wfx(_,Store,_EnumFinished,_),
  my_get_fdvars_att(Store,FDVARS),
  add_fd_var_to_FDVARS(FDVariable,FDVARS,NewFDVARS),
   %%  fd_size(FDVariable,Size),print(added(FDVariable,Size,NewFDVARS,old(FDVARS))),nl,
    %%  print(added_fd_variable(FDVariable,_Size,NewFDVARS)),nl,portray_waitflags_store(Store),nl,nl,%%
  put_mutable_attr(Store,wf_fdvars(NewFDVARS)).

add_fd_var_to_FDVARS(FDVariable,FDVARS,NewFDVARS) :-
  %insert_fd_var(FDVARS,FDVariable,Size,LWF,NewFDVARS),
  (FDVARS=[HH|_],HH==FDVariable -> NewFDVARS = FDVARS % variable already in list [cheap check at front]
    ; NewFDVARS = [FDVariable|FDVARS]). %% this adds new variables at front; will be given priority over older ones
  %% add_fd_var(FDVARS,FDVariable,NewFDVARS),  %% this adds new variables at the end and does more thorough duplicate check


% a bit like append: but cleans up list: removes numbers + checks if var already in list
% runtimes for tests: 349 stays at 220ms, test 1088 4660 -> 5040ms
%add_fd_var([],Var,[Var]).
%add_fd_var([V|T],Var,Res) :-
%  (V==Var -> Res = [V|T] %,print(dup(V)),nl,
%   ; nonvar(V) -> add_fd_var(T,Var,Res) %,print(nonvar(V)),nl
%   ; Res = [V|RT], add_fd_var(T,Var,RT)).


:- use_module(library(random),[random_select/3]).
% reverse the list at each step; alternating taking from end and front; makes no sense with randomise_enumeration_order
%my_get_next_fdvar_to_enumerate_rev(FDVARS,NextFDVar,RemainingFDVARS) :- lists:reverse(FDVARS,RevFDVARS),
%   my_get_next_fdvar_to_enumerate(RevFDVARS,NextFDVar,RemainingFDVARS).

my_get_next_fdvar_to_enumerate([],_NextFDVar,_RemainingFDVARS) :- !,fail.
my_get_next_fdvar_to_enumerate([V1|FDVARS],NextFDVar,RemainingFDVARS) :-
   nonvar(V1),
   !, % skip over non-variables (numbers); clpfd_get_next_variable_to_label does not delete variable from list and may crash with leading non-variable terms !
   my_get_next_fdvar_to_enumerate(FDVARS,NextFDVar,RemainingFDVARS).
my_get_next_fdvar_to_enumerate(FDVARS,NextFDVar,RemainingFDVARS) :-
   preferences:preference(randomise_enumeration_order,true),
   !,
   get_min_fd_size_elements(FDVARS,MinFDVars),
   random_select(NextFDVar,MinFDVars,_),
   %print(random_enum(NextFDVar,FDVARS,MinFDVars)),nl,
   RemainingFDVARS=FDVARS. % still contains variable; but no problem as it will be discarded later
my_get_next_fdvar_to_enumerate(FDVARS,NextFDVar,RemainingFDVARS) :-
   %clpfd_interface:clpfd_get_next_variable_to_label_ffc(FDVARS,NextFDVar,RemainingFDVARS).
   clpfd_interface:clpfd_get_next_variable_to_label(FDVARS,NextFDVar),
   % print(enumerating(NextFDVar,FDVARS)),nl, maplist(tools_printing:print_arg,FDVARS),nl,
   RemainingFDVARS=FDVARS.
%   ((FDVARS=[H|T],H==NextFDVar) -> RemainingFDVARS=T ;  RemainingFDVARS=FDVARS).% does not seem to buy any speed

get_min_fd_size_elements([V1|FDVARS],MinFDVars) :- my_get_fd_size(V1,Size),
   get_min_aux(FDVARS,Size,[V1],MinFDVars).

get_min_aux([],_,MinFDVars,MinFDVars).
get_min_aux([V1|FDVARS],Size,MAcc,MinFDVars) :- number(V1),!,
   get_min_aux(FDVARS,Size,MAcc,MinFDVars).
get_min_aux([V1|FDVARS],Size,MAcc,MinFDVars) :-
   my_get_fd_size(V1,Size1),
   (lt_size(Size1,Size)
    -> get_min_aux(FDVARS,Size1,[V1],MinFDVars)
     ; Size1=Size -> get_min_aux(FDVARS,Size,[V1|MAcc],MinFDVars)
     ; get_min_aux(FDVARS,Size,MAcc,MinFDVars)
    ).

lt_size(X,Y) :- number(X),(Y=sup -> true ; X<Y).

% -------------

% using this is relevant for e.g. tests 415, 416, 1096 : to give variables without frozen goals a lower priority
% TO DO: we could filter out those before we call clpfd_get_next_variable_to_label !?
fd_priority_leq_limit(FDVAR,Prio,Limit) :-
  my_get_fd_size(FDVAR,Size),
  number(Size), % not =sup or =inf
  geq_limit(Limit,Size), % the calculation below will only increase the Size
  (Size < 2 -> Prio = Size
   ; geq_limit(Limit,Size*4) -> Prio = Size % avoid calling frozen and fd_degree, Priority only used for warnings and bisect decisions
   ; (enumeration_only_fdvar(FDVAR) -> Prio is Size*4, Prio =< Limit
           ; Prio = Size)
  ).

my_get_fd_size(FDVAR,ResSize) :- fd_size(FDVAR,Size),!, ResSize=Size.
  %(Size=sup -> ResSize = 10000000 /* same as for inf */  ; ResSize=Size).
my_get_fd_size(FDVAR,Size) :- add_internal_error('Could not get fd_size: ',my_get_fd_size(FDVAR,Size)), Size=sup.

gt_limit(inf,_) :- !.
gt_limit(sup,_) :- !.
gt_limit(X,Limit) :- X > Limit.

geq_limit(inf,_) :- !.
geq_limit(sup,_) :- !.
geq_limit(X,Limit) :- X >= Limit.



% -------------


enumeration_only_fdvar(FDVAR) :- var(FDVAR),fd_degree(FDVAR,D),!,
  D=0,
  frozen(FDVAR,Goal), %print(frozen(FDVAR,D,Goal)),nl,
  enum_only_frozen_goal(Goal,FDVAR).
enumeration_only_fdvar(_).

% TO DO: maybe use enumeration_only_goal in b_enumerate
enum_only_frozen_goal((A,B),Var) :- !, enum_only_frozen_goal(A,Var), enum_only_frozen_goal(B,Var).
enum_only_frozen_goal(clpfd:_,_) :- !.
enum_only_frozen_goal(kernel_objects:G,V) :- !, enum_only_frozen_goal_k_obj(G,V).
enum_only_frozen_goal(true,_).

enum_only_frozen_goal_k_obj(safe_less_than_equal(_,X,Y),_) :- (number(X);number(Y)), !.
enum_only_frozen_goal_k_obj(call_enumerate_int(X,_,_,_),Var) :- Var==X,!.
enum_only_frozen_goal_k_obj(enumerate_int_wf(X,_,_),Var) :- Var==X,!.
enum_only_frozen_goal_k_obj(true,_).


% get a wait-flag that will be triggered first next time;
% can be used to ensure that all pending co-routines complete
get_idle_wait_flag(Info,wfx(WF0,Store,EnumFinished,_),LWF) :- !,
    (var(WF0) -> LWF=WF0
     ; nonvar(EnumFinished) -> LWF=1  % waitflag store already grounded to completion; nobody may drive it anymore
     ; get_waitflag_from_store(0.9,Info,Store,LWF)).
get_idle_wait_flag(_Info,no_wf_available,1).


get_wait_flag(Prio,Store,WF) :- get_wait_flag(Prio,'??',Store,WF).
get_wait_flag(0,_Info,wfx(WF0,_Store,_EnumFinished,_),WF) :- !, WF=WF0.
get_wait_flag(1.0,Info,wfx(WF0,Store,_EnumFinished,_),WF) :-
   % WF for deterministic computations, not guaranteed to generate efficient representations
   % (those are triggered in WF0 phase)
   !,
   %used to b: (var(WF0) -> get_waitflag_from_store(1.0,Info,Store,WF) ; WF=1).
   (ground(WF0) % if WF0 is ground the WF0 phase is completed fully;
    -> WF=1   % we can return this waitflag immediately;
              % good idea also if we are already in the later stages of propagation
              % so as not to delay the attached deterministic propagations (see test 383, SudokuHexAsConstant.mch)
    ; get_waitflag_from_store(1.0,Info,Store,WF)).
get_wait_flag(Prio,Info,wfx(_,Store,EnumFinished,_),WF) :- !,
   (ground(EnumFinished) -> WF=Prio /* enumeration has finished: return a ground WF */
    ; get_waitflag_from_store(Prio,Info,Store,WF)).
get_wait_flag(Prio,Info,Store,WF) :- check_is_wait_flag(Store),
   (Store=no_wf_available -> WF=Prio ;
    Store=none -> print(get_wait_flag_deprecated(Prio,Info,Store,WF)),nl, WF=Prio ;
    add_internal_error('Illegal call: ',get_wait_flag(Prio,Info,Store,WF))).


get_waitflag_from_store(Prio,Info,Store,WF) :-
   %get_debug_info(Info0,Info),
   % print(get_waitflag_from_store(Prio,Info,Store,WF)),nl,
  (Prio=inf -> integer_priority(X), RealPrio = float(X)
       % , print(infinite_priority(Prio,Info)),nl %%
     % ; (Prio>10, Prio<10000) -> RealPrio = float(100) % perform left-to-right enumeration in that range
      ; RealPrio=float(Prio)),
	my_get_wf_store_att(Store,Heap),
	(avl_fetch(RealPrio,Heap,wf(WFs,_OldInfo))
	 -> (is_finite_priority(RealPrio) % TO DO: investigate performance if reduce the threshold
	          % here we re-use the waitflag instead of storing a new one
	      -> (var(WFs) -> WF1 = WFs
	          ; WFs = [WF1|_] -> true % because add_wf can construct lists
	          ; print(wf_already_grounded(Prio,Info,WFs)),nl,
	            WFs = WF1
	         )
	      ;  (var(WFs) -> WFs2 = [WF1,WFs] ; WFs2 = [WF1|WFs]),
	         %print(multiple_wf(Prio,Info,_OldInfo,WFs)),nl,  % poses problems for 1003, possibly 1194; possibly better to keep to treat WFs as a FIFO stack to have localised enumeration?
	         %(var(WFs) -> WFs2 = [WFs,WF1]
	         % ; append(WFs,[WF1],WFs2) -> true
	         % ; add_internal_error('Illegal WF list:',wf(WFs,RealPrio,_OldInfo)),
	         %   WFs2 = [WF1|WFs]),
	         % else: store waitflags as list; important for test 1194
	         %print(not_reusing(WFs2)),nl,
	         avl_store(RealPrio,Heap,wf(WFs2,Info),NewHeap),
	         put_mutable_attr(Store,wf_store(NewHeap))
	    )
	  ; avl_store(RealPrio,Heap,wf(WF1,Info),NewHeap),
	    put_mutable_attr(Store,wf_store(NewHeap)) % put_atts seems to be expensive
	),!,
	WF=WF1.
get_waitflag_from_store(Prio,Info,Store,WF) :-
    add_internal_error('Error getting Waitflag: ', get_waitflag_from_store(Prio,Info,Store,WF)),fail.

% a version which delays getting a waitflag until the Prio is known:
:- block block_get_wait_flag(-,?,?,?).
block_get_wait_flag(Prio,Info,WFX,WF) :- get_wait_flag(Prio,Info,WFX,WF).

% check if there are no waitflags and no FD variables; note: there could be pending co-routines on WF0
no_pending_waitflags(no_wf_available).
no_pending_waitflags(wfx(_WF0,Store,_WFE,_)) :-
   my_get_wf_store_fdvars_atts(Store,Heap,FDList),
   FDList=[], %  no FD Variables
   empty_avl(Heap). % no waitflags pending
% Note grounding could still be in progress
% TO DO: we could add attribute before grounding and remove after grounding !?


% -------------------

% storing Predicates for Kodkod external function:
my_get_kodkod_predicates(wfx(_,Store,_,_),ID,PredicateList) :- my_get_wf_preds(Store,ID,kodod,PredicateList).
my_get_wf_preds(Store,ID,Type,PredicateList) :-
   (my_get_wf_preds_att(Store,ID,Type,PredicateList) -> true
	  ;  var(Store),PredicateList = []
	).
%my_store_wf_kodkod(Store,ID,PredicateList) :-  put_atts(Store,+wf_kodkod(ID,PredicateList)).
my_add_predicate_to_kodkod(wfx(_,Store,_,_),ID,Predicate) :-
   my_add_predicate_aux(wfx(_,Store,_,_),ID,kodod,Predicate).
my_add_predicate_aux(wfx(_,Store,_,_),ID,Type,Predicate) :-
   my_get_wf_preds(Store,ID,Type,PredicateList),
   put_mutable_attr(Store,wf_preds(ID,Type,[Predicate|PredicateList])).

% -------------------

% storing Predicates for external function calling satsolver:
my_get_satsolver_predicates(wfx(_,Store,_,_),ID,PredicateList) :- my_get_wf_preds(Store,ID,satsolver,PredicateList).
my_add_predicate_to_satsolver(wfx(_,Store,_,_),ID,Predicate) :-
   my_add_predicate_aux(wfx(_,Store,_,_),ID,satsolver,Predicate).

% -------------------


ground_wait_flag0(wfx(WF0,_,_,_)) :- grd_wf0(WF0).

ground_wait_flags(wfx(WF0,Store,EnumFinish,WFInfos)) :-  !,
    grd_wf0(WF0),
    deref_store(Store,DStore),
    init_wf_infos_for_grounding(WFInfos,WFInfos2),
    ground_waitflags_store_clpfd(DStore,WFInfos2),
    grd_ef(EnumFinish).
ground_wait_flags(no_wf_available) :-  !.
ground_wait_flags(W) :-
    add_internal_error('Waitflags in wrong format: ',ground_wait_flags(W)).

grd_wf0(E) :- var(E) -> E=s(E2), grd_wf01(E2) %%,print(grd_wf0_done),nl
                     ; (E=s(0) -> true ; add_internal_error('Illegal WF0 Waitflag value: ',grd_wf0(E))).
grd_ef(E) :- var(E) -> E=0
             ; ((E==0;E==wd_guarded) -> true ; add_internal_error('Illegal EF Waitflag value: ',grd_ef(E))).
grd_wf01(0).

/* currently unused:
create_wdguarded_wait_flags(wfx(WF0,Store,EnumFinish,Info),Res,Expected,
                            wfx(WF0,Store,LocalEnumFinish,Info)) :-
  %% if Res=Expected then the computations of the inner waitflags should be well-defined;
  %%   otherwise they may not
    copy_wfe(EnumFinish,Res,Expected,LocalEnumFinish).
:- block copy_wfe(-,?,?,?), copy_wfe(?,-,?,?).
copy_wfe(EnumFinish,Res,Expected,LocalEnumFinish) :-
  (Res==Expected -> LocalEnumFinish=EnumFinish
   ; LocalEnumFinish=wd_guarded %% branch is pruned by well-definedness condition
   ).
*/

% ground_constraintprop_wait_flags will not ground WFE flag

:- load_files(library(system), [when(compile_time), imports([environ/2])]).
:- if(environ(prob_safe_mode,true)).
ground_constraintprop_wait_flags(W) :- check_is_wait_flag(W),fail.
:- endif.
ground_constraintprop_wait_flags(wfx(WF0,Store,_WFE,WFInfos)) :-
  %% print(ground_constraintprop_wait_flags(WF0,Store,_WFE)),nl, %%
  !,grd_wf0(WF0),% print(finished_grounding_wf0),nl,portray_waitflags_store(Store),
  deref_store(Store,DStore),
  init_wf_infos_for_grounding(WFInfos,WFInfos2),
  ground_waitflags_store_clpfd(DStore,WFInfos2).
ground_constraintprop_wait_flags(no_wf_available) :-  !.
ground_constraintprop_wait_flags(W) :-
   add_internal_error('Waitflags in wrong format',ground_constraintprop_wait_flags(W)).



:- use_module(clpfd_interface,[clpfd_in_domain/1, clpfd_labeling/2]).
:- use_module(library(lists),[reverse/2]).

%ground_wf(Prio,WF) :- nl,print(ground(WF,Prio)),nl,WF=Prio, print(done_ground(WF,Prio)),nl.
% ground_wf(Prio,WF) :- Prio=WF.
ground_wf(Prio,WF) :-
    (var(WF) -> WF=Prio ; if(WF=Prio,true,add_internal_error('Illegal waitflag: ',ground_wf(Prio,WF)))).

:- if(environ(debug_waitflags,true)).
% provide feedback about waitflag grounding
init_wf_infos_for_grounding(Infos,NewInfos) :-
   %print(start_wf_grounding(Infos)),nl,
   ord_add_element(Infos,wf_indent(_),NewInfos).

% provided indentation to see scope of grounding
indent_wf(WFInfos) :- member(wf_indent(Var),WFInfos),!,indent_var(Var).
indent_wf(_).
% indent based upon a shared variable and increment it
indent_var(X) :- var(X),!, X=s(_).
indent_var(s(X)) :- !,print('.'), indent_var(X).
indent_var(_) :- print('***').

ground_prio(Prio,WFMin,Info,WFInfos) :- WFInfos \= '$debug_done',
    (member(IM,WFInfos),get_info_context_description(IM,Msg) -> true ; Msg=''),
    (Prio=float(P) -> true ; P=Prio),
    indent_wf(WFInfos),
    format(' -> ~w --> ~w :: ~w~n',[P,Info,Msg]),
    !,
    ground_prio(Prio,WFMin,Info,'$debug_done'). % recursive, but this time normal clauses will apply
:- else.
init_wf_infos_for_grounding(I,I).
:- endif.
ground_prio(Prio,WF,Info,_WFInfos) :- %print(ground_prio(Prio,WF,Info)),nl,
   nonvar(WF),!,
   (WF=[H|T] ->
    % we have a list of waitflags; ground one-by one; TO DO: only ground one and store remaining list in wfx Heap
          reverse([H|T],Ls),  %print(list(Ls)),nl,
          maplist(ground_wf(Prio),Ls)
        ; ground(WF) -> true % already grounded by somebody else
        ; add_internal_error('Waitflag already partially grounded: ',ground_prio(Prio,WF,Info))).
%ground_prio(Prio,WF,Info,WFInfos) :- get_debug_kernel_waitflags(Cntr),!,
%    print(ground_prio(Cntr,Prio,WF,Info)),nl, WF=Prio, print(grounded_prio(Cntr,Prio,WF,Info)),nl.
ground_prio(Prio,WF,_Info,_WFInfos) :-
   %external_functions:indent_format(user_output,'~w (~w)~n',[Prio,Info],no_trace),portray_attached_goals(WF),nl,
   %% (debug:debug_mode(on) -> external_functions:indent_format(user_output,'~w (~w)~n',[Prio,Info],no_trace) ; true),
		        % print(ground(Prio,WF,Info)),nl, translate:print_frozen_info(WF),nl,
				WF = Prio.                              % ground waitflag
	            % print(grounded(Prio,Info)),nl,nl, %%

%:- use_module(kernel_objects,[enum_warning_large/3]).
label_clpfd_variable(Variable) :- label_clpfd_variable(2097153,Variable).
label_clpfd_variable(Prio,Variable) :-
    %print(label_clpfd_variable(Prio,Variable)),nl,
    (gt_limit(Prio,2097152),  % note Prio > test also works with float(X) terms !
      check_if_labeling_domain_large(Variable,2097152,Size,RANGE)  % {x|x:1..2**23 & x mod 2 = x mod 1001} takes about 2 minutes; {x|x:1..2**21 & x mod 2 = x mod 1001} takes about 30 seconds; 2**21 = 2097152, 2**23 = 8388608
		-> %kernel_objects:enum_warning_large(Variable,'INTEGER VARIABLE',RANGE),  % could in principle also be enumerated set; but extremely unlikely given size; enum_warning will generate virtual time-outs, ... -> we no longer do this; for AMASS scheduler.mch there is a time-out after step 301 otherwise
		   performance_messages:perfmessage(enumerating_large_integer_variable(RANGE)),
		   (number(Size)
		    -> clpfd_labeling([Variable],[bisect]) % ffc no longer required: just a single variable
		    ;  add_warning(label_clpfd_variable,'Trying to enumerate FD Variable with infinite domain (possibly due to WD error) ',RANGE)
		   )
		    % we use bisecting rather than default enumeration; e.g., x:1..8388610 & y:1..8388610 & x*x*x = y*y + y + 10 is relatively quickly determined to be unsatisfiable (380 ms)
		    % Note: test 1099 tests this
		    % TO DO: investigate whether it makes sense to use bisect for smaller values already
		; preferences:preference(randomise_enumeration_order,true) -> clpfd_interface:clpfd_randomised_in_domain(Variable)
		; clpfd_in_domain(Variable)). % TO DO: use random ?

check_if_labeling_domain_large(H,Limit,Size,Res) :-
    my_get_fd_size(H,Size),
    gt_limit(Size,Limit),
    clpfd_interface:clpfd_domain(H,From,To),
    Res=From:To.

%check_if_labeling_domain_finite(H) :-
%    my_get_fd_size(H,Size), number(Size).

% mix ProB labeling with CLP(FD) labeling
ground_waitflags_store_clpfd(Store,WFInfos) :-
			my_get_wf_store_fdvars_atts(Store,Heap,FD),!,
			ground_waitflags_store_clpfd_aux(Store,Heap,FD,WFInfos).
ground_waitflags_store_clpfd(Store,WFInfos) :-
      add_internal_error('Illegal Waitflag: ',ground_waitflags_store_clpfd(Store,WFInfos)),fail.

ground_waitflags_store_clpfd_aux(Store,Heap,FDVars,WFInfos) :- FDVars \= [],
			%my_get_fdvars_att(Store,FDVars),
			%get_atts(Store,wf_fdvars(FDVars)), % there is a list of FD vars available
			!,
			( avl_del_min(Heap,Prio,wf(WFMin,Info),NewHeap) ->     % we have non-ground waitflags with a matching priority
			 (my_get_next_fdvar_to_enumerate(FDVars,NextFDVar,RemainingFDVARS),
			   fd_priority_leq_limit(NextFDVar,FDPrio,Prio)
			   -> enum_fd_variable_store(FDPrio,Store,NextFDVar,RemainingFDVARS,WFInfos)
			   ;  put_mutable_attr(Store,wf_store(NewHeap)),   % write the modified heap back
			      % debug:hit_counter(XXX),print(grounding(XXX,Prio,_Info)),nl, (XXX=16556->trace ; true),
				    ground_prio(Prio,WFMin,Info,WFInfos),
			      ground_waitflags_store_clpfd(Store,WFInfos)    % continue recursively
	     )
			;
	        (my_get_next_fdvar_to_enumerate(FDVars,NextFDVar,RemainingFDVARS)
			      -> enum_fd_variable_store(Store,NextFDVar,RemainingFDVARS,WFInfos)
			       ; true  % heap and FDVars list is empty, nothing more to do
			    )
			).
ground_waitflags_store_clpfd_aux(Store,Heap,[],WFInfos) :- % no FDVars are available to enumerate
        avl_del_min(Heap,Prio,wf(WFMin,Info),NewHeap),
        !,
			  put_mutable_attr(Store,wf_store(NewHeap)),   % write the modified heap back
				% debug:hit_counter(XXX),print(grounding(XXX,Prio,_Info)),nl, (XXX=16556->trace ; true),
			  ground_prio(Prio,WFMin,Info,WFInfos),                % ground waitflag
			  ground_waitflags_store_clpfd(Store,WFInfos).         % continue recursively
ground_waitflags_store_clpfd_aux(_Store,_Heap,[],_).   % heap and FDVars list is empty, nothing more to do

%:- use_module(library(lists),[exclude/3]).
enum_fd_variable_store(Store,NextFDVar,RemainingFDVARS,WFInfos) :-
   %exclude(nonvar,RemainingFDVARS,RemainingFDVARS2),
   put_mutable_attr(Store,wf_fdvars(RemainingFDVARS)),
    %%  print(labeling(NextFDVar,RemainingFDVARS)),nl, portray_fd_vars([NextFDVar|RemainingFDVARS]),nl,nl,
   label_clpfd_variable(NextFDVar),
   ground_waitflags_store_clpfd(Store,WFInfos).
% a version of the above where we know the priority
enum_fd_variable_store(FDPrio,Store,NextFDVar,RemainingFDVARS,WFInfos) :-
   %exclude(nonvar,RemainingFDVARS,RemainingFDVARS2), % does not seem to buy a lot
   put_mutable_attr(Store,wf_fdvars(RemainingFDVARS)),
    %%  print(labeling(NextFDVar,RemainingFDVARS)),nl, portray_fd_vars([NextFDVar|RemainingFDVARS]),nl,nl,
   label_clpfd_variable(FDPrio,NextFDVar),
   ground_waitflags_store_clpfd(Store,WFInfos).

:- if(false).
% small utility to trace through waitflag enumeration
:- dynamic spying_on/1.
spying_on(1).
spy_waitflags(_,_) :- \+ spying_on(_) , !.
spy_waitflags(Msg,Store) :- retract(spying_on(Nr)),print(Msg),nl,
   (Nr=1
    -> assert(spying_on(Nr)),
       print(' (j,t,p,q) ==> '),read_term(T,[]),
       action(T,Store)
    ;  N1 is Nr-1, assert(spying_on(N1))
   ).
action(t,_Store) :- !,trace.
action(p,Store) :- !,portray_waitflags_store(Store), spy_waitflags('',Store).
action(q,_Store) :- !,retractall(spying_on).
action(Nr,_) :- number(Nr),!, retractall(spying_on(_)), assert(spying_on(Nr)).
action(_,_).
:- endif.

:- assert_must_succeed(( kernel_waitflags:init_wait_flags(WF), kernel_waitflags:portray_waitflags(WF),
      kernel_waitflags:get_wait_flag(2,'Test',WF,WF2), when(nonvar(WF2),print(wf2)),
      clpfd:in(FD,'..'(1,2)),kernel_waitflags:add_fd_variable_for_labeling(FD,WF),
      kernel_waitflags:portray_waitflags(WF), kernel_waitflags:ground_wait_flags(WF)  )).
% PORTRAYING Waitflags
portray_waitflags(wfx(WF0,Store,WFE,Infos)) :- flush_output,
            print('*WAITFLAG STORE: '),print(wfx(WF0,'STORE',WFE,Infos)),nl,
            (var(WF0) -> print('> WF0 :: '), portray_attached_goals(WF0),nl ; true),
            portray_waitflags_store(Store),!,
            (var(WFE) -> print('> WFE :: '), portray_attached_goals(WFE),nl ; true),
            print('*END WAITFLAG STORE'),nl.
			%my_portray_avl(Heap),nl.
portray_waitflags(none) :- !, print('*WAITFLAG STORE:  none'),nl.
portray_waitflags(no_wf_available) :- !, print('*WAITFLAG STORE:  none'),nl.
portray_waitflags(Store) :-
   add_internal_error('Illegal Waitflag: ',portray_waitflags(Store)),flush_output.

portray_fd_vars([]) :- !, '$NOT_COVERED'('This is debugging code!').
portray_fd_vars([V|T]) :- print('> '), print(V), nonvar(V),!, nl,
   portray_fd_vars(T), '$NOT_COVERED'('This is debugging code!').
portray_fd_vars([V|T]) :- my_get_fd_size(V,Size), !, print(' : '), print(Size),
   portray_attached_goals(V),
   portray_fd_vars(T),
   '$NOT_COVERED'('This is debugging code!').
portray_fd_vars(X) :- print('> *** UNKNOWN FDVARS: '), print(X), nl,
   '$NOT_COVERED'('This is debugging code!').

portray_waitflags_store(Store) :-
            my_get_fdvars_att(Store,FDList),
            (FDList=[] -> true ; print('FDVARS:'),nl,portray_fd_vars(FDList),nl),
			my_get_wf_store_att(Store,Heap),
			avl_to_list(Heap,List), portray_avl_els(List).

portray_avl_els([]).
portray_avl_els([Prio-wf(LWF,Info)|T]) :-
  print('> '),print(Prio), print(' : '), print(LWF), print(' : '), portray_info(Info),nl,
  portray_attached_goals(LWF),
  portray_avl_els(T).

portray_info(label_clpfd_vars(Vars)) :- print('label_clpfd_vars :: '),
        translate:l_print_frozen_info(Vars).
portray_info(I) :- print(I).

portray_attached_goals(V) :- V==[],!.
portray_attached_goals(LWF) :- nonvar(LWF),LWF=[H|T],!,
   portray_attached_goals(H),portray_attached_goals(T).
portray_attached_goals(LWF) :- frozen(LWF,Goal), print('   :: '), portray_goal(Goal),nl.

:- use_module(tools_printing,[print_term_summary/1]).
portray_goal((A,B)) :- !, portray_goal(A), print(','), print('     '), portray_goal(B).
portray_goal(A) :- print_term_summary(A).

:- public my_portray_avl/1.
my_portray_avl(V) :- var(V), !, add_internal_error('Variable: ',my_portray_avl(V)).
my_portray_avl(V) :- portray_avl(V).


quick_portray_waitflags(wfx(WF0,Store,WFE,_)) :-
      my_get_wf_store_fdvars_atts(Store,Heap,FDList),
			avl_domain(Heap,List),
			format('% Waitflags-Store WF0=~w,WFE=~w, Flags=~w, FDVars=~w~n',[WF0,WFE,List,FDList]).
quick_portray_waitflags_store(Store) :-
      my_get_wf_store_fdvars_atts(Store,Heap,FDList),
			avl_domain(Heap,List),
			format('% Waitflags-Store Flages=~w, FDVars=~w~n',[List,FDList]).

:- assert_must_succeed(( kernel_waitflags:init_wait_flags(WF),
 kernel_waitflags:get_wait_flag1(WF,WF1), when(ground(WF1),X=a),
 kernel_waitflags:ground_wait_flags(WF),  X==a )).

% create a copy of the waitflags with WF0 and WFE as variable; should be finished with copy_wf_finish
% can be used if you temporarily want to disable non-deterministic enumeration, ...
%copy_wf_start(WF,CWF) :- copy_wf_start(WF,[],CWF).
% now a version with an info field:
copy_wf_start(wfx(_,WFStore,_,Infos),Info,wfx(_,WFStore,_,NewInfos)) :-
   add_debug_wait_flag_info(Infos,wf_copy(Info),NewInfos).
copy_wf_start(no_wf_available,_,no_wf_available).

% just copy WF0 and WFE over to new waitflag
copy_wf_finish(wfx(WF0,_,WFE,_),wfx(WF0,_,WFE,_)).
copy_wf_finish(no_wf_available,no_wf_available).


/* create a copy where the enumeration_finished waitflag is shared;
  the inner enumeration_finished wait flags should normally not be grounded */
create_inner_wait_flags(wfx(_,_,WFE,Infos),Info,wfx(_WF0Inner,S,WFEInner,NewInfos)) :-
   add_debug_wait_flag_info(Infos,Info,NewInfos),
   init_wait_flags_store(S), %print(copy_wfe_to_inner(_Info,WFE,WFEInner)),nl,
   copy_wfe_to_inner(WFE,WFEInner).
create_inner_wait_flags(no_wf_available,_,no_wf_available).

% was using: get_inner_enumeration_over_wait_flag/get_enumeration_almost_finished_wait_flag
%:- block copy_wfe_to_inner(-,?,?).
%copy_wfe_to_inner(_,WFE,WFE).

:- block copy_wfe_to_inner(-,-).
copy_wfe_to_inner(WFE,_) :- var(WFE),!. % inner was grounded before outer
copy_wfe_to_inner(WFE,WFE).
% copy enumeration finished waitflag from inner computation if WFE was not grounded there
% assumes inner WF enumeration is finished
copy_wfe_from_inner_if_necessary(wfx(_,_,WFEInner,_),wfx(_,_,WFE,_)) :- var(WFEInner),!, WFE=WFEInner.
copy_wfe_from_inner_if_necessary(_,_).

% is very similar to ground_constraintprop_wait_flags, but will also ground EF if no abort is pending
ground_inner_wait_flags(WF) :-
   ground_constraintprop_wait_flags(WF),
   ground_ef_wait_flag_unless_abort_pending(WF).

% copy WF0 and WFE, set up a new store for the rest
copy_wf01e_wait_flags(wfx(WF0,WFX,WFE,Infos),wfx(WF0,S,WFE,Infos)) :-
   init_wait_flags_store(S),
   get_waitflag_from_store(1,copy_wf01e,WFX,WF1), % now ground wait flag 1 if it is grounded in the orginal WF store
   copy_wf01e_wait_flags_aux(WF1,S).
copy_wf01e_wait_flags(no_wf_available,no_wf_available).

:- block copy_wf01e_wait_flags_aux(-,?).
copy_wf01e_wait_flags_aux(_,S) :-
    ground_waitflags_store_up_to_no_clpfd(1,S,[]). % no CLPFD - labeling required until prio 1


clone_wait_flags_from1(wfx(_WF0,S,WFE,Infos),wfx(_NewWF0,S2,WFE,Infos)) :- deref_store(S,S2).
clone_wait_flags_from1(no_wf_available,no_wf_available).
% creates a copy of the wait_flags, except for WF0; this needs to be grounded separately with grd_wf0/1
clone_wait_flags_from1_finish(wfx(WF0,_,_,_),wfx(WF0,_,_,_)). % finish copying by copying over WF0
clone_wait_flags_from1_finish(no_wf_available,no_wf_available).

get_wait_flag0(wfx(WF0,_,_,_),WF0).
get_wait_flag0(no_wf_available,0).

%get_wait_flag0(WFX,WF0) :-   get_wait_flag(0,WFX,WF0).  /* when this is ground you can do det. propagatons */
get_wait_flag1(WFX,WF1) :-   get_wait_flag(1.0,WFX,WF1). /* when this is ground you can do det. propagatons */
get_wait_flag1(Info,WFX,WF1) :-   %print(getwf1(Info)),nl, portray_waitflags(WFX),nl,
   get_wait_flag(1.0,Info,WFX,WF1).

get_binary_choice_wait_flag(Info,WFX,WF2) :- !, get_binary_choice_wait_flag_exp_backoff(4,Info,WFX,WF2).
%get_binary_choice_wait_flag(Info,WFX,WF2) :- % use 3 rather than 2 to give priority over enumeration choicepoints
%  get_wait_flag(4,Info,WFX,WF2). /* when this is ground you can do binary propagatons */

%get_middle_wait_flag(Info,WFX,WFn) :- get_wait_flag(10,Info,WFX,WFn).

get_last_wait_flag(Info,WFX,Res) :- last_finite_priority(P),get_wait_flag(P,Info,WFX,Res).

% ------------------------
get_wait_flag_infos(wfx(_,_,_,Infos),Infos).
get_wait_flag_infos(no_wf_available,[]).

:- use_module(library(ordsets),[ord_member/2,ord_add_element/3]).
get_wait_flag_info(wfx(_,_,_,Infos),Info) :- member(Info,Infos).

% use if the Info field is known, requires ordered/sorted Infos list
% used for wfx_no_enumeration information, see test 1368
is_wait_flag_info(wfx(_,_,_,Infos),Info) :-
  (nonvar(Infos) -> ord_member(Info,Infos) ; print(var_waitflag),nl,fail).

%set_wait_flag_infos(wfx(WF0,Store,WFE,_),Infos,wfx(WF0,Store,WFE,Infos)).
%set_wait_flag_infos(no_wf_available,_,no_wf_available).

add_wait_flag_info(wfx(WF0,Store,WFE,Infos),Info,wfx(WF0,Store,WFE,NewInfos)) :-
   (ord_add_element(Infos,Info,NewInfos) -> true %, print(added_wf_infos(NewInfos)),nl
      ; add_internal_error('Could not add_wait_flag_info:',(Info,Infos)),
        NewInfos = [Info|Infos]).
add_wait_flag_info(no_wf_available,_,no_wf_available).

:- if((environ(prob_safe_mode,true) ; environ(prob_src_profile,true) ; environ(prob_profile,true))).
add_debug_wait_flag_info(Infos,Info,NewInfos) :- ord_add_element(Infos,Info,NewInfos).
init_debug_wait_flag_info([],Infos) :- !, Infos=[].
init_debug_wait_flag_info([H|T],Infos) :- !, Infos=[H|T].
init_debug_wait_flag_info(Info,[Info]).
:- else.
add_debug_wait_flag_info(Infos,_,Infos).
init_debug_wait_flag_info(_,[]).
:- endif.
% ------------------------

%  BINARY CHOICE WAITFLAGS

:- use_module(tools,[max_tagged_pow2/1]).
% last power of 2 exponent that is still a finite priority
last_finite_pow2_prio_exponent(Lim) :- max_tagged_pow2(Pow),
   Lim is Pow-1. % stay below last_finite_priority, i.e., max_tagged_integer-1024

% utility to get power of two priority:
get_pow2_binary_choice_priority(Exp,Prio) :- last_finite_pow2_prio_exponent(Lim),
  (Exp =< Lim -> Prio is integer(2**Exp)
   ; Prio is integer(2**Lim)).

% get a binary choice wait flag, start with 2 but for every further call double the priority (exponential backoff)
% idea: give priority to data enumerations when too many choice points arise
get_binary_choice_wait_flag_exp_backoff(Info,WFX,WF2) :-
    get_binary_choice_wait_flag_exp_backoff(2,Info,WFX,WF2).
get_binary_choice_wait_flag_exp_backoff(_,_Info,no_wf_available,WF2) :- !, WF2=2.

% StartPrio should normally be a power of 2:
get_binary_choice_wait_flag_exp_backoff(MinPrio,Info,wfx(_,Store,_,_),WF2) :-
	my_get_wf_store_att(Store,Heap),
  large_finite_priority(Lim), % avoid creating too big numbers and stay in finite area
	(MinPrio < 1 -> StartPrio=1 ; StartPrio=MinPrio),
	get_bin_aux(StartPrio,Lim,Info,Heap,NewHeap,WF2),
	put_mutable_attr(Store,wf_store(NewHeap)).

% to do : maybe store attribute of current exponential ??
get_bin_aux(Prio,Lim,Info,Heap,NewHeap,WF2) :-
    avl_fetch(float(Prio),Heap,wf(WFs,OldInfo)),
    !,
    (Prio >= Lim -> WF2 = WFs, NewHeap=Heap  /* for large finite priorities single waitflag stored */
     ; OldInfo = '$binary'(_) -> P2 is Prio*2, get_bin_aux(P2,Lim,Info,Heap,NewHeap,WF2)
     ; otherwise -> WF2=WFs, avl_store(float(Prio),Heap,wf(WFs,'$binary'(Info)),NewHeap) % just update Info; so that next time we double
     ).
get_bin_aux(Prio,_,Info,Heap,NewHeap,WF2) :- % we have found an unused power of 2
   %print(get_bin(Prio,Info)),nl,
   avl_store(float(Prio),Heap,wf(WF2,'$binary'(Info)),NewHeap).

% ------------------------

get_enumeration_finished_wait_flag(wfx(_,_,WFE,_),Res) :- !,Res=WFE.
get_enumeration_finished_wait_flag(no_wf_available,WFE) :-  !,WFE=1.
get_enumeration_finished_wait_flag(W,E) :-
   add_internal_error('Waitflags in wrong format: ',get_enumeration_finished_wait_flag(W,E)).


ground_det_wait_flag(wfx(WF0,Store,_,WFInfos)) :- !, %% print(ground_det_wait_flag),nl,
     grd_wf0(WF0),
     deref_store(Store,DStore),
     init_wf_infos_for_grounding(WFInfos,WFInfos2),
     ground_waitflags_store_up_to_no_clpfd(1,DStore,WFInfos2). % no CLPFD labeling required for prio until 1
ground_det_wait_flag(no_wf_available) :-  !.
ground_det_wait_flag(W) :-
    add_internal_error('Waitflags in wrong format: ',ground_det_wait_flag(W)).

ground_wait_flag_to(wfx(WF0,Store,_,WFInfos),Limit) :- !,%% print(ground_det_wait_flag),nl,
     grd_wf0(WF0),
     deref_store(Store,DStore),
     init_wf_infos_for_grounding(WFInfos,WFInfos2),
     ground_waitflags_store_clpfd_up_to(Limit,DStore,WFInfos2).
ground_wait_flag_to(no_wf_available,_) :-  !.
ground_wait_flag_to(W,Limit) :-
    add_internal_error('Waitflags in wrong format: ',ground_wait_flag_to(W,Limit)).

ground_waitflags_store_up_to_no_clpfd(Prio,Store,WFInfos) :- % print(prio(Prio)),nl,
	my_get_wf_store_att(Store,Heap),  % portray_waitflags_store(Store), %
	( avl_min(Heap,Min,wf(WFMin,Info)) -> % heap is non-empty, first element is WF with priority Min
	    ( Min =< Prio ->          % we have non-ground waitflags with a matching priority
        (avl_del_min(Heap,AMin,_,NewHeap),AMin==Min     % remove the first element
         -> true
          ; add_internal_error('Could not remove min. in ground_waitflags_store_up_to_no_clpfd: ',Min:Info:Heap), fail
        ),
        put_mutable_attr(Store,wf_store(NewHeap)),   % write the modified heap back
            %%  print(grounding2(Prio,_Info)),nl, %%
        ground_prio(Min,WFMin,Info,WFInfos),
        ground_waitflags_store_up_to_no_clpfd(Prio,Store,WFInfos)         % continue recursively
	    ; /* otherwise -> */            % no more matching waitflags present
		true)                 % stop here
	; /* otherwise -> */               % heap is empty, nothing more to do
	    true).                    % stop here

ground_waitflags_store_clpfd_up_to(Limit,Store,WFInfos) :-
			my_get_wf_store_fdvars_atts(Store,Heap,FDVars),
			!,
			( avl_del_min(Heap,Prio,wf(WFMin,Info),NewHeap) ->     % we have non-ground waitflags with a matching priority
			  ( my_get_next_fdvar_to_enumerate(FDVars,NextFDVar,RemainingFDVARS),
			    fd_priority_leq_limit(NextFDVar,FDPrio,Limit)
			   -> (FDPrio > Limit -> true
			         ; enum_fd_variable_store(FDPrio,Store,NextFDVar,RemainingFDVARS,WFInfos))
			   ;  (Prio > Limit -> true
			       ; put_mutable_attr(Store,wf_store(NewHeap)),   % write the modified heap back
			         %print(grounding(Prio,_Info)),nl,
				       ground_prio(Prio,WFMin,Info,WFInfos),                   % ground waitflag
			         ground_waitflags_store_clpfd_up_to(Limit,Store,WFInfos)         % continue recursively
			      )
	      )
			; (my_get_next_fdvar_to_enumerate(FDVars,NextFDVar,RemainingFDVARS),
	       fd_priority_leq_limit(NextFDVar,FDPrio,Limit)
			      -> enum_fd_variable_store(FDPrio,Store,NextFDVar,RemainingFDVARS,WFInfos)
			       ; true  % heap and FDVars list is empty, nothing more to do
			   )
			).
ground_waitflags_store_clpfd_up_to(Limit,Store,WFInfos) :-
     add_internal_error('Illegal Waitflag: ',ground_waitflags_store_clpfd_up_to(Limit,Store,WFInfos)),fail.


update_waitflag(Prio,CurrentWaitflag,NewWaitFlag,WF) :-
  /* if the CurrentWaitflag is already ground and now a new WF with lower priority has been added to the store:
      create a new non-ground waitflag to give priority to new WF with lower priority value */
   (var(CurrentWaitflag) -> NewWaitFlag=CurrentWaitflag
     ; WF=wfx(_WF0,Store,_,_),my_get_wf_store_att(Store,Heap),
       (avl_min(Heap,Min,wf(_NewerWF,_Info)),
        (Prio=inf;Min<Prio)
         -> get_waitflag_from_store(Prio,updated_wf,Store,NewWaitFlag)  % could be optimized
           % ,print(updating_wf(Prio,CurrentWaitflag,NewWaitFlag,Min)),nl
         ;  fdvar_with_higher_prio_exists(Store,Prio) -> get_waitflag_from_store(Prio,updated_wf,Store,NewWaitFlag)
         ;  NewWaitFlag=CurrentWaitflag
       )
   ).

fdvar_with_higher_prio_exists(Store,Limit) :-
   my_get_fdvars_att(Store,FDVars),
   my_get_next_fdvar_to_enumerate(FDVars,NextFDVar,_),
   fd_priority_leq_limit(NextFDVar,_,Limit).

% get the minimum waitflag, if it exists
get_minimum_waitflag_prio(wfx(_,Store,_,_),MinPrio,Info) :-
    my_get_wf_store_att(Store,Heap),
    avl_min(Heap,MinPrio,wf(_LWF,Info)).


add_wd_error(Msg,Term,WF) :- add_wd_error_span(Msg,Term,unknown,WF).
add_wd_error_set_result(Msg,Term,Result,ResultValue,Span,WF) :-
   is_wd_guarded_result(Result),  % just dummy co-routine to detect variables which have WD assignments pending on them; should not be enumerated
   % add well-definedness error but also set Result Variable to ResultValue to trigger pending co-routines if wd_guarded
   add_abort_error7(well_definedness_error,Msg,Term,Result,ResultValue,Span,WF).
add_wd_error_span(Msg,Term,Span,WF) :- add_abort_error_span(well_definedness_error,Msg,Term,Span,WF).

:- block is_wd_guarded_result(-).
is_wd_guarded_result(_).

add_abort_error_span(ErrType,Msg,Term,Span,WF) :- add_abort_error7(ErrType,Msg,Term,0,0,Span,WF).


add_abort_error7(_ErrType,_Msg,_Term,_Result,_ResultValue,_Span,_WF) :-
    preference(disprover_mode,true),
    !, % When Disproving we can assume well-definedness of the PO; sometimes the relevant hypotheses will be filtered out
    fail.
add_abort_error7(ErrType,Msg,Term,Result,ResultValue,Span,WF) :-
    preference(raise_abort_immediately,F), F \= false,
    !,
    (F=full -> AWF=1 % really raise immediately; may entail more spurious messages, particularly in WF0
     ; get_idle_wait_flag(add_abort_error7,WF,AWF), %get_wait_flag0(WF,AWF),
       mark_pending_abort_error(WF,Msg,Term,Span)
     ),
    %when(nonvar(AWF),

    %(tools_printing:print_error('Raising error immediately without checking rest of predicate for unsatisfiability (because RAISE_ABORT_IMMEDIATELY set to TRUE): error could be spurious!'),true)),
    string_concatenate('Raising error immediately without checking rest of predicate for unsatisfiability (because RAISE_ABORT_IMMEDIATELY set to TRUE): error could be spurious!\n! ',Msg,NewMsg),
    %translate:explain_span(Span,TS,[]), format('~s~n',[TS]), trace,
    add_abort_error2(AWF,ErrType,NewMsg,Term,Result,ResultValue,Span).
add_abort_error7(ErrType,Msg,Term,Result,ResultValue,Span,WF) :-
    %nl,print(possible_wd_error(WF,Msg)),nl,
    mark_pending_abort_error(WF),
    get_enumeration_finished_wait_flag(WF,AWF), % infinite enumeration will never result in failure ? and thus never discard a WD error ? --> better to activate WD/abort error earlier, but the following does not fully work:
    %integer_priority(P), get_wait_flag(P,add_abort_error7,WF,AWF), % causes problem for test 1122
    add_abort_error2(AWF,ErrType,Msg,Term,Result,ResultValue,Span).


% no longer used:
%add_wd_error_span_now(Msg,Term,Span) :- add_abort_error2(true,well_definedness_error,Msg,Term,0,0,Span).

:- use_module(state_space,
  [store_abort_error_for_context_state_if_possible/4]).
:- block add_abort_error2(-,?,?,?,?,?,?).
add_abort_error2(wd_guarded,_Err,_Msg,_Term,Result,ResultValue,_Span) :- !,% ignoring error message; no-wd problem
  % set Result to default ResultValue; useful to get rid of pending co-routines; result will not be used
  (Result=ResultValue -> true ; print(add_abort_error_failure(Result,ResultValue)),nl).
add_abort_error2(_AWF,ErrType,Msg,Term,_,_,Span) :-  % print(adding_abort_error(Msg)),nl,trace,
  % add_error(abort_error,Msg,Term),
    (store_abort_error_for_context_state_if_possible(ErrType,Msg,Term,Span)
      -> true
       ; add_error(add_abort_error2,'Could not store error: ',(Msg,Term),Span)),
    % print(done),nl,
    fail.

/*
ground_wait_flags_catch_enumeration_warnings(WF) :-
    on_exception(enumeration_warning(A,B,C,D,E),
                 ground_wait_flags(WF),
                 (pending_abort_error(WF,Msg,Term,Span),
                  format('WD-Error may be causing Enumeration Warning : ~w~n',[Msg]),
                  fail
                  ;
                  throw(enumeration_warning(A,B,C,D,E)))).
*/

% ground enumeration finished waitflag unless abort is pending
% useful inside of exists, ... to complete all pending co-routines
ground_ef_wait_flag_unless_abort_pending(wfx(_WF0,_WFX,WFE,_)) :-
   var(WFE), %print(check_abort_pending(_WF0,_WFX,WFE)),nl,
   get_atts(WFE,-wf_abort_pending),
   !, %print(grd_ef(WFE)),nl,
   grd_ef(WFE).
ground_ef_wait_flag_unless_abort_pending(_).

% try and get a user-readable description of the context in which the WF store was created
get_wait_flags_context_msg(wfx(_WF0,_WFX,_WFE,Infos),Msg) :-
   member(Info,Infos),
   get_info_context_description(Info,Msg).

:- use_module(bsyntaxtree, [get_texpr_ids/2]).
get_info_context_description(expansion_context(Type,Parameters),Msg) :-
  (Parameters=[b(identifier(_),_,_)|_]
   -> get_texpr_ids(Parameters,Ids),
      get_msg_aux(Type,Ids,Parameters,Msg)
    ; get_msg_aux(Type,Parameters,[],Msg)
   ).
get_info_context_description(expansion_context_with_pos(Type,Parameters,Info),Msg) :-
   get_msg_aux(Type,Parameters,Info,Msg).
% also deal with : check_element_of_function_closure_nowf(MemoID)

get_msg_aux(Type,Parameters,Info,Msg) :-
    extract_span_description(Info,PosMsg),!, % we could do this only if preference trace_info set ?
    ajoin_with_sep(['expanding', Type, 'at', PosMsg, 'over ids'|Parameters],' ',Msg).
get_msg_aux(Type,[],_Info,Msg) :- !,
     ajoin(['expanding ', Type],Msg).
get_msg_aux(Type,Parameters,_Info,Msg) :-
     ajoin_with_sep(['expanding', Type, 'over ids'|Parameters],' ',Msg).


% check whether an abort error is pending.
pending_abort_error(wfx(_WF0,_WFX,WFE,_)) :- var(WFE),
    get_atts(WFE,+wf_abort_pending). %print(got_abort_pending),nl.
% succeed once for every pending abort error in Waitflag store
pending_abort_error(wfx(_WF0,_WFX,WFE,_),Msg,Term,Span) :- var(WFE),
  frozen(WFE,Goal), %print(goal(Goal)),nl,
  pending_abort_error_aux(Goal,Msg,Term,Span).

pending_abort_error_aux(kernel_waitflags:add_abort_error2(_,_ErrType,Msg,Term,_Result,_ResultValue,Span), Msg,Term,Span).
pending_abort_error_aux(add_abort_error2(_,_ErrType,Msg,Term,_Result,_ResultValue,Span), Msg,Term,Span).
pending_abort_error_aux(kernel_waitflags:mark_pending_abort_error2(_,Msg,Term,Span), Msg,Term,Span).
pending_abort_error_aux((A,B),Msg,Term,Span) :-
   (pending_abort_error_aux(A,Msg,Term,Span) ;
    pending_abort_error_aux(B,Msg,Term,Span) ).

% explicitly mark that an abort error is pending
mark_pending_abort_error(wfx(_WF0,_WFX,WFE,_)) :- % print(mark_abort_pending(_WFX)),nl,
    (var(WFE) -> put_atts(WFE,+wf_abort_pending)
              ;  true).
% explicitly mark that an abort error is pending with storing information
mark_pending_abort_error(wfx(_WF0,_WFX,WFE,_),Msg,Term,Span) :-
    (var(WFE) -> put_atts(WFE,+wf_abort_pending),
                 mark_pending_abort_error2(WFE,Msg,Term,Span)
              ;  true).
mark_pending_abort_error(no_wf_available,_Msg,_Term,_Span).

:- block mark_pending_abort_error2(-,?,?,?).
mark_pending_abort_error2(_,_,_,_).
% ---------------------------------------------


get_large_finite_wait_flag(Info,WFX,WF2) :-
    large_finite_priority(P),
    get_wait_flag(P,Info,WFX,WF2).

get_enumeration_starting_wait_flag(Info,WFX,WF2) :-
    last_finite_priority(P),
    get_wait_flag(P,enumeration_starting(Info),WFX,WF2).
%get_enumeration_almost_finished_wait_flag(Info,WFX,WF2) :-
%    tools:max_tagged_integer(P), % largest tagged value
%    get_wait_flag(P,enumeration_starting(Info),WFX,WF2).


get_integer_enumeration_wait_flag(Info,WFX,WF2) :-
  integer_priority(Prio),
  get_wait_flag(Prio,Info,WFX,WF2).

% A few hardcoded priorities
large_finite_priority(P) :- integer_priority(I), P is I-1000.
last_finite_priority(P) :- integer_priority(I), P is I-2. % Note: enumerate_integer_with_min_range_wf did set its priority to I-1
%integer_priority(10000000). % old value
integer_priority(Prio) :-  % priority for X : INTEGER enumerations
  tools:max_tagged_integer(P),
  Prio is P - 1024.
  % note inf_type_prio sets some priorities such as seq(INTEGER) ... higher

% priority for sets of infinite type (POW(INTEGER)...)
integer_pow2_priority(Prio) :-  % was 10000010
   integer_priority(P), Prio is P+10.


is_finite_priority(P) :-
   P < 100000, % TO DO: put max_tagged_integer for 32-bit system minus 1026 or so (see above)
   last_finite_priority(LFP),
   P =< LFP.

%integer_pow_priority(10000020) :- tools:max_tagged_integer(MAX).

% ensure that we do not exceed priority of type enumeration or use too big numbers
get_bounded_wait_flag(Prio,Info,WF,LWF) :-
    bound_prio(Prio,P),
    get_wait_flag(P,Info,WF,LWF).

bound_prio(Prio,P) :- number(Prio),
    Prio>268435452,  % 268435455 is max_tagged_integer on 32 bit platforms
    last_finite_priority(MAX),
    Prio>MAX,
    !, P=MAX.
bound_prio(P,P).

% -------------

% copy waitflags from one waitflag store to another; warning: can ground WF0 from outer store
% the assumption is that the outer store will be grounded/enumerated
% the inner store 1 will now point to the outer store 2 in case some co-routines still add waitflags
copy_waitflag_store(wfx(WF0,Store1,WFE1,_),wfx(WF0,Store2,WFE2,_Info2)) :-
    %print(copy_wf(WF0,Store1,Store2,WFE1,WFE2)),nl,
   my_get_wf_store_fdvars_atts(Store1,Heap,FDList1),
   (FDList1 = [] -> true
     ; my_get_fdvars_att(Store2,FDList2),
       %append(FDList1,FDList2,NewFDList), % used to be like this
       append_wf(FDList2,FDList1,NewFDList), % now we add inner FD variables to end
       put_mutable_attr(Store2,wf_fdvars(NewFDList))
   ),
   avl_to_list(Heap,WF_List),
   %print(copying(FDList1,WF_List)),nl,
	 my_get_wf_store_att(Store2,Heap2),
   add_wf(WF_List,Heap2,NewHeap2),
   put_mutable_attr(Store2,wf_store(NewHeap2)),
   set_mutable_ref(Store1,Store2),
   % ensures that calls which add waitflags now get-redirected to waitflag store 2
   % get_ef from Store2 and add trigger to enumerate Store1?
   %portray_waitflags(wfx(WF0,Store2,WFE2,Info2)),
   WFE1=WFE2.

append_wf([],FDL,FDL).
append_wf([H|T],FDL2,Res) :-
   (ground(H) -> append_wf(T,FDL2,Res) ; Res=[H|RT], append_wf(T,FDL2,RT)).

% Note: add_wf will construct multiple WF variables also for infinite_priorities,
%       get_waitflag_from_store will only keep multiple entries for infinite ones
add_wf([],Heap,Heap).
add_wf([Prio-wf(WF1,Info)|T],Heap,NewHeap) :- % purpose: copy waitflag WFlag to WF store; if same priority exists it will be appended at end
   (Prio=float(_) -> P=Prio ; P=float(Prio)),
   (avl_fetch(P,Heap,wf(WFs,_OldInfo))
     -> % priority waitflag(s) exist; add inner store WF1 at end
        merge_wf(WF1,WFs,MergedWFs),
	      avl_store(P,Heap,wf(MergedWFs,Info),Heap2)
     ; % this priority does not yet exist in outer store
       avl_store(P,Heap,wf(WF1,Info),Heap2)
    ),
    add_wf(T,Heap2,NewHeap).

% merge_wf(InnerWF,OuterWF,MergedWF)
merge_wf(WF1,WFs,MergedWFs) :- var(WF1),!,
   (var(WFs) -> MergedWFs=[WFs,WF1] ; append(WFs,[WF1],MergedWFs)).
merge_wf(WF1,WFs,MergedWFs) :- var(WFs),!, MergedWFs = [WFs|WF1].
merge_wf(WF1,WFs,MergedWFs) :- append(WFs,WF1,MergedWFs),!.
merge_wf(WF1,WFs,MergedWFs) :- add_internal_error('Illegal WF list:',merge_wf(WF1,WFs,MergedWFs)),
	         MergedWFs = [WFs|WF1].


% try and find an exact match for a waitflag info; this is linear in size of waitflag store !
find_waitflag_info(wfx(_WF0,Store,_WFE1,_),Info,Prio,WF1) :-
    my_get_wf_store_att(Store,Heap),
    avl_member(Prio,Heap,wf(WF1,Info)).

% dereference the waitflag store, see SICStus for SPRM-20503
% should be called e.g., in WHILE loops of interpreter to prevent degradation of put_atts performance
deref_wf(wfx(WF0,S,EF,I),R) :- !, R=wfx(WF0,S2,EF,I), deref_store(S,S2).
deref_wf(WF,WF).

% D is the dereferenced value of V.
% see email by SICStus for SPRM-20503; gets rid of variable chains; was useful for old Waitflag store with attributes
%deref(V,D) :- var(V), !, D = V.
%deref(X,X).

% --------------------------
% DEBUGGING UTILITIES:

% you also need to comment in code above using it
/*
:- dynamic debug_kernel_waitflags/1.
set_debug_kernel_waitflags :- (debug_kernel_waitflags(_) -> true ; assert(debug_kernel_waitflags(0))).

% the counter allows to set trace points
get_debug_kernel_waitflags(Counter) :- retract(debug_kernel_waitflags(C)), C1 is C+1,
   assert(debug_kernel_waitflags(C1)), Counter=C1.

get_debug_info(Info,Res) :- debug_kernel_waitflags(Counter),!,
   (Counter==39 -> trace ; true),
   Res=info(Counter,Info).
get_debug_info(Info,Info).
*/
% --------------------------

 :- assert_must_succeed(( kernel_waitflags:test_waitflags(1) )).
 :- assert_must_succeed(( kernel_waitflags:test_waitflags(2) )).

test_waitflags(1) :-
	init_wait_flags(Waitflags,[test_waitflags]),
	get_wait_flag(10,Waitflags,WF1),
	when(nonvar(WF1),print('WF1 - Prio 10\n')),
	get_wait_flag(5,Waitflags,WF2),
	when(nonvar(WF2),(print('WF2 - Prio 5\n'),
	       get_wait_flag(6,Waitflags,WF4),
         	when(nonvar(WF4),print('WF4 - Prio 6\n')),
	       get_wait_flag(2,Waitflags,WF5),
	      when(nonvar(WF5),print('WF5 - Prio2\n')))),
        get_wait_flag(5,Waitflags,WF3),
	when(nonvar(WF3),print('WF3 - Prio 5\n')),
	ground_wait_flags(Waitflags),nl.

test_waitflags(2) :-
  init_wait_flags(Waitflags,[test_waitflags]),
  get_wait_flag(10,Waitflags,WF1),
  when(nonvar(WF1),print('WF1 Prio 10\n')),
  get_wait_flag(5,Waitflags,WF2),
   when(nonvar(WF2),(print('WF2 - Prio 5\n'),
   get_wait_flag(6,Waitflags,WF4),
 	when(nonvar(WF4),print('WF4 - Prio 6\n')),
   get_wait_flag(2,Waitflags,WF5),
  when(nonvar(WF5),print('WF5 - Prio2\n')))),
  get_wait_flag(5,Waitflags,WF3),
  when(nonvar(WF3),print('WF3 - Prio 5\n')),
  ground_wait_flags(Waitflags),nl.


/*

bench_waitflags(N) :-
  statistics(walltime,[Tot1,_]),
  init_wait_flags(Waitflags),
  getwf(N,Waitflags),
  ground_wait_flags(Waitflags),nl,statistics(walltime,[Tot2,_]), Tot is Tot2-Tot1,
  format('Waitflag test for size ~w : ~w ms walltime (wf=~w)]~n',[N,Tot,Waitflags]).
getwf(0,_).
getwf(N,Waitflags) :- N>0, N1 is N-1, get_wait_flag(N,bench(N),Waitflags,_), getwf(N1,Waitflags).

% old attribute based:
| ?- kernel_waitflags:bench_waitflags(1000).

Waitflag test for size 1000 : 37 ms walltime (wf=wfx(0,_205205,0,[]))]
yes
| ?- kernel_waitflags:bench_waitflags(10000).

Waitflag test for size 10000 : 5859 ms walltime (wf=wfx(0,_2512071,0,[]))]
yes

% new mutable based:
| ?- kernel_waitflags:bench_waitflags(1000).

Waitflag test for size 1000 : 4 ms walltime (wf=wfx(0,$mutable(wfm_store(empty,[],[]),3008),0,[]))]
yes
| ?- kernel_waitflags:bench_waitflags(10000).

Waitflag test for size 10000 : 24 ms walltime (wf=wfx(0,$mutable(wfm_store(empty,[],[]),30008),0,[]))]
yes
| ?- kernel_waitflags:bench_waitflags(100000).

Waitflag test for size 100000 : 497 ms walltime (wf=wfx(0,$mutable(wfm_store(empty,[],[]),9),0,[]))]
yes

*/
