% (c) 2009-2020 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)


:- module(kodkod,[apply_kodkod/3, apply_kodkod/4,
                  replace_by_kodkod/3,
                  current_solver_is_not_incremental/0,
                  kodkod_request/5,
                  %apply_kodkod_to_counterexample_search/5,
                  kodkod_reset/0]).

:- use_module(library(lists)).
:- use_module(library(avl)).

:- use_module(probsrc(tools)).
:- use_module(probsrc(bmachine)).
:- use_module(probsrc(self_check)).
:- use_module(probsrc(bsyntaxtree)).
:- use_module(probsrc(translate)).
:- use_module(probsrc(predicate_analysis)).
:- use_module(probsrc(btypechecker)).
:- use_module(probsrc(b_global_sets),[b_global_set/1,lookup_global_constant/2]).
:- use_module(probsrc(preferences),[get_preference/2]).

:- use_module(kodkod_translate).
:- use_module(kodkod_tools).
:- use_module(kodkod_rewrite).
:- use_module(kodkod2).
:- use_module(kodkod_typing).
:- use_module(kodkod_process).
:- use_module(kodkod_annotator).

:- use_module(probsrc(module_information),[module_info/2]).
:- module_info(group,kodkod).
:- module_info(description,'This is the central module of the B to Kodkod translation (still experimental).').

%:- dynamic kodkod_error/1.

kodkod_reset :-
    ( stop_kodkod -> true; true).

% Signum is either pos (when we want the predicate to succeed) or neg (when we want the predicate to be false)
kodkod_request(Id,Signum,LocalState,State,LWF) :-
    % MaxNrOfSols is the maximum number of solutions found in one go by Kodkod
    get_max_nr_of_solutions(MaxNrOfSols),
    kodkod_request(Id,Signum,MaxNrOfSols,LocalState,State,LWF).

get_max_nr_of_solutions(MaxNrOfSols) :-
    current_solver_is_not_incremental,!,
    MaxNrOfSols = 1.
get_max_nr_of_solutions(MaxNrOfSols) :-
    get_preference(kodkod_max_nr_solutions,MaxNrOfSols). % TO DO: we could look at kodkod_for_components and maxNrOfInitialisations preference

current_solver_is_not_incremental :-
    get_preference(kodkod_sat_solver,Solver),
    not_incremental(Solver).
not_incremental(lingeling).

:- use_module(probsrc(error_manager)).

:- block kodkod_request(?,?,?,?,?,-).
kodkod_request(Id,Signum,MaxNrOfSols,LocalState,State,_LWF) :-
    on_exception(error(system_error,system_error(E)), % often SPIO_E_NET_CONNRESET
       kodkod_request_intern(Id,Signum,MaxNrOfSols,LocalState,State),
       (add_internal_error('System Error while processing Kodkod request: ',E),
        fail
        )).

apply_kodkod(Identifiers,Predicate,NewPredicate) :-
    get_preference(try_kodkod_on_load,true),
    !,
    (replace_by_kodkod(Identifiers,Predicate,NewPredicate) -> true
      ; get_texpr_ids(Identifiers,Ids),
        add_warning(kodkod_fail,'Could not use KODKOD to translate predicate to SAT, identifiers = ', Ids, Predicate),
        NewPredicate = Predicate).
apply_kodkod(_Identifiers,Predicate,Predicate).

create_id(Id,Type,b(identifier(Id),Type,[])).

% a version which takes ids and types list:
apply_kodkod(Ids,Types,Predicate,NewPredicate) :-
    get_preference(try_kodkod_on_load,true),
    maplist(create_id,Ids,Types,Identifiers),
    !,
    apply_kodkod(Identifiers,Predicate,NewPredicate).
apply_kodkod(_Identifiers,_Types,Predicate,Predicate).

%
:- use_module(probsrc(b_expression_sharing),[expand_cse_lazy_lets_if_cse_enabled/2]).
replace_by_kodkod(Constants,Predicate,NewPredicate) :-
    % Remove any CSE lazy_lets as Kodkod translation cannot deal with them yet
    expand_cse_lazy_lets_if_cse_enabled(Predicate,Predicate0),
    % The Predicate may already contain Kodkod calls, replace
    % them by their original B predicate
    undo_kodkod_replacements(Predicate0,Predicate1),
    % first we do a predicate analysis to determine ranges for integers
    try_predicate_analysis_with_global_sets(Predicate1,Constants,TaggedConstants,TaggedPredicate),
    replace_by_kodkod2(TaggedConstants,TaggedPredicate,NewPredicate).

:- use_module(probsrc(preferences),[temporary_set_preference/3,reset_temporary_preference/2]).
replace_by_kodkod2(Constants,_Predicate,NewPredicate) :-
    % If the static analysis states that there is no possible solution for one
    % of the identifiers, we just say "false"
    somechk(is_inconsistent_expression,Constants),!,
    write('kodkod: Static analysis detected inconsistencies, new predicate is "false"'),nl,
    create_texpr(falsity,pred,[],NewPredicate).
replace_by_kodkod2(Constants,Predicate,NewPredicate) :-
    temporary_set_preference(partition_predicates_inline_equalities,false,CHNG),
    split_predicates(Predicate,PList), % will split into components if KODKOD_ONLY_FULL = FALSE
    reset_temporary_preference(partition_predicates_inline_equalities,CHNG),
    % we try to translate each predicate alone
    try_apply_transformation_l(PList,Constants,1,Result),
    print_results(Result),
    % split the result into two lists: the successfull translatable predicates (ToKodkod)
    % and those which were not translatable (StayProb)
    filter_translatable(Result,ToKodkod,StayProb),
    %%print('Stay ProB: '), conjunct_predicates(StayProb,StayPred), translate:print_bexpr(StayPred),nl,
    % fail if there are no translatable predicates
    ToKodkod = [_|_],
    % now, translate all translatable predicates in one run
    (StayProb = [_|_] -> add_message(kodkod,'Not translating all components/predicates with Kodkod') ; true ),
    % with a special preference to force full evaluation we could also abort here and return UNKNOWN
    apply_transformation(Constants,ToKodkod,KPred),
    % the new
    %%print('For Kodkod: '), translate:print_bexpr(KPred),nl,
    conjunct_predicates([KPred|StayProb],NewPredicate).


split_predicates(Predicate,Parts) :-
    get_preference(kodkod_for_components,true),!, % only translate full components (or component not at all)
    predicate_components(Predicate,Components),
    findall(P,member(component(P,_),Components),Parts).
split_predicates(Predicate,Parts) :-
    conjunction_to_list(Predicate,Parts).

split_simple_equalities(InPredicate,Equalities,Rest) :-
    conjunction_to_list(InPredicate,InPredicates),
    split_list(is_simple_equality,InPredicates,EqualitiesL,RestL),
    conjunct_predicates(EqualitiesL,Equalities),
    conjunct_predicates(RestL,Rest).

is_simple_equality(TExpr) :-
    get_texpr_expr(TExpr,equal(A,B)),
    ( get_texpr_id(A,_),is_base_expression(B)
    ; get_texpr_id(B,_),is_base_expression(A)),!.

is_base_expression(TExpr) :-
    get_texpr_expr(TExpr,Expr),
    is_base_expression2(Expr).
is_base_expression2(identifier(Id)) :- b_global_set(Id),!. % a deferred set
is_base_expression2(identifier(Id)) :- lookup_global_constant(Id,_). % an enumerated set element
is_base_expression2(integer(_)).
is_base_expression2(couple(A,B)) :- is_base_expression(A),is_base_expression(B).
is_base_expression2(set_extension(L)) :- maplist(is_base_expression,L).
is_base_expression2(value(_)).
is_base_expression2(boolean_true).
is_base_expression2(boolean_false).
is_base_expression2(empty_set).
is_base_expression2(bool_set).
is_base_expression2(add(A,B)) :- is_base_expression(A),is_base_expression(B).
is_base_expression2(minus(A,B)) :- is_base_expression(A),is_base_expression(B).
is_base_expression2(unary_minus(A)) :- is_base_expression(A).
is_base_expression2(multiplication(A,B)) :- is_base_expression(A),is_base_expression(B).
is_base_expression2(div(A,B)) :- is_base_expression(A),is_base_expression(B).
is_base_expression2(floored_div(A,B)) :- is_base_expression(A),is_base_expression(B).
is_base_expression2(modulo(A,B)) :- is_base_expression(A),is_base_expression(B).
is_base_expression2(power_of(A,B)) :- is_base_expression(A),is_base_expression(B).


undo_kodkod_replacements(Old,New) :-
    contains_kodkod_call(Old),!,
    undo_kodkod_replacements2(Old,New).
undo_kodkod_replacements(Old,Old).
undo_kodkod_replacements2(TOld,TNew) :-
    get_texpr_expr(TOld,Old),
    undo_kodkod_replacements3(Old,TOld,TNew).
undo_kodkod_replacements3(conjunct(A,B),_TOld,TNew) :-
    !,create_texpr(conjunct(NewA,NewB),pred,[],TNew),
    undo_kodkod_replacements2(A,NewA),
    undo_kodkod_replacements2(B,NewB).
undo_kodkod_replacements3(kodkod(Id,_),_TOld,Orig) :-
    !,get_original_problem(Id,Orig).
undo_kodkod_replacements3(_Old,TOld,TOld).

contains_kodkod_call(TExpr) :-
    get_texpr_expr(TExpr,Expr),
    contains_kodkod_call2(Expr),!.
contains_kodkod_call2(kodkod(_,_)).
contains_kodkod_call2(conjunct(A,_)) :- contains_kodkod_call(A).
contains_kodkod_call2(conjunct(_,B)) :- contains_kodkod_call(B).


try_predicate_analysis_with_global_sets(Predicate,Constants,TaggedConstants,TaggedPredicate) :-
    catch(predicate_analysis_with_global_sets(Predicate,[],Constants,TaggedConstants,TaggedPredicate),
          failed_analysis(Expr),
          ( write('kodkod fail: predicate analysis failed on expression '),
              translate_bexpression_with_limit(Expr,80,Msg),
              write(Msg),nl,fail)),
    !.
try_predicate_analysis_with_global_sets(_Predicate,_Constants,_TaggedConstants,_TaggedPredicate) :-
    write('kodkod fail: predicate analysis failed.\n'),
    fail.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- meta_predicate try_call(-,-,0).

try_apply_transformation(Predicate,Constants,Index,Result) :-
    %%print('TRY KODKOD ON: '), translate:print_bexpr(Predicate),nl,
    Log = log(Predicate,Result,Index,Ints,Bits),
    try_call( 'applying rewriting rules', Log,
              apply_rewriting_rules(Predicate,Rewritten)),
    %%print('REWRITTEN: '), translate:print_bexpr(Rewritten),nl,
    try_call( 'extracting used types', Log,
              extract_used_types(Rewritten,Types,Ints,Bits)),
    try_call( 'creating Kodkod atoms for types', Log,
              create_atoms_for_types(Types,Ints,Bits,Typemap,Idmap1,_Valuemap)),
    try_call( 'translate and type Kodkod formula', Log,
              typed_translation(Constants,Predicate,Bits,Typemap,Idmap1,_Idmap,Rewritten,_KExpr)),
    finalize_result(Log).
try_call(_Msg,log(_Predicate,Result,_Index,_,_),_Call) :-
    nonvar(Result),!.
try_call(Msg,log(Predicate,Result,Index,_,_),Call) :-
    %%format('Try ~w ~n',[Msg]),
    ( catch(Call,
            kodkod(T),
            ( %%format('Exception: ~w~n',[T]),
            Result = failure(Index,Predicate,Msg,kodkod_exception(T)))
            ) -> true
    ; otherwise ->
        %%format('Failure for: ~w~n',[Msg]),
        Result = failure(Index,Predicate,Msg,failed)
    ).
finalize_result(log(Predicate,Result,Index,Ints,Bits)) :-
    var(Result),!,Result=ok(Index,Predicate,Ints,Bits).
finalize_result(_).


typed_translation(Constants,Predicate,Bits,Typemap,Idmap1,Idmap,Rewritten,KExpr) :-
    % Create the relations for the constants that are actually used
    restrict_to_used_constants(Constants,Predicate,UsedConstants),
    create_identifier_relations(UsedConstants,Typemap,Bits,Idmap1,Idmap),
    % translate the formula in two steps:
    kodkod_translate(Rewritten,Typemap,Idmap,Formula),
    kodkod_insert_casts(Formula,Typemap,Idmap,KExpr),
    !.

restrict_to_used_constants(Constants,Predicate,UsedConstants) :-
    find_identifier_uses(Predicate,['$examine_value_was_identifier_info'],UsedIds), % flag says: also get value(_) with was_identifier(_) field
    include(constant_is_used(UsedIds),Constants,UsedConstants).
constant_is_used(UsedIds,C) :-
    get_texpr_id(C,Id),memberchk(Id,UsedIds).

try_apply_transformation_l([],_Constants,_,[]).
try_apply_transformation_l([Predicate|Prest],Constants,I,[Result|Rrest]) :-
    try_apply_transformation(Predicate,Constants,I,Result),
    I2 is I+1,
    try_apply_transformation_l(Prest,Constants,I2,Rrest).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

filter_translatable([],[],[]).
filter_translatable([Res|Rrest],ToKodkod,StayProb) :-
    ( Res = ok(_,P,_,_)      -> ToKodkod=[P|Krest], StayProb=Prest
    ; Res = failure(_,P,_,_) -> ToKodkod=Krest, StayProb=[P|Prest]),
    filter_translatable(Rrest,Krest,Prest).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
:- use_module(probsrc(error_manager),[add_warning/3,add_warning/4]).
:- use_module(probsrc(debug),[debug_mode/1,silent_mode/1,nls/0,debug_println/2]).

apply_transformation(AllConstants,PredicateList,Result) :-
    conjunct_predicates(PredicateList,Predicate1),
    (debug_mode(on) -> print('TRANSLATING FOR KODKOD: '),nl,translate:print_bexpr_with_limit(Predicate1,1500),nl ; true),
    apply_transformation2(AllConstants,Predicate1,Result).

apply_transformation2(AllConstants,Predicate1,Result) :-
    split_simple_equalities(Predicate1,Equalities,Predicate),
    (debug_mode(on) -> print('Equalities: '), translate:print_bexpr_with_limit(Equalities,1500),nl ; true),
    (is_truth(Predicate)
     -> debug_println(19,'Predicate for Kodkod has only equalities. Predicate will be solved by ProB.'),
        Result=Predicate1
        % we could also fail, but then we would create a kodkod_fail warning
      ; apply_transformation3(AllConstants,Predicate,Equalities,Result)
       -> (debug_mode(on) -> print('RESULT: '),nl,translate:print_bexpr_with_limit(Result,1500),nl ; true)
      ; add_warning(kodkod,'Kodkod Transformation Failed for Predicate over: ',AllConstants),
        %trace, apply_transformation3(AllConstants,Predicate,Equalities,Result),
        fail).

apply_transformation3(AllConstants,Predicate,Equalities,Result) :-
    restrict_to_used_constants(AllConstants,Predicate,Constants),
    apply_rewriting_rules(Predicate,Rewritten),
    extract_used_types(Rewritten,Types,Ints,Bits),
    create_atoms_for_types(Types,Ints,Bits,Typemap,Idmap1,Valuemap),
    % check if any types are used, if not Kodkod would exit with an
    % empty universe exception
    % (java.lang.IllegalArgumentException: Cannot create an empty universe. or something similar)
    \+ empty_avl(Typemap),
    typed_translation(Constants,Predicate,Bits,Typemap,Idmap1,Idmap,Rewritten,KExpr),

    register_problem(Constants,KExpr,Idmap,Typemap,Valuemap,Predicate,ProblemNr),
    create_texpr(kodkod(ProblemNr,Constants),pred,[],KPred),
    conjunct_predicates([Equalities,KPred],Result).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

print_results([]).
print_results([R|Rrest]) :-
    print_result(R),nls,
    print_results(Rrest).
print_result(ok(_Ids,Index,P,Ints,Bits)) :-
    print_result(ok(Index,P,Ints,Bits)).
print_result(ok(_Index,P,Ints,Bits)) :-
    (silent_mode(on) -> true
      ; write('kodkod ok:   '),print_bexpr_with_limit(P,50),
        write('  ints: '),write(Ints),write(', intatoms: '),write(Bits)
     ).
print_result(failure(_Ids,Index,P,Msg,Reason)) :-
    print_result(failure(Index,P,Msg,Reason)).
print_result(failure(_Index,P,Msg,Reason)) :-
    write('kodkod fail: '),
    translate_bexpression_with_limit(P,50,PS), write(PS),
    write('  , reason: '),write_reason(Reason),write(' while '),write(Msg),
    (get_preference(kodkod_raise_warnings,true)
      -> add_warning(kodkod,'Cannot translate predicate for Kodkod: ',PS,P)
      ;  true).
write_reason(kodkod_exception(KE)) :-
    write_reason2(KE),!.
write_reason(E) :- write(E).

write_reason2(unsupported_type(T)) :-
    pretty_type(T,PT),
    write('unsupported type '),write(PT).
write_reason2(unsupported_expression(TE)) :-
    get_texpr_expr(TE,T),
    functor(T,Functor,Arity),
    translate_bexpression(TE,PE),
    format('unsupported expression ~w/~w: ~w',[Functor,Arity,PE]).
write_reason2(unsupported_int(Msg)) :-
    write(Msg).
write_reason2(side_condition(Msg,PPExpr)) :-
    format('~w: ~w',[Msg,PPExpr]).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- public pdebug/0.
pdebug :-
    install_b_portray_hook,
    assert(( user:portray(X) :- is_avl(X),!,portray_avl(X) )).
