:- module(kodkod_tools,[ extract_finite_integer_range/3
                       , unique_identifier/1
                       , unique_identifier/2
                       , unique_identifier/3
                       , merge_integer_info/3
                       , kodkod_conjunction/2
                       , kodkod_implication/3
                       , disassemble_conjunction/2

                       , pow2integer_relation_kodkod_name/1
                       , intset_relation_kodkod_name/1

                       , is_inconsistent_expression/1
                       ]).

:- use_module(probsrc(tools), [ajoin/2]).
:- use_module(probsrc(module_information)).
:- use_module(probsrc(translate), [translate_bexpression/2]).
:- use_module(probsrc(bsyntaxtree)).
:- use_module(probsrc(predicate_analysis), [info_disjunct/3]).

:- module_info(group,kodkod).
:- module_info(description,'This module contains helper predicates for the Kodkod translation part of ProB').

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% constants
pow2integer_relation_kodkod_name(p2i).
intset_relation_kodkod_name(is).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% helper to extract the integer range of a type expression
extract_finite_integer_range(TExpr,ScopeList,IRange) :-
    convert_scopelist_to_scope(ScopeList,interval,Scope),
    extract_finite_integer_range2(TExpr,Scope,IRange).
convert_scopelist_to_scope([],Scope,Scope).
convert_scopelist_to_scope([Access|Rest],OScope,Scope) :-
    functor(NScope,Access,1),arg(1,NScope,OScope),
    convert_scopelist_to_scope(Rest,NScope,Scope).
extract_finite_integer_range2(TExpr,Scope,IRange) :-
    get_texpr_info(TExpr,Info),
    memberchk(analysis(Analysis),Info),
    memberchk(Scope:IRange0,Analysis),
    ( is_inconsistent_intrange(IRange0) ->
        IRange = irange(0,0)    % throwing an exception is not really sensible, e.g.
                                % if we have an inconsistent range of a variable in a
                                % comprehension set - then the set would just be empty
                                % We can safely(?) assume a larger range (here 0..0)
    ; otherwise ->
        is_finite(IRange0),IRange0=IRange),!.
extract_finite_integer_range2(TExpr,_Scope,_IRange) :-
    translate_bexpression(TExpr,SExpr),
    ajoin(['integer without upper and lower bound in: ', SExpr], Msg),
    throw(kodkod(unsupported_int(Msg))).
is_finite(irange(Min,Max)) :- number(Min),number(Max).
is_inconsistent_intrange(irange(Min,Max)) :-
    number(Min),number(Max),Max < Min.
is_inconsistent_expression(b(E,_,Info)) :-
    memberchk(analysis(Analysis),Info),
    member(Type:IRange,Analysis), % elem(interval)
    \+(Type=elem(_)), % otherwise, e.g., for elem(interval) we talk about possible elements, an inconsistency here simply means that the set is empty
    is_inconsistent_intrange(IRange),!,
    debug:debug_println(19,inconsistent_range(E,IRange)).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% unique identifier
:- dynamic unique_counter/1, debug_identifier/0.


unique_identifier(Prefix,Name,Id) :-
    ( debug_identifier ->
        increase_counter(C),
        ajoin([Prefix,C,'_',Name],Id)
    ; otherwise ->
        unique_identifier(Prefix,Id)).
unique_identifier(Prefix,Id) :-
    increase_counter(C),
    ajoin([Prefix,C],Id).
unique_identifier(Id) :-
    unique_identifier(tmpid__,Id).
increase_counter(C2) :-
    ( retract(unique_counter(C)) -> true; C=0 ),
    C2 is C+1,
    assert(unique_counter(C2)).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% merge two integer ranges
merge_integer_info(none,A,A) :- !.
merge_integer_info(A,none,A) :- !.
merge_integer_info(IRangeA,IRangeB,IRange) :-
    info_disjunct(IRangeA,IRangeB,IRange).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% combining Kodkod expressions

kodkod_conjunction(List,Result) :-
    findall(E,(member(E,List),E \= true),L2),
    kodkod_conjunction2(L2,Result).
kodkod_conjunction2([],true).
kodkod_conjunction2([F],F) :- !.
kodkod_conjunction2([A,B|R],and(A,F)) :- kodkod_conjunction2([B|R],F).

kodkod_implication(true,Formula,Formula) :- !.
kodkod_implication(false,_,true) :- !.
kodkod_implication(F1,F2,implies(F1,F2)).

disassemble_conjunction(Formula,Result) :-
    disassemble_conjunction2(Formula,Result,[]),!.
disassemble_conjunction2(and(A,B)) -->
    disassemble_conjunction2(A),
    disassemble_conjunction2(B).
disassemble_conjunction2(X) --> {X\=and(_,_)},[X].
