% (c) 2012-2020 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(ample_sets, [
      compute_ample_set2/3,
      clear_dynamic_predicates_for_POR/0,
      reset_runtime_dynamic_predicates_POR/0, % now done via event-handling
      stutter_action/1, visible_action/1,check_cycle_proviso/0 % exported for ltl.pl for asserting and retracting stutter/visible actions
      ]).

/* SICSTUS Libraries */
:- use_module(library(ordsets)).
:- use_module(library(random)).
:- use_module(library(lists)).
:- use_module(library(ugraphs)).
%:- use_module(library(timeout)).

/* B module sources */
:- use_module(probsrc(bmachine),[b_get_invariant_from_machine/1,b_operation_preserves_full_invariant/1]).
:- use_module(probsrc(bsyntaxtree),[conjunction_to_list/2,occurs_in_expr/2]).
:- use_module(probsrc(specfile),[prepare_state_for_specfile_trans/2,specfile_possible_trans_name/2,specfile_trans/6]).
:- use_module(probsrc(state_space),[visited_expression/2,use_no_timeout/1]).

/* POR modules */
:- use_module(probporsrc(enable_graph)).
  % --> /* The Static Analysis' module */
:- use_module(probporsrc(static_analysis),[enable_graph/1,dependent_act/4]).
:- use_module(probpgesrc(pge_algo),[is_pge_opt_on/1,compute_and_update_disabled_enabled_operations/3]).

/* Importing predicates for handling the states of a respective B/Event-B machine */
:- use_module(probsrc(specfile),[state_corresponds_to_initialised_b_machine/2]).
:- use_module(probsrc(store), [empty_state/1]).

/* Other modules from ProB */
:- use_module(probsrc(debug),[debug_println/2]).
:- use_module(probsrc(preferences), [get_preference/2]).
:- use_module(probsrc(error_manager), [time_out_with_enum_warning_for_findall/3, add_warning/3]).

% Cosider to remove this import
:- use_module(probsrc(b_read_write_info),[b_get_read_write/3]).

/* Modules and Infos for the code coverage analysis */
:- use_module(probsrc(module_information)).
:- module_info(group,model_checker).
:- module_info(description,'This module contains predictes for partial order reduction using the ample sets\' theory.').

% Dynamic predicates for keeping track of visible and stutter actions,
% as well as of all dependent reations between the actions of the B model
:- dynamic stutter_action/1,visible_action/1,enables/2,explored_state/1,check_cycle_proviso/0.


:- public reset_static_dynamic_predicates_POR/0.
clear_dynamic_predicates_for_POR :-
    reset_static_dynamic_predicates_POR,
    reset_runtime_dynamic_predicates_POR.

:- use_module(probsrc(eventhandling),[register_event_listener/3]).
:- register_event_listener(clear_specification,clear_dynamic_predicates_for_POR,
                           'Reset Ample Sets.').
:- register_event_listener(reset_specification,reset_static_dynamic_predicates_POR,
                           'Reset Ample Sets.').

% predicates that will be asserted once on every loading of the machine
reset_static_dynamic_predicates_POR :-
   retractall(dependent_act(_,_,_,_)),
   retractall(enable_graph(_)).

% predicates that will be asserted at runtime
reset_runtime_dynamic_predicates_POR :-
    retractall(stutter_action(_)),
    retractall(visible_action(_)),
    retractall(explored_state(_)),
    retractall(enables(_,_)),
    retractall(check_cycle_proviso). % used to not track in the enable graph

/* This implementation of the ample sets approach sastisfies all four conditions of the ample sets theory (not proved yet).
   The conditions are taken from the book "Principles of Model Checking" written by C.Baier and J.P.Katoen.

For every state s in the state space, where Act(s) stands for all enabled actions in state s, the following conditions
must be satisfied:

1. Nonemptiness condition (A1) and terminal states
   (A 1.1) ample(s) <: Act(s) (set equality is possible)
   (A 1.2) ample(s) = {} <=> Act(s) = {}

2. Dependency condition
   (A 2) Let s -a_{1}> s_{1} -a_{2}> ... -a_{n}> s_{n} -a> s' be a finite execution
         fragment in the original state space. If a depends on ample(s), then a_{i}:ample(s) for some 0<i<=n.

   This condition can be satisfied by the following conditions (Local Criteria):
      (A 2.1) Any action b: Act(s)\ample(s) is independent of any a:ample(s) (Here we use the dependency graph generated from the Enabling Analysis)
      (A 2.2) Any disabled action b:Act\en(s) which is dependent on ample(s) cannot become enabled by a possible sequence of actions starting with an action
              from en(s)\ample(s). (We check this condition by means of using the enable graph generated from the Enabling Analysis)

      Note that determining the local criteria's conditions above is dependent on the formalism on which we are applying the partial order reduction method.
      The conditions (A 2.1) and (A 2.2) are adjusted to the semantics of the B method/Event-B. For other formalisms such as CSP we will need to define
      the conditions differently. On the one hand because we cannot use the Constraint Based Checker of ProB (it is not implemented for CSP yet), on the other hand
      because CSP differs semantically from B/Event-B (the process structure). An idea to apply Partial Order Reduction on CSP is to make the reduction
      on the CSP operator-level ('|||', [|X|] and so on). For more detailed description of how to apply POR on CSP read the following papers:
            - H. Wehrheim. Partial order reductions for failure refinement.
            - J. Sung, Y. Liu and J.S. Dong. Model Checking CSP Revisited: Introducing a Process Analysis Toolkit.

3. Stutter condition
   (A 3) If ample(s) /= Act(s) then any action t:ample(s) is a stutter action.

4. Cycle condition
   (A 4) For any cycle s_{0}s_{1}...s_{n} in the reduced state space and actions t:Act(s_{i}), for some
         0 < i <= n, there exists j:{1..n} such that t:ample(s_{j}). (s_{0},...,s_{n} are states)

 */

% compute_ample_set2(+CurID,+CurState,+POROptions)
%
% The main predicate for computing ample sets by respecting the four ample set conditions specified above.
% A subset of all possible enabled outgoing transitions in CurState will be computed and added to the state space. The predicate is similar
% to the compute_all_transitions/2 predicate implemented in tcltk_interface.pl.

:- use_module(probsrc(succeed_max),[reset_max_reached/0, max_reached/0]).
:- use_module(probsrc(state_space),[assert_max_reached_for_node/1]).
:- use_module(probsrc(value_persistance),[save_constants/1, add_new_transitions_to_cache/1]).

% Arguments:
% CurID - the current ID of the state.
% CurState - the current evaluation of all global variables (and constants) in the machine.
% FindInvViolations (obsolete) - possible values 1 (option setted) and 0 (option not setted). Conditions (A 3) and (A 4) checked if check_cycle_proviso/0 is set.
% POROptions - the options being set for exploring the state space by means of partial order reduction
%
compute_ample_set2(CurID,CurState,por(TYPE,UseEnableGraph,Depth,PGE)) :-
   reset_max_reached,
   prepare_state_for_specfile_trans(CurState,PreparedCurState),
   get_enabled_actions_at_state(CurID,PreparedCurState,PGE,TopLevelActions,Transitions,DisabledTopLevelActions),
   % we want only unique action names for analyzing and computing of ample sets
   list_to_ord_set(TopLevelActions,TopLevelActionsSet),
   % compute a sound ample set from all actions enabled in state CurID
   compute_ample_actions(CurID,TopLevelActionsSet,por(TYPE,UseEnableGraph,Depth,PGE),AmpleSet),
   debug_println(20,sets(CurID,allActions(TopLevelActionsSet),ampleSet(AmpleSet))),
   % mark state as visited (Cycle Condition (A 4))
   assert(explored_state(CurID)),
   ( (AmpleSet \= []) ->
      % compute just the actions from the ample set (case of state space reduction)
      % AmpleSet is a non-empty subset of enabled(CurID), AmpleSet can be equal to AllEnabledActions
      add_transitions(AmpleSet,CurID,Transitions,ActNameSuccessors),
      update_enabled_disabled_information_if_necessary(PGE,ActNameSuccessors,DisabledTopLevelActions,AmpleSet)
   ; otherwise ->
      true /* DEADLOCK */
   ),
   (max_reached ->
     assert_max_reached_for_node(CurID),
     MaxReached=true
   ; MaxReached=false
   ),
   (CurID==root ->
      save_constants(MaxReached)
   ;  add_new_transitions_to_cache(CurID)
   ),!. % once we have computed the ample set in CurState we don't want to backtrack

% computing a sound ample set of actions.
compute_ample_actions(CurID,AllEnabledActions,POROptions,AmpleSet) :-
  % get_ample_actions/5 call only reasonable to be executed when more than one actions are enabled in CurID
  (more_than_one_event(AllEnabledActions) ->
    get_ample_actions(CurID,AllEnabledActions,POROptions,AmpleSet)
  ; otherwise ->
    AmpleSet = AllEnabledActions
  ).

more_than_one_event(SetOfActions) :-
  length(SetOfActions,L),L>1.

/************************** ADD TRANSITIONS (BEGIN) ******************************/

% +ActNameSuccessors: a list of tuples (Act,SuccIds), where 'SuccIds' is the list of all state ids that are successor states of Act
add_transitions(AmpleSet,CurID,Transitions,ActNameSuccessors) :-
  add_transitions(no,AmpleSet,CurID,Transitions,[],ActNameSuccessors).

% first argument tests if the state has been fully explored while adding the
% ample set transitions to the state space ('yes' means state has been fully explored, so we can stop,
% 'no' means that the rest of the ample set transitions should be added to the state space until
% no more ample actions in the list or until the state is fully explored)
add_transitions(yes,_AmpleSet,_CurID,_Transitions,ActNameSuccessors,ActNameSuccessors).
add_transitions(_FullyExplored,[],_CurID,_Transitions,ActNameSuccessors,ActNameSuccessors).
add_transitions(no,[ActionName|ActNames],CurID,Transitions,ActNameSuccessorsGen,ActNameSuccessors) :-
  add_action_name_transitions(Transitions,CurID,ActionName,[],RemTransitions,FullyExplored,ActNameSuccessorsGen,ActNameSuccessorsGen1),
  add_transitions(FullyExplored,ActNames,CurID,RemTransitions,ActNameSuccessorsGen1,ActNameSuccessors).

add_action_name_transitions([],_CurID,_ActionName,RemTransitions,RemTransitions,no,ActSuccsGen,ActSuccsGen).
add_action_name_transitions([(ActionName1,Transitions)|R],CurID,ActionName,IgnoredActions_so_far,
                            RemTransitions,FullyExplored,ActSuccsGen,ActSuccs) :-
  (ActionName == ActionName1 ->
    % in this case we do not need to go through the rest of the transitions list
    add_action_transitions(ActionName,Transitions,CurID,NewIds),
    ActSuccsGen1 = [(ActionName,NewIds)|ActSuccsGen],
    append(IgnoredActions_so_far,R,AllRemainedActions),
    (cycle_condition_check_positive_l(NewIds) ->
      debug_println(9,'*********** POSSIBLE CYCLE DETECTED *************'),
      debug_println(9,computing_all_other_enabled_actions(back_transition(ActionName),AllRemainedActions)),
      add_all_other_enabled_transitions(AllRemainedActions,CurID,ActSuccsGen1,ActSuccs),
      RemTransitions=[],
      FullyExplored = yes
    ; otherwise -> FullyExplored = no, RemTransitions = AllRemainedActions, ActSuccs=ActSuccsGen1)
  ; otherwise ->
    % keeping track of all ignored actions so far, we don't want to go through the same order of actions every time we
    % want to add a transition from AmpleSet to the state space
    ord_add_element(IgnoredActions_so_far, (ActionName1,Transitions), IgnoredActions_so_far1),
    add_action_name_transitions(R,CurID,ActionName,IgnoredActions_so_far1,RemTransitions,FullyExplored,ActSuccsGen,ActSuccs)
  ).

% cycle condition A4 check
cycle_condition_check_positive_l(NewIds) :-
  check_cycle_proviso,!,
  member(NewId,NewIds),
  \+var(NewId),
  explored_state(NewId).

add_all_other_enabled_transitions([],_CurID,ActSuccsGen,ActSuccsGen).
add_all_other_enabled_transitions([(ActionName,Transitions)|R],CurID,ActSuccsGen,ActSuccs) :-
  add_action_transitions(ActionName,Transitions,CurID,NewIds),
  ActSuccsGen1 = [(ActionName,NewIds)|ActSuccsGen],
  add_all_other_enabled_transitions(R,CurID,ActSuccsGen1,ActSuccs).

add_action_transitions(_ActionName,Transitions,CurID,NewIds) :-
  findall(NewId,
    (member((Act,NewExpression,TransInfo,Residue),Transitions),
      tcltk_interface: add_trans_id(CurID,Act,NewExpression,Residue,NewId,TransInfo,_)),NewIds).

update_enabled_disabled_information_if_necessary(PGE,ActNameSuccessors,DisabledActionsAtCurState,EnabledActionsAtCurState) :-
  (is_pge_opt_on(PGE) ->
     compute_and_update_disabled_enabled_operations(ActNameSuccessors,DisabledActionsAtCurState,EnabledActionsAtCurState)
     ; true).

/************************** ADD TRANSITIONS (END) ******************************/

/* Auxiliary predicates for getting action names enabled at the current state */

% getting action names from the action tuples
%get_action_names_only(Actions,ActionNames) :-
%    maplist(get_action_name,Actions,ActionNames).
%get_action_name((ActionName,_Representations),ActionName).

get_enabled_actions_at_state(CurID,CurState,PGE,EnabledTopLevelActions,Transitions,DisabledTopLevelActions) :-
  get_specfile_possible_trans_names(CurID,CurState,PGE,TopLevelActions,DisabledOperations),
  get_enabled_actions_at_state_aux(TopLevelActions,CurID,CurState,EnabledTopLevelActions,Transitions,DisabledOperationsNew),
  list_to_ord_set(DisabledOperationsNew,DisabledOperationsNewSet),
  ord_union(DisabledOperations,DisabledOperationsNewSet,DisabledTopLevelActions).

get_specfile_possible_trans_names(CurID,CurState,PGE,TopLevelActions,DisabledOperations) :-
  ( PGE\= off ->
    % get enabled top level actoions using the information from the PGE algo
    pge_algo: get_possible_enabled_trans_names(CurID,CurState,DisabledOperations,_EnabledOperations,TopLevelActions)
  ; otherwise ->
    get_specfile_possible_trans_names(CurState,TopLevelActions),
    DisabledOperations = [] % in case PGE is not on we don't have any information about the disabledness of actions
  ).

get_specfile_possible_trans_names(CurState,TopLevelActions) :-
  findall(ActionName,specfile_possible_trans_name(CurState,ActionName),TopLevelActions).

get_enabled_actions_at_state_aux(TopLevelActions,CurID,CurState,EnabledTopLevelActions,Transitions,DisabledTopLevelActionsNew) :-
  get_enabled_actions_at_state_aux2(TopLevelActions,CurID,CurState,[],EnabledTopLevelActions,[],Transitions,[],DisabledTopLevelActionsNew).

get_enabled_actions_at_state_aux2([],_CurID,_CurState,EnabledTopLevelActions,EnabledTopLevelActions,Transitions,Transitions,DisabledNew,DisabledNew).
get_enabled_actions_at_state_aux2([TopLevelAction|R],CurID,CurState,EnabledTopLevelActionsGen,EnabledTopLevelActions,TransitionsGen,Transitions,DisabledNewGen,DisabledNew) :-
  (get_top_level_action_transitions(CurID,CurState,TopLevelAction,ActionTransitionsTuple) ->
    EnabledTopLevelActionsGen1 = [TopLevelAction|EnabledTopLevelActionsGen],
    TransitionsGen1 = [ActionTransitionsTuple|TransitionsGen],
    DisabledNewGen1 = DisabledNewGen
  ; otherwise ->
    EnabledTopLevelActionsGen1 = EnabledTopLevelActionsGen,
    TransitionsGen1 = TransitionsGen,
    DisabledNewGen1 = [TopLevelAction|DisabledNewGen]
  ),
  get_enabled_actions_at_state_aux2(R,CurID,CurState,EnabledTopLevelActionsGen1,EnabledTopLevelActions,TransitionsGen1,Transitions,DisabledNewGen1,DisabledNew).

get_top_level_action_transitions(CurID,CurState,TopLevelAction,ActionTransitionsTuple) :-
  findall((Act,NewExpression,TransInf,Residue),
          specfile_trans_timeout(CurID,CurState,TopLevelAction,Act,NewExpression,TransInf,Residue),
          Transitions),
  % in case TopLevelAction is not enabled we do not save the transitions in the list
  Transitions\=[],
  ActionTransitionsTuple = (TopLevelAction,Transitions).

specfile_trans_timeout(CurID,CurState,ActionName,Act,NewExpression,TransInf,Residue) :-
    (use_no_timeout(CurID) ->
        specfile_trans(CurState,ActionName,Act,NewExpression,TransInf,Residue)
    ;   preferences:preference(time_out,CurTO),
        time_out_with_enum_warning_for_findall(specfile_trans(CurState,ActionName,Act,NewExpression,TransInf,Residue),CurTO,TimeOutRes),
        pge_algo: check_timeout_result(CurID,ActionName,CurTO,TimeOutRes),
        Residue==[]
    ).
/* Auxiliary predicates for getting action names enabled at the curent state */
%---------------------------------------------------------------------------------------------------------------------------------
/************************** DETERMINING POSSIBLE AMPLE SETS (BEGIN) ******************************/

/************************** PREDICATES COMPUTING AMPLE SETS WITH RESPECT TO CONDITIONS (A 2) and (A 3) *************************/

/*

   implementation of the ample-sets approach satisfying the dependency condition (A 2)
   by using enabling and dependence analysis (see module enabling analysis)

*/

get_ample_actions(CurID,AllEnabledActions,POROptions,AmpleSet) :-
    get_preference(por_heur,Pref),
    get_ample_actions_pref(Pref,CurID,AllEnabledActions,POROptions,AmpleSet).

get_ample_actions_pref(minimal,CurID,AllEnabledActions,POROptions,AmpleSet) :- !,
    get_possible_ample_actions_minimal(AllEnabledActions,CurID,AllEnabledActions,POROptions,AmpleSet).
get_ample_actions_pref(Pref,CurID,AllEnabledActions,POROptions,AmpleSet) :-
    get_possible_ample_actions(Pref,AllEnabledActions,CurID,AllEnabledActions,POROptions,AmpleSet).

get_possible_ample_actions(_Pref,[],_CurID,AllEnabledActions,_POROptions,AllEnabledActions).
get_possible_ample_actions(Pref,Actions,CurID,AllEnabledActions,POROptions,AmpleSet) :-
    % choosing of action for computing a possible ample set for CurID should be random
    select_enabled_action_pref(Pref,Actions,Act,RemainedActions),
    (get_valid_ample_set_act(CurID,POROptions,Act,AllEnabledActions,AmpleSet) -> % get ample set satisfying (A 2) and (A 3) i.r.t. Act
      true
    ; otherwise -> % search for another valid ample set
      get_possible_ample_actions(Pref,RemainedActions,CurID,AllEnabledActions,POROptions,AmpleSet)
    ).

get_valid_ample_set_act(CurID,POROptions,Act,AllEnabledActions,AmpleSet) :-
  POROptions = por(ample_sets2,_,_,_),!,
  %% trying to find an ample set by the closure method
  get_possible_ample_set_closure(CurID,Act,AllEnabledActions,AmpleSet).
get_valid_ample_set_act(CurID,POROptions,Act,AllEnabledActions,AmpleSet) :-
  get_valid_ample_set_act_normal(CurID,POROptions,Act,AllEnabledActions,AmpleSet).

get_valid_ample_set_act_normal(CurID,POROptions,Act,AllEnabledActions,AmpleSet) :-
    % finding a set (DepActs) satisfying (A 2.1) // there is only one possible set, do not recall here
    get_dependent_actions(CurID,POROptions,Act,AllEnabledActions,DepActs,IndActs),!,
    ( (is_enabling_dependency_condition_satisfied(IndActs,DepActs,AllEnabledActions,CurID,POROptions), % check whether DepActs satisfies (A 2.2)
       check_stutter_condition(DepActs,AllEnabledActions)) % checking (A 3)
      -> AmpleSet = DepActs
    ; fail). % should fail in order to try to get another valid ample set

select_enabled_action_pref(random,Actions,Act,RemainedActions):- !,
  random_select(Act,Actions,RemainedActions).
select_enabled_action_pref(first,[Act|RemainedActions],Act,RemainedActions):- !.
select_enabled_action_pref(Pref,Actions,Act,RemainedActions) :-
  add_warning(ample_sets,'Unknown preference for choosing an action from the set of enabled actions: ', Pref),
  random_select(Act,Actions,RemainedActions). % choose the default option for choosing an action

check_stutter_condition(AmpleSet,AllEnabledActions) :-
    \+ ord_seteq(AmpleSet,AllEnabledActions),!,
    check_cycle_proviso(AmpleSet).

check_cycle_proviso(AmpleSet) :-
    % in case that we don't check for invariant violations, we don't have to prove the stuttering condition for AmpleSet
    (check_cycle_proviso ->
      stutter_actions(AmpleSet)
    ; true).

get_possible_ample_actions_minimal(AllEnabledActions,CurID,AllEnabledActions,POROptions,AmpleSet) :-
    maplist(get_valid_ample_set_act_nonfail(CurID,AllEnabledActions,POROptions),AllEnabledActions,AmpleSets),
    get_minimal_set(AmpleSets,AmpleSet),
    debug_println(9,minimal_ample_set(AmpleSet,AmpleSets)).

get_minimal_set([AmpleSet|Rest],MinAmpleSet) :-
    length(AmpleSet,Length),
    get_minimal_set_length(Length,AmpleSet,Rest,MinAmpleSet).

get_minimal_set_length(1,AmpleSet,_Rest,AmpleSet).
get_minimal_set_length(_Length,AmpleSet,[],AmpleSet).
get_minimal_set_length(Length,AmpleSet,[AmpleSet2|Rest],MinAmpleSet) :-
    length(AmpleSet2,Length2),
    (Length2<Length ->
      get_minimal_set_length(Length2,AmpleSet2,Rest,MinAmpleSet)
    ; otherwise ->
      get_minimal_set_length(Length,AmpleSet,Rest,MinAmpleSet)).

get_valid_ample_set_act_nonfail(CurID,AllEnabledActions,POROptions,Act,AmpleSet) :-
    (get_valid_ample_set_act(CurID,POROptions,Act,AllEnabledActions,AmpleSet)
      -> true
      ; AmpleSet = AllEnabledActions).

%%%%%% the closure method for computing ample sets %%%%%%
get_possible_ample_set_closure(CurID,Act,AllEnabledActions,AmpleSet) :-
  get_valid_ample_set_act_closure(CurID,Act,AllEnabledActions,AmpleSet),!,
  check_stutter_condition(AmpleSet,AllEnabledActions). % checking (A 3)

get_valid_ample_set_act_closure(CurID,Act,AllEnabledActions,AmpleSet) :-
  %(length(AllEnabledActions,3) -> trace;true),
  list_to_ord_set([Act],WorkSet),
  list_to_ord_set([],ClosureSet),
  compute_closure_set(WorkSet,CurID,ClosureSet,AllEnabledActions,Result),
  ord_intersection(Result,AllEnabledActions,AmpleSet).

compute_closure_set([],_CurID,ClosureSet,_AllEnabledActions,ClosureSet).
compute_closure_set([Act|Acts],CurID,ClosureSet,AllEnabledActions,Result) :-
  ord_add_element(ClosureSet,Act,ClosureSet1),
  get_new_work_set(Act,Acts,CurID,AllEnabledActions,ClosureSet1,NewWorkSet),
  compute_closure_set(NewWorkSet,CurID,ClosureSet1,AllEnabledActions,Result).

get_new_work_set(Act,WorkSet,CurID,AllEnabledActions,ClosureSet1,NewWorkSet) :-
  ord_member(Act,AllEnabledActions),!,
  get_dependency_set_closure(Act,WorkSet,CurID,ClosureSet1,NewWorkSet).
get_new_work_set(Act,WorkSet,CurID,AllEnabledActions,ClosureSet1,NewWorkSet) :-
  get_enabling_set_for_act(Act,WorkSet,CurID,AllEnabledActions,ClosureSet1,NewWorkSet).

get_dependency_set_closure(Act,WorkSet,CurID,ClosureSet1,NewWorkSet) :-
  findall(Act1,(is_dependent_to_and_coenabled(Act,Act1,CurID),ord_nonmember(Act1,ClosureSet1)),NewActions),
  list_to_ord_set(NewActions,NewActionsSet),
  ord_union(WorkSet,NewActionsSet,NewWorkSet).

get_enabling_set_for_act(Act,WorkSet,CurID,AllEnabledActions,ClosureSet1,NewWorkSet) :-
  % Act is disabled at current state
  findall(Act1,find_enabling_action(Act,CurID,AllEnabledActions,ClosureSet1,Act1),NewActions),
  list_to_ord_set(NewActions,NewActionsSet),
  ord_union(WorkSet,NewActionsSet,NewWorkSet).

find_enabling_action(Act,CurID,AllEnabledActions,ClosureSet,Act1) :-
  member(Act1,AllEnabledActions),
  ord_nonmember(Act1,ClosureSet),
  can_enable_certain_action(Act1,Act,CurID).

can_enable_certain_action(Act,ActionToBeEnabled,CurID) :-
  ( enables(Act,ActionToBeEnabled) ->
    debug_println(9,enables(Act,ActionToBeEnabled))
  ; otherwise ->
    enable_graph(EnableGraph),
    min_path(Act,ActionToBeEnabled,EnableGraph,P,L),
    debug_print_path(CurID,P,L),
    assert(enables(Act,ActionToBeEnabled))).

% get_dependent_actions(+CurID,+POROptions,+Act,+AllEnabledActions,+EnGraph,-DepActs)
% CurID - the ID of the current state (used here only for debugging purposes)
% POROptions - partial order reduction options (e.g. UseEnableGraph, Depth)
% Act - the action with which the current dependency class will be computed
% AllEnabledActions - the set of all in CurID enabled actions and it is an ordinary set
% DepActs - the computed candidate for ample set
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% EnableGraph - the enable graph for all possible enable actions' relations in the current machine. The enable graph will
%               be computed only single time (before starting with the verification of the state space).

get_dependent_actions(CurID,_POROptions,Act,AllEnabledActions,DepActs,IndActions) :-
    % finding all dependent actions to Act (A 2.1)
    get_dependency_set(Act,AllEnabledActions,CurID,DepActs),
    ord_subtract(AllEnabledActions,DepActs,IndActions).

is_enabling_dependency_condition_satisfied([],_DepActs,_AllEnabledActions,_CurID,_POROptions) :- !,fail.
is_enabling_dependency_condition_satisfied(IndActs,DepActs,AllEnabledActions,CurID,POROptions) :-
  enable_graph(EnableGraph),
  vertices(EnableGraph,Vertices), % getting all possible actions that can be enabled
  % DisabledActions = Vertices\AllEnabledActions
  ord_subtract(Vertices,AllEnabledActions,DisabledActions),
  % searching for an independent action to Act that could enable a currently disabled action that is dependent to Act (A 2.2)
  \+may_enable_dep_action_list(IndActs,DepActs,DisabledActions,EnableGraph,CurID,POROptions).

may_enable_dep_action_list(IndActs,DepActs,DisabledActions,EnableGraph,CurID,POROptions) :-
  member(Act2,IndActs),
  may_enable_dep_action(Act2,DepActs,DisabledActions,EnableGraph,CurID,POROptions).


get_dependency_set(Act,AllEnabledActions,CurID,DepActions_so_far) :-
  vertices_edges_to_ugraph([],[],G),
  get_dependency_tuples_converted_to_edges(Act,AllEnabledActions,RestEnabledActions,CurID,Edges),
  add_edges(G,Edges,G1),
  get_dependency_set1(Act,G1,[],RestEnabledActions,CurID,DepActions_so_far).

get_dependency_tuples_converted_to_edges(Act,AllEnabledActions,RestEnabledActions,CurID,Res) :-
  findall(Act1,(member(Act1,AllEnabledActions), is_dependent_to(Act,Act1,CurID)),DepActs),
  findall([Act-Act1,Act1-Act],member(Act1,DepActs),Edges),
  ord_subtract(AllEnabledActions,DepActs,RestEnabledActions),
  append(Edges,Res).

get_dependency_set1(Act,G,G,_Actions,_CurID,Res) :-
  reachable(Act,G,Res).
get_dependency_set1(Act,G,_GPrime,Actions,CurID,Res) :-
  GPrime1 = G,
  vertices(G,Vertices),
  findall([V-Act1,Act1-V],(member(Act1,Actions),member(V,Vertices),is_dependent_to(V,Act1,CurID)),NewEdges),
  append(NewEdges,NewEdges1),
  add_edges(G,NewEdges1,G1),vertices(G1,Vertices1),
  ord_subtract(Actions,Vertices1,Actions1),
  get_dependency_set1(Act,G1,GPrime1,Actions1,CurID,Res).

is_dependent_to(Act,Act1,CurID) :-
  (dependent_act(Act,Act1,Status,_Coenabled) ->
    (Status=predicate(Pred) -> check_enable_dependency(Act,Act1,Pred,CurID); true)
  ; otherwise ->
    fail
  ).

is_dependent_to_and_coenabled(Act,Act1,CurID) :-
  dependent_act(Act,Act1,Status,true),
  (Status=predicate(Pred) -> check_enable_dependency(Act,Act1,Pred,CurID); true).

check_enable_dependency(Act,Act1,Pred,CurID) :-
  debug_println(9,checking_guard_dependency(Act,Act1,Pred)),
  visited_expression(CurID,RawState),
  state_corresponds_to_initialised_b_machine(RawState,State),
  empty_state(LocalState),
  % if the enable predicate Pred evaluates to false we know that (Act,Act1) is a dependent pair
  \+enable_graph: b_test_boolean_expression_wf(Pred,LocalState,State).

/* searching for enabling path beginning from an action which is not from AmpleSet and
   which enables action dependent on AmpleSet */
may_enable_dep_action(Act,AmpleSet,DisabledActions,EnableGraph,CurID,por(_TYPE,UseEnableGraph,Depth,_PGE)) :-
  ( UseEnableGraph = true ->
    check_enabling_path(CurID,Act,DisabledActions,AmpleSet,EnableGraph,Depth)
  ; otherwise ->
    member(DisAct,DisabledActions),
    % we check here only currently disabled actions dependent on AmpleSet that can be enabled later
    depends_on_ample_set_and_coenabled(DisAct,AmpleSet,_DepAct),
    % is there a path that can enable Act1, which is dependent on Act
    enables_action(CurID,Act,DisAct,AmpleSet,EnableGraph)
  ),
  !.

enables_action(CurID,Act,DisAct,AmpleSet,EnableGraph) :-
  ( enables(Act,DisAct) ->
    % should improve performance a little bit (not applicable when using enable graphs)
    debug_println(9,enables(Act,DisAct))
  ; check_if_enabled(CurID,Act,DisAct,AmpleSet,EnableGraph) ->
    true
  ; otherwise ->
    fail
  ).

check_if_enabled(CurID,Act,DisAct,AmpleSet,EnableGraph) :-
  min_path(Act,DisAct,EnableGraph,P,L),
  sets_disjoint(AmpleSet,P),
  debug_print_path(CurID,P,L),
  assert(enables(Act,DisAct)).

% predicates used for the enable graph feature
check_enabling_path(CurID,Act,DisabledActions,AmpleSet,EnableGraph,Depth) :-
  findall(DepDisAct,(member(DepDisAct,DisabledActions),depends_on_ample_set_and_coenabled(DepDisAct,AmpleSet,_Act)),DependentDisabledActions),
  ( DependentDisabledActions = [] ->
    fail % in this case we know that no action dependent on AmpleSet can be enabled after executing Act
  ; otherwise ->
    enables_pred_action1(Act,DependentDisabledActions,AmpleSet,CurID,EnableGraph,Depth)).

enables_pred_action1(Act,DepDisActs,AmpleSet,CurID,EnableGraph,Depth) :-
    % visited_expression/2 and state_corresponds_to_initialised_b_machine/2 should be called once for each state,
    % avoid multiple calls per state
    (Depth==0 ->
       empty_state(ST)
    ; otherwise ->
       visited_expression(CurID,RawState),
       state_corresponds_to_initialised_b_machine(RawState,ST)
    ),
    %Pred =.. [evaluate_enabling_predicate_for_por, ST, AmpleSet],
    enable_graph: path_set(EnableGraph,Act,DepDisActs,enable_graph,
                           evaluate_enabling_predicate_for_por(ST,AmpleSet),Depth,Path,Length),
    debug_print_path(CurID,Path,Length).

% guaranteeing that the dependent action to AmpleSet has been enabled from an action off AmpleSet
sets_disjoint(AmpleSet,Path) :-
    list_to_ord_set(Path,OrdSet),
    \+ord_intersection(AmpleSet,OrdSet).

depends_on_ample_set_and_coenabled(Act,AmpleSet,Act1) :-
  member(Act1,AmpleSet),
  dependent_act(Act,Act1,_Status,Coenabled),
  Coenabled==true.

debug_print_path(CurID,Path,Length) :-
   debug_println(20,'***** Enable Path *****'),
   debug_println(20,path(CurID,Path,Length)),
   debug_println(20,'***********************').

/************************** DETERMINING POSSIBLE AMPLE SETS (END) ******************************/

/************************** CHECKING IF ALL ACTIONS FROM AMPLE SET ARE STUTTER ACTIONS (BEGIN) WITH REGARD TO THE INVARIANT ******************************/
/************************** PREDICATES WHICH SATISFY CONDITION (A3) ******************************/

stutter_actions(Actions) :-
    maplist(is_stutter_action,Actions).

% visible actions are non stutter actions
is_stutter_action(Act) :-
   ( stutter_action(Act) ->
       debug_println(9,stutter_action(Act))
   ; visible_action(Act) ->
       debug_println(9,visible_action(Act)),
       fail
   ; otherwise ->
       check_if_action_is_stutter_wrt_inv(Act)
   ).

% A stutter action cannot change the current invariant value.
% That's why we must check if the particular action preserves
% the full invariant in order to determine if the action is
% stutter under the following LTL formula:
% G Phi, where Phi is the invariant.
check_if_action_is_stutter_wrt_inv(Act) :-
    b_operation_preserves_full_invariant(Act),!,
    assert(stutter_action(Act)),
    debug_println(9,stutter_according_to_proof_info(Act)).
check_if_action_is_stutter_wrt_inv(Act) :-
   b_get_invariant_from_machine(Invariant),
   conjunction_to_list(Invariant,Conjunctions), % getting all conjunctions from the invariant
   b_get_read_write(Act,_Reads,Writes),
   ( action_does_not_modify_important_variables_in_the_inv(Writes,Conjunctions)
     -> assert(stutter_action(Act)),
        debug_println(9,stutter_according_to_read_write_info(Act))
   ; otherwise
     -> assert(visible_action(Act)),fail
   ).

action_does_not_modify_important_variables_in_the_inv([],_Conjuncions).
action_does_not_modify_important_variables_in_the_inv([Var|WrVars],Conjunctions) :-
   var_not_changing_visible_condition_in_the_inv(Var,Conjunctions),
   action_does_not_modify_important_variables_in_the_inv(WrVars,Conjunctions).

var_not_changing_visible_condition_in_the_inv(_Var,[]).
var_not_changing_visible_condition_in_the_inv(Var,[Conj|Conjs]) :-
   ( occurs_in_expr(Var,Conj)
     -> (\+ trivial_predicate(Var,Conj) -> fail; true)
   ; otherwise -> true
   ),
   var_not_changing_visible_condition_in_the_inv(Var,Conjs).

trivial_predicate(Var,Conj) :-
   \+var(Conj),
   trivial_member_predicate(Var,Conj).

/*

   TODO: Define formal when an action in Event-B/B can be seen as a stutter action.
   Maybe we should use here the predicate b_specialized_invariant_for_op/2 from bmachine.pl.

*/

/*trivial_member_predicate(Var,Conj) :-
   Conj=b(member(b(identifier(Var),integer,_),b(integer_set('NATURAL'),set(integer),_)),pred,_).
trivial_member_predicate(Var,Conj) :-
   Conj=b(member(b(identifier(Var),integer,_),b(integer_set('NAT'),set(integer),_)),pred,_).*/
trivial_member_predicate(Var,Conj) :-
   Conj=b(member(b(identifier(Var),integer,_),b(integer_set('INT'),set(integer),_)),pred,_).
trivial_member_predicate(Var,Conj) :-
   Conj=b(member(b(identifier(Var),integer,_),b(integer_set('INTEGER'),set(integer),_)),pred,_).

/************************** CHECKING IF ALL ACTIONS FROM AMPLE SET ARE STUTTER ACTIONS (END) ******************************/
