% (c) 2009-2020 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(prob_socketserver, [start/1, start_prob_socketserver/2]).

:- use_module(module_information,[module_info/2]).
:- module_info(group,cli).
:- module_info(description,'The socket server is used to interface to Java and other languages.').

:- use_module(version).
:- use_module(eclipse_interface).
:- use_module(prob2_interface).
:- use_module(extension('user_signal/user_signal'),[get_user_signal_ref/1,
                                                    user_interruptable_call_det/2,
                                                    ignore_user_interrupt_det/1]).

:- use_module(logger, [writeln_log/1, get_log_file/1,
                              logging_is_enabled/0]).
:- use_module(library(sockets)).
:- use_module(library(lists)).

:- op(1150, fx, type).

%prolog_socket :- start(_).

:- dynamic socket_opened/1.

start(Port) :- start_prob_socketserver(Port,true).

start_prob_socketserver(Port,Loopback) :-
	on_exception(E, ignore_user_interrupt_det(start2(Port,Loopback)),
	    ( E=halt(C) -> throw(halt(C))
        ; otherwise ->
            my_format_error("Exception in Socket Server: ~w~n",[E]),fail)),
	!.
start_prob_socketserver(Port,Loopback) :-
   my_format_error("Listening to Port failed: ~w (loopback=~w)~n",[Port,Loopback]),
   retract(socket_opened(Socket)),
   my_format("Attempting to close socket: ~w~n",[Socket]),
   socket_server_close_sp4(Socket),
   my_format("Socket closed: ~w~n",[Socket]).


socket_server_open_sp4(Port,Socket,Loopback) :-
     (socket_server_open(Port,Socket,[loopback(Loopback)]) % was true
      -> true
	     ; my_format_error("Unable to bind sp4 socket: ~w to localhost:~w.~n",[Socket,Port]),fail).

% Page 725, Chapter 10 of SICStus Manual for loopback(Bool) option:
% If true then the nodename will be ignored and the socket will only listen to connection
% from the loopback device, i.e. the local machine.

socket_server_close_sp4(Socket) :- socket_server_close(Socket).

socket_accept_sp4(Socket, Stream) :-
   socket_server_accept(Socket, Client, Stream, [type(text),encoding('UTF-8')]),
   my_format("Connected: ~w~n",[Client]).

start2(Port,Loopback) :-
	% print(socket('AF_INET', Socket)),nl,
	socket_server_open_sp4(Port,Socket,Loopback),
    assert(socket_opened(Socket)),
	%open(prologport, write,Stream),
	%format(Stream,'~w~n',[Port]),
	%close(Stream),
	prolog_flag(user_output, Stdout, Stdout),
	format(Stdout,'Port: ~w~n', [Port]), % ProB2 matches against this in PortPattern.java
    writeln_log(port(Port)),
    revision(Revision),
    format(Stdout,'probcli revision: ~w~n',[Revision]),
    get_user_signal_ref(Ref),
    format(Stdout,'user interrupt reference id: ~w~n',[Ref]),  % ProB2 matches against this in InterruptRefPattern.java
    format(Stdout,'-- starting command loop --~n', []), % ProB2 matches against this in CliStarter.java
	flush_output(Stdout),
	wait(Socket),
	socket_server_close_sp4(Socket),!,
    my_format("Socket closed: ~w~n",[Socket]),
    writeln_log(socket_closed(Socket)).


%% Wait for a connection on the given socket
wait(Socket) :-
	socket_accept_sp4(Socket, Stream),
	read_socket(Socket, Stream).

% adding meta_predicate makes ProB2 fail:
%:- meta_predicate make_call(-,0,-).
%:- meta_predicate do_call(0,-,-).

%% We have a connection so just wait for valid terms, and return either
%% OK X=val\000Y=val1
%% FAIL\000
%% ERROR SHORT long message here\000
read_socket(Socket,Stream) :-
   %print(listening_on_socket(Socket)),nl, flush_output(user_output),
	on_exception(E,
		     read_term(Stream, Term, [variable_names(Vars)]),
		     (portray_clause(user_output,goal_parse_error),Parse=fail,exception(E,Stream))),

	%print(received_term(Term)),nl, flush_output(user_output),
	(Parse=ok ->
	      ((Term == end_of_file ; Term==halt) ->
		       socket_server_close_sp4(Socket),
		       throw(halt(0))
	      ;  %portray_clause(user_output,calling(Term)),
		       make_call(Stream, Term, Vars)
		       %portray_clause(user_output,exit(Term))
	      )
	;
	      %% Parse failed, but we have already written error message
	      %print(an_exception_occurred(E)),nl,
	      portray_clause(user_output, an_exception_occurred(E))
	),
	%print(done(Term)),nl,
	%flush_output(user_output),
	%flush_output,
	%put_code(Stream, 1),

	flush_output(Stream),
	flush_output(user_output),
	read_socket(Socket, Stream).


%% We have a valid term so make the call
make_call(Stream,Term, Vars) :-
	% write(make_call(Term)),nl,flush_output(user_output),
	on_exception(CallEx,
		       do_call(Term,Vars,Result),
	           (portray_clause(user_output, exception(make_call/3, CallEx)),
              writeln_log(exception(CallEx)),
		          Result=exception(CallEx)
	           )),
	write_canonical(Stream,Result), %writeln_log(result(Result)),
	nl(Stream), put_code(Stream,1),
	flush_output(Stream).

do_call(Call,Vars,ResOut) :- %print(do_call(Call)),nl,flush_output(user_output),
	   writeln_logtime_call_start(Call,Timer), %logger:writeln_log_time(do_call(Call)),
     (white_list(Call)
       -> (do_not_interrupt_this_call(Call)
           ->  call(Call),
               IRes = ok
            ;  user_interruptable_call_det(call(Call),IRes)
           ),
           (IRes = interrupted -> ResOut = interrupted ; ResOut = yes(Vars))
       ;  my_format_error('Call not on whitelist: ~w.\n',[Call]),fail
     ),
     ((ground(Vars);Call=debug_console(_,_);ResOut = interrupted)
	 % print(success(Call)),nl,flush_output(user_output),
	  -> writeln_logtime_call_end(Call,Timer)
	  %-> logger:writeln_log_time(success(Call))
	  ; raise_exception(result_is_not_ground)),
%     \+ next_char_is_semicolon(Stream), /*Otherwise force backtracking */
     !.
do_call(_Call,_Vars,no) :-
     writeln_log(no).

:- dynamic exported_by_eclipse_interface/2.
white_list(halt). % needed to stop ProB
white_list(true). % to prevent fails because of empty queries
white_list((A,B)) :- white_list(A),white_list(B). % concatenation used to send several commands at once
white_list(C) :- nonvar(C),functor(C,Functor,Arity), % all public predicates of the module
    is_exported_by_eclipse_interface(Functor,Arity). % eclipse_interface can be used
is_exported_by_eclipse_interface(Functor,Arity) :-
    exported_by_eclipse_interface(Functor,Arity),!.
is_exported_by_eclipse_interface(Functor,Arity) :-
    % check that the list has not already been generated
    \+ exported_by_eclipse_interface(_,_),
    % generate the list
    create_exported_by_eclipse_interface_list,
    % do the check again
    exported_by_eclipse_interface(Functor,Arity).
create_exported_by_eclipse_interface_list :-
    % fail-driven loop over the predicates imported from eclipse_interface
    predicate_property(prob_socketserver:P,imported_from(eclipse_interface)),
    functor(P,Functor,Arity),
    assert(exported_by_eclipse_interface(Functor,Arity)),
    fail.
create_exported_by_eclipse_interface_list :-
    % fail-driven loop over the predicates imported from prob2_interface
    predicate_property(prob_socketserver:P,imported_from(prob2_interface)),
    functor(P,Functor,Arity),
    assert(exported_by_eclipse_interface(Functor,Arity)),
    fail.
create_exported_by_eclipse_interface_list.


do_not_interrupt_this_call((A,B)) :-
   (do_not_interrupt_this_call(A) ; do_not_interrupt_this_call(B)).
do_not_interrupt_this_call(do_modelchecking(_,_,_,_,_)). % interrupting prob can lead to spurious deadlocks at the moment

%next_char_is_semicolon(Stream) :- %print(peeking(Stream)),nl,
%    peek_code(Stream,X),!,
%    ((X=10;X=13)
%      -> (get_code(Stream,_),next_char_is_semicolon(Stream))
%      ; (X=59,get_code(Stream,_))).


get_stream_info(S,Info) :-
	on_exception(E,get_stream_info1(S,Info),(portray_clause(user_output, exception(E)))).

get_stream_info1(S,blank):-
	portray_clause(user_output, stream_is(S)),
	%true.
	fail.
get_stream_info1(S, Info) :-
	stream_property(S,file_name(Info)),!.
get_stream_info1(S, Info) :-
	stream_property(S,alias(Info)),
	!.
get_stream_info1(_S, unknown).


tokens_to_string([],[]).
tokens_to_string([A|As], [B|Bs]) :-
	token_to_string1(A,B),
	tokens_to_string(As,Bs).

token_to_string1(atom(A)-_, A) :- !.
token_to_string1(var(_,S,_)-_, A) :- !,
	atom_codes(A,S).
token_to_string1(A, A).


%% Exception handlers
%% First Argument is Sicstus exception, second is verbose message for user
exception(E,_Stream) :-
 portray_clause(user_output,E),fail.


exception(syntax_error(Call, Pos, Msg, Tokens,Lineend),Stream) :-
	!,
	portray_clause(user_output,syntax_stream(Call)),
	(Call = read_term(Err,_,_) ->
	    get_stream_info(Err,Info)
	;
	    Info = unknown_stream
	),
	tokens_to_string(Tokens,Where),

	my_format_error(Stream,'exception(ERR-~@PARSE~@~w in ~w\nlines::~w-~w\nCheck ~w)\n', [put_code(1),put_code(1),Msg,Info,Pos,Lineend,Where]),
	put_code(Stream, 1).

exception(existence_error(_, _, _, _, past_end_of_stream),Stream) :-
	!,
	my_format_error(Stream,'exception(ERR-~@EXISTENCE~@Attempted to read past end of stream)', [put_code(1),put_code(1)]),
	put_code(Stream, 1)	.

exception(existence_error(_, _,procedure,PredSpec,_),Stream) :-
	!,
	my_format_error(Stream,'exception(ERR-~@EXISTENCE~@Unable to call predicate ~w)', [put_code(1),put_code(1),PredSpec]),
	put_code(Stream, 1)	.

exception(existence_error(_, _,file,PredSpec,_),Stream) :-
	!,
	my_format_error(Stream,'exception(ERR-~@EXISTENCE~@Unable to open file ~w)', [put_code(1),put_code(1),PredSpec]),
	put_code(Stream, 1)	.

exception(E,Stream) :-
	my_format_error(Stream,'exception(ERR-~@UNKNOWN~@~w)', [put_code(1),put_code(1),E]),
	put_code(Stream, 1),
	portray_clause(user_output,unknown_exception(E)).



my_format(FS,Args) :- my_format(user_output,FS,Args).
my_format_error(FS,Args) :- my_format(user_error,FS,Args).
my_format_error(Stream,FS,Args) :- my_format(Stream,FS,Args).

my_format(Stream,FS,Args) :- format(Stream,FS,Args),
    (get_log_file(F)
      ->  open(F,append,S),
          format(S,FS,Args),
          close(S)
     ; true
    ).

writeln_logtime_call_start(Term,timer(Time,WTime)) :-
    ((logging_is_enabled ;  runtime_profile_available)  ->
      statistics(runtime,[Time,_]),
      statistics(walltime,[WTime,_]),
      (logging_is_enabled
        -> get_functors(Term,Fs),
           writeln_log(start_call(walltime(WTime),Fs))
        ;  true)
    ; true).

:- use_module(runtime_profiler,[register_profiler_runtime/4, runtime_profile_available/0]).
writeln_logtime_call_end(Term,timer(Time,WTime)) :-
    %format(user_output,'End for ~w~n',[Time]),
    ((logging_is_enabled ;  runtime_profile_available) ->
      statistics(runtime,[Time2,_]), RT is Time2-Time,
      statistics(walltime,[WTime2,_]), WT is WTime2-WTime,
      get_functors(Term,Fs),
      (logging_is_enabled
        -> statistics(memory_used,M), MB is M  / 1000000, % used instead of deprecated 1048576
           writeln_log(end___call(walltime(WTime2),delta_rt_wt(RT,WT),mb_used(MB),Fs))
        ;  true),
      register_profiler_runtime(Fs,unknown,RT,WT)
    ; true).

get_functors(Var,V) :- var(Var),!, V='_'.
get_functors((A,B),(FA,FB)) :- !, get_functors(A,FA), get_functors(B,FB).
%get_functors(A,F/list(Len)) :- functor(A,F,1), arg(1,A,Ls), nonvar(Ls), Ls = [_|_],
%    ground(Ls), length(Ls,Len),!.
get_functors(A,F/N) :- functor(A,F,N).

%runtime_entry(start) :- go_cli.

%save :- save_program('probcli.sav').
