% (c) 2009-2020 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(z_typechecker, [
                          %z_type/4,
                          global_environment/3,
                          extract_btypes/3,
                          internal2b/2,
                          %remove_typeinfos/2,
                          ti_expr/2,
                          ti_type/2,
                          zfields2b/3,
                          remove_fields_by_name/3
                          %unify_generic/4
                          ]).

:- use_module(library(lists)).

:- use_module(probsrc(error_manager),[add_error/3,add_warning/3]).
:- use_module(probsrc(tools)).

:- use_module(probsrc(module_information)).
:- module_info(group,proz).
:- module_info(description,'This module contains a type checker for Z specifications, mainly because the types are needed for the translation to B.').

:- use_module(fuzzfile).
:- use_module(proz).
:- use_module(schemavars).
:- use_module(z_tools).
:- use_module(subexpressions).
:- use_module(zparameters).
:- use_module(zenvironment).

%********************************************************************************
% Type checking expressions
z_type([],_,[],[]) :- !.
z_type([Expr|Rest],Env,[Type|TRest],[AExpr|ARest]) :-
    !,z_type(Expr,Env,Type,AExpr),
    z_type(Rest,Env,TRest,ARest).
z_type(Expr,Env,Type,ti(AnnotatedExpr,Type)) :-
    if(z_type2(Expr,Env,AnnotatedExpr,Type),true,
       (add_warning(z_typechecker,'Typing expression failed: ',Expr),
        %trace, z_type(Expr,Env,Type,ti(AnnotatedExpr,Type)),
        fail)).

ti_expr(ti(Expr,_),Expr).
ti_type(ti(_,Type),Type).

update_params(name('\\emptyset',_),[_],Params) :- !, Params=[]. % remove now useless parameter of \emptyset; otherwise Z consistency check will complain as currently the Param of \emptyset is not typed itself (I don't know why; probably because \emptyset is registered as a 0-ary construct)
update_params(_,P,P).

z_type2(ref(Name,Params),Env,ref(Name,NewParams),Type) :-
    ( lookup_type(Env,Name,T) ->
        ( nonvar(T),T=generic(GenParams,GenType) ->
            prepare_generic(Params,Env,GenType,GenParams,Type)
        ; Type=T -> true
        ;  otherwise -> add_warning(proz,'Type mismatch: ',Name:Type:T),fail
        ),
        update_params(Name,Params,NewParams)
    ; otherwise ->
        add_error(proz,'Internal error: Unknown type of ', Name), fail).

z_type2(number(N),_,number(N),num).

z_type2(apply(Ref,Arg),Env,apply(RefTI,ArgTI),ResultType) :-
    z_type(Ref,Env,FunType,RefTI),
    zrel(ArgType,ResultType,FunType),
    z_type(Arg,Env,ArgType,ArgTI).

z_type2(inop(Name,A,B),Env,inop(Name,Ati,Bti),Type) :-
    ti_expr(ApplyTI,apply(_,ArgTI)),
    ti_expr(ArgTI,tuple([Ati,Bti])),
    z_type(apply(ref(Name,[]),tuple([A,B])),Env,Type,ApplyTI).
z_type2(ingen(Name,A,B),Env,ingen(Name,Ati,Bti),Type) :-
    ti_expr(InopTI,inop(_,Ati,Bti)),
    z_type(inop(Name,A,B),Env,Type,InopTI).

z_type2(postop(Name,Arg),Env,postop(Name,ArgTI),Type) :-
    ti_expr(ApplyTI,apply(_,ArgTI)),
    z_type(apply(ref(Name,[]),Arg),Env,Type,ApplyTI).
z_type2(pregen(Name,Arg),Env,pregen(Name,ArgTI),Type) :-
    z_type2(postop(Name,Arg),Env,postop(_,ArgTI),Type).

z_type2(inrel(Name,A,B),Env,inrel(Name,Ati,Bti),Type) :-
    ti_expr(MemTI,member(TupleTI,_)),
    ti_expr(TupleTI,tuple([Ati,Bti])),
    z_type(member(tuple([A,B]),ref(Name,[])),Env,Type,MemTI).
z_type2(prerel(Name,Arg),Env,prerel(Name,ArgTI),Type) :-
    ti_expr(MemTI,member(ArgTI,_)),
    z_type(member(Arg,ref(Name,[])),Env,Type,MemTI).

z_type2(equal(A,B), Env, equal(Ati,Bti), bool) :-
    z_types_common([A,B],Env,_,[Ati,Bti]).
z_type2(member(A,B), Env, member(Ati,Bti), bool) :-
    z_type(A,Env,Type,Ati),
    z_type(B,Env,set(Type),Bti).
z_type2(power(A), Env, power(Ati),set(Type)) :-
    z_type(A,Env,Type,Ati).

z_type2(cross(Args), Env, cross(TypeInfos), set(tuple(Types))) :-
    z_type_sets(Args, Env, Types, TypeInfos).
z_type2(tuple(Args), Env, tuple(TypeInfos), tuple(Types)) :-
    z_type_l(Args, Env, Types,TypeInfos).
z_type2(if(If,Then,Else), Env, if(Ifti,Thenti,Elseti), Type) :-
    z_type(If,Env,bool,Ifti),
    z_type(Then,Env,Type,Thenti),
    z_type(Else,Env,Type,Elseti).

z_type2(ext(Expressions), Env, ext(TypeInfos), set(Type)) :-
    z_types_common(Expressions,Env,Type,TypeInfos).
z_type2(seq(Expressions), Env, seq(TypeInfos), Type) :-
    zseq(IType,Type),
    z_types_common(Expressions,Env,IType,TypeInfos).
z_type2(bag(Expressions), Env, bag(TypeInfos), set(tuple([Type,num]))) :-
    z_types_common(Expressions,Env,Type,TypeInfos).

z_type2(theta(Name,Deco,Renamings), Env, theta(Name,Deco,Renamings), Type) :-
    z_type(sref(Name,'',[],[]),Env,set(Type),_).

z_type2(sexpr(Sexpr), Env, sexpr(SexprTi), Type) :-
    z_type(Sexpr,Env,Type,SexprTi).
z_type2(select(Record,Field), Env, select(RecordTi,Field), Type) :-
    z_type(Record,Env,schema(Fields),RecordTi),
    type_of_field(Fields,Field,Type).

z_type2(lambda(Body,Expr), Env, lambda(BodyTi,ExprTi), Type) :-
    Body = body(Decls,_),
    declaration_tuple(Decls,Env,Tuple),
    ti_expr(TypeInfo,comp(BodyTi,JustTi)),
    ti_expr(InopTi,inop(_,_,ExprTi)),
    ti_expr(JustTi,just(InopTi)),
    z_type(comp(Body,just(inop(name('\\mapsto',''),Tuple,Expr))),
           Env,Type,TypeInfo).
z_type2(comp(Body,Expr), Env, comp(BodyTi,ExprTi), set(Type)) :-
    ti_expr(TypeInfo,mu(BodyTi,ExprTi)),
    z_type(mu(Body,Expr),Env,Type,TypeInfo).
z_type2(mu(Body,MuExpr), Env, mu(BodyTi,ExprTypeInfo), Type) :-
    ( MuExpr = nothing ->
        ExprTypeInfo = ti(nothing,none),
        Body = body(Decls,_),declaration_tuple(Decls,Env,Tuple),
        ti_expr(TypeInfo,mu(BodyTi,_)),
        z_type(mu(Body,just(Tuple)), Env, Type, TypeInfo)
    ; MuExpr = just(Expr) ->
        z_type_body(Body,Env,_,BodyTi,Subenv),
        z_type(just(Expr),Subenv,_,ExprTypeInfo),
        ti_expr(ExprTypeInfo,just(ExprTi)),
        ti_type(ExprTi,Type) ).
z_type2(letexpr(Letdefs,Expr), Env, letexpr(Lettis,ExprTi), Type) :-
    types_in_let(Letdefs,Env,LetTypes,Lettis),
    create_subenv(Env,LetTypes,LetEnv),
    z_type(Expr,LetEnv,Type,ExprTi).

z_type2(nothing,_,nothing,none).
z_type2(just(Expr),Env,just(ExprTi),none) :-
    z_type(Expr,Env,_,ExprTi).


%********************************************************************************
% Type checking Predicates

z_type2(falsity, _, falsity, bool).
z_type2(truth, _, truth, bool).
z_type2(forall(Body,Expr), Env, forall(BodyTi,ExprTi), bool) :-
    z_type_body(Body,Env,_,BodyTi,Subenv),
    z_type(Expr,Subenv,bool,ExprTi).
z_type2(exists(Body,Expr), Env, exists(BodyTi,ExprTi), Type) :-
    z_type2(forall(Body,Expr),Env,forall(BodyTi,ExprTi),Type).
z_type2(exists1(Body,Expr), Env, exists1(BodyTi,ExprTi), Type) :-
    z_type2(forall(Body,Expr),Env,forall(BodyTi,ExprTi),Type).
z_type2(and(A,B), Env, and(Ati,Bti), bool) :-
    z_type(A,Env,bool,Ati),z_type(B,Env,bool,Bti).
z_type2(or(A,B), Env, or(Ati,Bti), Type) :-
    z_type2(and(A,B),Env,and(Ati,Bti),Type).
z_type2(implies(A,B), Env, implies(Ati,Bti), Type) :-
    z_type2(and(A,B),Env,and(Ati,Bti),Type).
z_type2(equiv(A,B), Env, equiv(Ati,Bti), Type) :-
    z_type2(and(A,B),Env,and(Ati,Bti),Type).
z_type2(not(E), Env, not(Eti), bool) :- z_type(E,Env,bool,Eti).
z_type2(letpred(Letdefs,Expr), Env, letpred(Letties,ExprTi), bool) :-
    z_type2(letexpr(Letdefs,Expr), Env, letexpr(Letties,ExprTi), bool).
z_type2(spred(SExpr), Env, spred(SExprTi), bool) :-
    z_type(SExpr,Env,set(schema(_)),SExprTi).

%********************************************************************************
% Type checking schema expressions

z_type2(sref(Name,Deco,ParamValues,Renamings), Env, sref(Name,Deco,ParamValues,Renamings), Type) :-
    zlookup_schema(Env,Name,ParamNames,Expr),
    z_type_schemabody(Expr,Deco,ParamNames,ParamValues,Renamings,Env,Type,_).

z_type2(sand(A,B), Env, sand(Ati,Bti), Type) :-
    logical_bin_schema(A,B,Env,Ati,Bti,Type).
z_type2(sor(A,B), Env, sor(Ati,Bti), Type) :-
    logical_bin_schema(A,B,Env,Ati,Bti,Type).
z_type2(simplies(A,B), Env, simplies(Ati,Bti), Type) :-
    logical_bin_schema(A,B,Env,Ati,Bti,Type).
z_type2(sequiv(A,B), Env, sequiv(Ati,Bti), Type) :-
    logical_bin_schema(A,B,Env,Ati,Bti,Type).
z_type2(snot(A), Env, snot(Ati), Type) :-
    z_type(A,Env,Type,Ati).

z_type2(sforall(Body,Schema), Env, sforall(BodyTi,SchemaTi), Type) :-
    type_schema_quantifier(Body,Schema,Env,BodyTi,SchemaTi,Type).
z_type2(sexists(Body,Schema), Env, sexists(BodyTi,SchemaTi), Type) :-
    type_schema_quantifier(Body,Schema,Env,BodyTi,SchemaTi,Type).
z_type2(sexists1(Body,Schema), Env, sexists1(BodyTi,SchemaTi), Type) :-
    type_schema_quantifier(Body,Schema,Env,BodyTi,SchemaTi,Type).
z_type2(hide(Schema,Vars), Env, hide(SchemaTi,Vars), set(schema(Fields))) :-
    z_type(Schema,Env,set(schema(InnerFields)),SchemaTi),
    remove_fields_by_name(InnerFields,Vars,UFields),
    normalize_fields(UFields,Fields).

z_type2(fatsemi(A,B), Env, fatsemi(Ati,Bti), set(schema(Fields))) :-
    get_fields(A,B,Env,Ati,Bti,FieldsA,FieldsB),
    composite_fields(FieldsA,FieldsB,Out,In),
    remove_fields_by_name(FieldsA,Out,Aall),
    remove_fields_by_name(FieldsB,In,Ball),
    append_not_duplicate(Aall,Ball,UFields),
    normalize_fields(UFields,Fields).
z_type2(project(A,B), Env, project(Ati,Bti), set(schema(Fields))) :-
    z_type(A,Env,set(schema(_)),Ati),
    z_type(B,Env,set(schema(Fields)),Bti).
z_type2(pre(Schema), Env, pre(SchemaTi), set(schema(Fields))) :-
    z_type(Schema,Env,set(schema(AllFields)),SchemaTi),
    findall(name(Name,Deco),
            (member(field(name(Name,Deco),_),AllFields),
                member(Deco,['\'','!'])),
            ToRemove),
    remove_fields_by_name(AllFields,ToRemove,UFields),
    normalize_fields(UFields,Fields).
z_type2(pipe(A,B), Env, pipe(Ati,Bti), set(schema(Fields))) :-
    z_type(A,Env,set(schema(FieldsA)),Ati),
    z_type(B,Env,set(schema(FieldsB)),Bti),
    pipe_fields(FieldsA,FieldsB,Out,In),
    remove_fields_by_name(FieldsA,Out,AwoO),
    remove_fields_by_name(FieldsB,In,BwoI),
    append_not_duplicate(AwoO,BwoI,UFields),

    normalize_fields(UFields,Fields).

z_type2(body(Decl,Where), Env, TypedBody, Type) :-
    z_type_body(body(Decl,Where),Env,Type,BodyTi,_),
    ti_expr(BodyTi,TypedBody).
z_type2(text(Text), Env, text(TextTi), Type) :-
    z_type(Text,Env,Type,TextTi).

z_type2(given(Vars), _, given(Vars), none).
z_type2(namedset(SetName,Vars), _, namedset(SetName,Vars), none).
z_type2(data(Name,Arms), Env, data(Name,ArmsTi), none) :-
    z_type_arms(Arms,Env,Name,ArmsTi).
z_type2(sdef(Head,Body), Env, sdef(Head,BodyTi), none) :-
    z_type_schema(Head,Body,Env,BodyTi).
z_type2(defeq(Head,Body), Env, defeq(Head,BodyTi), none) :-
    z_type_schema(Head,Body,Env,BodyTi).
z_type2(eqeq(lhs(Name,ParamNames),Expr), Env, eqeq(lhs(Name,ParamNames),ExprTi), none) :-
    create_param_env(ParamNames,Env,ParamEnv),
    z_type(Expr,ParamEnv,PType,ExprTi),
    generic_type(ParamNames,PType,Type),
    lookup_type(Env,Name,Type).
z_type2(axdef(Body), Env, axdef(BodyTi), none) :-
    z_type_globaldef(Body,[],Env,BodyTi).
z_type2(define(ParamNames,Body), Env, define(ParamNames,BodyTi), none) :-
    z_type_globaldef(Body,ParamNames,Env,BodyTi).
z_type2(pred(Pred), Env, pred(PredTi), none) :-
    z_type(Pred,Env,bool,PredTi).


% internal expressions
z_type2(basetype(Type), _, basetype(Type), Type).

z_type2(binding(Bindings), Env, binding(BindingTis), schema(Fields)) :-
    z_type_binding(Bindings,Env,UFields,BindingTis),
    normalize_fields(UFields,Fields).

z_type2(ftconstant(Freetype,Name), _, ftconstant(Freetype,Name),freetype(Freetype)).
z_type2(ftconstructor(Freetype,Name,Expr), Env,
        ftconstructor(Freetype,Name,ExprTi), freetype(Freetype)) :-
    z_type(Expr,Env,_,ExprTi).
z_type2(ftdestructor(Freetype,Name,Expr), Env, ftdestructor(Freetype,Name,ExprTi), Type) :-
    z_type(Expr,Env,freetype(Freetype),ExprTi),
    lookup_type(Env,Name,ConstructorType),
    zrel(Type,freetype(Freetype),ConstructorType).
z_type2(ftcase(Freetype,Name,Expr), Env, ftcase(Freetype,Name,ExprTi), bool) :-
    z_type(Expr,Env,freetype(Freetype),ExprTi).



%********************************************************************************
% Type checking for lists of expressions

% z_type_l(Expressions, Environment, Types, TypeInfos).
%   Computes the type of each expression in the list and returns the
%   type informations and types as lists.
z_type_l([],_,[],[]).
z_type_l([Expr|Rest],Env,[Type|RestTypes],[TypeInfo|RestInfos]) :-
    z_type(Expr, Env, Type, TypeInfo),
    z_type_l(Rest, Env, RestTypes, RestInfos).

% z_type_sets(Expressions, Environment, Types, TypeInfos).
%   Computes the type of each (set) expression in the list and returns the
%   the type informations and the types with the "set" removed.
z_type_sets([],_,[],[]).
z_type_sets([Expr|Rest],Env,[Type|RestTypes],[TypeInfo|RestInfos]) :-
    z_type(Expr,Env,set(Type),TypeInfo),
    z_type_sets(Rest,Env,RestTypes,RestInfos).

z_type_binding([],_,[],[]).
z_type_binding([bindfield(Name,Expr)|BRest],Env,[field(Name,Type)|FRest],
               [ti(bindfield(Name,TypeInfo),none)|RestInfos]) :-
    z_type(Expr,Env,Type,TypeInfo),
    z_type_binding(BRest,Env,FRest,RestInfos).

% z_types_common(Expressions, Environemnt, Type, TypeInfos).
%   Assumes that each expression in the list has the same Type and
%   returns the list of all type informations.
z_types_common([],_,_,[]).
z_types_common([Expr|Rest],Env,Type,[TypeInfo|RestInfos]) :-
    z_type(Expr,Env,Type,TypeInfo),
    z_types_common(Rest,Env,Type,RestInfos).

% types_in_let(Letdefs,Environment,Variables,TypeInfos).
%   Computes the type of each expression in the list of let definitions and
%   returns a list of variables with their corresponding type and a list
%   of type informations for the let definitions.
types_in_let([],_,[],[]).
types_in_let([letdef(Name,Expr)|Rest],Env,[variable(Name,Type)|RTypes],[LetTi|RestInfos]) :-
    LetTi = ti(letdef(Name,TypeInfo),none),
    z_type(Expr,Env,Type,TypeInfo),
    types_in_let(Rest,Env,RTypes,RestInfos).

z_type_arms([],_,_,[]).
z_type_arms([Arm|ArmRest],Env,Freetype,[ti(AnnotatedArm,none)|TiRest]) :-
    z_type_arm(Arm,Env,Freetype,AnnotatedArm),
    z_type_arms(ArmRest,Env,Freetype,TiRest).
z_type_arm(arm(Name,nothing), Env, _, arm(Name,TypeInfo)) :-
    z_type(nothing,Env,_,TypeInfo).
z_type_arm(arm(Name,just(Expr)), Env, Freetype, arm(Name,ti(just(TypeInfo),none))) :-
    zrel(DomType,RanType,Type),
    z_type(Expr,Env,set(DomType),TypeInfo),
    lookup_type(Env,Freetype,set(RanType)),
    lookup_type(Env,Name,Type).

% z_type_body(Body,BodyTypeInfo,Subenvironment).
%   Type checks the given body and returns the type information and the
%   subenvironment specified by the body
z_type_body(body(Decl,Where),Env,Type,BodyTi,Subenv) :-
    BodyTi = ti(body(DeclTi,WhereTi),Type),
    Type = set(schema(Fields)),
    type_decls(Decl,Env,DeclTi,Fields),
    fields2var(Fields,Vars),
    create_subenv(Env,Vars,Subenv),
    z_types_common(Where,Subenv,bool,WhereTi).

fields2var([],[]).
fields2var([field(N,T)|FRest],[variable(N,T)|VRest]) :- fields2var(FRest,VRest).

%********************************************************************************
% predicates for type checking schemas
z_type_schema(shead(Name,ParamNames),Body,Env,BodyTi) :-
    findall(ref(P,[]),member(P,ParamNames),ParamValues),
    z_type_schemabody(Body,'',ParamNames,ParamValues,[],Env,PType,BodyTi),
    generic_type(ParamNames,PType,Type),
    lookup_type(Env,name(Name,''),Type).

z_type_schemabody(Body,Deco,PNames,PValues,Renamings,Env,set(schema(Fields)),TypeInfo) :-
    create_param_env(PNames,Env,ParamEnv),
    insert_parameter(Body,PNames,PValues,Expr),
    z_type(Expr,ParamEnv,set(schema(Fields1)),TypeInfo),
    decorate_fields(Fields1,Deco,Fields2),
    rename_fields(Fields2,Renamings,UFields),
    normalize_fields(UFields,Fields).

z_type_globaldef(Body,ParamNames,Env,BodyTi) :-
    create_param_env(ParamNames,Env,ParamEnv),
    z_type(Body,ParamEnv,set(schema(Fields)),BodyTi),
    z_type_globaldef2(Fields,ParamNames,Env).
z_type_globaldef2([],_,_).
z_type_globaldef2([field(Name,PType)|Rest],ParamNames,Env) :-
    generic_type(ParamNames,PType,Type),
    lookup_type(Env,Name,Type),
    z_type_globaldef2(Rest,ParamNames,Env).


pipe_fields(AFields,BFields,Out,In) :-
    intersect_fields(AFields,BFields,'!','?',Out,In).
composite_fields(AFields,BFields,Out,In) :-
    intersect_fields(AFields,BFields,'\'','',Out,In).
intersect_fields([],_,_,_,[],[]).
intersect_fields([field(name(Name,DecoA),Type)|ARest],BFields,DecoA,DecoB,
                 [name(Name,DecoA)|ORest],[name(Name,DecoB)|IRest]) :-
    member(field(name(Name,DecoB),TypeB),BFields),
    !,Type = TypeB,
    intersect_fields(ARest,BFields,DecoA,DecoB,ORest,IRest).
intersect_fields([_|ARest],BFields,DecoA,DecoB,ORest,IRest) :-
    intersect_fields(ARest,BFields,DecoA,DecoB,ORest,IRest).



generic_type(ParamNames,Type,GenericType) :-
    (ParamNames = [] ->
        GenericType = Type;
        GenericType = generic(ParamNames,Type)).


:- use_module(probsrc(error_manager),[add_internal_error/2]).
global_environment(Defs,Env,DefsTi) :-
    create_initial_zenvironment(Defs,GlobalEnv),

    predefined_variables(Predefs),
    create_subenv(GlobalEnv,Predefs,PreEnv),
    global_variables(PreEnv,Usertypes),
    create_subenv(PreEnv,Usertypes,Env),

    (z_type(Defs,Env,_,DefsTi) -> true ; add_internal_error('Typing Definitions failed: ',global_environment),fail).

% only for debugging:
:- public remove_zdef/2.
remove_zdef(zdef(_),zdef([])).
remove_zdef(env(E,D),env(R,D)) :- remove_zdef(E,R).

predefined_variables(Predefs) :-
    findall(variable(Name,Type),predefinedvar(Name,Type),Predefs).

global_variables(Env,Types) :-
    % given types
    findall(variable(Name,set(given(Name))),(zlookup(Env,given(L)),member(Name,L)),Given),
    % named sets
    findall(variable(Name,set(given(Name))),zlookup(Env,namedset(Name,_)),NamedSets),
    findall(variable(Name,given(SetName)),
            (zlookup(Env,namedset(SetName,L)),member(Name,L)),NamedSetMembers),
    % free types
    findall(variable(Name,set(freetype(Name))),zlookup(Env,data(Name,_)),FreeTypes),
    % free type constants
    findall(variable(Name,freetype(Freetype)),
            (zlookup(Env,data(Freetype,Arms)),member(arm(Name,nothing),Arms)),
            FTConstants),
    % free type constructors: not defined yet
    findall(variable(Name,set(tuple([_,freetype(Freetype)]))),
            (zlookup(Env,data(Freetype,Arms)),member(arm(Name,just(_)),Arms)),
            FTConstructors),
    % schema types: not defined yet
    env_z_defs(Env,Defs),
    findall(variable(name(Name,''),set(schema(_))),
            (member(Def,Defs),
                (Def = sdef(shead(Name,[]),_);
                    Def = defeq(shead(Name,[]),_))),
            Schemas),
    findall(variable(name(Name,''),generic(Params,set(schema(_)))),
            (member(Def,Defs),
                Params = [_|_],
                (Def = sdef(shead(Name,Params),_);
                    Def = defeq(shead(Name,Params),_))),
            ParamSchemas),
    % constants
    findall(variable(Name,_),
            (zlookup(Env,axdef(body(Decls,_))),decl_member(Decls,Name,_)),
            Constants),
    % abbreviations
    findall(variable(Name,_),zlookup(Env,eqeq(lhs(Name,[]),_)),Abbrevs),
    findall(variable(Name,generic(Params,_)),
            (zlookup(Env,eqeq(lhs(Name,Params),_)),Params=[_|_]),
            GenAbbrevs),
    % generic definitions
    findall(variable(Name,generic(Params,_)),
            (zlookup(Env,define(Params,body(Decls,_))),decl_member(Decls,Name,_)),
            Generics),

    append([Given,NamedSets,NamedSetMembers,FreeTypes,FTConstants,FTConstructors,
        Schemas,ParamSchemas,Constants,Abbrevs,GenAbbrevs,Generics],
           Types).


type_of_field(Fields,Fieldname,Type) :-
    when(nonvar(Fields),type_of_field2(Fields,Fieldname,Type)).
type_of_field2([F|R],Fieldname,Type) :-
    when(nonvar(F),type_of_field3([F|R],Fieldname,Type)).
type_of_field3([field(Fieldname,Type)|_],Fieldname,Type) :- !.
type_of_field3([_|Rest],Fieldname,Type) :- type_of_field2(Rest,Fieldname,Type).


create_param_env([],Env,Env) :- !.
create_param_env(Params,Env,NewEnv) :-
    Params = [_|_],
    findall(variable(Name,set(param(Name))),member(Name,Params),PVars),
    create_subenv(Env,PVars,NewEnv).

decl_member([decl(Vars,Type)|_],Var,Type) :- member(Var,Vars).
decl_member([_|Rest],Var,Type) :- decl_member(Rest,Var,Type).


prepare_generic(Params,Env,Gentype,Ids,Types) :-
    get_params(Params,Env,PTypes),
    unify_generic(Gentype,Ids,PTypes,Types).

unify_generic(Gentype,Ids,PTypes,Types) :-
    same_length(Ids,PTypes),
    when(ground(Gentype),prepare_generic2(Gentype,Ids,PTypes,Types)).

prepare_generic2(param(P),Ids,Vars,Type) :-
    nth0(N,Ids,P) -> nth0(N,Vars,Type) ; Type = param(P).
prepare_generic2(set(T),Ids,Vars,set(Type)) :- prepare_generic2(T,Ids,Vars,Type).
prepare_generic2(tuple(T),Ids,Vars,tuple(Types)) :- prepare_generic_l(T,Ids,Vars,Types).
prepare_generic2(schema(T),Id,Vars,schema(Types)) :- prepare_generic_fields(T,Id,Vars,Types).
prepare_generic2(num,_,_,num).
prepare_generic2(given(G),_,_,given(G)).
prepare_generic2(freetype(F),_,_,freetype(F)).

prepare_generic_l([],_,_,[]).
prepare_generic_l([T|Rest],Ids,Vars,[Type|TRest]) :-
    prepare_generic2(T,Ids,Vars,Type),
    prepare_generic_l(Rest,Ids,Vars,TRest).

prepare_generic_fields([],_,_,[]).
prepare_generic_fields([field(Name,T)|Rest],Ids,Vars,[field(Name,Type)|TRest]) :-
    prepare_generic2(T,Ids,Vars,Type),
    prepare_generic_fields(Rest,Ids,Vars,TRest).

get_params([],_,_).
get_params([EH|ER],Env,Types) :- z_type_sets([EH|ER],Env,Types,_).



% basic
predefinedvar(name('\\neq',''),generic([x],T)) :- zprel(x,x,T).

% sets
predefinedvar(name('\\emptyset',''),generic([x],set(param(x)))).
%predefinedvar(name('\\emptyset',''),generic([x],T)) :-  % emptyset with parameter e.g., \emptyset[\nat]
%    Set=set(param(x)), zrel(Set,Set,T).
predefinedvar(name('\\power_1',''),generic([x],T)) :- zrel(set(param(x)),set(set(param(x))),T).
predefinedvar(name('\\finset',''),T) :- predefinedvar(name('\\power_1',''),T).
predefinedvar(name('\\finset','_1'),T) :- predefinedvar(name('\\power_1',''),T).
% sets: operations
predefinedvar(name('\\cup',''),generic([x],T)) :-
    Set=set(param(x)), zrel(tuple([Set,Set]),Set,T).
predefinedvar(name('\\cap',''),T) :- predefinedvar(name('\\cup',''),T).
predefinedvar(name('\\setminus',''),T) :- predefinedvar(name('\\cup',''),T).
predefinedvar(name('\\bigcup',''),generic([x],T)) :- zrel(set(set(param(x))),set(param(x)),T).
predefinedvar(name('\\bigcap',''),T) :- predefinedvar(name('\\bigcup',''),T).
predefinedvar(name('\\#',''),generic([x],T)) :- zrel(set(param(x)),num,T).
% sets: relations
predefinedvar(name('\\subseteq',''),generic([x],T)) :- zrel(set(param(x)),set(param(x)),T).
predefinedvar(name('\\subset',''),T) :- predefinedvar(name('\\subseteq',''),T).
predefinedvar(name('\\notin',''),generic([x],T)) :- zrel(param(x),set(param(x)),T).

% relations: sets
predefinedvar(name('\\rel',''),generic([x,y],T)) :-
    zrel(tuple([set(param(x)),set(param(y))]),set(R),T), zprel(x,y,R).
predefinedvar(name('\\pfun',''),T) :- predefinedvar(name('\\rel',''),T).
predefinedvar(name('\\fun',''),T) :- predefinedvar(name('\\rel',''),T).
predefinedvar(name('\\pinj',''),T) :- predefinedvar(name('\\rel',''),T).
predefinedvar(name('\\inj',''),T) :- predefinedvar(name('\\rel',''),T).
predefinedvar(name('\\psurj',''),T) :- predefinedvar(name('\\rel',''),T).
predefinedvar(name('\\surj',''),T) :- predefinedvar(name('\\rel',''),T).
predefinedvar(name('\\bij',''),T) :- predefinedvar(name('\\rel',''),T).
predefinedvar(name('\\ffun',''),T) :- predefinedvar(name('\\rel',''),T).
predefinedvar(name('\\finj',''),T) :- predefinedvar(name('\\rel',''),T).
% relations: functions
predefinedvar(name('\\dom',''),generic([x,y],T)) :- zprel(x,y,R),zrel(R,set(param(x)),T).
predefinedvar(name('\\ran',''),generic([x,y],T)) :- zprel(x,y,R),zrel(R,set(param(y)),T).
% relations: operations
predefinedvar(name('\\id',''),generic([x],T)) :- zrel(set(param(x)),R,T),zprel(x,x,R).
predefinedvar(name('\\inv',''),generic([x,y],T)) :- zprelfunc(x,y,y,x,T).
predefinedvar(name('\\plus',''),generic([x],T)) :- zprelfunc(x,x,x,x,T).
predefinedvar(name('\\star',''),T) :- predefinedvar(name('\\plus',''),T).
predefinedvar(name('\\comp',''),generic([x,y,z],T)) :-
    zprel(x,y,A),zprel(y,z,B),zprel(x,z,C),zrel(tuple([A,B]),C,T).
predefinedvar(name('\\circ',''),generic([x,y,z],T)) :-
    zprel(x,y,A),zprel(y,z,B),zprel(x,z,C),zrel(tuple([B,A]),C,T).
predefinedvar(name('\\dres',''),generic([x,y],T)) :-
    zprel(x,y,R),zrel(tuple([set(param(x)),R]),R,T).
predefinedvar(name('\\rres',''),generic([x,y],T)) :-
    zprel(x,y,R),zrel(tuple([R,set(param(y))]),R,T).
predefinedvar(name('\\ndres',''),T) :- predefinedvar(name('\\dres',''),T).
predefinedvar(name('\\nrres',''),T) :- predefinedvar(name('\\rres',''),T).
predefinedvar(name('_(|_|)',''),generic([x,y],T)) :-
    zprel(x,y,R),zrel(tuple([R,set(param(x))]),set(param(y)),T).
predefinedvar(name('\\oplus',''),generic([x,y],T)) :-
    zprel(x,y,R),zrel(tuple([R,R]),R,T).
predefinedvar(name(iter,''),generic([x],T)) :-
    zprelfunc(x,x,x,x,F),zrel(num,F,T).

% numbers: sets
predefinedvar(name('\\num',''),set(num)).
predefinedvar(name('\\nat',''),set(num)).
predefinedvar(name('\\nat','_1'),set(num)).
predefinedvar(name('\\upto',''),T) :- zrel(tuple([num,num]),set(num),T).
% numbers: operations
predefinedvar(name('(-)',''),T) :- zrel(num,num,T).
predefinedvar(name('+',''),T) :- zrel(tuple([num,num]),num,T).
predefinedvar(name('-',''),T) :- predefinedvar(name('+',''),T).
predefinedvar(name('*',''),T) :- predefinedvar(name('+',''),T).
predefinedvar(name('\\div',''),T) :- predefinedvar(name('+',''),T).
predefinedvar(name('\\mod',''),T) :- predefinedvar(name('+',''),T).
% numbers: functions
predefinedvar(name('succ',''),T) :- zrel(num,num,T).
predefinedvar(name('min',''),T) :- zrel(set(num),num,T).
predefinedvar(name('max',''),T) :- predefinedvar(name('min',''),T).
% numbers: relations
predefinedvar(name('<',''),T) :- zrel(num,num,T).
predefinedvar(name('\\leq',''),T) :- predefinedvar(name('<',''),T).
predefinedvar(name('\\geq',''),T) :- predefinedvar(name('<',''),T).
predefinedvar(name('>',''),T) :- predefinedvar(name('<',''),T).

% sequences: sequences
predefinedvar(name('\\seq',''),generic([x],T)) :- zseq(param(x),Seq),zrel(set(param(x)),set(Seq),T).
predefinedvar(name('\\seq','_1'),T) :- predefinedvar(name('\\seq',''),T).
predefinedvar(name('\\iseq',''),T) :- predefinedvar(name('\\seq',''),T).
% sequences: operations
predefinedvar(name('\\cat',''),generic([x],T)) :-
    zseq(param(x),S),zrel(tuple([S,S]),S,T).
% sequences: functions
predefinedvar(name('rev',''),generic([x],T)) :- zseq(param(x),S),zrel(S,S,T).
predefinedvar(name('head',''),generic([x],T)) :- zseq(param(x),S),zrel(S,param(x),T).
predefinedvar(name('last',''),T) :- predefinedvar(name('head',''),T).
predefinedvar(name('tail',''),T) :- predefinedvar(name('rev',''),T).
predefinedvar(name('front',''),T) :- predefinedvar(name('tail',''),T).
predefinedvar(name('\\extract',''),generic([x],T)) :-
    zseq(param(x),S),zrel(tuple([set(num),S]),S,T).
predefinedvar(name('\\filter',''),generic([x],T)) :-
    zseq(param(x),S),zrel(tuple([S,set(param(x))]),S,T).
predefinedvar(name('squash',''),generic([x],T)) :- zseq(param(x),S),zrel(S,S,T).
predefinedvar(name('\\dcat',''),generic([x],T)) :-
    zseq(param(x),S),zseq(S,SS),zrel(SS,S,T).
predefinedvar(name('\\disjoint',''),generic([i,x],set(P))) :-
    zrel(param(i),set(param(x)),P).
predefinedvar(name('\\partition',''),generic([i,x],T)) :-
    zrel(P,set(param(x)),T),zrel(param(i),set(param(x)),P).
% sequences: relations
predefinedvar(name('\\prefix',''),generic([x],set(tuple([S,S])))) :- zseq(param(x),S).
predefinedvar(name('\\suffix',''),T) :- predefinedvar(name('\\prefix',''),T).
predefinedvar(name('\\inseq',''),T) :- predefinedvar(name('\\prefix',''),T).

% bag: set
predefinedvar(name('\\bag',''),generic([x],T)) :- zbag(param(x),Bag),zrel(set(param(x)),set(Bag),T).
% bag: functions
predefinedvar(name('items',''),generic([x],T)) :-
    zseq(param(x),Seq), zbag(param(x),Bag), zrel(Seq,Bag,T).
predefinedvar(name('count',''),generic([x],T)) :-
    zrel(param(x),num,ConcreteCount),zbag(param(x),Bag),zrel(Bag,ConcreteCount,T).
predefinedvar(name('\\bcount',''),generic([x],T)) :-
    zbag(param(x),Bag),zrel(tuple([Bag,param(x)]),num,T).
predefinedvar(name('\\otimes',''),generic([x],T)) :-
    zbag(param(x),Bag),zrel(tuple([num,Bag]),Bag,T).
predefinedvar(name('\\uplus',''),generic([x],T)) :-
    zbag(param(x),Bag),zrel(tuple([Bag,Bag]),Bag,T).
predefinedvar(name('\\uminus',''),T) :- predefinedvar(name('\\uplus',''),T).
predefinedvar(name('\\inbag',''),generic([x],Rel)) :-
    zbag(param(x),Bag),zrel(param(x),Bag,Rel).
predefinedvar(name('\\subbageq',''),generic([x],Rel)) :-
    zbag(param(x),Bag),zrel(Bag,Bag,Rel).

% tuples: operations
predefinedvar(name('\\mapsto',''),generic([x,y],T)) :-
    zrel(tuple([param(x),param(y)]),tuple([param(x),param(y)]),T).
% tuples: functions
predefinedvar(name('first',''),generic([x,y],T)) :- zrel(tuple([param(x),param(y)]),param(x),T).
predefinedvar(name('second',''),generic([x,y],T)) :- zrel(tuple([param(x),param(y)]),param(y),T).


% internal constructs
predefinedvar(name('\\prozignore',''),generic([x],T)) :- zrel(set(param(x)),set(param(x)),T).


zrel(A,B,set(tuple([A,B]))).
zprel(A,B,R) :- zrel(param(A),param(B),R).
zseq(A,T) :- zrel(num,A,T).
zrelfunc(A,B,C,D,F) :- zrel(A,B,X),zrel(C,D,Y),zrel(X,Y,F).
zprelfunc(A,B,C,D,F) :- zrelfunc(param(A),param(B),param(C),param(D),F).

zbag(A,T) :- zrel(A,num,T).


/******************************************************************************/

type_schema_quantifier(Body,Schema,Env,BodyTi,SchemaTi,set(schema(Fields))) :-
    z_type(Body,Env,set(schema(Fieldlist)),BodyTi),
    z_type(Schema,Env,set(schema(InnerFields)),SchemaTi),
    eliminate_fields(Fieldlist,InnerFields,UFields),
    normalize_fields(UFields,Fields).

logical_bin_schema(A,B,Env,Ati,Bti,set(schema(Fields))) :-
    get_fields(A,B,Env,Ati,Bti,FieldsA,FieldsB),
    merge_fields(FieldsA,FieldsB,Fields).

get_fields(A,B,Env,Ati,Bti,FieldsA,FieldsB) :-
    z_type(A,Env,set(schema(FieldsA)),Ati),
    z_type(B,Env,set(schema(FieldsB)),Bti).


type_decls(Decls,Env,DeclTi,Fields) :-
    type_decls2(Decls,Env,DeclTi,UFields),
    normalize_fields(UFields,Fields).
type_decls2([],_,[],[]).
type_decls2([decl(Vars,TypeExpr)|Rest],Env,[DeclTi|RestTi],Fields) :-
    !,DeclTi = ti(decl(Vars,TypeInfo),none),
    z_type(TypeExpr,Env,set(Type),TypeInfo),
    create_field_types(Vars,Type,LocalFields),
    append(LocalFields,RestFields,Fields),
    type_decls2(Rest,Env,RestTi,RestFields).
type_decls2([sdecl(Sref)|Rest],Env,[SDeclTi|RestTi],Fields) :-
    SDeclTi = ti(sdecl(SrefTi),none),
    z_type(Sref,Env,set(schema(LocalFields)),SrefTi),
    append(LocalFields,RestFields,Fields),
    type_decls2(Rest,Env,RestTi,RestFields).
create_field_types([],_,[]).
create_field_types([Name|NRest],Type,[field(Name,Type)|FRest]) :-
    create_field_types(NRest,Type,FRest).


merge_fields(FieldsA,FieldsB,Fields) :-
    eliminate_fields(FieldsA,FieldsB,NewFieldsB),
    append(FieldsA,NewFieldsB,UnsortedFields),
    normalize_fields(UnsortedFields,Fields).

eliminate_fields([],_,[]).
eliminate_fields([field(Name,TypeA)|Rest],Eliminate,Fields) :-
    (member(field(Name,TypeB),Eliminate) ->
        (TypeA = TypeB, Fields = Rest);
        Fields = [field(Name,TypeA)|ERest]),
    eliminate_fields(Rest,Eliminate,ERest).

normalize_fields(Fields,Normalized) :-
    findall(Name,member(field(Name,_),Fields),Fieldnames),
    sort(Fieldnames,SFieldnames),
    normalize_fields2(SFieldnames,Fields,Normalized).
normalize_fields2([],_,[]).
normalize_fields2([Name|Rest],Fields,[field(Name,Type)|NRest]) :-
    member(field(Name,Type),Fields),!,
    normalize_fields2(Rest,Fields,NRest).

decorate_fields(A,Deco,B) :- (Deco = '' -> A=B; decorate_fields2(A,Deco,B)).
decorate_fields2([],_,[]).
decorate_fields2([field(name(Name,DecoA),Type)|Rest],Deco,[field(name(Name,DecoB),Type)|DRest]) :-
    atom_concat(DecoA,Deco,DecoB),
    decorate_fields2(Rest,Deco,DRest).

rename_fields([],_,[]).
rename_fields([field(NameA,Type)|Rest],Renamings,[field(NameB,Type)|RenamedRest]) :-
    (member(rename(NameB,NameA),Renamings) ->
        true;
        NameA = NameB),
    rename_fields(Rest,Renamings,RenamedRest).


% declaration_tuple(Declarations,Environment,Tuple)
%   returns the expression for the tuple representing the declaration,
%   if only one variable is declared, a reference to the variable is returned.
declaration_tuple(Decl,Env,Tuple) :-
    schema_vars_decl(Decl,Env,Vars),
    declaration_tuple2(Vars,Tuple).
declaration_tuple2([Name],ref(Name,[])) :- !.
declaration_tuple2([A,B|R],tuple(Refs)) :-
    findall(ref(Name,[]),member(Name,[A,B|R]),Refs).


%********************************************************************************

remove_fields_by_name([],_,[]).
remove_fields_by_name([field(Name,Type)|FRest],ToRemove,Result) :-
    (member(Name,ToRemove) ->
        !,Result = Rest;
        Result = [field(Name,Type)|Rest]),
    remove_fields_by_name(FRest,ToRemove,Rest).

%********************************************************************************
% extract_btypes/3 returns the list of all variables declared in the given
% schema definition. It returns also a list of the types in the internal ProB
% syntax.
extract_btypes(SchemaDef,Vars,Types) :-
    ( ti_expr(SchemaDef,sdef(_,BodyTi))
    ; ti_expr(SchemaDef,defeq(_,BodyTi))
    ; ti_expr(SchemaDef,axdef(BodyTi))),
    !,
    ti_type(BodyTi,set(schema(Fields))),
    zfields2b(Fields,Vars,Types).

zfields2b([],[],[]).
zfields2b([field(Name,IType)|FRest],[Name|IRest],[BType|BRest]) :-
    internal2b(IType,BType),
    zfields2b(FRest,IRest,BRest).

internal2b(num,integer).
internal2b(given(Name),global(Id)) :- z_translate_identifier(Name,Id).
internal2b(freetype(Name),freetype(Id)) :- z_translate_identifier(Name,Id).
internal2b(set(X),set(T)) :- internal2b(X,T).
internal2b(tuple([A,B]),couple(Ab,Bb)) :- internal2b(A,Ab),internal2b(B,Bb).
internal2b(tuple([A,B,C|Rest]),couple(Ab,Rb)) :-
    internal2b(A,Ab),internal2b(tuple([B,C|Rest]),Rb).
internal2b(schema(Fi),record(Fb)) :- internal2b_fields(Fi,Fb).

internal2b(bool,pred).

internal2b_fields([],[]).
internal2b_fields([field(Name,Ti)|IRest],[field(Id,Tb)|BRest]) :-
    z_translate_identifier(Name,Id),
    internal2b(Ti,Tb),
    internal2b_fields(IRest,BRest).
