% (c) 2015-2020 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(model_translation, [translate_smt_model_to_b_values/4]).

:- use_module(probsrc(module_information),[module_info/2]).
:- module_info(group,smt_solvers).
:- module_info(description,'This module translates the models found by an SMT-solvers to a ProB state.').

:- use_module(library(lists), [maplist/3,maplist/4,
                               %convlist/3,delete/3,
                               append/2,
                               is_list/1]).
:- use_module(library(avl)).
:- use_module(library(codesio), [write_to_codes/2]).

:- use_module(probsrc(translate), [translate_bvalue/2]).
:- use_module(probsrc(tools), [arg_is_number/2,
                               split_atom/3,
                               atom_to_number/2,
                               ajoin/2]).
:- use_module(probsrc(tools_strings), [convert_to_number/2]).
:- use_module(probsrc(error_manager), [add_error/3,
                                       add_error_fail/3]).
:- use_module(probsrc(bsyntaxtree), [safe_create_texpr/3,
                                     unique_typed_id/3,
                                     create_equality/3,
                                     conjunct_predicates/2,
                                     disjunct_predicates/2,
                                     get_texpr_type/2,
                                     %create_implication/3,
                                     create_negation/2]).
:- use_module(probsrc(custom_explicit_sets), [try_expand_custom_set_to_list/4]).
:- use_module(probsrc(b_global_sets), [b_global_set/1, b_type2_set/2, global_type/2]).

%:- use_module(probsrc(self_check), [assert_must_succeed/1]).

:- use_module(solver_dispatcher).

translate_smt_model_to_b_values(Solver,State,Identifiers,Bindings) :-
    maplist(translate_smt_model_to_b_value(Solver,State),Identifiers,Bindings).

translate_smt_model_to_b_value(Solver,State,b(identifier(Id),Type,_Infos),binding(Id,Value,PPValue)) :-
    member(id(Id,Expr),State),
    smt_solver_interface_call(Solver,get_model_string(Expr,String)), !, % MAYBE WE SHOULD USE Z3_model_eval
    atom_codes(String,Codes),
    %format('Result of- get_model_string:~n~s~n---~n',[Codes]),
    parse_lisp_style_expressions([E],Codes,[]), !,
    (Solver == cvc4 -> translate_cvc4_tree(E,Type,Value) ;
     Solver == z3   -> translate_z3_tree_for_id(Id,E,Type,Value) ;
     otherwise      -> fail),

    (translate_bvalue(Value,PPValue) -> true ; PPValue='**pretty-print failed**').

:- use_module(probsrc(tools),[split_atom/3]).
translate_z3_tree_for_id(Id,Id,Type,Result) :- !,
    % variable is not assigned (hence z3 returns its name)
    % just ground it somehow
    % This NO LONGER SEEMS TO BE CORRECT FOR Mac/Darwin 4.8.8 or 4.8.9!!
    (%tools:host_platform(darwin),
     smt_solver_version(z3,Version),
     split_atom(Version,['.'],['4','8',B |_]), (B='8' ; B='9' ; B='10')
     -> format('Warning: No value for ~w in Z3 model (Z3 version ~w); try installing Z3 4.8.7~n',[Id,Version])
      ; z3_ground_result(Type,Result)).
translate_z3_tree_for_id(Id,EscapedId,Type,Result) :-
    % the id might have been escaped in z3 (e.g. due to ticks)
    ajoin(['|',Id,'|'],EscapedId), !,
    % variable is not assigned (hence z3 returns its name)
    % just ground it somehow
    z3_ground_result(Type,Result).
translate_z3_tree_for_id(_,E,Type,Result) :-
    translate_z3_tree_and_check(E,Type,Result,translate_z3_tree4).

translate_z3_tree(Tree,Type,Res) :-
     translate_z3_expr_using_full_model(Tree,Type,BExpr),
     b_compute_expression_nowf(BExpr,[],[],Res).

translate_z3_record(Env, field(Name,BType), CVCVal, field(Name,BVal)) :-
    translate_z3_expr(CVCVal, Env, BType, BVal).

% -----------------------------

% performs additional sanity check of computed values:
translate_z3_tree_and_check(SingleVal,ReturnType,Value,PP) :-
    translate_z3_tree(SingleVal,ReturnType,Value),
    check_value(ReturnType,Value,SingleVal,PP).

:- use_module(probsrc(error_manager),[add_internal_error/2]).
check_value(_,identifier(A),_,_) :- atom(A),!.
check_value(integer,I,SV,PP) :- \+ (I=int(Nr), number(Nr)),!,
     add_internal_error('Illegal Z3 value, not number computed:',PP:SV:I),fail.
check_value(boolean,B,SV,PP) :- B \= pred_true , B\= pred_false, !,
     add_internal_error('Illegal Z3 value, not boolean computed:',PP:SV:B),fail.
check_value(_,_,_,_).

% -------------------------------

translate_z3_expr_using_full_model(Expr,Type,Result) :-
    % get and parse the full model
    smt_solver_interface_call(z3,get_full_model_string(FullModel)),
    % format('z3 full model:~n~w~n',[FullModel]),
    z3_model_to_avl(FullModel,AVL),
    translate_z3_expr(Expr,env([],AVL),Type,Result).

map_translate_z3_predicate([], _, Acc, Acc).
map_translate_z3_predicate([Z3Pred|T], Env, Acc, Preds) :-
    translate_z3_predicate(Z3Pred, Env, Pred),
    map_translate_z3_predicate(T, Env, [Pred|Acc], Preds).

translate_z3_quantified_vars([], _, Acc, Acc).
translate_z3_quantified_vars([[Name,Type]|T], Env, Acc, Preds) :-
    z3_type_to_b_type(Type,BType),
    Id = b(identifier(Name),BType,[]),
    translate_z3_quantified_vars(T, Env, [Id|Acc], Preds).

% translate a z3 predicate to a B AST predicate
translate_z3_predicate([not,Left], Env, Res) :- !,
    translate_z3_predicate(Left, Env, TL),
    create_negation(TL,Res).
translate_z3_predicate([ZOP|T], Env, Res) :-
    z3_connective_to_b(ZOP, BOP),!,
    map_translate_z3_predicate(T, Env, [], NT),
    (   BOP == conjunct
    ->  conjunct_predicates(NT, Res)
    ;   BOP == disjunct
    ->  disjunct_predicates(NT, Res)
    ;   Pred =.. [BOP|NT],
        safe_create_texpr(Pred, pred, Res)
    ).
translate_z3_predicate([let,[[Id,Val]],Function], Env, Res) :- !,
    z3_replace_let_id_by_value(Id, Val, Function, NewFunction),
    translate_z3_predicate(NewFunction, Env, Res).
translate_z3_predicate([let,[[Id,Val]|T],Function], Env, Res) :-
    T \== [],
    z3_replace_let_id_by_value(Id, Val, Function, NewFunction),
    translate_z3_predicate([let,T,NewFunction], Env, Res).
translate_z3_predicate([ZOP,Left,Right], Env, Res) :-
     z3_binary_pred_to_b(ZOP, BOP,TypeL,TypeR),!,
     translate_z3_expr(Left, Env, TypeL, TL),
     translate_z3_expr(Right, Env, TypeR, TR),
     (BOP=equal -> create_equality(TL,TR,Res);
       Pred =.. [BOP,TL,TR],safe_create_texpr(Pred,pred,Res)).
translate_z3_predicate([select,Array,Entry], Env, Res) :-
    !,
    translate_z3_expr(Array, Env, _, BSet),
    translate_z3_expr(Entry, Env, _, BEntry),
    Res = b(member(BEntry,BSet),pred,[]).
translate_z3_predicate([exists,Vars,[Body]], Env, Res) :-
     translate_z3_quantified_vars(Vars, Env, [], BVars),
     Env = env(Ids,AVL),
     append(Ids, BVars, NewIds),
     translate_z3_predicate(Body, env(NewIds,AVL), BBody),!,
     Res = b(exists(BVars,BBody),pred,[]).
translate_z3_predicate(In, Env, Res) :-
    translate_z3_expr(In, Env, boolean, TRes),
    !,
    % expr to pred
    Res = b(equal(TRes,b(boolean_true,boolean,[])),pred,[]).
%translate_z3_predicate(Pred,E,_) :- print(translate_z3_predicate_failed(Pred,E)),nl,trace, translate_z3_predicate(Pred,E,_),fail.
translate_z3_predicate(Pred,_,_) :-
    add_error_fail(translate_z3_predicate, 'Missing translation for Z3 predicate:', Pred).

% see also smt_id_to_prob_id_preds in smtlib_solver
z3_connective_to_b(and,conjunct).
z3_connective_to_b(or,disjunct).
z3_connective_to_b('=>',implication).

% see also smt_id_to_prob_id in smtlib_solver
z3_binary_pred_to_b(<=,less_equal,integer,integer).
z3_binary_pred_to_b(<,less,integer,integer).
z3_binary_pred_to_b(>,greater,integer,integer). % not used ?
z3_binary_pred_to_b(>=,greater_equal,integer,integer). % not used ?
z3_binary_pred_to_b(=,equal,Type,Type).

map_translate_z3_expr([], _, Acc, Acc).
map_translate_z3_expr([Z3Pred|T], Env, Acc, Preds) :-
    translate_z3_expr(Z3Pred, Env, _, Pred),
    map_translate_z3_expr(T, Env, [Pred|Acc], Preds).

% translate a z3 expression to a B AST expression
%translate_z3_expr(Pred,E,T,_) :- print(translate_z3_expr(Pred,E,T)),nl,fail.
translate_z3_expr(['-',Number],_,integer,Res) :-
    !,
    Number1 is -Number,
    Res=b(integer(Number1),integer,[]).
translate_z3_expr(Nr,_,integer,Res) :- number(Nr),!, Res=b(integer(Nr),integer,[]).
translate_z3_expr(true,_,boolean,Res) :- !, Res=b(boolean_true,boolean,[]).
translate_z3_expr(false,_,boolean,Res) :- !, Res=b(boolean_false,boolean,[]).
translate_z3_expr([-,RealAtom], Env, real, Res) :-
    atom(RealAtom),
    atom_concat('-', RealAtom, NRealAtom),
    translate_z3_expr(NRealAtom, Env, real, Res).
translate_z3_expr(RealAtom, _, real, Res) :-
    atom(RealAtom),
    atom_to_number(RealAtom, Real),
    float(Real), !,
    Res = b(real(RealAtom),real,[]).
translate_z3_expr(AtomNr,_,global(Set),Res) :-
    atom(AtomNr),
    split_atom(AtomNr,['!'],[Set,val,InternalNumber]),
    !,
    convert_to_number(InternalNumber,Nr),
    ResNumber is Nr + 1,
    Val = fd(ResNumber,Set),
    global_type(Val,Set),
    Res=b(value(Val),global(Set),[]).
translate_z3_expr(Nr,_,global(GS),Res) :- number(Nr),!, % Z3 represents first element of GS by 0, ...
    Nr1 is Nr+1, Val = fd(Nr1,GS),
    global_type(Val,GS), % check if valid
    Res=b(value(Val),global(GS),[]).
translate_z3_expr([Couple,A,B],Env,couple(X,Y),b(couple(AT,BT),couple(X,Y),[])) :-
    atom(Couple),
    atom_concat('couple', _, Couple),
    translate_z3_expr(A,Env,X,AT),
    translate_z3_expr(B,Env,Y,BT).
translate_z3_expr([CouplePrj,CIdName], env(Ids,_), Type, Res) :-
    atom(CouplePrj),
    (   atom_concat('get_x_couple', _, CouplePrj),
        PrjNode = first_of_pair(CId)
    ;   atom_concat('get_y_couple', _, CouplePrj),
        PrjNode = second_of_pair(CId)
    ),
    !,
    CId = b(identifier(CIdName),_,_),
    member(CId, Ids),
    Res = b(PrjNode,Type,[]).
translate_z3_expr(Id, env(Env,_AVL), Type, Res) :-
    atom(Id),
    TId = b(identifier(Id),Type,_),
    member(TId, Env),
    !,
    Res=TId.
translate_z3_expr(CstFunc,env(Env,AVL),Type,CompSet) :- atom(CstFunc),
    Type = set(couple(PType,RType)),
    avl_fetch(CstFunc,AVL,'z3-define-fun'([[ParamName,ParameterType]],ReturnType,BodyTree)),
    z3_type_to_b_type(ParameterType,PType), % check if type matches
    z3_type_to_b_type(ReturnType,RType), % ditto
    !,
    PID = b(identifier(ParamName),PType,[]),
    %print(translate_body(CstFunc,PID)),nl,
    (translate_z3_expr(BodyTree,env([PID|Env],AVL),RType,BodyExpr) -> true
    ; trace, translate_z3_expr(BodyTree,env([PID|Env],AVL),RType,BodyExpr)),
    %print(body(CstFunc)),translate:print_bexpr(BodyExpr),nl,
    unique_typed_id("lambda",RType,LambdaID),
    create_equality(LambdaID,BodyExpr,CPred),
    % {PID,lambda | lambda = BodyExpr}
    safe_create_texpr(comprehension_set([PID,LambdaID],CPred),set(couple(PType,RType)),CompSet),
    translate:print_bexpr(CompSet),nl.
translate_z3_expr([CstFunc,Arg],env(Env,AVL),Type,Res) :- % function application
    avl_fetch(CstFunc,AVL,'z3-define-fun'([[ParamID,ParameterType]],ReturnType,BodyTree)),
    z3_type_to_b_type(ReturnType,Type), % check if type matches
    !,
    (ParamID == Arg   % special case if argument and parameter identical
    -> translate_z3_expr(BodyTree,env(Env,AVL),Type,Res)
    ; z3_type_to_b_type(ParameterType,PType), % TO DO: couple-> ground type; PType could be not ground !
     translate_z3_expr(CstFunc,env(Env,AVL),set(couple(PType,Type)),TFun),
     translate_z3_expr(Arg,env(Env,AVL),PType,TArg),
     safe_create_texpr(function(TFun,TArg),Type,Res)
    ).
translate_z3_expr([let,[[Id,Val]],Function],Env,Type,Res) :- !,
    z3_replace_let_id_by_value(Id,Val,Function,NewFunction),
    translate_z3_expr(NewFunction,Env,Type,Res).
translate_z3_expr([let,[[Id,Val]|T],Function],Env,Type,Res) :-
    T \== [],
    z3_replace_let_id_by_value(Id,Val,Function,NewFunction),
    translate_z3_expr([let,T,NewFunction],Env,Type,Res).
translate_z3_expr([ZOP, Left], Env, Type,Res) :-
    z3_unary_expr_to_b(ZOP,BOP,Type),!,
    translate_z3_expr(Left,Env, Type, TL),
    Expr =.. [BOP,TL],
    safe_create_texpr(Expr,Type,Res).
translate_z3_expr([ZOP, Left, Right], Env, Type, Res) :-
    z3_binary_expr_to_b(ZOP,BOP,TA,TB,Type),!,
    translate_z3_expr(Left, Env, TA, TL), get_texpr_type(TL,TA),
    translate_z3_expr(Right,Env, TB, TR), get_texpr_type(TR,TB),
    Expr =.. [BOP,TL,TR],
    safe_create_texpr(Expr,Type,Res).
translate_z3_expr([ite, Test, Left, Right], Env, Type, Res) :- !,
    translate_z3_predicate(Test,Env, BTest),
    translate_z3_expr(Left, Env, Type, BThen),
    translate_z3_expr(Right,Env, Type, BElse), get_texpr_type(BThen,Type),
    safe_create_texpr(if_then_else(BTest,BThen,BElse),Type,Res).
translate_z3_expr(Id,_,Type,Res) :- atom(Id), print(unknown_z3_id(Id)),nl,
    Res = b(identifier(Id),Type,[]).
translate_z3_expr(['_','as-array',AuxFunName],_,set(X),Res) :- !,
    % set is represented by aux function
    translate_z3_expr_using_full_model(AuxFunName,set(couple(X,boolean)),ChFunSet),
    % we computed the characteristic function of the set; now convert it to a set
    % Now we need to compute dom(ChFunSet |> {TRUE}) = ChFunSet~[{TRUE}] = {x|ChFunSet(x)=TRUE}
    unique_typed_id("el!",X,Element),
    safe_create_texpr(function(ChFunSet,Element),boolean,FCall),
    create_equality(b(boolean_true,boolean,[]),FCall,NewNewPred),
    safe_create_texpr(comprehension_set([Element],NewNewPred),set(X),CompSet),
    % translate:print_bexpr(CompSet),nl,
    compute_bcomp_set(CompSet,Set),
    Res = b(value(Set),set(X),[]).
translate_z3_expr([store,Sub,Entry,StoreRes],Env,set(Type),Res) :-
    translate_z3_expr(Sub,Env,set(Type),S1),
    translate_z3_expr(Entry,Env,Type,El2),
    S2 = b(set_extension([El2]),set(Type),[]),
    (  StoreRes = true -> Expr = union(S1,S2)
     ; StoreRes = false -> Expr = set_subtraction(S1,S2)
    ),
    !,
    b_compute_expression_nowf(b(Expr,set(Type),[]), [], [], List),
    Res = b(value(List),set(Type),[]).
translate_z3_expr([[as,const,['Array',Z3Type,'Bool']],false],_,Type,b(value([]),Type,[])) :-
    (Type=set(BType), z3_type_to_b_type(Z3Type,BType) -> true
      ; add_error(translate_z3_expr,'Unexpected type in Z3 model: ',types(Z3Type,Type))), !.
translate_z3_expr([[as,const,['Array',Z3Type,'Bool']],true],_,Type,b(value(Res),Type,[])) :-
    (Type=set(BType), z3_type_to_b_type(Z3Type,BType)
     -> b_type2_set(BType,Res),!
      ; add_error(translate_z3_expr,'Unexpected type in Z3 model: ',types(Z3Type,Type)),fail).
translate_z3_expr(['record'|Z3Fields],Env,record(Fields),b(rec(FieldsOut),record(Fields),[])) :-
    maplist(translate_z3_record(Env),Fields,Z3Fields,FieldsOut).
translate_z3_expr([lambda,Args,Body],env(Ids,_),set(BType),Res) :-
    Args = [[ID,Z3Type]],
    z3_type_to_b_type(Z3Type,BType),
    !,
    TID = b(identifier(ID),BType,[]),
    translate_z3_predicate(Body,env([TID|Ids],empty),TBody),
    safe_create_texpr(comprehension_set([TID],TBody),set(BType),CompSet),
    % print('lambda: '),translate:print_bexpr(CompSet),nl, print(CompSet),nl,
    compute_bcomp_set(CompSet,Set),
    Res = b(value(Set),set(BType),[]).
translate_z3_expr([['_',map,[AndOr,['Bool','Bool'],'Bool']]|T], Env, _, Res) :-
    (AndOr == and; AndOr == or),
    !,
    map_translate_z3_expr(T, Env, [], TT),
    TT = [b(_,SetType,_)|_],
    (   AndOr == and
    ->  Res = b(general_intersection(b(set_extension(TT),set(SetType),[])),SetType,[])
    ;   Res = b(general_union(b(set_extension(TT),set(SetType),[])),SetType,[])
    ).
 %translate_z3_expr(Pred,E,T,_) :- print(translate_z3_expr_failed(Pred,E,T)),nl,trace,translate_z3_expr(Pred,E,T,_),fail.
translate_z3_expr(Expr,_,Type,_) :-
    write(Expr),nl,
    add_error_fail(translate_z3_expr, 'Missing translation for Z3 expression:', Expr:Type).

z3_unary_expr_to_b('-',unary_minus ,integer).

% z3_binary_expr_to_b(Z3Op,BOp,TypeLeftArg,TypeRightArg,TypeOfExpr)
z3_binary_expr_to_b('*',multiplication, integer,integer,integer).
z3_binary_expr_to_b('+',add, integer,integer,integer).
z3_binary_expr_to_b('-',minus ,integer,integer,integer).
z3_binary_expr_to_b(Z3Couple,couple,TA,TB,couple(TA,TB)) :-
    atom(Z3Couple),
    atom_concat('couple', _, Z3Couple).

% --------------------

z3_ground_result(integer,int(0)).
z3_ground_result(boolean,pred_true).
z3_ground_result(couple(X,Y),(XG,YG)) :-
    z3_ground_result(X,XG), z3_ground_result(Y,YG).
z3_ground_result(set(_),[]).
z3_ground_result(record(Fields),rec(FieldsOut)) :-
    maplist(z3_ground_record_fields,Fields,FieldsOut).

z3_ground_record_fields(field(Name,BType),field(Name,BVal)) :-
    z3_ground_result(BType,BVal).

% --------------------

:- use_module(probsrc(b_interpreter),[b_compute_expression_nowf/4]).
compute_bcomp_set(CompSet,ResultOut) :-
    % translate:set_unicode_mode,translate:print_bexpr(CompSet),
    b_compute_expression_nowf(CompSet,[],[],Res),
    (try_expand_custom_set_to_list(Res,ResultOut,true,z3_set_to_b_set)
     -> true
     ;  ResultOut = Res).

z3_type_to_b_type('Int', integer).
z3_type_to_b_type('Bool', boolean).
z3_type_to_b_type(Z3Couple, couple(T1,T2)) :-
    atom(Z3Couple),
    atom_codes(Z3Couple, Codes),
    get_couple_type_from_atom([TT1,TT2], Codes, []),
    write_to_codes(T1, TT1),
    write_to_codes(T2, TT2).
z3_type_to_b_type(GlobalSet, global(GlobalSet)) :-
    b_global_set(GlobalSet).

get_couple_type_from_atom([T1,T2]) -->
    "couple(",
    type_symbol(T1),
    ",",
    type_symbol(T2),
    ")".

z3_replace_let_id_by_value(Id,Val,Function,NewFunction) :-
    is_list(Function), !,
    maplist(z3_replace_let_id_by_value(Id,Val),Function,NewFunction).
z3_replace_let_id_by_value(Id,Val,Function,NewFunction) :-
    atomic(Function), !,
    Id = Function -> NewFunction = Val ; NewFunction = Function.

z3_model_to_avl(FullModel,AVL) :-
    atom_codes(FullModel,Codes),
    %format('Full model:~n~s~n',[Codes]),nl,
    parse_lisp_style_expressions(E,Codes,[]),
    z3_lisp_to_avl(E,AVL).

z3_lisp_to_avl(E,AVL) :-
    empty_avl(In),
    z3_lisp_to_avl(E,In,AVL).

z3_lisp_to_avl([],A,A).
z3_lisp_to_avl([['define-fun',Name,Param,Return,Tree]|T],AIn,AOut) :-
    avl_store(Name,AIn,'z3-define-fun'(Param,Return,Tree),ATemp),
    z3_lisp_to_avl(T,ATemp,AOut).
% currently I think I do not need these facts...
% as far as I understand they represent enumerated set Elements
z3_lisp_to_avl([['declare-fun',_Name,_Params,_Type0]|T],AIn,AOut) :-
    z3_lisp_to_avl(T,AIn,AOut).
% these occur for cardinality constraints
z3_lisp_to_avl([['forall',_Vars,_Pred]|T],AIn,AOut) :-
    z3_lisp_to_avl(T,AIn,AOut).


% -----------------------------

translate_cvc4_tree(['CONST_RATIONAL',Arg],integer,int(Val)) :-
    arg_is_number(Arg,Val).
translate_cvc4_tree(['CONST_BOOLEAN',1],boolean,pred_true).
translate_cvc4_tree(['CONST_BOOLEAN',0],boolean,pred_false).
translate_cvc4_tree(['CONST_STRING',Content],string,string(Content)).
translate_cvc4_tree(['SINGLETON',ValIn],set(X),[ValOut]) :-
    translate_cvc4_tree(ValIn,X,ValOut).
translate_cvc4_tree(['EMPTYSET',emptyset,_Type],set(_),[]).
translate_cvc4_tree(['TUPLE',V1,V2],couple(Type1,Type2),(TV1,TV2)) :-
    translate_cvc4_tree(V1,Type1,TV1),
    translate_cvc4_tree(V2,Type2,TV2).
translate_cvc4_tree(['APPLY_CONSTRUCTOR','__cvc4_tuple_ctor',V1,V2],couple(Type1,Type2),(TV1,TV2)) :-
    translate_cvc4_tree(V1,Type1,TV1),
    translate_cvc4_tree(V2,Type2,TV2).
translate_cvc4_tree(['UNION',V1,V2],set(X),Val) :-
    translate_cvc4_tree(V1,set(X),TV1),
    translate_cvc4_tree(V2,set(X),TV2),
    append(TV1,TV2,Val).
translate_cvc4_tree(['UNINTERPRETED_CONSTANT',C],global(SId),fd(PBCId,SId)) :-
    atom_codes(C,CCodes),
    parse_uc(CId,CCodes,[]),
    PBCId is CId + 1.
translate_cvc4_tree(['RECORD',_CVCRecordType|Fields],record(RFields),rec(FieldsOut)) :-
    maplist(translate_cvc4_record,RFields,Fields,FieldsOut).

translate_cvc4_record(field(Name,BType),CVCVal,field(Name,BVal)) :-
    translate_cvc4_tree(CVCVal,BType,BVal).

parse_uc(Id) -->
    "uc__SORT_TYPE_var_", number(_),
    "__", number(N), {number_codes(Id,N)}.

parse_lisp_style_expressions(L) --> {reset_line_count},lisp_style_expressions(L).

lisp_style_expressions(Res) --> real_ws,!,
   lisp_style_expressions(Res).
lisp_style_expressions([E|Es]) -->
    lisp_style_expression(E), !,
    lisp_style_expressions(Es).
lisp_style_expressions([]) --> [],!.
%lisp_style_expressions(_,In,_) :- linecount(Line),
%    format("lisp-style parser failed on line ~w on: ~s~n",[Line,In]),fail.


closed_lisp_style_expressions(Res) --> real_ws,!,
   closed_lisp_style_expressions(Res).
closed_lisp_style_expressions([]) --> ")",!.
closed_lisp_style_expressions([E|Es]) -->
    lisp_style_expression(E), !,
    closed_lisp_style_expressions(Es).


:- dynamic linecount/1.
linecount(1).
reset_line_count :- retractall(linecount(_)), assert(linecount(1)).
inc_linecount :- retract(linecount(Nr)), N1 is Nr+1, assert(linecount(N1)).

% real whitespace
real_ws --> "\n", {inc_linecount},!, ws.
real_ws --> [W], {W = 32 ; W = 10 ; W = 13}, !, ws.
real_ws --> ";;", !, any_but_newline.

% optional whitespace
ws --> real_ws,!.
ws --> [].
any_but_newline --> [C], {C \= 10}, !, any_but_newline.
any_but_newline -->"\n", {inc_linecount},!, ws.

z3_op([42]) --> [42]. % *
z3_op([43]) --> [43]. % +
z3_op([45]) --> [45]. % -
z3_op([47]) --> [47]. % /
% TO DO: support all operators

lisp_style_expression(A) --> real(Cs), !, {atom_codes(A, Cs)}.
lisp_style_expression(L) --> "!", !, ws, lisp_style_expression(L), ws, ":weight", ws, number(_). % optional term parameter
lisp_style_expression(A) --> z3_op(Cs), !, {atom_codes(A, Cs)}.
lisp_style_expression(A) --> typed_couple(Cs), !, {atom_codes(A, Cs)}.
lisp_style_expression(A) --> non_empty_symbol(Cs), !, {atom_codes(A, Cs)}.
lisp_style_expression(N) --> number(Cs), !, {number_codes(N, Cs)}.
lisp_style_expression(N) --> "-", number(Cs), !, {number_codes(N2,Cs), N is -N2}.
lisp_style_expression(A) --> string(Codes), !, {atom_codes(A,Codes)}.
lisp_style_expression(List) --> "(", !, closed_lisp_style_expressions(List).
lisp_style_expression(record_type(Fields)) -->
    "[#", !, ws, record_type_fields(Fields),!,
             ws, expect("#]",record_type).
lisp_style_expression(_,In,_) :- In \= [], % not at end of string
    linecount(Line),
    format("lisp-style parser failed on line ~w on: ~s~n",[Line,In]),fail.

expect(L,_) --> L,!.
expect(L,Ctxt,In,_) :- linecount(Line),
    format("lisp-style parser failed on line ~w expecting ~s for ~w on: ~s~n",[Line,L,Ctxt, In]),fail.

record_type_fields([F|Fields]) -->
    record_type_field(F), ws, expect(",",record_type_field), !,
    ws, record_type_fields(Fields).
record_type_fields([F]) --> record_type_field(F).

record_type_field(field(NameA,TypeA)) -->
    non_empty_symbol(Name), ":(TYPE_CONSTANT ", non_empty_symbol(Type), " type)",
    {atom_codes(NameA,Name),atom_codes(TypeA,Type)}.

real(R) -->
    number(I),
    ".",
    number(D),
    (   "?",
        !
    ;   ""
    ),
    {append([I,[46],D],R)}.

number([D|Ds]) --> digit(D), number(Ds).
number([D]) --> digit(D).

typed_couple(Cs) -->
    "|couple(",
    type_symbol(Cs1), !,
    expect(",", typed_couple),
    type_symbol(Cs2),
    expect(")|", typed_couple),
    {append(["couple(",Cs1,",",Cs2,")"], Cs)}.

digit(D) --> [D], {48 =< D, D =< 57}.

string(Codes) --> "\"", non_empty_symbol(Codes), "\"".

type_symbol(A) --> non_empty_symbol(A1), type_symbol_term(A2,0), !, {append(A1, A2, A)}.
type_symbol(_,In,_) :- linecount(Line),
    format("lisp-style parser failed on line ~w expecting type_symbol on: ~s~n",[Line, In]),fail.

type_symbol_term([40|A],Lvl) --> "(", {L1 is Lvl+1}, type_symbol_term(A,L1).
type_symbol_term([41|A],Lvl) --> ")", {Lvl>0, L1 is Lvl-1}, type_symbol_term(A,L1).
type_symbol_term(A,Lvl) --> non_empty_symbol(A1), type_symbol_term(A2,Lvl), {append(A1, A2, A)}.
type_symbol_term(A,0) --> non_empty_symbol(A).
type_symbol_term([],0) --> "".

non_empty_symbol([A|As]) -->
    [A],
    { 65 =< A, A =< 90
    ; 97 =< A, A =< 122
    % ; 48 =< A, A =< 57
    ; A = 124 % "|"
    ; A = 95  % "_",
    ; A = 45  % "-"
    ; A = 39  % "'"
    ; A = 33  % "!"
    ; 60 =< A, A =< 62 % "<", "=", ">"
    },
    symbol(As).

symbol([A|As]) -->
    [A],
    { 65 =< A, A =< 90
    ; 97 =< A, A =< 122
    ; 48 =< A, A =< 57
    ; A = 124 % "|"
    ; A = 95  % "_",
    ; A = 45  % "-"
    ; A = 39  % "'"
    ; A = 33  % "!"
    ; 60 =< A, A =< 62 % "<", "=", ">"
    },
    symbol(As).
symbol([]) --> "".
