% (c) 2014-2020 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(smtlib2_environment,[empty_env/1,
                               add_identifier/4,

                               add_assertion/3,
                               get_assertions/2,
                               push_assertion_stack/3,
                               pop_assertion_stack/3,

                               get_type/3,
                               add_function/6,
                               get_function/5,
                               create_local_env/3,
                               get_custom_types/2,
                               is_custom_type/2,
                               add_custom_type/3]).

:- use_module(probsrc(module_information),[module_info/2]).
:- module_info(group,smtlib).
:- module_info(description,'The SMT-LIB 2 Interpreter - Environment').

:- use_module(library(lists)).
:- use_module(library(avl)).

% environment: env(Identifiers,Functions,Assertions,CustomTypes)
% identifiers: avl tree of b(identifier(ID),Type,[]) stored at ID
% functions: avl tree of function(Parameters,RetType,Definition) at ID
% Assertions: Stack (List) of Lists of predicates
% CustomTypes: List of custom types, i.e. deferred sets

empty_env(env(IDs,Funs,[[]],[])) :-
  empty_avl(IDs),
  empty_avl(Funs).

get_custom_types(env(_,_,_,Ct),Ct).
is_custom_type(env(_,_,_,Cts),Ct) :-
  member(Ct,Cts).
add_custom_type(env(IDs,Funs,As,Ct),ID,env(IDs,Funs,As,[ID|Ct])).

create_local_env(env(IDs,Funs,As,Ct),LocalIDs,env(NIDs,Funs,As,Ct)) :-
  avl_store_all_ids(LocalIDs,IDs,NIDs).

avl_store_all_ids([],A,A).
avl_store_all_ids([b(identifier(ID),Type,[])|T],AIn,AOut) :-
  avl_store(ID,AIn,b(identifier(ID),Type,[]),ATemp),
  avl_store_all_ids(T,ATemp,AOut).

add_function(env(IDs,Funs,As,Ct),ID,P,R,D,env(IDs,NFuns,As,Ct)) :-
  avl_store(ID,Funs,function(P,R,D),NFuns).
get_function(ID,env(_,Funs,_,_),Params,Results,Definition) :- atomic(ID),!,
  avl_fetch(ID,Funs,function(Params,Results,Definition)).
get_function(ID,env(_,Funs,_,_),Params,Results,Definition) :-
  print(var_get_function(ID)),nl,
  avl_member(ID,Funs,function(Params,Results,Definition)).

pop_assertion_stack(Env,0,Env) :- !.
pop_assertion_stack(env(IDs,Funs,[_|Ass],Types),Count,EnvOut) :- Count > 0, !,
  NCount is Count - 1,
  pop_assertion_stack(env(IDs,Funs,Ass,Types),NCount,EnvOut).

push_assertion_stack(Env,0,Env) :- !.
push_assertion_stack(env(IDs,Funs,Ass,Types),Count,EnvOut) :- Count > 0, !,
  NCount is Count - 1,
  pop_assertion_stack(env(IDs,Funs,[[]|Ass],Types),NCount,EnvOut).

get_assertions(env(_,_,As,_),AllAssertions) :- append(As,AllAssertions).
add_assertion(env(IDs,Funs,[CurrentAssertionSet|AssertionSets],Ct),Term,env(IDs,Funs,[NAs|AssertionSets],Ct)) :-
  NAs = [Term|CurrentAssertionSet].

add_identifier(env(IDs,Funs,As,Ct),ID,Type,env(NIDs,Funs,As,Ct)) :-
  avl_store(ID,IDs,b(identifier(ID),Type,[]),NIDs).

get_type(Id,env(IDs,_,_,_),Type) :- atomic(Id),!,
  avl_fetch(Id,IDs,b(identifier(Id),Type,[])).
get_type(Id,env(IDs,_,_,_),Type) :-
  print(var_get_type(Id)),nl,
  avl_member(Id,IDs,b(identifier(Id),Type,[])).
