% (c) 2009-2020 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(specfile, [
                    specfile_possible_trans_name/2, specfile_possible_trans_name_for_successors/2,
                    specfile_trans/6,
                    prepare_state_for_specfile_trans/2, prepare_state_for_specfile_trans/3,

                    property/2, elementary_prop/2,

                    partial_trans/4, specfile_trans_or_partial_trans/7,

                    animation_mode/1, set_animation_mode/1, add_animation_minor_mode/1,
                    animation_minor_mode/1, set_animation_minor_mode/1, remove_animation_minor_mode/0,
                    unset_animation_minor_modes/1, reset_animation_minor_modes/1,
                    b_mode/0, z_mode/0, b_or_z_mode/0, b_or_z_mode/1, eventb_mode/0,
                    csp_mode/0, csp_mode/1,
		                csp_with_bz_mode/0, z_or_tla_minor_mode/0,
					          process_algebra_mode/0,

					          type_check_csp_and_b/0,

                    currently_opened_file/1, % only true for successfully opened files
                    currently_opened_file_status/2, % second parameter: success or failure
                    %reset_currently_opened_file/0,
                    set_currently_opened_file/1, set_failed_to_open_file/1,
                    set_currently_opening_file/1,
                    set_currently_opened_package/1,
                    spec_file_has_been_successfully_loaded/0,

					state_corresponds_to_initialised_b_machine/1, state_corresponds_to_initialised_b_machine/2,
					state_corresponds_to_fully_setup_b_machine/1, state_corresponds_to_fully_setup_b_machine/2,
					state_corresponds_to_set_up_constants/1, state_corresponds_to_set_up_constants/2, state_corresponds_to_set_up_constants_only/2,
					current_b_expression/1,
					get_current_state_for_b_formula/2, get_state_for_b_formula/3,

					get_operation_name/2, get_operation_internal_name/2,
					translate_operation_name/2,
					get_operation_arguments/2, get_operation_return_values_and_arguments/3,
					build_up_b_real_operation/2,
					expand_const_and_vars/3, expand_const_and_vars_to_full_store/2,
                    expand_to_constants_and_variables/3,

					compute_operation_effect_max/6, % has replaced compute_operation_effect/5
					create_setup_constants_operation/2,
					get_internal_representation/1, get_internal_representation/2,
					get_possible_event/1, get_feasible_event/1,
					get_possible_language_specific_top_level_event/3,
					get_specification_description/2
					]).

:- use_module(tools).

:- use_module(module_information).
:- module_info(group,animator).
:- module_info(description,'This module computes transitions and properties depending on the current animator mode.').

:- use_module(library(lists)).
%:- use_module(library(avl)).
%:- use_module(library(clpfd)).

:- use_module(debug).
:- use_module(b_interpreter).
:- use_module(preferences).
:- use_module(self_check).
:- use_module(error_manager).
:- use_module(translate).
:- use_module(succeed_max,[succeed_max_call_id/3]).

:- use_module(store,[empty_state/1]).
:- use_module(bmachine).
:- use_module(bsyntaxtree).


%% :- use_module(delay).

:- use_module(value_persistance, [load_constants/2]).

/* --------------------------------- */




/* --------------------------------- */
/* ANIMATION & TEMPORAL VERIFICATION */
/* --------------------------------- */


:- dynamic animation_mode/1, animation_minor_mode/1.
% not marking as volatile; otherwise we startup with no animation_mode set
% not all parts of ProB can deal with this at the moment
%:- volatile animation_mode/1, animation_minor_mode/1.

animation_mode(b).
/* possible values: b, csp, xtl, z, csp_and_b, promela */

b_mode :- animation_mode(X), (X=b ; X=csp_and_b).
z_mode :- animation_mode(X), (X=b ; X=csp_and_b), animation_minor_mode(z).
eventb_mode :-
          animation_mode(X), (X=b ; X=csp_and_b),
          animation_minor_mode(eventb).
csp_with_bz_mode :- animation_mode(csp_and_b).
z_or_tla_minor_mode :- animation_minor_mode(X), (X=tla ; X=z).

csp_mode :- animation_mode(X), csp_mode(X).
csp_mode(cspm).
csp_mode(csp_and_b).

b_or_z_mode :-
  (animation_mode(X) -> (X=b -> true ; X=csp_and_b)). % basically the same as b_mode
b_or_z_mode(b). b_or_z_mode(csp_and_b).

process_algebra_mode :-
 (animation_mode(X) -> \+ X=b, \+X=z).

set_animation_mode(X) :-
   retractall(animation_mode(_)),
   remove_animation_minor_mode,
   assert(animation_mode(X)).

set_animation_minor_mode(X) :-
   remove_animation_minor_mode,
   assert(animation_minor_mode(X)).
remove_animation_minor_mode :-
   retractall(animation_minor_mode(_)).

% use to temporarily unset animation minor mode:
:- use_module(library(lists),[maplist/2]).
unset_animation_minor_modes(L) :- findall(X,retract(animation_minor_mode(X)),L).
reset_animation_minor_modes(L) :- maplist(add_animation_minor_mode,L).

add_animation_minor_mode(X) :-
   (animation_minor_mode(X) -> true ; assert(animation_minor_mode(X))).


:- dynamic currently_opened_file_status/2.
:- volatile currently_opened_file_status/2.
%currently_opened_file(none).

reset_currently_opened_file :-  retractall(currently_opened_file_status(_,_)).
reset_specfile :- reset_currently_opened_file,
   retractall(animation_minor_mode(_)).
:- use_module(eventhandling,[register_event_listener/3]).
:- register_event_listener(clear_specification,reset_specfile,
                           'Reset Specfile Currently Opened File.').

% call if a spec was not loaded from file but from package (e.g., via socket / prob2)
set_currently_opened_package(Type) :-
  set_currently_opened_file(package(Type)).

% indicates we are in the process of opening a file
set_currently_opening_file(File) :-
  reset_currently_opened_file,
  assert(currently_opened_file_status(File,opening)).

set_currently_opened_file(File) :-
  reset_currently_opened_file,
  assert(currently_opened_file_status(File,success)).

set_failed_to_open_file(File) :-
  reset_currently_opened_file,
  assert(currently_opened_file_status(File,failure)).

currently_opened_file(File) :- currently_opened_file_status(File,success).

spec_file_has_been_successfully_loaded :- currently_opened_file(_).

:- use_module(bmachine,[b_machine_has_variables/0]).
% true if state corresponds to an initialised B machine
state_corresponds_to_initialised_b_machine(const_and_vars(_,_)) :- !,
   b_or_z_mode.
state_corresponds_to_initialised_b_machine([]).
state_corresponds_to_initialised_b_machine([_|_]).
state_corresponds_to_initialised_b_machine(csp_and_b(_,BState)) :-
   csp_with_bz_mode,
   state_corresponds_to_initialised_b_machine(BState).


% check if state corresponds to an initialised B machine, and if so return expanded State
state_corresponds_to_initialised_b_machine(const_and_vars(ConstID,Vars),State) :- !,
   b_or_z_mode, expand_const_and_vars(ConstID,Vars,State).
state_corresponds_to_initialised_b_machine(expanded_const_and_vars(_ConstID,_Vars,FullStore),State) :- !,
   % generated by prepare_state_for_specfile_trans
   State=FullStore.
state_corresponds_to_initialised_b_machine([],[]).
state_corresponds_to_initialised_b_machine([H|T],[H|T]).
state_corresponds_to_initialised_b_machine(csp_and_b(_,BState),State) :-
   csp_with_bz_mode,
   state_corresponds_to_initialised_b_machine(BState,State).

% check if all variables, constants setup; also allows concrete_constants if no variables exist
state_corresponds_to_fully_setup_b_machine(concrete_constants(_)) :- !, b_or_z_mode,
   \+ b_machine_has_variables.
state_corresponds_to_fully_setup_b_machine(X) :- state_corresponds_to_initialised_b_machine(X).

state_corresponds_to_fully_setup_b_machine(concrete_constants(Constants),State) :- !, b_or_z_mode,
   \+ b_machine_has_variables, State=Constants.
state_corresponds_to_fully_setup_b_machine(X,State) :- state_corresponds_to_initialised_b_machine(X,State).

% check if we only have constants, no variable values yet
state_corresponds_to_set_up_constants_only(root,State) :- !, animation_mode(b),
   \+ b_machine_has_constants_or_properties, State=[]. % relevant for test 960
state_corresponds_to_set_up_constants_only(concrete_constants(Constants),State) :- !,
   b_or_z_mode, State=Constants.
state_corresponds_to_set_up_constants_only(csp_and_b(_,BState),State) :- !,
   csp_with_bz_mode,
   state_corresponds_to_set_up_constants_only(BState,State).

state_corresponds_to_set_up_constants(const_and_vars(_,_)) :- !. % avoid expansion
state_corresponds_to_set_up_constants(X) :- state_corresponds_to_set_up_constants(X,_).

state_corresponds_to_set_up_constants(const_and_vars(ConstID,Vars),State) :- !,
   b_or_z_mode, expand_const_and_vars(ConstID,Vars,State).
state_corresponds_to_set_up_constants(expanded_const_and_vars(_ConstID,_Vars,FullStore),State) :- !,
   % generated by prepare_state_for_specfile_trans
   State=FullStore.
state_corresponds_to_set_up_constants(concrete_constants(Constants),State) :- !,
   b_or_z_mode, State=Constants.
state_corresponds_to_set_up_constants(csp_and_b(_,BState),State) :- !,
   csp_with_bz_mode,
   state_corresponds_to_set_up_constants(BState,State).
state_corresponds_to_set_up_constants([],State) :- !, State=[],
   animation_mode(b).
state_corresponds_to_set_up_constants(State,RState) :-
   animation_mode(b),
   (State = root -> \+ b_machine_has_constants_or_properties,RState=[] ; RState=State).

:- use_module(state_space,[current_expression/2]).
current_b_expression(CurBState) :-
        current_expression(_CurID,CurState),
        state_corresponds_to_set_up_constants(CurState,CurBState).



:- use_module(state_space, [current_state_id/1]).
:- use_module(bmachine, [determine_type_of_formula/2]).
% get a state sufficient to evaluate a given formula (expression or predicate)
% it will give an error is the machine is not sufficiently initialised
get_current_state_for_b_formula(TypedExpr,BState) :-
    current_state_id(CurID),
    get_state_for_b_formula(CurID,TypedExpr,BState).
get_state_for_b_formula(StateID,TypedExpr,BState) :-
    determine_type_of_formula(TypedExpr,Type),
    get_state(Type,StateID,TypedExpr,BState).


get_state(requires_nothing,_ID,_TE,[]).
get_state(requires_constants,StateID,TypedExpr,BState) :-
    my_visited_expression(StateID,State),
    (state_corresponds_to_set_up_constants(State,BState) -> true ;
      add_error(get_state,'Please setup constants or initialise machine first, cannot evaluate:',TypedExpr,TypedExpr),fail).
get_state(requires_variables,StateID,TypedExpr,BState) :-
    my_visited_expression(StateID,State),
    (state_corresponds_to_initialised_b_machine(State,BState) -> true ;
      %translate_bexpression_with_limit(Cond,100,CS),
      add_error(get_state,'Please initialise machine first, cannot evaluate:',TypedExpr,TypedExpr),fail).

my_visited_expression(root,S) :- !,
     S=root. % ensure we succeed even if no state space set (e.g. when no machine provided to probcli)
my_visited_expression(StateID,State) :- visited_expression(StateID,State).

/* TRANSITIONS */

:- use_module(xtl_interface).
:- use_module(tools_meta,[call_residue/2]).
% ---------------------------------------------------


% Tries to find successor state for Op in State DB. Fails, if no solution is possible
% Input: State DB, Operation Op
% Output: DB2 the sucessor state
% Residue: Unfinished co-routines (should not happen!)
%specfile_trans(DB,Op,DB2,[]) :- !, specfile:i_transition(DB,Op,DB2).
specfile_trans(DB,OpName,Op,DB2,TransInfo,Residue) :-
%       i_transition(DB,OpName,Op,DB2,TransInfo), Residue = []. % is faster, but no residue checking
         animation_mode(MODE),
         call_residue(i_transition(MODE,DB,OpName,Op,DB2,TransInfo),Residue).

i_transition(b,State,OpName,Op,NewState,TransInfo) :- !,
    b_trans(State,OpName,Op,NewState,TransInfo).
i_transition(xtl,State,OpName,Op,NewState,[]) :- !,
    xtl_transition(State,Op,NewState), functor(Op,OpName,_).
i_transition(cspm,State,_OpName,Op,NewState,[]) :- !,
    % Note: the OpName is ignored: we cannot make use of it
    cspm_transition(State,Op,NewState).
%i_transition(promela,State,_OpName,Op,NewState,[]) :- !,
%    promela_transition(State,Op,NewState).
 % ; MODE=smv         -> smv_transition(State,Op,NewState),TransInfo=[]
i_transition(csp_and_b,State,OpName,Op,NewState,TransInfo) :- !,  %%  print(cspB(OpName,Op,State)),nl, %trace,
    % Note: the OpName is either the name of a B operation or '$CSP' (or a variable)
    csp_and_b_transition(State,OpName,Op,NewState,TransInfo).


% if you call specfile_trans multiple times for the same state, it is
% more efficient to call this predicate first once
prepare_state_for_specfile_trans(const_and_vars(ConstID,Vars),R) :- !, %print(expanding_const(ConstID)),nl,
      R = expanded_const_and_vars(ConstID,Vars,FullStore),
      expand_const_and_vars(ConstID,Vars,FullStore). %, print(full(FullStore)),nl.
% TO DO: also expand for csp_and_b
prepare_state_for_specfile_trans(X,X).

% a version with a memo argument to store expanded constants
prepare_state_for_specfile_trans(const_and_vars(ConstID,Vars),MEMO,R) :- !, %print(expanding_const(ConstID)),nl,
      R = expanded_const_and_vars(ConstID,Vars,FullStore),
      (nonvar(MEMO),MEMO = [ConstID/ConstantsStore]
        -> %tools_printing:print_term_summary(memo(ConstID,ConstantsStore)),nl,
           append(Vars,ConstantsStore,FullStore) % simply reuse expanded constants in MEMO
           % used to be constants first, but has to be compatible with b_initialise_machine2 and expand_const_and_vars/3
        ;  % expand_const_and_vars/3 inlined (so as to obtain ConstantsStore
          state_space:visited_expression(ConstID,concrete_constants(ConstantsStore)), %% This can take a while if the Constants are big !
           % print(no_memo(ConstID,MEMO)),tools_printing:print_term_summary(ConstantsStore),nl,
           append(Vars,ConstantsStore,FullStore),
           (var(MEMO) -> MEMO=[ConstID/ConstantsStore] ; true)
      ). %, print(full(FullStore)),nl.
% TO DO: also expand for csp_and_b
prepare_state_for_specfile_trans(X,_,X).


/* can be attempted if trans fails to attempt to use partially computed transitions */
partial_trans(DB,Op,DB2,Residue) :-
  call_residue(partial_i_transition(DB,Op,DB2),Residue).

partial_i_transition(root,'$partial_setup_constants',concrete_constants(ConstantsStore)) :-
     animation_mode(b),
     set_error_context(operation('$partial_setup_constants',root)),
     b_interpreter:b_partial_set_up_concrete_constants(ConstantsStore).

% a predicate which automatically calls partial_trans :
specfile_trans_or_partial_trans(DB,OpName,Op,DB2,TransInfo,Residue,Partial) :-
   if(specfile_trans(DB,OpName,Op,DB2,TransInfo,Residue),Partial=false,
      (OpName='$setup_constants', % OpName is often an input argument here
       TransInfo=[], Partial = true,
       partial_trans(DB,Op,DB2,Residue))).

/* build up skeletons of possible operations; to be used before time_out call is made */
specfile_possible_trans_name(State,OpName) :- animation_mode(MODE),
  (MODE=b -> b_possible_trans_name(State,OpName)
   ; MODE = csp_and_b -> csp_b_possible_trans_name(State,OpName)
   ; setup_transition_skeleton(MODE,OpName) /* TO DO: add more possible skeletons */
  ).

% only set up skeletons if it is worth it for computing all successors
% e.g., currently it is not worth it to setup skeletons for either CSP or CSP||B (due to renaming,... we cannot predict which parts of a CSP process can be excluded)
specfile_possible_trans_name_for_successors(State,OpName) :- animation_mode(MODE),
  (MODE=b -> b_possible_trans_name(State,OpName)
   ; setup_transition_skeleton(MODE,OpName)
  ).

setup_transition_skeleton(_,_).  /* TO DO: add more possible skeletons ? */

:- use_module(library(random)).
b_possible_trans_name(root,OpName) :- !,
 (b_machine_has_constants_or_properties
   -> OpName = '$setup_constants'
   ;  OpName = '$initialise_machine').
b_possible_trans_name(concrete_constants(_),OpName) :- !, OpName = '$initialise_machine'.
b_possible_trans_name(_State,OpName) :-
   (preferences:preference(randomise_operation_order,true),random(1,3,1)
      -> b_machine_operation_names_in_reverse_order(OpName) ; true),
   %b_get_machine_operation(OpName,_Res,_Params,_,_,_).
   b_top_level_feasible_operation(OpName). % TO DO: we could link also with cbc feasibility analysis if carried out
   % we could check: b_operation_cannot_modify_state(OpName) given a preference

csp_b_possible_trans_name(csp_and_b_root,OpName) :- !,
  (b_machine_has_constants_or_properties
   -> OpName = 'tau($setup_constants)'
   ;  OpName = 'tau($initialise_machine)').
csp_b_possible_trans_name(concrete_constants(_),OpName) :- !,
  OpName = 'tau($initialise_machine)'.
csp_b_possible_trans_name(State,OpName) :-
  b_possible_trans_name(State,OpName).
  % TO DO: add non-synchronised channels

% --------------

:- use_module(probcspsrc(haskell_csp),[channel_type_list/2]).
:- use_module(debug).
% check whether a currently open CSP is type compatible with a loaded B machine
type_check_csp_and_b :- animation_mode(csp_and_b), debug_println(9,type_check_csp_and_b),
   b_get_machine_operation(ChannelOpName,Res,Params,_,_,_), append(Params,Res,ParamsRes),
   channel_type_list(ChannelOpName,CSPTypeList),
   % TO DO: also check B variable names which match channel name
   debug_println(9,checking_synchronised_channel(ChannelOpName,ParamsRes,CSPTypeList)),
   l_check_compatibility(ParamsRes,CSPTypeList,ChannelOpName),
   fail.
type_check_csp_and_b.

%convert_csp_type_to_list(type(T),R) :- convert_csp_type_to_list2(T,R).
%convert_csp_type_to_list2(dotUnitType,[]).
%convert_csp_type_to_list2(dotTupleType(T),T).

l_check_compatibility([],[],_) :- !.
l_check_compatibility([B|BT],[CSP|CSPT],ChannelOpName) :- !, check_compatibility(B,CSP,ChannelOpName),
    l_check_compatibility(BT,CSPT,ChannelOpName).
l_check_compatibility([],CSP,ChannelOpName) :- !,
    add_error(type_check_csp_and_b,'CSP Channel has too many parameters: ', ChannelOpName:CSP).
l_check_compatibility(B,[],ChannelOpName) :- !,
    print('* CSP Channel has too few parameters: '), print(ChannelOpName),nl,
    print('* B Arguments will be ignored: '), l_print_bexpr_or_subst(B),nl.
    %add_error(type_check_csp_and_b,'CSP Channel has too few parameters: ', ChannelOpName:B).

check_compatibility(b(identifier(ID),TYPE,INFO),CSP_TYPE,ChannelOpName) :- !,
   (type_ok(TYPE,CSP_TYPE) -> true ;
   % TO DO: pretty print B and CSP Types
     add_error(type_check_csp_and_b,'Incompatible types between B and CSP: ',ChannelOpName:ID:(TYPE:CSP_TYPE),b(identifier(ID),TYPE,INFO))
   ).
check_compatibility(B,CSP,ChannelOpName) :- !,
   add_internal_error('Illegal types: ',check_compatibility(B,CSP,ChannelOpName)).

% TO DO: needs to be refined much more;
type_ok(Type,CSPType) :-
  (type_ok2(Type,CSPType) -> true
    ; add_error(type_check_csp_and_b,'Cannot be converted (B:CSP): ',Type:CSPType),fail).
type_ok2(boolean,X) :- !, is_csp_boolean_type(X).
type_ok2(integer,X) :- !, is_csp_integer_type(X).
type_ok2(string,CSP) :- !,is_csp_global_set_type(CSP,string).
type_ok2(set(T),CSP) :- !, is_csp_set_type(CSP,T).
% TO DO: what if we have set(couple(integer,T)) instead of seq(T) ?
% rye
% check set(couple(integer,T)) (seq type) in B
type_ok2(seq(T),CSP) :- !, is_csp_seq_type(CSP,T).
type_ok2(couple(A,B),CSP) :- !, is_csp_couple_type(CSP,A,B).
type_ok2(global(GS),CSP) :- !,is_csp_global_set_type(CSP,GS).
type_ok2(X,Y) :- print(unkown(X,Y)),nl.

is_csp_couple_type(setValue([na_tuple([A|Rest])|_]),TA,TB) :- type_ok(TA,setValue([A])),
     check_rest_couple_els(Rest,TB).
is_csp_couple_type(typeTuple([A|Rest]),TA,TB) :- type_ok(TA,A), check_rest_couple(Rest,TB).

check_rest_couple_els([B],TB) :- !, type_ok(TB,setValue([B])).
check_rest_couple_els([B1|BRest],couple(TB1,TBRest)) :- type_ok(TB1,setValue([B1])),
    check_rest_couple_els(BRest,TBRest).
check_rest_couple([B],TB) :- !, type_ok(TB,B).
check_rest_couple([B1|BRest],couple(TB1,TBRest)) :- type_ok(TB1,B1),
    check_rest_couple(BRest,TBRest).

% rye
% check if it is couple(integer, T) type and return T in 2nd arguments
couple_int(couple(integer, T), T). % possible sequence

% rye: it's not correct to check only Head
% % for example, for CSP type, {{}, {(1,2),(2,3)}}. Its head is emptyset, but the second one is set of tuple
is_csp_set_type('Set'([H|_]),Type) :- type_ok(Type,H). % just check Head
% check the 2nd element in set in case the 1st is emptyset
is_csp_set_type(setValue([_|T]),Type) :-
     T = [H1 | _], !,
     % if it is set(couple(integer, T)), it can be a sequence type
     (couple_int(Type, TH1) -> is_csp_seq_type(setValue(T), TH1)
        ; type_ok(Type,H1)
     ). % check the 2nd
% if it is only one element in set, just check the head
is_csp_set_type(setValue([H|T]),Type) :-
    (couple_int(Type, TH1) -> is_csp_seq_type(setValue([H|T]), TH1)
        ; type_ok(Type,H)
    ). % just check Head

is_csp_seq_type('Seq'(H),Type) :- type_ok(Type,H).
is_csp_seq_type(setValue([H|_]),Type) :- H=list(L), is_csp_list(L,Type).
is_csp_list([],_).
is_csp_list([H|_],Type) :- type_ok(Type,setValue([H])).

is_csp_boolean_type(boolType).
is_csp_boolean_type(setValue([H|_])) :- is_csp_boolean_value(H).
is_csp_boolean_value(true).
is_csp_boolean_value(false).

is_csp_integer_type(intType).
is_csp_integer_type(setFromTo(_,_)).
is_csp_integer_type(setFrom(_)).
is_csp_integer_type(setValue([int(_)|_])).

is_csp_global_set_type(dataType(_),_GS). % TO DO CHECK MEMBERS
is_csp_global_set_type(setValue([]),_GS). % TO DO CHECK MEMBERS
is_csp_global_set_type(setValue([_|_]),_GS). % TO DO CHECK MEMBERS

% --------------


:- use_module(bmachine,[b_get_machine_operation/4]).
:- use_module(bsets_clp,[tuple_of/3]).
% compute transitions for a CSP || B specification
% Note: the OpName is either the name of a B machine operation (even if hidden) or '$CSP' (useful for ltsmin)
csp_and_b_transition(root,OpName,Op,NewState,TransInfo) :- !,
  Op = start_cspm_MAIN, OpName=Op, NewState = csp_and_b_root,TransInfo = [].
csp_and_b_transition(csp_and_b_root,OpName,Trans,NewState,TransInfo) :- !,
  Trans = tau(Op), OpName = '$CSP',
  b_trans(root,_,Op,InitialBState,TransInfo),
  (InitialBState = concrete_constants(_)
    -> NewState = InitialBState
    ;  csp_initialisation_for_b(InitialCSPState),
       NewState = csp_and_b(InitialCSPState,InitialBState)
  ).
csp_and_b_transition(concrete_constants(C),OpName,Trans,NewState,TransInfo) :-  !,
  Trans = tau(Op), OpName = '$CSP',
  b_trans(concrete_constants(C),_,Op,InitialBState,TransInfo),
  csp_initialisation_for_b(InitialCSPState),
  NewState = csp_and_b(InitialCSPState,InitialBState).
csp_and_b_transition(csp_and_b(CSPState,BState),OpName,EventVisibleToUser,csp_and_b(NewCSP,NewB),TransInfo) :-
  !,
  csp_transition_for_b(CSPState,ChOpName,ChArgs,CSPAction,NewCSP),
  % TO DO: split up csp_transition_for_b: into two parts; second part converts datatypes and only called if linking op exists
  %%  print(csp_trans(ChOpName,ChArgs)),nl, %%
  length(ChArgs,Len), length(OpArgs,Len),
  b_transition_for_csp(ChOpName,CSPAction,ChArgs,OpArgs,BState,OpName,Operation,NewB,TransInfo),
  hide_csp_event(CSPAction,Operation,EventVisibleToUser).
csp_and_b_transition(State,'$CSP',Op,NewState,[]) :-
   % we have a pure CSP state: use CSP-M transition
   cspm_transition(State,Op,NewState).

:- use_module(bmachine,[b_is_variable/1,b_is_constant/1]).
% compute the B counterpart for a CSP transition:
b_transition_for_csp(ChOpName,_CSPAction,ChArgs,OpArgs,BState,OpName,Operation,NewB,TransInfo) :-
        build_up_b_operation_for_csp(ChOpName,OpArgs,Operation),
        !,
        OpName = ChOpName,
        %%  print_message(trying_b_trans(Operation,ChArgs,OpArgs)), %%
        generate_b_operationargs_from_csp(ChArgs,OpArgs),
        %% print(trying_b_trans(Operation)),nl, %%
        b_trans(BState,_,Operation,NewB,TransInfo).
        %%,   print_message(csp_btrans(Operation)), %%.
b_transition_for_csp(ChOpName,_CSPAction,ChArgs,OpArgs,BState,OpName,Operation,NewB,TransInfo) :-
        OpName = ChOpName,
        (b_is_variable(ChOpName) ; b_is_constant(ChOpName)),
         %% print(probing(ChOpName)),nl,
        expand_const_and_vars_to_full_store(BState,FBState),
        try_lookup_value_in_store_and_global_sets(ChOpName,FBState,IdValue),
        !,
        NewB=BState,TransInfo=[],
        /* interpret as probing operation: get value of var,cst,SET */
        generate_b_operationargs_from_csp(ChArgs,OpArgs),
        (OpArgs=[SingleVal]
		  ->  SingleVal = IdValue, Operation='-->'(ChOpName,[IdValue])
		  ;   OpArgs=[Val1,Val2],
			  tuple_of(Val1,Val2,IdValue), /* from bsets_clp, use exact_element_of ??? */
			  OpCall =.. [ChOpName,Val1],
			  Operation='-->'(OpCall,[Val2])
		).
b_transition_for_csp(_ChOpName,CSPAction,_ChArgs,_OpArgs,BState,OpName,Operation,NewB,TransInfo) :-
       OpName = '$CSP', % all CSP events grouped together
	   NewB=BState,TransInfo=[],
	   /* no B operation of the name; could be tau or tick or extra comm channel */
	   %Op1 =.. [ChOpName|OpArgs],
	   Operation = CSPAction.  % removed csp(.) wrapper [for LTL model checker]

hide_csp_event(CSPAction,_,EventVisibleToUser) :-
   haskell_csp:symbol('HIDE_CSPB',_,Span,_),!, % if there is a HIDE_CSPB definition then we do MAIN [{| HIDE_CSPB |}] B_MACHINE \ {| HIDE_CSPB |}
   haskell_csp:evaluate_argument(val_of('HIDE_CSPB',Span),EvCList),
   % print(hide(EvCList,CSPAction)),nl,
   haskell_csp:expand_channel_pattern_expression(EvCList,ECList,no_loc_info_available),
   haskell_csp:cspm_hide_action(CSPAction,omega,ECList, Span, EventVisibleToUser,_).
hide_csp_event(_CSPAction,Operation,EventVisibleToUser) :-
  % get_preference(csp_event_visible_to_user,true) -> EventVisibleToUser = CSPAction
     EventVisibleToUser = Operation.

build_up_b_operation_for_csp(OpName,OpArgs,Operation) :-  %print(csp(OpName,OpArgs,Operation)),nl,
  nonvar(OpName),
  b_get_machine_operation(OpName,Res,Params,_), %print(get_machine_op(OpName,Res,Params)),nl,
  specfile:generate_args(OpArgs,Params,Res, BArgs,BResults,OpName), %print(gen_args(BArgs,BResults)),nl,
  % TO DO: either also extract types from Params, Res or better write a static checker that tests that the B operation and CSP channel types are compatible
  safe_univ(Op1,[OpName|BArgs]), %print(op1(Op1)),nl,
  (BResults = []  -> Operation = Op1  ;  Operation =.. ['-->',Op1,BResults]).

:- use_module(bmachine,[b_get_machine_operation_for_animation/6]).
build_up_b_real_operation(OpName,Operation) :-
  b_get_machine_operation_for_animation(OpName,Res,Params,_,_,true), %true==TopLevel
  length(Params,Len), length(OpArgs,Len),
  length(Res,RLen), length(BResults,RLen),
  safe_univ(Op1,[OpName|OpArgs]),
  (BResults = []  -> Operation = Op1  ;  Operation =.. ['-->',Op1,BResults]).

generate_args([],[],Res,[],BRes,_) :- length(Res,L), length(BRes,L).
generate_args([],[_Param|T],Res,[_|TA],BRes,OpName) :-
   generate_args([],T,Res,TA,BRes,OpName).
generate_args([CSPArg|TC],[Param|T],Res,[CSPArg|TA],BRes,OpName) :- check_type_for_id(CSPArg,Param,OpName),
   generate_args(TC,T,Res,TA,BRes,OpName).
generate_args([CSPArg|TC],[],[ReturnParam|TRes],[],[CSPArg|TBRes],OpName) :-
    check_type_for_id(CSPArg,ReturnParam,OpName),
   generate_args(TC,[],TRes,[],TBRes,OpName).
generate_args([H|T],[],[],[],[],OpName) :-
   add_error(specfile,'CSP provides extra arguments: ',OpName:[H|T]).

check_type_for_id(Val,TID,OpName) :- get_texpr_type(TID,Type),
   check_type(Type,Val,OpName:TID).

:- block check_type(?,-,?).
check_type(boolean,pred_true,_) :- !.
check_type(boolean,pred_false,_) :- !.
check_type(integer,int(_),_) :- !.
check_type(string,string(_),_) :- !.
check_type(global(GS),fd(_,GS),_) :- !.
check_type(couple(A,B),(VA,VB),OID) :- !, check_type(A,VA,OID), check_type(B,VB,OID).
check_type(set(_T),[],_) :- !.
check_type(set(_T),closure(_,_,_),_) :- !. % TO DO: check type signature of closure
check_type(set(T),[H|_],OID) :- !, check_type(T,H,OID).
check_type(set(T),avl_set((node(Y,_,_,_L,_R))),OID) :- !, check_type(T,Y,OID).
check_type(seq(T),V,OID) :- !, check_type(set(couple(integer,T)),V,OID).
% TO DO: record, freetype ; but do not appear in translation from CSP to B yet
check_type(Type,Value,OpName:TID) :-
   def_get_texpr_id(TID,ID),
   add_error(check_type_for_id,'Illegal CSP value for argument: ',OpName:ID:type(Type):value(Value)),
   fail.

get_operation_internal_name(Op,Name) :- var(Op),!,
    add_error(get_operation_internal_name,'Variable Transition: ',Op), Name=Op.
get_operation_internal_name('$initialise_machine',N) :- !, N='$initialise_machine'.
get_operation_internal_name('$setup_constants',N) :- !, N='$setup_constants'.
get_operation_internal_name('$partial_setup_constants',N) :- !, N='$partial_setup_constants'.
get_operation_internal_name(Op,Name) :- get_operation_name(Op,Name).

% tool to get the basic operation/channel/event name of a transition:
get_operation_name(Op,Name) :- var(Op),!, add_error(get_operation_name,'Variable Transition: ',Op), Name=Op.
get_operation_name('-->'(F,_),N) :- !, functor(F,N,_).
%get_operation_name(io([Action|_],proc(_Nr,_PROC),_SPAN),F) :-
%     animation_mode(promela),!,functor(Action,F,_).
get_operation_name(io(_V,Ch,_SPAN),N) :- csp_mode,!, functor(Ch,N,_).
get_operation_name(FullOperation,Name) :- functor(FullOperation,Functor,_),
    translate_operation_name(Functor,Name).

get_operation_arguments(Op,Args) :- var(Op),!, add_error(get_operation_arguments,'Variable Transition: ',Op), Args=[].
get_operation_arguments('-->'(F,_ReturnValues),Args) :- !, F =.. [_|Args].
%get_operation_arguments(io([_Action|V],proc(_Nr,_PROC),_SPAN),Args) :-
%     animation_mode(promela),!,Args=V.
get_operation_arguments(io(V,_Ch,_SPAN),Args) :- csp_mode,!, Args = V.
get_operation_arguments(FullOperation,Args) :- FullOperation =.. [_|Args].

get_operation_return_values_and_arguments('-->'(F,ReturnValues),ReturnValues,Args) :- !,
    F =.. [_|Args].
get_operation_return_values_and_arguments(Op,[],Args) :- get_operation_arguments(Op,Args).

translate_operation_name('$initialise_machine',T) :- !,T='INITIALISATION'.
translate_operation_name('$setup_constants',T) :- !,T='SETUP_CONSTANTS'.
translate_operation_name('$partial_setup_constants',T) :- !,T='PARTIAL_SETUP_CONSTANTS'.
%translate_operation_name(InternalName,Res) :- !, Res=InternalName.
translate_operation_name(X,X).





:- use_module(state_space,[visited_expression/2]).
:- use_module(library(between),[between/3]).

b_trans(State,OpName,Operation,NewState,PathInfo) :-
	 compute_operation_effect_max(State,OpName,Operation,NewState,PathInfo,_Max).

compute_operation_effect_max([],OpName,Operation,NewState,PathInfo,Max) :-
     compute_operation_on_expanded_store([],OpName,Operation,Updates,PathInfo,Max),
     store:store_updates_and_normalise(Updates,[],NewState).
compute_operation_effect_max([H|T],OpName,Operation,NewState,PathInfo,Max) :-
     compute_operation_on_expanded_store([H|T],OpName,Operation,Updates,PathInfo,Max),
     store:store_updates_and_normalise(Updates,[H|T],NewState).
compute_operation_effect_max(const_and_vars(ConstID,Vars),OpName,Operation,ResultingStore,PathInfo,Max) :-
     prepare_state_for_specfile_trans(const_and_vars(ConstID,Vars),R),
     %print(expanded_const_and_vars(ConstID)),nl,
     compute_operation_effect_max(R,OpName,Operation,ResultingStore,PathInfo,Max).
compute_operation_effect_max(concrete_constants(ConstantsStore),'$initialise_machine',OpInit,ResultingStore,PathInfo,Max) :-
     get_max_enablings_for_init(Max,'$initialise_machine',MaxForCall),
     succeed_max_call_id('$initialise_machine',
                    b_interpreter:b_initialise_machine(ConstantsStore,InitialVars,InitialStore,PathInfo),MaxForCall),
     create_initialisation_operation(InitialVars,OpInit),
     %user:do_one_gui_event,
     (\+ preferences:preference(symmetry_mode,flood), /* TO DO : improve permutation to allow using const_and_vars optimisation */
       visited_expression(ConstID,concrete_constants(ConstantsStore))
      -> ResultingStore = const_and_vars(ConstID,InitialVars)  /* avoid storing constant values again */
      ;  ResultingStore = InitialStore).
compute_operation_effect_max(expanded_const_and_vars(ConstID,Vars,FullStore),OpName,Operation,ResultingStore,PathInfo,Max) :-
     %print(const_and_vars(ConstID,Vars)),nl,
     compute_operation_on_expanded_store(FullStore,OpName,Operation,Updates,PathInfo,Max),
     ResultingStore = const_and_vars(ConstID,NewVars),
     store:store_updates_and_normalise(Updates,Vars,NewVars).
     %store:store_updates(Updates,Vars,NewVars). % we could use this when operation re-used / cached
     %ErrorMsg = compute_operation_effect(const_and_vars(ConstID,Vars),Operation,ResultingStore,PathInfo),
     %reconstruct_constants_store(ConstID,ConstantsStore,NewState,ErrorMsg,ResultingStore).
compute_operation_effect_max(root,'$setup_constants',OpSetup,concrete_constants(FilteredConstantsStore),[],Max) :-
     b_machine_has_constants_or_properties,
     get_max_enablings_for_init(Max,'$setup_constants',MaxForCall),
     compute_constants(ConstantsStore,MaxForCall),
     create_setup_constants_operation(ConstantsStore,OpSetup),
     %%print_message('FOUND_CONSTANTS'(OpName)),
     (get_preference(filter_unused_constants,true)
       -> exclude(unused_binding,ConstantsStore,FilteredConstantsStore)
       ;  FilteredConstantsStore=ConstantsStore).
     %user:do_one_gui_event.
compute_operation_effect_max(root,'$initialise_machine',OpInit,InitialStore,PathInfo,Max) :-
     \+ b_machine_has_constants_or_properties,
     get_max_enablings_for_init(Max,'$initialise_machine',MaxForCall),
     %user:do_one_gui_event, user:do_one_gui_event, user:do_one_gui_event,
     empty_state(EmptyState),
     succeed_max_call_id('$initialise_machine',
                     b_interpreter:b_initialise_machine(EmptyState,InitialVars,InitialStore,PathInfo),MaxForCall),
     create_initialisation_operation(InitialVars,OpInit).
     %user:do_one_gui_event.

unused_binding(bind(C,_)) :- bmachine:b_is_unused_constant(C).

compute_constants(ConstantsStore,_Max) :-
    % check if values have been computed and stored into a file before -- if yes, use them
    load_constants(Stores,MaxReached),!,
    choose_loaded_solution(Stores,ConstantsStore,MaxReached).
compute_constants(ConstantsStore,Max) :-
    % print('Searching for valid CONSTANTS'),nl,!,
    succeed_max_call_id('$setup_constants',b_interpreter:b_set_up_concrete_constants(ConstantsStore),Max).

choose_loaded_solution(Stores,ConstantsStore,_MaxReached) :-
    member(ConstantsStore,Stores).
choose_loaded_solution(_Stores,_ConstantsStore,true) :-
    % provoke the setting of the "maximum reached flag"
    succeed_max_call_id('$setup_constants',member(_,_),1),fail.

expand_const_and_vars_to_full_store(root,R) :- !,R=[].
expand_const_and_vars_to_full_store(concrete_constants(ConstantsStore),FullStore) :- !,
    ConstantsStore = FullStore.
expand_const_and_vars_to_full_store(const_and_vars(ConstID,Vars),FullStore) :- !,
    expand_const_and_vars(ConstID,Vars,FullStore).
expand_const_and_vars_to_full_store(expanded_const_and_vars(_ConstID,_Vars,ExpStore),FullStore) :- !,
    FullStore = ExpStore.
expand_const_and_vars_to_full_store(R,R).

expand_const_and_vars(ConstID,Vars,FullStore) :-
    visited_expression(ConstID,concrete_constants(ConstantsStore)),
    %% This can take a while if the Constants are big !
     % print('exp: '),tools_printing:print_term_summary(ConstantsStore),nl,
    append(Vars,ConstantsStore,FullStore). % used to be Constants first, now consistent with b_initialise_machine2,
    % b_initialise_machine2 puts constants at end to make sharing of complete tail-list easier for successor states

expand_to_constants_and_variables(root,[],[]).
expand_to_constants_and_variables(concrete_constants(ConstStore),ConstStore,[]).
expand_to_constants_and_variables(const_and_vars(ConstID,VarStore),ConstStore,VarStore) :-
    visited_expression(ConstID,concrete_constants(ConstStore)).
expand_to_constants_and_variables([],[],[]).
expand_to_constants_and_variables([H|T],[],[H|T]).

:- use_module(b_operation_cache,[compute_operation_on_expanded_store_cache/5]).

% compute operation effect on a store where all constants have been put into the environment
compute_operation_on_expanded_store(InState,OpName,Operation,NewState,PathInfo,Max) :-
     get_preference(try_operation_reuse,full),
     animation_mode(b), % doesn't seem to work with CSP||B yet; see CSPB/LTSMin/Counters.mch + .csp
     \+ animation_minor_mode(eventb),  %  projection seems to confuse Event-B interpreter, see e.g. prob_examples/public_examples/EventBPrologPackages/ABZ_Landing_Gear/Ref3_ControllerSensors_mch.eventb
     % TO DO: fix
     !,
     get_max_enablings_per_operation(Max,OpName,MaxForCall),
     succeed_max_call_id(OpName,
         compute_operation_on_expanded_store_cache(OpName,Operation,InState,NewState,PathInfo),
         MaxForCall).
compute_operation_on_expanded_store(InState,OpName,Operation,NewState,PathInfo,Max) :-
     get_max_enablings_per_operation(Max,OpName,MaxForCall),
     succeed_max_call_id(OpName,
         b_interpreter:b_execute_top_level_operation_update(OpName,Operation,InState,NewState,PathInfo),
         MaxForCall).


create_initialisation_operation(_InitialVars,'$initialise_machine').

create_setup_constants_operation(_ConstantVars,'$setup_constants').

:- use_module(bmachine,[b_get_machine_operation_max/2]). % MAX_OPERATIONS_... DEFINITIONS
get_max_enablings_per_operation_aux(OpName,RMax,RandomisedRestart) :-
   b_get_machine_operation_max(OpName,Max),
   !,
   (Max<0 -> RMax is -Max, RandomisedRestart=true ; RMax=Max, RandomisedRestart=false).
get_max_enablings_per_operation_aux(_OpName,Max,false) :-
   preferences:preference(maxNrOfEnablingsPerOperation,Max).


get_max_enablings_per_operation(Max,OpName,MaxForCall) :-
     (var(Max) -> get_max_enablings_per_operation_aux(OpName,Max,RandomisedRestart) ; RandomisedRestart=false),
      (RandomisedRestart=true,
       get_preference(randomise_enumeration_order,true)
       -> % we succeed Max times with the value 1, thus forcing a randomised restart
          MaxForCall=1, between(1,Max,Retry), format('Randomised Restart ~w for ~w~n',[Retry,OpName])
       ; % we succeed once with the value Max
          MaxForCall is Max).

get_max_enablings_for_init(Max,Op,MaxForCall) :-
     (var(Max) -> get_preference(maxNrOfInitialisations,Max) ; true),
      (Max>1,
       get_preference(randomisedRestartInitalisations,true),
       get_preference(randomise_enumeration_order,true)
       -> % we succeed Max times with the value 1, thus forcing a randomised restart
          MaxForCall=1, between(1,Max,Retry), format('Randomised Restart ~w for ~w~n',[Retry,Op])
       ; % we succeed once with the value Max
          MaxForCall=Max).

/* PROPERTIES */

property(const_and_vars(ConstID,Vars),Property) :- b_or_z_mode,
    expand_const_and_vars(ConstID,Vars,FullStore),!,
    property(FullStore,Property).
property(csp_and_b(CSPState,BState),Property) :- animation_mode(csp_and_b),!,
    %% print(checking_csp_and_b(CSPState)),nl,
    expand_const_and_vars_to_full_store(BState,FBState),
   (xtl_interface:cspm_property(CSPState,Property) ;
    b_property(FBState,Property)).
property(State,Property) :- animation_mode(AM),
   (AM = b -> b_property(State,Property)
    ; AM=xtl -> xtl_interface:xtl_property(State,Property)
    ; AM=cspm -> xtl_interface:cspm_property(State,Property)
    ; AM=csp_and_b -> b_property(State,Property) % State should be root
   ).

%b_property(const_and_vars(ConstID,Vars),Prop) :- !,
%   expand_const_and_vars(ConstID,Vars,FullStore),
%   b_property(FullStore,Prop).
b_property(root,Prop) :- !,  b_preference_par(PARA,Op,VAL), Prop=.. [Op,PARA,VAL].
b_property([H|T],non_ground(Var)) :- debug:debug_mode(on),
         member(bind(Var,Val),[H|T]),\+(ground(Val)).
b_property([H|T],Prop) :- elementary_prop([H|T],Prop).
b_property([],Prop) :- elementary_prop([],Prop).
b_property(concrete_constants(DB),Prop) :- elementary_prop(DB,Prop).



:- use_module(preferences).
:- use_module(b_global_sets,[b_global_set/1, b_get_fd_type_bounds/3,inferred_minimum_global_set_cardinality/2,
                             inferred_maximum_global_set_cardinality/2, unfixed_deferred_set/1]).

b_preference_par('MAXINT','=',MaxInt) :- get_preference(maxint,MaxInt).
b_preference_par('MININT','=',MinInt) :- get_preference(minint,MinInt).
b_preference_par(card(GlobalSet),Op,Card) :-
    b_global_set(GlobalSet),
    b_get_fd_type_bounds(GlobalSet,Low,Up), RCard is Up-Low+1,
    % provide feedback if unfixed_deferred_set
    (inferred_maximum_global_set_cardinality(GlobalSet,MaxCard),
      (inferred_minimum_global_set_cardinality(GlobalSet,MinCard) -> true ; MinCard=1),
      MinCard \= MaxCard,
      Op = ':', Card = '..'(MinCard,MaxCard)

    ;

     inferred_minimum_global_set_cardinality(GlobalSet,MinCard),
     \+ (inferred_maximum_global_set_cardinality(GlobalSet,_)),
      Op='>=',Card=MinCard % show minimum cardinality info

      ;

     Op= '=', (unfixed_deferred_set(GlobalSet)
               -> ajoin([RCard,' (assumed for deferred set)'],Card) ; Card=RCard)
    ).

/* --------------------------------------------------------- */


elementary_prop(DB,Prop) :-
    member(bind(Var,Val),DB), %print(var_prop(Var)),nl_time,
    elementary_prop3(Val,Var,Prop).

:- use_module(custom_explicit_sets,[is_avl_relation/1,is_avl_partial_function/1,get_first_avl_elements/4, avl_approximate_size/3]).
elementary_prop3(Val,Var,Prop) :- var(Val),!, Prop = '='(Var,'_VAR_').
elementary_prop3(avl_set(Avl),Var,Prop) :- !,
   (show_avl_set(Avl) % \+ custom_explicit_sets:is_avl_sequence(Avl))
     -> elementary_fun_prop(avl_set(Avl),Var,Prop)
      ; elementary_var_prop(avl_set(Avl),Var,Prop)).
elementary_prop3([H|T],Var,Prop) :- H=(_,_),
   get_preference(show_function_tuples_in_property,true),!,
   elementary_fun_prop([H|T],Var,Prop).
elementary_prop3(Val,Var,Prop) :- elementary_var_prop(Val,Var,Prop).

show_avl_set(AVL) :- get_preference(show_function_tuples_in_property,true),
   is_avl_relation(AVL),
   avl_approximate_size(AVL,0,Size), Size < 257, % only the first 30 will be shown by get_relation_element anyway
   is_avl_partial_function(AVL).


:- use_module(bmachine,[b_is_variable/2]).
elementary_var_prop(Val,Var,Prop) :-
    % TODO: use translate:translate_bvalue_for_expression/3
    animation_minor_mode(tla),!,
    ( identifier_has_tla_type(Var,TlaType) ->
        translate_bvalue_with_tlatype(Val,TlaType,Text)
    ; otherwise ->
        translate_bvalue(Val,Text)),
    Prop = '='(Var,Text).
elementary_var_prop(Val,Var,Prop) :-
    (var_cst_type(Var,Type)
     -> translate_bvalue_with_type_and_limit(Val,Type,320,Text)
     ;  translate_bvalue_with_type_and_limit(Val,any,320,Text)),  % translate_properties_with_limit uses 320 Limit
    Prop = '='(Var,Text).

identifier_has_tla_type(Id,Type) :-
    get_texpr_id(TId,Id),
    (b_get_machine_constants(Ids)
    ;b_get_machine_variables(Ids)),
    member(TId,Ids),!,
    get_texpr_info(TId,Infos),
    memberchk(tla_type(Type),Infos).
% to do: use b_is_variable + translate_bvalue_with_type
elementary_fun_prop(Val,Var,Prop) :-
    nonvar(Val),
    (var_cst_type(Var,Type),dom_range_type(Type,Dom,Ran) -> true ; Type=any,Dom=any, Ran=any),
    get_relation_element(Val,X,Y), % Show value in form:  Var(X) = Y
    create_texpr(identifier(Var),Type,[],Fun),
    create_value(X,Dom,Arg),
    create_texpr(function(Fun,Arg),Ran,[],Lhs),
    create_value(Y,Ran,Rhs),
    create_texpr(equal(Lhs,Rhs),pred,[],Expr),
    translate_bexpression_with_limit(Expr,80,Prop).

var_cst_type(Name,Type) :- (b_is_variable(Name,Type) ; b_is_constant(Name,Type)).
dom_range_type(seq(R),integer,R).
dom_range_type(set(couple(D,R)),D,R).

get_relation_element(avl_set(A),X,Y) :- custom_explicit_sets:is_avl_relation(A),
    !, %preferences:preference(expand_avl_upto,Limit),
       Limit=30, % reduced limit, as now we can inspect with evaluation view
       get_first_avl_elements(A,Limit,Els,CutOff),
       (member((X,Y),Els), X\=term(_)
        ;
        CutOff=not_all, X='...', Y='...').
get_relation_element(Value,X,Y) :-
    member((X,Y),Value). /* otherwise we have a record structure */

create_value('...',_,R) :- !, R = b(string('...'),string,[]).
create_value(X,T,R) :-
    create_texpr(value(X),T,[],R).


:- use_module(bmachine,[b_show_machine_representation/3]).
:- use_module(probcspsrc(haskell_csp_analyzer),[get_internal_csp_representation/1]).
get_internal_representation(X) :- get_internal_representation(X,false).
get_internal_representation(X,UnsetMinorMode) :- animation_mode(AM), get_internal_aux(AM,X,UnsetMinorMode).
get_internal_aux(b,X,UnsetMinorMode) :- !, b_show_machine_representation(X,true,UnsetMinorMode). % UnsetMinorMode=true means we try and generate classical B syntax
get_internal_aux(cspm,X,_) :- !, get_internal_csp_representation(X).
get_internal_aux(csp_and_b,Res,UnsetMinorMode) :- !, append("CSP||B - B Part only",X,Res),
    b_show_machine_representation(X,true,UnsetMinorMode).
get_internal_aux(_,X,_) :- X="Internal representation only available in B or CSP mode".


:- use_module(probcspsrc(haskell_csp),[channel/2]).
:- use_module(bmachine,[b_top_level_operation/1, b_top_level_feasible_operation/1]).
get_possible_event(OpName) :- b_or_z_mode,
   %b_is_operation_name(OpName).
   b_top_level_operation(OpName).
get_possible_event(Channel) :- csp_mode,
   channel(Channel,_).
% what about csp_and_b mode ?

get_feasible_event(OpName) :- b_or_z_mode,
   %b_is_operation_name(OpName).
   b_top_level_feasible_operation(OpName).
get_feasible_event(Channel) :- csp_mode,
   channel(Channel,_).


get_possible_language_specific_top_level_event(OpName,ResultNames,ParameterNames) :-
    b_or_z_mode,!,
    get_possible_b_top_level_event(OpName,ResultNames,ParameterNames).
get_possible_language_specific_top_level_event(Channel,unknown,unknown) :- csp_mode,
    channel(Channel,_).

get_possible_b_top_level_event('$setup_constants',[],Ids) :-
    b_machine_has_constants_or_properties,
    b_get_machine_constants(TIds), maplist(get_texpr_id,TIds,Ids).
get_possible_b_top_level_event('$initialise_machine',[],Ids) :-
    b_get_machine_variables(TIds), maplist(get_texpr_id,TIds,Ids).
get_possible_b_top_level_event(OpName,ResultNames,ParameterNames) :-
    b_top_level_operation(OpName),
    b_get_machine_operation(OpName,Results,RealParameters,_RealBody,_OType,_OpPos),
    maplist(get_texpr_id,Results,ResultNames),
    maplist(get_texpr_id,RealParameters,ParameterNames).

% obtain a textual description of specification category names
get_specification_description(Category,Name) :-
  animation_mode(Mode),
  (animation_minor_mode(Minor) -> true ; Minor = none), % there could be several minor modes !
  (get_specific_description(Mode,Minor,Category,N) -> Name=N
   ; get_default_description(Category,N) -> Name = N
   ; add_error(get_specification_description,'Unknown category: ',Category),
     Name=Category).

get_specific_description(cspm,_,C,N) :- get_csp_description(C,N).
get_specific_description(b,eventb,C,N) :- get_eventb_description(C,N).
get_specific_description(b,tla,C,N) :- get_tla_description(C,N).
get_specific_description(b,alloy,C,N) :- get_alloy_description(C,N).

get_csp_description(operations,'CHANNELS').
get_csp_description(operation,'CHANNEL').
get_csp_description(assertions,'CSP_ASSERTIONS').
get_csp_description(machine,'CSP_SPECIFICATION').
get_csp_description(operations_lc,'events').

get_tla_description(properties,'ASSUME').
get_tla_description(machine,'MODULE').
get_tla_description(operation,'ACTION'). % no key word
get_tla_description(operations,'ACTIONS'). % no key word
get_tla_description(operations_lc,'actions').

get_eventb_description(properties,'AXIOMS').
get_eventb_description(assertions,'THEOREMS').
get_eventb_description(operations,'EVENTS').
get_eventb_description(operation,'EVENT').
get_eventb_description(operations_lc,'events').
get_eventb_description(machine,'MODEL').

get_alloy_description(machine,'MODULE').
get_alloy_description(sets,'SIGNATURES'). % abstract
get_alloy_description(operations,'RUNS').
get_alloy_description(operation,'RUN').

get_default_description(constraints,'CONSTRAINTS').
get_default_description(properties,'PROPERTIES').
get_default_description(assertions,'ASSERTIONS').
get_default_description(invariant,'INVARIANT').
get_default_description(invariants,'INVARIANTS').
get_default_description(operations,'OPERATIONS').
get_default_description(operation,'OPERATION').
get_default_description(operations_lc,'operations').
get_default_description(machine,'MACHINE').
get_default_description(constants,'CONSTANTS').
get_default_description(variables,'VARIABLES').
get_default_description(initialisation,'INITIALISATION').
get_default_description(sets,'SETS').
get_default_description(goal,'GOAL').
get_default_description(definition,'DEFINITION').
get_default_description(definitions,'DEFINITIONS').
get_default_description(variants,'VARIANTS').
