% (c) 2009-2020 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(state_space_exploration_modes,[
   depth_breadth_first_mode/1,
   set_depth_breadth_first_mode/1, reset_depth_breadth_first_mode/0,
   dbf_modes/3,
   is_not_interesting/1, % classifying certain nodes as not interesting
   get_open_node_to_check/2,
   %add_new_visited_expression/6,
   get_id_of_node_and_add_if_required/4,
   compute_hash/3, % maybe we should move this to state_space ?
   compute_heuristic_function_for_state_id/2,
   analyze_hash_collisions/0
   ]).

:- use_module(state_space).
:- use_module(error_manager).
:- use_module(tools).
:- use_module(debug).

:- use_module(module_information,[module_info/2]).
:- module_info(group,infrastructure).
:- module_info(description,'This module groups various techniques for exploring the state space.').


:- dynamic last_step_breadth_first_in_mixed_mode/0.
reset_state_space_exploration_modes :- retractall(last_step_breadth_first_in_mixed_mode).
:- use_module(eventhandling,[register_event_listener/3]).
:- register_event_listener(reset_specification,reset_state_space_exploration_modes,
                           'Reset Model Checker BF/DF Info.').


:- load_files(library(system), [when(compile_time), imports([environ/2])]).

:- dynamic depth_breadth_first_mode/1.
depth_breadth_first_mode(mixed).

dbf_modes(0,mixed,'Mixed DF/BF').
dbf_modes(1,breadth_first,'Breadth First').
dbf_modes(2,depth_first,'Depth First').
:- if(\+ environ(prob_myheap,false)).
dbf_modes(3,heuristic,'Heuristic Function / Random'). % Random if no Heuristic Function there
dbf_modes(4,hash,'Hash-Random').
dbf_modes(5,random,'Random'). % not sure this is necessary; Hash-Random seems better ?!
dbf_modes(6,out_degree_hash,'Out-Degree for Deadlock Checking').
%dbf_modes(7,disabled_actions,'Disabled Transitions in State').
:- endif.

% provide a few short-cuts for convenience
dbf_mode_alias(df,depth_first).
dbf_mode_alias('depth-first',depth_first).
dbf_mode_alias(bf,breadth_first).
dbf_mode_alias('breadth-first',breadth_first).
dbf_mode_alias(out_degree,out_degree_hash).
dbf_mode_alias(dlk,out_degree_hash).
%dbf_mode_alias(dis,disabled_actions).

reset_depth_breadth_first_mode :- set_depth_breadth_first_mode(mixed).

set_depth_breadth_first_mode(InMODE) :-
  (dbf_mode_alias(InMODE,MODE) -> true ; MODE = InMODE),
  retractall(depth_breadth_first_mode(_)),
  debug_println(9,depth_breadth_first_mode(MODE)),
  assert(depth_breadth_first_mode(MODE)),
  (dbf_modes(_,MODE,_) -> true ; add_internal_error('Unknown mode: ',set_depth_breadth_first_mode(MODE))).


/* ---------------------------------- */
/*       get_open_node_to_check/2     */
/* ---------------------------------- */

get_open_node_to_check(depth_first,NodeID) :- pop_id_from_front(NodeID).
get_open_node_to_check(hash,NodeID) :- pop_id_from_front(NodeID).  %,print(pop(NodeID)),nl.
get_open_node_to_check(heuristic,NodeID) :- pop_id_from_front(NodeID).
get_open_node_to_check(random,NodeID) :- pop_id_from_front(NodeID).
get_open_node_to_check(out_degree_hash,NodeID) :- pop_id_from_front(NodeID).
get_open_node_to_check(breadth_first,NodeID) :- pop_id_from_end(NodeID).
get_open_node_to_check(mixed,NodeID) :-  %%depth_breadth_first_mode(DFMODE),
   (random(1,3,1)
                  /* random(1,4,1) ==> 33 percent chance of going DF */
                  /* random(1,3,1) ==> 50 percent chance of going DF */
                  /* (random(1,3,1);random(1,3,1)) ==> 75 % chance of going DF */
    -> retractall(last_step_breadth_first_in_mixed_mode),
       % We are going depth_first
       pop_id_from_front(NodeID) %, print_message(a(NodeID)).
    ;  (last_step_breadth_first_in_mixed_mode -> true ; assert(last_step_breadth_first_in_mixed_mode)),
       pop_id_oldest(NodeID)
    ).



/* ---------------------------------- */
/*            add_open_node/3         */
/* ---------------------------------- */


add_open_node(NodeID,Hash,FromID) :- %print_message(aa(NodeID)),
   (is_not_interesting(NodeID) ->
      asserta(not_interesting(NodeID))
    ; add_open_node2(NodeID,Hash,FromID)
    ).
:- use_module(library(terms),[term_size/2]).
:- use_module(library(random)).
add_open_node2(NodeID,Hash,FromID) :- %print_message(aa(NodeID)),
   %(b_or_z_mode -> assert(not_invariant_checked(NodeID)) ; true), % is now added when we pop an open node
   (last_step_breadth_first_in_mixed_mode , \+ random(1,8,1)
          -> add_id_at_end(NodeID) /* do not add successor of BF step at top */
                                   /* avoids degeneration into BF when large branching factor */
          ;   depth_breadth_first_mode(Mode) -> add_open_node_in_mode(Mode,NodeID,Hash,FromID)
           % add_id_at_front(NodeID) % for default DF/BF behaviour

             % add_id_with_weight(NodeID,_Hash) % use hash value

             % add_id_random(NodeID)   % use a random number

             % visited_expression(NodeID,T), term_size(T,Sz),
             % add_id_with_weight(NodeID,Sz)  % use term_size as heuristic

             % add_id_to_process(NodeID,_Hash,_FromID) % for out_degree as heuristic

            %add_id_with_heuristic(NodeID)  % for using custom heuristic if possible
   ).

add_open_node_in_mode(breadth_first,NodeID,_Hash,_) :- add_id_at_front(NodeID).
add_open_node_in_mode(depth_first,NodeID,_Hash,_) :- add_id_at_front(NodeID).
add_open_node_in_mode(hash,NodeID,Hash,_) :- %print(add(NodeID,Hash)),nl,
    add_id_with_weight(NodeID,Hash). % use Hash as pseudo-random Weight
add_open_node_in_mode(heuristic,NodeID,_Hash,_) :- add_id_with_heuristic(NodeID).
add_open_node_in_mode(mixed,NodeID,_Hash,_) :- add_id_at_front(NodeID).
add_open_node_in_mode(random,NodeID,_Hash,_) :- add_id_random(NodeID).
add_open_node_in_mode(out_degree_hash,NodeID,Hash,FromID) :- add_id_to_process(NodeID,Hash,FromID).
% commented out to avoid dependence on pge_algo and because currently not used:
%add_open_node_in_mode(disabled_actions,NodeID,Hash,_) :-
%  (pge_algo:disabled_operations(NodeID,DisOps,_EnabledOps) ->
%    length(DisOps,Weight), add_id_with_weight(NodeID,Weight)
%  ; otherwise -> add_id_with_weight(NodeID,Hash)).

:- use_module(bmachine,[b_get_machine_heuristic_function/1]).
:- use_module(specfile,[state_corresponds_to_initialised_b_machine/2, b_or_z_mode/1, animation_mode/1]).
:- use_module(xtl_interface,[xtl_heuristic_function_result/2, xtl_heuristic_function_active/0]).

add_id_with_heuristic(NodeID) :- %print(adding(NodeID)),nl,
   compute_heuristic_function_for_state_id(NodeID,HRes),
   !,
   (HRes = int(H)
        -> add_id_with_weight(NodeID,H) % , print(added_id_with_weight(NodeID,H)),nl
        ;  print_message('*** Illegal HEURISTIC_FUNCTION result (must be INTEGER): '), print_message(HRes),
           add_id_random(NodeID)
    ).
add_id_with_heuristic(NodeID) :- add_id_random(NodeID).

compute_heuristic_function_for_state_id(NodeID,Value) :- animation_mode(Mode),
   compute_heuristic_function_for_state_id3(Mode,NodeID,Value).
compute_heuristic_function_for_state_id3(Mode,NodeID,Value) :-
   b_or_z_mode(Mode), !,
   b_get_machine_heuristic_function(HFunction),
   visited_expression(NodeID,State),
   % print(cur_expr(State)),nl,
   state_corresponds_to_initialised_b_machine(State,BState),
   (b_interpreter:b_compute_expression_nowf(HFunction,[],BState,HRes)
     -> Value=HRes
     ; print_message(heuristic_function_failed(NodeID,BState)),fail
    ).
compute_heuristic_function_for_state_id3(xtl,NodeID,Value) :- !,
   xtl_heuristic_function_active,
   visited_expression(NodeID,State),
   xtl_heuristic_function_result(State,Value).

% -------------------

:- use_module(bmachine,[b_get_machine_searchscope/1]).
:- use_module(bsyntaxtree,[create_negation/2]).
:- use_module(eclipse_interface,[test_boolean_expression_in_node/2]).
is_not_interesting(NodeID) :-
   b_get_machine_searchscope(Scope),
   %format('Testing SCOPE in ~w~n',[NodeID]),
   preferences:preference(use_scope_predicate,true),
   create_negation(Scope,NotScope),
   test_boolean_expression_in_node(NodeID,NotScope).


% -----------------------




/* check whether we have visited a node already: if so return
 id of that node; otherwise generate new id and add node */
:- use_module(state_packing,[pack_state/2]).
get_id_of_node_and_add_if_required(State,ID,Res,FromID) :-
  preferences:preference(symmetry_mode,SymMode),
 % print(computing_hash),nl,
  pack_state(State,PackedState), % TO DO: based on FromID and TransitionID copy unchanged variables if COMPRESSION TRUE
  compute_hash(SymMode,State,PackedState,Hash,Marker), % TO DO: we could use PackedState
 % print_message(hash(Hash)),
  ( packed_expression_already_visited(SymMode,PackedState,Hash,Marker,ID)
   -> Res = exists
   ;  Res = new,
      gen_new_state_id(ID), /* gensym('node',ID),*/

      % add_new_visited_expression:
      state_space_packed_add(ID,PackedState),
      add_open_node(ID,Hash,FromID),
      (SymMode=flood -> add_permutations(ID,State) ; true),

      assert(hash_to_id(Hash,ID)),
      (SymMode=hash -> assert_id_to_marker(ID,Marker) ; true)
   ).

packed_expression_already_visited(SymMode,_,Hash,_Marker,ID) :-
   (preferences:preference(ignore_hash_collisions,true) ; SymMode=nauty),
   !, hash_to_id(Hash,ID).
packed_expression_already_visited(hash,_,Hash,Marker,ID) :- Marker \= none,!,
    hash_to_id(Hash,ID),id_to_marker(ID,Marker).
packed_expression_already_visited(_SymMode,PackedState,Hash,_Marker,ID) :-
    find_hashed_packed_visited_expression(Hash,PackedState,ID).

assert_id_to_marker(ID,Marker) :-
    (Marker = none;preferences:preference(ignore_hash_collisions,true)
        -> true  % don't assert marker info if we ignore hash collisions anyway
        ; assert(id_to_marker(ID,Marker))
    ).


% Utilities for Permutation Flooding


:- use_module(state_permuter).
add_permutations(ID,Expression) :- %print(add_permutations(ID,Expression)),nl,
  permute_state(Expression,Perm), Perm \= Expression,
  generate_new_flood_id(Perm,PermID),
  state_space_add(PermID,Perm),
  (preferences:preference(forget_state_space,true)
          -> true ; store_transition(PermID,'*permute*',ID,_)),
  fail.
add_permutations(_,_).

generate_new_flood_id(State,ID) :-
  pack_state(State,PackedState),
  compute_hash(flood,State,PackedState,Hash,Marker),
  \+ packed_expression_already_visited(flood,PackedState,Hash,Marker,ID),
  gen_new_state_id(ID),
  assert(hash_to_id(Hash,ID)).



:- use_module(symmetry_marker,[compute_marker_for_state/2]).
:- use_module(hashing).


/* compute_hash(State, HashValue, Marker/SecondHashTerm) */
compute_hash(State,Hash,Marker) :- preferences:preference(symmetry_mode,SymMode),
  pack_state(State,PackedState),
  compute_hash(SymMode,State,PackedState,Hash,Marker).

%:- use_module(extension('probhash/probhash')).

compute_hash(off, _,PackedState,Hash,Marker) :- !,  Marker=none,
   my_term_hash(PackedState,Hash). %, tools_printing:print_term_summary(Env),nl.
compute_hash(hash,State,_PackedState,Hash,Marker) :-
  (compute_marker_for_state(State,HMarker)
    -> HMarker=Marker
      %raw_sha_hash(HMarker,Marker)
     ; print(compute_marker_failed(State)),nl,fail),
  my_term_hash(Marker,Hash),!.
compute_hash(nauty,State,_PackedState,Hash,Marker) :- !, Marker=none,
  my_term_hash(State,EnvHash),
  (hash_to_nauty_id(EnvHash,NautyID), hash_to_id(NautyID,RealID),
   visited_expression(RealID,State)
    -> %% print(exact_state_exists(EnvHash,NautyID,State)),nl, %%
      % the exact same state already exists; no need to compute normal form
       Hash=NautyID
   ;   (graph_canon:add_state_graph_to_nauty(State,Exists)
        -> true
        ; add_error(compute_hash,'Failed: ',add_state_graph_to_nauty(State,Exists), Exists=1)
        ),
        (Exists>0 -> Hash=Exists
             ; /* new node */ Hash is -Exists, assert(hash_to_nauty_id(EnvHash,Hash)))
   ),
  % print(nauty_add(Exists,Hash,State)),nl,
  Marker=none.
% my_term_hash now uses if_var(ignore) option of SICStus 4.1; no longer need to check ground
compute_hash(_, _Env,PackedState,Hash,none) :- my_term_hash(PackedState,Hash).


% --------------------------------------------

% some utilities to analyse state space:

hash_collision(ID,ID2,S1,S2) :-
  hash_to_id(Hash,ID),
  hash_to_id(Hash,ID2), ID2 \= root, (ID=root->true; ID2>ID),
  visited_expression(ID,S1),
  %print((ID,S1)),nl,
  visited_expression(ID2,S2),
  %print((ID2,S2)),nl,
  nl.

:- dynamic hash_collision/2. % hash_collision(NrStatesWithSameHash, NrOfSuchClusters)
analyze_hash_collisions :- retractall(hash_collision(_,_)),
    print('Analyzing Hash Collisions'),nl,
    hash_to_id(Hash,ID),
    \+ (hash_to_id(Hash,ID2), ID2 @< ID),
    findall(X,hash_to_id(Hash,X),L),
    length(L,Nr),
    (retract(hash_collision(Nr,C))
      -> (C1 is C+1, assert(hash_collision(Nr,C1)))
      ;  assert(hash_collision(Nr,1))
    ),fail.
analyze_hash_collisions :-
   print('Hash Collision Distribution (1=no collision): '),
   findall((Nr,X),hash_collision(Nr,X),Coll),
   sort(Coll,SC),
   print(SC),nl,
   (hash_collision(ID,ID2,S1,S2)  ->
       print('SAMPLE COLLISION: '),nl,
       print(ID),nl,print(S1),nl, print_hash_info(S1),
       print(ID2),nl,print(S2),nl,print_hash_info(S2)
     ; print('NO COLLISIONS')
   ),nl.


%ac :- analyse_collision([range(smallint),depth(infinite)]). % Runtime: 3890 ms
%ac0 :- analyse_collision([range(smallint),algorithm(sdbm),depth(infinite)]). % Runtime: 3760 ms
%ac1 :- analyse_collision([range(infinite),algorithm(sdbm),depth(infinite)]). % Runtime: 3730 ms
%ac2 :- analyse_collision([range(infinite),algorithm(hsieh),depth(infinite)]). % Runtime: 3970 ms
%ac3 :- analyse_collision([range(infinite),algorithm(jenkins),depth(infinite)]). % Runtime: 3970 ms
%ac4 :- analyse_collision([]). % Runtime: 3720 ms
%analyse_collision(Options) :- retractall(hash_collision(_,_)),
%  statistics(runtime,[_,_]),
%  prolog_flag(max_tagged_integer,Max),
%  prolog_flag(min_tagged_integer,Min),
%  visited_expression(ID,B),
%  term_hash(B,Options,HashB), %
%  %super_hash(B,HashB), %
%  ((HashB>=Min, HashB =< Max) -> true ; print('*** HASH NOT GOOD FOR INDEXING: '), print(ID),nl),
%  (hash_collision(HashB,OtherID)
%    -> print('COLLISION : '), print(ID), print(' <-> '), print(OtherID), print(' : '), print(HashB),nl
%    ;  assert(hash_collision(HashB,ID))
%  ),fail.
%analyse_collision(_Options) :-
%  statistics(runtime,[_,T]), print('Runtime: '), print(T), print(' ms').


:- use_module(library(terms),[term_hash/2]).
print_hash_info(S1) :-
       %env_hash(S1,EH1), print(env_hash(EH1)),nl,
       my_term_hash(S1,MTH1), print(my_term_hash(MTH1)),nl,
       term_hash(S1,TH1), print(sicstus_term_hash(TH1)),nl.

:- use_module(library(lists),[maplist/2]).
:- public print_current_state_term_sizes/0.
print_current_state_term_sizes :-
   current_expression(ID,State), %expanded
   state_corresponds_to_initialised_b_machine(State,BState),
   format('Term Sizes of variables and constants in current state: ~w~n',[ID]),
   findall( TS/Var, (member(bind(Var,Term),BState),
                     term_size(Term,TS)), List),
    sort(List,SL),
    maplist(print_ts,SL).

print_ts(TS/Var) :- format('~w : ~w~n',[Var,TS]).

% --------------------------------------------
