% (c) 2017 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(constraints,[get_behavioral_constraint_from_input/6,
                       get_behavioral_constraint/6,
                       get_behavioral_constraint_aux/7,
                       get_new_behavioral_constraint/5,
                       get_prevent_component_swap_constraint/4,
                       get_line_amount/4]).

:- use_module(synthesis(synthesis_util)).
:- use_module(synthesis(symmetry_reduction)).

:- use_module(probsrc(bsyntaxtree)).
:- use_module(probsrc(bmachine_structure),[get_section/3]).
:- use_module(probsrc(tools),[flatten/2]).
:- use_module(probsrc(module_information),[module_info/2]).
:- use_module(probsrc(bmachine),[full_b_machine/1,
                                 b_get_machine_constants/1,
                                 b_get_properties_from_machine/1,
                                 b_get_machine_operation/4,
                                 b_get_machine_variables/1,
                                 b_get_machine_set/1]).

:- use_module(library(lists)).

:- use_module(extension('plspec/plspec/plspec_core')).

:- module_info(group,synthesis).
:- module_info(description,'This module defines the constraint for synthesizing invariants, preconditions or complete operations using simple I/O examples.').

unfold_existential_quantification(true).

split_input_output_location_vars(LocationVars,LocationVarsIn,LocationVarsOut) :-
    split_input_output_location_vars(LocationVars,[],[],LocationVarsIn,LocationVarsOut).

split_input_output_location_vars([],InAcc,OutAcc,InAcc,OutAcc).
split_input_output_location_vars([LocationVar|T],InAcc,OutAcc,LocationVarsIn,LocationVarsOut) :-
    accumulate_location_var(LocationVar,InAcc,OutAcc,NewInAcc,NewOutAcc) ,
    split_input_output_location_vars(T,NewInAcc,NewOutAcc,LocationVarsIn,LocationVarsOut).

accumulate_location_var(b(identifier(LocationVarName),integer,Info),InAcc,OutAcc,[b(identifier(LocationVarName),integer,Info)|InAcc],OutAcc) :-
    atom_concat(li,_,LocationVarName).
accumulate_location_var(b(identifier(LocationVarName),integer,Info),InAcc,OutAcc,InAcc,[b(identifier(LocationVarName),integer,Info)|OutAcc]) :-
    atom_concat(lo,_,LocationVarName).
accumulate_location_var(_,InAcc,OutAcc,InAcc,OutAcc).

% Exclude a solution and add this node to the behavioral constraint. We exclude only the solution given
% by the program ast without involving generated deadcode.
get_new_behavioral_constraint(ProgramAST,LocationVars,solution(SolutionBindings),BehavioralConstraint,NewBehavioralConstraint) :-
    % get the used operators output location variables to exclude from the further solution
    get_used_location_vars_and_op_names(ProgramAST,LocationVars,[],[],_,UsedOperatorNames) ,
    exclude_solution_for_further_search(UsedOperatorNames,LocationVars,SolutionBindings,Exclusion) ,
    split_input_output_location_vars(LocationVars,LocationVarsIn,LocationVarsOut) ,
    used_library(Components) ,
    %findall(Component,(member(Component,AllComponents) , Component = (_,Name) , member(Name,UsedOperatorNames)),UsedComponents) ,
    get_post_symmetry_reduction_constraint(Components,SolutionBindings,LocationVarsIn,LocationVarsOut,SymmetryReduction) ,
    conjunct_predicates([BehavioralConstraint,Exclusion,SymmetryReduction],NewBehavioralConstraint).

get_behavioral_constraint_aux((ValidGuards,Operation),action,Library,PosInput,PosOutput,BehavioralConstraint,LocationVars) :-
    get_behavioral_constraint(Library,(ValidGuards,Operation),PosInput,PosOutput,BehavioralConstraint,LocationVars).
get_behavioral_constraint_aux((ValidInvariantsOrGuards,Operation),_,Library,PosInput,PosOutput,BehavioralConstraint,LocationVars) :-
    get_behavioral_constraint_from_input(Library,(ValidInvariantsOrGuards,Operation),PosInput,PosOutput,BehavioralConstraint,LocationVars).

% create outputs and get behavioral constraint
get_behavioral_constraint_from_input(Library,(ValidGuards,Operation),PosInput,NegInput,BehavioralConstraint,LocationVars) :-
    % get output values
    length(PosInput,PosLength) , length(NegInput,NegLength) ,
    % create empty lists
    length(TempPosOutput,PosLength) , length(TempNegOutput,NegLength) ,
    % for our task to repair invariant violations output is either "true" or "false" (interpreted as pred_true/pred_false (synthesis(type,pred))
    % but implemented as boolean_true/boolean_false, we set options each node's info representing the type which can differ from the ast node's type)
    findall(In,(member(In,TempPosOutput) , In = [b(boolean_true,boolean,[synthesis(type,pred)])]),PosOutput) ,
    findall(In,(member(In,TempNegOutput) , In = [b(boolean_false,boolean,[synthesis(type,pred)])]),NegOutput) ,
    append(PosInput,NegInput,Input) ,
    append(PosOutput,NegOutput,Output) ,
    get_behavioral_constraint(Library,(ValidGuards,Operation),Input,Output,BehavioralConstraint,LocationVars).

:- dynamic used_library/1.
:- volatile used_library/1.

% Get the functional constraint for each Input-Output Example
% It is sufficient to create one well-definedness constraint surrounding the behavioral constraint
% but the library constraint needs to be defined for every examples functional constraint.
get_behavioral_constraint(Library,(ValidGuards,Operation),InputExamples,OutputExamples,BehavioralConstraint,LocationVars) :-
    % lists of currently used components, necessary if a library is used several times
    retractall(used_library(_)) ,
    asserta(used_library([])) ,
    InputExamples = [InputValues|_] ,
    % create library constraint separated from the funtional constraint
    % to make sure to just obtain one set of location variables
    get_library_constraint(InputValues,Library,(ValidGuards,Operation),LibConstraint,InputVars,OutputVars,EnumeratedConstants) ,
    % set the temporary location variables with prefix "l"
    maplist(get_location_var_from_ast_node,InputVars,LocationVarsIn) ,
    maplist(get_location_var_from_ast_node,OutputVars,LocationVarsOut) ,
    % tuples of raw component name and internal component name like (add,add1)
    used_library(Components) ,
    get_preliminary_symmetry_reduction_constraint(Components,LocationVarsIn,LocationVarsOut,SymmetryReductionPred) ,
    conjunct_predicates([LibConstraint,SymmetryReductionPred],ExtendedLibConstraint) ,

    % get some information needed for well-definedness constraint
    get_line_amount(Operation,Library,InputValues,M) ,
    length(InputValues,VarAmount) ,
    % create well-definedness constraint separated from function constraints
    get_well_definedness_constraint(VarAmount,LocationVarsIn,LocationVarsOut,M,WellDefinednessConstraint) ,
    get_max_cardinality_from_examples(InputExamples,MaxCard) ,
    get_range_of_integer_nodes_from_examples(MaxCard,InputExamples,InMin,InMax) ,
    % create upper and lower bound for input values (kodkod warning) and constants
    create_texpr(integer(InMin),integer,[],IntLowerBound) ,
    create_texpr(integer(InMax),integer,[],IntUpperBound) ,
    create_texpr(integer(MaxCard),integer,[],CardUpperBound) , ! ,
    map_get_function_constraint(Operation,0,Library,ExtendedLibConstraint,InputVars,OutputVars,InputExamples,OutputExamples,FunctionConstraints,ProgramInLocationVars) ,
    enumerated_sets_constraint(EnumeratedSetsConstraint) ,
    append(OutputVars,EnumeratedConstants,ConstantsAndOutputs) ,
    % restrict domains outside of the existential quantifiers
    restrict_constant_domains(IntLowerBound,IntUpperBound,CardUpperBound,ConstantsAndOutputs,ConstantRestrictionList) ,
    % and create typing predicates for constants and input output values of components
    create_typing_predicates_from_info(ConstantsAndOutputs,TypingPredicatesConj) ,
    conjunct_predicates(ConstantRestrictionList,ConstantRestrictions) ,
    findall(InputRestriction,(member(In,InputVars) , get_bounds_for_node(In,IntLowerBound,IntUpperBound,CardUpperBound,InputRestriction)),TempInputRestrictions) ,
    conjunct_predicates(TempInputRestrictions,InputRestrictions) ,
    conjunct_predicates([EnumeratedSetsConstraint,WellDefinednessConstraint,TypingPredicatesConj,ConstantRestrictions,InputRestrictions|FunctionConstraints],BehavioralConstraint) ,
    append([LocationVarsIn,LocationVarsOut,ProgramInLocationVars],LocationVars).

enumerated_sets_constraint(EnumeratedSetsConstraint) :-
    findall(SetName,b_get_machine_set(SetName),EnumeratedSets) ,
    full_b_machine(BMachine) ,
    get_section(enumerated_elements,BMachine,Elems),
    enumerated_sets_constraint_aux(Elems,EnumeratedSets,TEnumeratedSetsConstraint) ,
    conjunct_predicates(TEnumeratedSetsConstraint,EnumeratedSetsConstraint).

enumerated_sets_constraint_aux(_,[],[]).
enumerated_sets_constraint_aux(Elems,[SetName|T],[b(conjunct(Equality,DistinctConstraint),pred,[])|NT]) :-
    findall(b(identifier(N),global(SetName),[]),member(b(identifier(N),global(SetName),_),Elems),EElems) ,
    EElems \== [] , ! ,
    Equality = b(equal(b(identifier(SetName),set(global(SetName)),[]),b(set_extension(EElems),set(global(SetName)),[])),pred,[]) ,
    get_consistency_constraint(EElems,DistinctConstraint) ,
    enumerated_sets_constraint_aux(Elems,T,NT).
enumerated_sets_constraint_aux(Elems,[_|T],NT) :-
    enumerated_sets_constraint_aux(Elems,T,NT).

map_get_function_constraint(_,_,_,_,_,_,[],[],[],[]).
map_get_function_constraint(Operation,ExampleCount,Library,LibConstraint,InputVars,OutputVars,[Input|TIn],[Output|TOut],[FunctionConstraint|NTFunction],LocationVars) :-
    get_function_constraint(Operation,ExampleCount,Library,LibConstraint,InputVars,OutputVars,Input,Output,FunctionConstraint,NewLocationVars) ,
    NewExampleCount is ExampleCount + 1 ,
    map_get_function_constraint(Operation,NewExampleCount,Library,LibConstraint,InputVars,OutputVars,TIn,TOut,NTFunction,NTLVars) ,
    append(NewLocationVars,NTLVars,LocationVars).

get_function_constraint(Operation,ExampleCount,Library,LibConstraint,InputVars,OutputVars,InputValues,OutputValues,FunctionConstraint,LocationVars) :-
    % add a predicate setting equalities to the machine variables for the current example, since we possibly consider additional set constants like c = {i}, where i is a machine variable
    synthesis_util:create_equality_nodes_from_example(InputValues,InputEqualityList) ,
    % in this context, we have to add the invariants setting the types of the machine variables
    conjunct_predicates(InputEqualityList,InputEquality) ,
    b_get_machine_variables(MachineVars) ,
    findall(MachineVar,(member(InputValue,InputValues) , get_texpr_info(InputValue,Info) , member(synthesis(machinevar,CurrentVarName),Info) ,
                        member(MachineVar,MachineVars) , get_texpr_id(MachineVar,CurrentVarName)),CurrentVars) ,
    b_get_typing_predicate_for_vars(CurrentVars,TypingMachineVarsPredicate) ,
    get_line_amount(Operation,Library,InputValues,M) ,
    % enumerate input/output values
    enumerate_inputs_and_outputs(ExampleCount,InputValues,input,InputValueNodes) ,
    enumerate_inputs_and_outputs(ExampleCount,OutputValues,output,OutputValueNodes) ,
    get_connectivity_constraint(InputValueNodes,OutputValueNodes,InputVars,OutputVars,ConnectivityConstraint) ,
    set_input_location_vars(InputValueNodes,InputLocationConstraint) ,
    set_output_location_vars(OutputValueNodes,M,OutputLocationConstraint,OutputLocationVars) ,
    conjunct_predicates([TypingMachineVarsPredicate,InputEquality,OutputLocationConstraint,InputLocationConstraint,LibConstraint,ConnectivityConstraint],ExistsBody) ,
    % get all used identifier as ast nodes
    find_typed_identifier_uses_with_info(ExistsBody,TListOfIDNodes) ,
    remove_dup_id_nodes(TListOfIDNodes,ListOfIDNodes) ,
    get_temporary_ids(ListOfIDNodes,TempTemporaryIDs) ,
    append(TempTemporaryIDs,CurrentVars,TemporaryIDs) , % the machine variables are local to the exists quantifier which are used within the equalities to the machine variables
    findall(LocationVar,(member(LocationVar,ListOfIDNodes) , LocationVar = b(identifier(Name),integer,_) , atom_concat(l,_,Name)),TempListOfLVarNodes) ,
    append(TempListOfLVarNodes,OutputLocationVars,LocationVars) ,
    % and their corresponding identifier names
    maplist(get_texpr_id,ListOfIDNodes,UsedIDs) ,
    (   unfold_existential_quantification(true)
    ->  set_unique_temporary_id_names(ExampleCount,TemporaryIDs,ExistsBody,FunctionConstraint)
    ;   FunctionConstraint = b(exists(TemporaryIDs,ExistsBody),pred,[used_ids(UsedIDs)])
    ).

set_unique_temporary_id_names(ExampleCount,TemporaryIDs,b(identifier(Name),Type,Info),b(identifier(UniqueName),Type,Info)) :-
    member(b(identifier(Name),Type,_),TemporaryIDs) , ! ,
    atom_codes(Name,NameCodes) ,
    number_codes(ExampleCount,SuffixNumber) ,
    append(NameCodes,[95|SuffixNumber],UniqueNameCodes) ,
    atom_codes(UniqueName,UniqueNameCodes).
set_unique_temporary_id_names(_,_,b(identifier(Name),Type,Info),b(identifier(Name),Type,Info)).
set_unique_temporary_id_names(ExampleCount,TemporaryIDs,b(Node,Type,Info),b(NewNode,Type,Info)) :-
    Node =.. [Functor,Arg1,Arg2] , ! ,
    set_unique_temporary_id_names(ExampleCount,TemporaryIDs,Arg1,NewArg1) ,
    set_unique_temporary_id_names(ExampleCount,TemporaryIDs,Arg2,NewArg2) ,
    NewNode =.. [Functor,NewArg1,NewArg2].
set_unique_temporary_id_names(ExampleCount,TemporaryIDs,b(Node,Type,Info),b(NewNode,Type,Info)) :-
    Node =.. [Functor,Arg] , ! ,
    set_unique_temporary_id_names(ExampleCount,TemporaryIDs,Arg,NewArg) ,
    NewNode =.. [Functor,NewArg].
set_unique_temporary_id_names(_,_,Ast,Ast).

create_typing_predicates_from_info([],b(truth,pred,[])).
create_typing_predicates_from_info([Node|T],TypingPredicatesConj) :-
    create_typing_predicate_from_info(Node,TypingPredicate) ,
    create_typing_predicates_from_info(T,TypingPredicate,TypingPredicatesConj).

create_typing_predicates_from_info([],Acc,Acc).
create_typing_predicates_from_info([Node|T],Acc,TypingPredicatesConj) :-
    create_typing_predicate_from_info(Node,TypingPredicate) ,
    create_texpr(conjunct(Acc,TypingPredicate),pred,[],NewAcc) ,
    create_typing_predicates_from_info(T,NewAcc,TypingPredicatesConj).

create_typing_predicate_from_info(Node,b(member(Node,TypeAst),pred,[])) :-
    get_texpr_info(Node,Info) ,
    member(synthesis(type,Type),Info) ,
    Type \= pred , ! ,
    get_type_ast(Type,TypeAst).
create_typing_predicate_from_info(_,b(truth,pred,[])).

get_type_ast(global(IndependentType),b(identifier(IndependentType),set(global(IndependentType)),[])).
get_type_ast(set(global(IndependentType)),b(pow_subset(b(identifier(IndependentType),set(global(IndependentType)),[])),set(set(global(IndependentType))),[])) :- !.
get_type_ast(Nested,b(pow_subset(TypeAst),set(InnerType),[])) :-
    Nested =.. [Functor,Type] ,
    (Functor = set ; Functor = seq) ,
    get_type_ast(Type,TypeAst),
    TypeAst = b(_,InnerType,_).
get_type_ast(couple(A,B),b(cartesian_product(TA,TB),set(couple(A,B)),[])) :-
    get_type_ast(A,TA) ,
    get_type_ast(B,TB).
get_type_ast(integer,b(integer_set('INTEGER'),set(integer),[])).
get_type_ast(real,b(real_set,set(real),[])).
get_type_ast(string,b(string_set,set(string),[])).
get_type_ast(boolean,b(bool_set,set(boolean),[])).

restrict_constant_domains(_,_,_,[],[]).
restrict_constant_domains(IntLowerBound,IntUpperBound,CardUpperBound,[Out|T],[ConstantRestriction|Restrictions]) :-
    is_integer_constant(Out) ,
    ConstantRestriction = b(member(Out,b(interval(IntLowerBound,IntUpperBound),set(integer),[])),pred,[]) ,
    restrict_constant_domains(IntLowerBound,IntUpperBound,CardUpperBound,T,Restrictions).
restrict_constant_domains(IntLowerBound,IntUpperBound,CardUpperBound,[Out|T],[ConstantRestriction|Restrictions]) :-
    is_set_or_seq_constant(Out) ,
    ConstantRestriction = b(less_equal(b(card(Out),integer,[]),CardUpperBound),pred,[]) ,
    restrict_constant_domains(IntLowerBound,IntUpperBound,CardUpperBound,T,Restrictions).
% skip constants like boolean
restrict_constant_domains(IntLowerBound,IntUpperBound,CardUpperBound,[_|T],Restrictions) :-
    restrict_constant_domains(IntLowerBound,IntUpperBound,CardUpperBound,T,Restrictions).

get_max_cardinality_from_examples(InputExamples,MaxCard) :-
    flatten(InputExamples,InputSet) ,
    findall(SetOrSeqNode,(member(SetOrSeqNode,InputSet) , get_texpr_type(SetOrSeqNode,Type) , (Type = set(_) ; Type = seq(_)) ,
                          get_texpr_info(SetOrSeqNode,Info) , member(synthesis(initial_example),Info)),SetOrSeqNodes) ,
    SetOrSeqNodes \= [] , ! ,
    findall(Card,(member(Node,SetOrSeqNodes) , b_compute_expression_nowf_no_bindings(b(card(Node),integer,[]),Card)),Cards) ,
    (Cards = []
    ->  MaxCardExamples = 0
    ;   b_compute_expression_nowf_no_bindings(b(max(b(value(Cards),set(integer),[])),integer,[]),Res) , Res = int(MaxCardExamples)) ,
    % subjectively add a limit of MaxCard + 3
    MaxCard is MaxCardExamples + 3.
% zero if no sets or seq are given
get_max_cardinality_from_examples(_,0).

% Get the range of integer nodes from the initial examples to restrict constant domains.
get_range_of_integer_nodes_from_examples(_,InputExamples,InMin,InMax) :-
    flatten(InputExamples,InputSet) ,
    get_values_of_integer_nodes(InputSet,[],IntegerValues) ,
    IntegerValues \= [] , ! ,
    min_member(TInMin,IntegerValues) , max_member(TInMax,IntegerValues) ,
    InMin is TInMin - 10 , InMax is TInMax + 10.
get_range_of_integer_nodes_from_examples(MaxCard,_,0,NMaxCard) :- % at least MaxCard + 5, e.g. card(x) > 5 but no integer is given
    NMaxCard is MaxCard + 5.

% TODO: consider nested sets and seqs of integer for domain restrictions?
get_values_of_integer_nodes([],Acc,Acc).
get_values_of_integer_nodes([Input|T],Acc,IntegerValues) :-
    % type is integer, set(integer) or seq(integer)
    get_texpr_type(Input,Type) ,
    get_values_of_integer_nodes_aux(Type,Input,Integer) ,
    get_values_of_integer_nodes(T,[Integer|Acc],IntegerValues).
get_values_of_integer_nodes([_|T],Acc,IntegerValues) :-
    % other type or empty set/seq
    get_values_of_integer_nodes(T,Acc,IntegerValues).
get_values_of_integer_nodes_aux(integer,Input,Integer) :-
    b_compute_expression_nowf_no_bindings(Input,Value) , Value = int(Integer).
get_values_of_integer_nodes_aux(set(integer),Input,Integer) :-
    not_empty_set_or_seq(Input) ,   % max([]) is not well-defined
    b_compute_expression_nowf_no_bindings(b(max(Input),integer,[]),Value) , Value = int(Integer).
get_values_of_integer_nodes_aux(set(integer),Input,Integer) :-
    not_empty_set_or_seq(Input) ,   % same for empty sequences
    b_compute_expression_nowf_no_bindings(b(max(b(range(Input),set(integer),[])),integer,[]),Value) , Value = int(Integer).

not_empty_set_or_seq(Input) :-
    b_compute_expression_nowf_no_bindings(Input,Value) , Value \= [].

% TODO: same here, consider nested sets and seqs of integer?
get_bounds_for_node(Node,LowerIntBound,UpperIntBound,_,Restriction) :-
    Node = b(_,integer,_) ,
    Restriction = b(member(Node,b(interval(LowerIntBound,UpperIntBound),set(integer),[])),pred,[]).
get_bounds_for_node(Node,LowerIntBound,UpperIntBound,CardUpperBound,Restriction) :-
    Node = b(_,Type,_) ,
    Type = set(integer) ,
    % we restrict the cardinality of sets of integers as well as the domain of their elements
    Restriction = b(conjunct(b(less_equal(b(card(Node),integer,[]),CardUpperBound),pred,[]),
        b(member(Node,b(pow_subset(b(interval(LowerIntBound,UpperIntBound),set(integer),[])),set(set(integer)),[])),pred,[])),pred,[]).
get_bounds_for_node(Node,_LowerIntBound,_UpperIntBound,CardUpperBound,Restriction) :-
    Node = b(_,Type,_) ,
    Type = seq(integer) ,
    % we restrict the size of seqs of integers
    % TODO: restrict domain of elements of seq(integer)
    Restriction = b(less_equal(b(card(Node),integer,[]),CardUpperBound),pred,[]).
get_bounds_for_node(Node,_,_,CardUpperBound,Restriction) :-
    Node = b(_,Type,_) ,
    % either card or size for sets or sequences
    (Type = set(InnerType) , Functor = card ; Type = seq(InnerType) , Functor = size) ,
    InnerType \= integer ,
    RestrNode =.. [Functor,Node] ,
    % we restrict cardinality/size of sets/seqs with an other type than integer
    Restriction = b(less_equal(b(RestrNode,integer,[]),CardUpperBound),pred,[]).

is_integer_constant(b(Node,integer,Info)) :-
    is_constant(b(Node,integer,Info)).

is_set_or_seq_constant(b(Node,Type,Info)) :-
    (Type = set(_) ; Type = seq(_)) ,
    is_constant(b(Node,Type,Info)).

% all temporary library variables like i0,i1,o1 in exists node except 'constant'
get_temporary_ids([],[]).
get_temporary_ids([ID|T],[ID|NT]) :-
    get_texpr_id(ID,Name) ,
    % filter identifier nodes e.g. referring to enumerated sets or independent types
    get_texpr_info(ID,Info) ,
    member(synthesis(_,_),Info) ,
    atom_codes(Name,[Char|_]) ,
    Char \= 108 ,
    \+atom_concat(constant,_,Name) , ! ,
    get_temporary_ids(T,NT).
get_temporary_ids([_|T],NT) :-
    get_temporary_ids(T,NT).

% we can have duplicate nodes because of undefined types of sets
remove_dup_id_nodes(Nodes,Distinct) :-
    remove_dup_id_nodes(Nodes,[],Distinct).
remove_dup_id_nodes([],Acc,Acc).
remove_dup_id_nodes([Node|T],Acc,Distinct) :-
    Node = b(identifier(Name),_,_) ,
    \+member(b(identifier(Name),_,_),Acc) , ! ,
    remove_dup_id_nodes(T,[Node|Acc],Distinct).
remove_dup_id_nodes([_|T],Acc,Distinct) :-
    remove_dup_id_nodes(T,Acc,Distinct).

% define connections of temporary inputs and outputs
get_connectivity_constraint(InputExampleNodes,OutputExampleNodes,InputVars,OutputVars,ConnectivityConstraint) :-
    append([InputVars,OutputVars,InputExampleNodes,OutputExampleNodes],SetOfVars) ,
    findall((X,Y),(member(X,SetOfVars) , member(Y,SetOfVars) , X \= Y , nth0(I1,SetOfVars,X) , nth0(I2,SetOfVars,Y) , I1 < I2),ListOfVarPairs) ,
    maplist(get_single_connectivity_constraint,ListOfVarPairs,ConstraintList) ,
    conjunct_predicates(ConstraintList,ConnectivityConstraint).

get_single_connectivity_constraint((Var1,Var2),ConnectivityConstraint) :-
    get_location_var_from_ast_node(Var1,LocationVar1) ,
    get_location_var_from_ast_node(Var2,LocationVar2) ,
    % only connect two variables if they have the same type
    (valid_connection(Var1,Var2)
    ->  LHS = b(equal(LocationVar1,LocationVar2),pred,[]) ,
        RHS = b(equal(Var1,Var2),pred,[]) ,
        create_implication(LHS,RHS,ConnectivityConstraint)
    % prevent equality of such incompatible location variables
    ;   ConnectivityConstraint = b(not_equal(LocationVar1,LocationVar2),pred,[])).

valid_connection(Var1,Var2) :-
    get_texpr_info(Var1,Info1) ,
    get_texpr_info(Var2,Info2) ,
    % same type, we use the term synthesis(type,Type) from each node's info instead of the ast node's type
    % reason: we use the type boolean for predicates and real boolean values since one cannot define an identifier of type pred, and thus,
    % a boolean node representing a predicate contains the term synthesis(type,pred) whereat the node itself has the type boolean.
    valid_type_connection(Info1,Info2) , ! ,
    valid_connection_aux(Info1,Info2).

% Check if a location variable refers to an input location.
refers_to_input(Info) :-
    member(synthesis(_,left_input),Info) ; member(synthesis(_,right_input),Info) ; member(synthesis(_,input),Info).

% TODO: maybe rewrite this more pretty, but we definitely have to consider both cases like we do in here
valid_connection_aux(Info1,Info2) :-
    % machineconstants have fixed values, and thus, are not mapped to program input parameters
    member(synthesis(machineconstant,_),Info1) ,
    \+member(synthesis(input(_)),Info2).
valid_connection_aux(Info2,Info1) :-
    member(synthesis(machineconstant,_),Info1) ,
    \+member(synthesis(input(_)),Info2).
valid_connection_aux(Info1,Info2) :-
    % if_then_else only on the top level of a program output
    member(synthesis(ComponentName,output),Info1) ,
    delete_numbers_from_atom(ComponentName,RawName) ,
    RawName = if_then_else ,
    member(synthesis(output(_)),Info2).
valid_connection_aux(Info2,Info1) :-
    member(synthesis(ComponentName,output),Info1) ,
    delete_numbers_from_atom(ComponentName,RawName) ,
    RawName = if_then_else , ! ,
    member(synthesis(output(_)),Info2).
% input of skip component only from program parameters, i.e., the first lines of the program, or the input of another component to not create a contradiction
valid_connection_aux(Info1,Info2) :-
    member(synthesis(ComponentName,input),Info1) ,
    delete_numbers_from_atom(ComponentName,RawName) ,
    RawName = skip ,
    (member(synthesis(input(_)),Info2) ; refers_to_input(Info2)).
valid_connection_aux(Info2,Info1) :-
    member(synthesis(ComponentName,input),Info1) ,
    delete_numbers_from_atom(ComponentName,RawName) ,
    RawName = skip , ! ,
    (member(synthesis(input(_)),Info2) ; refers_to_input(Info2)).
% output of skip component only for program output parameters, i.e., the last lines of the program, or as an if-statement's output
valid_connection_aux(Info1,Info2) :-
    member(synthesis(ComponentName,output),Info1) ,
    delete_numbers_from_atom(ComponentName,RawName) ,
    RawName = skip ,
    (member(synthesis(output(_)),Info2) ; is_if_statement_output(Info2)).
valid_connection_aux(Info2,Info1) :-
    member(synthesis(ComponentName,output),Info1) ,
    delete_numbers_from_atom(ComponentName,RawName) ,
    RawName = skip ,! ,
    (member(synthesis(output(_)),Info2) ; is_if_statement_output(Info2)).
% constant_set only right input of member
valid_connection_aux(Info1,Info2) :-
    member(synthesis(ConstantSet,output),Info1) ,
    delete_numbers_from_atom(ConstantSet,constant_set) , ! ,
    member(synthesis(Member,right_input),Info2) ,
    delete_numbers_from_atom(Member,Op) ,
    (Op = member ; Op = not_member).
valid_connection_aux(Info2,Info1) :-
    member(synthesis(ConstantSet,output),Info1) ,
    delete_numbers_from_atom(ConstantSet,constant_set) , ! ,
    member(synthesis(Member,right_input),Info2) ,
    delete_numbers_from_atom(Member,Op) ,
    (Op = member ; Op = not_member).
valid_connection_aux(Info1,Info2) :-
    member(synthesis(type,boolean),Info1) ,
    (member(synthesis(input(_)),Info2) ;
    (member(synthesis(Comp2,_),Info2) , atom_concat(constant,_,Comp2))).
valid_connection_aux(Info2,Info1) :-
    member(synthesis(type,boolean),Info1) ,
    (member(synthesis(input(_)),Info2) ;
    (member(synthesis(Comp2,_),Info2) , atom_concat(constant,_,Comp2))).
valid_connection_aux(Info1,Info2) :-
    % input to output mapping is not possible in the synthesis constraint anyway, but explicitly prevent this here to support the solver (we use skip components for that)
    member(synthesis(input(_)),Info1) , ! ,
    \+member(synthesis(output(_)),Info2).
valid_connection_aux(Info2,Info1) :-
    member(synthesis(input(_)),Info1) , ! ,
    \+member(synthesis(output(_)),Info2).
valid_connection_aux(_,_).

valid_type_connection(Info1,Info2) :-
    member(synthesis(type,Type1),Info1) ,
    member(synthesis(type,Type2),Info2) ,
    valid_type_connection_aux(Type1,Type2).

valid_type_connection_aux(Type,Type).
% Sequences can also be mapped to set of couples with integer keys. Well definedness is ensured
% by the well-definedness constraint, i.e., sequences have to be enumerated from 1..n
valid_type_connection_aux(Type1,Type2) :-
    (Type1 = seq(InnerType) ; Type1 = set(couple(integer,InnerType))) ,
    (Type2 = seq(InnerType) ; Type2 = set(couple(integer,InnerType))).

is_if_statement_output(Info) :-
    (member(synthesis(Component,true_out),Info) ; member(synthesis(Component,false_out),Info)) ,
    delete_numbers_from_atom(Component,RawComponent) , ! ,
    RawComponent = if_then_else.

get_well_definedness_constraint(VarAmount,LocationVarsIn,LocationVarsOut,M,WellDefinednessConstraint) :-
    get_consistency_constraint(LocationVarsOut,ConsistencyConstraint) ,
    get_acyclic_constraint(LocationVarsIn,LocationVarsOut,AcyclicConstraint) ,
    NewM is M - 1 , % minus one to use an interval
    NewMInput is NewM - 1 , % no input location can get its value from the last line of the internal program
    get_location_constraint(LocationVarsIn,0,NewMInput,InputRestriction) ,
    get_location_constraint(LocationVarsOut,VarAmount,NewM,OutputRestriction) ,
    restrict_if_statements_output_location(LocationVarsOut,VarAmount,NewM,IfOutputRestrictions) ,
    conjunct_predicates([IfOutputRestrictions,InputRestriction,OutputRestriction,ConsistencyConstraint,AcyclicConstraint],
        WellDefinednessConstraint).

% If-statements can only be used on the top level, that is, one of the program outputs.
restrict_if_statements_output_location(LocationVarsOut,VarAmount,NewM,IfOutputRestrictions) :-
    LowerBound is NewM - VarAmount + 1 , % + 1 to use an interval
    findall(IfOutputRestriction,(member(IfOutputLocation,LocationVarsOut) , get_texpr_info(IfOutputLocation,Info) , member(synthesis(Component,output),Info) ,
                                 delete_numbers_from_atom(Component,RawComponent) , RawComponent = if_then_else ,
                                 IfOutputRestriction = b(member(IfOutputLocation,b(interval(b(integer(LowerBound),integer,[]),b(integer(NewM),integer,[])),set(integer),[])),pred,[])),IfOutputRestrictionList) ,
    conjunct_predicates(IfOutputRestrictionList,IfOutputRestrictions).

% create restriction nodes (membership to an interval) to the domain of each location variable
get_location_constraint(LocationVars,A,B,RestrictionNode) :-
    maplist(get_single_location_constraint(A,B),LocationVars,RestrictionList) ,
    conjunct_predicates(RestrictionList,RestrictionNode).
get_single_location_constraint(A,B,LocationVar,RestrictionNode) :-
    RestrictionNode = b(member(LocationVar,
        b(interval(b(integer(A),integer,[]),b(integer(B),integer,[])),set(integer),[])),pred,[]).

% consistency constraint guarantees that there is only one output in each line
get_consistency_constraint(LocationVarsOut,Constraint) :-
    % find all pairs of location variables (not redundant) and create "not_equal"-Node for each pair
    get_consistency_constraint_aux(LocationVarsOut,TConstraint) ,
    %findall(InequalityNode,
    %    (member(L1,LocationVarsOut) , member(L2,LocationVarsOut) ,
    %     L1 \= L2 , nth0(I1,LocationVarsOut,L1) , nth0(I2,LocationVarsOut,L2) ,
    %     I1 < I2 , create_texpr(not_equal(L1,L2),pred,[],InequalityNode)),InequalityList) ,
    conjunct_predicates(TConstraint,Constraint).

get_consistency_constraint_aux([],[]).
get_consistency_constraint_aux([LocationVarOut|T],[Inequality|NT]) :-
    findall(b(not_equal(LocationVarOut,LVar),pred,[]),member(LVar,T),InequalityList) ,
    conjunct_predicates(InequalityList,Inequality) ,
    get_consistency_constraint_aux(T,NT).

% acyclic constraint guarantees that every operator's input is defined before its corresponding output
get_acyclic_constraint(LocationVarsIn,LocationVarsOut,Constraint) :-
    get_acyclic_constraint_list(LocationVarsIn,LocationVarsOut,ConstraintList) ,
    conjunct_predicates(ConstraintList,Constraint).
get_acyclic_constraint_list(_,[],[]).
get_acyclic_constraint_list(LocationVarsIn,[Out|ROut],[Constraint|ConstraintList]) :-
    get_texpr_info(Out,Info) , member(synthesis(ComponentID,output),Info) ,
    get_location_var_node_by_info_term(LocationVarsIn,synthesis(ComponentID,left_input),In1) ,
    get_location_var_node_by_info_term(LocationVarsIn,synthesis(ComponentID,right_input),In2) ,
    create_texpr(less(In1,Out),pred,[],Node1) ,
    create_texpr(less(In2,Out),pred,[],Node2) ,
    conjunct_predicates([Node1,Node2],Constraint) ,
    get_acyclic_constraint_list(LocationVarsIn,ROut,ConstraintList).
get_acyclic_constraint_list(LocationVarsIn,[Out|ROut],[Constraint|ConstraintList]) :-
    get_texpr_info(Out,Info) , member(synthesis(ComponentID,output),Info) ,
    get_location_var_node_by_info_term(LocationVarsIn,synthesis(ComponentID,input),In) ,
    create_texpr(less(In,Out),pred,[],Constraint) ,
    get_acyclic_constraint_list(LocationVarsIn,ROut,ConstraintList).
% skip constants
get_acyclic_constraint_list(LocationVarsIn,[_|ROut],ConstraintList) :-
    get_acyclic_constraint_list(LocationVarsIn,ROut,ConstraintList).

% Construct the library constraint by encoding the given library components as logical formulas.
get_library_constraint(InputValues,Components,(ValidGuards,Operation),Constraint,InputVars,OutputVars,EnumeratedConstants) :-
    create_fixed_constants_to_insert_to_sets(InputValues,FixedSetConstantsConstraint) ,
    maplist(get_single_component_constraint,Components,ConstraintList) ,
    get_machine_constants_constraint(MachineConstantsConstraint) ,
    get_operation_parameters_constraint(ValidGuards,Operation,OperationParametersConstraint) ,
    conjunct_predicates([OperationParametersConstraint,FixedSetConstantsConstraint,MachineConstantsConstraint|ConstraintList],TConstraint) ,
    find_typed_identifier_uses_with_info(TConstraint,ListOfIDNodes) ,
    % split used identifier by input and output
    findall(I,(member(I,ListOfIDNodes) , get_texpr_id(I,IName) , is_component_input(I) , atom_concat(i,_,IName)),InputVars) ,
    findall(O,(member(O,ListOfIDNodes) , get_texpr_id(O,OName) , is_component_output(O) , atom_concat(o,_,OName)),OutputVars) ,
    findall(C,(member(C,ListOfIDNodes) , is_enumerated_constant(C)),EnumeratedConstants) ,
    get_library_wd_constraints(ListOfIDNodes,ListOfIDNodes,[],WDConstraintList) ,
    conjunct_predicates([TConstraint|WDConstraintList],Constraint).

is_component_input(Node) :-
    get_texpr_info(Node,Info) ,
    (member(synthesis(_,left_input),Info) ; member(synthesis(_,right_input),Info) ; member(synthesis(_,input),Info)
    ; member(synthesis(_,condition),Info) ; member(synthesis(_,true_out),Info) ; member(synthesis(_,false_out),Info)) , !.

is_component_output(Node) :-
    get_texpr_info(Node,Info) ,
    member(synthesis(_,output),Info) , !.

is_enumerated_constant(b(identifier(Name),_,Info)) :-
    atom_concat(constant,_,Name) ,
    member(synthesis(enumerated_constant),Info) , !.

:- dynamic additional_set_constants/1.
:- volatile additional_set_constants/1.

% We analyze the current machine variables and add additional set constants.
% B does not provide an operator to directly insert an element in a set. We provide set constants that
% are enumerated by the solver, but in this connection the solver is not able to consider machine variables,
% like enumerating a set constant to be {1,2,i} where i is an integer valued machine variable.
% To that effect, we add set constants containing a single machine variable, like c = {i} which then
% can be unified with a set. This is only necessary if, for instance, a set of integers as well as at least one integer
% valued machine variable is given.
create_fixed_constants_to_insert_to_sets(InputValues,FixedSetConstantsConstraint) :-
    findall((MachineVarName,Type),(member(InputValue,InputValues) , get_texpr_info(InputValue,Info) ,
                                   member(synthesis(type,Type),Info) , member(synthesis(machinevar,MachineVarName),Info)),MachineVarNameTypeTuples) ,
    create_fixed_constants_to_insert_to_sets_aux(MachineVarNameTypeTuples,FixedSetConstantsConstraint).
% We do not need to add additional constants if no set is given.
create_fixed_constants_to_insert_to_sets_aux(MachineVarNameTypeTuples,b(truth,pred,[])) :-
    retractall(additional_set_constants(_)) ,
    assert(additional_set_constants([])) ,
    \+member((_,set(_)),MachineVarNameTypeTuples) , !.
% Otherwise, create additional set constants.
create_fixed_constants_to_insert_to_sets_aux(MachineVarNameTypeTuples,FixedSetConstantsConstraint) :-
    get_set_types(MachineVarNameTypeTuples,TempSetTypes) ,
    findall(NonSetTypeTuple,(member(NonSetTypeTuple,MachineVarNameTypeTuples) , NonSetTypeTuple = (_,Type) , \+member(Type,TempSetTypes)),NonSetTypeTupleTuples) ,
    remove_dups(TempSetTypes,SetTypes) , % we only need one constant for an arbitrary amount of sets of this specific type, like one constant c = {i} for several sets of type integer
    maplist(create_fixed_constants_to_insert_to_sets_aux2(SetTypes,[]),NonSetTypeTupleTuples,TempFixedSetConstantsConstraintList) ,
    findall(FixedConstraint,(member(FixedConstraint,TempFixedSetConstantsConstraintList) , FixedConstraint \= []),FixedSetConstantsConstraintList) ,
    retractall(additional_set_constants(_)) ,
    assert(additional_set_constants(FixedSetConstantsConstraintList)) ,
    conjunct_predicates(FixedSetConstantsConstraintList,FixedSetConstantsConstraint).

create_fixed_constants_to_insert_to_sets_aux2([],[],_,[]). % do not use conjunct_predicates if Acc is an empty list since this create b(truth,pred,[])
create_fixed_constants_to_insert_to_sets_aux2([],Acc,_,FixedSetConstantsConstraint) :-
    conjunct_predicates(Acc,FixedSetConstantsConstraint).
create_fixed_constants_to_insert_to_sets_aux2([SetType|T],Acc,NonSetTypeTuple,FixedSetConstantsConstraint) :-
    create_fixed_constant_for_set_and_machine_var_type(SetType,NonSetTypeTuple,FixedConstantValue) , ! ,
    unique_typed_id("o",SetType,TempOut) ,
    get_texpr_id(TempOut,OutId) ,
    add_texpr_infos(TempOut,[synthesis(OutId,output),synthesis(type,SetType),synthesis(machineconstant,FixedConstantValue)],Out) ,
    create_texpr(equal(Out,FixedConstantValue),pred,[],FixedConstant) ,
    create_fixed_constants_to_insert_to_sets_aux2(T,[FixedConstant|Acc],NonSetTypeTuple,FixedSetConstantsConstraint).
create_fixed_constants_to_insert_to_sets_aux2([_|T],Acc,NonSetTypeTuple,FixedSetConstantsConstraint) :-
    % skip: the set type does not match the machine var type, that is, NonSetTypeTuple
    create_fixed_constants_to_insert_to_sets_aux2(T,Acc,NonSetTypeTuple,FixedSetConstantsConstraint).

% Fails if the inner type of the set does not match the machine var type, that is, NonSetTypeTuple
create_fixed_constant_for_set_and_machine_var_type(InnerType,(MachineVarName,InnerType),MachineVar) :-
    b_get_machine_variables(MachineVars) ,
    member(MachineVar,MachineVars) ,
    get_texpr_id(MachineVar,IdName) , IdName = MachineVarName.
% We also consider nested sets of sets here.
create_fixed_constant_for_set_and_machine_var_type(set(InnerType),NonSetTypeTuple,b(set_extension([PartialFixedConstant]),set(InnerType),[])) :-
    create_fixed_constant_for_set_and_machine_var_type(InnerType,NonSetTypeTuple,PartialFixedConstant).

get_set_types(MachineVarNameTypeTuples,SetTypes) :-
    get_set_types_aux(MachineVarNameTypeTuples,[],SetTypes).
get_set_types_aux([],Acc,Acc).
get_set_types_aux([(_,set(InnerType))|T],Acc,SetTypes) :- ! ,
    get_set_types_aux(T,[set(InnerType)|Acc],SetTypes).
get_set_types_aux([_|T],Acc,SetTypes) :-
    get_set_types_aux(T,Acc,SetTypes).

% Create synthesis constants for operation parameters and consider the valid guards of the operation which hold
% the typing of each parameter (this assumption is a fact since otherwise the machine is not well-defined).
get_operation_parameters_constraint(ValidGuards,Operation,OperationParametersConstraint) :-
    b_get_machine_operation(Operation,_,OperationParameters,_) , ValidGuards \= [] , ! ,
    get_operation_parameters_constraint_aux(OperationParameters,[],ConstraintList) ,
    get_operation_parameters_constraint_aux2(ValidGuards,ConstraintList,OperationParametersConstraint).
% The operation does not exist yet, e.g., because we synthesize a new operation from scratch.
get_operation_parameters_constraint(_,_,b(truth,pred,[])).

get_operation_parameters_constraint_aux([],Acc,Acc).
get_operation_parameters_constraint_aux([OperationParameter|T],Acc,ConstraintList) :-
    get_texpr_type(OperationParameter,ParameterType) ,
    get_texpr_id(OperationParameter,ParameterName) ,
    unique_typed_id("o",ParameterType,TempOut) ,
    add_texpr_infos(TempOut,[synthesis(ParameterName,output),synthesis(type,ParameterType),synthesis(operationparameter,OperationParameter)],Out) ,
    create_texpr(equal(Out,OperationParameter),pred,[],Constraint) ,
    get_machine_constants_constraint_aux(T,[Constraint|Acc],ConstraintList).

get_operation_parameters_constraint_aux2(_,[],b(truth,pred,[])).
get_operation_parameters_constraint_aux2(ValidGuards,ConstraintList,OperationParametersConstraint) :-
    append(ValidGuards,ConstraintList,NewConstraintList) ,
    conjunct_predicates(NewConstraintList,OperationParametersConstraint).

% Create synthesis constants for machine constants and consider the machine properties defining the values of
% machine constants. Therefore, these synthesis constants will not be enumerated by the solver.
get_machine_constants_constraint(MachineConstantsConstraint) :-
    b_get_machine_constants(MachineConstants) ,
    get_machine_constants_constraint_aux(MachineConstants,[],ConstraintList) ,
    conjunct_predicates(ConstraintList,MachineConstantsConstraint).

get_machine_constants_constraint_aux([],Acc,[MachineProperties|Acc]) :-
    % append the machine properties defining the machine constants
    b_get_properties_from_machine(MachineProperties).
get_machine_constants_constraint_aux([MachineConstant|T],Acc,ConstraintList) :-
    get_texpr_type(MachineConstant,ConstantType) ,
    get_texpr_id(MachineConstant,ConstantName) ,
    unique_typed_id("o",ConstantType,TempOut) ,
    add_texpr_infos(TempOut,[synthesis(ConstantName,output),synthesis(type,ConstantType),synthesis(machineconstant,MachineConstant)],Out) ,
    create_texpr(equal(Out,MachineConstant),pred,[],Constraint) ,
    get_machine_constants_constraint_aux(T,[Constraint|Acc],ConstraintList).

% Create well-definedness constraints to prevent division by zero, max/min of empty sets or
% front/tail/last/first/perm on empty sequences or power_of with a negative number.
get_library_wd_constraints(_,[],Acc,Acc).
get_library_wd_constraints(ListOfIDNodes,[b(Node,Type,Info)|T],Acc,WDConstraintList) :-
    \+member(synthesis(_,output),Info) ,
    member(synthesis(OpName,_),Info) ,
    get_library_wd_constraints_aux(ListOfIDNodes,OpName,b(Node,Type,Info),AccWDConstraint) ,
    get_library_wd_constraints(ListOfIDNodes,T,[AccWDConstraint|Acc],WDConstraintList).
get_library_wd_constraints(ListOfIDNodes,[_|T],Acc,WDConstraintList) :-
    get_library_wd_constraints(ListOfIDNodes,T,Acc,WDConstraintList).

get_library_wd_constraints_aux(ListOfIDNodes,OpName,_,b(less_equal(LeftInput,RightInput),pred,[])) :-
    % left input of interval is smaller or equal than the right input
    delete_numbers_from_atom(OpName,RawOpName) ,
    RawOpName = interval ,
    get_var_node_by_info_term(ListOfIDNodes,synthesis(OpName,left_input),LeftInput) ,
    get_var_node_by_info_term(ListOfIDNodes,synthesis(OpName,right_input),RightInput).
get_library_wd_constraints_aux(ListOfIDNodes,OpName,_,b(conjunct(b(greater(LeftInput,b(integer(-1),integer,[])),pred,[]),b(greater(RightInput,b(integer(-1),integer,[])),pred,[])),pred,[])) :-
    % modulo only for positive numbers
    delete_numbers_from_atom(OpName,RawOpName) ,
    RawOpName = modulo ,
    get_var_node_by_info_term(ListOfIDNodes,synthesis(OpName,left_input),LeftInput) ,
    get_var_node_by_info_term(ListOfIDNodes,synthesis(OpName,right_input),RightInput).
get_library_wd_constraints_aux(ListOfIDNodes,OpName,_,b(not_equal(LeftInput,RightInput),pred,[])) :-
    % prohibit the same inputs for subset/subset_strict like 's <<: s'
    delete_numbers_from_atom(OpName,RawOpName) ,
    member(RawOpName,[subset]) , % subset_strict (union and intersection is too restrictive here)
    get_var_node_by_info_term(ListOfIDNodes,synthesis(OpName,left_input),LeftInput) ,
    get_var_node_by_info_term(ListOfIDNodes,synthesis(OpName,right_input),RightInput).
get_library_wd_constraints_aux(_,OpName,Node,b(not_equal(Node,b(integer(0),integer,[])),pred,[])) :-
    % to only check division by zero we would need to check that the variable is the right input, thus the denominator,
    % but we also want to prevent 0/c since this is just the constant value 0
    delete_numbers_from_atom(OpName,div).
get_library_wd_constraints_aux(ListOfIDNodes,OpName,_,b(member(RightInput,b(interval(b(integer(0),integer,[]),b(integer(10),integer,[])),set(integer),[])),pred,[])) :-
    % the exponent of power_of has to be positive and we furthermore restrict the exponent to be less or equal to 10
    delete_numbers_from_atom(OpName,RawOpName) ,
    RawOpName = power_of ,
    get_var_node_by_info_term(ListOfIDNodes,synthesis(OpName,right_input),RightInput).
get_library_wd_constraints_aux(_,OpName,Node,b(greater(b(card(Node),integer,[]),b(integer(0),integer,[])),pred,[])) :-
    delete_numbers_from_atom(OpName,RawOpName) ,
    member(RawOpName,[max,min,front,tail,first,last,perm,general_intersection]).

% Library components represented as logical formulas.

get_single_component_constraint(if_then_else-Type,Constraint) :-
    unique_typed_id("i",boolean,TempCondition) ,
    unique_typed_id("i",Type,TempIn1) ,
    unique_typed_id("i",Type,TempIn2) ,
    unique_typed_id("o",Type,TempOut) ,
    set_component_id(if_then_else,UniqueComponent) ,
    % set information of identifier nodes
    add_texpr_infos(TempCondition,[synthesis(type,pred),synthesis(UniqueComponent,condition)],Condition) ,
    add_texpr_infos(TempIn1,[synthesis(type,Type),synthesis(UniqueComponent,true_out)],In1) ,
    add_texpr_infos(TempIn2,[synthesis(type,Type),synthesis(UniqueComponent,false_out)],In2) ,
    add_texpr_infos(TempOut,[synthesis(type,Type),synthesis(UniqueComponent,output)],Out) ,
    Node =.. [if_then_else,Condition,In1,In2] ,
    Constraint = b(equal(b(Node,Type,[]),Out),pred,[]).
% (INTEGER,INTEGER) :: INTEGER
get_single_component_constraint(Component,Constraint) :-
    member(Component,[add,minus,multiplication,div,modulo,power_of]) , ! ,
    unique_typed_id("i",integer,TempIn1) ,
    unique_typed_id("i",integer,TempIn2) ,
    unique_typed_id("o",integer,TempOut) ,
    set_component_id(Component,UniqueComponent) ,
    % set information of identifier nodes
    add_texpr_infos(TempIn1,[synthesis(type,integer),synthesis(UniqueComponent,left_input)],In1) ,
    add_texpr_infos(TempIn2,[synthesis(type,integer),synthesis(UniqueComponent,right_input)],In2) ,
    add_texpr_infos(TempOut,[synthesis(type,integer),synthesis(UniqueComponent,output)],Out) ,
    Node =.. [Component,In1,In2] ,
    Constraint = b(equal(b(Node,integer,[]),Out),pred,[]).
% TYPE :: TYPE
get_single_component_constraint(skip-Type,b(equal(In,Out),pred,[])) :-
    % We can't map inputs to outputs directly since input parameters are located in the first lines of the program
    % whereat output parameter are located in the last lines, and thus, their locations can't match.
    % Therefore, we use a simple wrapper allowing to connect input to output locations implicitly to skip values.
    unique_typed_id("i",Type,TempIn) ,
    unique_typed_id("o",Type,TempOut) ,
    set_component_id(skip,ComponentName) ,
    add_texpr_infos(TempIn,[synthesis(ComponentName,input),synthesis(type,Type)],In) ,
    add_texpr_infos(TempOut,[synthesis(ComponentName,output),synthesis(type,Type)],Out).
% (BOOLEAN,BOOLEAN) :: BOOLEAN
/*get_single_component_constraint(Operator,Constraint) :-
    member(Operator,[conjunct,disjunct,implication,equivalence]) , ! ,
    unique_typed_id("i",boolean,TempIn1) ,
    unique_typed_id("i",boolean,TempIn2) ,
    unique_typed_id("o",boolean,TempOut) ,
    set_component_id(Operator,ComponentName) ,
    % set information of identifier nodes
    add_texpr_infos(TempIn1,[synthesis(ComponentName,left_input),synthesis(type,pred)],In1) ,
    add_texpr_infos(TempIn2,[synthesis(ComponentName,right_input),synthesis(type,pred)],In2) ,
    add_texpr_infos(TempOut,[synthesis(ComponentName,output),synthesis(type,pred)],Out) ,
    Node =.. [Operator,In1,In2] ,
    Constraint = b(equal(Out,b(convert_bool(b(Node,pred,[])),boolean,[])),pred,[]).*/
get_single_component_constraint(Operator,Constraint) :-
    member(Operator,[conjunct,disjunct,implication,equivalence]) , ! ,
    unique_typed_id("i",boolean,TempIn1) ,
    unique_typed_id("i",boolean,TempIn2) ,
    unique_typed_id("o",boolean,TempOut) ,
    set_component_id(Operator,ComponentName) ,
    % set information of identifier nodes
    add_texpr_infos(TempIn1,[synthesis(ComponentName,left_input),synthesis(type,pred)],In1) ,
    add_texpr_infos(TempIn2,[synthesis(ComponentName,right_input),synthesis(type,pred)],In2) ,
    add_texpr_infos(TempOut,[synthesis(ComponentName,output),synthesis(type,pred)],Out) ,
    get_logical_constraint(Operator,In1,In2,Out,Truth,Falsity) ,
    disjunct_predicates([Truth,Falsity],Constraint).
% (PRED) :: BOOLEAN
get_single_component_constraint(convert_bool,Constraint) :-
    unique_typed_id("i",boolean,TempIn) ,
    unique_typed_id("o",boolean,TempOut) ,
    set_component_id(convert_bool,ComponentName) ,
    % set information of identifier nodes
    add_texpr_infos(TempIn,[synthesis(ComponentName,input),synthesis(type,pred)],In) ,
    add_texpr_infos(TempOut,[synthesis(ComponentName,output),synthesis(type,boolean)],Out) ,
    Constraint = b(equal(Out,In),pred,[]).
% (BOOLEAN :: BOOLEAN)
get_single_component_constraint(negation,Constraint) :- ! ,
    unique_typed_id("i",boolean,TempIn) ,
    unique_typed_id("o",boolean,TempOut) ,
    set_component_id(negation,Negation) ,
    % set information of identifier nodes
    add_texpr_infos(TempIn,[synthesis(Negation,input),synthesis(type,pred)],In) ,
    add_texpr_infos(TempOut,[synthesis(Negation,output),synthesis(type,pred)],Out) ,
    Truth = b(conjunct(b(equal(In,b(boolean_false,boolean,[])),pred,[]),
        b(equal(Out,b(boolean_true,boolean,[])),pred,[])),pred,[]) ,
    Falsity = b(conjunct(b(equal(In,b(boolean_true,boolean,[])),pred,[]),
        b(equal(Out,b(boolean_false,boolean,[])),pred,[])),pred,[]) ,
    disjunct_predicates([Truth,Falsity],Constraint).
% explicit set constants referring to enumerated/deferred sets accessed via operation parameters
% create two constants p and {p} for each operation parameter p
% both constants are set up in one call to ensure that the shared identifier p is the same
get_single_component_constraint(param_constant-global(_),b(truth,pred,[])). % so we skip here and set the constraint below
get_single_component_constraint(param_constant-set(global(IndependentType)),Constraint) :-
    get_single_constant_or_set_operator_id(constant-global(IndependentType),TempOutSingle,ConstantSingle) ,
    get_single_constant_or_set_operator_id(constant-set(global(IndependentType)),TempOutSet,ConstantSet) , ! ,
    get_texpr_id(ConstantSingle,ConstantSingleName) ,
    get_texpr_id(ConstantSet,ConstantSetName) ,
    ParamId = b(identifier(ParamName),global(IndependentType),[synthesis(ConstantSingleName,output),synthesis(type,global(IndependentType)),synthesis(param_single,ParamName),synthesis(ConstantSetName,output),synthesis(type,set(global(IndependentType))),synthesis(param_set,ParamName)]) , % the same parameter identifier p for both constants c1 = p, c2 = {p}
    % set information of identifier node
    add_texpr_infos(TempOutSingle,[synthesis(ConstantSingleName,output),synthesis(type,global(IndependentType)),synthesis(param_single,ParamName)],OutSingle) ,
    add_texpr_infos(TempOutSet,[synthesis(ConstantSetName,output),synthesis(type,set(global(IndependentType))),synthesis(param_set,ParamName)],OutSet) ,
    atom_concat(p_,IndependentType,ParamName) ,
    MachineSet = b(identifier(IndependentType),set(global(IndependentType)),[]) ,
    Constraint = b(conjunct(b(conjunct(b(member(ParamId,MachineSet),pred,[]),b(equal(b(set_extension([ParamId]),set(global(IndependentType)),[]),OutSet),pred,[])),pred,[]),b(equal(ParamId,OutSingle),pred,[])),pred,[]).
% constant :: (INTEGER,BOOLEAN,SET)
get_single_component_constraint(constant-Type,Constraint) :-
    get_single_constant_or_set_operator_id(constant-Type,TempOut,Constant) , ! ,
    get_texpr_id(Constant,ConstantName) ,
    % set information of identifier node
    add_texpr_infos(TempOut,[synthesis(ConstantName,output),synthesis(type,Type)],Out) ,
    Constraint = b(equal(Constant,Out),pred,[]).
% bool_set, string_set, integer_set,..
get_single_component_constraint(Operator,Constraint) :-
    get_single_constant_or_set_operator_id(Operator,TempOut,b(_,Type,_)) , ! ,
    unique_typed_id("constant_set",Type,b(identifier(ConstantName),Type,Info)) ,
    add_texpr_infos(TempOut,[synthesis(ConstantName,output),synthesis(type,Type)],Out) ,
    % TODO: define as subset to avoid infinite sets (virtual timeout) ?
    Constraint = b(equal(Out,b(identifier(ConstantName),Type,Info)),pred,[]).
% set-expressions
% SET :: INTEGER
get_single_component_constraint(Operator,Constraint) :-
    member(Operator,[min,max]) ,
    unique_typed_id("i",set(integer),TempIn) ,
    unique_typed_id("o",integer,TempOut) ,
    set_component_id(Operator,InternalName) ,
    add_texpr_infos(TempIn,[synthesis(InternalName,input),synthesis(type,set(integer))],In) ,
    add_texpr_infos(TempOut,[synthesis(InternalName,output),synthesis(type,integer)],Out) ,
    Node =.. [Operator,In] ,
    Constraint = b(equal(b(Node,integer,[]),Out),pred,[]).
get_single_component_constraint(card-SetType,Constraint) :-
    unique_typed_id("i",SetType,TempIn) ,
    unique_typed_id("o",integer,TempOut) ,
    set_component_id(card,InternalName) ,
    add_texpr_infos(TempIn,[synthesis(InternalName,input),synthesis(type,SetType)],In) ,
    add_texpr_infos(TempOut,[synthesis(InternalName,output),synthesis(type,integer)],Out) ,
    Constraint = b(equal(b(card(In),integer,[]),Out),pred,[]).
% SET :: SET(SET)
get_single_component_constraint(Operator-SetType,Constraint) :-
    member(Operator,[pow_subset,pow_subset1,fin_subset,fin1_subset]) ,
    unique_typed_id("i",SetType,TempIn) ,
    unique_typed_id("o",set(SetType),TempOut) ,
    set_component_id(Operator,NOperator) ,
    add_texpr_infos(TempIn,[synthesis(NOperator,input),synthesis(type,SetType)],In) ,
    add_texpr_infos(TempOut,[synthesis(NOperator,output),synthesis(type,set(SetType))],Out) ,
    Node =.. [Operator,In] ,
    Constraint = b(equal(b(Node,set(SetType),[]),Out),pred,[]).
% SEQ(TYPE) :: SET(SEQ(TYPE))
get_single_component_constraint(Operator-SetType,Constraint) :-
    member(Operator,[seq,seq1,iseq]) ,
    SetType = set(InnerType) ,
    unique_typed_id("i",SetType,TempIn) ,
    unique_typed_id("o",set(seq(InnerType)),TempOut) ,
    set_component_id(Operator,InternalName) ,
    add_texpr_infos(TempIn,[synthesis(InternalName,input),synthesis(type,SetType)],In) ,
    add_texpr_infos(TempOut,[synthesis(InternalName,output),synthesis(type,set(seq(InnerType)))],Out) ,
    Node =.. [Operator,In] ,
    Constraint = b(equal(b(Node,set(seq(InnerType)),[]),Out),pred,[]).
% SEQ(TYPE) :: SET(SEQ(TYPE))
get_single_component_constraint(perm-SeqType,b(equal(b(perm(In),OutputType,[]),Out),pred,[])) :-
    get_perm_component_output_type(SeqType,OutputType) ,
    unique_typed_id("i",SeqType,TempIn) ,
    unique_typed_id("o",OutputType,TempOut) ,
    set_component_id(perm,InternalName) ,
    add_texpr_infos(TempIn,[synthesis(InternalName,input),synthesis(type,SeqType)],In) ,
    add_texpr_infos(TempOut,[synthesis(InternalName,output),synthesis(type,OutputType)],Out).
% SEQ(TYPE) :: SET(SEQ(TYPE))
get_single_component_constraint(interval,b(equal(b(interval(In1,In2),set(integer),[]),Out),pred,[])) :-
    unique_typed_id("i",integer,TempIn1) ,
    unique_typed_id("i",integer,TempIn2) ,
    unique_typed_id("o",set(integer),TempOut) ,
    set_component_id(interval,InternalName) ,
    add_texpr_infos(TempIn1,[synthesis(InternalName,left_input),synthesis(type,integer)],In1) ,
    add_texpr_infos(TempIn2,[synthesis(InternalName,right_input),synthesis(type,integer)],In2) ,
    add_texpr_infos(TempOut,[synthesis(InternalName,output),synthesis(type,set(integer))],Out).
% SET :: SET
get_single_component_constraint(Operator-SetType,Constraint) :-
    member(Operator,[closure,reflexive_closure]) ,
    unique_typed_id("i",SetType,TempIn) ,
    unique_typed_id("o",SetType,TempOut) ,
    set_component_id(Operator,NOperator) ,
    add_texpr_infos(TempIn,[synthesis(NOperator,input),synthesis(type,SetType)],In) ,
    add_texpr_infos(TempOut,[synthesis(NOperator,output),synthesis(type,SetType)],Out) ,
    Node =.. [Operator,In] ,
    Constraint = b(equal(b(Node,SetType,[]),Out),pred,[]).
% (SET,SET) :: SET
get_single_component_constraint(Operator-SetType,Constraint) :-
    member(Operator,[union,intersection,concat,insert_front,insert_tail,restrict_front,
                     restrict_tail,overwrite,set_subtraction,image,cartesian_product,
                     domain_restriction,domain_subtraction,range_restriction,range_subtraction]) ,
    get_single_component_constraint_aux3(Operator-SetType,TempIn1,TempIn2,TempOut,Operator) ,
    set_component_id(Operator,NOperator) ,
    get_texpr_type(TempIn1,In1Type) ,
    get_texpr_type(TempIn2,In2Type) ,
    get_texpr_type(TempOut,OutType) ,
    add_texpr_infos(TempIn1,[synthesis(NOperator,left_input),synthesis(type,In1Type)],In1) ,
    add_texpr_infos(TempIn2,[synthesis(NOperator,right_input),synthesis(type,In2Type)],In2) ,
    add_texpr_infos(TempOut,[synthesis(NOperator,output),synthesis(type,OutType)],Out) ,
    Node =.. [Operator,In1,In2] ,
    Constraint = b(equal(b(Node,SetType,[]),Out),pred,[]).
get_single_component_constraint(Operator-SetType,Constraint) :-
    member(Operator,[first,last,front,tail,general_union,general_intersection,reverse,range,domain]) ,
    get_single_component_constraint_aux3(Operator-SetType,TempIn,_,TempOut,Operator) ,
    set_component_id(Operator,NOperator) ,
    get_texpr_type(TempIn,InType) ,
    get_texpr_type(TempOut,OutType) ,
    add_texpr_infos(TempIn,[synthesis(NOperator,input),synthesis(type,InType)],In) ,
    add_texpr_infos(TempOut,[synthesis(NOperator,output),synthesis(type,OutType)],Out) ,
    Node =.. [Operator,In] ,
    Constraint = b(equal(b(Node,SetType,[]),Out),pred,[]).
get_single_component_constraint(size-SetType,Constraint) :-
    get_fn_types(SetType,integer,Type) ,
    unique_typed_id("i",seq(Type),TempIn) ,
    unique_typed_id("o",integer,TempOut) ,
    set_component_id(size,NOperator) ,
    add_texpr_infos(TempIn,[synthesis(NOperator,input),synthesis(type,seq(Type))],In) ,
    add_texpr_infos(TempOut,[synthesis(NOperator,output),synthesis(type,integer)],Out) ,
    Constraint = b(equal(b(size(In),integer,[]),Out),pred,[]).

% (INTEGER,INTEGER) :: BOOLEAN || (BOOLEAN,BOOLEAN) :: BOOLEAN || (SET,SET) :: BOOLEAN
get_single_component_constraint(Operator,Truth) :-
    get_single_component_constraint_aux2(Operator,TempIn1,TempIn2,TempOut,Op) ,
    set_component_id(Op,NOperator) ,
    add_texpr_infos(TempIn1,[synthesis(NOperator,left_input)],In1) ,
    add_texpr_infos(TempIn2,[synthesis(NOperator,right_input)],In2) ,
    add_texpr_infos(TempOut,[synthesis(NOperator,output)],Out) ,
    TempTrueNode =.. [Op,In1,In2] ,
    TrueNode = b(TempTrueNode,pred,[]) ,
    BoolNodeTrue = b(equal(Out,b(boolean_true,boolean,[])),pred,[]) ,
    create_equivalence(TrueNode,BoolNodeTrue,Truth).

get_single_component_constraint_aux2(Operator-Type,In1,In2,Out,Operator) :-
    member(Operator,[equal,not_equal,subset,subset_strict,not_subset,not_subset_strict]) , ! ,
    unique_typed_id("i",Type,TempIn1) ,
    unique_typed_id("i",Type,TempIn2) ,
    unique_typed_id("o",boolean,TempOut) ,
    add_texpr_infos(TempIn1,[synthesis(type,Type)],In1) ,
    add_texpr_infos(TempIn2,[synthesis(type,Type)],In2) ,
    add_texpr_infos(TempOut,[synthesis(type,pred)],Out).
get_single_component_constraint_aux2(Operator-SetType,In1,In2,Out,Operator) :-
    member(Operator,[member,not_member]) , ! ,
    (SetType = set(Type) ; SetType = seq(Type)) ,
    unique_typed_id("i",Type,TempIn1) ,
    unique_typed_id("i",SetType,TempIn2) ,
    unique_typed_id("o",boolean,TempOut) ,
    add_texpr_infos(TempIn1,[synthesis(type,Type)],In1) ,
    add_texpr_infos(TempIn2,[synthesis(type,SetType)],In2) ,
    add_texpr_infos(TempOut,[synthesis(type,pred)],Out).
% greater,greater_equal,less,..
get_single_component_constraint_aux2(Operator,In1,In2,Out,Operator) :-
    unique_typed_id("i",integer,TempIn1) ,
    unique_typed_id("i",integer,TempIn2) ,
    unique_typed_id("o",boolean,TempOut) ,
    add_texpr_infos(TempIn1,[synthesis(type,integer)],In1) ,
    add_texpr_infos(TempIn2,[synthesis(type,integer)],In2) ,
    add_texpr_infos(TempOut,[synthesis(type,pred)],Out).

% set- and seq-expressions
get_single_component_constraint_aux3(Operator-SetType,TempIn,_,TempOut,Operator) :-
    member(Operator,[general_union,general_intersection]) ,
    unique_typed_id("i",SetType,TempIn) ,
    SetType = set(InnerType) ,
    unique_typed_id("o",InnerType,TempOut).
get_single_component_constraint_aux3(image-SetType,TempIn1,TempIn2,TempOut,image) :-
    get_fn_types(SetType,A,B) ,
    unique_typed_id("i",set(couple(A,B)),TempIn1) ,
    unique_typed_id("i",set(A),TempIn2) ,
    unique_typed_id("o",set(B),TempOut).
get_single_component_constraint_aux3(cartesian_product-SetType,TempIn1,TempIn2,TempOut,cartesian_product) :-
    SetType = set(InnerType) ,      % TODO: what about cartesian product of two differently typed sets?
    unique_typed_id("i",set(InnerType),TempIn1) ,
    unique_typed_id("i",set(InnerType),TempIn2) ,
    unique_typed_id("o",set(couple(InnerType,InnerType)),TempOut).
get_single_component_constraint_aux3(Operator-SetType,TempIn1,TempIn2,TempOut,Operator) :-
    member(Operator,[domain_restriction,domain_subtraction]) ,
    get_fn_types(SetType,A,B) ,
    unique_typed_id("i",set(A),TempIn1) ,
    unique_typed_id("i",set(couple(A,B)),TempIn2) ,
    unique_typed_id("o",set(couple(A,B)),TempOut).
get_single_component_constraint_aux3(Operator-SetType,TempIn1,TempIn2,TempOut,Operator) :-
    member(Operator,[range_restriction,range_subtraction]) ,
    get_fn_types(SetType,A,B) ,
    unique_typed_id("i",set(couple(A,B)),TempIn1) ,
    unique_typed_id("i",set(B),TempIn2) ,
    unique_typed_id("o",set(couple(A,B)),TempOut).
get_single_component_constraint_aux3(reverse-SeqType,TempIn,_,TempOut,reverse) :-
    unique_typed_id("i",SeqType,TempIn) ,
    unique_typed_id("o",SeqType,TempOut).
get_single_component_constraint_aux3(domain-SetType,TempIn,_,TempOut,domain) :-
    get_fn_types(SetType,A,_) ,
    unique_typed_id("i",SetType,TempIn) ,
    unique_typed_id("o",set(A),TempOut).
get_single_component_constraint_aux3(range-SetType,TempIn,_,TempOut,range) :-
    get_fn_types(SetType,_,B) ,
    unique_typed_id("i",SetType,TempIn) ,
    unique_typed_id("o",set(B),TempOut).
get_single_component_constraint_aux3(Operator-SetType,TempIn1,TempIn2,TempOut,Operator) :-
    member(Operator,[union,intersection,set_subtraction]) ,
    unique_typed_id("i",SetType,TempIn1) ,
    unique_typed_id("i",SetType,TempIn2) ,
    unique_typed_id("o",SetType,TempOut).
get_single_component_constraint_aux3(overwrite-SetType,TempIn1,TempIn2,TempOut,overwrite) :-
    get_fn_types(SetType,integer,Type) ,
    unique_typed_id("i",seq(Type),TempIn1) ,
    unique_typed_id("i",seq(Type),TempIn2) ,
    unique_typed_id("o",seq(Type),TempOut).
get_single_component_constraint_aux3(concat-SeqType,TempIn1,TempIn2,TempOut,concat) :-
    unique_typed_id("i",SeqType,TempIn1) ,
    unique_typed_id("i",SeqType,TempIn2) ,
    unique_typed_id("o",SeqType,TempOut).
get_single_component_constraint_aux3(Operator-SetType,TempIn,_,TempOut,Operator) :-
    member(Operator,[first,last]) ,
    unique_typed_id("i",SetType,TempIn) ,
    (SetType = seq(InnerType) ; SetType = set(couple(integer,InnerType))) ,
    unique_typed_id("o",InnerType,TempOut).
get_single_component_constraint_aux3(Operator-SetType,TempIn,_,TempOut,Operator) :-
    member(Operator,[front,tail]) ,
    unique_typed_id("i",SetType,TempIn) ,
    unique_typed_id("o",SetType,TempOut).
get_single_component_constraint_aux3(insert_front-SetType,TempIn1,TempIn2,TempOut,insert_front) :-
    get_fn_types(SetType,integer,InnerType) ,
    unique_typed_id("i",InnerType,TempIn1) ,
    unique_typed_id("i",SetType,TempIn2) ,
    unique_typed_id("o",SetType,TempOut).
get_single_component_constraint_aux3(insert_tail-SetType,TempIn1,TempIn2,TempOut,insert_tail) :-
    get_fn_types(SetType,integer,InnerType) ,
    unique_typed_id("i",SetType,TempIn1) ,
    unique_typed_id("i",InnerType,TempIn2) ,
    unique_typed_id("o",SetType,TempOut).
get_single_component_constraint_aux3(Operator-SetType,TempIn1,TempIn2,TempOut,Operator) :-
    member(Operator,[restrict_front,restrict_tail]) ,
    unique_typed_id("i",SetType,TempIn1) ,
    unique_typed_id("i",integer,TempIn2) ,
    unique_typed_id("o",SetType,TempOut).

% Operators
% Use boolean instead of pred_true/pred_false. We can't define identifier ASTs of type predicate.
get_logical_constraint(implication,In1,In2,Out,Truth,Falsity) :-
    Truth = b(conjunct(b(disjunct(
        b(equal(In1,b(boolean_false,boolean,[])),pred,[]),
        b(equal(In2,b(boolean_true,boolean,[])),pred,[])),pred,[]),
        b(equal(Out,b(boolean_true,boolean,[])),pred,[])),pred,[]) ,
    Falsity = b(conjunct(b(conjunct(
        b(equal(In1,b(boolean_true,boolean,[])),pred,[]),
        b(equal(In2,b(boolean_false,boolean,[])),pred,[])),pred,[]),
        b(equal(Out,b(boolean_false,boolean,[])),pred,[])),pred,[]).
get_logical_constraint(equivalence,In1,In2,Out,Truth,Falsity) :-
    get_logical_constraint(implication,In1,In2,Out,TruthImpl1,FalsityImpl1) ,
    get_logical_constraint(implication,In2,In1,Out,TruthImpl2,FalsityImpl2) ,
    Truth = b(conjunct(TruthImpl1,TruthImpl2),pred,[]) ,
    Falsity = b(disjunct(FalsityImpl1,FalsityImpl2),pred,[]).
get_logical_constraint(conjunct,In1,In2,Out,Truth,Falsity) :-
    Truth = b(conjunct(b(conjunct(
        b(equal(In1,b(boolean_true,boolean,[])),pred,[]),
        b(equal(In2,b(boolean_true,boolean,[])),pred,[])),pred,[]),
        b(equal(Out,b(boolean_true,boolean,[])),pred,[])),pred,[]) ,
    Falsity = b(conjunct(b(disjunct(
        b(equal(In1,b(boolean_false,boolean,[])),pred,[]),
        b(equal(In2,b(boolean_false,boolean,[])),pred,[])),pred,[]),
        b(equal(Out,b(boolean_false,boolean,[])),pred,[])),pred,[]).
get_logical_constraint(disjunct,In1,In2,Out,Truth,Falsity) :-
    Truth = b(conjunct(b(disjunct(
        b(equal(In1,b(boolean_true,boolean,[])),pred,[]),
        b(equal(In2,b(boolean_true,boolean,[])),pred,[])),pred,[]),
        b(equal(Out,b(boolean_true,boolean,[])),pred,[])),pred,[]) ,
    Falsity = b(conjunct(b(conjunct(
        b(equal(In1,b(boolean_false,boolean,[])),pred,[]),
        b(equal(In2,b(boolean_false,boolean,[])),pred,[])),pred,[]),
        b(equal(Out,b(boolean_false,boolean,[])),pred,[])),pred,[]).

get_fn_types(seq(Type),integer,Type).
get_fn_types(set(couple(A,B)),A,B).

% Get the output type of permutations component, i.e., perm(s) for a sequence s.
get_perm_component_output_type(set(CoupleType),set(seq(CoupleType))) :-
    CoupleType = couple(integer,_).
get_perm_component_output_type(seq(InnerType),set(seq(couple(integer,InnerType)))).

% Add prefix "l" to Var for connectivity constraint, add the type to the location variables
% information and adopt the given information which encodes the location variables reference,
% e.g. [synthesis(type,integer),synthesis(left_input,greater)]
get_location_var_from_ast_node(b(identifier(Name),_,Info),b(identifier(LocationName),integer,Info)) :-
    ! , atom_concat(l,Name,LocationName).
get_location_var_from_ast_node(b(_,Type,Info),b(identifier(LocationName),integer,Info)) :-
    get_location_var_from_ast_node_aux(Type,ValueCode) ,
    % set input/output enumeration as prefix to get unique name
    % output location names are like 'lovalue0boolean_true'
    member(synthesis(example(ExampleCount)),Info) ,
    number_codes(ExampleCount,ExampleCountCode) ,
    (member(synthesis(input(C)),Info)
    ->  number_codes(C,NC) ,
        append([[105,110,112,117,116],NC,ExampleCountCode],CCode)
    ;   member(synthesis(output(C)),Info) ,
        number_codes(C,NumberCodes) ,
        append([[111,118,97,108,117,101],NumberCodes,ExampleCountCode],CCode)) ,
    append(CCode,ValueCode,Codes) ,
    atom_codes(LocationName,[108|Codes]).

get_location_var_from_ast_node_aux(Type,ValueCode) :-
    Type =.. [Functor|_] ,
    atom_codes(Functor,ValueCode).

% Enumerate used components for identifier nodes information so we definitely
% know which input/output locations belong together if a component is used multiple times.
% Here, the library components are represented as tuples of the component name and its internal name with id as postfix.
set_component_id(Component,ComponentWithId) :-
    used_library(Components) ,
    retract(used_library(_)) ,
    length(Components,OpAmount) ,
    ComponentId is OpAmount + 1 ,
    atom_number_concat(Component,ComponentId,ComponentWithId) ,
    assert(used_library([(Component,ComponentWithId)|Components])).

% Evaluate M := |I| + N, i.e., the amount of variables considered in the examples plus the amount of library components.
% We do not consider machine constants as explicit library components since their values are defined by the machine properties, i.e., the amount
% of machine constants is not considered within N. Therefore, we have to add x lines to the program where x is the amount of machine constants.
% Moreover, we may consider operation parameters if present which we handle equivalent to machine constants.
% Generally, the value of M determines the amount of lines in the program
get_line_amount(Operation,Components,InputVars,M) :-
    additional_set_constants(AdditionalSetConstants) ,
    length(AdditionalSetConstants,AC) ,
    length(InputVars,I) ,
    length(Components,N) ,
    b_get_machine_constants(MachineConstants) ,
    % if the operation exists consider the amount of operation parameters for that we consider synthesis constants
    (ground(Operation) , b_get_machine_operation(Operation,_,OperationParameters,_)
    ->  length(OperationParameters,OperationParameterAmount)
    ;   OperationParameterAmount = 0) ,
    length(MachineConstants,AmountOfMachineConstants) ,
    M is I + N + AmountOfMachineConstants + AC + OperationParameterAmount.

% Set the location variables of the program parameters and output given by I/O-Examples.
% Each input parameters are defined in the first lines of the program
set_input_location_vars(InputValueNodes,InputLocations) :-
    set_input_location_vars(InputValueNodes,0,[],InputLocationsList) ,
    conjunct_predicates(InputLocationsList,InputLocations).
set_input_location_vars([],_,Acc,Acc).
set_input_location_vars([InputValueNode|R],C,Acc,InputLocations) :-
    get_location_var_from_ast_node(InputValueNode,InputLocationNode) ,
    InputLocation = b(equal(InputLocationNode,b(integer(C),integer,[])),pred,[]) ,
    C1 is C + 1 ,
    set_input_location_vars(R,C1,[InputLocation|Acc],InputLocations).
% We dispose output values in descending order starting at the last line of the program
set_output_location_vars(OutputValueNodes,M,OutputLocations,OutputLocationVars) :-
    M1 is M - 1 ,
    set_output_location_vars(OutputValueNodes,M1,[],OutputLocationsList,[],OutputLocationVars) ,
    conjunct_predicates(OutputLocationsList,OutputLocations).
set_output_location_vars([],_,Acc,Acc,VarAcc,VarAcc).
set_output_location_vars([OutputValueNode|R],C,Acc,OutputLocations,VarAcc,OutputLocationVars) :-
    get_location_var_from_ast_node(OutputValueNode,OutputLocationNode) ,
    OutputLocation = b(equal(OutputLocationNode,b(integer(C),integer,[])),pred,[]) ,
    C1 is C - 1 ,
    add_texpr_infos(OutputLocationNode,[synthesis(program_line(C))],NewOutputLocationNode) ,
    set_output_location_vars(R,C1,[OutputLocation|Acc],OutputLocations,[NewOutputLocationNode|VarAcc],OutputLocationVars).

% Enumerate input nodes in the info to use within the connectivity constraint, the corresponding machine variable names have already been added to each node's info.
% We also enumerate the examples in the order they are given to get unique identifier names for program input and output parameter later on and especially unique names that
% can be reconstructed using the node's info. Using unique_typed_id/3 we get like any unique identifier without any pattern to reconstruct later on.
enumerate_inputs_and_outputs(ExampleCount,InputValues,StateType,InputNodes) :-
    enumerate_inputs_and_outputs(ExampleCount,InputValues,StateType,0,InputNodes).
enumerate_inputs_and_outputs(_,[],_,_,[]).
enumerate_inputs_and_outputs(ExampleCount,[b(Node,Type,Info)|R],input,C,[b(Node,Type,[synthesis(input(C)),synthesis(example(ExampleCount)),synthesis(type,Type)|Info])|T]) :-
    C1 is C + 1 ,
    enumerate_inputs_and_outputs(ExampleCount,R,input,C1,T).
enumerate_inputs_and_outputs(ExampleCount,[b(Node,Type,Info)|R],output,C,[b(Node,Type,[synthesis(output(C)),synthesis(example(ExampleCount)),synthesis(type,Type)|Info])|T]) :-
    C1 is C + 1 ,
    enumerate_inputs_and_outputs(ExampleCount,R,output,C1,T).

atom_number_concat(Atom,Number,Concatenation) :-
    number_codes(Number,NumberCodes) ,
    atom_codes(Atom,AtomCodes) ,
    append(AtomCodes,NumberCodes,ConcatCodes) ,
    atom_codes(Concatenation,ConcatCodes).

% Get an output ĺocation variable for a specific location value.
get_output_location_for_location_value(LocationVars,SolutionBindings,LocationValue,OutputLocationVar) :-
    get_binding_for_name_or_value(SolutionBindings,OutputLocationName,LocationValue) ,
    atom_concat(lo,_,OutputLocationName) , ! ,
    member(OutputLocationVar,LocationVars) , OutputLocationVar = b(identifier(OutputLocationName),integer,_).
% Return a node for the location value in case there is no explicit operators' output mapped, thus, it is an input parameter.
% We know the first input parameter is in the first line of the program etc. which will not change, therefore we
% do not need the explicit location variable name but only the location value.
get_output_location_for_location_value(_,_,int(Value),b(integer(Value),integer,[])).

% add input location variables since the program asts contain explicit values
accumulate_used_location_var_aux(Operator,LocationVars,Acc,NewAcc) :-
    findall(LocationVar,(member(LocationVar,LocationVars) , LocationVar = b(_,_,Info) ,
                         member(synthesis(Operator,LocationType),Info) ,
                         (LocationType = left_input ; LocationType = right_input)),InputLocationVars) ,
    append(Acc,InputLocationVars,NewAcc).

:- spec_pre(get_used_location_vars_and_op_names/6,[ast,list(ast),list(ast),list(atom),var,var]).
:- spec_invariant(get_used_location_vars_and_op_names/6,[ast,list(ast),list(ast),list(atom),list(ast),list(atom)]).
:- spec_post(get_used_location_vars_and_op_names/6,[ast,list(ast),list(ast),list(atom),var,var],
                                                   [ast,list(ast),list(ast),list(atom),list(ast),list(atom)]).
get_used_location_vars_and_op_names(b(Node,_,Info),LocationVars,LocationAcc,OpAcc,UsedLocationVars,UsedOperatorNames) :-
    Node =.. [if_then_else,Condition,TrueOut,FalseOut] ,
    accumulate_used_location_var(Info,LocationVars,LocationAcc,OpAcc,LocationAcc1,OpAcc1) ,
    get_used_location_vars_and_op_names(Condition,LocationVars,LocationAcc1,OpAcc1,LocationAcc2,OpAcc2) ,
    get_used_location_vars_and_op_names(TrueOut,LocationVars,LocationAcc2,OpAcc2,LocationAcc3,OpAcc3) ,
    get_used_location_vars_and_op_names(FalseOut,LocationVars,LocationAcc3,OpAcc3,UsedLocationVars,UsedOperatorNames) , !.
get_used_location_vars_and_op_names(b(Node,_,Info),LocationVars,LocationAcc,OpAcc,UsedLocationVars,UsedOperatorNames) :-
    Node =.. [_,LeftAST,RightAST] ,
    is_ast(LeftAST) , is_ast(RightAST) ,
    accumulate_used_location_var(Info,LocationVars,LocationAcc,OpAcc,LocationAcc1,OpAcc1) ,
    get_used_location_vars_and_op_names(LeftAST,LocationVars,LocationAcc1,OpAcc1,LocationAcc2,OpAcc2) ,
    get_used_location_vars_and_op_names(RightAST,LocationVars,LocationAcc2,OpAcc2,UsedLocationVars,UsedOperatorNames) , !.
    %findall(ParameterLocation,(member(ParameterLocation,LocationVars) ,
    %                           ParameterLocation = b(_,_,Info) , member(synthesis(input(_)),Info)),ParameterLocations) ,
    %append(TempUsedLocationVars,ParameterLocations,UsedLocationVars).
get_used_location_vars_and_op_names(b(Node,_,Info),LocationVars,LocationAcc,OpAcc,UsedLocationVars,UsedOperatorNames) :-
    Node =.. [_,AST] ,
    accumulate_used_location_var(Info,LocationVars,LocationAcc,OpAcc,NewLocationAcc,NewOpAcc) ,
    (is_ast(AST)
    ->  get_used_location_vars_and_op_names(AST,LocationVars,NewLocationAcc,NewOpAcc,UsedLocationVars,UsedOperatorNames)
    ;   UsedLocationVars = NewLocationAcc , UsedOperatorNames = NewOpAcc).
get_used_location_vars_and_op_names(_,_,_,_,Acc,Acc).

accumulate_used_location_var(Info,LocationVars,LocationAcc,OpAcc,NewLocationAcc,[Op|OpAcc]) :-
    member(synthesis(Op,output),Info) ,
    \+member(Op,OpAcc) ,
    member(b(identifier(_),integer,IdInfo),LocationVars) ,
    member(synthesis(Op,output),IdInfo) ,
    accumulate_used_location_var_aux(Op,LocationVars,LocationAcc,NewLocationAcc).
accumulate_used_location_var(_,_,LocationAcc,OpAcc,LocationAcc,OpAcc).

:- spec_pre(get_input_location_nodes_for_operator_name/4,[atom,list(ast),list(ast),var]).
:- spec_invariant(get_input_location_nodes_for_operator_name/4,[atom,list(ast),list(ast),list(ast)]).
:- spec_post(get_input_location_nodes_for_operator_name/4,[atom,list(ast),list(ast),var],
                                                          [atom,list(ast),list(ast),list(ast)]).
get_input_location_nodes_for_operator_name(_,[],Acc,Acc).
get_input_location_nodes_for_operator_name(OperatorName,[LocationVar|T],Acc,InputLocations) :-
    LocationVar = b(identifier(_),integer,Info) ,
    member(synthesis(OperatorName,LocationType),Info) ,
    LocationType \= output , ! ,
    get_input_location_nodes_for_operator_name(OperatorName,T,[LocationVar|Acc],InputLocations).
get_input_location_nodes_for_operator_name(OperatorName,[_|T],Acc,InputLocations) :-
    get_input_location_nodes_for_operator_name(OperatorName,T,Acc,InputLocations).

%get_output_location_node_for_operator_name(OperatorName,[OutputLocationNode|_],OutputLocationNode) :-
%    OutputLocationNode = b(identifier(_),integer,Info) ,
%    member(synthesis(OperatorName,output),Info) , !.
%get_output_location_node_for_operator_name(OperatorName,[_|T],OutputLocationNode) :-
%    get_output_location_node_for_operator_name(OperatorName,T,OutputLocationNode).

is_constant(b(identifier(Constant),_,_)) :-
    atom_concat(constant,_,Constant).
is_constant(b(_,_,Info)) :-
    member(synthesis(OpName,output),Info) ,
    atom_concat(constant,_,OpName).

exclude_solution_for_further_search(UsedOperatorNames,LocationVars,SolutionBindings,Exclusion) :-
    maplist(exclude_output_to_input_mapping_for_operator(LocationVars,SolutionBindings),UsedOperatorNames,TExclusionList) ,
    remove_dups(TExclusionList,ExclusionList) ,
    disjunct_predicates(ExclusionList,TempExclusion) ,
    exclude_solution_for_further_search_aux(TempExclusion,Exclusion).

exclude_solution_for_further_search_aux(b(falsity,_,_),b(truth,pred,[])) :- !.
exclude_solution_for_further_search_aux(Exclusion,Exclusion).

% constants: Inequality with the input location it is mapped to or with the explicit location value
%            in case the constant is the program output.
exclude_output_to_input_mapping_for_operator(LocationVars,SolutionBindings,OperatorName,Exclusion) :-
    atom_concat(constant,_,OperatorName) , ! ,
    get_location_var_node_by_info_term(LocationVars,synthesis(OperatorName,output),ConstantLocation) ,
    ConstantLocation = b(identifier(ConstantLocationName),integer,_) ,
    get_binding_for_name_or_value(SolutionBindings,ConstantLocationName,LocationValue) ,
    get_binding_for_name_or_value(SolutionBindings,MappedInputLocationName,LocationValue) ,
    MappedInputLocationName \= ConstantLocationName ,
    % exclude connection for all constants to prevent swapping with other constants of the same type
    findall(E,(member(ConstantNode,LocationVars) , is_constant(ConstantNode) , ConstantNode = b(identifier(CName),_,_) ,
               get_binding_for_name_or_value(SolutionBindings,CName,LValue) ,
               ConstantLocationName \= CName ,
               exclude_output_to_input_mapping_for_constant_aux(LocationVars,LValue,ConstantLocation,CName,E)),TExclusions) ,
    flatten(TExclusions,Exclusions) ,
    exclude_output_to_input_mapping_for_constant_aux(LocationVars,LocationValue,ConstantLocation,MappedInputLocationName,TExclusion) ,
    conjunct_predicates([TExclusion|Exclusions],Exclusion).
exclude_output_to_input_mapping_for_operator(LocationVars,SolutionBindings,OperatorName,Exclusion) :-
    get_location_var_node_by_info_term(LocationVars,synthesis(OperatorName,output),ConstantLocation) ,
    ConstantLocation = b(identifier(ConstantLocationName),integer,_) ,
    get_binding_for_name_or_value(SolutionBindings,ConstantLocationName,LocationValue) ,
    get_binding_for_name_or_value(SolutionBindings,MappedInputLocationName,LocationValue) ,
    MappedInputLocationName \= ConstantLocationName ,
    exclude_output_to_input_mapping_for_constant_aux(LocationVars,LocationValue,ConstantLocation,MappedInputLocationName,Exclusion).
% Exclude the mapping of a given operators input locations to its corresponding output locations for the given solution.
exclude_output_to_input_mapping_for_operator(LocationVars,SolutionBindings,OperatorName,Exclusion) :-
    get_input_location_nodes_for_operator_name(OperatorName,LocationVars,[],InputLocations) , InputLocations \= [] ,
    maplist(exclude_specific_input_output_mapping(LocationVars,SolutionBindings),InputLocations,InputOutputExclusions) ,
    disjunct_predicates(InputOutputExclusions,Exclusion).

exclude_output_to_input_mapping_for_constant_aux(LocationVars,_,ConstantLocation,MappedInputLocationName,Exclusion) :-
    member(MappedInputLocationVar,LocationVars) , MappedInputLocationVar = b(identifier(MappedInputLocationName),_,_) ,
    create_texpr(not_equal(ConstantLocation,MappedInputLocationVar),pred,[],Exclusion).
%exclude_output_to_input_mapping_for_constant_aux(_,LocationValue,ConstantLocation,_,Exclusion) :-
%    create_texpr(not_equal(ConstantLocation,b(value(LocationValue),integer,[])),pred,[],Exclusion).

exclude_specific_input_output_mapping(LocationVars,SolutionBindings,b(identifier(Id),integer,Info),Exclusion) :-
    get_binding_for_name_or_value(SolutionBindings,Id,LocationValue) ,
    get_output_location_for_location_value(LocationVars,SolutionBindings,LocationValue,OutputLocationVar) ,
    create_texpr(not_equal(b(identifier(Id),integer,Info),OutputLocationVar),pred,[],Exclusion).
exclude_specific_input_output_mapping(_,_,_,b(falsity,pred,[])).

% Using a component several times, e.g. with internal names add1,add2, we don't want to
% swap both components when searching for a further solution.
get_prevent_component_swap_constraint(UsedLVars,SolutionBindings,LocationVars,Constraint) :-
    get_input_location_vars_for_outputs(UsedLVars,LocationVars,OutputInputsTuples) ,
    maplist(exclude_swapping_for_component(SolutionBindings,LocationVars),OutputInputsTuples,TExclusions) ,
    remove_dups(TExclusions,Exclusions) ,
    conjunct_predicates(Exclusions,Constraint).

% TODO: Maybe consider constants too?
get_input_location_vars_for_outputs(UsedLVars,LocationVars,OutputInputsTuples) :-
    % create tuples of component and an empty list to use as an accumulator
    findall((Component,[]),(member(OutLVar,UsedLVars) , get_texpr_info(OutLVar,Info) ,
                            member(synthesis(Component,output),Info)),Acc) ,
    get_input_location_vars_for_outputs_aux(LocationVars,Acc,OutputInputsTuples).
get_input_location_vars_for_outputs_aux([],Acc,Acc).
get_input_location_vars_for_outputs_aux([LocationVar|T],Acc,OutputInputsTuples) :-
    get_texpr_info(LocationVar,Info) ,
    (member(synthesis(Component,left_input),Info) ; member(synthesis(Component,right_input),Info) ; member(synthesis(Component,input),Info)) ,
    member((Component,Inputs),Acc) ,
    delete(Acc,(Component,Inputs),TempAcc) ,
    remove_dups([LocationVar|Inputs],InputLVars),
    NewTuple = (Component,InputLVars) ,
    \+member(NewTuple,Acc) , % no duplicates
    get_input_location_vars_for_outputs_aux(T,[NewTuple|TempAcc],OutputInputsTuples).
get_input_location_vars_for_outputs_aux([_|T],Acc,OutputInputsTuples) :-
    get_input_location_vars_for_outputs_aux(T,Acc,OutputInputsTuples).

% Given a tuple of an output location var and the corresponding component's inputs.
% We search for the output locations that are mapped to the input locations.
% Then, we search all equivalent component inputs and exclude mapping the output locations to
% the equivalent input location.
% Thus, we exclude swapping two equivalent operators when searching for a further solution for a given
% synthesized program.
exclude_swapping_for_component(SolutionBindings,LocationVars,(OutLVar,InLVars),ExclusionConstraint) :-
    maplist(get_lvar_name_and_input_type,InLVars,InLVarNames,InLVarTypes) ,
    % get the location values the inputs of OutLVar are mapped to
    maplist(get_binding_for_name_or_value(SolutionBindings),InLVarNames,InLVarValues) ,
    % create tuples of (lvar,lvalue)
    zip(InLVarNames,InLVarValues,VarValueTuples) ,
    zip(InLVarNames,InLVarTypes,VarTypeTuples) ,
    % then, get the output locations those inputs are mapped to
    maplist(get_corresponding_mapped_location(SolutionBindings,LocationVars),VarValueTuples,CorrespondingLocations) ,
    % and create tuples of output locations and the type of the input they are mapped to, i.e.,
    % either left_input or right_input
    findall((OutLVarName,InputType),(member((InLVar,OutLVarName),CorrespondingLocations) ,
                                     member((InLVar,InputType),VarTypeTuples)),OutLVarInTypeTuples) ,
    exclude_swapping_for_component_aux(LocationVars,(OutLVar,OutLVarInTypeTuples),ExclusionConstraint).
exclude_swapping_for_component_aux(LocationVars,(OutLVarName,OutLVarInTypeTuples),ExclusionConstraint) :-
    get_equivalent_input_location_nodes(OutLVarName,LocationVars,EquivalentInputLocations) ,
    % find pairs of output and input location that correspond (are mapped) to the same input type, i.e.
    % either left_input, right_input or input
    findall(Exclusion,(member((CorOutLVarName,InputType),OutLVarInTypeTuples) ,
                       get_location_var_node_by_name(CorOutLVarName,LocationVars,OutLVarNode) ,
                       member((InLVarNode,InputType),EquivalentInputLocations) ,
                       create_texpr(not_equal(OutLVarNode,InLVarNode),pred,[],Exclusion)),Exclusions) ,
    conjunct_predicates(Exclusions,ExclusionConstraint).

% Get all input locations that are equivalent to the component that LocationVarName belongs to.
% Don't consider exactly the same location var.
get_equivalent_input_location_nodes(LocationVarName,LocationVars,InputLocations) :-
    delete_numbers_from_atom(LocationVarName,RawName) ,
    get_equivalent_input_location_nodes_aux(LocationVarName,RawName,LocationVars,[],InputLocations).
get_equivalent_input_location_nodes_aux(_,_,[],Acc,Acc).
get_equivalent_input_location_nodes_aux(LocationVarName,RawName,[LocationVar|T],Acc,InputLocations) :-
    get_texpr_info(LocationVar,Info) ,
    member(synthesis(LVarName,InputType),Info) ,
    (InputType = left_input ; InputType = right_input ; InputType = input) ,
    delete_numbers_from_atom(LVarName,RawLVarName) ,
    RawLVarName = RawName ,
    % not exactly the same location
    LVarName \= LocationVarName ,
    \+member((LocationVar,InputType),Acc) , % no duplicates
    get_equivalent_input_location_nodes_aux(LocationVarName,RawName,T,[(LocationVar,InputType)|Acc],InputLocations).
get_equivalent_input_location_nodes_aux(LocationVarName,RawName,[_|T],Acc,InputLocations) :-
    get_equivalent_input_location_nodes_aux(LocationVarName,RawName,T,Acc,InputLocations).

% Given a tuple of location var name and mapped value, we search for the other location var that
% is mapped to this location value (line of the program).
get_corresponding_mapped_location(SolutionBindings,_,(LVarName,LVarValue),(LVarName,LVarName2)) :-
    get_binding_for_name_or_value(SolutionBindings,LVarName2,LVarValue) ,
    delete_numbers_from_atom(LVarName2,RawName) ,
    RawName \= constant ,
    LVarName \= LVarName2.

get_lvar_name_and_input_type(b(identifier(LVarName),integer,Info),LVarName,InputType) :-
    get_input_type_from_info(Info,InputType).

get_input_type_from_info(Info,left_input) :-
    member(synthesis(_,left_input),Info).
get_input_type_from_info(Info,right_input) :-
    member(synthesis(_,right_input),Info).
get_input_type_from_info(Info,input) :-
    member(synthesis(_,input),Info).
