% (c) 2009-2020 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

%:- use_module('../extensions/profiler/profiler.pl').

:- module(test_runner, [
                        test_repl/0,
                        run_silently/0,
                        run_last_test/0, last/0,
                        run_tests_by_id/1, ids/1,
                        run_tests_all/0, run_tests_all_except/1,
                        run_tests_by_category/1, run_tests_by_category/2, category/1,
                        run_tests_by_first_category/1,
                        makefile_by_category/2, copy/1,
                        generate_makefile/2,
                        generate_makefile_all_tests/0,
                        generate_makefile_for_coverage/0,
                        full_flush/0,
                        jenkins_sanity_check/0,
                        sanity_check/1,
                        v/0, vv/0,
                        cache/0, % enable caching
                        all_categories/1,
                        all_ids/1,
                        ids_in_category/2,
                        set_prob_examples_location/1,
                        halt_tests/0 % use to call halt with proper exit code; however, normally test predicates will call halt1 directly, mainly useful in junit mode
                       ]).

% comment in to ensure that plspec loaded first:
% :- use_module('../extensions/plspec/plspec/plspec_core').

:- use_module(prob_cli). % we call go_cli
% disable clauses are not together warning, because
% cli_testcase/4 and cli_testcase/5 are sorted by id
:- set_prolog_flag(discontiguous_warnings,off).

%:- prolog_flag(compiling,_,debugcode).
%:- prolog_flag(source_info,_,on).
%:- prolog_flag(profiling,_,on).

/*
:- use_module('../extensions/profiler/profiler.pl').    % (adds the profiler himself)
:- use_module('../extensions/profiler/profiler_te.pl'). % (adds the term expander)
%:- use_module('../extensions/profiler/profiler_gui.pl'). % (adds the term expander)
:- enable_profiling_all(b_expression_sharing).
:- enable_profiling(bsyntaxtree:transform_bexpr_with_acc/5).
:- enable_profiling(bsyntaxtree:transform_bexpr/3).
:- enable_profiling(b_ast_cleanup:recompute_used_ids_info/2).
:- enable_profiling(bsyntaxtree:sub_expression_contains_wd_condition/1).

for starting from Spider:
set_prob_examples_location('../../prob_examples'), run_silently, test_repl.
or use test_runner_cov.pl
*/


:- use_module(library(system)).
:- use_module(library(lists)).
:- use_module(library(sets), [intersect/2, list_to_set/2]).
:- use_module(library(file_systems)).

:- use_module(junit_tests,[print_junit/2,set_junit_dir/1,create_junit_result/6]).
:- use_module(testcases).
:- use_module(system_call).
:- use_module(tools,[safe_number_codes/2,get_options/5]).
:- use_module(tools_commands,[edit_file/1]).
:- use_module(error_manager,[get_total_number_of_errors/1, get_all_errors/1]).
:- use_module(debug,[formatsilent/3, debug_mode/1]).
:- use_module(performance_messages,[toggle_perfmessages/0]).

:- use_module(module_information).
:- module_info(group,testing).
:- module_info(description,'This module runs the tests stored in testcases.pl.').


:- meta_predicate generate_makefile(-,1).
generate_makefile(Filename,WriteMakeTargetPredicate) :-
    tell(Filename),
    all_first_categories(AllTestCategories),
    write('testtarget: '),
    write_subtargets(AllTestCategories), nl, nl,
    maplist(WriteMakeTargetPredicate,AllTestCategories),
    told.

write_subtargets([G]) :- write(G).
write_subtargets([G|Gs]) :-
    write(G), write(' '),
    write_subtargets(Gs).

generate_makefile_all_tests :-
     generate_makefile('Makefile_All_Tests', write_makefile_target).
write_makefile_target(G) :-
     % run tests where category G is the first category to avoid running tests multiple times
     % also avoids concurrency issues when test run multiple times in parallel with diff checking of output
     format('~w:\n\tsicstus -l src/test_runner.pl --goal "run_silently, run_tests_by_first_category(\'~w\'), halt ; halt."',[G,G]),nl,nl.


generate_makefile_for_coverage :-
    generate_makefile('Makefile_Coverage_Groups', write_coverage_target).
write_coverage_target(G) :-
   format('~w:\n\tsicstus -l src/prob_cov_runner.pl --goal "coverage_group_data(\'~w\'), halt ; halt."',[G,G]),nl,nl.

all_first_categories(X) :- % only categories that appear as first category
    findall(Group1,
            (cli_testcase(_Id,[Group1|_],_Conf,_Args,_Desc),
             Group1 \= private_source_not_available), % these tests are not checked in; cannot be run on Jenkins, ...
            ListOfGroups),
    list_to_set(ListOfGroups,X).
all_categories(X) :-
    findall(Groups,
            (cli_testcase(_Id,Groups,_Conf,_Args,_Desc),
             Groups \= [private_source_not_available|_]), % these tests are not checked in; cannot be run on Jenkins, ...
            List),
    append(List,ListOfGroups),
    list_to_set(ListOfGroups,X).

all_ids(X) :-
    findall(Id,
            cli_testcase(Id,_Groups,_Conf,_Args,_Desc),
            List),
    list_to_set(List,X).

ids_in_category(Category,IDs) :-
    findall(Id,
            (cli_testcase(Id,Groups,_Conf,_Args,_Desc),member(Category,Groups)),
            List),
    list_to_set(List,IDs).

prob_junit(PROB_JUNIT) :-
    environ('PROB_JUNIT_DIR', Var), !,
    PROB_JUNIT = ['-junit', Var].
prob_junit([]).

dontstop(Var) :-
    environ('DONTSTOP', Var2), !, Var=Var2.
dontstop(false).

halt1_allowed :- \+ repl_mode,
   prob_junit([]), dontstop(false).

tests_multiply_timeout(Var) :-
    environ('TESTS_MULTIPLY_TIMEOUT', Var2), !, tools:arg_is_number(Var2,Var).
tests_multiply_timeout(Var) :- multiply_timeout(Var).

:- dynamic multiply_timeout/1.

% call updir from within Spider, when you are in the src subdirectory
set_cur_dir :- (directory_exists('./src') -> true /* we seem to be in the right directory */
          ; directory_exists('./cia') -> current_directory(_Old,'./..')
          ; print('Please start tests from ProB directory or ProB/src directory'),nl),
   print('Current directory: '),
   current_directory(CD), print(CD),nl.

reset_test_runner(SessionInfo) :-
    reset_nr_of_tests, set_cur_dir,
    retractall(test_skipped(_)), retractall(test_failed(_)), retractall(last_test_failed(_)),
    retractall(test_diff_failed(_,_,_)),
    statistics(walltime,[WStart,_]),
    bb_put(reset_test_runner_wtime,WStart),
    bb_put(reset_test_runner_info,SessionInfo).


ids(Nr) :- show_command(Nr).
show_command(Nr) :- cli_testcase(Nr,_,_,L,Desc),
  print('# '), print(Desc),nl, l_pr([probcli|L]).
l_pr([]) :- nl.
l_pr([H|T]) :- print(' '),print(H), l_pr(T).

:- volatile sanity_check_failed/0.
:- dynamic sanity_check_failed/0.
jenkins_sanity_check :-
   sanity_check(true),
   (sanity_check_failed -> halt1 ; halt(0)).
sanity_check(_Jenkins) :-
   cli_testcase(Id,Categories,_,Args,_Comm1),
   \+ member('-strict',Args),
   format_warning('*** test does not use -strict option: ',[]), print_test(Id,Categories),nl,
   assert(sanity_check_failed), fail.
sanity_check(Jenkins) :-
   Jenkins == false,
   cli_testcase_diff_check_output(Id,_F1,_F2),
   cli_testcase(Id,Categories,_I1,_A1,_Comm1),
   (Categories = [_]
    -> true
    ; format_warning('*** multiple categories for test with output file; can generate race conditions when run in parallel: ',[]),
      print_test(Id,Categories),
      nl  /* can generate race conditions when run in parallel */),
   assert(sanity_check_failed), fail.
sanity_check(_Jenkins) :-
   cli_testcase_diff_check_output(Id,_F1,_F2),
   \+ (cli_testcase(Id,_Categories,_I1,_A1,_Comm1)),
   format_warning('*** no cli_testcase, but diff output defined for id: ~w~n',[Id]),
   assert(sanity_check_failed), fail.
sanity_check(_Jenkins) :-
   cli_testcase(Id,C1,I1,A1,Comm1),
   cli_testcase(Id,C2,I2,A2,Comm2),
   [C1,I1,A1,Comm1] \= [C2,I2,A2,Comm2],
   format_warning('*** multiple cli_testcases for id: ~w~n',[Id]),
   assert(sanity_check_failed), fail.
sanity_check(_Jenkins).


format_warning(F,Args) :- format_colour(red,F,Args).
format_colour(Col,F,Args) :-
   start_terminal_colour(Col,user_output),
   format(user_output,F,Args),
   reset_terminal_colour(user_output).

print_test(Id,Comment) :- print(Id),print(' : '), print(Comment).

% run the latest test (with highest id)
last :- run_last_test.
run_last_test :- get_last_test_id(Id),
  run_tests_by_id(Id,_,no_skipping).

get_last_test_id(Id) :- set_last_test_id, last_test_id(Id).

set_last_test_id :-
   retractall(last_test_id(_)), assert(last_test_id(-1)),
   cli_testcase(Id,C1,I1,A1,Comm1),
   ((cli_testcase(Id,C2,I2,A2,Comm2),(C1,I1,A1,Comm1) \= (C2,I2,A2,Comm2))
     -> format_warning('~n*** multiple entries for test id: ~w~n',[Id])
     ; true),
   update_last(Id),
   fail.
set_last_test_id.

:- volatile last_test_id/1.
:- dynamic last_test_id/1.
update_last(Id) :- number(Id),!, % debug_print(9,Id), debug_print(9,' '),
   retract(last_test_id(L)),
   (Id>L -> Max = Id ; Max = L),
   assert(last_test_id(Max)).
update_last(Id) :- format_warning('~n*** test id not a number: ~w~n',[Id]).

repeat_id(ID,Nr) :- repeat_id_aux(ID,0,Nr,success,10000000,0,0).
repeat_id_aux(ID,Nr,TotNr,Status1,Min,Max,Tot) :- Nr<TotNr, !, N1 is Nr+1,
  statistics(walltime,[Start,_]),
  run_tests_by_id(ID,Status,no_skipping),
  (Status=success -> Status2=Status1, Col=blue ; Status2=Status, Col=red),
  statistics(walltime,[Stop,_]),
  WT is Stop - Start,
  NewMin is min(Min,WT), NewMax is max(Max,WT), NewTot is Tot+WT,
  Average is NewTot / N1,
  format_colour([Col],'Test ~w repeated ~w/~w times~n Walltime Minumum = ~w ms, Maximum = ~w ms, Average = ~w ms~n Status=~w~n',[ID,N1,TotNr,NewMin,NewMax,Average,Status]),
  repeat_id_aux(ID,N1,TotNr,Status2,NewMin,NewMax,NewTot).
repeat_id_aux(_,_,_TotNr,_Status,_Min,_Max,_).

% run a test with a specific id
run_id(X) :- run_tests_by_id(X,_,no_skipping).
run_tests_by_id(X) :- run_tests_by_id(X,_,allow_skipping).
run_tests_by_id(X,Status,AllowSkipping) :- reset_test_runner(id(X)),
  run_tests_by_id_aux(X,AllowSkipping),
  check_failed(Status).
run_tests_by_id_aux([],_).
run_tests_by_id_aux(From-To,AllowSkipping) :- !,
    (From>To -> true
     ; run_tests_by_id_aux(From,AllowSkipping), F1 is From+1,
       run_tests_by_id_aux(F1-To,AllowSkipping)).
run_tests_by_id_aux([Id | Ids],AllowSkipping) :- !,
    run_tests_by_id_aux(Id,AllowSkipping),
    run_tests_by_id_aux(Ids,AllowSkipping).
run_tests_by_id_aux(Id,AllowSkipping) :-
    cli_testcase(Id,_TestCategories,TestInfos,Arguments,Comment), !,
    run_list([testcase(Id,TestInfos,Arguments,Comment)],AllowSkipping).
run_tests_by_id_aux(Id,_AllowSkipping) :-
    \+(cli_testcase(Id,_Category,_Infos,_Arguments,_Comment)), !,
    format_warning('*** No testcase with ID ~w found~n', [Id]).

:- use_module(library(random),[random_select/3]).
run_random_tests(Nr) :-
   findall(testcase(Id,TestInfos,Arguments,Comment),
          cli_testcase(Id,_TestCategories,TestInfos,Arguments,Comment),AllTests),
   select_random_tests(Nr,AllTests,Tests),
   reset_test_runner(random(Nr)),
   run_list(Tests), check_failed(_).
select_random_tests(N,_,[]) :- N<1,!.
select_random_tests(_,[],[]) :- !.
select_random_tests(N,All,[X|Tests]) :-
   random_select(X,All,Rest),
   N1 is N-1,
   select_random_tests(N1,Rest,Tests).


run_tests_all :- run_tests_all_except([]).

:- use_module(library(ordsets)).
run_tests_all_except(Categories) :- reset_test_runner(all_except(Categories)),
    sort(Categories,SC),
    cli_testcase(Id,TestCategories,Infos,Arguments,Comment), sort(TestCategories,TC),
    \+ ord_intersect(TC,SC),
    run_list([testcase(Id,Infos,Arguments,Comment)]),
    fail.
run_tests_all_except(_) :- check_failed(_).


% run all tests with a specific category
run_silently :- add_additional_arguments(['-silent']), assert(silent_running).
:- dynamic silent_running/0.

category(X) :- run_tests_by_category(X).
run_tests_by_category(X) :- run_tests_by_category(X,all).
run_tests_by_first_category(X) :- % used as entry in Jenkins when running all tests in parallel by category
  run_silently,
  run_tests_by_category(X,first).
run_tests_by_category(X,FirstOnly) :-
    reset_test_runner(category(X,FirstOnly)),
    (is_list(X) -> run_tests_by_category_aux(X,FirstOnly) ; run_tests_by_category_aux([X],FirstOnly)),
    check_failed(_).
run_tests_by_category_aux(Categories,FirstOnly) :-
    get_tests_for_categories(Categories,FirstOnly,List),
    ( List=[]   -> format_warning('*** No testcase with categories ~w found~n', [Categories])
    ; otherwise -> run_list(List)).

get_tests_for_categories(Categories,List) :-
    get_tests_for_categories(Categories,all,List).
get_tests_for_categories(Categories,FirstOnly,List) :-
    findall(testcase(Id, Infos, Arguments, Comment),
            (cli_testcase(Id, TestCategories, Infos, Arguments, Comment),
                (FirstOnly=first -> TestCategories=[C1|_], member(C1,Categories)
                  ; intersect(Categories, TestCategories)) ),
            List).

run_tests_using_command(Command) :- reset_test_runner(cmd(Command)),
    findall(testcase(Id, Infos, Arguments, Comment),
            (cli_testcase(Id, _TestCategories, Infos, Arguments, Comment),
                member(Command, Arguments)),
            List),
    ( List=[]   -> format_warning('*** No testcase using command ~w found~n', [Command])
    ; otherwise -> run_list(List), check_failed(_)).

run_tests_using_preference(Pref,Val) :- reset_test_runner(using_pref(Pref,Val)),
    findall(testcase(Id, Infos, Arguments, Comment),
            (cli_testcase(Id, _TestCategories, Infos, Arguments, Comment),
                append(_,['-p',Pref,Val|_],Arguments) ),
            List),
    ( List=[]   -> format_warning('*** No testcase using preference ~w with value ~w found~n', [Pref,Val])
    ; otherwise -> run_list(List), check_failed(_)).

% Generate file list (for copying) of a category :
show_files(Cat) :-
   cli_testcase(Id, TestCategories, _Infos, Arguments, _Comment),
   member(Cat, TestCategories),  % print(Arguments),nl,
   ( file_in_arguments(File,Arguments)
     ;
     get_testcase_diff_check_output(Id,_GeneratedFile,File)),
   format('~w ',[File]),
   fail.
show_files(_) :- nl.

file_in_arguments(F,['-p',_,_|T]) :- !, file_in_arguments(F,T).
file_in_arguments(F,[C,_|T]) :- binary_command(C),!, file_in_arguments(F,T).
file_in_arguments(F,[H|T]) :- is_file(H),H=F ; file_in_arguments(F,T).
binary_command(eval). binary_command(evalt). binary_command(evalf).
binary_command('-mc').
is_file(F) :- atom(F), atom_codes(F,Codes), member(47,Codes),!.


% Generate a Makefile for all listed Categories
% example: makefile_by_category('AlstomMakefile',[animate,history,sptxt,rel_fnc])
% the Makefile can be run to perform the listed tests on a compiled version of probcli
makefile_by_category(File, Categories) :-
    (is_list(Categories) -> Cats = Categories ; Cats = [Categories]),
    my_open(File, Stream),
    format(Stream, 'PROBCLI=probcli~n',[]),
    format(Stream, 'test_all:', []),
    makefile_write_categories(Stream, Cats),
    makefile_by_category_aux(Stream, Cats),
    my_close(Stream).
makefile_by_category_aux(_Stream, []).
makefile_by_category_aux(Stream, [Cat | Cats]) :-
    makefile_by_category_single(Stream, Cat),
    makefile_by_category_aux(Stream, Cats).
makefile_by_category_single(Stream, Cat) :-
    findall(testcase(Id, Infos, Arguments, Comment),
            (cli_testcase(Id, TestCategories, Infos, Arguments, Comment),
                member(Cat, TestCategories)),
            List),
    (List=[] -> format_warning('*** No testcase with category ~w found~n', [Cat])
             ;  format(Stream, '~n~w:~n', [Cat]), makefile_write_calls(Stream, List)).

my_open(user_output,S) :- !, S=user_output.
my_open(File,S) :- open(File,write,S).
my_close(user_output) :- !.
my_close(S) :- close(S).

makefile_write_categories(Stream, []) :-
    format(Stream, '~n', []).
makefile_write_categories(Stream, [Cat | Cats]) :-
    format(Stream, ' ~a', [Cat]),
    makefile_write_categories(Stream, Cats).

makefile_write_calls(_Stream, []).
makefile_write_calls(Stream, [testcase(Id, _Infos, Arguments, Comment) | Tests]) :-
    (Comment='' -> true ; format(Stream,'\techo \"Test ~w : ~w\"~n',[Id,Comment])),
    format(Stream, '\t$(PROBCLI)', []),
    makefile_write_arguments(Stream, Arguments),
    makefile_write_diff(Stream, Id),
    makefile_write_calls(Stream, Tests).

makefile_write_arguments(Stream, []) :-
    format(Stream, '~n', []).
makefile_write_arguments(Stream, [Arg | Args]) :-
    format(Stream, ' ~w', [Arg]),
    (quote_next_arg(Arg) -> makefile_write_arguments_quoted(Stream,Args)
      ; makefile_write_arguments(Stream, Args)).

quote_next_arg('-goal'). % this will contain spaces ,... surround in "" for shell
quote_next_arg('--check_goal').
quote_next_arg('-check_goal').
quote_next_arg('-cbc_deadlock_pred').
quote_next_arg('-eval').
quote_next_arg('-evalt').
quote_next_arg('-evalf').
quote_next_arg('-cbc_sequence_with_target'). % actually quotes next two !
quote_next_arg('-cbc_sequence_with_target_all'). % ditto <- TO DO


makefile_write_arguments_quoted(Stream, []) :-
    format(Stream, '~n', []).
makefile_write_arguments_quoted(Stream, [Arg | Args]) :-
    format(Stream, ' \"~w\"', [Arg]),
    makefile_write_arguments(Stream, Args).

makefile_write_diff(Stream, ID) :-
    get_testcase_diff_check_output(ID, File1, File2),
    format(Stream, '\tdiff -b ~w ~w~n', [File1, File2]),
    fail.
makefile_write_diff(_Stream, _ID).

% -------------------------

copy(Cat) :- (Cat=[_|_] -> C=Cat ; C=[Cat]),
   generate_copy_commands(C,'testarchive/').

:- use_module(b_trace_checking,[get_default_trace_file/2]).
generate_copy_commands(Categories,Dest) :-
    cli_testcase(ID, TestCategories, _Infos, Arguments, _Comment),
    non_empty_inter(Categories,TestCategories), %print(ID),nl,
    Arguments=[MainFile|_], generate_copy_command(MainFile,Dest), % print(MainFile),nl,
    additional_testcase_file(ID,MainFile,Arguments,ExtraFile),
    generate_copy_command(ExtraFile,Dest),
    fail.
generate_copy_commands(_,_).

additional_testcase_file(ID,_,_,EFile) :- extra_testcase_file(ID,EFile).
additional_testcase_file(_ID,File,Arguments,TFile) :- member('-t',Arguments),
    get_default_trace_file(File,TFile).
additional_testcase_file(ID,_,_,RefFile2) :- get_testcase_diff_check_output(ID,_File1,RefFile2).

non_empty_inter(A,B) :- member(X,A), member(X,B),!.

:- use_module(tools_strings,[string_concatenate/3]).
:- use_module(tools,[get_parent_directory/2]).
generate_copy_command(File,Dest) :-
    safe_file_exists(File),
    get_parent_directory(File,Dir),!,
    string_concatenate(Dest,Dir,DestDir),
    string_concatenate(Dest,File,DestFile),
    format(user_output,'\tmkdir -p ~w~n',[DestDir]),
    get_command_path(mkdir,MkCmdPath),
    system_call(MkCmdPath, ['-p',DestDir],_Text1,_JExit1),
    format(user_output,'\tcp ~w ~w~n',[File,DestFile]),
    get_command_path(cp,CpCmdPath),
    system_call(CpCmdPath, [File,DestFile],_Text2,_JExit2).
generate_copy_command(_,_).

% -------------------------

:- dynamic test_failed/1, last_test_failed/1, test_diff_failed/3, test_skipped/1.
:- use_module(tools_printing,[start_terminal_colour/2, reset_terminal_colour/1, print_red/1]).
halt_tests :- test_failed(_),!,
   halt1.
halt_tests :-
   nl,print('TEST RUN SUCCESSFUL'),nl,
   halt. % regular halt
halt1 :-
   format_colour([red,bold],'TEST RUN FAILED~n',[]),
   halt(1).

check_failed(failure) :- test_failed(X),!,print_failed_tests,
   print('Use the following command to run individual tests: '),nl,
   print('   sicstus -l src/test_runner.pl --goal "run_id('),print(X),print(')."'),nl,
   (halt1_allowed -> halt1 ; print('halt(1) :: junit / dontstop mode'),nl).
check_failed(success) :- number_of_tests_run(Nr),
    start_terminal_colour([green,bold],user_output),
    findall(Y,test_skipped(Y),Skips), length(Skips,NrSkips),
    bb_get(reset_test_runner_info,Info),
   (Nr=1,NrSkips=0,Info\=category(_,_) -> format(user_output,'Test successful.~n',[])
    ; NrSkips>0 -> format(user_output,'All ~w tests successful, ~w skipped (for ~w).~n',[Nr,NrSkips,Info])
    ; format(user_output,'All ~w tests successful (for ~w).~n',[Nr,Info])),
    reset_terminal_colour(user_output).
print_failed_tests :- number_of_tests_run(Nr),
   findall(Y,test_failed(Y),Fails), length(Fails,NrFails),
   findall(Y,test_skipped(Y),Skips), length(Skips,NrSkips),
   start_terminal_colour([red,bold],user_error),
   bb_get(reset_test_runner_info,Info),
   format(user_error,'** Tests run: ~w, skipped: ~w, failed: ~w (for ~w)**~n** Failed tests:~n',[Nr,NrSkips,NrFails,Info]),
   test_failed(X), lookup_test_description(X,Desc),
   format(user_error,'~w ~w~n',[X,Desc]),
   fail.
print_failed_tests :- nl(user_error),
   (user_interrupt_signal_received -> format(user_error,'Tests were interrupted by CTRL-C (user_interrupt)~n',[])
     ; true),
   reset_terminal_colour(user_error).

lookup_test_description(Id,Desc) :- cli_testcase(Id,_,_,_,Desc).

:- dynamic user_interrupt_signal_received/0.
run_list(List) :- run_list(List,allow_skipping).
run_list(List,AllowSkipping) :- retractall(user_interrupt_signal_received),
     length(List,Len),
     run_single_testcase_list_aux(List,Len,AllowSkipping).
run_single_testcase_list_aux([],_,_) :- !.
run_single_testcase_list_aux(List,_Len,allow_skipping) :- user_interrupt_signal_received,!,
       length(List,RestLen),
       format('Skipping ~w remaining tests~n',[RestLen]),
       maplist(assert_test_skipped,List).
run_single_testcase_list_aux([TC|Tail],Len,AllowSkipping) :-
     print_progress_stats(Len),
     run_single_testcase(TC,AllowSkipping),
     run_single_testcase_list_aux(Tail,Len,AllowSkipping).
assert_test_skipped(Id) :- assert(test_skipped(Id)).

print_progress_stats(All) :- number_of_tests_run(Nr), Nr>0,!,
    findall(Y,test_failed(Y),Fails), length(Fails,NrFails),
    (last_test_failed(Failed) -> ajoin(['(e.g. ', Failed, ')'],FailInfo) ; FailInfo = ''),
    findall(Y,test_skipped(Y),Skips), length(Skips,NrSkips),
    bb_get(reset_test_runner_wtime,WStart),
    statistics(walltime,[WNow,_]), Delta is round((WNow - WStart) / 1000),
    bb_get(reset_test_runner_info,Info),
    format_colour([blue],'Progress:  ~w/~w tests run, ~w skipped, ~w failed ~w (running since ~w sec for ~w)~n',[Nr,All,NrSkips,NrFails,FailInfo,Delta,Info]).
print_progress_stats(_).

cache :- add_additional_arguments(['-cache','/Users/leuschel/svn_root/NewProB/examples/cache/']).
v :- add_additional_arguments(['-v']). % verbose
vv :- add_additional_arguments(['-vv']). % very_verbose

:- volatile additional_arguments/1.
:- dynamic additional_arguments/1.
% add additional cli arguments when running tests:
add_additional_arguments(List) :-
  (retract(additional_arguments(Old)) -> true ; Old=[]),
   append(Old,List,New),
   assert(additional_arguments(New)).

% auto complete a preference name and print error if no match found
get_preference_name(Prefix,Name) :- auto_complete_preference(Prefix,Name),!.
get_preference_name(Prefix,_) :-
    format_colour([red],'No matching preference found for: ~w~n',[Prefix]),fail.

:- use_module(preferences,[eclipse_preference/2]).
auto_complete_preference(Prefix,Name) :-
   atom_codes(Prefix,PC),
   eclipse_preference(EP,Name),
   (atom_codes(EP,EPC) ; atom_codes(Name,EPC)),
   append(PC,_,EPC).

% add additional preference when running tests:
add_additional_preference(PREF,PREFVAL) :-
  (retract(additional_arguments(Old0)) -> remove_matching_pref(Old0,PREF,Old) ; Old=[]),
   New = ['-p',PREF,PREFVAL|Old],
   format('New additional arguments: ~w~n',[New]),
   assert(additional_arguments(New)).

% remove all preferences conflicting with PREF
remove_matching_pref([],_PREF,[]).
remove_matching_pref(['-p',P,OLD|T],PREF,Res) :- !,
   (P=PREF -> Res=T ; Res = ['-p',P,OLD|RT], remove_matching_pref(T,PREF,RT)).
remove_matching_pref([H|T],PREF,[H|RT]) :- remove_matching_pref(T,PREF,RT).

% remove all preferencs conflicting with other list of prefs
remove_matching_prefs([],P,P).
remove_matching_prefs(['-cache',_File|T],InPrefs,Res) :- !,
   remove_matching_prefs(T,InPrefs,Res).
remove_matching_prefs(['-p',PREF,_|T],InPrefs,Res) :-
   remove_matching_pref(InPrefs,PREF,In2),
   remove_matching_prefs(T,In2,Res).
remove_matching_prefs([_|T],InPrefs,Res) :-
   remove_matching_prefs(T,InPrefs,Res).


% RUNNING SINGLE TESTCASE:
% ------------------------
run_single_testcase(testcase(Id,Infos,Arguments,_),AllowSkipping) :-
    % allows to run a test only using its Id
    var(Infos),
    var(Arguments),
    cli_testcase(Id, _, TestInfos, GroundArguments, Description),
    run_single_testcase(testcase(Id,TestInfos,GroundArguments,Description),AllowSkipping).
run_single_testcase(testcase(Id,Infos,Arguments,_Comment),allow_skipping) :-
    skip_test(Infos), !,
    full_flush,
    format_colour([blue,bold],'Skipping test ~w ~n', [Id]), print_junit_skip(Arguments),
    assert(test_skipped(Id)),
    full_flush.
run_single_testcase(testcase(Id,_Infos,Arguments,Comment),_) :-
    full_flush,
    format_colour([blue],'Running test ~w ~n      ~w~n',[Id,Comment]),
    (debug_mode(off) -> true ; print('testcase: probcli '), print_args(Arguments),nl),
    clear_diff_output(Id),
    clear_logxml_output(Id),
    prob_junit(JUnit),
    maplist(patch_prob_examples_loc,Arguments,Arguments0), % update path to prob_examples if necessary
    append(Arguments0, JUnit, Arguments1),
    (additional_arguments(ExtraArgs)
     -> remove_matching_prefs(ExtraArgs,Arguments1,RemArguments1), % remove conflicting arguments now overriden
        append(ExtraArgs,RemArguments1,Arguments2)
     ; Arguments1=Arguments2),
    (tests_multiply_timeout(Factor) -> modify_timeout(Factor,Arguments2,Arguments3) ; Arguments3 = Arguments2),
    (silent_running -> true ; print('executing: probcli '), print_args(Arguments3),nl),
    full_flush,
    test_started(Id),
    catch(prob_cli:go_cli(Arguments3,[test_runner]), Exception, true), !,
    test_finished(Id,Walltime),
    (
      Exception == halt(0)                        -> check_diff_output(Id), check_logxml(Id) ;
      var(Exception)                              -> check_diff_output(Id), check_logxml(Id) ;
      Exception = error(X,Y)                      -> (halt1_allowed -> format_warning('error on test execution: error(~w,~w)~n',[X,Y]), halt1 ; test_failed(Id,Exception)) ;
      Exception = enumeration_warning(_,_,_,_,_)  -> exception_during_test(Id,'Enumeration',Exception) ;
      Exception = solver_and_provers_too_weak     -> exception_during_test(Id,'Solver',Exception) ;
      Exception == halt(1)                        -> test_failed(Id, Exception) ;
      Exception == parse_errors(_PE)              -> test_failed(Id, Exception) ;
      Exception == user_interrupt_signal          -> assert(user_interrupt_signal_received),
                                                     format_warning('CTRL-C received, aborting tests~n',[]),
                                                     test_failed(Id, Exception) ;
      otherwise                                   -> test_failed(Id, Exception), format_warning('Exception not caught in test_runner: ~w~n',[Exception]), halt1
     ), !,
    (test_failed(Id)
     -> true % messages already printed above
     ; format_colour([green],'Test ~w completed successfully (in ~w ms)~n~n',[Id,Walltime])
    ),
    on_exception(ResetExc,
       prob_cli:reset_cli, % when bundles/extensions are missing we can get exceptions here
       exception_during_test(Id,'Reset probcli',ResetExc)),
    full_flush.

exception_during_test(Id,Type,Exception) :-
    (halt1_allowed -> format_warning('exception (~w) on test execution: ~w)~n',[Type,Exception]), halt1
     ; test_failed(Id,Exception)
    ).

:- dynamic test_took_aux/1.
:- dynamic last_testcase_run/1, number_of_tests_run/1.
number_of_tests_run(0).
reset_nr_of_tests :- retractall(number_of_tests_run(_)), assert(number_of_tests_run(0)).

:- use_module(library(system),[now/1, datime/2]).
:- dynamic performance_session_running/1, performance_session_stats/4.

performance_session_start :-
    now(When),
    datime(When,datime(Year,Month,Day,Hour,Min,Sec)),
    format('~nStarting Codespeed Performance Monitoring session ~w:~w:~w:~w:~w:~w~n',[Year,Month,Day,Hour,Min,Sec]),
    retractall(performance_session_running(_)),
    assert(performance_session_running(When)).

:- use_module(parsercall,[get_parser_version/1]).
:- use_module(version, [version_str/1, revision/1, lastchangeddate/1]).
performance_session_end(FilePrefix) :-
    performance_session_running(When),
    datime(When,datime(Year,Month,Day,Hour,Min,Sec)),
    %tools:ajoin([FilePrefix,':',Year,Month,Day,Hour,Min,Sec],FileName),
    format('~nFinishing Codespeed session ~w:~w:~w:~w:~w:~w~n -> File : ~w~n',[Year,Month,Day,Hour,Min,Sec,FilePrefix]),
    open(FilePrefix,append,S),
    format(S,'~n/* Codespeed session ~w:~w:~w:~w:~w:~w */~n',[Year,Month,Day,Hour,Min,Sec]),
    version_str(Vers), portray_clause(S, session_prob_version(When,Vers)),
    revision(Rev), portray_clause(S, session_prob_revision(When,Rev)),
    lastchangeddate(DD), portray_clause(S, session_prob_lastchangeddate(When,DD)),
    get_parser_version(PV), portray_clause(S, session_prob_parser_version(When,PV)),
    write_perf_data(When,S).

write_perf_data(When,S) :- additional_arguments(New),
    portray_clause(S, stored_additional_arguments(When,New)),
    fail.
write_perf_data(When,S) :- performance_session_stats(When,Id,Time,WTime),
    portray_clause(S, stored_performance_test_stats(When,Id,Time,WTime)),
    fail.
write_perf_data(_When,S) :- nl(S), nl(S), close(S).



test_started(Id) :-
    retractall(last_testcase_run(_)), assert(last_testcase_run(Id)),
    retractall(test_took_aux(_)),
    statistics(runtime,[Start,_]),
    statistics(walltime,[WStart,_]),
    bb_put(test_started,Start),
    bb_put(test_started_wtime,WStart).

:- dynamic test_stats/5.
test_finished(Id,WTime) :-
    statistics(runtime,[End,_]),
    statistics(walltime,[WEnd,_]),
    bb_get(test_started,Start),
    bb_get(test_started_wtime,WStart),
    Time is End - Start, WTime is WEnd- WStart,
    (retract(number_of_tests_run(Nr)) -> N1 is Nr+1 ; N1=1),
     assert(number_of_tests_run(N1)),
    (retract(test_stats(Id,PrevTime,PrevWTime,_,_))
       -> assert(test_stats(Id,Time,WTime,PrevTime,PrevWTime))
        ; assert(test_stats(Id,Time,WTime,-1,-1))
    ),
    (performance_session_running(When)
     -> assert(performance_session_stats(When,Id,Time,WTime))
     ;  true).

print_delta_stats :- print('Comparing walltimes with previous test run: '),nl,
   findall(delta(DeltaPerc,DeltaWTime,Id),test_delta_stat(Id,DeltaPerc,DeltaWTime),L),
   (L=[] -> print('No previous run information available'),nl
    ; print(' ID |  % (delta absolute) | walltime (runtime)~n'),nl,
      sort(L,SL),
      maplist(print_delta,SL)).
test_delta_stat(Id,DeltaPerc,DeltaWTime) :-
   test_stats(Id,_RTime,WTime,_PrevRTime,PrevWTime),
   PrevWTime>0,
   DeltaWTime is WTime - PrevWTime,
   DeltaPerc is (100*DeltaWTime) / PrevWTime.
print_delta(delta(DeltaPerc,DeltaWTime,Id)) :-
   test_stats(Id,RTime,WTime,_PrevRTime,PrevWTime),
   format(' ~w | ~2f % (~w ms) | ~w ms (~w ms runtime) [~w walltime ms previously]~n',
          [Id,DeltaPerc,DeltaWTime,WTime,RTime,PrevWTime]).

:- use_module(tools,[print_memory_used_wo_gc/0]).
print_current_stats :-
   print_memory_used_wo_gc,nl,
   format('  ID  | OK | WALLTIME | RUNTIME~n',[]),
   test_stats(Id,RTime,WTime,_PrevRTime,_PrevWTime),
   (test_failed(Id) -> OK = '*FAILED*' ; OK = ' OK '),
   format(' ~w | ~w | ~w ms | ~w ms runtime~n',
          [Id,OK,WTime,RTime]),
  fail.
print_current_stats.

test_took(X) :- test_took_aux(X), !.
test_took(XInSeconds) :-
    statistics(runtime,[End,_]),
    bb_get(test_started,Start),
    X is End - Start,
    XInSeconds is X / 1000,
    asserta(test_took_aux(XInSeconds)).

print_junit_skip(Arguments) :-
    prob_junit(['-junit', Dir])
    -> set_junit_dir(Dir), test_took(X),
       create_junit_result(Arguments,'Integration Tests','Integration_Test',X,skip,Result),
       print_junit([Result],'Integration_Test')
    ;  true.

% if the test expects a time_out error, the timeout is not expanded
% otherwise, timeout is increased to allow coverage analysis / junit / etc to finish
modify_timeout(_,OldOptions,OldOptions) :- segment(OldOptions,['-expcterr','time_out']), !.
modify_timeout(Factor,[],['-p','TIME_OUT',NVal]) :- % timeout was not set at all - set it to Factor*Default
   % Note: there is a potential problem when the time_out is set inside the machine and not in the test !! TO DO: fix
   preferences:preference_default_value(time_out,DEFAULT),
   NVal is Factor * DEFAULT.
modify_timeout(Factor,['-p','TIME_OUT',OLD|T],['-p','TIME_OUT',NewT|T]) :-
   tools:arg_is_number(OLD,OLDT), !,
   % preferences:preference_default_value(time_out,DEFAULT),
   % OLDT < DEFAULT,!, % we explicitly set a TIME_OUT lower than the default value
   NewT is OLDT * Factor.
modify_timeout(Factor,[H|T],[H|MT]) :- modify_timeout(Factor,T,MT).

full_flush :- flush_output(user_output), flush_output(user_error).

print_args([]).
print_args([H|T]) :- print(H), print(' '), print_args(T).

test_failed(Id,Msg) :-
  test_failed(Id,Msg,'').
test_failed(Id,Msg1,Msg2) :-
  (Msg1 = user_interrupt_signal -> assert(user_interrupt_signal_received) ; true),
  test_took(X),
  cli_testcase(Id,Categories,_Infos,Arguments,Comment),
  Categories = [FirstCat|_], !,
  ErrorMessage = ['Test with Id',Id,'failed.','Test Categories: ',Categories,'Test Arguments: ',Arguments,'Test Comment: ',Comment,'Error Messages in Test Runner: ',Msg1,Msg2,'Content of Error Manager: '],
  (get_all_errors(AllErrors) -> true ; AllErrors = []),
  append(ErrorMessage,AllErrors,FullErrorMessage),
  create_junit_result(Id,'Integration Tests',FirstCat,X,error(FullErrorMessage),Result),
  print_junit([Result],'Integration_Test'),
  assert_test_failed(Id),
  start_terminal_colour(red,user_error),
  format(user_error,'*** Test ~w FAILED: ~w~w~n', [Id,Msg1,Msg2]),
  reset_terminal_colour(user_error).

assert_test_failed(Id) :-
   assert(test_failed(Id)),
   retractall(last_test_failed(_)),
   assert(last_test_failed(Id)).

diff_failed(Id,F1,F2,DiffOutputText,CmpOutputText) :-
  test_took(X),
  atom_codes(DiffOutputAtom,DiffOutputText),
  atom_codes(CmpOutputAtom,CmpOutputText),
  ErrMsg = ['Diff for test with Id\n',Id,'\nfailed:\n','Output file\n',F1,'\ndoes not correspond to stored version\n',F2,
            '\nOutput of Diff:\n',DiffOutputAtom,
            '\nOutput of Cmp:\n',CmpOutputAtom],
  create_junit_result(Id,'Diff Checking','Diff_checking',X,
              error(ErrMsg),
              Result),
  print_junit([Result],'Diff_checking'),
  assert_test_failed(Id),
  assert(test_diff_failed(Id,F1,F2)),
  start_terminal_colour(red,user_error),
  format(user_error,'*** Test ~w FAILED: Diff failed:~nOutput file ~w~ndoes not correspond to stored version~n~w~n', [Id,F1,F2]),
  format(user_error,'Diff:~n~s~n',[DiffOutputText]),
  format(user_error,'Cmp:~n~s~n',[CmpOutputText]),
  reset_terminal_colour(user_error).

:- use_module(tools_commands,[diff_files_with_editor/2]).
diff_in_editor :- findall(I,test_diff_failed(I,_,_),LI), sort(LI,SI),
  format('Opening failed diff files in editor: ~w~n',[SI]),
  test_diff_failed(Id,F1,F2),
  format('Test ~w~n~w ~w~n',[Id,F1,F2]),
  diff_files_with_editor(F1,F2),
  fail.
diff_in_editor.


clear_diff_output(Id) :- % clear all files that should be generated
  get_testcase_diff_check_output(Id,GeneratedFile,_StoredReferenceFile),
  safe_file_exists(GeneratedFile),
  (cli_testcase_do_not_delete(Id,GeneratedFile) -> formatsilent(user_output,'% Keeping: ~w~n',[GeneratedFile])
   ; formatsilent(user_output,'% Deleting: ~w~n',[GeneratedFile]),
     delete_file(GeneratedFile)
  ),fail.
clear_diff_output(_).

check_diff_output(Id) :-
  findall(diff(Id,GeneratedFile,StoredReferenceFile),
          get_testcase_diff_check_output(Id,GeneratedFile,StoredReferenceFile),
      ListOfDiffsToCheck),
  maplist(check_diff_output2, ListOfDiffsToCheck).

check_diff_output2(diff(Id,GeneratedFile,StoredReferenceFile)) :-
  \+ safe_file_exists(GeneratedFile)       -> test_failed(Id,'Output file does not exist:',GeneratedFile) ;
  \+ safe_file_exists(StoredReferenceFile) -> test_failed(Id,'Stored file does not exist:',StoredReferenceFile) ;
  otherwise                           -> diff(Id,GeneratedFile,StoredReferenceFile).

diff(Id,F1,F2) :-
  formatsilent(user_output,'% Checking: diff / cmp ~w ~w~n',[F1,F2]),
  get_command_path(diff,DiffPath),
  get_command_path(cmp,CmpPath),
  (system_call(DiffPath,['-b',F1,F2],DiffOutputText,_ErrTextDiff,ExitDiff) % use -q for quiet
    -> true
     ; DiffOutputText = "*** CALLING DIFF FAILED !", ExitDiff = fail
  ),
  formatsilent(user_output,'% Checking: cmp ~w ~w~n',[F1,F2]),
  (system_call(CmpPath,['-b',F1,F2],CmpOutputText,_ErrTextCmp,_ExitCmp)
    -> true
     ; CmpOutputText = "*** CALLING CMP FAILED !"
  ),
  (ExitDiff = exit(0)%, ExitCmp = exit(0)
   -> true
   ; diff_failed(Id,F1,F2,DiffOutputText,CmpOutputText)).


% ----------------
:- volatile logxml_file/1.
% logxml file used automatically when started by test_runner
:- dynamic logxml_file/1.
clear_logxml_output(_Id) :- logxml_file(GeneratedFile),
  safe_file_exists(GeneratedFile),
  delete_file(GeneratedFile),
  fail.
clear_logxml_output(_).

set_logxml_file(File) :- retractall(logxml_file(_)), assert(logxml_file(File)),
    format('Adding -logxml ~w.~n',[File]),
    add_additional_arguments(['-logxml',File]).

:- use_module(logger,[read_xml_log_file/2]).
check_logxml(Id) :-
   logxml_file(File),!,
   xmllint(Id,File),
   check_logxml(Id,File).
check_logxml(_).

xmllint(Id,File) :-
   (call_xmllint(File,ErrText,ExitLint)
    -> (ExitLint = exit(0) -> true
        ;  test_failed(Id,'xmllint reported errors for file:',File),
           formatsilent(user_error,'% xmllint result~n~s~n',[ErrText])
       )
     ; test_failed(Id,'Calling xmllint failed for file:',File)
   ).

:- dynamic use_logxml_schema/0.
set_use_logxml_schema :- (use_logxml_schema -> true ; assert(use_logxml_schema)).
call_xmllint(File,ErrText,ExitLint) :- use_logxml_schema,
   absolute_file_name(prob_home('doc/logxml_xsd.xml'),SchemaFile),
   file_exists(SchemaFile),!,
   get_command_path(xmllint,LintPath),
   format('Calling ~w with schema ~w for logxml file ~w~n',[LintPath,SchemaFile,File]),
   system_call(LintPath,['--schema', SchemaFile, File, '-noout'],_OutputText,ErrText,ExitLint).
call_xmllint(File,ErrText,ExitLint) :-
   get_command_path(xmllint,LintPath),
   format('Calling ~w without schema for logxml file ~w~n',[LintPath,File]),
   system_call(LintPath,[File],_OutputText,ErrText,ExitLint).

check_logxml(Id,File) :-
   on_exception(E, read_xml_log_file(File,Infos),
       test_failed(Id,'Exception trying to read logxml file:',E)),
   !,
   format_colour([blue],'Read logxml file ~w, contains ~w.~n',[File,Infos]),
   check_error_cound(Id,Infos).
check_logxml(Id,File) :-
   test_failed(Id,'Logxml file could not be read:',File).

check_error_cound(Id,Infos) :-
   testcase_expects_errors(Id),!,
   (member(expected_errors/E,Infos),E>0 -> true ; test_failed(Id,'Logxml file does not report errors:',Infos)).
check_error_cound(Id,Infos) :-
   (member(errors/0,Infos),member(expected_errors/0,Infos) -> true
    ; test_failed(Id,'Logxml file reports errors:',Infos)).

testcase_expects_errors(Id) :-
   cli_testcase(Id,_Cat,_Infos,Arguments,_Comment),
   member(A,Arguments), expect_error_arg(A).

expect_error_arg('-expcterr').
expect_error_arg('-expecterr').
expect_error_arg('-expect').
expect_error_arg('-expcterrpos').

% ------------------

:- volatile repl_mode/0.
:- dynamic repl_mode/0.

:- dynamic prob_examples_location/1.
set_prob_examples_location(Dir) :-
   retractall(prob_examples_location(_)),
   (atom(Dir) -> atom_codes(Dir,DC) ; DC=Dir),
   format('Setting location of prob_examples directory to: ~s~n',[DC]),
   assert(prob_examples_location(DC)).

% update path to prob_examples if necessary:
patch_prob_examples_loc(Arg,PatchedArg) :-  atom(Arg),
   prob_examples_location(NewLoc),
   atom_codes(Arg,ArgC),
   append("../prob_examples",Rest,ArgC),
   !,
   append(NewLoc,Rest,NewArgC),
   atom_codes(PatchedArg,NewArgC),
   format('Patched ~w to ~w~n',[Arg,PatchedArg]).
patch_prob_examples_loc(A,A).

get_testcase_diff_check_output(Id,PF1,PF2) :-
   cli_testcase_diff_check_output(Id,F1,F2),
   patch_prob_examples_loc(F1,PF1),
   patch_prob_examples_loc(F2,PF2).

:- use_module(extension('counter/counter'),[counter_init/0]).
:- use_module(prob_startup, [startup_prob/0]).
init_test_runner :- startup_prob,counter_init.
:- use_module(library(lists),[maplist/2]).
% a minimal shell to execute tests:
test_repl :-
   init_test_runner,
   print_prob_version,
   assert(repl_mode),
   prolog_flag(argv,ArgV), treat_argv(ArgV),
   test_repl_loop,
   retractall(repl_mode).

print_prob_version :-
   version_str(VS), revision(Rev), lastchangeddate(DD),
   format('ProB ~w~n Revision: ~w~n Date: ~w~n',[VS,Rev,DD]).


treat_argv(['-prob-examples',Dir|T]) :- !, set_prob_examples_location(Dir),
  treat_argv(T).
treat_argv(Args) :- maplist(eval_argv,Args).

% execute tests provided on the command-line:
eval_argv(Cmd) :- format('ARGV ==> ~w~n',[Cmd]),
   atom_codes(Cmd,C), safe_number_codes(Nr,C), !, test_eval(Nr).
eval_argv(Cmd) :- test_eval(Cmd),!.

test_repl_loop :- safe_read(T), test_eval(T), !, test_repl_loop.
test_repl_loop.

safe_read(T) :- print('TEST ==> '),
   on_exception(error(syntax_error(E),_),read(T),
       (format_warning('*** Syntax error: ~w~n*** Type Prolog term followed by a dot(.) and enter.~n',[E]),
        safe_read(T))).

:- meta_predicate wall(0).
wall(Call) :-
    statistics(walltime,[Start,_]),
    call(Call),
    statistics(walltime,[Stop,_]), WT is Stop-Start,
    format('Walltime: ~w ms~n',[WT]),
    print_memory_used_wo_gc,nl.

% -------------------------

:- use_module(library(file_systems)).
:- use_module(tools,[get_options/5]).
% true for test_files
test_file(Id,File,AbsFileName) :- cli_testcase(Id,_Cat,_Infos,Arguments,_Comment),
   get_options(Arguments,prob_cli:recognised_cli_option,_Options,Files,fail),
   member(File,Files),
   is_existing_file(File),
   absolute_file_name(File,AbsFileName).

is_existing_file(X) :- \+ number(X), atom(X),
   atom_codes(X,Codes),[BS] = "/", (member(BS,Codes) -> true),
   file_exists(X).

% obtain a list of all files used in tests
all_files(Files) :- findall(F,test_file(_,_,F),A), sort(A,Files).

% a test file that can be loaded:
valid_test_file(Id,File,AbsFileName,XT) :-
   test_file(Id,File,AbsFileName),
   cli_testcase(Id,_Cat,_Infos,Arguments,_Comment),
   \+ append(_,['-expcterr', load_main_file |_],Arguments),
   tools:get_filename_extension(File,XT).
all_valid_files(Files,Mode) :-
   findall(F, (valid_test_file(_,_,F,XT),
               relevant_extension(XT,Mode)),
               A),
   sort(A,Files).

% traverse a directory and indicate which specification files are used in tests and which ones not
traverse :- traverse('../prob_examples/public_examples/').
traverse(SD) :- all_files(Files), absolute_file_name(SD,StartDir),
    format('Examining files in ~w~n +  means file is used in some test~n~n',[StartDir]),
    traverse(StartDir,Files).

traverse(Dir,AllFiles) :- file_member_of_directory(Dir,_,FullFile),
    tools:get_filename_extension(FullFile,XT),
    (member(FullFile,AllFiles) -> format(' +  ~w~n',[FullFile])
      ; relevant_extension(XT,_) -> format('--- ~w~n',[FullFile])),
    fail.
traverse(Dir,AllFiles) :- directory_member_of_directory(Dir,_,SubDir),
    %format('~nSTART ~w~n',[SubDir]),
    traverse(SubDir,AllFiles),
    %format('~n  END ~w~n',[SubDir]),
    fail.
traverse(_,_).

relevant_extension('mch',b).
relevant_extension('ref',b).
relevant_extension('imp',b).
relevant_extension('tla',tla).
relevant_extension('fuzz',z).
relevant_extension('tex',z).
relevant_extension('csp',csp).
relevant_extension('cspm',csp).
relevant_extension('eventb',eventb).

% --------------------------

test_eval(quit) :- !,fail.
test_eval(q) :- !,fail.
test_eval(end_of_file) :- !,fail.  % Ctrl-D
test_eval(Cmd) :- test_eval1(Cmd),!.
test_eval(Cmd) :-
    format_colour([red],'Error executing command: ~w~n',[Cmd]).

test_eval1(N) :- number(N),!, wall(run_id(N)).
test_eval1(last) :- !, wall(run_last_test).
test_eval1(N-M) :- number(N), number(M),!, wall(run_tests_by_id(N-M)).
test_eval1(repeat(ID,M)) :- !, repeat_id(ID,M).
test_eval1(r) :- !, run_random_tests(25).
test_eval1(v) :- !,v.
test_eval1(verbose) :- !,v.
test_eval1(all_files) :- !, all_files(Files), length(Files,Len),
   format('~nFiles = ~n~w~n # Files = ~w~n',[Files,Len]).
test_eval1(valid_files(Mode)) :- !, all_valid_files(Files,Mode), length(Files,Len),
   format('~nValid ~w Files = ~n~w~n # Files = ~w~n',[Mode,Files,Len]).
test_eval1(files) :- !, traverse.
test_eval1(files(Dir)) :- !, traverse(Dir).
test_eval1(ex(Dir)) :- !, set_prob_examples_location(Dir).
test_eval1(cache) :- !,print('Enabling cache'),nl,
    cache.
test_eval1(debug) :- !,print('Enabling Prolog debugging mode (use -v or -vv for ProB debugging info)'),nl,
    debug,
    retractall(multiply_timeout(_)),
    assert(multiply_timeout(10)).
test_eval1(factor(X)) :- !,
    retractall(multiply_timeout(_)),
    format('Setting timeout factor to ~w~n',[X]),
    assert(multiply_timeout(X)).
test_eval1(timeout(X)) :- !,
    format('Adding -timeout ~w for model checking, disproving~nUse factor(X) to set TIME_OUT factor.~n',[X]),
    add_additional_arguments(['-timeout',X]).
test_eval1(logxml) :- !, File = './test_runner_logxml.xml',
    set_logxml_file(File).
test_eval1(xsd) :- !, File = './test_runner_logxml.xml',
    set_logxml_file(File), set_use_logxml_schema.
test_eval1(debug_off) :- !,print('Disabling Prolog debugging mode'),nl,
    nodebug,
    retractall(multiply_timeout(_)).
test_eval1(clpfd) :- !,print('Enabling CLPFD'),nl,
    add_additional_preference('CLPFD','TRUE').
test_eval1(clpfd_off) :- !,print('Disabling CLPFD'),nl,
    add_additional_preference('CLPFD','FALSE').
test_eval1(smt) :- !,print('Enabling SMT'),nl,
    add_additional_preference('SMT','TRUE').
test_eval1(smt_off) :- !,print('Disabling SMT'),nl,
    add_additional_preference('SMT','FALSE').
test_eval1(chr) :- !,print('Enabling CHR'),nl,
    add_additional_preference('CHR','TRUE').
test_eval1(chr_off) :- !,print('Disabling CHR'),nl,
    add_additional_preference('CHR','FALSE').
test_eval1(cse_off) :- !,print('Disabling CSE'),nl,
    add_additional_preference('CSE','FALSE').
test_eval1(cse) :- !,print('Enabling CSE'),nl,
    add_additional_preference('CSE','TRUE').
test_eval1(cse_subst) :- !,print('Enabling CSE_SUBST'),nl,
    add_additional_preference('CSE','TRUE'),
    add_additional_preference('CSE_SUBST','TRUE').
test_eval1(trace_info) :- !,print('Enabling TRACE_INFO'),nl,
    add_additional_preference('TRACE_INFO','TRUE').
%
test_eval1(p(PREF)) :- !,
    get_preference_name(PREF,PName),
    print('Enabling Preference '),print(PName),nl,
    add_additional_preference(PName,'true').
test_eval1(p(PREF,VAL)) :- !,
    get_preference_name(PREF,PName),
    print('Setting Preference '),print(PName),nl,
    add_additional_preference(PName,VAL).
test_eval1(random) :- !,print('Enabling RANDOMISE_ENUMERATION_ORDER'),nl,
    add_additional_preference('RANDOMISE_ENUMERATION_ORDER','TRUE').
test_eval1(random_off) :- !,print('Enabling RANDOMISE_ENUMERATION_ORDER'),nl,
    add_additional_preference('RANDOMISE_ENUMERATION_ORDER','FALSE').
test_eval1(sanity_check) :- !, sanity_check(false).
test_eval1(sc) :- !, sanity_check(false).
test_eval1(trace) :- !, print('Enabling TRACE_UPON_ERROR'),nl,
    add_additional_preference('TRACE_UPON_ERROR','TRUE').
test_eval1(trace_off) :- !, print('Disabling TRACE_UPON_ERROR'),nl,
    add_additional_preference('TRACE_UPON_ERROR','FALSE').
test_eval1(raise) :- !,print('Enabling STRICT_RAISE_ENUM_WARNINGS'),nl,
    add_additional_preference('STRICT_RAISE_ENUM_WARNINGS','TRUE').
test_eval1(nopt) :- !,print('Disabling OPTIMIZE_AST'),nl,
    add_additional_preference('OPTIMIZE_AST','FALSE').
test_eval1(vv) :- !,vv.
test_eval1(silent) :- !, run_silently.
test_eval1(x) :- !,halt.
test_eval1(reload) :- !,use_module(probsrc(test_runner)), use_module(probsrc(testcases)).
test_eval1(edit) :- last_testcase_run(Id),
   cli_testcase(Id,_,_Infos,Arguments,_Comment),
   member(File,Arguments),  safe_file_exists(File),!,
   edit_file(File).
test_eval1(e) :- !, test_eval1(edit).
test_eval1(diff) :- !, diff_in_editor.
test_eval1(halt) :- !,halt.
test_eval1(info) :- !, get_total_number_of_errors(X), format('~nTotal number of errors: ~w~n~n',[X]).
test_eval1(cat) :- !, print('Categories: '),
   findall(Cat,
           (cli_testcase(_, TestCategories, _, _, _),  member(Cat, TestCategories)), List),
   sort(List,SL), print(SL),nl.
test_eval1(cat(Category)) :- !,
   wall(run_tests_by_category(Category,all)).
test_eval1(list(Category)) :- !,
   get_tests_for_categories([Category],TList),
   findall(Id,member(testcase(Id, _, _, _),TList),List),
   format('Tests for category: ~w~n',[List]).
test_eval1(make(Categories)) :- !,
   wall(makefile_by_category(user_output,Categories)).
test_eval1(make(File,Categories)) :- !,
   wall(makefile_by_category(File,Categories)).
test_eval1(files(Category)) :- !, show_files(Category).
test_eval1(uses(Command)) :- !, wall(run_tests_using_command(Command)).
test_eval1(uses(Pref,Val)) :- !, wall(run_tests_using_preference(Pref,Val)).
test_eval1(p) :- !, test_eval1(profile).
test_eval1(ps) :- !, test_eval1(print_profile).
test_eval1(pc) :- !, test_eval1(print_coverage).
test_eval1(delta) :- !, print_delta_stats.
test_eval1(stats) :- !, print_current_stats.
test_eval1(start) :- !, performance_session_start.
test_eval1(stop) :- !, performance_session_end('log/test_runner_performance_log.pl').
test_eval1(codespeed) :- !,
   performance_session_start,
   test_eval1(cat(codespeed)),
   performance_session_end('log/test_runner_performance_codespeed_log.pl').
test_eval1(profile) :- !, print('PROFILING : '), %spy([avl:avl_size/2]),
   (current_prolog_flag(profiling,on)
    -> set_prolog_flag(profiling,off), print('OFF') ;
       set_prolog_flag(profiling,on), print('ON')),
    nl,print('USE ps to print_profile or pc to print_coverage info'),nl.
test_eval1(profile_stats) :- !, test_eval1(print_profile).
test_eval1(print_profile) :- !, nl,print('PROFILE INFORMATION:'), nl,
   on_exception(error(existence_error(_,_),_),print_profile, print('CAN ONLY BE USED WHEN RUNNING PROB FROM SOURCE')),
   nl, debug:timer_statistics.
test_eval1(print_coverage) :- !, nl,print('COVERAGE INFORMATION:'), nl,
   (prolog_flag(source_info,on) -> true ; print_red('Only useful when prolog_flag(source_info,on)!'),nl),
   on_exception(error(existence_error(_,_),_),print_coverage, print('CAN ONLY BE USED WHEN RUNNING PROB FROM SOURCE')),
   nl.
test_eval1(profile_reset) :- !, nl,print('RESETTING PROFILE INFORMATION'), nl,
   on_exception(error(existence_error(_,_),_),profile_reset, print('CAN ONLY BE USED WHEN RUNNING PROB FROM SOURCE')),
   nl.
test_eval1(perf) :- !, toggle_perfmessages.
test_eval1(Category) :- valid_category(Category),!,
   wall(run_tests_by_category(Category,all)).
test_eval1(cs(Id)) :- !, codespeed_id(Id).
test_eval1('-'(ProBCLICommand)) :- atom(ProBCLICommand),
   atom_concat('-',ProBCLICommand,Atom),
   (prob_cli:recognised_option(Atom,Call) -> true % from probcli.pl
     ; format('Unknown probcli command: -~w~n',[ProBCLICommand]),fail),
   !,
   format('Executing probcli command ~w~n',[Call]),
   call(prob_cli:Call).
test_eval1([H|T]) :- number(H),!, wall(run_tests_by_id([H|T],_,no_skipping)).
test_eval1([Category|T]) :- valid_category(Category),!,
   wall(run_tests_by_category([Category|T],all)).
test_eval1(C) :-
     (C=help -> true ; format('*** Unknown command ~w~n',[C])),
     print(' Commands:   Nr,  Nr-Nr,  last, q, x, v, vv, uses(Cmd), uses(Pref,Val),'),nl,
     print('             repeat(id,nr), timeout(ms),factor(f), e,edit,diff, r (for 25 random tests),'),nl,
     print('             cat, cat(Cats),make(Cats),make(File,Cats),files(Cat),'),nl,
     print('             profile, profile_stats, (to turn Prolog profiling on and print info)'),nl,
     print('             debug,debug_off,        (to turn Prolog debug mode on or off)'),nl,
     print('             perf, reload, sc,'),nl,
     print('             -CMD,  (for probcli commands like -profile)'),nl,
     print('  * for setting preferences:'),nl,
     print('             p(PREF), p(PREF,VAL),'),nl,
     print('             clpfd,clpfd_off, smt,smt_off, chr,chr_off, cse,cse_subst,cse_off,'),nl,
     print('             random,random_off, trace_info, nopt,'),nl,
     print('             cache, perf,       (turn ProB caching or performance messages on)'),nl,
     print('             trace,trace_off,   (set TRACE_UPON_ERROR preference)'),nl,
     print('  * statistics:'),nl,
     print('             delta, stats, info.'),nl.

safe_file_exists(F) :- atom(F), file_exists(F).

valid_category(Cat) :-
   cli_testcase(_Id, TestCategories, _Infos, _Arguments, _Comment),
   member(Cat, TestCategories).


% ---------------------------

% run test using binary
% binaries are registered in a file probcli_bak/codespeed_versions.pl
% results for test ID are written to probcli_bak/codespeed/res_ID.csv

:- use_module(tools, [ajoin/2,ajoin_with_sep/3]).
codespeed_id([]) :- !.
codespeed_id([H|T]) :- !, codespeed_id(H),!, codespeed_id(T).
codespeed_id(Id) :-
    use_module('probcli_bak/codespeed_versions.pl'), % contains probcli_binary
    cli_testcase(Id,_TestCategories,_TestInfos,Arguments,Comment),
    format('Benchmarking test ~w using probcli binaries~nTest Description: ~w~n',[Id,Comment]),
    maplist(convert_arg_to_atom,Arguments,Args2),
    findall(probcli_binary(V1,V2,V3,F,Path,Sics,Date),probcli_binary(V1,V2,V3,F,Path,_Hash,Sics,Date),Binaries),
    Repeats=3,
    maplist(bench_probcli_binary_testcase(Id,Args2,Repeats),Binaries,Walltimes,Oks),

    ajoin(['codespeed/res_',Id,'.csv'],FName),
    absolute_file_name(FName,AF,[relative_to(probcli_bak)]),
    format('Writing codespeed results to ~w~n',[AF]),
    open(AF,write,Stream),
    format(Stream,'\"Test ~w\"~n',[Id]),
    format(Stream,'\"~w\"~n',[Comment]),
    ajoin_with_sep(Arguments,' ',ArgsStr),
    format(Stream,'Command:,\"probcli ~w\"~n~n',[ArgsStr]),
    format(Stream,'\"~w.~w.~w-~w\",~w,~w,~w,~w,~w,~w,\"~w\"~n',[v1,v2,v3,f,sics,date,avg,min,max,ok,walltimes]),
    maplist(print_codespeed_results(Stream,Repeats),Binaries,Walltimes,Oks),
    close(Stream),
    maplist(print_codespeed_results(user_output,Repeats),Binaries,Walltimes,Oks).

%:- use_module(library(statistics),[min_max/3]).
print_codespeed_results(Stream,Repeats,probcli_binary(V1,V2,V3,F,_Path,sicstus(S1,S2,S3),date(Y,M,D)),Walltimes,Ok) :-
   sumlist(Walltimes,Sum),
   Average is Sum / Repeats,
   min_member(Min,Walltimes),
   max_member(Max,Walltimes),
   ajoin([S1,'.',S2,'.',S3],Sics),
   ajoin([Y,'/',M,'/',D],Date),
   format(Stream,'\"~w.~w.~w-~w\",~w,~w,~w,~w,~w,~w,\"~w\"~n',[V1,V2,V3,F,Sics,Date,Average,Min,Max,Ok,Walltimes]).

bench_probcli_binary_testcase(Id,Args,Repeats,probcli_binary(V1,V2,V3,F,Path,_,_),Walltimes,Ok) :-
    format('~nRunning test ~w using version ~w.~w.~w-~w (~w)~n',[Id,V1,V2,V3,F,Path]),
    run_probcli_binary_testcase(Id,Path,Args,_WT,Ok), % run once for parser
    rep_bench(Repeats,Id,Path,Args,Walltimes).

rep_bench(0,_,_,_,[]).
rep_bench(Nr,Id,Path,Args,[WT1|WTR]) :- Nr>0,
    run_probcli_binary_testcase(Id,Path,Args,WT1,_),
    N1 is Nr-1,
    rep_bench(N1,Id,Path,Args,WTR).

:- use_module(system_call, [system_call/5]).
run_probcli_binary_testcase(Id,Path,Arguments,WT,Ok) :-
    statistics(walltime,[Start,_]),
%    absolute_file_name('probcli_bak/',BakPath),
%    atom_concat(BakPath,Path,Cmd),
    absolute_file_name(Path,Cmd,[relative_to(probcli_bak)]),
    format(' Test ~w :: ~w ~w~n',[Id,Cmd,Arguments]),
    system_call(Cmd,Arguments,_OutputText,ErrText,Exit),
    statistics(walltime,[Stop,_]), WT is Stop-Start,
    format(' Walltime: ~w ms; ~w~n',[WT,Exit]),
    (Exit=exit(0), ErrText = [] -> Ok=true
      ;  format_colour([red,bold],'STD-ERROR (~w):~n~s~n',[Exit,ErrText]), Ok=false
    ).

% convert for system_call/process_create which do not accept numbers:
convert_arg_to_atom(Nr,Atom) :- number(Nr), number_codes(Nr,C),!,atom_codes(Atom,C).
convert_arg_to_atom(A,A).
