% (c) 2009-2020 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)


:- module(tools_strings,
                 [string_concatenate/3,
                  ajoin/2, ajoin_with_limit/3, ajoin_with_sep/3,
                  safe_name/2,
                  predicate_functor/3,
                  atom_codes_with_limit/2, atom_codes_with_limit/3,
                  truncate_atom/3,
                  convert_cli_arg/2,
                  convert_to_number/2,
                  convert_atom_to_number/2,
                  get_hex_codes/2,
                  get_hex_bytes/2
                  ]).

:- use_module(module_information).

:- module_info(group,infrastructure).
:- module_info(description,'A few utilities on strings/atoms seperated out from tools.pl to avoid cyclic module dependencies.').


:- mode string_concatenate(i,i,o).
string_concatenate(X,Y,XY) :- atom(X),atom(Y),!, atom_concat(X,Y,XY).
string_concatenate(X,Y,XY) :- atom(X),number(Y),!,
   % convert Y to atom, atom_concat is much faster than converting to codes list and back
   number_codes(Y,YC),
   atom_codes(YA,YC),atom_concat(X,YA,XY).
string_concatenate(X,Y,XY) :-
    safe_name(X,Xs),safe_name(Y,Ys),append(Xs,Ys,XYs),atom_codes(XY,XYs).

% tests are stored in tools.pl to avoid cyclic module dependencies
%:- assert_must_succeed((tools_strings: ajoin_with_sep([link,a,xa],'.',Text), Text == 'link.a.xa')).
%:- assert_must_succeed((tools_strings: ajoin_with_sep([link],'.',Text), Text == 'link')).

ajoin_with_sep(List,Sep,Text) :- ajoin_with_sep_aux(List,'',Sep,Text).

ajoin_with_sep_aux([],Txt,_Sep,Text) :- Text=Txt.
ajoin_with_sep_aux([H|T],Txt,Sep,Text) :-
   ( Txt == '' -> toAtom(H,TTxt)
   ; ajoin([Txt,Sep,H],TTxt)), ajoin_with_sep_aux(T,TTxt,Sep,Text).


/* Concats a list of atoms, but checks if numbers or compound terms are
   involved, which are converted to simple atoms */
ajoin(ListOfAtoms,Atom) :- ajoin_with_limit(ListOfAtoms,10000000,Atom).
% atom_codes with list of 100000000 takes very long, with 10000000 a few seconds

%safe_atom_concat(A,B,C,ExceptionOcc) :-
%    on_exception(error(representation_error(max_atom_length),_),atom_concat(A,B,C),
%       (print(exception(max_atom_length)),nl,A=C,ExceptionOcc=true)).

:- use_module(library(codesio),[write_to_codes/2]).

toAtom(Number,Atom) :- number(Number),!,number_chars(Number,C),atom_chars(Atom,C).
toAtom(Atom,Atom) :- atomic(Atom),!.
toAtom(Term,Atom) :- write_to_codes(Term,Codes),safe_atom_codes(Atom,Codes).

toCodes(Number,Codes) :- number(Number),!,number_codes(Number,Codes).
toCodes(Atom,Codes) :- atomic(Atom),!,atom_codes(Atom,Codes).
toCodes(Term,Codes) :- write_to_codes(Term,Codes).

% a copy of safe_atom_codes/2 from tools to avoid module dependency on error_manager
safe_atom_codes(V,C) :- var(V),var(C),!,
   print_error('Variables in call: '),print_error(safe_atom_codes(V,C)),
   C='$VARIABLE$'.
safe_atom_codes(A,C) :-
    on_exception(error(representation_error(max_atom_length),_),atom_codes(A,C),
       (print(exception(max_atom_length)),nl,atom_codes_with_limit(A,1000,C))).

% will concatenate until the Limit is reached or exceeded; it may produce atoms longer than Limit
% (if first atom already longer than limit + it adds ...
%:- assert_must_succeed((tools: ajoin_with_limit(['A','B','C','D'],100,Text), Text == 'ABCD')).
%:- assert_must_succeed((tools: ajoin_with_limit(['A','B','C','D'],2,Text), Text == 'AB...')).


ajoin_with_limit(Atoms,Limit,Result) :-
    ajoin_codes_with_limit(Atoms,Limit,Codes),
    safe_atom_codes(Atom,Codes), Result=Atom.

%:- assert_must_succeed((tools: ajoin_codes_with_limit(['A','B','C','D'],100,Text), Text == "ABCD")).
ajoin_codes_with_limit([],_,[]).
ajoin_codes_with_limit([Atom|TAtoms],Limit,Res) :-
    toCodes(Atom,AtomCodes),
    add_codes(AtomCodes,TAtoms,Limit,Res).

add_codes([],TAtoms,Limit,Res) :- !, ajoin_codes_with_limit(TAtoms,Limit,Res).
add_codes(_,_,Limit,Res) :- Limit < 1, !, Res = "...".
add_codes([H|T],TAtoms,Limit,[H|TR]) :- L1 is Limit-1,
   add_codes(T,TAtoms,L1,TR).



:- use_module(tools_printing,[print_error/1]).
safe_name(X,N) :- atom(X),!, atom_codes(X,N).
safe_name(X,N) :- number(X),!, number_codes(X,N).
safe_name(X,N) :- var(X),!, N="var".
safe_name(lambda_res(X),[114,101,115,95|N]) :- !, atom_codes(X,N).
safe_name(X,N) :- functor(X,F,_),atom_codes(F,N), print_error(non_atomic_in_safe_name(X)).


predicate_functor(X,F,N) :- var(X),!, print_error(var_in_predicate_functor),F='$VAR',N=0.
predicate_functor(_Module:Pred,F,N) :- !,predicate_functor(Pred,F,N).
predicate_functor(P,F,N) :- functor(P,F,N).



atom_codes_with_limit(A,C) :-
    on_exception(error(representation_error(max_atom_length),_),
                  atom_codes(A,C),
       (print(exception(max_atom_length)),nl,atom_codes_with_limit(A,1000,C))).


atom_codes_with_limit(A,Limit,Codes) :- var(A), Limit >= 0, !,
    truncate_codes(Codes,Limit,TCodes,_),
    atom_codes(A,TCodes).
atom_codes_with_limit(A,Limit,Codes) :- Limit < 1, !, atom_codes(A,Codes).
atom_codes_with_limit(A,Limit,Codes) :- atom_codes(A,Codes1),
    truncate_codes(Codes,Limit,Codes1,_).


truncate_codes([],_,[],false).
truncate_codes([H|T],Count,Res,Trunc) :-
   Count<1 -> Res = [46,46,46],Trunc=true /* '...' */
           ; Res = [H|TT], C1 is Count-1, truncate_codes(T,C1,TT,Trunc).

%:- assert_must_succeed((tools_strings:truncate_atom(abcd,100,Text), Text == 'abcd')).
%:- assert_must_succeed((tools_strings:truncate_atom(abcd,2,Text), Text == 'ab...')).
%:- assert_must_succeed((tools_strings:truncate_atom(abcd,0,Text), Text == '...')).
% TO DO: could be made more efficient by using something like sub_atom(Atom,0,Limit,_,NewAtom)
truncate_atom(Atom,Limit,NewAtom) :-
   atom_codes(Atom,Codes),
   truncate_codes(Codes,Limit,TCodes,Trunc),
   (Trunc=true -> atom_codes(NewAtom,TCodes) ; NewAtom=Atom).


convert_cli_arg(PrefVal,Value) :- compound(PrefVal),!,Value=PrefVal.
convert_cli_arg(Atom,Value) :-
    convert_atom_to_number(Atom,Nr),!,  /* convert '12' to 12 */
    Value=Nr.
convert_cli_arg(V,V).

convert_to_number(Nr,Res) :- number(Nr),!,Res=Nr.
convert_to_number(Atom,Nr) :- convert_atom_to_number(Atom,Nr).

convert_atom_to_number(Atom,Nr) :-
    atom(Atom), atom_codes(Atom,C),
    on_exception(error(syntax_error(_N),_),number_codes(Nr,C),
       fail). % in this case safe_number_codes fails ; we cannot convert the codes into a number



% ---------- hex utilities: to do move to tools ?


%:- assert_must_succeed(tools_strings: get_hex_codes(255,"ff").
get_hex_codes(0,Chars) :- !, Chars="0".
get_hex_codes(Nr,Chars) :- get_hex_codes(Nr,[],Chars).

get_hex_codes(0,Acc,R) :- !, R=Acc.
get_hex_codes(Nr,Acc,R) :-
    DigNr is Nr mod 16,
    get_hex_code(DigNr,Char),
    N1 is Nr // 16,
    get_hex_codes(N1,[Char|Acc],R).

%:- assert_must_succeed(tools_strings: get_hex_bytes([255,3],"ff03")).
% use for converting output of sha hash library:
get_hex_bytes([],[]).
get_hex_bytes([Byte|T],[C1,C2|HT]) :- get_hex_byte(Byte,C1,C2), get_hex_bytes(T,HT).

get_hex_byte(Byte,C1,C2) :-
   N1 is Byte // 16, N2 is Byte mod 16,
   get_hex_code(N1,C1), get_hex_code(N2,C2).
get_hex_code(Nr,Digit) :- Nr<10,!, Digit is Nr+48. % "0" = [48]
get_hex_code(Nr,Digit) :- Nr<16,!, Digit is Nr+87. % "A" = [65], "a" = [97]


