% (c) 2009-2020 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(xtl_interface, [open_xtl_file/1,
                           xtl_transition/3,
                           xtl_property/2,
                           xtl_animation_function_result/2, xtl_animation_image/2,
                           xtl_heuristic_function_active/0,
                           xtl_heuristic_function_result/2,
                           xtl_animation_image_click_transition/6,
                           xtl_animation_image_right_click_transition/3,
                           xtl_get_definition_string/2,

                           csp_initialisation_for_b/1,
                           csp_transition_for_b/5,
                           generate_b_operationargs_from_csp/2,

                           %open_promela_file/1,
                           %promela_transition/3,
                           %promela_property/2,

                           %open_smv_file/1, smv_transition/3, smv_property/2, % SMV mode broken

                           open_cspm_file/1, last_opened_cspm_file/1,
                           cspm_transition/3,
                           cspm_property/2,
                           set_cspm_main_process/1,
                           reset_xtl_interface/0]).


:- use_module(module_information).
:- module_info(group,animator).
:- module_info(description,'Provides an interface to the non-B animators depending on animation-mode.').

/* --------------- XTL ----------------- */
:- volatile prop/2, trans/3, start/1, animation_function_result/2, animation_image/2.
:- volatile animation_image_click_transition/6, animation_image_right_click_transition/3.
:- volatile heuristic_function_active/0, heuristic_function_result/2.
:- volatile prob_pragma_string/2.
:- dynamic prop/2.
:- dynamic trans/3.
:- dynamic start/1.
:- dynamic animation_function_result/2.
:- dynamic animation_image/2.
:- dynamic animation_image_click_transition/6, animation_image_right_click_transition/3.
:- dynamic heuristic_function_active/0.
:- dynamic heuristic_function_result/2.
:- dynamic prob_pragma_string/2.

% the following imports are required so that XTL .P files can make use of these functions:
:- use_module(library(lists)).
:- use_module(library(between)).
:- use_module(library(ordsets)).
:- use_module(library(samsort)).
:- use_module(library(random)).
:- use_module(library(avl)).
:- use_module(library(logarr)).
:- use_module(library(heaps)).

% ProB utilities (which can also be used by XTL code)
:- use_module(error_manager).
:- use_module(preferences,[get_preference/2]).
:- use_module(debug).
:- use_module(tools).

open_xtl_file(File) :-
    abolish([prop/2, trans/3, start/1], [force(true),tree(true)]),
    abolish([animation_image/2,animation_function_result/2,
             animation_image_click_transition/6,animation_image_right_click_transition/3,
             heuristic_function_active/0,
             prob_pragma_string/2],
         [force(true),tree(true)]),
    assert((heuristic_function_active :- fail)),
    assert((animation_image(_,_) :- fail)),
    assert((animation_function_result(_,_) :- fail)),
    assert((animation_image_click_transition(_,_,_,_,_,_) :- fail)),
    assert((animation_image_right_click_transition(_,_,_) :- fail)),
    assert((prob_pragma_string(_,_) :- fail)),
    debug_println(9,tcltk_open_xtl_file(File)),
    consult_without_redefine_warning(File),
    debug_println(9,new_xtl_file(File)).


xtl_transition(root,start_xtl_system,NewState) :- start(NewState).
xtl_transition(State,Operation,NewState) :- State \= root, trans(Operation,State,NewState).

xtl_property(State,Property) :- State \= root, prop(State,Property).

xtl_animation_function_result(State,AnimationMatrix) :- State \= root,
    animation_function_result(State,AnimationMatrix).

xtl_animation_image(Nr,PathToGif) :-
    %on_exception(error(existence_error(_,_),_),
    animation_image(Nr,PathToGif).

% return a transition template to execute for simple clicks (From=To) or drags
% OperationTemplate can either be the template of an operation to match or a list of such templates
% (the operations will then be executed in order)
xtl_animation_image_click_transition(FromX,FromY,ToX,ToY,OperationTemplate,Image) :-
    animation_image_click_transition(FromX,FromY,ToX,ToY,OperationTemplate,Image).

xtl_animation_image_right_click_transition(X,Y,OperationTemplate) :-
    animation_image_right_click_transition(X,Y,OperationTemplate).

xtl_heuristic_function_active :-
    heuristic_function_active.
xtl_heuristic_function_result(State,int(IntegerVal)) :- State \= root,
    heuristic_function_result(State,Res),
    (Res=int(R) -> IntegerVal=R
     ; number(Res) -> IntegerVal=Res
     ; add_error(xtl_heuristic_function_result,'heuristic_function_result must be integer: ',Res),fail
    ).

% way to mimic DEFINITION Strings in XTL mode, such as ASSERT_LTL
xtl_get_definition_string(Def_Name,DefString) :-
   prob_pragma_string(Def_Name,DefString).

consult_without_redefine_warning(File) :-
    prolog_flag(redefine_warnings, Old, off),
    prolog_flag(single_var_warnings, Old2, off),
    (on_exception(error(existence_error(_,_),_),
      my_compile(File),
      add_error_fail(xtl,'XTL File does not exist:',File))
      -> OK=true ; OK=false),
    prolog_flag(redefine_warnings, _, Old),
    prolog_flag(single_var_warnings, _, Old2),
    OK=true.

my_compile(F) :- %get_preference(user_is_an_expert_with_accessto_source_distribution,true),
                 !, % it seems it is ok to call compile also in probcli binary; it may do consult though
                 compile(F).
my_compile(F) :- consult(F).


/* --------------- Promela ----------------- */

%:- use_module('promela/h_int').

/* --------------- SMV ----------------- */

% :- use_module('smv/smv_trans').


/* --------------- CSP-M ----------------- */

:- use_module(probcspsrc(haskell_csp),[parse_and_load_cspm_file/1,
         cspm_trans_enum/3,
         animatable_process/1, animatable_process_without_arguments/1,
         get_symbol_span/2,force_evaluate_argument/2,normalise_cspm_state/2]).
:- use_module(probcspsrc(haskell_csp_analyzer),[cspPrintCompiled/2]).
:- use_module(probsrc(translate),[translate_cspm_state/2]).

:- dynamic last_opened_cspm_file/1. % useful for csp_and_b mode

open_cspm_file(File) :-
    retractall(last_opened_cspm_file(_)),
    debug_println(15,open_cspm_file(File)), flush_output(user_output),
    parse_and_load_cspm_file(File),
    assert(last_opened_cspm_file(File)).

:- dynamic cspm_main_process/1.
cspm_main_process('MAIN').
set_cspm_main_process(M) :-
  retractall(cspm_main_process(_)),
  assert(cspm_main_process(M)).

reset_xtl_interface :- retractall(last_opened_cspm_file(_)),
    reset_cspm_main_process.
reset_cspm_main_process :- set_cspm_main_process('MAIN').

:- use_module(eventhandling,[register_event_listener/3]).
:- register_event_listener(clear_specification,reset_xtl_interface,
                           'Reset XTL Interface.').

cspm_transition(root,start_cspm_MAIN,NormalisedNewState) :-
    cspm_main_process(MAIN),
    animatable_process_without_arguments(MAIN),
    get_start_expr(MAIN,NewState),
    normalise_cspm_state(NewState,NormalisedNewState).
cspm_transition(root,start_cspm(X),NormalisedNewState) :- cspm_main_process(MAIN),
    (get_preference(cspm_animate_all_processes_without_arguments,true)
     ; \+ animatable_process_without_arguments(MAIN)),
    animatable_process_without_arguments(X),
    X\=MAIN,
    get_start_expr(X,NewState),
    normalise_cspm_state(NewState,NormalisedNewState).
cspm_transition(root,start_cspm(X),NormalisedNewState) :- cspm_main_process(MAIN),
    get_preference(cspm_animate_all_processes,true),
    animatable_process(X),
    X\=MAIN,
    get_start_expr(X,NewState),
    normalise_cspm_state(NewState,NormalisedNewState).
cspm_transition(root,io([V1],print,no_loc_info_available),root) :-
    cspPrintCompiled(Expr,CompiledExpr), debug_println(9,cspPrintCompiled(Expr,CompiledExpr)),
    nl, translate:print_csp_value(Expr),
    print(' == '), nl, print('  '),
    force_evaluate_argument(CompiledExpr,V1),
    translate:print_csp_value(V1),nl.
cspm_transition(root,no_process_to_animate,root) :-
    ( get_preference(cspm_animate_all_processes,true) ->
        \+ animatable_process(_)
    ; \+ animatable_process_without_arguments(_)).
cspm_transition(State,Action,NormalisedNewState) :- State \= root,
   %print(comp),nl,
   cspm_trans_enum(State,Action,NewState),
   normalise_cspm_state(NewState,NormalisedNewState).
   %(ActionS = io(V,Ch,_Span) -> Action = io(V,Ch) ; Action=ActionS).
   %print(new(NewState)),nl. /* TO DO: Normalise */

cspm_property(State,Property) :-
   translate_cspm_state(State,Property).

/* --------------- CSP ----------------- */


get_start_expr(Proc,val_of(Proc,Span)) :- get_symbol_span(Proc,Span).



csp_initialisation_for_b(NewState) :-  cspm_main_process(MAIN),
  (animatable_process_without_arguments(MAIN) -> get_start_expr(MAIN,NewState);
    (animatable_process_without_arguments(X)
      ->  add_error(csp_transition_for_b,'No MAIN process in the CSP file! I am animating:',X),
          NewState = val_of(X)
       ;  add_error(csp_transition_for_b,'No animatable process in the CSP file!'), NewState = stop)
    ).

csp_transition_for_b(State,Ch,Args,Action,NewState) :- State \= root,
  % print(cspm_trans_enum(State,Action,NewState)),nl,
   cspm_trans_enum(State,Action,NewState),  %% TO DO: delay enumeration until B operation has been setup ?
  % print(cspm_trans_enum(Action,NewState)),nl,
   decompose_event(Action,Ch,Args).
  % print(b(Ch,BArgs)),nl.


   /* needed: an any operation: map any operation<------------- */

decompose_event(io(V,Ch,_Src),Ch,V).
decompose_event(tau(S),tau(S),[]).
%% decompose_event(i(S),i(S),[]). %% deprecated
decompose_event(tick(S),tick(S),[]).

generate_b_operationargs_from_csp(V,BArgs) :- l_copy_args_to_b(V,BArgs).


l_copy_args_to_b(tail_in(X),[Y]) :- translate_and_normalise_arg_to_b(X,Y).
l_copy_args_to_b([],[]).
l_copy_args_to_b([HCSP|T],[HB|TB]) :-
  copy_args_to_b(HCSP,HB),
  l_copy_args_to_b(T,TB).

copy_args_to_b(dot(X),Y) :- !,translate_and_normalise_arg_to_b(X,Y).  /* is this still required with the new eval ?? */
copy_args_to_b(in(X),Y) :- !,translate_and_normalise_arg_to_b(X,Y).
copy_args_to_b(out(X),Y) :- !,translate_and_normalise_arg_to_b(X,Y).
copy_args_to_b(X,Y) :- translate_and_normalise_arg_to_b(X,Y).

:- use_module(store,[normalise_value_for_var/4]).

translate_and_normalise_arg_to_b(CSP,BN) :- translate_arg_to_b(CSP,B), normalise_value_for_var(csp,true,B,BN).

:- use_module(tools,[print_message/1, convert_list_into_pairs/2]).
:- use_module(custom_explicit_sets,[construct_avl_from_lists/2]).

%translate_arg_to_b(X,Y) :- print(translate_arg_to_b(X,Y)),nl,fail.
translate_arg_to_b(X,X) :- var(X),!.
translate_arg_to_b(X,int(X)) :- number(X),!,print_message(converted_int(X)).
translate_arg_to_b(fd(N,S),fd(N,S)) :- !. /* copy B SET element across */
translate_arg_to_b(string(S),string(S)) :- !. /* copy B STRING element across */
translate_arg_to_b(int(N),int(N)) :- !.
translate_arg_to_b(true,pred_true /* bool_true */) :- !.
translate_arg_to_b(false,pred_false /* bool_false */) :- !.
translate_arg_to_b(global_set(N),global_set(N)) :- !.
translate_arg_to_b(freetype(N),freetype(N)) :- !.
translate_arg_to_b(avl_set(N),avl_set(N)) :- !.
translate_arg_to_b(closure(A,B,C),closure(A,B,C)) :- !.
translate_arg_to_b(closure(A,B,C,E),closure(A,B,C,E)) :- !.
translate_arg_to_b(setValue(S),R) :- !, translate_arg_to_b(S,R1),
   construct_avl_from_lists(R1,R).
   %sort(R1,R). % IS SORTING NECESSARY?; we could translate to AVL
translate_arg_to_b(list(L),R) :- !, translate_list_to_b(L,1,R1),
   custom_explicit_sets:construct_avl_from_lists(R1,R).
translate_arg_to_b([],[]) :- !.
translate_arg_to_b([H|T],[TH|TT]) :- !,translate_arg_to_b(H,TH), translate_arg_to_b(T,TT).
translate_arg_to_b((H,T),(TH,TT)) :- !,translate_arg_to_b(H,TH), translate_arg_to_b(T,TT).
translate_arg_to_b(na_tuple(L),Res) :- !,l_translate_arg_to_b(L,TL),
   convert_list_into_pairs(TL,Res).
translate_arg_to_b(Constant,BRep) :- translate_b_constant(Constant,BRep),!. /* clause necessary?? */
translate_arg_to_b(term(Constant),BRep) :- translate_b_constant(Constant,BRep),!.
% TO DO: treat floats/reals
translate_arg_to_b(term(N),term(N)) :- !.
translate_arg_to_b(DeferredSetEl,FD) :-
   is_deferred_set_element_name(DeferredSetEl,FD),!.
translate_arg_to_b(X,string(X)) :- atomic(X),!. % if the identfier X is not known: translate it to a string
% TO DO: some static checking: if no operation has a STRING parameter type, then we can skip this clause and generate an error message straightaway
translate_arg_to_b(X,term(X)) :- add_error(translate_arg_to_b,'Unknown CSP datatype, cannot convert to B:',X).
/* extend for other types */

translate_list_to_b([],_,[]).
translate_list_to_b([H|T],Nr,[(int(Nr),TH)|TT]) :- translate_arg_to_b(H,TH),
  N1 is Nr+1, translate_list_to_b(T,N1,TT).

l_translate_arg_to_b([],[]).
l_translate_arg_to_b([H|T],[TH|TT]) :- translate_arg_to_b(H,TH),
  l_translate_arg_to_b(T,TT).

:- use_module(tools,[safe_atom_codes/2]).
:- use_module(self_check).
:- assert_must_succeed( (xtl_interface:is_deferred_set_element_name('Code1',R),R=fd(1,'Code')) ).
:- assert_must_fail( xtl_interface:is_deferred_set_element_name('CodeXX',_R) ).
is_deferred_set_element_name(DeferredSetEl,fd(Nr,Set)) :- atomic(DeferredSetEl),
   b_global_sets:b_global_deferred_set(Set), atom_codes(Set,SetCodes),
   append(SetCodes,NrCodes,DC),
   safe_atom_codes(DeferredSetEl,DC),
   catch(number_codes(Nr,NrCodes),_,fail).

:- use_module(b_global_sets,[b_global_set/1, all_elements_of_type/2]).

translate_b_constant(GS,BRep) :- nonvar(GS),b_global_set(GS),all_elements_of_type(GS,BRep),!.
translate_b_constant(Constant,BRep) :- nonvar(Constant),b_global_sets:lookup_global_constant(Constant,BRep),!.
