% (c) 2016-2020 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(kernel_cardinality,[finite_int_cardinality/3]).

:- use_module(module_information,[module_info/2]).
:- module_info(group,kernel).
:- module_info(description,'This module provides operations for cardinality of sets.').

:- use_module(self_check).
:- use_module(preferences).
%:- use_module(library(avl),[avl_size/2]).
:- use_module(custom_explicit_sets,[explicit_set_cardinality_wf/3]).
:- use_module(kernel_waitflags,[add_wd_error/3]).
:- use_module(kernel_objects,[int_plus/3, empty_set_wf/2]).
:- use_module(library(clpfd)).
:- use_module(clpfd_interface,[clpfd_geq/3]).

% the co-routine block_finite_int_cardinality/3 is attached to free variables and re-used when possible
% contrary to earlier versions in kernel_objects we do *not* instantiate list skeletons to avoid introducing unnecessary variables und un-ordered enumeration
% re-usage is currently only checked once per top-level finite_int_cardinality call
% TO DO: either check continuously or catch unifications in equal_object or use attributed variables unification hook
%  card(N)=5 & card(M)=4 & card(O)>3 & (N=M or N=O) would not catch N=M inconsistency straight away

% TO DO: add upper-bound for card based upon type
% TO DO: maybe instantiate list skeletons if CLPFD=FALSE
% TO DO: add support for not_empty as well (difference with card(.)>0  ==> card will raise WD error
% TO DO: add version which also allows inf
% TO DO: maybe add domain/range as co-routines in a similar fashion to card
% TO DO: tests 1418, 1419; e.g., card(x)=10 & !y.(y:x & y<50 => y+1:x) & x<:1..50
% DONE: test 292: % {x,y|y : INTEGER & (y = card(x * INTEGER) & x <<: {TRUE})} = {{} |-> 0}
/*
x:seq({1}) & 55:dom(x) & card(x)=10  with CLPFD=FALSE
x:seq({1}) & 5:dom(x) & card(x)>6 with CLPFD=TRUE:
TO DO: instantiate list skeleton for last_finite_priority
** Tests run: 612, failed: 8 **
** Failed tests:
496 Many explicit computations which check that all operators work as expected on various concrete data.
1187 Ensure proper enumeration of sequences.
1272 ensure proper reification of card
1358 Check feasibility analysis works
1418 Test this is solved quickly using new ordered_value check for card.
1419 Test this is solved quickly using new ordered_value check for card & avoid bug in expand all quantifeir.
1540 Test Integer Z laws and that the division semantics of Z is preserved.
1562 Test no complicated set-unification with backtracking
1611   x:seq({1}) & 5:dom(x) & card(x)>6

1624 Test cardinality reification propagates through check_finite_card.
1625 Check card of interval propagates
1626 A few card tests; test reification involving records
1627 Check card propagates without need for t:1..20 constraint


*/


:- assert_must_succeed(( finite_int_cardinality(V,int(X),no_wf_available), X in 1..5, finite_int_cardinality(V,int(X2),no_wf_available), X2 in 5..10, X2==5)).
:- assert_must_succeed(( finite_int_cardinality(V,int(X),no_wf_available), X in 1..5, finite_int_cardinality(V,int(X2),no_wf_available), X2 in 5..10, V=[_A,_B,_C,_D|T], T==[] )).

finite_int_cardinality(V,int(Card),WF) :- finite_int_cardinality4(V,0,Card,WF).

finite_int_cardinality4(V,SoFar,Card,WF) :- var(V),!,
   int_plus(int(SoFar),int(VCard),int(Card)),
   attach_int_cardinality(V,VCard,WF).
finite_int_cardinality4([],Acc,Card,_) :- !, Card=Acc.
finite_int_cardinality4([_|T],Acc,N,WF) :- !, A1 is Acc+1, finite_int_cardinality4(T,A1,N,WF).
finite_int_cardinality4(CS,Acc,Card,WF) :-
    explicit_set_cardinality_wf(CS,CSCard,WF),  % TO DO: use finite version which itself raises WD error when inf and use clpfd addition
    add_cs_card(CS,CSCard,Acc,Card,WF),
    % TO DO: instantiate list skeleton if last_finite_priority encountered as this sometimes enables detection of inconsistencies; see test 1611
    (var(Card) -> clpfd_geq(Card,Acc,_) ; true).

:- block add_cs_card(?,-,?,?,?).
add_cs_card(_CS,CSCard,Acc,Res,_WF) :- number(CSCard),!,Res is Acc+CSCard.
add_cs_card(CS,_,_Acc,_Res,WF) :- add_wd_error('card applied to infinite (or very large) set: ',CS,WF).


attach_int_cardinality(Var,VCard,_WF) :-
    frozen(Var,Goals), % print(frozen(Var,Goals)),nl,
    get_finite_int_card(Goals,Var,GCard),!, % there is already co-routine attached
    print(reusing_coroutine(Var,GCard)),nl,
    VCard=GCard.
attach_int_cardinality(Var,VCard,WF) :-  geqzero(VCard),
    force_last_empty_set(Var,VCard,WF),
    block_finite_int_cardinality(Var,VCard,WF).

get_finite_int_card(kernel_cardinality:block_finite_int_cardinality(Var,VarCard,_),W,GCard) :- W==Var,!, GCard=VarCard.
get_finite_int_card((A,B),W,GCard) :- get_finite_int_card(A,W,GCard) -> true ; get_finite_int_card(B,W,GCard).

:- block block_finite_int_cardinality(-,?,?).
%block_finite_int_cardinality(S,C,_) :- print(bfis(S,C)),nl,fail.
block_finite_int_cardinality([],Card,_) :- !, Card=0.
block_finite_int_cardinality([_|T],N,WF) :- !,plus1(N1,N),block_finite_int_cardinality(T,N1,WF).
%block_finite_int_cardinality(avl_set(A),N,_) :- !,avl_size(A,N).
block_finite_int_cardinality(CS,Res,WF) :-
    explicit_set_cardinality_wf(CS,CSCard,WF), % TO DO: use finite version and unify Card and CSCard
    add_cs_card(CS,CSCard,0,Res,WF).

% co-routine to force empty set once cardinality known and set sufficiently instantiated
:- block force_last_empty_set(?,-,?).
force_last_empty_set(Set,Nr,WF) :- force_empty_set1(Set,Nr,WF).

force_empty_set1(Set,0,WF) :- !, (var(Set) -> empty_set_wf(Set,WF) ; true).
force_empty_set1(Set,Nr,WF) :- force_empty_set2(Set,Nr,WF).

:- block force_empty_set2(-,?,?).
force_empty_set2([_|T],Nr,WF) :- !, N1 is Nr-1, force_empty_set1(T,N1,WF).
force_empty_set2(_,_,_).

geqzero(N) :- (preference(use_clpfd_solver,false) -> true ; N #>=0).

% should probably be moved to clpfd_interface.pl
plus1(N,N1) :- number(N),!,N1 is N+1.
plus1(N,N1) :- preference(use_clpfd_solver,false),!, delay_plus1(N,N1).
plus1(N,N1) :- N1 #= N+1.

:- block delay_plus1(-,-).
delay_plus1(N,N1) :- var(N1),!, N1 is N+1.
delay_plus1(N,N1) :- N is N1-1.
